macro_rules! hashmap {
    ($( $key: expr => $val: expr ),*) => {{
         let mut map = ::std::collections::HashMap::new();
         $( map.insert($key, $val); )*
         map
    }}
}

pub fn esri_wkt_from_epsg(code: u16) -> String {
    let hm = hashmap![
        3819=>"GEOGCS[\"GCS_HD1909\",DATUM[\"D_Hungarian_Datum_1909\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",3819]]",
        3821=>"GEOGCS[\"GCS_TWD_1967\",DATUM[\"D_TWD_1967\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",3821]]",
        3824=>"GEOGCS[\"GCS_TWD_1997\",DATUM[\"D_TWD_1997\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",3824]]",
        3889=>"GEOGCS[\"GCS_IGRS\",DATUM[\"D_Iraqi_Geospatial_Reference_System\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",3889]]",
        3906=>"GEOGCS[\"GCS_MGI_1901\",DATUM[\"D_MGI_1901\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",3906]]",
        4001=>"GEOGCS[\"GCS_Airy_1830\",DATUM[\"D_Airy_1830\",SPHEROID[\"Airy_1830\",6377563.396,299.3249646]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4001]]",
        4002=>"GEOGCS[\"GCS_Airy_Modified\",DATUM[\"D_Airy_Modified\",SPHEROID[\"Airy_Modified\",6377340.189,299.3249646]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4002]]",
        4003=>"GEOGCS[\"GCS_Australian\",DATUM[\"D_Australian\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4003]]",
        4004=>"GEOGCS[\"GCS_Bessel_1841\",DATUM[\"D_Bessel_1841\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4004]]",
        4005=>"GEOGCS[\"GCS_Bessel_Modified\",DATUM[\"D_Bessel_Modified\",SPHEROID[\"Bessel_Modified\",6377492.018,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4005]]",
        4006=>"GEOGCS[\"GCS_Bessel_Namibia\",DATUM[\"D_Bessel_Namibia\",SPHEROID[\"Bessel_Namibia\",6377483.865280418,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4006]]",
        4007=>"GEOGCS[\"GCS_Clarke_1858\",DATUM[\"D_Clarke_1858\",SPHEROID[\"Clarke_1858\",6378293.645208759,294.260676369]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4007]]",
        4008=>"GEOGCS[\"GCS_Clarke_1866\",DATUM[\"D_Clarke_1866\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4008]]",
        4009=>"GEOGCS[\"GCS_Clarke_1866_Michigan\",DATUM[\"D_Clarke_1866_Michigan\",SPHEROID[\"Clarke_1866_Michigan\",6378450.047,294.978684677]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4009]]",
        4010=>"GEOGCS[\"GCS_Clarke_1880_Benoit\",DATUM[\"D_Clarke_1880_Benoit\",SPHEROID[\"Clarke_1880_Benoit\",6378300.789,293.4663155389802]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4010]]",
        4011=>"GEOGCS[\"GCS_Clarke_1880_IGN\",DATUM[\"D_Clarke_1880_IGN\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4011]]",
        4012=>"GEOGCS[\"GCS_Clarke_1880_RGS\",DATUM[\"D_Clarke_1880_RGS\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4012]]",
        4013=>"GEOGCS[\"GCS_Clarke_1880_Arc\",DATUM[\"D_Clarke_1880_Arc\",SPHEROID[\"Clarke_1880_Arc\",6378249.145,293.466307656]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4013]]",
        4014=>"GEOGCS[\"GCS_Clarke_1880_SGA\",DATUM[\"D_Clarke_1880_SGA\",SPHEROID[\"Clarke_1880_SGA\",6378249.2,293.46598]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4014]]",
        4015=>"GEOGCS[\"GCS_Everest_Adj_1937\",DATUM[\"D_Everest_Adj_1937\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4015]]",
        4016=>"GEOGCS[\"GCS_Everest_def_1967\",DATUM[\"D_Everest_Def_1967\",SPHEROID[\"Everest_Definition_1967\",6377298.556,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4016]]",
        4018=>"GEOGCS[\"GCS_Everest_Modified\",DATUM[\"D_Everest_Modified\",SPHEROID[\"Everest_1830_Modified\",6377304.063,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4018]]",
        4019=>"GEOGCS[\"GCS_GRS_1980\",DATUM[\"D_GRS_1980\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4019]]",
        4020=>"GEOGCS[\"GCS_Helmert_1906\",DATUM[\"D_Helmert_1906\",SPHEROID[\"Helmert_1906\",6378200.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4020]]",
        4021=>"GEOGCS[\"GCS_Indonesian\",DATUM[\"D_Indonesian\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4021]]",
        4022=>"GEOGCS[\"GCS_International_1924\",DATUM[\"D_International_1924\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4022]]",
        4023=>"GEOGCS[\"GCS_MOLDREF99\",DATUM[\"D_MOLDREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4023]]",
        4024=>"GEOGCS[\"GCS_Krasovsky_1940\",DATUM[\"D_Krasovsky_1940\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4024]]",
        4025=>"GEOGCS[\"GCS_NWL_9D\",DATUM[\"D_NWL_9D\",SPHEROID[\"NWL_9D\",6378145.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4025]]",
        4027=>"GEOGCS[\"GCS_Plessis_1817\",DATUM[\"D_Plessis_1817\",SPHEROID[\"Plessis_1817\",6376523.0,308.64]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4027]]",
        4028=>"GEOGCS[\"GCS_Struve_1860\",DATUM[\"D_Struve_1860\",SPHEROID[\"Struve_1860\",6378298.3,294.73]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4028]]",
        4029=>"GEOGCS[\"GCS_War_Office\",DATUM[\"D_War_Office\",SPHEROID[\"War_Office\",6378300.0,296.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4029]]",
        4031=>"GEOGCS[\"GCS_GEM_10C\",DATUM[\"D_GEM_10C\",SPHEROID[\"GEM_10C\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4031]]",
        4032=>"GEOGCS[\"GCS_OSU_86F\",DATUM[\"D_OSU_86F\",SPHEROID[\"OSU_86F\",6378136.2,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4032]]",
        4033=>"GEOGCS[\"GCS_OSU_91A\",DATUM[\"D_OSU_91A\",SPHEROID[\"OSU_91A\",6378136.3,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4033]]",
        4034=>"GEOGCS[\"GCS_Clarke_1880\",DATUM[\"D_Clarke_1880\",SPHEROID[\"Clarke_1880\",6378249.144808011,293.4663076556253]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4034]]",
        4035=>"GEOGCS[\"GCS_Sphere\",DATUM[\"D_Sphere\",SPHEROID[\"Sphere\",6371000.0,0.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4035]]",
        4036=>"GEOGCS[\"GCS_GRS_1967\",DATUM[\"D_GRS_1967\",SPHEROID[\"GRS_1967\",6378160.0,298.247167427]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4036]]",
        4042=>"GEOGCS[\"GCS_Everest_1830\",DATUM[\"D_Everest_1830\",SPHEROID[\"Everest_1830\",6377299.36,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4042]]",
        4044=>"GEOGCS[\"GCS_Everest_def_1962\",DATUM[\"D_Everest_Def_1962\",SPHEROID[\"Everest_Definition_1962\",6377301.243,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4044]]",
        4045=>"GEOGCS[\"GCS_Everest_def_1975\",DATUM[\"D_Everest_Def_1975\",SPHEROID[\"Everest_Definition_1975\",6377299.151,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4045]]",
        4046=>"GEOGCS[\"GCS_RGRDC_2005\",DATUM[\"D_Reseau_Geodesique_de_la_RDC_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4046]]",
        4047=>"GEOGCS[\"GCS_Sphere_GRS_1980_Authalic\",DATUM[\"D_Sphere_GRS_1980_Authalic\",SPHEROID[\"Sphere_GRS_1980_Authalic\",6371007.0,0.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4047]]",
        4052=>"GEOGCS[\"GCS_Sphere_Clarke_1866_Authalic\",DATUM[\"D_Sphere_Clarke_1866_Authalic\",SPHEROID[\"Sphere_Clarke_1866_Authalic\",6370997.0,0.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4052]]",
        4053=>"GEOGCS[\"GCS_Sphere_International_1924_Authalic\",DATUM[\"D_Sphere_International_1924_Authalic\",SPHEROID[\"Sphere_International_1924_Authalic\",6371228.0,0.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4053]]",
        4054=>"GEOGCS[\"GCS_Hughes_1980\",DATUM[\"D_Hughes_1980\",SPHEROID[\"Hughes_1980\",6378273.0,298.279411123064]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4054]]",
        4055=>"GEOGCS[\"GCS_WGS_1984_Major_Auxiliary_Sphere\",DATUM[\"D_WGS_1984_Major_Auxiliary_Sphere\",SPHEROID[\"WGS_1984_Major_Auxiliary_Sphere\",6378137.0,0.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4055]]",
        4075=>"GEOGCS[\"GCS_SREF98\",DATUM[\"D_Serbian_Reference_Network_1998\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4075]]",
        4081=>"GEOGCS[\"GCS_REGCAN95\",DATUM[\"D_Red_Geodesica_de_Canarias_1995\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4081]]",
        4120=>"GEOGCS[\"GCS_Greek\",DATUM[\"D_Greek\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4120]]",
        4121=>"GEOGCS[\"GCS_GGRS_1987\",DATUM[\"D_GGRS_1987\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4121]]",
        4122=>"GEOGCS[\"GCS_ATS_1977\",DATUM[\"D_ATS_1977\",SPHEROID[\"ATS_1977\",6378135.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4122]]",
        4123=>"GEOGCS[\"GCS_KKJ\",DATUM[\"D_KKJ\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4123]]",
        4124=>"GEOGCS[\"GCS_RT_1990\",DATUM[\"D_RT_1990\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4124]]",
        4125=>"GEOGCS[\"GCS_Samboja\",DATUM[\"D_Samboja\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4125]]",
        4126=>"GEOGCS[\"GCS_LKS_1994\",DATUM[\"D_Lithuania_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4669]]",
        4127=>"GEOGCS[\"GCS_Tete\",DATUM[\"D_Tete\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4127]]",
        4128=>"GEOGCS[\"GCS_Madzansua\",DATUM[\"D_Madzansua\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4128]]",
        4129=>"GEOGCS[\"GCS_Observatario\",DATUM[\"D_Observatario\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4129]]",
        4130=>"GEOGCS[\"GCS_Moznet\",DATUM[\"D_Moznet\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4130]]",
        4131=>"GEOGCS[\"GCS_Indian_1960\",DATUM[\"D_Indian_1960\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4131]]",
        4132=>"GEOGCS[\"GCS_FD_1958\",DATUM[\"D_FD_1958\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4132]]",
        4133=>"GEOGCS[\"GCS_Estonia_1992\",DATUM[\"D_Estonia_1992\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4133]]",
        4134=>"GEOGCS[\"GCS_PDO_1993\",DATUM[\"D_PDO_1993\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4134]]",
        4135=>"GEOGCS[\"GCS_Old_Hawaiian\",DATUM[\"D_Old_Hawaiian\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4135]]",
        4136=>"GEOGCS[\"GCS_St_Lawrence_Island\",DATUM[\"D_St_Lawrence_Island\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4136]]",
        4137=>"GEOGCS[\"GCS_St_Paul_Island\",DATUM[\"D_St_Paul_Island\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4137]]",
        4138=>"GEOGCS[\"GCS_St_George_Island\",DATUM[\"D_St_George_Island\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4138]]",
        4139=>"GEOGCS[\"GCS_Puerto_Rico\",DATUM[\"D_Puerto_Rico\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4139]]",
        4140=>"GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4617]]",
        4141=>"GEOGCS[\"GCS_Israel\",DATUM[\"D_Israel\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4141]]",
        4142=>"GEOGCS[\"GCS_Locodjo_1965\",DATUM[\"D_Locodjo_1965\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4142]]",
        4143=>"GEOGCS[\"GCS_Abidjan_1987\",DATUM[\"D_Abidjan_1987\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4143]]",
        4144=>"GEOGCS[\"GCS_Kalianpur_1937\",DATUM[\"D_Kalianpur_1937\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4144]]",
        4145=>"GEOGCS[\"GCS_Kalianpur_1962\",DATUM[\"D_Kalianpur_1962\",SPHEROID[\"Everest_Definition_1962\",6377301.243,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4145]]",
        4146=>"GEOGCS[\"GCS_Kalianpur_1975\",DATUM[\"D_Kalianpur_1975\",SPHEROID[\"Everest_Definition_1975\",6377299.151,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4146]]",
        4147=>"GEOGCS[\"GCS_Hanoi_1972\",DATUM[\"D_Hanoi_1972\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4147]]",
        4148=>"GEOGCS[\"GCS_Hartebeesthoek_1994\",DATUM[\"D_Hartebeesthoek_1994\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4148]]",
        4149=>"GEOGCS[\"GCS_CH1903\",DATUM[\"D_CH1903\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4149]]",
        4150=>"GEOGCS[\"GCS_CH1903+\",DATUM[\"D_CH1903+\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4150]]",
        4151=>"GEOGCS[\"GCS_Swiss_TRF_1995\",DATUM[\"D_Swiss_TRF_1995\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4151]]",
        4152=>"GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4152]]",
        4153=>"GEOGCS[\"GCS_Rassadiran\",DATUM[\"D_Rassadiran\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4153]]",
        4154=>"GEOGCS[\"GCS_European_1950_ED77\",DATUM[\"D_European_1950_ED77\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4154]]",
        4155=>"GEOGCS[\"GCS_Dabola_1981\",DATUM[\"D_Dabola_1981\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4155]]",
        4156=>"GEOGCS[\"GCS_S_JTSK\",DATUM[\"D_S_JTSK\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4156]]",
        4157=>"GEOGCS[\"GCS_Mount_Dillon\",DATUM[\"D_Mount_Dillon\",SPHEROID[\"Clarke_1858\",6378293.645208759,294.260676369]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4157]]",
        4158=>"GEOGCS[\"GCS_Naparima_1955\",DATUM[\"D_Naparima_1955\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4158]]",
        4159=>"GEOGCS[\"GCS_European_Libyan_Datum_1979\",DATUM[\"D_European_Libyan_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4159]]",
        4160=>"GEOGCS[\"GCS_Chos_Malal_1914\",DATUM[\"D_Chos_Malal_1914\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4160]]",
        4161=>"GEOGCS[\"GCS_Pampa_del_Castillo\",DATUM[\"D_Pampa_del_Castillo\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4161]]",
        4162=>"GEOGCS[\"GCS_Korean_Datum_1985\",DATUM[\"D_Korean_Datum_1985\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4162]]",
        4163=>"GEOGCS[\"GCS_Yemen_NGN_1996\",DATUM[\"D_Yemen_NGN_1996\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4163]]",
        4164=>"GEOGCS[\"GCS_South_Yemen\",DATUM[\"D_South_Yemen\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4164]]",
        4165=>"GEOGCS[\"GCS_Bissau\",DATUM[\"D_Bissau\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4165]]",
        4166=>"GEOGCS[\"GCS_Korean_Datum_1995\",DATUM[\"D_Korean_Datum_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4166]]",
        4167=>"GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4167]]",
        4168=>"GEOGCS[\"GCS_Accra\",DATUM[\"D_Accra\",SPHEROID[\"War_Office\",6378300.0,296.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4168]]",
        4169=>"GEOGCS[\"GCS_American_Samoa_1962\",DATUM[\"D_American_Samoa_1962\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4169]]",
        4170=>"GEOGCS[\"GCS_SIRGAS\",DATUM[\"D_SIRGAS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4170]]",
        4171=>"GEOGCS[\"GCS_RGF_1993\",DATUM[\"D_RGF_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4171]]",
        4172=>"GEOGCS[\"GCS_POSGAR\",DATUM[\"D_POSGAR\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4172]]",
        4173=>"GEOGCS[\"GCS_IRENET95\",DATUM[\"D_IRENET95\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4173]]",
        4174=>"GEOGCS[\"GCS_Sierra_Leone_1924\",DATUM[\"D_Sierra_Leone_1924\",SPHEROID[\"War_Office\",6378300.0,296.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4174]]",
        4175=>"GEOGCS[\"GCS_Sierra_Leone_1968\",DATUM[\"D_Sierra_Leone_1968\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4175]]",
        4176=>"GEOGCS[\"GCS_Australian_Antarctic_1998\",DATUM[\"D_Australian_Antarctic_1998\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4176]]",
        4178=>"GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4178]]",
        4179=>"GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4179]]",
        4180=>"GEOGCS[\"GCS_Estonia_1997\",DATUM[\"D_Estonia_1997\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4180]]",
        4181=>"GEOGCS[\"GCS_Luxembourg_1930\",DATUM[\"D_Luxembourg_1930\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4181]]",
        4182=>"GEOGCS[\"GCS_Azores_Occidental_1939\",DATUM[\"D_Azores_Occidental_Islands_1939\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4182]]",
        4183=>"GEOGCS[\"GCS_Azores_Central_1948\",DATUM[\"D_Azores_Central_Islands_1948\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4183]]",
        4184=>"GEOGCS[\"GCS_Azores_Oriental_1940\",DATUM[\"D_Azores_Oriental_Islands_1940\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4184]]",
        4185=>"GEOGCS[\"GCS_Madeira_1936\",DATUM[\"D_Madeira_1936\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4185]]",
        4188=>"GEOGCS[\"GCS_OSNI_1952\",DATUM[\"D_OSNI_1952\",SPHEROID[\"Airy_1830\",6377563.396,299.3249646]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4188]]",
        4189=>"GEOGCS[\"GCS_REGVEN\",DATUM[\"D_REGVEN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4189]]",
        4190=>"GEOGCS[\"GCS_POSGAR_1998\",DATUM[\"D_POSGAR_1998\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4190]]",
        4191=>"GEOGCS[\"GCS_Albanian_1987\",DATUM[\"D_Albanian_1987\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4191]]",
        4192=>"GEOGCS[\"GCS_Douala_1948\",DATUM[\"D_Douala_1948\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4192]]",
        4193=>"GEOGCS[\"GCS_Manoca_1962\",DATUM[\"D_Manoca_1962\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4193]]",
        4194=>"GEOGCS[\"GCS_Qornoq_1927\",DATUM[\"D_Qornoq_1927\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4194]]",
        4195=>"GEOGCS[\"GCS_Scoresbysund_1952\",DATUM[\"D_Scoresbysund_1952\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4195]]",
        4196=>"GEOGCS[\"GCS_Ammassalik_1958\",DATUM[\"D_Ammassalik_1958\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4196]]",
        4197=>"GEOGCS[\"GCS_Garoua\",DATUM[\"D_Garoua\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4197]]",
        4198=>"GEOGCS[\"GCS_Kousseri\",DATUM[\"D_Kousseri\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4198]]",
        4199=>"GEOGCS[\"GCS_Egypt_1930\",DATUM[\"D_Egypt_1930\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4199]]",
        4200=>"GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4200]]",
        4201=>"GEOGCS[\"GCS_Adindan\",DATUM[\"D_Adindan\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4201]]",
        4202=>"GEOGCS[\"GCS_Australian_1966\",DATUM[\"D_Australian_1966\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4202]]",
        4203=>"GEOGCS[\"GCS_Australian_1984\",DATUM[\"D_Australian_1984\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4203]]",
        4204=>"GEOGCS[\"GCS_Ain_el_Abd_1970\",DATUM[\"D_Ain_el_Abd_1970\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4204]]",
        4205=>"GEOGCS[\"GCS_Afgooye\",DATUM[\"D_Afgooye\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4205]]",
        4206=>"GEOGCS[\"GCS_Agadez\",DATUM[\"D_Agadez\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4206]]",
        4207=>"GEOGCS[\"GCS_Lisbon\",DATUM[\"D_Lisbon\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4207]]",
        4208=>"GEOGCS[\"GCS_Aratu\",DATUM[\"D_Aratu\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4208]]",
        4209=>"GEOGCS[\"GCS_Arc_1950\",DATUM[\"D_Arc_1950\",SPHEROID[\"Clarke_1880_Arc\",6378249.145,293.466307656]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4209]]",
        4210=>"GEOGCS[\"GCS_Arc_1960\",DATUM[\"D_Arc_1960\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4210]]",
        4211=>"GEOGCS[\"GCS_Batavia\",DATUM[\"D_Batavia\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4211]]",
        4212=>"GEOGCS[\"GCS_Barbados_1938\",DATUM[\"D_Barbados_1938\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4212]]",
        4213=>"GEOGCS[\"GCS_Beduaram\",DATUM[\"D_Beduaram\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4213]]",
        4214=>"GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4214]]",
        4215=>"GEOGCS[\"GCS_Belge_1950\",DATUM[\"D_Belge_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4215]]",
        4216=>"GEOGCS[\"GCS_Bermuda_1957\",DATUM[\"D_Bermuda_1957\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4216]]",
        4218=>"GEOGCS[\"GCS_Bogota\",DATUM[\"D_Bogota\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4218]]",
        4219=>"GEOGCS[\"GCS_Bukit_Rimpah\",DATUM[\"D_Bukit_Rimpah\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4219]]",
        4220=>"GEOGCS[\"GCS_Camacupa\",DATUM[\"D_Camacupa\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4220]]",
        4221=>"GEOGCS[\"GCS_Campo_Inchauspe\",DATUM[\"D_Campo_Inchauspe\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4221]]",
        4222=>"GEOGCS[\"GCS_Cape\",DATUM[\"D_Cape\",SPHEROID[\"Clarke_1880_Arc\",6378249.145,293.466307656]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4222]]",
        4223=>"GEOGCS[\"GCS_Carthage\",DATUM[\"D_Carthage\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4223]]",
        4224=>"GEOGCS[\"GCS_Chua\",DATUM[\"D_Chua\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4224]]",
        4225=>"GEOGCS[\"GCS_Corrego_Alegre\",DATUM[\"D_Corrego_Alegre\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4225]]",
        4226=>"GEOGCS[\"GCS_Cote_d_Ivoire\",DATUM[\"D_Cote_d_Ivoire\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4226]]",
        4227=>"GEOGCS[\"GCS_Deir_ez_Zor\",DATUM[\"D_Deir_ez_Zor\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4227]]",
        4228=>"GEOGCS[\"GCS_Douala\",DATUM[\"D_Douala\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4228]]",
        4229=>"GEOGCS[\"GCS_Egypt_1907\",DATUM[\"D_Egypt_1907\",SPHEROID[\"Helmert_1906\",6378200.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4229]]",
        4230=>"GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4230]]",
        4231=>"GEOGCS[\"GCS_European_1987\",DATUM[\"D_European_1987\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4231]]",
        4232=>"GEOGCS[\"GCS_Fahud\",DATUM[\"D_Fahud\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4232]]",
        4233=>"GEOGCS[\"GCS_Gandajika_1970\",DATUM[\"D_Gandajika_1970\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4233]]",
        4234=>"GEOGCS[\"GCS_Garoua\",DATUM[\"D_Garoua\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4197]]",
        4235=>"GEOGCS[\"GCS_Guyane_Francaise\",DATUM[\"D_Guyane_Francaise\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4235]]",
        4236=>"GEOGCS[\"GCS_Hu_Tzu_Shan\",DATUM[\"D_Hu_Tzu_Shan\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4236]]",
        4237=>"GEOGCS[\"GCS_Hungarian_1972\",DATUM[\"D_Hungarian_1972\",SPHEROID[\"GRS_1967\",6378160.0,298.247167427]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4237]]",
        4238=>"GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4238]]",
        4239=>"GEOGCS[\"GCS_Indian_1954\",DATUM[\"D_Indian_1954\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4239]]",
        4240=>"GEOGCS[\"GCS_Indian_1975\",DATUM[\"D_Indian_1975\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4240]]",
        4241=>"GEOGCS[\"GCS_Jamaica_1875\",DATUM[\"D_Jamaica_1875\",SPHEROID[\"Clarke_1880\",6378249.144808011,293.4663076556253]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4241]]",
        4242=>"GEOGCS[\"GCS_Jamaica_1969\",DATUM[\"D_Jamaica_1969\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4242]]",
        4243=>"GEOGCS[\"GCS_Kalianpur_1880\",DATUM[\"D_Kalianpur_1880\",SPHEROID[\"Everest_1830\",6377299.36,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4243]]",
        4244=>"GEOGCS[\"GCS_Kandawala\",DATUM[\"D_Kandawala\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4244]]",
        4245=>"GEOGCS[\"GCS_Kertau\",DATUM[\"D_Kertau\",SPHEROID[\"Everest_1830_Modified\",6377304.063,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4245]]",
        4246=>"GEOGCS[\"GCS_Kuwait_Oil_Company\",DATUM[\"D_Kuwait_Oil_Company\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4246]]",
        4247=>"GEOGCS[\"GCS_La_Canoa\",DATUM[\"D_La_Canoa\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4247]]",
        4248=>"GEOGCS[\"GCS_Provisional_S_American_1956\",DATUM[\"D_Provisional_S_American_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4248]]",
        4249=>"GEOGCS[\"GCS_Lake\",DATUM[\"D_Lake\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4249]]",
        4250=>"GEOGCS[\"GCS_Leigon\",DATUM[\"D_Leigon\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4250]]",
        4251=>"GEOGCS[\"GCS_Liberia_1964\",DATUM[\"D_Liberia_1964\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4251]]",
        4252=>"GEOGCS[\"GCS_Lome\",DATUM[\"D_Lome\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4252]]",
        4253=>"GEOGCS[\"GCS_Luzon_1911\",DATUM[\"D_Luzon_1911\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4253]]",
        4254=>"GEOGCS[\"GCS_Hito_XVIII_1963\",DATUM[\"D_Hito_XVIII_1963\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4254]]",
        4255=>"GEOGCS[\"GCS_Herat_North\",DATUM[\"D_Herat_North\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4255]]",
        4256=>"GEOGCS[\"GCS_Mahe_1971\",DATUM[\"D_Mahe_1971\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4256]]",
        4257=>"GEOGCS[\"GCS_Makassar\",DATUM[\"D_Makassar\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4257]]",
        4258=>"GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4258]]",
        4259=>"GEOGCS[\"GCS_Malongo_1987\",DATUM[\"D_Malongo_1987\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4259]]",
        4260=>"GEOGCS[\"GCS_Manoca\",DATUM[\"D_Manoca\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4260]]",
        4261=>"GEOGCS[\"GCS_Merchich\",DATUM[\"D_Merchich\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Grad\",0.01570796326794897],AUTHORITY[\"EPSG\",4261]]",
        4262=>"GEOGCS[\"GCS_Massawa\",DATUM[\"D_Massawa\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4262]]",
        4263=>"GEOGCS[\"GCS_Minna\",DATUM[\"D_Minna\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4263]]",
        4264=>"GEOGCS[\"GCS_Mhast\",DATUM[\"D_Mhast\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4264]]",
        4265=>"GEOGCS[\"GCS_Monte_Mario\",DATUM[\"D_Monte_Mario\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4265]]",
        4266=>"GEOGCS[\"GCS_Mporaloko\",DATUM[\"D_Mporaloko\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4266]]",
        4267=>"GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4267]]",
        4268=>"GEOGCS[\"GCS_North_American_Michigan\",DATUM[\"D_North_American_Michigan\",SPHEROID[\"Clarke_1866_Michigan\",6378450.047,294.978684677]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4268]]",
        4269=>"GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4269]]",
        4270=>"GEOGCS[\"GCS_Nahrwan_1967\",DATUM[\"D_Nahrwan_1967\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4270]]",
        4271=>"GEOGCS[\"GCS_Naparima_1972\",DATUM[\"D_Naparima_1972\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4271]]",
        4272=>"GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4272]]",
        4273=>"GEOGCS[\"GCS_NGO_1948\",DATUM[\"D_NGO_1948\",SPHEROID[\"Bessel_Modified\",6377492.018,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4273]]",
        4274=>"GEOGCS[\"GCS_Datum_73\",DATUM[\"D_Datum_73\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4274]]",
        4275=>"GEOGCS[\"GCS_NTF\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4275]]",
        4276=>"GEOGCS[\"GCS_NSWC_9Z_2\",DATUM[\"D_NSWC_9Z_2\",SPHEROID[\"NWL_9D\",6378145.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4276]]",
        4277=>"GEOGCS[\"GCS_OSGB_1936\",DATUM[\"D_OSGB_1936\",SPHEROID[\"Airy_1830\",6377563.396,299.3249646]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4277]]",
        4278=>"GEOGCS[\"GCS_OSGB_1970_SN\",DATUM[\"D_OSGB_1970_SN\",SPHEROID[\"Airy_1830\",6377563.396,299.3249646]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4278]]",
        4279=>"GEOGCS[\"GCS_OS_SN_1980\",DATUM[\"D_OS_SN_1980\",SPHEROID[\"Airy_1830\",6377563.396,299.3249646]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4279]]",
        4280=>"GEOGCS[\"GCS_Padang_1884\",DATUM[\"D_Padang_1884\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4280]]",
        4281=>"GEOGCS[\"GCS_Palestine_1923\",DATUM[\"D_Palestine_1923\",SPHEROID[\"Clarke_1880_Benoit\",6378300.789,293.4663155389802]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4281]]",
        4282=>"GEOGCS[\"GCS_Pointe_Noire\",DATUM[\"D_Pointe_Noire\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4282]]",
        4283=>"GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4283]]",
        4284=>"GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4284]]",
        4285=>"GEOGCS[\"GCS_Qatar_1974\",DATUM[\"D_Qatar\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4285]]",
        4286=>"GEOGCS[\"GCS_Qatar_1948\",DATUM[\"D_Qatar_1948\",SPHEROID[\"Helmert_1906\",6378200.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4286]]",
        4287=>"GEOGCS[\"GCS_Qornoq\",DATUM[\"D_Qornoq\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4287]]",
        4288=>"GEOGCS[\"GCS_Loma_Quintana\",DATUM[\"D_Loma_Quintana\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4288]]",
        4289=>"GEOGCS[\"GCS_Amersfoort\",DATUM[\"D_Amersfoort\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4289]]",
        4291=>"GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4618]]",
        4292=>"GEOGCS[\"GCS_Sapper_Hill_1943\",DATUM[\"D_Sapper_Hill_1943\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4292]]",
        4293=>"GEOGCS[\"GCS_Schwarzeck\",DATUM[\"D_Schwarzeck\",SPHEROID[\"Bessel_Namibia\",6377483.865280418,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4293]]",
        4294=>"GEOGCS[\"GCS_Segora\",DATUM[\"D_Segora\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4294]]",
        4295=>"GEOGCS[\"GCS_Serindung\",DATUM[\"D_Serindung\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4295]]",
        4296=>"GEOGCS[\"GCS_Sudan\",DATUM[\"D_Sudan\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4296]]",
        4297=>"GEOGCS[\"GCS_Tananarive_1925\",DATUM[\"D_Tananarive_1925\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4297]]",
        4298=>"GEOGCS[\"GCS_Timbalai_1948\",DATUM[\"D_Timbalai_1948\",SPHEROID[\"Everest_Definition_1967\",6377298.556,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4298]]",
        4299=>"GEOGCS[\"GCS_TM65\",DATUM[\"D_TM65\",SPHEROID[\"Airy_Modified\",6377340.189,299.3249646]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4299]]",
        4300=>"GEOGCS[\"GCS_TM75\",DATUM[\"D_TM75\",SPHEROID[\"Airy_Modified\",6377340.189,299.3249646]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4300]]",
        4301=>"GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4301]]",
        4302=>"GEOGCS[\"GCS_Trinidad_1903\",DATUM[\"D_Trinidad_1903\",SPHEROID[\"Clarke_1858\",6378293.645208759,294.260676369]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4302]]",
        4303=>"GEOGCS[\"GCS_Trucial_Coast_1948\",DATUM[\"D_Trucial_Coast_1948\",SPHEROID[\"Helmert_1906\",6378200.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4303]]",
        4304=>"GEOGCS[\"GCS_Voirol_1875\",DATUM[\"D_Voirol_1875\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4304]]",
        4306=>"GEOGCS[\"GCS_Bern_1938\",DATUM[\"D_Bern_1938\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4306]]",
        4307=>"GEOGCS[\"GCS_Nord_Sahara_1959\",DATUM[\"D_Nord_Sahara_1959\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4307]]",
        4308=>"GEOGCS[\"GCS_RT38\",DATUM[\"D_Stockholm_1938\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4308]]",
        4309=>"GEOGCS[\"GCS_Yacare\",DATUM[\"D_Yacare\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4309]]",
        4310=>"GEOGCS[\"GCS_Yoff\",DATUM[\"D_Yoff\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4310]]",
        4311=>"GEOGCS[\"GCS_Zanderij\",DATUM[\"D_Zanderij\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4311]]",
        4312=>"GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4312]]",
        4313=>"GEOGCS[\"GCS_Belge_1972\",DATUM[\"D_Belge_1972\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4313]]",
        4314=>"GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4314]]",
        4315=>"GEOGCS[\"GCS_Conakry_1905\",DATUM[\"D_Conakry_1905\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4315]]",
        4316=>"GEOGCS[\"GCS_Dealul_Piscului_1933\",DATUM[\"D_Dealul_Piscului_1933\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4316]]",
        4317=>"GEOGCS[\"GCS_Dealul_Piscului_1970\",DATUM[\"D_Dealul_Piscului_1970\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4317]]",
        4318=>"GEOGCS[\"GCS_NGN\",DATUM[\"D_NGN\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4318]]",
        4319=>"GEOGCS[\"GCS_KUDAMS\",DATUM[\"D_Kuwait_Utility\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4319]]",
        4322=>"GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4322]]",
        4324=>"GEOGCS[\"GCS_WGS_1972_BE\",DATUM[\"D_WGS_1972_BE\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4324]]",
        4326=>"GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4326]]",
        4463=>"GEOGCS[\"GCS_RGSPM_2006\",DATUM[\"D_Reseau_Geodesique_de_St_Pierre_et_Miquelon_2006\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4463]]",
        4470=>"GEOGCS[\"GCS_RGM_2004\",DATUM[\"D_Reseau_Geodesique_de_Mayotte_2004\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4470]]",
        4475=>"GEOGCS[\"GCS_Cadastre_1997\",DATUM[\"D_Cadastre_1997\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4475]]",
        4483=>"GEOGCS[\"GCS_Mexican_Datum_of_1993\",DATUM[\"D_Mexican_Datum_of_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4483]]",
        4490=>"GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4490]]",
        4555=>"GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4555]]",
        4558=>"GEOGCS[\"GCS_RRAF_1991\",DATUM[\"D_RRAF_1991\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4558]]",
        4600=>"GEOGCS[\"GCS_Anguilla_1957\",DATUM[\"D_Anguilla_1957\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4600]]",
        4601=>"GEOGCS[\"GCS_Antigua_1943\",DATUM[\"D_Antigua_1943\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4601]]",
        4602=>"GEOGCS[\"GCS_Dominica_1945\",DATUM[\"D_Dominica_1945\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4602]]",
        4603=>"GEOGCS[\"GCS_Grenada_1953\",DATUM[\"D_Grenada_1953\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4603]]",
        4604=>"GEOGCS[\"GCS_Montserrat_1958\",DATUM[\"D_Montserrat_1958\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4604]]",
        4605=>"GEOGCS[\"GCS_St_Kitts_1955\",DATUM[\"D_St_Kitts_1955\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4605]]",
        4606=>"GEOGCS[\"GCS_St_Lucia_1955\",DATUM[\"D_St_Lucia_1955\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4606]]",
        4607=>"GEOGCS[\"GCS_St_Vincent_1945\",DATUM[\"D_St_Vincent_1945\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4607]]",
        4608=>"GEOGCS[\"GCS_NAD_1927_Definition_1976\",DATUM[\"D_NAD_1927_Definition_1976\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4608]]",
        4609=>"GEOGCS[\"GCS_NAD_1927_CGQ77\",DATUM[\"D_NAD_1927_CGQ77\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4609]]",
        4610=>"GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4610]]",
        4611=>"GEOGCS[\"GCS_Hong_Kong_1980\",DATUM[\"D_Hong_Kong_1980\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4611]]",
        4612=>"GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4612]]",
        4613=>"GEOGCS[\"GCS_Gunung_Segara\",DATUM[\"D_Gunung_Segara\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4613]]",
        4614=>"GEOGCS[\"GCS_QND_1995\",DATUM[\"D_QND_1995\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4614]]",
        4615=>"GEOGCS[\"GCS_Porto_Santo_1936\",DATUM[\"D_Porto_Santo_1936\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4615]]",
        4616=>"GEOGCS[\"GCS_Selvagem_Grande_1938\",DATUM[\"D_Selvagem_Grande_1938\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4616]]",
        4617=>"GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4617]]",
        4618=>"GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4618]]",
        4619=>"GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4619]]",
        4620=>"GEOGCS[\"GCS_Point_58\",DATUM[\"D_Point_58\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4620]]",
        4621=>"GEOGCS[\"GCS_Fort_Marigot\",DATUM[\"D_Fort_Marigot\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4621]]",
        4622=>"GEOGCS[\"GCS_Sainte_Anne\",DATUM[\"D_Sainte_Anne\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4622]]",
        4623=>"GEOGCS[\"GCS_CSG_1967\",DATUM[\"D_CSG_1967\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4623]]",
        4624=>"GEOGCS[\"GCS_RGFG_1995\",DATUM[\"D_RGFG_1995\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4624]]",
        4625=>"GEOGCS[\"GCS_Fort_Desaix\",DATUM[\"D_Fort_Desaix\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4625]]",
        4626=>"GEOGCS[\"GCS_Reunion_1947\",DATUM[\"D_Reunion_1947\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4626]]",
        4627=>"GEOGCS[\"GCS_RGR_1992\",DATUM[\"D_RGR_1992\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4627]]",
        4628=>"GEOGCS[\"GCS_Tahiti_1952\",DATUM[\"D_Tahiti_1952\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4628]]",
        4629=>"GEOGCS[\"GCS_Tahaa_1954\",DATUM[\"D_Tahaa_1954\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4629]]",
        4630=>"GEOGCS[\"GCS_IGN72_Nuku_Hiva\",DATUM[\"D_IGN72_Nuku_Hiva\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4630]]",
        4631=>"GEOGCS[\"GCS_Kerguelen_Island_1949\",DATUM[\"D_Kerguelen_Island_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4698]]",
        4632=>"GEOGCS[\"GCS_Combani_1950\",DATUM[\"D_Combani_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4632]]",
        4633=>"GEOGCS[\"GCS_IGN56_Lifou\",DATUM[\"D_IGN56_Lifou\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4633]]",
        4636=>"GEOGCS[\"GCS_Petrels_1972\",DATUM[\"D_Petrels_1972\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4636]]",
        4637=>"GEOGCS[\"GCS_Pointe_Geologie_Perroud_1950\",DATUM[\"D_Pointe_Geologie_Perroud_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4637]]",
        4638=>"GEOGCS[\"GCS_Saint_Pierre_et_Miquelon_1950\",DATUM[\"D_Saint_Pierre_et_Miquelon_1950\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4638]]",
        4639=>"GEOGCS[\"GCS_MOP78\",DATUM[\"D_MOP78\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4639]]",
        4641=>"GEOGCS[\"GCS_IGN53_Mare\",DATUM[\"D_IGN53_Mare\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4641]]",
        4642=>"GEOGCS[\"GCS_ST84_Ile_des_Pins\",DATUM[\"D_ST84_Ile_des_Pins\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4642]]",
        4643=>"GEOGCS[\"GCS_ST71_Belep\",DATUM[\"D_ST71_Belep\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4643]]",
        4644=>"GEOGCS[\"GCS_NEA74_Noumea\",DATUM[\"D_NEA74_Noumea\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4644]]",
        4645=>"GEOGCS[\"GCS_RGNC_1991\",DATUM[\"D_RGNC_1991\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4645]]",
        4646=>"GEOGCS[\"GCS_Grand_Comoros\",DATUM[\"D_Grand_Comoros\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4646]]",
        4657=>"GEOGCS[\"GCS_Reykjavik_1900\",DATUM[\"D_Reykjavik_1900\",SPHEROID[\"Danish_1876\",6377019.27,300.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4657]]",
        4658=>"GEOGCS[\"GCS_Hjorsey_1955\",DATUM[\"D_Hjorsey_1955\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4658]]",
        4659=>"GEOGCS[\"GCS_ISN_1993\",DATUM[\"D_Islands_Network_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4659]]",
        4660=>"GEOGCS[\"GCS_Helle_1954\",DATUM[\"D_Helle_1954\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4660]]",
        4661=>"GEOGCS[\"GCS_LKS_1992\",DATUM[\"D_Latvia_1992\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4661]]",
        4662=>"GEOGCS[\"GCS_IGN72_Grande_Terre\",DATUM[\"D_IGN72_Grande_Terre\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4662]]",
        4663=>"GEOGCS[\"GCS_Porto_Santo_1995\",DATUM[\"D_Porto_Santo_1995\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4663]]",
        4664=>"GEOGCS[\"GCS_Azores_Oriental_1995\",DATUM[\"D_Azores_Oriental_Islands_1995\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4664]]",
        4665=>"GEOGCS[\"GCS_Azores_Central_1995\",DATUM[\"D_Azores_Central_Islands_1995\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4665]]",
        4666=>"GEOGCS[\"GCS_Lisbon_1890\",DATUM[\"D_Lisbon_1890\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4666]]",
        4667=>"GEOGCS[\"GCS_IKBD_1992\",DATUM[\"D_Iraq_Kuwait_Boundary_1992\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4667]]",
        4668=>"GEOGCS[\"GCS_European_1979\",DATUM[\"D_European_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4668]]",
        4669=>"GEOGCS[\"GCS_LKS_1994\",DATUM[\"D_Lithuania_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4669]]",
        4670=>"GEOGCS[\"GCS_IGM_1995\",DATUM[\"D_IGM_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4670]]",
        4671=>"GEOGCS[\"GCS_Voirol_1879\",DATUM[\"D_Voirol_1879\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4671]]",
        4672=>"GEOGCS[\"GCS_Chatham_Island_1971\",DATUM[\"D_Chatham_Island_1971\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4672]]",
        4673=>"GEOGCS[\"GCS_Chatham_Islands_1979\",DATUM[\"D_Chatham_Islands_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4673]]",
        4674=>"GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4674]]",
        4675=>"GEOGCS[\"GCS_Guam_1963\",DATUM[\"D_Guam_1963\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4675]]",
        4676=>"GEOGCS[\"GCS_Vientiane_1982\",DATUM[\"D_Vientiane_1982\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4676]]",
        4677=>"GEOGCS[\"GCS_Lao_1993\",DATUM[\"D_Lao_1993\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4677]]",
        4678=>"GEOGCS[\"GCS_Lao_1997\",DATUM[\"D_Lao_National_Datum_1997\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4678]]",
        4679=>"GEOGCS[\"GCS_Jouik_1961\",DATUM[\"D_Jouik_1961\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4679]]",
        4680=>"GEOGCS[\"GCS_Nouakchott_1965\",DATUM[\"D_Nouakchott_1965\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4680]]",
        4682=>"GEOGCS[\"GCS_Gulshan_303\",DATUM[\"D_Gulshan_303\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4682]]",
        4683=>"GEOGCS[\"GCS_PRS_1992\",DATUM[\"D_Philippine_Reference_System_1992\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4683]]",
        4684=>"GEOGCS[\"GCS_Gan_1970\",DATUM[\"D_Gan_1970\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4684]]",
        4686=>"GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4686]]",
        4687=>"GEOGCS[\"GCS_RGPF\",DATUM[\"D_Reseau_Geodesique_de_la_Polynesie_Francaise\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4687]]",
        4688=>"GEOGCS[\"GCS_Fatu_Iva_1972\",DATUM[\"D_Fatu_Iva_1972\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4688]]",
        4689=>"GEOGCS[\"GCS_IGN63_Hiva_Oa\",DATUM[\"D_IGN63_Hiva_Oa\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4689]]",
        4690=>"GEOGCS[\"GCS_Tahiti_1979\",DATUM[\"D_Tahiti_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4690]]",
        4691=>"GEOGCS[\"GCS_Moorea_1987\",DATUM[\"D_Moorea_1987\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4691]]",
        4692=>"GEOGCS[\"GCS_Maupiti_1983\",DATUM[\"D_Maupiti_1983\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4692]]",
        4693=>"GEOGCS[\"GCS_Nakhl-e_Ghanem\",DATUM[\"D_Nakhl-e_Ghanem\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4693]]",
        4694=>"GEOGCS[\"GCS_POSGAR_1994\",DATUM[\"D_POSGAR_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4694]]",
        4695=>"GEOGCS[\"GCS_Katanga_1955\",DATUM[\"D_Katanga_1955\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4695]]",
        4696=>"GEOGCS[\"GCS_Kasai_1953\",DATUM[\"D_Kasai_1953\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4696]]",
        4697=>"GEOGCS[\"GCS_IGC_1962_6th_Parallel_South\",DATUM[\"D_IGC_1962_Arc_of_the_6th_Parallel_South\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4697]]",
        4698=>"GEOGCS[\"GCS_Kerguelen_Island_1949\",DATUM[\"D_Kerguelen_Island_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4698]]",
        4699=>"GEOGCS[\"GCS_Le_Pouce_1934\",DATUM[\"D_Le_Pouce_1934\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4699]]",
        4700=>"GEOGCS[\"GCS_IGN_Astro_1960\",DATUM[\"D_IGN_Astro_1960\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4700]]",
        4701=>"GEOGCS[\"GCS_IGCB_1955\",DATUM[\"D_Institut_Geographique_du_Congo_Belge_1955\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4701]]",
        4702=>"GEOGCS[\"GCS_Mauritania_1999\",DATUM[\"D_Mauritania_1999\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4702]]",
        4703=>"GEOGCS[\"GCS_Mhast_1951\",DATUM[\"D_Mhast_1951\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4703]]",
        4704=>"GEOGCS[\"GCS_Mhast_Onshore\",DATUM[\"D_Mhast_Onshore\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4704]]",
        4705=>"GEOGCS[\"GCS_Mhast_Offshore\",DATUM[\"D_Mhast_Offshore\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4705]]",
        4706=>"GEOGCS[\"GCS_Egypt_Gulf_of_Suez_S-650_TL\",DATUM[\"D_Egypt_Gulf_of_Suez_S-650_TL\",SPHEROID[\"Helmert_1906\",6378200.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4706]]",
        4707=>"GEOGCS[\"GCS_Tern_Island_1961\",DATUM[\"D_Tern_Island_1961\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4707]]",
        4708=>"GEOGCS[\"GCS_Anna_1_1965\",DATUM[\"D_Anna_1_1965\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4708]]",
        4709=>"GEOGCS[\"GCS_Beacon_E_1945\",DATUM[\"D_Beacon_E_1945\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4709]]",
        4710=>"GEOGCS[\"GCS_DOS_71_4\",DATUM[\"D_DOS_71_4\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4710]]",
        4711=>"GEOGCS[\"GCS_Astro_1952\",DATUM[\"D_Astro_1952\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4711]]",
        4712=>"GEOGCS[\"GCS_Ascension_Island_1958\",DATUM[\"D_Ascension_Island_1958\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4712]]",
        4713=>"GEOGCS[\"GCS_Ayabelle\",DATUM[\"D_Ayabelle\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4713]]",
        4714=>"GEOGCS[\"GCS_Bellevue_IGN\",DATUM[\"D_Bellevue_IGN\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4714]]",
        4715=>"GEOGCS[\"GCS_Camp_Area\",DATUM[\"D_Camp_Area\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4715]]",
        4716=>"GEOGCS[\"GCS_Canton_1966\",DATUM[\"D_Canton_1966\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4716]]",
        4717=>"GEOGCS[\"GCS_Cape_Canaveral\",DATUM[\"D_Cape_Canaveral\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4717]]",
        4718=>"GEOGCS[\"GCS_Solomon_1968\",DATUM[\"D_Solomon_1968\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4718]]",
        4719=>"GEOGCS[\"GCS_Easter_Island_1967\",DATUM[\"D_Easter_Island_1967\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4719]]",
        4720=>"GEOGCS[\"GCS_Fiji_1986\",DATUM[\"D_Fiji_1986\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4720]]",
        4721=>"GEOGCS[\"GCS_Fiji_1956\",DATUM[\"D_Fiji_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4721]]",
        4722=>"GEOGCS[\"GCS_ISTS_061_1968\",DATUM[\"D_ISTS_061_1968\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4722]]",
        4723=>"GEOGCS[\"GCS_Grand_Cayman_1959\",DATUM[\"D_Grand_Cayman_1959\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4723]]",
        4724=>"GEOGCS[\"GCS_ISTS_073_1969\",DATUM[\"D_ISTS_073_1969\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4724]]",
        4725=>"GEOGCS[\"GCS_Johnston_Island_1961\",DATUM[\"D_Johnston_Island_1961\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4725]]",
        4726=>"GEOGCS[\"GCS_Little_Cayman_1961\",DATUM[\"D_Little_Cayman_1961\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4726]]",
        4727=>"GEOGCS[\"GCS_Midway_1961\",DATUM[\"D_Midway_1961\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4727]]",
        4728=>"GEOGCS[\"GCS_Pico_de_Las_Nieves\",DATUM[\"D_Pico_de_Las_Nieves\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4728]]",
        4729=>"GEOGCS[\"GCS_Pitcairn_1967\",DATUM[\"D_Pitcairn_1967\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4729]]",
        4730=>"GEOGCS[\"GCS_Santo_DOS_1965\",DATUM[\"D_Santo_DOS_1965\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4730]]",
        4731=>"GEOGCS[\"GCS_Viti_Levu_1916\",DATUM[\"D_Viti_Levu_1916\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4731]]",
        4732=>"GEOGCS[\"GCS_Wake_Eniwetok_1960\",DATUM[\"D_Wake_Eniwetok_1960\",SPHEROID[\"Hough_1960\",6378270.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4732]]",
        4733=>"GEOGCS[\"GCS_Wake_Island_1952\",DATUM[\"D_Wake_Island_1952\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4733]]",
        4734=>"GEOGCS[\"GCS_Tristan_1968\",DATUM[\"D_Tristan_1968\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4734]]",
        4735=>"GEOGCS[\"GCS_Kusaie_1951\",DATUM[\"D_Kusaie_1951\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4735]]",
        4736=>"GEOGCS[\"GCS_Deception_Island\",DATUM[\"D_Deception_Island\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4736]]",
        4737=>"GEOGCS[\"GCS_Korea_2000\",DATUM[\"D_Korea_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4737]]",
        4738=>"GEOGCS[\"GCS_Hong_Kong_1963\",DATUM[\"D_Hong_Kong_1963\",SPHEROID[\"Clarke_1858\",6378293.645208759,294.260676369]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4738]]",
        4739=>"GEOGCS[\"GCS_Hong_Kong_1963_67\",DATUM[\"D_Hong_Kong_1963_67\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4739]]",
        4740=>"GEOGCS[\"GCS_PZ_1990\",DATUM[\"D_Parametrop_Zemp_1990\",SPHEROID[\"PZ_1990\",6378136.0,298.257839303]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4740]]",
        4741=>"GEOGCS[\"GCS_FD_1954\",DATUM[\"D_Faroe_Datum_1954\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4741]]",
        4742=>"GEOGCS[\"GCS_GDM_2000\",DATUM[\"D_GDM_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4742]]",
        4743=>"GEOGCS[\"GCS_Karbala_1979_Polservice\",DATUM[\"D_Karbala_1979_Polservice\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4743]]",
        4744=>"GEOGCS[\"GCS_Nahrwan_1934\",DATUM[\"D_Nahrwan_1934\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4744]]",
        4745=>"GEOGCS[\"GCS_RD/83\",DATUM[\"D_Rauenberg_1983\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4745]]",
        4746=>"GEOGCS[\"GCS_PD/83\",DATUM[\"D_Potsdam_1983\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4746]]",
        4747=>"GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4747]]",
        4748=>"GEOGCS[\"GCS_Vanua_Levu_1915\",DATUM[\"D_Vanua_Levu_1915\",SPHEROID[\"Clarke_1880_Intl_Ft\",6378306.3696,293.4663076556299]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4748]]",
        4749=>"GEOGCS[\"GCS_RGNC_1991-93\",DATUM[\"D_Reseau_Geodesique_de_Nouvelle_Caledonie_1991-93\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4749]]",
        4750=>"GEOGCS[\"GCS_ST87_Ouvea\",DATUM[\"D_ST87_Ouvea\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4750]]",
        4751=>"GEOGCS[\"GCS_Kertau_RSO\",DATUM[\"D_Kertau_RSO\",SPHEROID[\"Everest_Modified_1969\",6377295.664,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4751]]",
        4752=>"GEOGCS[\"GCS_Viti_Levu_1912\",DATUM[\"D_Viti_Levu_1912\",SPHEROID[\"Clarke_1880_Intl_Ft\",6378306.3696,293.4663076556299]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4752]]",
        4753=>"GEOGCS[\"GCS_fk89\",DATUM[\"D_fk89\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4753]]",
        4754=>"GEOGCS[\"GCS_LGD2006\",DATUM[\"D_Libyan_Geodetic_Datum_2006\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4754]]",
        4755=>"GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4755]]",
        4756=>"GEOGCS[\"GCS_VN_2000\",DATUM[\"D_Vietnam_2000\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4756]]",
        4757=>"GEOGCS[\"GCS_SVY21\",DATUM[\"D_SVY21\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4757]]",
        4758=>"GEOGCS[\"GCS_JAD_2001\",DATUM[\"D_Jamaica_2001\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4758]]",
        4759=>"GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4759]]",
        4760=>"GEOGCS[\"GCS_WGS_1966\",DATUM[\"D_WGS_1966\",SPHEROID[\"WGS_1966\",6378145.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4760]]",
        4761=>"GEOGCS[\"GCS_HTRS96\",DATUM[\"D_Croatian_Terrestrial_Reference_System\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4761]]",
        4762=>"GEOGCS[\"GCS_Bermuda_2000\",DATUM[\"D_Bermuda_2000\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4762]]",
        4763=>"GEOGCS[\"GCS_Pitcairn_2006\",DATUM[\"D_Pitcairn_2006\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4763]]",
        4764=>"GEOGCS[\"GCS_RSRGD2000\",DATUM[\"D_Ross_Sea_Region_Geodetic_Datum_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4764]]",
        4765=>"GEOGCS[\"GCS_Slovenia_1996\",DATUM[\"D_Slovenia_Geodetic_Datum_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4765]]",
        4801=>"GEOGCS[\"GCS_Bern_1898_Bern\",DATUM[\"D_Bern_1898\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Bern\",7.439583333333333],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4801]]",
        4802=>"GEOGCS[\"GCS_Bogota_Bogota\",DATUM[\"D_Bogota\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Bogota\",-74.08091666666667],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4802]]",
        4803=>"GEOGCS[\"GCS_Lisbon_Lisbon\",DATUM[\"D_Lisbon\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Lisbon\",-9.131906111111112],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4803]]",
        4804=>"GEOGCS[\"GCS_Makassar_Jakarta\",DATUM[\"D_Makassar\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Jakarta\",106.8077194444444],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4804]]",
        4805=>"GEOGCS[\"GCS_MGI_Ferro\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4805]]",
        4806=>"GEOGCS[\"GCS_Monte_Mario_Rome\",DATUM[\"D_Monte_Mario\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Rome\",12.45233333333333],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4806]]",
        4807=>"GEOGCS[\"GCS_NTF_Paris\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897],AUTHORITY[\"EPSG\",4807]]",
        4808=>"GEOGCS[\"GCS_Padang_1884_Jakarta\",DATUM[\"D_Padang_1884\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Jakarta\",106.8077194444444],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4808]]",
        4809=>"GEOGCS[\"GCS_Belge_1950_Brussels\",DATUM[\"D_Belge_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Brussels\",4.367975],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4809]]",
        4810=>"GEOGCS[\"GCS_Tananarive_1925_Paris\",DATUM[\"D_Tananarive_1925\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897],AUTHORITY[\"EPSG\",4810]]",
        4811=>"GEOGCS[\"GCS_Voirol_1875_Paris\",DATUM[\"D_Voirol_1875\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897],AUTHORITY[\"EPSG\",4811]]",
        4813=>"GEOGCS[\"GCS_Batavia_Jakarta\",DATUM[\"D_Batavia\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Jakarta\",106.8077194444444],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4813]]",
        4814=>"GEOGCS[\"GCS_RT38_Stockholm\",DATUM[\"D_Stockholm_1938\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Stockholm\",18.05827777777778],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4814]]",
        4815=>"GEOGCS[\"GCS_Greek_Athens\",DATUM[\"D_Greek\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Athens\",23.7163375],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4815]]",
        4816=>"GEOGCS[\"GCS_Carthage_Paris\",DATUM[\"D_Carthage\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897],AUTHORITY[\"EPSG\",4816]]",
        4817=>"GEOGCS[\"GCS_NGO_1948_Oslo\",DATUM[\"D_NGO_1948\",SPHEROID[\"Bessel_Modified\",6377492.018,299.1528128]],PRIMEM[\"Oslo\",10.72291666666667],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4817]]",
        4818=>"GEOGCS[\"GCS_S_JTSK_Ferro\",DATUM[\"D_S_JTSK\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4818]]",
        4819=>"GEOGCS[\"GCS_Nord_Sahara_1959_Paris\",DATUM[\"D_Nord_Sahara_1959\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897],AUTHORITY[\"EPSG\",4819]]",
        4820=>"GEOGCS[\"GCS_Gunung_Segara_Jakarta\",DATUM[\"D_Gunung_Segara\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Jakarta\",106.8077194444444],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4820]]",
        4821=>"GEOGCS[\"GCS_Voirol_1879_Paris\",DATUM[\"D_Voirol_1879\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897],AUTHORITY[\"EPSG\",4821]]",
        4823=>"GEOGCS[\"GCS_Sao_Tome\",DATUM[\"D_Sao_Tome\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4823]]",
        4824=>"GEOGCS[\"GCS_Principe\",DATUM[\"D_Principe\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4824]]",
        4901=>"GEOGCS[\"GCS_ATF_Paris\",DATUM[\"D_ATF\",SPHEROID[\"Plessis_1817\",6376523.0,308.64]],PRIMEM[\"Paris_RGS\",2.337208333333333],UNIT[\"Grad\",0.01570796326794897],AUTHORITY[\"EPSG\",4901]]",
        4902=>"GEOGCS[\"GCS_Nord_de_Guerre_Paris\",DATUM[\"D_Nord_de_Guerre\",SPHEROID[\"Plessis_1817\",6376523.0,308.64]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897],AUTHORITY[\"EPSG\",4902]]",
        4903=>"GEOGCS[\"GCS_Madrid_1870_Madrid\",DATUM[\"D_Madrid_1870\",SPHEROID[\"Struve_1860\",6378298.3,294.73]],PRIMEM[\"Madrid\",-3.687938888888889],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4903]]",
        4904=>"GEOGCS[\"GCS_Lisbon_1890_Lisbon\",DATUM[\"D_Lisbon_1890\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Lisbon\",-9.131906111111112],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",4904]]",
        5013=>"GEOGCS[\"GCS_PTRA08\",DATUM[\"D_PTRA08\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5013]]",
        5228=>"GEOGCS[\"GCS_S_JTSK/05\",DATUM[\"D_S_JTSK_05\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5228]]",
        5229=>"GEOGCS[\"GCS_S_JTSK/05_Ferro\",DATUM[\"D_S_JTSK_05\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5229]]",
        5233=>"GEOGCS[\"GCS_SLD99\",DATUM[\"D_Sri_Lanka_Datum_1999\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5233]]",
        5246=>"GEOGCS[\"GCS_GDBD2009\",DATUM[\"D_GDBD2009\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5246]]",
        5252=>"GEOGCS[\"GCS_TUREF\",DATUM[\"D_Turkish_National_Reference_Frame\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5252]]",
        5264=>"GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5264]]",
        5324=>"GEOGCS[\"GCS_ISN_2004\",DATUM[\"D_Islands_Network_2004\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5324]]",
        5340=>"GEOGCS[\"GCS_POSGAR_2007\",DATUM[\"D_POSGAR_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5340]]",
        5354=>"GEOGCS[\"GCS_MARGEN\",DATUM[\"D_Marco_Geodesico_Nacional\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5354]]",
        5360=>"GEOGCS[\"GCS_SIRGAS-Chile\",DATUM[\"D_SIRGAS-Chile\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5360]]",
        5365=>"GEOGCS[\"GCS_CR05\",DATUM[\"D_Costa_Rica_2005\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5365]]",
        5371=>"GEOGCS[\"GCS_MARCARIO_SOLIS\",DATUM[\"D_SGNP_MARCARIO_SOLIS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5371]]",
        5373=>"GEOGCS[\"GCS_Peru96\",DATUM[\"D_Peru96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5373]]",
        5381=>"GEOGCS[\"GCS_SIRGAS-ROU98\",DATUM[\"D_SIRGAS-ROU98\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5381]]",
        5393=>"GEOGCS[\"GCS_SIRGAS_ES2007.8\",DATUM[\"D_SIRGAS_ES2007.8\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5393]]",
        5451=>"GEOGCS[\"GCS_Ocotepeque_1935\",DATUM[\"D_Ocotepeque_1935\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5451]]",
        5464=>"GEOGCS[\"GCS_Sibun_Gorge_1922\",DATUM[\"D_Sibun_Gorge_1922\",SPHEROID[\"Clarke_1858\",6378293.645208759,294.260676369]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5464]]",
        5467=>"GEOGCS[\"GCS_Panama-Colon_1911\",DATUM[\"D_Panama-Colon-1911\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5467]]",
        5489=>"GEOGCS[\"GCS_RGAF09\",DATUM[\"Reseau_Geodesique_des_Antilles_Francaises_2009\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5489]]",
        5524=>"GEOGCS[\"GCS_Corrego_Alegre_1961\",DATUM[\"D_Corrego_Alegre_1961\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5524]]",
        5527=>"GEOGCS[\"GCS_SAD_1969_96\",DATUM[\"D_South_American_Datum_1969_96\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5527]]",
        5546=>"GEOGCS[\"GCS_PNG94\",DATUM[\"D_Papua_New_Guinea_Geodetic_Datum_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5546]]",
        5561=>"GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5561]]",
        5593=>"GEOGCS[\"GCS_FEH2010\",DATUM[\"D_Fehmarnbelt_Datum_2010\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5593]]",
        5681=>"GEOGCS[\"GCS_DB_REF\",DATUM[\"D_Deutsche_Bahn_Reference_System\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5681]]",
        5886=>"GEOGCS[\"GCS_TGD2005\",DATUM[\"D_Tonga_Geodetic_Datum_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",5886]]",
        6135=>"GEOGCS[\"GCS_CIGD11\",DATUM[\"D_Cayman_Islands_Geodetic_Datum_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",6135]]",
        6207=>"GEOGCS[\"GCS_Nepal_Nagarkot\",DATUM[\"D_Nepal_Nagarkot\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",6207]]",
        6311=>"GEOGCS[\"GCS_CGRS_1993\",DATUM[\"D_Cyprus_Geodetic_Reference_System_1993\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",6311]]",
        6318=>"GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",6318]]",
        6322=>"GEOGCS[\"GCS_NAD_1983_PA11\",DATUM[\"D_NAD_1983_PA11\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",6322]]",
        6325=>"GEOGCS[\"GCS_NAD_1983_MA11\",DATUM[\"D_NAD_1983_MA11\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",6325]]",
        6365=>"GEOGCS[\"GCS_Mexico_ITRF2008\",DATUM[\"D_Mexico_ITRF2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",6365]]",
        6668=>"GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",6668]]",
        6706=>"GEOGCS[\"GCS_RDN2008\",DATUM[\"D_Rete_Dinamica_Nazionale_2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",6706]]",
        6783=>"GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",6783]]",
        6881=>"GEOGCS[\"GCS_Aden_1925\",DATUM[\"D_Aden_1925\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",6881]]",
        6882=>"GEOGCS[\"GCS_Bekaa_Valley_1920\",DATUM[\"D_Bekaa_Valley_1920\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",6882]]",
        6883=>"GEOGCS[\"GCS_Bioko\",DATUM[\"D_Bioko\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",6883]]",
        6892=>"GEOGCS[\"GCS_South_East_Island_1943\",DATUM[\"D_South_East_Island_1943\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",6892]]",
        6894=>"GEOGCS[\"GCS_Gambia\",DATUM[\"D_Gambia\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",6894]]",
        6980=>"GEOGCS[\"IGD05\",DATUM[\"Israel_Geodetic_Datum_2005\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7136]]",
        6983=>"GEOGCS[\"IG05_Intermediate_CRS\",DATUM[\"IG05_Intermediate_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",6983]]",
        6987=>"GEOGCS[\"IGD05(2012)\",DATUM[\"Israeli_Geodetic_Datum_2005(2012)\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7139]]",
        6990=>"GEOGCS[\"IG05(2012)_Intermediate_CRS\",DATUM[\"IG05(2012)_Intermediate_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",6990]]",
        7035=>"GEOGCS[\"RGSPM06_(lon-lat)\",DATUM[\"D_Reseau_Geodesique_de_St_Pierre_et_Miquelon_2006\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7035]]",
        7037=>"GEOGCS[\"RGR92_(lon-lat)\",DATUM[\"D_RGR_1992\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7037]]",
        7039=>"GEOGCS[\"RGM04_(lon-lat)\",DATUM[\"D_Reseau_Geodesique_de_Mayotte_2004\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7039]]",
        7041=>"GEOGCS[\"RGFG95_(lon-lat)\",DATUM[\"D_RGFG_1995\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7041]]",
        7073=>"GEOGCS[\"GCS_RGTAAF07\",DATUM[\"D_Reseau_Geodesique_des_Terres_Australes_et_Antarctiques_Francaises_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7073]]",
        7084=>"GEOGCS[\"RGF93_(lon-lat)\",DATUM[\"D_RGF_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7084]]",
        7086=>"GEOGCS[\"RGAF09_(lon-lat)\",DATUM[\"Reseau_Geodesique_des_Antilles_Francaises_2009\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7086]]",
        7133=>"GEOGCS[\"RGTAAF07_(lon-lat)\",DATUM[\"D_Reseau_Geodesique_des_Terres_Australes_et_Antarctiques_Francaises_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7133]]",
        7136=>"GEOGCS[\"IGD05\",DATUM[\"Israel_Geodetic_Datum_2005\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7136]]",
        7139=>"GEOGCS[\"IGD05(2012)\",DATUM[\"Israeli_Geodetic_Datum_2005(2012)\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7139]]",
        7373=>"GEOGCS[\"ONGD14\",DATUM[\"Oman_National_Geodetic_Datum_2014\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7373]]",
        7683=>"GEOGCS[\"GSK-2011\",DATUM[\"Geodezicheskaya_Sistema_Koordinat_2011\",SPHEROID[\"GSK-2011\",6378136.5,298.2564151]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7683]]",
        7798=>"GEOGCS[\"BGS2005\",DATUM[\"Bulgaria_Geodetic_System_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7798]]",
        7844=>"GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7844]]",
        7881=>"GEOGCS[\"St_Helena_Tritan\",DATUM[\"St_Helena_Tritan\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7881]]",
        7886=>"GEOGCS[\"SHGD2015\",DATUM[\"St_Helena_Geodetic_Datum_2015\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",7886]]",
        8042=>"GEOGCS[\"Gusterberg(Ferro)\",DATUM[\"Gusterberg_(Ferro)\",SPHEROID[\"Zach_1812\",6376045.0,310.0]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",8042]]",
        8043=>"GEOGCS[\"St._Stephen(Ferro)\",DATUM[\"St._Stephen_(Ferro)\",SPHEROID[\"Zach_1812\",6376045.0,310.0]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",8043]]",
        8086=>"GEOGCS[\"ISN2016\",DATUM[\"Islands_Net_2016\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",8086]]",
        8232=>"GEOGCS[\"NAD83(CSRS96)\",DATUM[\"North_American_Datum_of_1983_(CSRS96)\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",8232]]",
        8237=>"GEOGCS[\"NAD83(CSRS)v2\",DATUM[\"North_American_Datum_of_1983_(CSRS)_version_2\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",8237]]",
        8240=>"GEOGCS[\"NAD83(CSRS)v3\",DATUM[\"North_American_Datum_of_1983_(CSRS)_version_3\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",8240]]",
        8246=>"GEOGCS[\"NAD83(CSRS)v4\",DATUM[\"North_American_Datum_of_1983_(CSRS)_version_4\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",8246]]",
        8249=>"GEOGCS[\"NAD83(CSRS)v5\",DATUM[\"North_American_Datum_of_1983_(CSRS)_version_5\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",8249]]",
        8252=>"GEOGCS[\"NAD83(CSRS)v6\",DATUM[\"North_American_Datum_of_1983_(CSRS)_version_6\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",8252]]",
        8255=>"GEOGCS[\"NAD83(CSRS)v7\",DATUM[\"North_American_Datum_of_1983_(CSRS)_version_7\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433],AUTHORITY[\"EPSG\",8255]]",
        2000=>"PROJCS[\"Anguilla_1957_British_West_Indies_Grid\",GEOGCS[\"GCS_Anguilla_1957\",DATUM[\"D_Anguilla_1957\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-62.0],PARAMETER[\"Scale_Factor\",0.9995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2000]]",
        2001=>"PROJCS[\"Antigua_1943_British_West_Indies_Grid\",GEOGCS[\"GCS_Antigua_1943\",DATUM[\"D_Antigua_1943\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-62.0],PARAMETER[\"Scale_Factor\",0.9995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2001]]",
        2002=>"PROJCS[\"Dominica_1945_British_West_Indies_Grid\",GEOGCS[\"GCS_Dominica_1945\",DATUM[\"D_Dominica_1945\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-62.0],PARAMETER[\"Scale_Factor\",0.9995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2002]]",
        2003=>"PROJCS[\"Grenada_1953_British_West_Indies_Grid\",GEOGCS[\"GCS_Grenada_1953\",DATUM[\"D_Grenada_1953\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-62.0],PARAMETER[\"Scale_Factor\",0.9995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2003]]",
        2004=>"PROJCS[\"Montserrat_1958_British_West_Indies_Grid\",GEOGCS[\"GCS_Montserrat_1958\",DATUM[\"D_Montserrat_1958\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-62.0],PARAMETER[\"Scale_Factor\",0.9995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2004]]",
        2005=>"PROJCS[\"St_Kitts_1955_British_West_Indies_Grid\",GEOGCS[\"GCS_St_Kitts_1955\",DATUM[\"D_St_Kitts_1955\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-62.0],PARAMETER[\"Scale_Factor\",0.9995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2005]]",
        2006=>"PROJCS[\"St_Lucia_1955_British_West_Indies_Grid\",GEOGCS[\"GCS_St_Lucia_1955\",DATUM[\"D_St_Lucia_1955\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-62.0],PARAMETER[\"Scale_Factor\",0.9995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2006]]",
        2007=>"PROJCS[\"St_Vincent_1945_British_West_Indies_Grid\",GEOGCS[\"GCS_St_Vincent_1945\",DATUM[\"D_St_Vincent_1945\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-62.0],PARAMETER[\"Scale_Factor\",0.9995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2007]]",
        2008=>"PROJCS[\"NAD_1927_CGQ77_MTM_2_SCoPQ\",GEOGCS[\"GCS_NAD_1927_CGQ77\",DATUM[\"D_NAD_1927_CGQ77\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-55.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2008]]",
        2009=>"PROJCS[\"NAD_1927_CGQ77_MTM_3_SCoPQ\",GEOGCS[\"GCS_NAD_1927_CGQ77\",DATUM[\"D_NAD_1927_CGQ77\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-58.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2009]]",
        2010=>"PROJCS[\"NAD_1927_CGQ77_MTM_4_SCoPQ\",GEOGCS[\"GCS_NAD_1927_CGQ77\",DATUM[\"D_NAD_1927_CGQ77\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-61.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2010]]",
        2011=>"PROJCS[\"NAD_1927_CGQ77_MTM_5_SCoPQ\",GEOGCS[\"GCS_NAD_1927_CGQ77\",DATUM[\"D_NAD_1927_CGQ77\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-64.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2011]]",
        2012=>"PROJCS[\"NAD_1927_CGQ77_MTM_6_SCoPQ\",GEOGCS[\"GCS_NAD_1927_CGQ77\",DATUM[\"D_NAD_1927_CGQ77\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-67.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2012]]",
        2013=>"PROJCS[\"NAD_1927_CGQ77_MTM_7_SCoPQ\",GEOGCS[\"GCS_NAD_1927_CGQ77\",DATUM[\"D_NAD_1927_CGQ77\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2013]]",
        2014=>"PROJCS[\"NAD_1927_CGQ77_MTM_8_SCoPQ\",GEOGCS[\"GCS_NAD_1927_CGQ77\",DATUM[\"D_NAD_1927_CGQ77\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-73.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2014]]",
        2015=>"PROJCS[\"NAD_1927_CGQ77_MTM_9_SCoPQ\",GEOGCS[\"GCS_NAD_1927_CGQ77\",DATUM[\"D_NAD_1927_CGQ77\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-76.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2015]]",
        2016=>"PROJCS[\"NAD_1927_CGQ77_MTM_10_SCoPQ\",GEOGCS[\"GCS_NAD_1927_CGQ77\",DATUM[\"D_NAD_1927_CGQ77\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2016]]",
        2017=>"PROJCS[\"NAD_1927_DEF_1976_MTM_8\",GEOGCS[\"GCS_NAD_1927_Definition_1976\",DATUM[\"D_NAD_1927_Definition_1976\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-73.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2017]]",
        2018=>"PROJCS[\"NAD_1927_DEF_1976_MTM_9\",GEOGCS[\"GCS_NAD_1927_Definition_1976\",DATUM[\"D_NAD_1927_Definition_1976\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-76.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2018]]",
        2019=>"PROJCS[\"NAD_1927_DEF_1976_MTM_10\",GEOGCS[\"GCS_NAD_1927_Definition_1976\",DATUM[\"D_NAD_1927_Definition_1976\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2019]]",
        2020=>"PROJCS[\"NAD_1927_DEF_1976_MTM_11\",GEOGCS[\"GCS_NAD_1927_Definition_1976\",DATUM[\"D_NAD_1927_Definition_1976\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2020]]",
        2021=>"PROJCS[\"NAD_1927_DEF_1976_MTM_12\",GEOGCS[\"GCS_NAD_1927_Definition_1976\",DATUM[\"D_NAD_1927_Definition_1976\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2021]]",
        2022=>"PROJCS[\"NAD_1927_DEF_1976_MTM_13\",GEOGCS[\"GCS_NAD_1927_Definition_1976\",DATUM[\"D_NAD_1927_Definition_1976\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2022]]",
        2023=>"PROJCS[\"NAD_1927_DEF_1976_MTM_14\",GEOGCS[\"GCS_NAD_1927_Definition_1976\",DATUM[\"D_NAD_1927_Definition_1976\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2023]]",
        2024=>"PROJCS[\"NAD_1927_DEF_1976_MTM_15\",GEOGCS[\"GCS_NAD_1927_Definition_1976\",DATUM[\"D_NAD_1927_Definition_1976\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2024]]",
        2025=>"PROJCS[\"NAD_1927_DEF_1976_MTM_16\",GEOGCS[\"GCS_NAD_1927_Definition_1976\",DATUM[\"D_NAD_1927_Definition_1976\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2025]]",
        2026=>"PROJCS[\"NAD_1927_DEF_1976_MTM_17\",GEOGCS[\"GCS_NAD_1927_Definition_1976\",DATUM[\"D_NAD_1927_Definition_1976\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-96.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2026]]",
        2027=>"PROJCS[\"NAD_1927_DEF_1976_UTM_Zone_15N\",GEOGCS[\"GCS_NAD_1927_Definition_1976\",DATUM[\"D_NAD_1927_Definition_1976\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2027]]",
        2028=>"PROJCS[\"NAD_1927_DEF_1976_UTM_Zone_16N\",GEOGCS[\"GCS_NAD_1927_Definition_1976\",DATUM[\"D_NAD_1927_Definition_1976\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2028]]",
        2029=>"PROJCS[\"NAD_1927_DEF_1976_UTM_Zone_17N\",GEOGCS[\"GCS_NAD_1927_Definition_1976\",DATUM[\"D_NAD_1927_Definition_1976\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2029]]",
        2030=>"PROJCS[\"NAD_1927_DEF_1976_UTM_Zone_18N\",GEOGCS[\"GCS_NAD_1927_Definition_1976\",DATUM[\"D_NAD_1927_Definition_1976\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2030]]",
        2031=>"PROJCS[\"NAD_1927_CGQ77_UTM_Zone_17N\",GEOGCS[\"GCS_NAD_1927_CGQ77\",DATUM[\"D_NAD_1927_CGQ77\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2031]]",
        2032=>"PROJCS[\"NAD_1927_CGQ77_UTM_Zone_18N\",GEOGCS[\"GCS_NAD_1927_CGQ77\",DATUM[\"D_NAD_1927_CGQ77\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2032]]",
        2033=>"PROJCS[\"NAD_1927_CGQ77_UTM_Zone_19N\",GEOGCS[\"GCS_NAD_1927_CGQ77\",DATUM[\"D_NAD_1927_CGQ77\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2033]]",
        2034=>"PROJCS[\"NAD_1927_CGQ77_UTM_Zone_20N\",GEOGCS[\"GCS_NAD_1927_CGQ77\",DATUM[\"D_NAD_1927_CGQ77\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2034]]",
        2035=>"PROJCS[\"NAD_1927_CGQ77_UTM_Zone_21N\",GEOGCS[\"GCS_NAD_1927_CGQ77\",DATUM[\"D_NAD_1927_CGQ77\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2035]]",
        2036=>"PROJCS[\"NAD_1983_CSRS_New_Brunswick_Stereographic\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",7500000.0],PARAMETER[\"Central_Meridian\",-66.5],PARAMETER[\"Scale_Factor\",0.999912],PARAMETER[\"Latitude_Of_Origin\",46.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2953]]",
        2037=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_19N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2960]]",
        2038=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_20N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2961]]",
        2039=>"PROJCS[\"Israel_TM_Grid\",GEOGCS[\"GCS_Israel\",DATUM[\"D_Israel\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",219529.584],PARAMETER[\"False_Northing\",626907.39],PARAMETER[\"Central_Meridian\",35.20451694444445],PARAMETER[\"Scale_Factor\",1.0000067],PARAMETER[\"Latitude_Of_Origin\",31.73439361111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2039]]",
        2040=>"PROJCS[\"Locodjo_1965_UTM_Zone_30N\",GEOGCS[\"GCS_Locodjo_1965\",DATUM[\"D_Locodjo_1965\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2040]]",
        2041=>"PROJCS[\"Abidjan_1987_UTM_Zone_30N\",GEOGCS[\"GCS_Abidjan_1987\",DATUM[\"D_Abidjan_1987\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2041]]",
        2042=>"PROJCS[\"Locodjo_1965_UTM_Zone_29N\",GEOGCS[\"GCS_Locodjo_1965\",DATUM[\"D_Locodjo_1965\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2042]]",
        2043=>"PROJCS[\"Abidjan_1987_UTM_Zone_29N\",GEOGCS[\"GCS_Abidjan_1987\",DATUM[\"D_Abidjan_1987\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2043]]",
        2044=>"PROJCS[\"Hanoi_1972_GK_Zone_18\",GEOGCS[\"GCS_Hanoi_1972\",DATUM[\"D_Hanoi_1972\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",18500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2044]]",
        2045=>"PROJCS[\"Hanoi_1972_GK_Zone_19\",GEOGCS[\"GCS_Hanoi_1972\",DATUM[\"D_Hanoi_1972\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",19500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2045]]",
        2056=>"PROJCS[\"CH1903+_LV95\",GEOGCS[\"GCS_CH1903+\",DATUM[\"D_CH1903+\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Center\"],PARAMETER[\"False_Easting\",2600000.0],PARAMETER[\"False_Northing\",1200000.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Azimuth\",90.0],PARAMETER[\"Longitude_Of_Center\",7.439583333333333],PARAMETER[\"Latitude_Of_Center\",46.95240555555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2056]]",
        2057=>"PROJCS[\"Rassadiran_Nakhl_e_Taqi\",GEOGCS[\"GCS_Rassadiran\",DATUM[\"D_Rassadiran\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Center\"],PARAMETER[\"False_Easting\",658377.437],PARAMETER[\"False_Northing\",3044969.194],PARAMETER[\"Scale_Factor\",0.999895934],PARAMETER[\"Azimuth\",0.5716611944444444],PARAMETER[\"Longitude_Of_Center\",52.60353916666667],PARAMETER[\"Latitude_Of_Center\",27.51882880555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2057]]",
        2058=>"PROJCS[\"ED_1950_ED77_UTM_Zone_38N\",GEOGCS[\"GCS_European_1950_ED77\",DATUM[\"D_European_1950_ED77\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2058]]",
        2059=>"PROJCS[\"ED_1950_ED77_UTM_Zone_39N\",GEOGCS[\"GCS_European_1950_ED77\",DATUM[\"D_European_1950_ED77\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2059]]",
        2060=>"PROJCS[\"ED_1950_ED77_UTM_Zone_40N\",GEOGCS[\"GCS_European_1950_ED77\",DATUM[\"D_European_1950_ED77\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2060]]",
        2061=>"PROJCS[\"ED_1950_ED77_UTM_Zone_41N\",GEOGCS[\"GCS_European_1950_ED77\",DATUM[\"D_European_1950_ED77\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2061]]",
        2062=>"PROJCS[\"Madrid_1870_Madrid_Spain\",GEOGCS[\"GCS_Madrid_1870_Madrid\",DATUM[\"D_Madrid_1870\",SPHEROID[\"Struve_1860\",6378298.3,294.73]],PRIMEM[\"Madrid\",-3.687938888888889],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",600000.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Scale_Factor\",0.9988085293],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2062]]",
        2063=>"PROJCS[\"Dabola_1981_UTM_Zone_28N\",GEOGCS[\"GCS_Dabola_1981\",DATUM[\"D_Dabola_1981\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3461]]",
        2064=>"PROJCS[\"Dabola_1981_UTM_Zone_29N\",GEOGCS[\"GCS_Dabola_1981\",DATUM[\"D_Dabola_1981\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3462]]",
        2065=>"PROJCS[\"S-JTSK_Ferro_Krovak\",GEOGCS[\"GCS_S_JTSK_Ferro\",DATUM[\"D_S_JTSK\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Krovak\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Pseudo_Standard_Parallel_1\",78.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Azimuth\",30.28813975277778],PARAMETER[\"Longitude_Of_Center\",42.5],PARAMETER[\"Latitude_Of_Center\",49.5],PARAMETER[\"X_Scale\",1.0],PARAMETER[\"Y_Scale\",1.0],PARAMETER[\"XY_Plane_Rotation\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2065]]",
        2066=>"PROJCS[\"Mount_Dillon_Tobago_Grid\",GEOGCS[\"GCS_Mount_Dillon\",DATUM[\"D_Mount_Dillon\",SPHEROID[\"Clarke_1858\",6378293.645208759,294.260676369]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",187500.0],PARAMETER[\"False_Northing\",180000.0],PARAMETER[\"Central_Meridian\",-60.68600888888889],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",11.25217861111111],UNIT[\"Link_Clarke\",0.201166195164],AUTHORITY[\"EPSG\",2066]]",
        2067=>"PROJCS[\"Naparima_1955_UTM_Zone_20N\",GEOGCS[\"GCS_Naparima_1955\",DATUM[\"D_Naparima_1955\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2067]]",
        2068=>"PROJCS[\"ELD_1979_Libya_5\",GEOGCS[\"GCS_European_Libyan_Datum_1979\",DATUM[\"D_European_Libyan_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2068]]",
        2069=>"PROJCS[\"ELD_1979_Libya_6\",GEOGCS[\"GCS_European_Libyan_Datum_1979\",DATUM[\"D_European_Libyan_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",11.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2069]]",
        2070=>"PROJCS[\"ELD_1979_Libya_7\",GEOGCS[\"GCS_European_Libyan_Datum_1979\",DATUM[\"D_European_Libyan_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",13.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2070]]",
        2071=>"PROJCS[\"ELD_1979_Libya_8\",GEOGCS[\"GCS_European_Libyan_Datum_1979\",DATUM[\"D_European_Libyan_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2071]]",
        2072=>"PROJCS[\"ELD_1979_Libya_9\",GEOGCS[\"GCS_European_Libyan_Datum_1979\",DATUM[\"D_European_Libyan_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",17.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2072]]",
        2073=>"PROJCS[\"ELD_1979_Libya_10\",GEOGCS[\"GCS_European_Libyan_Datum_1979\",DATUM[\"D_European_Libyan_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",19.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2073]]",
        2074=>"PROJCS[\"ELD_1979_Libya_11\",GEOGCS[\"GCS_European_Libyan_Datum_1979\",DATUM[\"D_European_Libyan_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2074]]",
        2075=>"PROJCS[\"ELD_1979_Libya_12\",GEOGCS[\"GCS_European_Libyan_Datum_1979\",DATUM[\"D_European_Libyan_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",23.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2075]]",
        2076=>"PROJCS[\"ELD_1979_Libya_13\",GEOGCS[\"GCS_European_Libyan_Datum_1979\",DATUM[\"D_European_Libyan_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",25.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2076]]",
        2077=>"PROJCS[\"ELD_1979_UTM_Zone_32N\",GEOGCS[\"GCS_European_Libyan_Datum_1979\",DATUM[\"D_European_Libyan_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2077]]",
        2078=>"PROJCS[\"ELD_1979_UTM_Zone_33N\",GEOGCS[\"GCS_European_Libyan_Datum_1979\",DATUM[\"D_European_Libyan_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2078]]",
        2079=>"PROJCS[\"ELD_1979_UTM_Zone_34N\",GEOGCS[\"GCS_European_Libyan_Datum_1979\",DATUM[\"D_European_Libyan_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2079]]",
        2080=>"PROJCS[\"ELD_1979_UTM_Zone_35N\",GEOGCS[\"GCS_European_Libyan_Datum_1979\",DATUM[\"D_European_Libyan_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2080]]",
        2081=>"PROJCS[\"Chos_Malal_1914_Argentina_2\",GEOGCS[\"GCS_Chos_Malal_1914\",DATUM[\"D_Chos_Malal_1914\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2081]]",
        2082=>"PROJCS[\"Pampa_del_Castillo_Argentina_2\",GEOGCS[\"GCS_Pampa_del_Castillo\",DATUM[\"D_Pampa_del_Castillo\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2082]]",
        2083=>"PROJCS[\"Hito_XVIII_1963_Argentina_2\",GEOGCS[\"GCS_Hito_XVIII_1963\",DATUM[\"D_Hito_XVIII_1963\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2083]]",
        2084=>"PROJCS[\"Hito_XVIII_1963_UTM_19S\",GEOGCS[\"GCS_Hito_XVIII_1963\",DATUM[\"D_Hito_XVIII_1963\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2084]]",
        2085=>"PROJCS[\"NAD_1927_Cuba_Norte\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",280296.016],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",22.35],PARAMETER[\"Scale_Factor\",0.99993602],PARAMETER[\"Latitude_Of_Origin\",22.35],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2085]]",
        2086=>"PROJCS[\"NAD_1927_Cuba_Sur\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",229126.939],PARAMETER[\"Central_Meridian\",-76.83333333333333],PARAMETER[\"Standard_Parallel_1\",20.71666666666667],PARAMETER[\"Scale_Factor\",0.99994848],PARAMETER[\"Latitude_Of_Origin\",20.71666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2086]]",
        2087=>"PROJCS[\"ELD_1979_TM_12_NE\",GEOGCS[\"GCS_European_Libyan_Datum_1979\",DATUM[\"D_European_Libyan_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2087]]",
        2088=>"PROJCS[\"Carthage_TM_11_NE\",GEOGCS[\"GCS_Carthage\",DATUM[\"D_Carthage\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",11.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2088]]",
        2089=>"PROJCS[\"Yemen_NGN_1996_UTM_Zone_38N\",GEOGCS[\"GCS_Yemen_NGN_1996\",DATUM[\"D_Yemen_NGN_1996\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2089]]",
        2090=>"PROJCS[\"Yemen_NGN_1996_UTM_Zone_39N\",GEOGCS[\"GCS_Yemen_NGN_1996\",DATUM[\"D_Yemen_NGN_1996\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2090]]",
        2091=>"PROJCS[\"South_Yemen_GK_Zone_8\",GEOGCS[\"GCS_South_Yemen\",DATUM[\"D_South_Yemen\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",8500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2395]]",
        2092=>"PROJCS[\"South_Yemen_GK_Zone_9\",GEOGCS[\"GCS_South_Yemen\",DATUM[\"D_South_Yemen\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",9500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2396]]",
        2093=>"PROJCS[\"Hanoi_1972_GK_106_NE\",GEOGCS[\"GCS_Hanoi_1972\",DATUM[\"D_Hanoi_1972\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",106.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2093]]",
        2094=>"PROJCS[\"WGS_1972_BE_TM_106_NE\",GEOGCS[\"GCS_WGS_1972_BE\",DATUM[\"D_WGS_1972_BE\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",106.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2094]]",
        2095=>"PROJCS[\"Bissau_UTM_Zone_28N\",GEOGCS[\"GCS_Bissau\",DATUM[\"D_Bissau\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2095]]",
        2096=>"PROJCS[\"Korean_1985_Korea_East_Belt\",GEOGCS[\"GCS_Korean_Datum_1985\",DATUM[\"D_Korean_Datum_1985\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2096]]",
        2097=>"PROJCS[\"Korean_1985_Korea_Central_Belt\",GEOGCS[\"GCS_Korean_Datum_1985\",DATUM[\"D_Korean_Datum_1985\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",127.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2097]]",
        2098=>"PROJCS[\"Korean_1985_Korea_West_Belt\",GEOGCS[\"GCS_Korean_Datum_1985\",DATUM[\"D_Korean_Datum_1985\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",125.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2098]]",
        2099=>"PROJCS[\"Qatar_1948_Qatar_Grid\",GEOGCS[\"GCS_Qatar_1948\",DATUM[\"D_Qatar_1948\",SPHEROID[\"Helmert_1906\",6378200.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",50.76138888888889],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",25.38236111111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2099]]",
        2100=>"PROJCS[\"Greek_Grid\",GEOGCS[\"GCS_GGRS_1987\",DATUM[\"D_GGRS_1987\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2100]]",
        2101=>"PROJCS[\"Lake_Maracaibo_Grid_M1\",GEOGCS[\"GCS_Lake\",DATUM[\"D_Lake\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",-52684.972],PARAMETER[\"Central_Meridian\",-71.60561777777777],PARAMETER[\"Standard_Parallel_1\",10.16666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",10.16666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2101]]",
        2102=>"PROJCS[\"Lake_Maracaibo_Grid\",GEOGCS[\"GCS_Lake\",DATUM[\"D_Lake\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",147315.028],PARAMETER[\"Central_Meridian\",-71.60561777777777],PARAMETER[\"Standard_Parallel_1\",10.16666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",10.16666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2102]]",
        2103=>"PROJCS[\"Lake_Maracaibo_Grid_M3\",GEOGCS[\"GCS_Lake\",DATUM[\"D_Lake\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",447315.028],PARAMETER[\"Central_Meridian\",-71.60561777777777],PARAMETER[\"Standard_Parallel_1\",10.16666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",10.16666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2103]]",
        2104=>"PROJCS[\"Lake_Maracaibo_La_Rosa_Grid\",GEOGCS[\"GCS_Lake\",DATUM[\"D_Lake\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",-17044.0],PARAMETER[\"False_Northing\",-23139.97],PARAMETER[\"Central_Meridian\",-71.60561777777777],PARAMETER[\"Standard_Parallel_1\",10.16666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",10.16666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2104]]",
        2105=>"PROJCS[\"NZGD_2000_Mount_Eden_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",174.7641666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",-36.87972222222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2105]]",
        2106=>"PROJCS[\"NZGD_2000_Bay_of_Plenty_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",176.4661111111111],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-37.76111111111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2106]]",
        2107=>"PROJCS[\"NZGD_2000_Poverty_Bay_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",177.8855555555556],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-38.62444444444444],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2107]]",
        2108=>"PROJCS[\"NZGD_2000_Hawkes_Bay_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",176.6736111111111],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-39.65083333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2108]]",
        2109=>"PROJCS[\"NZGD_2000_Taranaki_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",174.2277777777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-39.13555555555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2109]]",
        2110=>"PROJCS[\"NZGD_2000_Tuhirangi_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",175.64],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-39.51222222222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2110]]",
        2111=>"PROJCS[\"NZGD_2000_Wanganui_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",175.4880555555555],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-40.24194444444444],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2111]]",
        2112=>"PROJCS[\"NZGD_2000_Wairarapa_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",175.6472222222222],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-40.92527777777777],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2112]]",
        2113=>"PROJCS[\"NZGD_2000_Wellington_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",174.7763888888889],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-41.3011111111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2113]]",
        2114=>"PROJCS[\"NZGD_2000_Collingwood_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",172.6719444444444],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-40.71472222222223],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2114]]",
        2115=>"PROJCS[\"NZGD_2000_Nelson_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",173.2991666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-41.27444444444444],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2115]]",
        2116=>"PROJCS[\"NZGD_2000_Karamea_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",172.1088888888889],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-41.28972222222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2116]]",
        2117=>"PROJCS[\"NZGD_2000_Buller_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",171.5811111111111],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-41.81055555555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2117]]",
        2118=>"PROJCS[\"NZGD_2000_Grey_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",171.5497222222222],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-42.33361111111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2118]]",
        2119=>"PROJCS[\"NZGD_2000_Amuri_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",173.01],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-42.68888888888888],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2119]]",
        2120=>"PROJCS[\"NZGD_2000_Marlborough_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",173.8019444444444],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-41.54444444444444],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2120]]",
        2121=>"PROJCS[\"NZGD_2000_Hokitika_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",170.9797222222222],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-42.88611111111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2121]]",
        2122=>"PROJCS[\"NZGD_2000_Okarito_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",170.2608333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-43.11],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2122]]",
        2123=>"PROJCS[\"NZGD_2000_Jacksons_Bay_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",168.6061111111111],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-43.97777777777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2123]]",
        2124=>"PROJCS[\"NZGD_2000_Mount_Pleasant_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",172.7269444444445],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-43.59055555555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2124]]",
        2125=>"PROJCS[\"NZGD_2000_Gawler_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",171.3605555555555],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-43.74861111111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2125]]",
        2126=>"PROJCS[\"NZGD_2000_Timaru_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",171.0572222222222],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-44.40194444444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2126]]",
        2127=>"PROJCS[\"NZGD_2000_Lindis_Peak_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",169.4675],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-44.735],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2127]]",
        2128=>"PROJCS[\"NZGD_2000_Mount_Nicholas_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",168.3986111111111],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-45.13277777777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2128]]",
        2129=>"PROJCS[\"NZGD_2000_Mount_York_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",167.7386111111111],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-45.56361111111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2129]]",
        2130=>"PROJCS[\"NZGD_2000_Observation_Point_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",170.6283333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-45.81611111111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2130]]",
        2131=>"PROJCS[\"NZGD_2000_North_Taieri_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",170.2825],PARAMETER[\"Scale_Factor\",0.99996],PARAMETER[\"Latitude_Of_Origin\",-45.86138888888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2131]]",
        2132=>"PROJCS[\"NZGD_2000_Bluff_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",168.3427777777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-46.6],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2132]]",
        2133=>"PROJCS[\"NZGD_2000_UTM_Zone_58S\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2133]]",
        2134=>"PROJCS[\"NZGD_2000_UTM_Zone_59S\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2134]]",
        2135=>"PROJCS[\"NZGD_2000_UTM_Zone_60S\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2135]]",
        2136=>"PROJCS[\"Accra_Ghana_Grid\",GEOGCS[\"GCS_Accra\",DATUM[\"D_Accra\",SPHEROID[\"War_Office\",6378300.0,296.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",900000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-1.0],PARAMETER[\"Scale_Factor\",0.99975],PARAMETER[\"Latitude_Of_Origin\",4.666666666666667],UNIT[\"Foot_Gold_Coast\",0.3047997101815088],AUTHORITY[\"EPSG\",2136]]",
        2137=>"PROJCS[\"Accra_TM_1_NW\",GEOGCS[\"GCS_Accra\",DATUM[\"D_Accra\",SPHEROID[\"War_Office\",6378300.0,296.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-1.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2137]]",
        2138=>"PROJCS[\"NAD_1927_CGQ77_Quebec_Lambert\",GEOGCS[\"GCS_NAD_1927_CGQ77\",DATUM[\"D_NAD_1927_CGQ77\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-68.5],PARAMETER[\"Standard_Parallel_1\",46.0],PARAMETER[\"Standard_Parallel_2\",60.0],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2138]]",
        2139=>"PROJCS[\"NAD_1983_CSRS_MTM_2_SCoPQ\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-55.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2944]]",
        2140=>"PROJCS[\"NAD_1983_CSRS_MTM_3\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-58.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2945]]",
        2141=>"PROJCS[\"NAD_1983_CSRS_MTM_4\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-61.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2946]]",
        2142=>"PROJCS[\"NAD_1983_CSRS_MTM_5\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-64.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2947]]",
        2143=>"PROJCS[\"NAD_1983_CSRS_MTM_6\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-67.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2948]]",
        2144=>"PROJCS[\"NAD_1983_CSRS_MTM_7\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2949]]",
        2145=>"PROJCS[\"NAD_1983_CSRS_MTM_8\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-73.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2950]]",
        2146=>"PROJCS[\"NAD_1983_CSRS_MTM_9\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-76.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2951]]",
        2147=>"PROJCS[\"NAD_1983_CSRS_MTM_10\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2952]]",
        2148=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_21N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2962]]",
        2149=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_18N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2959]]",
        2150=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_17N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2958]]",
        2151=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_13N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2957]]",
        2152=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_12N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2956]]",
        2153=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_11N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2955]]",
        2154=>"PROJCS[\"RGF_1993_Lambert_93\",GEOGCS[\"GCS_RGF_1993\",DATUM[\"D_RGF_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",6600000.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Standard_Parallel_1\",44.0],PARAMETER[\"Standard_Parallel_2\",49.0],PARAMETER[\"Latitude_Of_Origin\",46.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2154]]",
        2155=>"PROJCS[\"Samoa_1962_Samoa_Lambert\",GEOGCS[\"GCS_American_Samoa_1962\",DATUM[\"D_American_Samoa_1962\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",312234.65],PARAMETER[\"Central_Meridian\",-170.0],PARAMETER[\"Standard_Parallel_1\",-14.26666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-14.26666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3102]]",
        2157=>"PROJCS[\"IRENET95_Irish_Transverse_Mercator\",GEOGCS[\"GCS_IRENET95\",DATUM[\"D_IRENET95\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",750000.0],PARAMETER[\"Central_Meridian\",-8.0],PARAMETER[\"Scale_Factor\",0.99982],PARAMETER[\"Latitude_Of_Origin\",53.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2157]]",
        2158=>"PROJCS[\"IRENET95_UTM_Zone_29N\",GEOGCS[\"GCS_IRENET95\",DATUM[\"D_IRENET95\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2158]]",
        2159=>"PROJCS[\"Sierra_Leone_1924_New_Colony_Grid\",GEOGCS[\"GCS_Sierra_Leone_1924\",DATUM[\"D_Sierra_Leone_1924\",SPHEROID[\"War_Office\",6378300.0,296.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",6.666666666666667],UNIT[\"Foot_Gold_Coast\",0.3047997101815088],AUTHORITY[\"EPSG\",2159]]",
        2160=>"PROJCS[\"Sierra_Leone_1924_New_War_Office_Grid\",GEOGCS[\"GCS_Sierra_Leone_1924\",DATUM[\"D_Sierra_Leone_1924\",SPHEROID[\"War_Office\",6378300.0,296.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",600000.0],PARAMETER[\"Central_Meridian\",-12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",6.666666666666667],UNIT[\"Foot_Gold_Coast\",0.3047997101815088],AUTHORITY[\"EPSG\",2160]]",
        2161=>"PROJCS[\"Sierra_Leone_1968_UTM_Zone_28N\",GEOGCS[\"GCS_Sierra_Leone_1968\",DATUM[\"D_Sierra_Leone_1968\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2161]]",
        2162=>"PROJCS[\"Sierra_Leone_1968_UTM_Zone_29N\",GEOGCS[\"GCS_Sierra_Leone_1968\",DATUM[\"D_Sierra_Leone_1968\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2162]]",
        2163=>"PROJCS[\"US_National_Atlas_Equal_Area\",GEOGCS[\"GCS_Sphere_Clarke_1866_Authalic\",DATUM[\"D_Sphere_Clarke_1866_Authalic\",SPHEROID[\"Sphere_Clarke_1866_Authalic\",6370997.0,0.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2163]]",
        2164=>"PROJCS[\"Locodjo_1965_TM_5_NW\",GEOGCS[\"GCS_Locodjo_1965\",DATUM[\"D_Locodjo_1965\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-5.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2164]]",
        2165=>"PROJCS[\"Abidjan_1987_TM_5_NW\",GEOGCS[\"GCS_Abidjan_1987\",DATUM[\"D_Abidjan_1987\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-5.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2165]]",
        2166=>"PROJCS[\"Pulkovo_1942_Adj_1983_3_Degree_GK_Zone_3\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2397]]",
        2167=>"PROJCS[\"Pulkovo_1942_Adj_1983_3_Degree_GK_Zone_4\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2398]]",
        2168=>"PROJCS[\"Pulkovo_1942_Adj_1983_3_Degree_GK_Zone_5\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2399]]",
        2169=>"PROJCS[\"Luxembourg_1930_Gauss\",GEOGCS[\"GCS_Luxembourg_1930\",DATUM[\"D_Luxembourg_1930\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",80000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",6.166666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",49.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2169]]",
        2170=>"PROJCS[\"MGI_Slovenia_Grid\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2170]]",
        2172=>"PROJCS[\"Pulkovo_1942_Adj_1958_Poland_Zone_II\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",4603000.0],PARAMETER[\"False_Northing\",5806000.0],PARAMETER[\"Central_Meridian\",21.50277777777778],PARAMETER[\"Scale_Factor\",0.9998],PARAMETER[\"Latitude_Of_Origin\",53.00194444444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2172]]",
        2173=>"PROJCS[\"Pulkovo_1942_Adj_1958_Poland_Zone_III\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",3501000.0],PARAMETER[\"False_Northing\",5999000.0],PARAMETER[\"Central_Meridian\",17.00833333333333],PARAMETER[\"Scale_Factor\",0.9998],PARAMETER[\"Latitude_Of_Origin\",53.58333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2173]]",
        2174=>"PROJCS[\"Pulkovo_1942_Adj_1958_Poland_Zone_IV\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",3703000.0],PARAMETER[\"False_Northing\",5627000.0],PARAMETER[\"Central_Meridian\",16.67222222222222],PARAMETER[\"Scale_Factor\",0.9998],PARAMETER[\"Latitude_Of_Origin\",51.67083333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2174]]",
        2175=>"PROJCS[\"Pulkovo_1942_Adj_1958_Poland_Zone_V\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",237000.0],PARAMETER[\"False_Northing\",-4700000.0],PARAMETER[\"Central_Meridian\",18.95833333333333],PARAMETER[\"Scale_Factor\",0.999983],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2175]]",
        2176=>"PROJCS[\"ETRS_1989_Poland_CS2000_Zone_5\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.999923],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2176]]",
        2177=>"PROJCS[\"ETRS_1989_Poland_CS2000_Zone_6\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",6500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",18.0],PARAMETER[\"Scale_Factor\",0.999923],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2177]]",
        2178=>"PROJCS[\"ETRS_1989_Poland_CS2000_Zone_7\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",7500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.999923],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2178]]",
        2179=>"PROJCS[\"ETRS_1989_Poland_CS2000_Zone_8\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",8500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",0.999923],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2179]]",
        2180=>"PROJCS[\"ETRS_1989_Poland_CS92\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",-5300000.0],PARAMETER[\"Central_Meridian\",19.0],PARAMETER[\"Scale_Factor\",0.9993],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2180]]",
        2188=>"PROJCS[\"Azores_Occidental_1939_UTM_Zone_25N\",GEOGCS[\"GCS_Azores_Occidental_1939\",DATUM[\"D_Azores_Occidental_Islands_1939\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2188]]",
        2189=>"PROJCS[\"Azores_Central_1948_UTM_Zone_26N\",GEOGCS[\"GCS_Azores_Central_1948\",DATUM[\"D_Azores_Central_Islands_1948\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2189]]",
        2190=>"PROJCS[\"Azores_Oriental_1940_UTM_Zone_26N\",GEOGCS[\"GCS_Azores_Oriental_1940\",DATUM[\"D_Azores_Oriental_Islands_1940\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2190]]",
        2191=>"PROJCS[\"Madeira_1936_UTM_Zone_28N\",GEOGCS[\"GCS_Madeira_1936\",DATUM[\"D_Madeira_1936\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2191]]",
        2192=>"PROJCS[\"ED_1950_France_EuroLambert\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",2200000.0],PARAMETER[\"Central_Meridian\",2.337229166666667],PARAMETER[\"Standard_Parallel_1\",46.8],PARAMETER[\"Scale_Factor\",0.99987742],PARAMETER[\"Latitude_Of_Origin\",46.8],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2192]]",
        2193=>"PROJCS[\"NZGD_2000_New_Zealand_Transverse_Mercator\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1600000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",173.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2193]]",
        2195=>"PROJCS[\"NAD_1983_HARN_UTM_Zone_2S\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2195]]",
        2196=>"PROJCS[\"ETRS_1989_Kp2000_Jutland\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.5],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2196]]",
        2197=>"PROJCS[\"ETRS_1989_Kp2000_Zealand\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2197]]",
        2198=>"PROJCS[\"ETRS_1989_Kp2000_Bornholm\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",900000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2198]]",
        2200=>"PROJCS[\"ATS_1977_New_Brunswick_Stereographic\",GEOGCS[\"GCS_ATS_1977\",DATUM[\"D_ATS_1977\",SPHEROID[\"ATS_1977\",6378135.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",-66.5],PARAMETER[\"Scale_Factor\",0.999912],PARAMETER[\"Latitude_Of_Origin\",46.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2200]]",
        2201=>"PROJCS[\"REGVEN_UTM_Zone_18N\",GEOGCS[\"GCS_REGVEN\",DATUM[\"D_REGVEN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2201]]",
        2202=>"PROJCS[\"REGVEN_UTM_Zone_19N\",GEOGCS[\"GCS_REGVEN\",DATUM[\"D_REGVEN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2202]]",
        2203=>"PROJCS[\"REGVEN_UTM_Zone_20N\",GEOGCS[\"GCS_REGVEN\",DATUM[\"D_REGVEN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2203]]",
        2204=>"PROJCS[\"NAD_1927_StatePlane_Tennessee_FIPS_4100\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-86.0],PARAMETER[\"Standard_Parallel_1\",35.25],PARAMETER[\"Standard_Parallel_2\",36.41666666666666],PARAMETER[\"Latitude_Of_Origin\",34.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2204]]",
        2205=>"PROJCS[\"NAD_1983_StatePlane_Kentucky_North_FIPS_1601\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.25],PARAMETER[\"Standard_Parallel_1\",37.96666666666667],PARAMETER[\"Standard_Parallel_2\",38.96666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2205]]",
        2206=>"PROJCS[\"ED_1950_3_Degree_GK_Zone_9\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",9500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2206]]",
        2207=>"PROJCS[\"ED_1950_3_Degree_GK_Zone_10\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",10500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2207]]",
        2208=>"PROJCS[\"ED_1950_3_Degree_GK_Zone_11\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",11500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2208]]",
        2209=>"PROJCS[\"ED_1950_3_Degree_GK_Zone_12\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",12500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",36.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2209]]",
        2210=>"PROJCS[\"ED_1950_3_Degree_GK_Zone_13\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",13500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2210]]",
        2211=>"PROJCS[\"ED_1950_3_Degree_GK_Zone_14\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",14500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",42.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2211]]",
        2212=>"PROJCS[\"ED_1950_3_Degree_GK_Zone_15\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",15500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2212]]",
        2213=>"PROJCS[\"ETRS_1989_TM_30_NE\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2213]]",
        2214=>"PROJCS[\"Douala_1948_AEF_West\",GEOGCS[\"GCS_Douala_1948\",DATUM[\"D_Douala_1948\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",10.5],PARAMETER[\"Scale_Factor\",0.999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3119]]",
        2215=>"PROJCS[\"Manoca_1962_UTM_Zone_32N\",GEOGCS[\"GCS_Manoca_1962\",DATUM[\"D_Manoca_1962\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2215]]",
        2216=>"PROJCS[\"Qornoq_1927_UTM_Zone_22N\",GEOGCS[\"GCS_Qornoq_1927\",DATUM[\"D_Qornoq_1927\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2216]]",
        2217=>"PROJCS[\"Qornoq_1927_UTM_Zone_23N\",GEOGCS[\"GCS_Qornoq_1927\",DATUM[\"D_Qornoq_1927\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2217]]",
        2219=>"PROJCS[\"ATS_1977_UTM_Zone_19N\",GEOGCS[\"GCS_ATS_1977\",DATUM[\"D_ATS_1977\",SPHEROID[\"ATS_1977\",6378135.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2219]]",
        2220=>"PROJCS[\"ATS_1977_UTM_Zone_20N\",GEOGCS[\"GCS_ATS_1977\",DATUM[\"D_ATS_1977\",SPHEROID[\"ATS_1977\",6378135.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2220]]",
        2222=>"PROJCS[\"NAD_1983_StatePlane_Arizona_East_FIPS_0201_Feet_Intl\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-110.1666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2222]]",
        2223=>"PROJCS[\"NAD_1983_StatePlane_Arizona_Central_FIPS_0202_Feet_Intl\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.9166666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2223]]",
        2224=>"PROJCS[\"NAD_1983_StatePlane_Arizona_West_FIPS_0203_Feet_Intl\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-113.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2224]]",
        2225=>"PROJCS[\"NAD_1983_StatePlane_California_I_FIPS_0401_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Standard_Parallel_2\",41.66666666666666],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2225]]",
        2226=>"PROJCS[\"NAD_1983_StatePlane_California_II_FIPS_0402_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",38.33333333333334],PARAMETER[\"Standard_Parallel_2\",39.83333333333334],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2226]]",
        2227=>"PROJCS[\"NAD_1983_StatePlane_California_III_FIPS_0403_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",37.06666666666667],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2227]]",
        2228=>"PROJCS[\"NAD_1983_StatePlane_California_IV_FIPS_0404_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-119.0],PARAMETER[\"Standard_Parallel_1\",36.0],PARAMETER[\"Standard_Parallel_2\",37.25],PARAMETER[\"Latitude_Of_Origin\",35.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2228]]",
        2229=>"PROJCS[\"NAD_1983_StatePlane_California_V_FIPS_0405_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-118.0],PARAMETER[\"Standard_Parallel_1\",34.03333333333333],PARAMETER[\"Standard_Parallel_2\",35.46666666666667],PARAMETER[\"Latitude_Of_Origin\",33.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2229]]",
        2230=>"PROJCS[\"NAD_1983_StatePlane_California_VI_FIPS_0406_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-116.25],PARAMETER[\"Standard_Parallel_1\",32.78333333333333],PARAMETER[\"Standard_Parallel_2\",33.88333333333333],PARAMETER[\"Latitude_Of_Origin\",32.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2230]]",
        2231=>"PROJCS[\"NAD_1983_StatePlane_Colorado_North_FIPS_0501_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.000316083],PARAMETER[\"False_Northing\",999999.999996],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",39.71666666666667],PARAMETER[\"Standard_Parallel_2\",40.78333333333333],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2231]]",
        2232=>"PROJCS[\"NAD_1983_StatePlane_Colorado_Central_FIPS_0502_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.000316083],PARAMETER[\"False_Northing\",999999.999996],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",38.45],PARAMETER[\"Standard_Parallel_2\",39.75],PARAMETER[\"Latitude_Of_Origin\",37.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2232]]",
        2233=>"PROJCS[\"NAD_1983_StatePlane_Colorado_South_FIPS_0503_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.000316083],PARAMETER[\"False_Northing\",999999.999996],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",37.23333333333333],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2233]]",
        2234=>"PROJCS[\"NAD_1983_StatePlane_Connecticut_FIPS_0600_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",999999.999996],PARAMETER[\"False_Northing\",499999.999998],PARAMETER[\"Central_Meridian\",-72.75],PARAMETER[\"Standard_Parallel_1\",41.2],PARAMETER[\"Standard_Parallel_2\",41.86666666666667],PARAMETER[\"Latitude_Of_Origin\",40.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2234]]",
        2235=>"PROJCS[\"NAD_1983_StatePlane_Delaware_FIPS_0700_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.41666666666667],PARAMETER[\"Scale_Factor\",0.999995],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2235]]",
        2236=>"PROJCS[\"NAD_1983_StatePlane_Florida_East_FIPS_0901_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2236]]",
        2237=>"PROJCS[\"NAD_1983_StatePlane_Florida_West_FIPS_0902_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2237]]",
        2238=>"PROJCS[\"NAD_1983_StatePlane_Florida_North_FIPS_0903_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.5],PARAMETER[\"Standard_Parallel_1\",29.58333333333333],PARAMETER[\"Standard_Parallel_2\",30.75],PARAMETER[\"Latitude_Of_Origin\",29.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2238]]",
        2239=>"PROJCS[\"NAD_1983_StatePlane_Georgia_East_FIPS_1001_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2239]]",
        2240=>"PROJCS[\"NAD_1983_StatePlane_Georgia_West_FIPS_1002_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2296583.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2240]]",
        2241=>"PROJCS[\"NAD_1983_StatePlane_Idaho_East_FIPS_1101_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-112.1666666666667],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2241]]",
        2242=>"PROJCS[\"NAD_1983_StatePlane_Idaho_Central_FIPS_1102_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-114.0],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2242]]",
        2243=>"PROJCS[\"NAD_1983_StatePlane_Idaho_West_FIPS_1103_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-115.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2243]]",
        2244=>"PROJCS[\"NAD_1983_StatePlane_Indiana_East_FIPS_1301_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",328083.3333333333],PARAMETER[\"False_Northing\",820208.3333333333],PARAMETER[\"Central_Meridian\",-85.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2965]]",
        2245=>"PROJCS[\"NAD_1983_StatePlane_Indiana_West_FIPS_1302_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2952750.0],PARAMETER[\"False_Northing\",820208.3333333333],PARAMETER[\"Central_Meridian\",-87.08333333333333],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2966]]",
        2246=>"PROJCS[\"NAD_1983_StatePlane_Kentucky_North_FIPS_1601_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.25],PARAMETER[\"Standard_Parallel_1\",37.96666666666667],PARAMETER[\"Standard_Parallel_2\",38.96666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2246]]",
        2247=>"PROJCS[\"NAD_1983_StatePlane_Kentucky_South_FIPS_1602_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Standard_Parallel_1\",36.73333333333333],PARAMETER[\"Standard_Parallel_2\",37.93333333333333],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2247]]",
        2248=>"PROJCS[\"NAD_1983_StatePlane_Maryland_FIPS_1900_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.0],PARAMETER[\"Standard_Parallel_1\",38.3],PARAMETER[\"Standard_Parallel_2\",39.45],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2248]]",
        2249=>"PROJCS[\"NAD_1983_StatePlane_Massachusetts_Mainland_FIPS_2001_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",2460625.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Standard_Parallel_1\",41.71666666666667],PARAMETER[\"Standard_Parallel_2\",42.68333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2249]]",
        2250=>"PROJCS[\"NAD_1983_StatePlane_Massachusetts_Island_FIPS_2002_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.5],PARAMETER[\"Standard_Parallel_1\",41.28333333333333],PARAMETER[\"Standard_Parallel_2\",41.48333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2250]]",
        2251=>"PROJCS[\"NAD_1983_StatePlane_Michigan_North_FIPS_2111_Feet_Intl\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",26246719.16010498],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Standard_Parallel_1\",45.48333333333333],PARAMETER[\"Standard_Parallel_2\",47.08333333333334],PARAMETER[\"Latitude_Of_Origin\",44.78333333333333],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2251]]",
        2252=>"PROJCS[\"NAD_1983_StatePlane_Michigan_Central_FIPS_2112_Feet_Intl\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",19685039.37007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.36666666666666],PARAMETER[\"Standard_Parallel_1\",44.18333333333333],PARAMETER[\"Standard_Parallel_2\",45.7],PARAMETER[\"Latitude_Of_Origin\",43.31666666666667],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2252]]",
        2253=>"PROJCS[\"NAD_1983_StatePlane_Michigan_South_FIPS_2113_Feet_Intl\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",13123359.58005249],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.36666666666666],PARAMETER[\"Standard_Parallel_1\",42.1],PARAMETER[\"Standard_Parallel_2\",43.66666666666666],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2253]]",
        2254=>"PROJCS[\"NAD_1983_StatePlane_Mississippi_East_FIPS_2301_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.83333333333333],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",29.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2254]]",
        2255=>"PROJCS[\"NAD_1983_StatePlane_Mississippi_West_FIPS_2302_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2296583.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.33333333333333],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",29.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2255]]",
        2256=>"PROJCS[\"NAD_1983_StatePlane_Montana_FIPS_2500_Feet_Intl\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968503.937007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-109.5],PARAMETER[\"Standard_Parallel_1\",45.0],PARAMETER[\"Standard_Parallel_2\",49.0],PARAMETER[\"Latitude_Of_Origin\",44.25],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2256]]",
        2257=>"PROJCS[\"NAD_1983_StatePlane_New_Mexico_East_FIPS_3001_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",541337.5],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-104.3333333333333],PARAMETER[\"Scale_Factor\",0.9999090909090909],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2257]]",
        2258=>"PROJCS[\"NAD_1983_StatePlane_New_Mexico_Central_FIPS_3002_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-106.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2258]]",
        2259=>"PROJCS[\"NAD_1983_StatePlane_New_Mexico_West_FIPS_3003_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2723091.666666666],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-107.8333333333333],PARAMETER[\"Scale_Factor\",0.9999166666666667],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2259]]",
        2260=>"PROJCS[\"NAD_1983_StatePlane_New_York_East_FIPS_3101_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",492125.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",38.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2260]]",
        2261=>"PROJCS[\"NAD_1983_StatePlane_New_York_Central_FIPS_3102_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",820208.3333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-76.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2261]]",
        2262=>"PROJCS[\"NAD_1983_StatePlane_New_York_West_FIPS_3103_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1148291.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-78.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2262]]",
        2263=>"PROJCS[\"NAD_1983_StatePlane_New_York_Long_Island_FIPS_3104_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.0],PARAMETER[\"Standard_Parallel_1\",40.66666666666666],PARAMETER[\"Standard_Parallel_2\",41.03333333333333],PARAMETER[\"Latitude_Of_Origin\",40.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2263]]",
        2264=>"PROJCS[\"NAD_1983_StatePlane_North_Carolina_FIPS_3200_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.002616666],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.0],PARAMETER[\"Standard_Parallel_1\",34.33333333333334],PARAMETER[\"Standard_Parallel_2\",36.16666666666666],PARAMETER[\"Latitude_Of_Origin\",33.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2264]]",
        2265=>"PROJCS[\"NAD_1983_StatePlane_North_Dakota_North_FIPS_3301_Feet_Intl\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968503.937007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",47.43333333333333],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2265]]",
        2266=>"PROJCS[\"NAD_1983_StatePlane_North_Dakota_South_FIPS_3302_Feet_Intl\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968503.937007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",46.18333333333333],PARAMETER[\"Standard_Parallel_2\",47.48333333333333],PARAMETER[\"Latitude_Of_Origin\",45.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2266]]",
        2267=>"PROJCS[\"NAD_1983_StatePlane_Oklahoma_North_FIPS_3501_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",35.56666666666667],PARAMETER[\"Standard_Parallel_2\",36.76666666666667],PARAMETER[\"Latitude_Of_Origin\",35.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2267]]",
        2268=>"PROJCS[\"NAD_1983_StatePlane_Oklahoma_South_FIPS_3502_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",33.93333333333333],PARAMETER[\"Standard_Parallel_2\",35.23333333333333],PARAMETER[\"Latitude_Of_Origin\",33.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2268]]",
        2269=>"PROJCS[\"NAD_1983_StatePlane_Oregon_North_FIPS_3601_Feet_Intl\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",8202099.737532808],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",44.33333333333334],PARAMETER[\"Standard_Parallel_2\",46.0],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2269]]",
        2270=>"PROJCS[\"NAD_1983_StatePlane_Oregon_South_FIPS_3602_Feet_Intl\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4921259.842519685],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",42.33333333333334],PARAMETER[\"Standard_Parallel_2\",44.0],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2270]]",
        2271=>"PROJCS[\"NAD_1983_StatePlane_Pennsylvania_North_FIPS_3701_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",40.88333333333333],PARAMETER[\"Standard_Parallel_2\",41.95],PARAMETER[\"Latitude_Of_Origin\",40.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2271]]",
        2272=>"PROJCS[\"NAD_1983_StatePlane_Pennsylvania_South_FIPS_3702_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",39.93333333333333],PARAMETER[\"Standard_Parallel_2\",40.96666666666667],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2272]]",
        2273=>"PROJCS[\"NAD_1983_StatePlane_South_Carolina_FIPS_3900_Feet_Intl\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",32.5],PARAMETER[\"Standard_Parallel_2\",34.83333333333334],PARAMETER[\"Latitude_Of_Origin\",31.83333333333333],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2273]]",
        2274=>"PROJCS[\"NAD_1983_StatePlane_Tennessee_FIPS_4100_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-86.0],PARAMETER[\"Standard_Parallel_1\",35.25],PARAMETER[\"Standard_Parallel_2\",36.41666666666666],PARAMETER[\"Latitude_Of_Origin\",34.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2274]]",
        2275=>"PROJCS[\"NAD_1983_StatePlane_Texas_North_FIPS_4201_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-101.5],PARAMETER[\"Standard_Parallel_1\",34.65],PARAMETER[\"Standard_Parallel_2\",36.18333333333333],PARAMETER[\"Latitude_Of_Origin\",34.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2275]]",
        2276=>"PROJCS[\"NAD_1983_StatePlane_Texas_North_Central_FIPS_4202_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",6561666.666666666],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",32.13333333333333],PARAMETER[\"Standard_Parallel_2\",33.96666666666667],PARAMETER[\"Latitude_Of_Origin\",31.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2276]]",
        2277=>"PROJCS[\"NAD_1983_StatePlane_Texas_Central_FIPS_4203_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2296583.333333333],PARAMETER[\"False_Northing\",9842500.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",30.11666666666667],PARAMETER[\"Standard_Parallel_2\",31.88333333333333],PARAMETER[\"Latitude_Of_Origin\",29.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2277]]",
        2278=>"PROJCS[\"NAD_1983_StatePlane_Texas_South_Central_FIPS_4204_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",13123333.33333333],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Standard_Parallel_1\",28.38333333333333],PARAMETER[\"Standard_Parallel_2\",30.28333333333333],PARAMETER[\"Latitude_Of_Origin\",27.83333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2278]]",
        2279=>"PROJCS[\"NAD_1983_StatePlane_Texas_South_FIPS_4205_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",16404166.66666666],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",26.16666666666667],PARAMETER[\"Standard_Parallel_2\",27.83333333333333],PARAMETER[\"Latitude_Of_Origin\",25.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2279]]",
        2280=>"PROJCS[\"NAD_1983_StatePlane_Utah_North_FIPS_4301_Feet_Intl\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640419.947506561],PARAMETER[\"False_Northing\",3280839.895013123],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",40.71666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.33333333333334],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2280]]",
        2281=>"PROJCS[\"NAD_1983_StatePlane_Utah_Central_FIPS_4302_Feet_Intl\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640419.947506561],PARAMETER[\"False_Northing\",6561679.790026246],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",39.01666666666667],PARAMETER[\"Standard_Parallel_2\",40.65],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2281]]",
        2282=>"PROJCS[\"NAD_1983_StatePlane_Utah_South_FIPS_4303_Feet_Intl\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640419.947506561],PARAMETER[\"False_Northing\",9842519.685039369],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",37.21666666666667],PARAMETER[\"Standard_Parallel_2\",38.35],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2282]]",
        2283=>"PROJCS[\"NAD_1983_StatePlane_Virginia_North_FIPS_4501_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",11482916.66666666],PARAMETER[\"False_Northing\",6561666.666666666],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",38.03333333333333],PARAMETER[\"Standard_Parallel_2\",39.2],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2283]]",
        2284=>"PROJCS[\"NAD_1983_StatePlane_Virginia_South_FIPS_4502_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",11482916.66666666],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",36.76666666666667],PARAMETER[\"Standard_Parallel_2\",37.96666666666667],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2284]]",
        2285=>"PROJCS[\"NAD_1983_StatePlane_Washington_North_FIPS_4601_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.8333333333333],PARAMETER[\"Standard_Parallel_1\",47.5],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2285]]",
        2286=>"PROJCS[\"NAD_1983_StatePlane_Washington_South_FIPS_4602_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",45.83333333333334],PARAMETER[\"Standard_Parallel_2\",47.33333333333334],PARAMETER[\"Latitude_Of_Origin\",45.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2286]]",
        2287=>"PROJCS[\"NAD_1983_StatePlane_Wisconsin_North_FIPS_4801_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",45.56666666666667],PARAMETER[\"Standard_Parallel_2\",46.76666666666667],PARAMETER[\"Latitude_Of_Origin\",45.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2287]]",
        2288=>"PROJCS[\"NAD_1983_StatePlane_Wisconsin_Central_FIPS_4802_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",44.25],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2288]]",
        2289=>"PROJCS[\"NAD_1983_StatePlane_Wisconsin_South_FIPS_4803_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",42.73333333333333],PARAMETER[\"Standard_Parallel_2\",44.06666666666667],PARAMETER[\"Latitude_Of_Origin\",42.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2289]]",
        2290=>"PROJCS[\"Prince_Edward_Island_Stereographic\",GEOGCS[\"GCS_ATS_1977\",DATUM[\"D_ATS_1977\",SPHEROID[\"ATS_1977\",6378135.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",400000.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.999912],PARAMETER[\"Latitude_Of_Origin\",47.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2290]]",
        2291=>"PROJCS[\"NAD_1983_CSRS_Prince_Edward_Island\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.999912],PARAMETER[\"Latitude_Of_Origin\",47.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2954]]",
        2292=>"PROJCS[\"NAD_1983_CSRS_Prince_Edward_Island\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.999912],PARAMETER[\"Latitude_Of_Origin\",47.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2954]]",
        2294=>"PROJCS[\"ATS_1977_MTM_4_Nova_Scotia\",GEOGCS[\"GCS_ATS_1977\",DATUM[\"D_ATS_1977\",SPHEROID[\"ATS_1977\",6378135.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-61.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2294]]",
        2295=>"PROJCS[\"ATS_1977_MTM_5_Nova_Scotia\",GEOGCS[\"GCS_ATS_1977\",DATUM[\"D_ATS_1977\",SPHEROID[\"ATS_1977\",6378135.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-64.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2295]]",
        2308=>"PROJCS[\"Batavia_TM_109_SE\",GEOGCS[\"GCS_Batavia\",DATUM[\"D_Batavia\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",109.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2308]]",
        2309=>"PROJCS[\"WGS_1984_TM_116_SE\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",116.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2309]]",
        2310=>"PROJCS[\"WGS_1984_TM_132_SE\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",132.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2310]]",
        2311=>"PROJCS[\"WGS_1984_TM_6_NE\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",6.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2311]]",
        2312=>"PROJCS[\"Garoua_UTM_Zone_33N\",GEOGCS[\"GCS_Garoua\",DATUM[\"D_Garoua\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2312]]",
        2313=>"PROJCS[\"Kousseri_UTM_Zone_33N\",GEOGCS[\"GCS_Kousseri\",DATUM[\"D_Kousseri\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2313]]",
        2314=>"PROJCS[\"Trinidad_1903_Trinidad_Grid_Feet_Clarke\",GEOGCS[\"GCS_Trinidad_1903\",DATUM[\"D_Trinidad_1903\",SPHEROID[\"Clarke_1858\",6378293.645208759,294.260676369]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",283800.0],PARAMETER[\"False_Northing\",214500.0],PARAMETER[\"Central_Meridian\",-61.33333333333334],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",10.44166666666667],UNIT[\"Foot_Clarke\",0.3047972654],AUTHORITY[\"EPSG\",2314]]",
        2315=>"PROJCS[\"Campo_Inchauspe_UTM_19S\",GEOGCS[\"GCS_Campo_Inchauspe\",DATUM[\"D_Campo_Inchauspe\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2315]]",
        2316=>"PROJCS[\"Campo_Inchauspe_UTM_20S\",GEOGCS[\"GCS_Campo_Inchauspe\",DATUM[\"D_Campo_Inchauspe\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2316]]",
        2317=>"PROJCS[\"PSAD_1956_ICN_Regional\",GEOGCS[\"GCS_Provisional_S_American_1956\",DATUM[\"D_Provisional_S_American_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-66.0],PARAMETER[\"Standard_Parallel_1\",3.0],PARAMETER[\"Standard_Parallel_2\",9.0],PARAMETER[\"Latitude_Of_Origin\",6.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2317]]",
        2318=>"PROJCS[\"Ain_el_Abd_Aramco_Lambert\",GEOGCS[\"GCS_Ain_el_Abd_1970\",DATUM[\"D_Ain_el_Abd_1970\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",48.0],PARAMETER[\"Standard_Parallel_1\",17.0],PARAMETER[\"Standard_Parallel_2\",33.0],PARAMETER[\"Latitude_Of_Origin\",25.08951],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2318]]",
        2319=>"PROJCS[\"ED_1950_TM27\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2319]]",
        2320=>"PROJCS[\"ED_1950_TM30\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2320]]",
        2321=>"PROJCS[\"ED_1950_TM33\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2321]]",
        2322=>"PROJCS[\"ED_1950_TM36\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",36.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2322]]",
        2323=>"PROJCS[\"ED_1950_TM39\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2323]]",
        2324=>"PROJCS[\"ED_1950_TM42\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",42.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2324]]",
        2325=>"PROJCS[\"ED_1950_TM45\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2325]]",
        2326=>"PROJCS[\"Hong_Kong_1980_Grid\",GEOGCS[\"GCS_Hong_Kong_1980\",DATUM[\"D_Hong_Kong_1980\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",836694.05],PARAMETER[\"False_Northing\",819069.8],PARAMETER[\"Central_Meridian\",114.1785555555556],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",22.31213333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2326]]",
        2327=>"PROJCS[\"Xian_1980_GK_Zone_13\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",13500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2327]]",
        2328=>"PROJCS[\"Xian_1980_GK_Zone_14\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",14500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2328]]",
        2329=>"PROJCS[\"Xian_1980_GK_Zone_15\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",15500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2329]]",
        2330=>"PROJCS[\"Xian_1980_GK_Zone_16\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",16500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2330]]",
        2331=>"PROJCS[\"Xian_1980_GK_Zone_17\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",17500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2331]]",
        2332=>"PROJCS[\"Xian_1980_GK_Zone_18\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",18500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2332]]",
        2333=>"PROJCS[\"Xian_1980_GK_Zone_19\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",19500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2333]]",
        2334=>"PROJCS[\"Xian_1980_GK_Zone_20\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",20500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2334]]",
        2335=>"PROJCS[\"Xian_1980_GK_Zone_21\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",21500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2335]]",
        2336=>"PROJCS[\"Xian_1980_GK_Zone_22\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",22500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2336]]",
        2337=>"PROJCS[\"Xian_1980_GK_Zone_23\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",23500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2337]]",
        2338=>"PROJCS[\"Xian_1980_GK_CM_75E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2338]]",
        2339=>"PROJCS[\"Xian_1980_GK_CM_81E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2339]]",
        2340=>"PROJCS[\"Xian_1980_GK_CM_87E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2340]]",
        2341=>"PROJCS[\"Xian_1980_GK_CM_93E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2341]]",
        2342=>"PROJCS[\"Xian_1980_GK_CM_99E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2342]]",
        2343=>"PROJCS[\"Xian_1980_GK_CM_105E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2343]]",
        2344=>"PROJCS[\"Xian_1980_GK_CM_111E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2344]]",
        2345=>"PROJCS[\"Xian_1980_GK_CM_117E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2345]]",
        2346=>"PROJCS[\"Xian_1980_GK_CM_123E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2346]]",
        2347=>"PROJCS[\"Xian_1980_GK_CM_129E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2347]]",
        2348=>"PROJCS[\"Xian_1980_GK_CM_135E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2348]]",
        2349=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_25\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",25500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2349]]",
        2350=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_26\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",26500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",78.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2350]]",
        2351=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_27\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",27500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2351]]",
        2352=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_28\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",28500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",84.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2352]]",
        2353=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_29\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",29500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2353]]",
        2354=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_30\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",30500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2354]]",
        2355=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_31\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",31500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2355]]",
        2356=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_32\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",32500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",96.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2356]]",
        2357=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_33\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",33500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2357]]",
        2358=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_34\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",34500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",102.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2358]]",
        2359=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_35\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",35500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2359]]",
        2360=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_36\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",36500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",108.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2360]]",
        2361=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_37\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",37500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2361]]",
        2362=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_38\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",38500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",114.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2362]]",
        2363=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_39\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",39500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2363]]",
        2364=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_40\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",40500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",120.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2364]]",
        2365=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_41\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",41500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2365]]",
        2366=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_42\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",42500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",126.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2366]]",
        2367=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_43\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",43500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2367]]",
        2368=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_44\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",44500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",132.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2368]]",
        2369=>"PROJCS[\"Xian_1980_3_Degree_GK_Zone_45\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",45500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2369]]",
        2370=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_75E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2370]]",
        2371=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_78E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",78.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2371]]",
        2372=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_81E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2372]]",
        2373=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_84E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",84.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2373]]",
        2374=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_87E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2374]]",
        2375=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_90E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2375]]",
        2376=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_93E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2376]]",
        2377=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_96E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",96.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2377]]",
        2378=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_99E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2378]]",
        2379=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_102E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",102.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2379]]",
        2380=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_105E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2380]]",
        2381=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_108E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",108.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2381]]",
        2382=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_111E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2382]]",
        2383=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_114E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",114.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2383]]",
        2384=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_117E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2384]]",
        2385=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_120E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",120.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2385]]",
        2386=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_123E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2386]]",
        2387=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_126E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",126.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2387]]",
        2388=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_129E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2388]]",
        2389=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_132E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",132.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2389]]",
        2390=>"PROJCS[\"Xian_1980_3_Degree_GK_CM_135E\",GEOGCS[\"GCS_Xian_1980\",DATUM[\"D_Xian_1980\",SPHEROID[\"Xian_1980\",6378140.0,298.257]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2390]]",
        2391=>"PROJCS[\"Finland_Zone_1\",GEOGCS[\"GCS_KKJ\",DATUM[\"D_KKJ\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2391]]",
        2392=>"PROJCS[\"Finland_Zone_2\",GEOGCS[\"GCS_KKJ\",DATUM[\"D_KKJ\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2392]]",
        2393=>"PROJCS[\"Finland_Zone_3\",GEOGCS[\"GCS_KKJ\",DATUM[\"D_KKJ\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2393]]",
        2394=>"PROJCS[\"Finland_Zone_4\",GEOGCS[\"GCS_KKJ\",DATUM[\"D_KKJ\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2394]]",
        2395=>"PROJCS[\"South_Yemen_GK_Zone_8\",GEOGCS[\"GCS_South_Yemen\",DATUM[\"D_South_Yemen\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",8500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2395]]",
        2396=>"PROJCS[\"South_Yemen_GK_Zone_9\",GEOGCS[\"GCS_South_Yemen\",DATUM[\"D_South_Yemen\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",9500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2396]]",
        2397=>"PROJCS[\"Pulkovo_1942_Adj_1983_3_Degree_GK_Zone_3\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2397]]",
        2398=>"PROJCS[\"Pulkovo_1942_Adj_1983_3_Degree_GK_Zone_4\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2398]]",
        2399=>"PROJCS[\"Pulkovo_1942_Adj_1983_3_Degree_GK_Zone_5\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2399]]",
        2400=>"PROJCS[\"RT90_25_gon_W\",GEOGCS[\"GCS_RT_1990\",DATUM[\"D_RT_1990\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.80827777777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2400]]",
        2401=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_25\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",25500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2401]]",
        2402=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_26\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",26500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",78.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2402]]",
        2403=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_27\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",27500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2403]]",
        2404=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_28\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",28500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",84.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2404]]",
        2405=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_29\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",29500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2405]]",
        2406=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_30\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",30500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2406]]",
        2407=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_31\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",31500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2407]]",
        2408=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_32\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",32500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",96.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2408]]",
        2409=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_33\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",33500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2409]]",
        2410=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_34\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",34500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",102.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2410]]",
        2411=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_35\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",35500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2411]]",
        2412=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_36\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",36500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",108.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2412]]",
        2413=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_37\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",37500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2413]]",
        2414=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_38\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",38500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",114.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2414]]",
        2415=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_39\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",39500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2415]]",
        2416=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_40\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",40500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",120.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2416]]",
        2417=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_41\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",41500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2417]]",
        2418=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_42\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",42500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",126.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2418]]",
        2419=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_43\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",43500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2419]]",
        2420=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_44\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",44500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",132.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2420]]",
        2421=>"PROJCS[\"Beijing_1954_3_Degree_GK_Zone_45\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",45500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2421]]",
        2422=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_75E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2422]]",
        2423=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_78E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",78.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2423]]",
        2424=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_81E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2424]]",
        2425=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_84E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",84.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2425]]",
        2426=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_87E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2426]]",
        2427=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_90E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2427]]",
        2428=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_93E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2428]]",
        2429=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_96E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",96.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2429]]",
        2430=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_99E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2430]]",
        2431=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_102E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",102.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2431]]",
        2432=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_105E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2432]]",
        2433=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_108E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",108.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2433]]",
        2434=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_111E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2434]]",
        2435=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_114E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",114.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2435]]",
        2436=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_117E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2436]]",
        2437=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_120E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",120.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2437]]",
        2438=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_123E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2438]]",
        2439=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_126E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",126.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2439]]",
        2440=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_129E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2440]]",
        2441=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_132E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",132.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2441]]",
        2442=>"PROJCS[\"Beijing_1954_3_Degree_GK_CM_135E\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2442]]",
        2443=>"PROJCS[\"JGD_2000_Japan_Zone_1\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",33.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2443]]",
        2444=>"PROJCS[\"JGD_2000_Japan_Zone_2\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",131.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",33.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2444]]",
        2445=>"PROJCS[\"JGD_2000_Japan_Zone_3\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",132.1666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2445]]",
        2446=>"PROJCS[\"JGD_2000_Japan_Zone_4\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",133.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",33.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2446]]",
        2447=>"PROJCS[\"JGD_2000_Japan_Zone_5\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",134.3333333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2447]]",
        2448=>"PROJCS[\"JGD_2000_Japan_Zone_6\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",136.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2448]]",
        2449=>"PROJCS[\"JGD_2000_Japan_Zone_7\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",137.1666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2449]]",
        2450=>"PROJCS[\"JGD_2000_Japan_Zone_8\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",138.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2450]]",
        2451=>"PROJCS[\"JGD_2000_Japan_Zone_9\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",139.8333333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2451]]",
        2452=>"PROJCS[\"JGD_2000_Japan_Zone_10\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",140.8333333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2452]]",
        2453=>"PROJCS[\"JGD_2000_Japan_Zone_11\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",140.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2453]]",
        2454=>"PROJCS[\"JGD_2000_Japan_Zone_12\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",142.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2454]]",
        2455=>"PROJCS[\"JGD_2000_Japan_Zone_13\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",144.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2455]]",
        2456=>"PROJCS[\"JGD_2000_Japan_Zone_14\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",142.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2456]]",
        2457=>"PROJCS[\"JGD_2000_Japan_Zone_15\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",127.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2457]]",
        2458=>"PROJCS[\"JGD_2000_Japan_Zone_16\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",124.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2458]]",
        2459=>"PROJCS[\"JGD_2000_Japan_Zone_17\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",131.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2459]]",
        2460=>"PROJCS[\"JGD_2000_Japan_Zone_18\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",136.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",20.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2460]]",
        2461=>"PROJCS[\"JGD_2000_Japan_Zone_19\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",154.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2461]]",
        2462=>"PROJCS[\"Albanian_1987_GK_Zone_4\",GEOGCS[\"GCS_Albanian_1987\",DATUM[\"D_Albanian_1987\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2462]]",
        2463=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_21E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2463]]",
        2464=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_27E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2464]]",
        2465=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_33E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2465]]",
        2466=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_39E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2466]]",
        2467=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_45E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2467]]",
        2468=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_51E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2468]]",
        2469=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_57E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2469]]",
        2470=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_63E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2470]]",
        2471=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_69E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2471]]",
        2472=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_75E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2472]]",
        2473=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_81E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2473]]",
        2474=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_87E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2474]]",
        2475=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_93E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2475]]",
        2476=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_99E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2476]]",
        2477=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_105E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2477]]",
        2478=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_111E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2478]]",
        2479=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_117E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2479]]",
        2480=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_123E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2480]]",
        2481=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_129E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2481]]",
        2482=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_135E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2482]]",
        2483=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_141E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2483]]",
        2484=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_147E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2484]]",
        2485=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_153E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2485]]",
        2486=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_159E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2486]]",
        2487=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_165E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2487]]",
        2488=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_171E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2488]]",
        2489=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_177E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2489]]",
        2490=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_177W\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2490]]",
        2491=>"PROJCS[\"Pulkovo_1995_Gauss-Kruger_CM_171W\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2491]]",
        2494=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_21E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2494]]",
        2495=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_27E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2495]]",
        2496=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_33E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2496]]",
        2497=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_39E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2497]]",
        2498=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_45E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2498]]",
        2499=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_51E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2499]]",
        2500=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_57E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2500]]",
        2501=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_63E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2501]]",
        2502=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_69E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2502]]",
        2503=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_75E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2503]]",
        2504=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_81E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2504]]",
        2505=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_87E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2505]]",
        2506=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_93E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2506]]",
        2507=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_99E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2507]]",
        2508=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_105E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2508]]",
        2509=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_111E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2509]]",
        2510=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_117E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2510]]",
        2511=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_123E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2511]]",
        2512=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_129E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2512]]",
        2513=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_135E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2513]]",
        2514=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_141E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2514]]",
        2515=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_147E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2515]]",
        2516=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_153E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2516]]",
        2517=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_159E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2517]]",
        2518=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_165E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2518]]",
        2519=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_171E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2519]]",
        2520=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_177E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2520]]",
        2521=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_177W\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2521]]",
        2522=>"PROJCS[\"Pulkovo_1942_Gauss-Kruger_CM_171W\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2522]]",
        2523=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_7\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",7500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2523]]",
        2524=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_8\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",8500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2524]]",
        2525=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_9\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",9500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2525]]",
        2526=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_10\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",10500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2526]]",
        2527=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_11\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",11500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2527]]",
        2528=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_12\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",12500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",36.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2528]]",
        2529=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_13\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",13500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2529]]",
        2530=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_14\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",14500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",42.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2530]]",
        2531=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_15\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",15500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2531]]",
        2532=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_16\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",16500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",48.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2532]]",
        2533=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_17\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",17500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2533]]",
        2534=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_18\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",18500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",54.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2534]]",
        2535=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_19\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",19500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2535]]",
        2536=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_20\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",20500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",60.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2536]]",
        2537=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_21\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",21500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2537]]",
        2538=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_22\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",22500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",66.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2538]]",
        2539=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_23\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",23500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2539]]",
        2540=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_24\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",24500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",72.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2540]]",
        2541=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_25\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",25500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2541]]",
        2542=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_26\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",26500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",78.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2542]]",
        2543=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_27\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",27500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2543]]",
        2544=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_28\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",28500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",84.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2544]]",
        2545=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_29\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",29500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2545]]",
        2546=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_30\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",30500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2546]]",
        2547=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_31\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",31500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2547]]",
        2548=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_32\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",32500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",96.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2548]]",
        2549=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_33\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",33500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2549]]",
        2550=>"PROJCS[\"Samboja_UTM_Zone_50S\",GEOGCS[\"GCS_Samboja\",DATUM[\"D_Samboja\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2550]]",
        2551=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_34\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",34500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",102.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2551]]",
        2552=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_35\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",35500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2552]]",
        2553=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_36\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",36500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",108.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2553]]",
        2554=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_37\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",37500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2554]]",
        2555=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_38\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",38500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",114.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2555]]",
        2556=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_39\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",39500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2556]]",
        2557=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_40\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",40500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",120.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2557]]",
        2558=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_41\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",41500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2558]]",
        2559=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_42\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",42500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",126.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2559]]",
        2560=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_43\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",43500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2560]]",
        2561=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_44\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",44500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",132.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2561]]",
        2562=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_45\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",45500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2562]]",
        2563=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_46\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",46500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",138.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2563]]",
        2564=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_47\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",47500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2564]]",
        2565=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_48\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",48500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",144.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2565]]",
        2566=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_49\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",49500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2566]]",
        2567=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_50\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",50500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",150.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2567]]",
        2568=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_51\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",51500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2568]]",
        2569=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_52\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",52500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",156.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2569]]",
        2570=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_53\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",53500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2570]]",
        2571=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_54\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",54500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",162.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2571]]",
        2572=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_55\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",55500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2572]]",
        2573=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_56\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",56500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",168.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2573]]",
        2574=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_57\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",57500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2574]]",
        2575=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_58\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",58500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",174.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2575]]",
        2576=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_59\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",59500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2576]]",
        2577=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_60\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",60500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",180.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3389]]",
        2578=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_61\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",61500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2578]]",
        2579=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_62\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",62500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-174.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2579]]",
        2580=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_63\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",63500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2580]]",
        2581=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_64\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",64500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-168.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2581]]",
        2582=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_21E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2582]]",
        2583=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_24E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2583]]",
        2584=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_27E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2584]]",
        2585=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_30E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2585]]",
        2586=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_33E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2586]]",
        2587=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_36E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",36.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2587]]",
        2588=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_39E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2588]]",
        2589=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_42E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",42.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2589]]",
        2590=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_45E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2590]]",
        2591=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_48E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",48.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2591]]",
        2592=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_51E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2592]]",
        2593=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_54E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",54.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2593]]",
        2594=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_57E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2594]]",
        2595=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_60E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",60.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2595]]",
        2596=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_63E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2596]]",
        2597=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_66E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",66.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2597]]",
        2598=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_69E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2598]]",
        2599=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_72E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",72.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2599]]",
        2600=>"PROJCS[\"LKS_1994_Lithuania_TM\",GEOGCS[\"GCS_LKS_1994\",DATUM[\"D_Lithuania_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",0.9998],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3346]]",
        2601=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_75E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2601]]",
        2602=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_78E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",78.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2602]]",
        2603=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_81E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2603]]",
        2604=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_84E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",84.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2604]]",
        2605=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_87E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2605]]",
        2606=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_90E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2606]]",
        2607=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_93E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2607]]",
        2608=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_96E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",96.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2608]]",
        2609=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_99E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2609]]",
        2610=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_102E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",102.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2610]]",
        2611=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_105E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2611]]",
        2612=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_108E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",108.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2612]]",
        2613=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_111E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2613]]",
        2614=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_114E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",114.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2614]]",
        2615=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_117E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2615]]",
        2616=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_120E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",120.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2616]]",
        2617=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_123E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2617]]",
        2618=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_126E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",126.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2618]]",
        2619=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_129E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2619]]",
        2620=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_132E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",132.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2620]]",
        2621=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_135E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2621]]",
        2622=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_138E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",138.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2622]]",
        2623=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_141E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2623]]",
        2624=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_144E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",144.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2624]]",
        2625=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_147E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2625]]",
        2626=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_150E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",150.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2626]]",
        2627=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_153E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2627]]",
        2628=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_156E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",156.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2628]]",
        2629=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_159E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2629]]",
        2630=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_162E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",162.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2630]]",
        2631=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_165E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2631]]",
        2632=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_168E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",168.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2632]]",
        2633=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_171E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2633]]",
        2634=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_174E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",174.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2634]]",
        2635=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_177E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2635]]",
        2636=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_180E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",180.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2636]]",
        2637=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_177W\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2637]]",
        2638=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_174W\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-174.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2638]]",
        2639=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_171W\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2639]]",
        2640=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_168W\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-168.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2640]]",
        2641=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_7\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",7500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2641]]",
        2642=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_8\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",8500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2642]]",
        2643=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_9\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",9500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2643]]",
        2644=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_10\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",10500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2644]]",
        2645=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_11\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",11500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2645]]",
        2646=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_12\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",12500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",36.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2646]]",
        2647=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_13\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",13500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2647]]",
        2648=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_14\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",14500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",42.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2648]]",
        2649=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_15\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",15500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2649]]",
        2650=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_16\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",16500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",48.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2650]]",
        2651=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_17\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",17500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2651]]",
        2652=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_18\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",18500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",54.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2652]]",
        2653=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_19\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",19500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2653]]",
        2654=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_20\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",20500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",60.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2654]]",
        2655=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_21\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",21500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2655]]",
        2656=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_22\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",22500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",66.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2656]]",
        2657=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_23\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",23500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2657]]",
        2658=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_24\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",24500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",72.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2658]]",
        2659=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_25\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",25500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2659]]",
        2660=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_26\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",26500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",78.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2660]]",
        2661=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_27\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",27500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2661]]",
        2662=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_28\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",28500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",84.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2662]]",
        2663=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_29\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",29500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2663]]",
        2664=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_30\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",30500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2664]]",
        2665=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_31\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",31500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2665]]",
        2666=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_32\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",32500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",96.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2666]]",
        2667=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_33\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",33500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2667]]",
        2668=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_34\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",34500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",102.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2668]]",
        2669=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_35\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",35500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2669]]",
        2670=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_36\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",36500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",108.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2670]]",
        2671=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_37\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",37500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2671]]",
        2672=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_38\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",38500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",114.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2672]]",
        2673=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_39\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",39500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2673]]",
        2674=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_40\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",40500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",120.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2674]]",
        2675=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_41\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",41500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2675]]",
        2676=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_42\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",42500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",126.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2676]]",
        2677=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_43\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",43500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2677]]",
        2678=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_44\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",44500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",132.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2678]]",
        2679=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_45\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",45500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2679]]",
        2680=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_46\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",46500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",138.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2680]]",
        2681=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_47\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",47500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2681]]",
        2682=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_48\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",48500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",144.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2682]]",
        2683=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_49\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",49500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2683]]",
        2684=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_50\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",50500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",150.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2684]]",
        2685=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_51\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",51500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2685]]",
        2686=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_52\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",52500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",156.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2686]]",
        2687=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_53\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",53500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2687]]",
        2688=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_54\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",54500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",162.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2688]]",
        2689=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_55\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",55500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2689]]",
        2690=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_56\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",56500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",168.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2690]]",
        2691=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_57\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",57500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2691]]",
        2692=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_58\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",58500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",174.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2692]]",
        2693=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_59\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",59500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2693]]",
        2694=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_60\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",60500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",180.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3390]]",
        2695=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_61\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",61500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2695]]",
        2696=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_62\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",62500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-174.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2696]]",
        2697=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_63\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",63500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2697]]",
        2698=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_64\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",64500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-168.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2698]]",
        2699=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_21E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2699]]",
        2700=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_24E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2700]]",
        2701=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_27E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2701]]",
        2702=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_30E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2702]]",
        2703=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_33E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2703]]",
        2704=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_36E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",36.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2704]]",
        2705=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_39E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2705]]",
        2706=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_42E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",42.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2706]]",
        2707=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_45E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2707]]",
        2708=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_48E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",48.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2708]]",
        2709=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_51E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2709]]",
        2710=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_54E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",54.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2710]]",
        2711=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_57E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2711]]",
        2712=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_60E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",60.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2712]]",
        2713=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_63E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2713]]",
        2714=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_66E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",66.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2714]]",
        2715=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_69E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2715]]",
        2716=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_72E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",72.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2716]]",
        2717=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_75E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2717]]",
        2718=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_78E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",78.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2718]]",
        2719=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_81E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2719]]",
        2720=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_84E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",84.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2720]]",
        2721=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_87E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2721]]",
        2722=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_90E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2722]]",
        2723=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_93E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2723]]",
        2724=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_96E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",96.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2724]]",
        2725=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_99E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2725]]",
        2726=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_102E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",102.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2726]]",
        2727=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_105E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2727]]",
        2728=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_108E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",108.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2728]]",
        2729=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_111E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2729]]",
        2730=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_114E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",114.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2730]]",
        2731=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_117E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2731]]",
        2732=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_120E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",120.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2732]]",
        2733=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_123E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2733]]",
        2734=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_126E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",126.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2734]]",
        2735=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_129E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2735]]",
        2736=>"PROJCS[\"Tete_UTM_Zone_36S\",GEOGCS[\"GCS_Tete\",DATUM[\"D_Tete\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2736]]",
        2737=>"PROJCS[\"Tete_UTM_Zone_37S\",GEOGCS[\"GCS_Tete\",DATUM[\"D_Tete\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2737]]",
        2738=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_132E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",132.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2738]]",
        2739=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_135E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2739]]",
        2740=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_138E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",138.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2740]]",
        2741=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_141E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2741]]",
        2742=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_144E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",144.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2742]]",
        2743=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_147E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2743]]",
        2744=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_150E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",150.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2744]]",
        2745=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_153E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2745]]",
        2746=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_156E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",156.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2746]]",
        2747=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_159E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2747]]",
        2748=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_162E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",162.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2748]]",
        2749=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_165E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2749]]",
        2750=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_168E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",168.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2750]]",
        2751=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_171E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2751]]",
        2752=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_174E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",174.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2752]]",
        2753=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_177E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2753]]",
        2754=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_180E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",180.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2754]]",
        2755=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_177W\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2755]]",
        2756=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_174W\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-174.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2756]]",
        2757=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_171W\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2757]]",
        2758=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_168W\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-168.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2758]]",
        2759=>"PROJCS[\"NAD_1983_HARN_StatePlane_Alabama_East_FIPS_0101\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-85.83333333333333],PARAMETER[\"Scale_Factor\",0.99996],PARAMETER[\"Latitude_Of_Origin\",30.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2759]]",
        2760=>"PROJCS[\"NAD_1983_HARN_StatePlane_Alabama_West_FIPS_0102\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.5],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2760]]",
        2761=>"PROJCS[\"NAD_1983_HARN_StatePlane_Arizona_East_FIPS_0201\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",213360.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-110.1666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2761]]",
        2762=>"PROJCS[\"NAD_1983_HARN_StatePlane_Arizona_Central_FIPS_0202\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",213360.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.9166666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2762]]",
        2763=>"PROJCS[\"NAD_1983_HARN_StatePlane_Arizona_West_FIPS_0203\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",213360.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-113.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2763]]",
        2764=>"PROJCS[\"NAD_1983_HARN_StatePlane_Arkansas_North_FIPS_0301\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",34.93333333333333],PARAMETER[\"Standard_Parallel_2\",36.23333333333333],PARAMETER[\"Latitude_Of_Origin\",34.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2764]]",
        2765=>"PROJCS[\"NAD_1983_HARN_StatePlane_Arkansas_South_FIPS_0302\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",400000.0],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",33.3],PARAMETER[\"Standard_Parallel_2\",34.76666666666667],PARAMETER[\"Latitude_Of_Origin\",32.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2765]]",
        2766=>"PROJCS[\"NAD_1983_HARN_StatePlane_California_I_FIPS_0401\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Standard_Parallel_2\",41.66666666666666],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2766]]",
        2767=>"PROJCS[\"NAD_1983_HARN_StatePlane_California_II_FIPS_0402\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",38.33333333333334],PARAMETER[\"Standard_Parallel_2\",39.83333333333334],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2767]]",
        2768=>"PROJCS[\"NAD_1983_HARN_StatePlane_California_III_FIPS_0403\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",37.06666666666667],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2768]]",
        2769=>"PROJCS[\"NAD_1983_HARN_StatePlane_California_IV_FIPS_0404\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-119.0],PARAMETER[\"Standard_Parallel_1\",36.0],PARAMETER[\"Standard_Parallel_2\",37.25],PARAMETER[\"Latitude_Of_Origin\",35.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2769]]",
        2770=>"PROJCS[\"NAD_1983_HARN_StatePlane_California_V_FIPS_0405\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-118.0],PARAMETER[\"Standard_Parallel_1\",34.03333333333333],PARAMETER[\"Standard_Parallel_2\",35.46666666666667],PARAMETER[\"Latitude_Of_Origin\",33.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2770]]",
        2771=>"PROJCS[\"NAD_1983_HARN_StatePlane_California_VI_FIPS_0406\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-116.25],PARAMETER[\"Standard_Parallel_1\",32.78333333333333],PARAMETER[\"Standard_Parallel_2\",33.88333333333333],PARAMETER[\"Latitude_Of_Origin\",32.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2771]]",
        2772=>"PROJCS[\"NAD_1983_HARN_StatePlane_Colorado_North_FIPS_0501\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",914401.8289],PARAMETER[\"False_Northing\",304800.6096],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",39.71666666666667],PARAMETER[\"Standard_Parallel_2\",40.78333333333333],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2772]]",
        2773=>"PROJCS[\"NAD_1983_HARN_StatePlane_Colorado_Central_FIPS_0502\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",914401.8289],PARAMETER[\"False_Northing\",304800.6096],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",38.45],PARAMETER[\"Standard_Parallel_2\",39.75],PARAMETER[\"Latitude_Of_Origin\",37.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2773]]",
        2774=>"PROJCS[\"NAD_1983_HARN_StatePlane_Colorado_South_FIPS_0503\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",914401.8289],PARAMETER[\"False_Northing\",304800.6096],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",37.23333333333333],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2774]]",
        2775=>"PROJCS[\"NAD_1983_HARN_StatePlane_Connecticut_FIPS_0600\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",304800.6096],PARAMETER[\"False_Northing\",152400.3048],PARAMETER[\"Central_Meridian\",-72.75],PARAMETER[\"Standard_Parallel_1\",41.2],PARAMETER[\"Standard_Parallel_2\",41.86666666666667],PARAMETER[\"Latitude_Of_Origin\",40.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2775]]",
        2776=>"PROJCS[\"NAD_1983_HARN_StatePlane_Delaware_FIPS_0700\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.41666666666667],PARAMETER[\"Scale_Factor\",0.999995],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2776]]",
        2777=>"PROJCS[\"NAD_1983_HARN_StatePlane_Florida_East_FIPS_0901\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2777]]",
        2778=>"PROJCS[\"NAD_1983_HARN_StatePlane_Florida_West_FIPS_0902\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2778]]",
        2779=>"PROJCS[\"NAD_1983_HARN_StatePlane_Florida_North_FIPS_0903\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.5],PARAMETER[\"Standard_Parallel_1\",29.58333333333333],PARAMETER[\"Standard_Parallel_2\",30.75],PARAMETER[\"Latitude_Of_Origin\",29.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2779]]",
        2780=>"PROJCS[\"NAD_1983_HARN_StatePlane_Georgia_East_FIPS_1001\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2780]]",
        2781=>"PROJCS[\"NAD_1983_HARN_StatePlane_Georgia_West_FIPS_1002\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2781]]",
        2782=>"PROJCS[\"NAD_1983_HARN_StatePlane_Hawaii_1_FIPS_5101\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-155.5],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",18.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2782]]",
        2783=>"PROJCS[\"NAD_1983_HARN_StatePlane_Hawaii_2_FIPS_5102\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-156.6666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",20.33333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2783]]",
        2784=>"PROJCS[\"NAD_1983_HARN_StatePlane_Hawaii_3_FIPS_5103\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-158.0],PARAMETER[\"Scale_Factor\",0.99999],PARAMETER[\"Latitude_Of_Origin\",21.16666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2784]]",
        2785=>"PROJCS[\"NAD_1983_HARN_StatePlane_Hawaii_4_FIPS_5104\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-159.5],PARAMETER[\"Scale_Factor\",0.99999],PARAMETER[\"Latitude_Of_Origin\",21.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2785]]",
        2786=>"PROJCS[\"NAD_1983_HARN_StatePlane_Hawaii_5_FIPS_5105\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-160.1666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",21.66666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2786]]",
        2787=>"PROJCS[\"NAD_1983_HARN_StatePlane_Idaho_East_FIPS_1101\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-112.1666666666667],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2787]]",
        2788=>"PROJCS[\"NAD_1983_HARN_StatePlane_Idaho_Central_FIPS_1102\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-114.0],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2788]]",
        2789=>"PROJCS[\"NAD_1983_HARN_StatePlane_Idaho_West_FIPS_1103\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-115.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2789]]",
        2790=>"PROJCS[\"NAD_1983_HARN_StatePlane_Illinois_East_FIPS_1201\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.33333333333333],PARAMETER[\"Scale_Factor\",0.999975],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2790]]",
        2791=>"PROJCS[\"NAD_1983_HARN_StatePlane_Illinois_West_FIPS_1202\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.16666666666667],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2791]]",
        2792=>"PROJCS[\"NAD_1983_HARN_StatePlane_Indiana_East_FIPS_1301\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",250000.0],PARAMETER[\"Central_Meridian\",-85.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2792]]",
        2793=>"PROJCS[\"NAD_1983_HARN_StatePlane_Indiana_West_FIPS_1302\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",900000.0],PARAMETER[\"False_Northing\",250000.0],PARAMETER[\"Central_Meridian\",-87.08333333333333],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2793]]",
        2794=>"PROJCS[\"NAD_1983_HARN_StatePlane_Iowa_North_FIPS_1401\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",42.06666666666667],PARAMETER[\"Standard_Parallel_2\",43.26666666666667],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2794]]",
        2795=>"PROJCS[\"NAD_1983_HARN_StatePlane_Iowa_South_FIPS_1402\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",40.61666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2795]]",
        2796=>"PROJCS[\"NAD_1983_HARN_StatePlane_Kansas_North_FIPS_1501\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",38.71666666666667],PARAMETER[\"Standard_Parallel_2\",39.78333333333333],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2796]]",
        2797=>"PROJCS[\"NAD_1983_HARN_StatePlane_Kansas_South_FIPS_1502\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",400000.0],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",37.26666666666667],PARAMETER[\"Standard_Parallel_2\",38.56666666666667],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2797]]",
        2798=>"PROJCS[\"NAD_1983_HARN_StatePlane_Kentucky_North_FIPS_1601\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.25],PARAMETER[\"Standard_Parallel_1\",37.96666666666667],PARAMETER[\"Standard_Parallel_2\",38.96666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2798]]",
        2799=>"PROJCS[\"NAD_1983_HARN_StatePlane_Kentucky_South_FIPS_1602\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Standard_Parallel_1\",36.73333333333333],PARAMETER[\"Standard_Parallel_2\",37.93333333333333],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2799]]",
        2800=>"PROJCS[\"NAD_1983_HARN_StatePlane_Louisiana_North_FIPS_1701\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.5],PARAMETER[\"Standard_Parallel_1\",31.16666666666667],PARAMETER[\"Standard_Parallel_2\",32.66666666666666],PARAMETER[\"Latitude_Of_Origin\",30.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2800]]",
        2801=>"PROJCS[\"NAD_1983_HARN_StatePlane_Louisiana_South_FIPS_1702\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-91.33333333333333],PARAMETER[\"Standard_Parallel_1\",29.3],PARAMETER[\"Standard_Parallel_2\",30.7],PARAMETER[\"Latitude_Of_Origin\",28.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2801]]",
        2802=>"PROJCS[\"NAD_1983_HARN_StatePlane_Maine_East_FIPS_1801\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-68.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2802]]",
        2803=>"PROJCS[\"NAD_1983_HARN_StatePlane_Maine_West_FIPS_1802\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",900000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.16666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2803]]",
        2804=>"PROJCS[\"NAD_1983_HARN_StatePlane_Maryland_FIPS_1900\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.0],PARAMETER[\"Standard_Parallel_1\",38.3],PARAMETER[\"Standard_Parallel_2\",39.45],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2804]]",
        2805=>"PROJCS[\"NAD_1983_HARN_StatePlane_Massachusetts_Mainland_FIPS_2001\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",750000.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Standard_Parallel_1\",41.71666666666667],PARAMETER[\"Standard_Parallel_2\",42.68333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2805]]",
        2806=>"PROJCS[\"NAD_1983_HARN_StatePlane_Massachusetts_Island_FIPS_2002\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.5],PARAMETER[\"Standard_Parallel_1\",41.28333333333333],PARAMETER[\"Standard_Parallel_2\",41.48333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2806]]",
        2807=>"PROJCS[\"NAD_1983_HARN_StatePlane_Michigan_North_FIPS_2111\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",8000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Standard_Parallel_1\",45.48333333333333],PARAMETER[\"Standard_Parallel_2\",47.08333333333334],PARAMETER[\"Latitude_Of_Origin\",44.78333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2807]]",
        2808=>"PROJCS[\"NAD_1983_HARN_StatePlane_Michigan_Central_FIPS_2112\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.36666666666666],PARAMETER[\"Standard_Parallel_1\",44.18333333333333],PARAMETER[\"Standard_Parallel_2\",45.7],PARAMETER[\"Latitude_Of_Origin\",43.31666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2808]]",
        2809=>"PROJCS[\"NAD_1983_HARN_StatePlane_Michigan_South_FIPS_2113\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.36666666666666],PARAMETER[\"Standard_Parallel_1\",42.1],PARAMETER[\"Standard_Parallel_2\",43.66666666666666],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2809]]",
        2810=>"PROJCS[\"NAD_1983_HARN_StatePlane_Minnesota_North_FIPS_2201\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-93.1],PARAMETER[\"Standard_Parallel_1\",47.03333333333333],PARAMETER[\"Standard_Parallel_2\",48.63333333333333],PARAMETER[\"Latitude_Of_Origin\",46.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2810]]",
        2811=>"PROJCS[\"NAD_1983_HARN_StatePlane_Minnesota_Central_FIPS_2202\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-94.25],PARAMETER[\"Standard_Parallel_1\",45.61666666666667],PARAMETER[\"Standard_Parallel_2\",47.05],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2811]]",
        2812=>"PROJCS[\"NAD_1983_HARN_StatePlane_Minnesota_South_FIPS_2203\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-94.0],PARAMETER[\"Standard_Parallel_1\",43.78333333333333],PARAMETER[\"Standard_Parallel_2\",45.21666666666667],PARAMETER[\"Latitude_Of_Origin\",43.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2812]]",
        2813=>"PROJCS[\"NAD_1983_HARN_StatePlane_Mississippi_East_FIPS_2301\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.83333333333333],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",29.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2813]]",
        2814=>"PROJCS[\"NAD_1983_HARN_StatePlane_Mississippi_West_FIPS_2302\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.33333333333333],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",29.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2814]]",
        2815=>"PROJCS[\"NAD_1983_HARN_StatePlane_Missouri_East_FIPS_2401\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.5],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",35.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2815]]",
        2816=>"PROJCS[\"NAD_1983_HARN_StatePlane_Missouri_Central_FIPS_2402\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.5],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",35.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2816]]",
        2817=>"PROJCS[\"NAD_1983_HARN_StatePlane_Missouri_West_FIPS_2403\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",850000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-94.5],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",36.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2817]]",
        2818=>"PROJCS[\"NAD_1983_HARN_StatePlane_Montana_FIPS_2500\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-109.5],PARAMETER[\"Standard_Parallel_1\",45.0],PARAMETER[\"Standard_Parallel_2\",49.0],PARAMETER[\"Latitude_Of_Origin\",44.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2818]]",
        2819=>"PROJCS[\"NAD_1983_HARN_StatePlane_Nebraska_FIPS_2600\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Standard_Parallel_2\",43.0],PARAMETER[\"Latitude_Of_Origin\",39.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2819]]",
        2820=>"PROJCS[\"NAD_1983_HARN_StatePlane_Nevada_East_FIPS_2701\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",8000000.0],PARAMETER[\"Central_Meridian\",-115.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2820]]",
        2821=>"PROJCS[\"NAD_1983_HARN_StatePlane_Nevada_Central_FIPS_2702\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",6000000.0],PARAMETER[\"Central_Meridian\",-116.6666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2821]]",
        2822=>"PROJCS[\"NAD_1983_HARN_StatePlane_Nevada_West_FIPS_2703\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",4000000.0],PARAMETER[\"Central_Meridian\",-118.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2822]]",
        2823=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_Hampshire_FIPS_2800\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2823]]",
        2824=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_Jersey_FIPS_2900\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",38.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2824]]",
        2825=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_Mexico_East_FIPS_3001\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",165000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-104.3333333333333],PARAMETER[\"Scale_Factor\",0.9999090909090909],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2825]]",
        2826=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_Mexico_Central_FIPS_3002\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-106.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2826]]",
        2827=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_Mexico_West_FIPS_3003\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",830000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-107.8333333333333],PARAMETER[\"Scale_Factor\",0.9999166666666667],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2827]]",
        2828=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_York_East_FIPS_3101\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",38.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2828]]",
        2829=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_York_Central_FIPS_3102\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-76.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2829]]",
        2830=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_York_West_FIPS_3103\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",350000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-78.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2830]]",
        2831=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_York_Long_Island_FIPS_3104\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.0],PARAMETER[\"Standard_Parallel_1\",40.66666666666666],PARAMETER[\"Standard_Parallel_2\",41.03333333333333],PARAMETER[\"Latitude_Of_Origin\",40.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2831]]",
        2832=>"PROJCS[\"NAD_1983_HARN_StatePlane_North_Dakota_North_FIPS_3301\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",47.43333333333333],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2832]]",
        2833=>"PROJCS[\"NAD_1983_HARN_StatePlane_North_Dakota_South_FIPS_3302\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",46.18333333333333],PARAMETER[\"Standard_Parallel_2\",47.48333333333333],PARAMETER[\"Latitude_Of_Origin\",45.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2833]]",
        2834=>"PROJCS[\"NAD_1983_HARN_StatePlane_Ohio_North_FIPS_3401\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",40.43333333333333],PARAMETER[\"Standard_Parallel_2\",41.7],PARAMETER[\"Latitude_Of_Origin\",39.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2834]]",
        2835=>"PROJCS[\"NAD_1983_HARN_StatePlane_Ohio_South_FIPS_3402\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",38.73333333333333],PARAMETER[\"Standard_Parallel_2\",40.03333333333333],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2835]]",
        2836=>"PROJCS[\"NAD_1983_HARN_StatePlane_Oklahoma_North_FIPS_3501\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",35.56666666666667],PARAMETER[\"Standard_Parallel_2\",36.76666666666667],PARAMETER[\"Latitude_Of_Origin\",35.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2836]]",
        2837=>"PROJCS[\"NAD_1983_HARN_StatePlane_Oklahoma_South_FIPS_3502\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",33.93333333333333],PARAMETER[\"Standard_Parallel_2\",35.23333333333333],PARAMETER[\"Latitude_Of_Origin\",33.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2837]]",
        2838=>"PROJCS[\"NAD_1983_HARN_StatePlane_Oregon_North_FIPS_3601\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",44.33333333333334],PARAMETER[\"Standard_Parallel_2\",46.0],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2838]]",
        2839=>"PROJCS[\"NAD_1983_HARN_StatePlane_Oregon_South_FIPS_3602\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",42.33333333333334],PARAMETER[\"Standard_Parallel_2\",44.0],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2839]]",
        2840=>"PROJCS[\"NAD_1983_HARN_StatePlane_Rhode_Island_FIPS_3800\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Scale_Factor\",0.99999375],PARAMETER[\"Latitude_Of_Origin\",41.08333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2840]]",
        2841=>"PROJCS[\"NAD_1983_HARN_StatePlane_South_Dakota_North_FIPS_4001\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",44.41666666666666],PARAMETER[\"Standard_Parallel_2\",45.68333333333333],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2841]]",
        2842=>"PROJCS[\"NAD_1983_HARN_StatePlane_South_Dakota_South_FIPS_4002\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",42.83333333333334],PARAMETER[\"Standard_Parallel_2\",44.4],PARAMETER[\"Latitude_Of_Origin\",42.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2842]]",
        2843=>"PROJCS[\"NAD_1983_HARN_StatePlane_Tennessee_FIPS_4100\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-86.0],PARAMETER[\"Standard_Parallel_1\",35.25],PARAMETER[\"Standard_Parallel_2\",36.41666666666666],PARAMETER[\"Latitude_Of_Origin\",34.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2843]]",
        2844=>"PROJCS[\"NAD_1983_HARN_StatePlane_Texas_North_FIPS_4201\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-101.5],PARAMETER[\"Standard_Parallel_1\",34.65],PARAMETER[\"Standard_Parallel_2\",36.18333333333333],PARAMETER[\"Latitude_Of_Origin\",34.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2844]]",
        2845=>"PROJCS[\"NAD_1983_HARN_StatePlane_Texas_North_Central_FIPS_4202\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",32.13333333333333],PARAMETER[\"Standard_Parallel_2\",33.96666666666667],PARAMETER[\"Latitude_Of_Origin\",31.66666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2845]]",
        2846=>"PROJCS[\"NAD_1983_HARN_StatePlane_Texas_Central_FIPS_4203\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",3000000.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",30.11666666666667],PARAMETER[\"Standard_Parallel_2\",31.88333333333333],PARAMETER[\"Latitude_Of_Origin\",29.66666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2846]]",
        2847=>"PROJCS[\"NAD_1983_HARN_StatePlane_Texas_South_Central_FIPS_4204\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",4000000.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Standard_Parallel_1\",28.38333333333333],PARAMETER[\"Standard_Parallel_2\",30.28333333333333],PARAMETER[\"Latitude_Of_Origin\",27.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2847]]",
        2848=>"PROJCS[\"NAD_1983_HARN_StatePlane_Texas_South_FIPS_4205\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",5000000.0],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",26.16666666666667],PARAMETER[\"Standard_Parallel_2\",27.83333333333333],PARAMETER[\"Latitude_Of_Origin\",25.66666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2848]]",
        2849=>"PROJCS[\"NAD_1983_HARN_StatePlane_Utah_North_FIPS_4301\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",40.71666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2849]]",
        2850=>"PROJCS[\"NAD_1983_HARN_StatePlane_Utah_Central_FIPS_4302\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",39.01666666666667],PARAMETER[\"Standard_Parallel_2\",40.65],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2850]]",
        2851=>"PROJCS[\"NAD_1983_HARN_StatePlane_Utah_South_FIPS_4303\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",3000000.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",37.21666666666667],PARAMETER[\"Standard_Parallel_2\",38.35],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2851]]",
        2852=>"PROJCS[\"NAD_1983_HARN_StatePlane_Vermont_FIPS_4400\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-72.5],PARAMETER[\"Scale_Factor\",0.9999642857142857],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2852]]",
        2853=>"PROJCS[\"NAD_1983_HARN_StatePlane_Virginia_North_FIPS_4501\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",38.03333333333333],PARAMETER[\"Standard_Parallel_2\",39.2],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2853]]",
        2854=>"PROJCS[\"NAD_1983_HARN_StatePlane_Virginia_South_FIPS_4502\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",36.76666666666667],PARAMETER[\"Standard_Parallel_2\",37.96666666666667],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2854]]",
        2855=>"PROJCS[\"NAD_1983_HARN_StatePlane_Washington_North_FIPS_4601\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.8333333333333],PARAMETER[\"Standard_Parallel_1\",47.5],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2855]]",
        2856=>"PROJCS[\"NAD_1983_HARN_StatePlane_Washington_South_FIPS_4602\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",45.83333333333334],PARAMETER[\"Standard_Parallel_2\",47.33333333333334],PARAMETER[\"Latitude_Of_Origin\",45.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2856]]",
        2857=>"PROJCS[\"NAD_1983_HARN_StatePlane_West_Virginia_North_FIPS_4701\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Standard_Parallel_1\",39.0],PARAMETER[\"Standard_Parallel_2\",40.25],PARAMETER[\"Latitude_Of_Origin\",38.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2857]]",
        2858=>"PROJCS[\"NAD_1983_HARN_StatePlane_West_Virginia_South_FIPS_4702\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",37.48333333333333],PARAMETER[\"Standard_Parallel_2\",38.88333333333333],PARAMETER[\"Latitude_Of_Origin\",37.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2858]]",
        2859=>"PROJCS[\"NAD_1983_HARN_StatePlane_Wisconsin_North_FIPS_4801\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",45.56666666666667],PARAMETER[\"Standard_Parallel_2\",46.76666666666667],PARAMETER[\"Latitude_Of_Origin\",45.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2859]]",
        2860=>"PROJCS[\"NAD_1983_HARN_StatePlane_Wisconsin_Central_FIPS_4802\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",44.25],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2860]]",
        2861=>"PROJCS[\"NAD_1983_HARN_StatePlane_Wisconsin_South_FIPS_4803\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",42.73333333333333],PARAMETER[\"Standard_Parallel_2\",44.06666666666667],PARAMETER[\"Latitude_Of_Origin\",42.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2861]]",
        2862=>"PROJCS[\"NAD_1983_HARN_StatePlane_Wyoming_East_FIPS_4901\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.1666666666667],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2862]]",
        2863=>"PROJCS[\"NAD_1983_HARN_StatePlane_Wyoming_East_Central_FIPS_4902\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-107.3333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2863]]",
        2864=>"PROJCS[\"NAD_1983_HARN_StatePlane_Wyoming_West_Central_FIPS_4903\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-108.75],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2864]]",
        2865=>"PROJCS[\"NAD_1983_HARN_StatePlane_Wyoming_West_FIPS_4904\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-110.0833333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2865]]",
        2866=>"PROJCS[\"NAD_1983_HARN_StatePlane_Puerto_Rico_Virgin_Islands_FIPS_5200\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Central_Meridian\",-66.43333333333334],PARAMETER[\"Standard_Parallel_1\",18.03333333333333],PARAMETER[\"Standard_Parallel_2\",18.43333333333333],PARAMETER[\"Latitude_Of_Origin\",17.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2866]]",
        2867=>"PROJCS[\"NAD_1983_HARN_StatePlane_Arizona_East_FIPS_0201_Feet_Intl\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-110.1666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2867]]",
        2868=>"PROJCS[\"NAD_1983_HARN_StatePlane_Arizona_Central_FIPS_0202_Feet_Intl\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.9166666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2868]]",
        2869=>"PROJCS[\"NAD_1983_HARN_StatePlane_Arizona_West_FIPS_0203_Feet_Intl\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-113.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2869]]",
        2870=>"PROJCS[\"NAD_1983_HARN_StatePlane_California_I_FIPS_0401_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Standard_Parallel_2\",41.66666666666666],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2870]]",
        2871=>"PROJCS[\"NAD_1983_HARN_StatePlane_California_II_FIPS_0402_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",38.33333333333334],PARAMETER[\"Standard_Parallel_2\",39.83333333333334],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2871]]",
        2872=>"PROJCS[\"NAD_1983_HARN_StatePlane_California_III_FIPS_0403_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",37.06666666666667],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2872]]",
        2873=>"PROJCS[\"NAD_1983_HARN_StatePlane_California_IV_FIPS_0404_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-119.0],PARAMETER[\"Standard_Parallel_1\",36.0],PARAMETER[\"Standard_Parallel_2\",37.25],PARAMETER[\"Latitude_Of_Origin\",35.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2873]]",
        2874=>"PROJCS[\"NAD_1983_HARN_StatePlane_California_V_FIPS_0405_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-118.0],PARAMETER[\"Standard_Parallel_1\",34.03333333333333],PARAMETER[\"Standard_Parallel_2\",35.46666666666667],PARAMETER[\"Latitude_Of_Origin\",33.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2874]]",
        2875=>"PROJCS[\"NAD_1983_HARN_StatePlane_California_VI_FIPS_0406_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-116.25],PARAMETER[\"Standard_Parallel_1\",32.78333333333333],PARAMETER[\"Standard_Parallel_2\",33.88333333333333],PARAMETER[\"Latitude_Of_Origin\",32.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2875]]",
        2876=>"PROJCS[\"NAD_1983_HARN_StatePlane_Colorado_North_FIPS_0501_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.000316083],PARAMETER[\"False_Northing\",999999.999996],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",39.71666666666667],PARAMETER[\"Standard_Parallel_2\",40.78333333333333],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2876]]",
        2877=>"PROJCS[\"NAD_1983_HARN_StatePlane_Colorado_Central_FIPS_0502_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.000316083],PARAMETER[\"False_Northing\",999999.999996],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",38.45],PARAMETER[\"Standard_Parallel_2\",39.75],PARAMETER[\"Latitude_Of_Origin\",37.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2877]]",
        2878=>"PROJCS[\"NAD_1983_HARN_StatePlane_Colorado_South_FIPS_0503_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.000316083],PARAMETER[\"False_Northing\",999999.999996],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",37.23333333333333],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2878]]",
        2879=>"PROJCS[\"NAD_1983_HARN_StatePlane_Connecticut_FIPS_0600_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",999999.999996],PARAMETER[\"False_Northing\",499999.999998],PARAMETER[\"Central_Meridian\",-72.75],PARAMETER[\"Standard_Parallel_1\",41.2],PARAMETER[\"Standard_Parallel_2\",41.86666666666667],PARAMETER[\"Latitude_Of_Origin\",40.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2879]]",
        2880=>"PROJCS[\"NAD_1983_HARN_StatePlane_Delaware_FIPS_0700_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.41666666666667],PARAMETER[\"Scale_Factor\",0.999995],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2880]]",
        2881=>"PROJCS[\"NAD_1983_HARN_StatePlane_Florida_East_FIPS_0901_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2881]]",
        2882=>"PROJCS[\"NAD_1983_HARN_StatePlane_Florida_West_FIPS_0902_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2882]]",
        2883=>"PROJCS[\"NAD_1983_HARN_StatePlane_Florida_North_FIPS_0903_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.5],PARAMETER[\"Standard_Parallel_1\",29.58333333333333],PARAMETER[\"Standard_Parallel_2\",30.75],PARAMETER[\"Latitude_Of_Origin\",29.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2883]]",
        2884=>"PROJCS[\"NAD_1983_HARN_StatePlane_Georgia_East_FIPS_1001_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2884]]",
        2885=>"PROJCS[\"NAD_1983_HARN_StatePlane_Georgia_West_FIPS_1002_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2296583.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2885]]",
        2886=>"PROJCS[\"NAD_1983_HARN_StatePlane_Idaho_East_FIPS_1101_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-112.1666666666667],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2886]]",
        2887=>"PROJCS[\"NAD_1983_HARN_StatePlane_Idaho_Central_FIPS_1102_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-114.0],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2887]]",
        2888=>"PROJCS[\"NAD_1983_HARN_StatePlane_Idaho_West_FIPS_1103_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-115.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2888]]",
        2891=>"PROJCS[\"NAD_1983_HARN_StatePlane_Kentucky_North_FIPS_1601_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.25],PARAMETER[\"Standard_Parallel_1\",37.96666666666667],PARAMETER[\"Standard_Parallel_2\",38.96666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2891]]",
        2892=>"PROJCS[\"NAD_1983_HARN_StatePlane_Kentucky_South_FIPS_1602_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Standard_Parallel_1\",36.73333333333333],PARAMETER[\"Standard_Parallel_2\",37.93333333333333],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2892]]",
        2893=>"PROJCS[\"NAD_1983_HARN_StatePlane_Maryland_FIPS_1900_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.0],PARAMETER[\"Standard_Parallel_1\",38.3],PARAMETER[\"Standard_Parallel_2\",39.45],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2893]]",
        2894=>"PROJCS[\"NAD_1983_HARN_StatePlane_Massachusetts_Mainland_FIPS_2001_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",2460625.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Standard_Parallel_1\",41.71666666666667],PARAMETER[\"Standard_Parallel_2\",42.68333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2894]]",
        2895=>"PROJCS[\"NAD_1983_HARN_StatePlane_Massachusetts_Island_FIPS_2002_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.5],PARAMETER[\"Standard_Parallel_1\",41.28333333333333],PARAMETER[\"Standard_Parallel_2\",41.48333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2895]]",
        2896=>"PROJCS[\"NAD_1983_HARN_StatePlane_Michigan_North_FIPS_2111_Feet_Intl\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",26246719.16010498],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Standard_Parallel_1\",45.48333333333333],PARAMETER[\"Standard_Parallel_2\",47.08333333333334],PARAMETER[\"Latitude_Of_Origin\",44.78333333333333],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2896]]",
        2897=>"PROJCS[\"NAD_1983_HARN_StatePlane_Michigan_Central_FIPS_2112_Feet_Intl\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",19685039.37007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.36666666666666],PARAMETER[\"Standard_Parallel_1\",44.18333333333333],PARAMETER[\"Standard_Parallel_2\",45.7],PARAMETER[\"Latitude_Of_Origin\",43.31666666666667],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2897]]",
        2898=>"PROJCS[\"NAD_1983_HARN_StatePlane_Michigan_South_FIPS_2113_Feet_Intl\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",13123359.58005249],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.36666666666666],PARAMETER[\"Standard_Parallel_1\",42.1],PARAMETER[\"Standard_Parallel_2\",43.66666666666666],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2898]]",
        2899=>"PROJCS[\"NAD_1983_HARN_StatePlane_Mississippi_East_FIPS_2301_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.83333333333333],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",29.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2899]]",
        2900=>"PROJCS[\"NAD_1983_HARN_StatePlane_Mississippi_West_FIPS_2302_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2296583.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.33333333333333],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",29.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2900]]",
        2901=>"PROJCS[\"NAD_1983_HARN_StatePlane_Montana_FIPS_2500_Feet_Intl\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968503.937007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-109.5],PARAMETER[\"Standard_Parallel_1\",45.0],PARAMETER[\"Standard_Parallel_2\",49.0],PARAMETER[\"Latitude_Of_Origin\",44.25],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2901]]",
        2902=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_Mexico_East_FIPS_3001_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",541337.5],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-104.3333333333333],PARAMETER[\"Scale_Factor\",0.9999090909090909],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2902]]",
        2903=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_Mexico_Central_FIPS_3002_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-106.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2903]]",
        2904=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_Mexico_West_FIPS_3003_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2723091.666666666],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-107.8333333333333],PARAMETER[\"Scale_Factor\",0.9999166666666667],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2904]]",
        2905=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_York_East_FIPS_3101_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",492125.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",38.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2905]]",
        2906=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_York_Central_FIPS_3102_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",820208.3333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-76.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2906]]",
        2907=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_York_West_FIPS_3103_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1148291.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-78.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2907]]",
        2908=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_York_Long_Island_FIPS_3104_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.0],PARAMETER[\"Standard_Parallel_1\",40.66666666666666],PARAMETER[\"Standard_Parallel_2\",41.03333333333333],PARAMETER[\"Latitude_Of_Origin\",40.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2908]]",
        2909=>"PROJCS[\"NAD_1983_HARN_StatePlane_North_Dakota_North_FIPS_3301_Feet_Intl\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968503.937007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",47.43333333333333],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2909]]",
        2910=>"PROJCS[\"NAD_1983_HARN_StatePlane_North_Dakota_South_FIPS_3302_Feet_Intl\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968503.937007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",46.18333333333333],PARAMETER[\"Standard_Parallel_2\",47.48333333333333],PARAMETER[\"Latitude_Of_Origin\",45.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2910]]",
        2911=>"PROJCS[\"NAD_1983_HARN_StatePlane_Oklahoma_North_FIPS_3501_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",35.56666666666667],PARAMETER[\"Standard_Parallel_2\",36.76666666666667],PARAMETER[\"Latitude_Of_Origin\",35.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2911]]",
        2912=>"PROJCS[\"NAD_1983_HARN_StatePlane_Oklahoma_South_FIPS_3502_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",33.93333333333333],PARAMETER[\"Standard_Parallel_2\",35.23333333333333],PARAMETER[\"Latitude_Of_Origin\",33.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2912]]",
        2913=>"PROJCS[\"NAD_1983_HARN_StatePlane_Oregon_North_FIPS_3601_Feet_Intl\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",8202099.737532808],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",44.33333333333334],PARAMETER[\"Standard_Parallel_2\",46.0],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2913]]",
        2914=>"PROJCS[\"NAD_1983_HARN_StatePlane_Oregon_South_FIPS_3602_Feet_Intl\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4921259.842519685],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",42.33333333333334],PARAMETER[\"Standard_Parallel_2\",44.0],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2914]]",
        2915=>"PROJCS[\"NAD_1983_HARN_StatePlane_Tennessee_FIPS_4100_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-86.0],PARAMETER[\"Standard_Parallel_1\",35.25],PARAMETER[\"Standard_Parallel_2\",36.41666666666666],PARAMETER[\"Latitude_Of_Origin\",34.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2915]]",
        2916=>"PROJCS[\"NAD_1983_HARN_StatePlane_Texas_North_FIPS_4201_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-101.5],PARAMETER[\"Standard_Parallel_1\",34.65],PARAMETER[\"Standard_Parallel_2\",36.18333333333333],PARAMETER[\"Latitude_Of_Origin\",34.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2916]]",
        2917=>"PROJCS[\"NAD_1983_HARN_StatePlane_Texas_North_Central_FIPS_4202_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",6561666.666666666],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",32.13333333333333],PARAMETER[\"Standard_Parallel_2\",33.96666666666667],PARAMETER[\"Latitude_Of_Origin\",31.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2917]]",
        2918=>"PROJCS[\"NAD_1983_HARN_StatePlane_Texas_Central_FIPS_4203_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2296583.333333333],PARAMETER[\"False_Northing\",9842500.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",30.11666666666667],PARAMETER[\"Standard_Parallel_2\",31.88333333333333],PARAMETER[\"Latitude_Of_Origin\",29.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2918]]",
        2919=>"PROJCS[\"NAD_1983_HARN_StatePlane_Texas_South_Central_FIPS_4204_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",13123333.33333333],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Standard_Parallel_1\",28.38333333333333],PARAMETER[\"Standard_Parallel_2\",30.28333333333333],PARAMETER[\"Latitude_Of_Origin\",27.83333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2919]]",
        2920=>"PROJCS[\"NAD_1983_HARN_StatePlane_Texas_South_FIPS_4205_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",16404166.66666666],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",26.16666666666667],PARAMETER[\"Standard_Parallel_2\",27.83333333333333],PARAMETER[\"Latitude_Of_Origin\",25.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2920]]",
        2921=>"PROJCS[\"NAD_1983_HARN_StatePlane_Utah_North_FIPS_4301_Feet_Intl\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640419.947506561],PARAMETER[\"False_Northing\",3280839.895013123],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",40.71666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.33333333333334],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2921]]",
        2922=>"PROJCS[\"NAD_1983_HARN_StatePlane_Utah_Central_FIPS_4302_Feet_Intl\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640419.947506561],PARAMETER[\"False_Northing\",6561679.790026246],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",39.01666666666667],PARAMETER[\"Standard_Parallel_2\",40.65],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2922]]",
        2923=>"PROJCS[\"NAD_1983_HARN_StatePlane_Utah_South_FIPS_4303_Feet_Intl\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640419.947506561],PARAMETER[\"False_Northing\",9842519.685039369],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",37.21666666666667],PARAMETER[\"Standard_Parallel_2\",38.35],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2923]]",
        2924=>"PROJCS[\"NAD_1983_HARN_StatePlane_Virginia_North_FIPS_4501_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",11482916.66666666],PARAMETER[\"False_Northing\",6561666.666666666],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",38.03333333333333],PARAMETER[\"Standard_Parallel_2\",39.2],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2924]]",
        2925=>"PROJCS[\"NAD_1983_HARN_StatePlane_Virginia_South_FIPS_4502_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",11482916.66666666],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",36.76666666666667],PARAMETER[\"Standard_Parallel_2\",37.96666666666667],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2925]]",
        2926=>"PROJCS[\"NAD_1983_HARN_StatePlane_Washington_North_FIPS_4601_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.8333333333333],PARAMETER[\"Standard_Parallel_1\",47.5],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2926]]",
        2927=>"PROJCS[\"NAD_1983_HARN_StatePlane_Washington_South_FIPS_4602_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",45.83333333333334],PARAMETER[\"Standard_Parallel_2\",47.33333333333334],PARAMETER[\"Latitude_Of_Origin\",45.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2927]]",
        2928=>"PROJCS[\"NAD_1983_HARN_StatePlane_Wisconsin_North_FIPS_4801_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",45.56666666666667],PARAMETER[\"Standard_Parallel_2\",46.76666666666667],PARAMETER[\"Latitude_Of_Origin\",45.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2928]]",
        2929=>"PROJCS[\"NAD_1983_HARN_StatePlane_Wisconsin_Central_FIPS_4802_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",44.25],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2929]]",
        2930=>"PROJCS[\"NAD_1983_HARN_StatePlane_Wisconsin_South_FIPS_4803_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",42.73333333333333],PARAMETER[\"Standard_Parallel_2\",44.06666666666667],PARAMETER[\"Latitude_Of_Origin\",42.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2930]]",
        2931=>"PROJCS[\"Beduaram_TM_13_NE\",GEOGCS[\"GCS_Beduaram\",DATUM[\"D_Beduaram\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",13.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2931]]",
        2932=>"PROJCS[\"QND_1995_Qatar_National_Grid\",GEOGCS[\"GCS_QND_1995\",DATUM[\"D_QND_1995\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",300000.0],PARAMETER[\"Central_Meridian\",51.21666666666667],PARAMETER[\"Scale_Factor\",0.99999],PARAMETER[\"Latitude_Of_Origin\",24.45],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2932]]",
        2933=>"PROJCS[\"Gunung_Segara_UTM_Zone_50S\",GEOGCS[\"GCS_Gunung_Segara\",DATUM[\"D_Gunung_Segara\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2933]]",
        2934=>"PROJCS[\"Gunung_Segara_Jakarta_NEIEZ\",GEOGCS[\"GCS_Gunung_Segara_Jakarta\",DATUM[\"D_Gunung_Segara\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Jakarta\",106.8077194444444],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Mercator\"],PARAMETER[\"False_Easting\",3900000.0],PARAMETER[\"False_Northing\",900000.0],PARAMETER[\"Central_Meridian\",3.192280555555556],PARAMETER[\"Standard_Parallel_1\",4.45405154589751],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5329]]",
        2935=>"PROJCS[\"Pulkovo_1942_CS63_Zone_A1\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",1300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",41.53333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.1166666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2935]]",
        2936=>"PROJCS[\"Pulkovo_1942_CS63_Zone_A2\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",2300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",44.53333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.1166666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2936]]",
        2937=>"PROJCS[\"Pulkovo_1942_CS63_Zone_A3\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",47.53333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.1166666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2937]]",
        2938=>"PROJCS[\"Pulkovo_1942_CS63_Zone_A4\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",50.53333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.1166666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2938]]",
        2939=>"PROJCS[\"Pulkovo_1942_CS63_Zone_K2\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",2300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",50.76666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.1333333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2939]]",
        2940=>"PROJCS[\"Pulkovo_1942_CS63_Zone_K3\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",53.76666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.1333333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2940]]",
        2941=>"PROJCS[\"Pulkovo_1942_CS63_Zone_K4\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",56.76666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.1333333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2941]]",
        2942=>"PROJCS[\"Porto_Santo_1936_UTM_Zone_28N\",GEOGCS[\"GCS_Porto_Santo_1936\",DATUM[\"D_Porto_Santo_1936\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2942]]",
        2943=>"PROJCS[\"Selvagem_Grande_1938_UTM_Zone_28N\",GEOGCS[\"GCS_Selvagem_Grande_1938\",DATUM[\"D_Selvagem_Grande_1938\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2943]]",
        2944=>"PROJCS[\"NAD_1983_CSRS_MTM_2_SCoPQ\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-55.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2944]]",
        2945=>"PROJCS[\"NAD_1983_CSRS_MTM_3\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-58.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2945]]",
        2946=>"PROJCS[\"NAD_1983_CSRS_MTM_4\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-61.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2946]]",
        2947=>"PROJCS[\"NAD_1983_CSRS_MTM_5\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-64.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2947]]",
        2948=>"PROJCS[\"NAD_1983_CSRS_MTM_6\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-67.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2948]]",
        2949=>"PROJCS[\"NAD_1983_CSRS_MTM_7\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2949]]",
        2950=>"PROJCS[\"NAD_1983_CSRS_MTM_8\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-73.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2950]]",
        2951=>"PROJCS[\"NAD_1983_CSRS_MTM_9\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-76.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2951]]",
        2952=>"PROJCS[\"NAD_1983_CSRS_MTM_10\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2952]]",
        2953=>"PROJCS[\"NAD_1983_CSRS_New_Brunswick_Stereographic\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",7500000.0],PARAMETER[\"Central_Meridian\",-66.5],PARAMETER[\"Scale_Factor\",0.999912],PARAMETER[\"Latitude_Of_Origin\",46.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2953]]",
        2954=>"PROJCS[\"NAD_1983_CSRS_Prince_Edward_Island\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.999912],PARAMETER[\"Latitude_Of_Origin\",47.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2954]]",
        2955=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_11N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2955]]",
        2956=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_12N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2956]]",
        2957=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_13N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2957]]",
        2958=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_17N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2958]]",
        2959=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_18N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2959]]",
        2960=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_19N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2960]]",
        2961=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_20N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2961]]",
        2962=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_21N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2962]]",
        2964=>"PROJCS[\"NAD_1927_Alaska_Albers_Feet\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-154.0],PARAMETER[\"Standard_Parallel_1\",55.0],PARAMETER[\"Standard_Parallel_2\",65.0],PARAMETER[\"Latitude_Of_Origin\",50.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2964]]",
        2965=>"PROJCS[\"NAD_1983_StatePlane_Indiana_East_FIPS_1301_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",328083.3333333333],PARAMETER[\"False_Northing\",820208.3333333333],PARAMETER[\"Central_Meridian\",-85.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2965]]",
        2966=>"PROJCS[\"NAD_1983_StatePlane_Indiana_West_FIPS_1302_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2952750.0],PARAMETER[\"False_Northing\",820208.3333333333],PARAMETER[\"Central_Meridian\",-87.08333333333333],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2966]]",
        2967=>"PROJCS[\"NAD_1983_HARN_StatePlane_Indiana_East_FIPS_1301_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",328083.3333333333],PARAMETER[\"False_Northing\",820208.3333333333],PARAMETER[\"Central_Meridian\",-85.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2967]]",
        2968=>"PROJCS[\"NAD_1983_HARN_StatePlane_Indiana_West_FIPS_1302_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2952750.0],PARAMETER[\"False_Northing\",820208.3333333333],PARAMETER[\"Central_Meridian\",-87.08333333333333],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2968]]",
        2969=>"PROJCS[\"Fort_Marigot_UTM_20N\",GEOGCS[\"GCS_Fort_Marigot\",DATUM[\"D_Fort_Marigot\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2969]]",
        2970=>"PROJCS[\"Sainte_Anne_UTM_20N\",GEOGCS[\"GCS_Sainte_Anne\",DATUM[\"D_Sainte_Anne\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2970]]",
        2971=>"PROJCS[\"CSG_1967_UTM_22N\",GEOGCS[\"GCS_CSG_1967\",DATUM[\"D_CSG_1967\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2971]]",
        2972=>"PROJCS[\"RGFG_1995_UTM_22N\",GEOGCS[\"GCS_RGFG_1995\",DATUM[\"D_RGFG_1995\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2972]]",
        2973=>"PROJCS[\"Fort_Desaix_UTM_20N\",GEOGCS[\"GCS_Fort_Desaix\",DATUM[\"D_Fort_Desaix\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2973]]",
        2975=>"PROJCS[\"RGR_1992_UTM_40S\",GEOGCS[\"GCS_RGR_1992\",DATUM[\"D_RGR_1992\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2975]]",
        2976=>"PROJCS[\"Tahiti_1952_UTM_6S\",GEOGCS[\"GCS_Tahiti_1952\",DATUM[\"D_Tahiti_1952\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2976]]",
        2977=>"PROJCS[\"Tahaa_1954_UTM_5S\",GEOGCS[\"GCS_Tahaa_1954\",DATUM[\"D_Tahaa_1954\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2977]]",
        2978=>"PROJCS[\"IGN72_Nuku_Hiva_UTM_7S\",GEOGCS[\"GCS_IGN72_Nuku_Hiva\",DATUM[\"D_IGN72_Nuku_Hiva\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2978]]",
        2979=>"PROJCS[\"Kerguelen_Island_1949_UTM_42S\",GEOGCS[\"GCS_Kerguelen_Island_1949\",DATUM[\"D_Kerguelen_Island_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3336]]",
        2980=>"PROJCS[\"Combani_1950_UTM_38S\",GEOGCS[\"GCS_Combani_1950\",DATUM[\"D_Combani_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2980]]",
        2981=>"PROJCS[\"IGN56_Lifou_UTM_58S\",GEOGCS[\"GCS_IGN56_Lifou\",DATUM[\"D_IGN56_Lifou\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2981]]",
        2982=>"PROJCS[\"IGN72_Grande_Terre_UTM_58S\",GEOGCS[\"GCS_IGN72_Grande_Terre\",DATUM[\"D_IGN72_Grande_Terre\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3060]]",
        2984=>"PROJCS[\"RGNC_1991_Lambert_New_Caledonia\",GEOGCS[\"GCS_RGNC_1991\",DATUM[\"D_RGNC_1991\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",300000.0],PARAMETER[\"Central_Meridian\",166.0],PARAMETER[\"Standard_Parallel_1\",-20.66666666666667],PARAMETER[\"Standard_Parallel_2\",-22.33333333333333],PARAMETER[\"Latitude_Of_Origin\",-21.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2984]]",
        2985=>"PROJCS[\"Petrels_1972_Terre_Adelie_Polar_Stereographic\",GEOGCS[\"GCS_Petrels_1972\",DATUM[\"D_Petrels_1972\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic_South_Pole\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",-2299363.487823496],PARAMETER[\"Central_Meridian\",140.0],PARAMETER[\"Standard_Parallel_1\",-67.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2985]]",
        2986=>"PROJCS[\"Perroud_1950_Terre_Adelie_Polar_Stereographic\",GEOGCS[\"GCS_Pointe_Geologie_Perroud_1950\",DATUM[\"D_Pointe_Geologie_Perroud_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic_South_Pole\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",-2299363.487823496],PARAMETER[\"Central_Meridian\",140.0],PARAMETER[\"Standard_Parallel_1\",-67.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2986]]",
        2987=>"PROJCS[\"Saint_Pierre_et_Miquelon_1950_UTM_21N\",GEOGCS[\"GCS_Saint_Pierre_et_Miquelon_1950\",DATUM[\"D_Saint_Pierre_et_Miquelon_1950\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2987]]",
        2988=>"PROJCS[\"MOP78_UTM_1S\",GEOGCS[\"GCS_MOP78\",DATUM[\"D_MOP78\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2988]]",
        2991=>"PROJCS[\"NAD_1983_Oregon_Statewide_Lambert\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",43.0],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2991]]",
        2992=>"PROJCS[\"NAD_1983_Oregon_Statewide_Lambert_Feet_Intl\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312335.958005249],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",43.0],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2992]]",
        2993=>"PROJCS[\"NAD_1983_HARN_Oregon_Statewide_Lambert\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",43.0],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2993]]",
        2994=>"PROJCS[\"NAD_1983_HARN_Oregon_Statewide_Lambert_Feet_Intl\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312335.958005249],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",43.0],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",2994]]",
        2995=>"PROJCS[\"IGN53_Mare_UTM_58S\",GEOGCS[\"GCS_IGN53_Mare\",DATUM[\"D_IGN53_Mare\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2995]]",
        2996=>"PROJCS[\"ST84_Ile_des_Pins_UTM_58S\",GEOGCS[\"GCS_ST84_Ile_des_Pins\",DATUM[\"D_ST84_Ile_des_Pins\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2996]]",
        2997=>"PROJCS[\"ST71_Belep_UTM_58S\",GEOGCS[\"GCS_ST71_Belep\",DATUM[\"D_ST71_Belep\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2997]]",
        2998=>"PROJCS[\"NEA74_Noumea_UTM_58S\",GEOGCS[\"GCS_NEA74_Noumea\",DATUM[\"D_NEA74_Noumea\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2998]]",
        2999=>"PROJCS[\"Grand_Comoros_UTM_38S\",GEOGCS[\"GCS_Grand_Comoros\",DATUM[\"D_Grand_Comoros\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2999]]",
        3000=>"PROJCS[\"Gunung_Segara_NEIEZ\",GEOGCS[\"GCS_Gunung_Segara\",DATUM[\"D_Gunung_Segara\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Mercator\"],PARAMETER[\"False_Easting\",3900000.0],PARAMETER[\"False_Northing\",900000.0],PARAMETER[\"Central_Meridian\",110.0],PARAMETER[\"Standard_Parallel_1\",4.45405154589751],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3000]]",
        3001=>"PROJCS[\"Batavia_NEIEZ\",GEOGCS[\"GCS_Batavia\",DATUM[\"D_Batavia\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Mercator\"],PARAMETER[\"False_Easting\",3900000.0],PARAMETER[\"False_Northing\",900000.0],PARAMETER[\"Central_Meridian\",110.0],PARAMETER[\"Standard_Parallel_1\",4.45405154589751],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3001]]",
        3002=>"PROJCS[\"Makassar_NEIEZ\",GEOGCS[\"GCS_Makassar\",DATUM[\"D_Makassar\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Mercator\"],PARAMETER[\"False_Easting\",3900000.0],PARAMETER[\"False_Northing\",900000.0],PARAMETER[\"Central_Meridian\",110.0],PARAMETER[\"Standard_Parallel_1\",4.45405154589751],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3002]]",
        3003=>"PROJCS[\"Monte_Mario_Italy_1\",GEOGCS[\"GCS_Monte_Mario\",DATUM[\"D_Monte_Mario\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3003]]",
        3004=>"PROJCS[\"Monte_Mario_Italy_2\",GEOGCS[\"GCS_Monte_Mario\",DATUM[\"D_Monte_Mario\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2520000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3004]]",
        3005=>"PROJCS[\"NAD_1983_BC_Environment_Albers\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-126.0],PARAMETER[\"Standard_Parallel_1\",50.0],PARAMETER[\"Standard_Parallel_2\",58.5],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3005]]",
        3006=>"PROJCS[\"SWEREF99_TM\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3006]]",
        3007=>"PROJCS[\"SWEREF99_12_00\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3007]]",
        3008=>"PROJCS[\"SWEREF99_13_30\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",13.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3008]]",
        3009=>"PROJCS[\"SWEREF99_15_00\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3009]]",
        3010=>"PROJCS[\"SWEREF99_16_30\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",16.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3010]]",
        3011=>"PROJCS[\"SWEREF99_18_00\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",18.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3011]]",
        3012=>"PROJCS[\"SWEREF99_14_15\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",14.25],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3012]]",
        3013=>"PROJCS[\"SWEREF99_15_45\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.75],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3013]]",
        3014=>"PROJCS[\"SWEREF99_17_15\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",17.25],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3014]]",
        3015=>"PROJCS[\"SWEREF99_18_45\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",18.75],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3015]]",
        3016=>"PROJCS[\"SWEREF99_20_15\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",20.25],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3016]]",
        3017=>"PROJCS[\"SWEREF99_21_45\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.75],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3017]]",
        3018=>"PROJCS[\"SWEREF99_23_15\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",23.25],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3018]]",
        3019=>"PROJCS[\"RT90_75_gon_V\",GEOGCS[\"GCS_RT_1990\",DATUM[\"D_RT_1990\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",11.30827777777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3019]]",
        3020=>"PROJCS[\"RT90_5_gon_V\",GEOGCS[\"GCS_RT_1990\",DATUM[\"D_RT_1990\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",13.55827777777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3020]]",
        3021=>"PROJCS[\"RT90_25_gon_V\",GEOGCS[\"GCS_RT_1990\",DATUM[\"D_RT_1990\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.80827777777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3021]]",
        3022=>"PROJCS[\"RT90_0_gon\",GEOGCS[\"GCS_RT_1990\",DATUM[\"D_RT_1990\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",18.05827777777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3022]]",
        3023=>"PROJCS[\"RT90_25_gon_O\",GEOGCS[\"GCS_RT_1990\",DATUM[\"D_RT_1990\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",20.30827777777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3023]]",
        3024=>"PROJCS[\"RT90_5_gon_O\",GEOGCS[\"GCS_RT_1990\",DATUM[\"D_RT_1990\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",22.55827777777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3024]]",
        3025=>"PROJCS[\"RT38_75_gon_V\",GEOGCS[\"GCS_RT38\",DATUM[\"D_Stockholm_1938\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",11.30827777777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3025]]",
        3026=>"PROJCS[\"RT38_5_gon_V\",GEOGCS[\"GCS_RT38\",DATUM[\"D_Stockholm_1938\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",13.55827777777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3026]]",
        3027=>"PROJCS[\"RT38_25_gon_V\",GEOGCS[\"GCS_RT38\",DATUM[\"D_Stockholm_1938\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.80827777777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3027]]",
        3028=>"PROJCS[\"RT38_0_gon\",GEOGCS[\"GCS_RT38\",DATUM[\"D_Stockholm_1938\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",18.05827777777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3028]]",
        3029=>"PROJCS[\"RT38_25_gon_O\",GEOGCS[\"GCS_RT38\",DATUM[\"D_Stockholm_1938\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",20.30827777777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3029]]",
        3030=>"PROJCS[\"RT38_5_gon_O\",GEOGCS[\"GCS_RT38\",DATUM[\"D_Stockholm_1938\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",22.55827777777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3030]]",
        3031=>"PROJCS[\"WGS_1984_Antarctic_Polar_Stereographic\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic_South_Pole\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",-71.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3031]]",
        3032=>"PROJCS[\"WGS_1984_Australian_Antarctic_Polar_Stereographic\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic_South_Pole\"],PARAMETER[\"False_Easting\",6000000.0],PARAMETER[\"False_Northing\",6000000.0],PARAMETER[\"Central_Meridian\",70.0],PARAMETER[\"Standard_Parallel_1\",-71.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3032]]",
        3033=>"PROJCS[\"WGS_1984_Australian_Antarctic_Lambert\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6000000.0],PARAMETER[\"False_Northing\",6000000.0],PARAMETER[\"Central_Meridian\",70.0],PARAMETER[\"Standard_Parallel_1\",-68.5],PARAMETER[\"Standard_Parallel_2\",-74.5],PARAMETER[\"Latitude_Of_Origin\",-50.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3033]]",
        3034=>"PROJCS[\"ETRS_1989_LCC\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4000000.0],PARAMETER[\"False_Northing\",2800000.0],PARAMETER[\"Central_Meridian\",10.0],PARAMETER[\"Standard_Parallel_1\",35.0],PARAMETER[\"Standard_Parallel_2\",65.0],PARAMETER[\"Latitude_Of_Origin\",52.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3034]]",
        3035=>"PROJCS[\"ETRS_1989_LAEA\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",4321000.0],PARAMETER[\"False_Northing\",3210000.0],PARAMETER[\"Central_Meridian\",10.0],PARAMETER[\"Latitude_Of_Origin\",52.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3035]]",
        3036=>"PROJCS[\"Moznet_UTM_Zone_36S\",GEOGCS[\"GCS_Moznet\",DATUM[\"D_Moznet\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3036]]",
        3037=>"PROJCS[\"Moznet_UTM_Zone_37S\",GEOGCS[\"GCS_Moznet\",DATUM[\"D_Moznet\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3037]]",
        3038=>"PROJCS[\"ETRS_1989_ETRS-TM26\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3038]]",
        3039=>"PROJCS[\"ETRS_1989_ETRS-TM27\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3039]]",
        3040=>"PROJCS[\"ETRS_1989_ETRS-TM28\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3040]]",
        3041=>"PROJCS[\"ETRS_1989_ETRS-TM29\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3041]]",
        3042=>"PROJCS[\"ETRS_1989_ETRS-TM30\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3042]]",
        3043=>"PROJCS[\"ETRS_1989_ETRS-TM31\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3043]]",
        3044=>"PROJCS[\"ETRS_1989_ETRS-TM32\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3044]]",
        3045=>"PROJCS[\"ETRS_1989_ETRS-TM33\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3045]]",
        3046=>"PROJCS[\"ETRS_1989_ETRS-TM34\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3046]]",
        3047=>"PROJCS[\"ETRS_1989_ETRS-TM35\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3047]]",
        3048=>"PROJCS[\"ETRS_1989_ETRS-TM36\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3048]]",
        3049=>"PROJCS[\"ETRS_1989_ETRS-TM37\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3049]]",
        3050=>"PROJCS[\"ETRS_1989_ETRS-TM38\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3050]]",
        3051=>"PROJCS[\"ETRS_1989_ETRS-TM39\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3051]]",
        3054=>"PROJCS[\"Hjorsey_1955_UTM_Zone_26N\",GEOGCS[\"GCS_Hjorsey_1955\",DATUM[\"D_Hjorsey_1955\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3054]]",
        3055=>"PROJCS[\"Hjorsey_1955_UTM_Zone_27N\",GEOGCS[\"GCS_Hjorsey_1955\",DATUM[\"D_Hjorsey_1955\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3055]]",
        3056=>"PROJCS[\"Hjorsey_1955_UTM_Zone_28N\",GEOGCS[\"GCS_Hjorsey_1955\",DATUM[\"D_Hjorsey_1955\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3056]]",
        3057=>"PROJCS[\"ISN_1993_Lambert_1993\",GEOGCS[\"GCS_ISN_1993\",DATUM[\"D_Islands_Network_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-19.0],PARAMETER[\"Standard_Parallel_1\",64.25],PARAMETER[\"Standard_Parallel_2\",65.75],PARAMETER[\"Latitude_Of_Origin\",65.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3057]]",
        3058=>"PROJCS[\"Helle_1954_Jan_Mayen_Grid\",GEOGCS[\"GCS_Helle_1954\",DATUM[\"D_Helle_1954\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",-7800000.0],PARAMETER[\"Central_Meridian\",-8.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3058]]",
        3059=>"PROJCS[\"LKS_1992_Latvia_TM\",GEOGCS[\"GCS_LKS_1992\",DATUM[\"D_Latvia_1992\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",-6000000.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3059]]",
        3060=>"PROJCS[\"IGN72_Grande_Terre_UTM_58S\",GEOGCS[\"GCS_IGN72_Grande_Terre\",DATUM[\"D_IGN72_Grande_Terre\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3060]]",
        3061=>"PROJCS[\"Porto_Santo_1995_UTM_Zone_28N\",GEOGCS[\"GCS_Porto_Santo_1995\",DATUM[\"D_Porto_Santo_1995\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3061]]",
        3062=>"PROJCS[\"Azores_Oriental_1995_UTM_Zone_26N\",GEOGCS[\"GCS_Azores_Oriental_1995\",DATUM[\"D_Azores_Oriental_Islands_1995\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3062]]",
        3063=>"PROJCS[\"Azores_Central_1995_UTM_Zone_26N\",GEOGCS[\"GCS_Azores_Central_1995\",DATUM[\"D_Azores_Central_Islands_1995\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3063]]",
        3064=>"PROJCS[\"IGM_1995_UTM_Zone_32N\",GEOGCS[\"GCS_IGM_1995\",DATUM[\"D_IGM_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3064]]",
        3065=>"PROJCS[\"IGM_1995_UTM_Zone_33N\",GEOGCS[\"GCS_IGM_1995\",DATUM[\"D_IGM_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3065]]",
        3066=>"PROJCS[\"ED_1950_Jordan_TM\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",-3000000.0],PARAMETER[\"Central_Meridian\",37.0],PARAMETER[\"Scale_Factor\",0.9998],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3066]]",
        3067=>"PROJCS[\"EUREF_FIN_TM35FIN\",GEOGCS[\"GCS_EUREF_FIN\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3067]]",
        3068=>"PROJCS[\"DHDN_Soldner_Berlin\",GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",40000.0],PARAMETER[\"False_Northing\",10000.0],PARAMETER[\"Central_Meridian\",13.62720366666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",52.41864827777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3068]]",
        3069=>"PROJCS[\"NAD_1927_Wisconsin_TM\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",-4500000.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3069]]",
        3070=>"PROJCS[\"NAD_1983_Wisconsin_TM\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",520000.0],PARAMETER[\"False_Northing\",-4480000.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3070]]",
        3071=>"PROJCS[\"NAD_1983_HARN_Wisconsin_TM\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",520000.0],PARAMETER[\"False_Northing\",-4480000.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3071]]",
        3072=>"PROJCS[\"NAD_1983_Maine_2000_East_Zone\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-67.875],PARAMETER[\"Scale_Factor\",0.99998],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3072]]",
        3073=>"PROJCS[\"NAD_1983_Maine_2000_Central_Zone\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.125],PARAMETER[\"Scale_Factor\",0.99998],PARAMETER[\"Latitude_Of_Origin\",43.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3463]]",
        3074=>"PROJCS[\"NAD_1983_Maine_2000_West_Zone\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.375],PARAMETER[\"Scale_Factor\",0.99998],PARAMETER[\"Latitude_Of_Origin\",42.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3074]]",
        3075=>"PROJCS[\"NAD_1983_HARN_Maine_2000_East_Zone\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-67.875],PARAMETER[\"Scale_Factor\",0.99998],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3075]]",
        3076=>"PROJCS[\"NAD_1983_HARN_Maine_2000_Central_Zone\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.125],PARAMETER[\"Scale_Factor\",0.99998],PARAMETER[\"Latitude_Of_Origin\",43.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3464]]",
        3077=>"PROJCS[\"NAD_1983_HARN_Maine_2000_West_Zone\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.375],PARAMETER[\"Scale_Factor\",0.99998],PARAMETER[\"Latitude_Of_Origin\",42.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3077]]",
        3078=>"PROJCS[\"NAD_1983_Michigan_GeoRef_Meters\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Natural_Origin\"],PARAMETER[\"False_Easting\",2546731.496],PARAMETER[\"False_Northing\",-4354009.816],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Azimuth\",337.25556],PARAMETER[\"Longitude_Of_Center\",-86.0],PARAMETER[\"Latitude_Of_Center\",45.30916666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3078]]",
        3079=>"PROJCS[\"NAD_1983_HARN_Michigan_GeoRef_Meters\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Natural_Origin\"],PARAMETER[\"False_Easting\",2546731.496],PARAMETER[\"False_Northing\",-4354009.816],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Azimuth\",337.25556],PARAMETER[\"Longitude_Of_Center\",-86.0],PARAMETER[\"Latitude_Of_Center\",45.30916666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3079]]",
        3080=>"PROJCS[\"NAD_1927_Texas_Statewide_Mapping_System\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.0],PARAMETER[\"False_Northing\",3000000.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",27.41666666666667],PARAMETER[\"Standard_Parallel_2\",34.91666666666666],PARAMETER[\"Latitude_Of_Origin\",31.16666666666667],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3080]]",
        3081=>"PROJCS[\"NAD_1983_Texas_Statewide_Mapping_System\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",27.41666666666667],PARAMETER[\"Standard_Parallel_2\",34.91666666666666],PARAMETER[\"Latitude_Of_Origin\",31.16666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3081]]",
        3082=>"PROJCS[\"NAD_1983_Texas_Centric_Mapping_System_Lambert\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",5000000.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",27.5],PARAMETER[\"Standard_Parallel_2\",35.0],PARAMETER[\"Latitude_Of_Origin\",18.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3082]]",
        3083=>"PROJCS[\"NAD_1983_Texas_Centric_Mapping_System_Albers\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",6000000.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",27.5],PARAMETER[\"Standard_Parallel_2\",35.0],PARAMETER[\"Latitude_Of_Origin\",18.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3083]]",
        3084=>"PROJCS[\"NAD_1983_HARN_Texas_Centric_Mapping_System_Lambert\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",5000000.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",27.5],PARAMETER[\"Standard_Parallel_2\",35.0],PARAMETER[\"Latitude_Of_Origin\",18.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3084]]",
        3085=>"PROJCS[\"NAD_1983_HARN_Texas_Centric_Mapping_System_Albers\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",6000000.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",27.5],PARAMETER[\"Standard_Parallel_2\",35.0],PARAMETER[\"Latitude_Of_Origin\",18.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3085]]",
        3086=>"PROJCS[\"NAD_1983_Florida_GDL_Albers\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.0],PARAMETER[\"Standard_Parallel_1\",24.0],PARAMETER[\"Standard_Parallel_2\",31.5],PARAMETER[\"Latitude_Of_Origin\",24.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3086]]",
        3087=>"PROJCS[\"NAD_1983_HARN_Florida_GDL_Albers\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.0],PARAMETER[\"Standard_Parallel_1\",24.0],PARAMETER[\"Standard_Parallel_2\",31.5],PARAMETER[\"Latitude_Of_Origin\",24.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3087]]",
        3088=>"PROJCS[\"NAD_1983_StatePlane_Kentucky_FIPS_1600\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Standard_Parallel_1\",37.08333333333334],PARAMETER[\"Standard_Parallel_2\",38.66666666666666],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3088]]",
        3089=>"PROJCS[\"NAD_1983_StatePlane_Kentucky_FIPS_1600_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4921250.0],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Standard_Parallel_1\",37.08333333333334],PARAMETER[\"Standard_Parallel_2\",38.66666666666666],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3089]]",
        3090=>"PROJCS[\"NAD_1983_HARN_StatePlane_Kentucky_FIPS_1600\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Standard_Parallel_1\",37.08333333333334],PARAMETER[\"Standard_Parallel_2\",38.66666666666666],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3090]]",
        3091=>"PROJCS[\"NAD_1983_HARN_StatePlane_Kentucky_FIPS_1600_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4921250.0],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Standard_Parallel_1\",37.08333333333334],PARAMETER[\"Standard_Parallel_2\",38.66666666666666],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3091]]",
        3092=>"PROJCS[\"Tokyo_UTM_Zone_51N\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3092]]",
        3093=>"PROJCS[\"Tokyo_UTM_Zone_52N\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3093]]",
        3094=>"PROJCS[\"Tokyo_UTM_Zone_53N\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3094]]",
        3095=>"PROJCS[\"Tokyo_UTM_Zone_54N\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3095]]",
        3096=>"PROJCS[\"Tokyo_UTM_Zone_55N\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3096]]",
        3097=>"PROJCS[\"JGD_2000_UTM_Zone_51N\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3097]]",
        3098=>"PROJCS[\"JGD_2000_UTM_Zone_52N\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3098]]",
        3099=>"PROJCS[\"JGD_2000_UTM_Zone_53N\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3099]]",
        3100=>"PROJCS[\"JGD_2000_UTM_Zone_54N\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3100]]",
        3101=>"PROJCS[\"JGD_2000_UTM_Zone_55N\",GEOGCS[\"GCS_JGD_2000\",DATUM[\"D_JGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3101]]",
        3102=>"PROJCS[\"Samoa_1962_Samoa_Lambert\",GEOGCS[\"GCS_American_Samoa_1962\",DATUM[\"D_American_Samoa_1962\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",312234.65],PARAMETER[\"Central_Meridian\",-170.0],PARAMETER[\"Standard_Parallel_1\",-14.26666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-14.26666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3102]]",
        3106=>"PROJCS[\"Gulshan_303_Bangladesh_TM\",GEOGCS[\"GCS_Gulshan_303\",DATUM[\"D_Gulshan_303\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3106]]",
        3107=>"PROJCS[\"GDA_1994_South_Australia_Lambert\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Standard_Parallel_1\",-28.0],PARAMETER[\"Standard_Parallel_2\",-36.0],PARAMETER[\"Latitude_Of_Origin\",-32.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3107]]",
        3108=>"PROJCS[\"ETRS_1989_Guernsey_Grid\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",47000.0],PARAMETER[\"False_Northing\",50000.0],PARAMETER[\"Central_Meridian\",-2.416666666666667],PARAMETER[\"Scale_Factor\",0.999997],PARAMETER[\"Latitude_Of_Origin\",49.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3108]]",
        3109=>"PROJCS[\"ETRS_1989_Jersey_Transverse_Mercator\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",40000.0],PARAMETER[\"False_Northing\",70000.0],PARAMETER[\"Central_Meridian\",-2.135],PARAMETER[\"Scale_Factor\",0.9999999],PARAMETER[\"Latitude_Of_Origin\",49.225],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3109]]",
        3110=>"PROJCS[\"AGD_1966_VICGRID\",GEOGCS[\"GCS_Australian_1966\",DATUM[\"D_Australian_1966\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",145.0],PARAMETER[\"Standard_Parallel_1\",-36.0],PARAMETER[\"Standard_Parallel_2\",-38.0],PARAMETER[\"Latitude_Of_Origin\",-37.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3110]]",
        3111=>"PROJCS[\"GDA_1994_VICGRID94\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",2500000.0],PARAMETER[\"Central_Meridian\",145.0],PARAMETER[\"Standard_Parallel_1\",-36.0],PARAMETER[\"Standard_Parallel_2\",-38.0],PARAMETER[\"Latitude_Of_Origin\",-37.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3111]]",
        3112=>"PROJCS[\"GDA_1994_Geoscience_Australia_Lambert\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",134.0],PARAMETER[\"Standard_Parallel_1\",-18.0],PARAMETER[\"Standard_Parallel_2\",-36.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3112]]",
        3113=>"PROJCS[\"GDA_1994_BCSG02\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",0.99999],PARAMETER[\"Latitude_Of_Origin\",-28.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3113]]",
        3114=>"PROJCS[\"MAGNA_Colombia_Oeste_Oeste\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-80.07750791666666],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",4.596200416666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3114]]",
        3115=>"PROJCS[\"MAGNA_Colombia_Oeste\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-77.07750791666666],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",4.596200416666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3115]]",
        3116=>"PROJCS[\"MAGNA_Colombia_Bogota\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-74.07750791666666],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",4.596200416666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3116]]",
        3117=>"PROJCS[\"MAGNA_Colombia_Este\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-71.07750791666666],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",4.596200416666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3117]]",
        3118=>"PROJCS[\"MAGNA_Colombia_Este_Este\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-68.07750791666666],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",4.596200416666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3118]]",
        3119=>"PROJCS[\"Douala_1948_AEF_West\",GEOGCS[\"GCS_Douala_1948\",DATUM[\"D_Douala_1948\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",10.5],PARAMETER[\"Scale_Factor\",0.999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3119]]",
        3120=>"PROJCS[\"Pulkovo_1942_Adj_1958_Poland_Zone_I\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",4637000.0],PARAMETER[\"False_Northing\",5467000.0],PARAMETER[\"Central_Meridian\",21.08333333333333],PARAMETER[\"Scale_Factor\",0.9998],PARAMETER[\"Latitude_Of_Origin\",50.625],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3120]]",
        3121=>"PROJCS[\"PRS_1992_Philippines_Zone_I\",GEOGCS[\"GCS_PRS_1992\",DATUM[\"D_Philippine_Reference_System_1992\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3121]]",
        3122=>"PROJCS[\"PRS_1992_Philippines_Zone_II\",GEOGCS[\"GCS_PRS_1992\",DATUM[\"D_Philippine_Reference_System_1992\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",119.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3122]]",
        3123=>"PROJCS[\"PRS_1992_Philippines_Zone_III\",GEOGCS[\"GCS_PRS_1992\",DATUM[\"D_Philippine_Reference_System_1992\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",121.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3123]]",
        3124=>"PROJCS[\"PRS_1992_Philippines_Zone_IV\",GEOGCS[\"GCS_PRS_1992\",DATUM[\"D_Philippine_Reference_System_1992\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3124]]",
        3125=>"PROJCS[\"PRS_1992_Philippines_Zone_V\",GEOGCS[\"GCS_PRS_1992\",DATUM[\"D_Philippine_Reference_System_1992\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",125.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3125]]",
        3126=>"PROJCS[\"ETRS_1989_ETRS-GK19FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",19.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3126]]",
        3127=>"PROJCS[\"ETRS_1989_ETRS-GK20FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",20.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3127]]",
        3128=>"PROJCS[\"ETRS_1989_ETRS-GK21FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3128]]",
        3129=>"PROJCS[\"ETRS_1989_ETRS-GK22FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",22.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3129]]",
        3130=>"PROJCS[\"ETRS_1989_ETRS-GK23FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",23.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3130]]",
        3131=>"PROJCS[\"ETRS_1989_ETRS-GK24FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3131]]",
        3132=>"PROJCS[\"ETRS_1989_ETRS-GK25FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",25.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3132]]",
        3133=>"PROJCS[\"ETRS_1989_ETRS-GK26FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",26.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3133]]",
        3134=>"PROJCS[\"ETRS_1989_ETRS-GK27FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3134]]",
        3135=>"PROJCS[\"ETRS_1989_ETRS-GK28FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",28.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3135]]",
        3136=>"PROJCS[\"ETRS_1989_ETRS-GK29FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",29.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3136]]",
        3137=>"PROJCS[\"ETRS_1989_ETRS-GK30FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3137]]",
        3138=>"PROJCS[\"ETRS_1989_ETRS-GK31FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",31.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3138]]",
        3141=>"PROJCS[\"Fiji_1956_UTM_Zone_60S\",GEOGCS[\"GCS_Fiji_1956\",DATUM[\"D_Fiji_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3141]]",
        3142=>"PROJCS[\"Fiji_1956_UTM_Zone_1S\",GEOGCS[\"GCS_Fiji_1956\",DATUM[\"D_Fiji_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3142]]",
        3146=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_6\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",6500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",18.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3146]]",
        3147=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_CM_18E\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",18.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3147]]",
        3148=>"PROJCS[\"Indian_1960_UTM_Zone_48N\",GEOGCS[\"GCS_Indian_1960\",DATUM[\"D_Indian_1960\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3148]]",
        3149=>"PROJCS[\"Indian_1960_UTM_Zone_49N\",GEOGCS[\"GCS_Indian_1960\",DATUM[\"D_Indian_1960\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3149]]",
        3150=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_6\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",6500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",18.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3150]]",
        3151=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_CM_18E\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",18.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3151]]",
        3153=>"PROJCS[\"NAD_1983_CSRS_BC_Environment_Albers\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-126.0],PARAMETER[\"Standard_Parallel_1\",50.0],PARAMETER[\"Standard_Parallel_2\",58.5],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3153]]",
        3154=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_7N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3154]]",
        3155=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_8N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3155]]",
        3156=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_9N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3156]]",
        3157=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_10N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3157]]",
        3158=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_14N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3158]]",
        3159=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_15N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3159]]",
        3160=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_16N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3160]]",
        3161=>"PROJCS[\"NAD_1983_Ontario_MNR_Lambert\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",930000.0],PARAMETER[\"False_Northing\",6430000.0],PARAMETER[\"Central_Meridian\",-85.0],PARAMETER[\"Standard_Parallel_1\",44.5],PARAMETER[\"Standard_Parallel_2\",53.5],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3161]]",
        3162=>"PROJCS[\"NAD_1983_CSRS_Ontario_MNR_Lambert\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",930000.0],PARAMETER[\"False_Northing\",6430000.0],PARAMETER[\"Central_Meridian\",-85.0],PARAMETER[\"Standard_Parallel_1\",44.5],PARAMETER[\"Standard_Parallel_2\",53.5],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3162]]",
        3163=>"PROJCS[\"RGNC_1991_93_Lambert_New_Caledonia\",GEOGCS[\"GCS_RGNC_1991-93\",DATUM[\"D_Reseau_Geodesique_de_Nouvelle_Caledonie_1991-93\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",300000.0],PARAMETER[\"Central_Meridian\",166.0],PARAMETER[\"Standard_Parallel_1\",-20.66666666666667],PARAMETER[\"Standard_Parallel_2\",-22.33333333333333],PARAMETER[\"Latitude_Of_Origin\",-21.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3163]]",
        3164=>"PROJCS[\"ST87_Ouvea_UTM_58S\",GEOGCS[\"GCS_ST87_Ouvea\",DATUM[\"D_ST87_Ouvea\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3164]]",
        3165=>"PROJCS[\"NEA74_Noumea_Lambert\",GEOGCS[\"GCS_NEA74_Noumea\",DATUM[\"D_NEA74_Noumea\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.66],PARAMETER[\"False_Northing\",1.02],PARAMETER[\"Central_Meridian\",166.44242575],PARAMETER[\"Standard_Parallel_1\",-22.24469175],PARAMETER[\"Standard_Parallel_2\",-22.29469175],PARAMETER[\"Latitude_Of_Origin\",-22.26969175],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3165]]",
        3166=>"PROJCS[\"NEA74_Noumea_Lambert_2\",GEOGCS[\"GCS_NEA74_Noumea\",DATUM[\"D_NEA74_Noumea\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",8.313],PARAMETER[\"False_Northing\",-2.354],PARAMETER[\"Central_Meridian\",166.4425],PARAMETER[\"Standard_Parallel_1\",-22.24472222222222],PARAMETER[\"Standard_Parallel_2\",-22.29472222222222],PARAMETER[\"Latitude_Of_Origin\",-22.26972222222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3166]]",
        3167=>"PROJCS[\"Kertau_RSO_RSO_Malaya_ChSears1922trunc\",GEOGCS[\"GCS_Kertau_RSO\",DATUM[\"D_Kertau_RSO\",SPHEROID[\"Everest_Modified_1969\",6377295.664,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Rectified_Skew_Orthomorphic_Natural_Origin\"],PARAMETER[\"False_Easting\",40000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Scale_Factor\",0.99984],PARAMETER[\"Azimuth\",323.0257905],PARAMETER[\"Longitude_Of_Center\",102.25],PARAMETER[\"Latitude_Of_Center\",4.0],PARAMETER[\"XY_Plane_Rotation\",323.1301023611111],UNIT[\"Chain_Sears_1922_Truncated\",20.116756],AUTHORITY[\"EPSG\",3167]]",
        3168=>"PROJCS[\"Kertau_RSO_RSO_Malaya\",GEOGCS[\"GCS_Kertau_RSO\",DATUM[\"D_Kertau_RSO\",SPHEROID[\"Everest_Modified_1969\",6377295.664,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Rectified_Skew_Orthomorphic_Natural_Origin\"],PARAMETER[\"False_Easting\",804670.24],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Scale_Factor\",0.99984],PARAMETER[\"Azimuth\",323.0257905],PARAMETER[\"Longitude_Of_Center\",102.25],PARAMETER[\"Latitude_Of_Center\",4.0],PARAMETER[\"XY_Plane_Rotation\",323.1301023611111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3168]]",
        3169=>"PROJCS[\"RGNC_1991-93_UTM_Zone_57S\",GEOGCS[\"GCS_RGNC_1991-93\",DATUM[\"D_Reseau_Geodesique_de_Nouvelle_Caledonie_1991-93\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3169]]",
        3170=>"PROJCS[\"RGNC_1991-93_UTM_Zone_58S\",GEOGCS[\"GCS_RGNC_1991-93\",DATUM[\"D_Reseau_Geodesique_de_Nouvelle_Caledonie_1991-93\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3170]]",
        3171=>"PROJCS[\"RGNC_1991-93_UTM_Zone_59S\",GEOGCS[\"GCS_RGNC_1991-93\",DATUM[\"D_Reseau_Geodesique_de_Nouvelle_Caledonie_1991-93\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3171]]",
        3172=>"PROJCS[\"IGN53_Mare_UTM_Zone_59S\",GEOGCS[\"GCS_IGN53_Mare\",DATUM[\"D_IGN53_Mare\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3172]]",
        3174=>"PROJCS[\"NAD_1983_Great_Lakes_Basin_Albers\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-84.455955],PARAMETER[\"Standard_Parallel_1\",42.122774],PARAMETER[\"Standard_Parallel_2\",49.01518],PARAMETER[\"Latitude_Of_Origin\",45.568977],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3174]]",
        3175=>"PROJCS[\"NAD_1983_Great_Lakes_and_St_Lawrence_Albers\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-83.248627],PARAMETER[\"Standard_Parallel_1\",42.122774],PARAMETER[\"Standard_Parallel_2\",49.01518],PARAMETER[\"Latitude_Of_Origin\",45.568977],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3175]]",
        3176=>"PROJCS[\"Indian_1960_TM_106NE\",GEOGCS[\"GCS_Indian_1960\",DATUM[\"D_Indian_1960\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",106.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3176]]",
        3177=>"PROJCS[\"LGD2006_Libya_TM\",GEOGCS[\"GCS_LGD2006\",DATUM[\"D_Libyan_Geodetic_Datum_2006\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",17.0],PARAMETER[\"Scale_Factor\",0.9965],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3177]]",
        3178=>"PROJCS[\"Greenland_1996_UTM_Zone_18N\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3178]]",
        3179=>"PROJCS[\"Greenland_1996_UTM_Zone_19N\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3179]]",
        3180=>"PROJCS[\"Greenland_1996_UTM_Zone_20N\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3180]]",
        3181=>"PROJCS[\"Greenland_1996_UTM_Zone_21N\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3181]]",
        3182=>"PROJCS[\"Greenland_1996_UTM_Zone_22N\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3182]]",
        3183=>"PROJCS[\"Greenland_1996_UTM_Zone_23N\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3183]]",
        3184=>"PROJCS[\"Greenland_1996_UTM_Zone_24N\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3184]]",
        3185=>"PROJCS[\"Greenland_1996_UTM_Zone_25N\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3185]]",
        3186=>"PROJCS[\"Greenland_1996_UTM_Zone_26N\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3186]]",
        3187=>"PROJCS[\"Greenland_1996_UTM_Zone_27N\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3187]]",
        3188=>"PROJCS[\"Greenland_1996_UTM_Zone_28N\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3188]]",
        3189=>"PROJCS[\"Greenland_1996_UTM_Zone_29N\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3189]]",
        3190=>"PROJCS[\"LGD2006_Libya_TM_Zone_5\",GEOGCS[\"GCS_LGD2006\",DATUM[\"D_Libyan_Geodetic_Datum_2006\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3190]]",
        3191=>"PROJCS[\"LGD2006_Libya_TM_Zone_6\",GEOGCS[\"GCS_LGD2006\",DATUM[\"D_Libyan_Geodetic_Datum_2006\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",11.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3191]]",
        3192=>"PROJCS[\"LGD2006_Libya_TM_Zone_7\",GEOGCS[\"GCS_LGD2006\",DATUM[\"D_Libyan_Geodetic_Datum_2006\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",13.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3192]]",
        3193=>"PROJCS[\"LGD2006_Libya_TM_Zone_8\",GEOGCS[\"GCS_LGD2006\",DATUM[\"D_Libyan_Geodetic_Datum_2006\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3193]]",
        3194=>"PROJCS[\"LGD2006_Libya_TM_Zone_9\",GEOGCS[\"GCS_LGD2006\",DATUM[\"D_Libyan_Geodetic_Datum_2006\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",17.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3194]]",
        3195=>"PROJCS[\"LGD2006_Libya_TM_Zone_10\",GEOGCS[\"GCS_LGD2006\",DATUM[\"D_Libyan_Geodetic_Datum_2006\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",19.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3195]]",
        3196=>"PROJCS[\"LGD2006_Libya_TM_Zone_11\",GEOGCS[\"GCS_LGD2006\",DATUM[\"D_Libyan_Geodetic_Datum_2006\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3196]]",
        3197=>"PROJCS[\"LGD2006_Libya_TM_Zone_12\",GEOGCS[\"GCS_LGD2006\",DATUM[\"D_Libyan_Geodetic_Datum_2006\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",23.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3197]]",
        3198=>"PROJCS[\"LGD2006_Libya_TM_Zone_13\",GEOGCS[\"GCS_LGD2006\",DATUM[\"D_Libyan_Geodetic_Datum_2006\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",25.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3198]]",
        3199=>"PROJCS[\"LGD2006_UTM_Zone_32N\",GEOGCS[\"GCS_LGD2006\",DATUM[\"D_Libyan_Geodetic_Datum_2006\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3199]]",
        3200=>"PROJCS[\"FD_1958_Iraq\",GEOGCS[\"GCS_FD_1958\",DATUM[\"D_FD_1958\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",1166200.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Standard_Parallel_1\",32.5],PARAMETER[\"Scale_Factor\",0.99878640777],PARAMETER[\"Latitude_Of_Origin\",32.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3200]]",
        3201=>"PROJCS[\"LGD2006_UTM_Zone_33N\",GEOGCS[\"GCS_LGD2006\",DATUM[\"D_Libyan_Geodetic_Datum_2006\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3201]]",
        3202=>"PROJCS[\"LGD2006_UTM_Zone_34N\",GEOGCS[\"GCS_LGD2006\",DATUM[\"D_Libyan_Geodetic_Datum_2006\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3202]]",
        3203=>"PROJCS[\"LGD2006_UTM_Zone_35N\",GEOGCS[\"GCS_LGD2006\",DATUM[\"D_Libyan_Geodetic_Datum_2006\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3203]]",
        3294=>"PROJCS[\"WGS_1984_USGS_Transantarctic_Mountains\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",162.0],PARAMETER[\"Standard_Parallel_1\",-76.66666666666667],PARAMETER[\"Standard_Parallel_2\",-79.33333333333333],PARAMETER[\"Latitude_Of_Origin\",-78.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3294]]",
        3295=>"PROJCS[\"Guam_1963_Yap_Islands\",GEOGCS[\"GCS_Guam_1963\",DATUM[\"D_Guam_1963\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Azimuthal_Equidistant\"],PARAMETER[\"False_Easting\",40000.0],PARAMETER[\"False_Northing\",60000.0],PARAMETER[\"Central_Meridian\",138.1687444444444],PARAMETER[\"Latitude_Of_Origin\",9.546708333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3295]]",
        3296=>"PROJCS[\"RGPF_UTM_Zone_5S\",GEOGCS[\"GCS_RGPF\",DATUM[\"D_Reseau_Geodesique_de_la_Polynesie_Francaise\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3296]]",
        3297=>"PROJCS[\"RGPF_UTM_Zone_6S\",GEOGCS[\"GCS_RGPF\",DATUM[\"D_Reseau_Geodesique_de_la_Polynesie_Francaise\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3297]]",
        3298=>"PROJCS[\"RGPF_UTM_Zone_7S\",GEOGCS[\"GCS_RGPF\",DATUM[\"D_Reseau_Geodesique_de_la_Polynesie_Francaise\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3298]]",
        3299=>"PROJCS[\"RGPF_UTM_Zone_8S\",GEOGCS[\"GCS_RGPF\",DATUM[\"D_Reseau_Geodesique_de_la_Polynesie_Francaise\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3299]]",
        3300=>"PROJCS[\"Estonian_Coordinate_System_of_1992\",GEOGCS[\"GCS_Estonia_1992\",DATUM[\"D_Estonia_1992\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",6375000.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Standard_Parallel_1\",58.0],PARAMETER[\"Standard_Parallel_2\",59.33333333333334],PARAMETER[\"Latitude_Of_Origin\",57.51755393055556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3300]]",
        3301=>"PROJCS[\"Estonia_1997_Estonia_National_Grid\",GEOGCS[\"GCS_Estonia_1997\",DATUM[\"D_Estonia_1997\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",6375000.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Standard_Parallel_1\",58.0],PARAMETER[\"Standard_Parallel_2\",59.33333333333334],PARAMETER[\"Latitude_Of_Origin\",57.51755393055556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3301]]",
        3302=>"PROJCS[\"IGN63_Hiva_Oa_UTM_Zone_7S\",GEOGCS[\"GCS_IGN63_Hiva_Oa\",DATUM[\"D_IGN63_Hiva_Oa\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3302]]",
        3303=>"PROJCS[\"Fatu_Iva_1972_UTM_Zone_7S\",GEOGCS[\"GCS_Fatu_Iva_1972\",DATUM[\"D_Fatu_Iva_1972\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3303]]",
        3304=>"PROJCS[\"Tahiti_1979_UTM_Zone_6S\",GEOGCS[\"GCS_Tahiti_1979\",DATUM[\"D_Tahiti_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3304]]",
        3305=>"PROJCS[\"Moorea_1987_UTM_Zone_6S\",GEOGCS[\"GCS_Moorea_1987\",DATUM[\"D_Moorea_1987\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3305]]",
        3306=>"PROJCS[\"Maupiti_1983_UTM_Zone_5S\",GEOGCS[\"GCS_Maupiti_1983\",DATUM[\"D_Maupiti_1983\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3306]]",
        3307=>"PROJCS[\"Nakhl-e_Ghanem_UTM_Zone_39N\",GEOGCS[\"GCS_Nakhl-e_Ghanem\",DATUM[\"D_Nakhl-e_Ghanem\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3307]]",
        3308=>"PROJCS[\"GDA_1994_NSW_Lambert\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",9300000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Standard_Parallel_1\",-30.75],PARAMETER[\"Standard_Parallel_2\",-35.75],PARAMETER[\"Latitude_Of_Origin\",-33.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3308]]",
        3309=>"PROJCS[\"NAD_1927_California_Teale_Albers\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",-4000000.0],PARAMETER[\"Central_Meridian\",-120.0],PARAMETER[\"Standard_Parallel_1\",34.0],PARAMETER[\"Standard_Parallel_2\",40.5],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3309]]",
        3310=>"PROJCS[\"NAD_1983_California_Teale_Albers\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",-4000000.0],PARAMETER[\"Central_Meridian\",-120.0],PARAMETER[\"Standard_Parallel_1\",34.0],PARAMETER[\"Standard_Parallel_2\",40.5],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3310]]",
        3311=>"PROJCS[\"NAD_1983_HARN_California_Teale_Albers\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",-4000000.0],PARAMETER[\"Central_Meridian\",-120.0],PARAMETER[\"Standard_Parallel_1\",34.0],PARAMETER[\"Standard_Parallel_2\",40.5],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3311]]",
        3312=>"PROJCS[\"CSG_1967_UTM_Zone_21N\",GEOGCS[\"GCS_CSG_1967\",DATUM[\"D_CSG_1967\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3312]]",
        3313=>"PROJCS[\"RGFG_1995_UTM_Zone_21N\",GEOGCS[\"GCS_RGFG_1995\",DATUM[\"D_RGFG_1995\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3313]]",
        3315=>"PROJCS[\"Katanga_1955_Katanga_TM\",GEOGCS[\"GCS_Katanga_1955\",DATUM[\"D_Katanga_1955\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",26.0],PARAMETER[\"Scale_Factor\",0.9998],PARAMETER[\"Latitude_Of_Origin\",-9.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3315]]",
        3316=>"PROJCS[\"Kasai_1953_Congo_TM_Zone_22\",GEOGCS[\"GCS_Kasai_1953\",DATUM[\"D_Kasai_1953\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",22.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3316]]",
        3317=>"PROJCS[\"Kasai_1953_Congo_TM_Zone_24\",GEOGCS[\"GCS_Kasai_1953\",DATUM[\"D_Kasai_1953\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3317]]",
        3318=>"PROJCS[\"IGC_1962_Congo_TM_Zone_12\",GEOGCS[\"GCS_IGC_1962_6th_Parallel_South\",DATUM[\"D_IGC_1962_Arc_of_the_6th_Parallel_South\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3318]]",
        3319=>"PROJCS[\"IGC_1962_Congo_TM_Zone_14\",GEOGCS[\"GCS_IGC_1962_6th_Parallel_South\",DATUM[\"D_IGC_1962_Arc_of_the_6th_Parallel_South\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",14.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3319]]",
        3320=>"PROJCS[\"IGC_1962_Congo_TM_Zone_16\",GEOGCS[\"GCS_IGC_1962_6th_Parallel_South\",DATUM[\"D_IGC_1962_Arc_of_the_6th_Parallel_South\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",16.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3320]]",
        3321=>"PROJCS[\"IGC_1962_Congo_TM_Zone_18\",GEOGCS[\"GCS_IGC_1962_6th_Parallel_South\",DATUM[\"D_IGC_1962_Arc_of_the_6th_Parallel_South\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",18.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3321]]",
        3322=>"PROJCS[\"IGC_1962_Congo_TM_Zone_20\",GEOGCS[\"GCS_IGC_1962_6th_Parallel_South\",DATUM[\"D_IGC_1962_Arc_of_the_6th_Parallel_South\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",20.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3322]]",
        3323=>"PROJCS[\"IGC_1962_Congo_TM_Zone_22\",GEOGCS[\"GCS_IGC_1962_6th_Parallel_South\",DATUM[\"D_IGC_1962_Arc_of_the_6th_Parallel_South\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",22.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3323]]",
        3324=>"PROJCS[\"IGC_1962_Congo_TM_Zone_24\",GEOGCS[\"GCS_IGC_1962_6th_Parallel_South\",DATUM[\"D_IGC_1962_Arc_of_the_6th_Parallel_South\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3324]]",
        3325=>"PROJCS[\"IGC_1962_Congo_TM_Zone_26\",GEOGCS[\"GCS_IGC_1962_6th_Parallel_South\",DATUM[\"D_IGC_1962_Arc_of_the_6th_Parallel_South\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",26.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3325]]",
        3326=>"PROJCS[\"IGC_1962_Congo_TM_Zone_28\",GEOGCS[\"GCS_IGC_1962_6th_Parallel_South\",DATUM[\"D_IGC_1962_Arc_of_the_6th_Parallel_South\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",28.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3326]]",
        3327=>"PROJCS[\"IGC_1962_Congo_TM_Zone_30\",GEOGCS[\"GCS_IGC_1962_6th_Parallel_South\",DATUM[\"D_IGC_1962_Arc_of_the_6th_Parallel_South\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3327]]",
        3328=>"PROJCS[\"Pulkovo_1942_Adj_1958_GUGiK-80\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",19.16666666666667],PARAMETER[\"Scale_Factor\",0.9997143],PARAMETER[\"Latitude_Of_Origin\",52.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3328]]",
        3329=>"PROJCS[\"Pulkovo_1942_Adj_1958_3_Degree_GK_Zone_5\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3329]]",
        3330=>"PROJCS[\"Pulkovo_1942_Adj_1958_3_Degree_GK_Zone_6\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",6500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",18.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3330]]",
        3331=>"PROJCS[\"Pulkovo_1942_Adj_1958_3_Degree_GK_Zone_7\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",7500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3331]]",
        3332=>"PROJCS[\"Pulkovo_1942_Adj_1958_3_Degree_GK_Zone_8\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",8500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3332]]",
        3333=>"PROJCS[\"Pulkovo_1942_Adj_1958_GK_Zone_3\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3333]]",
        3334=>"PROJCS[\"Pulkovo_1942_Adj_1958_GK_Zone_4\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3334]]",
        3335=>"PROJCS[\"Pulkovo_1942_Adj_1958_GK_Zone_5\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3335]]",
        3336=>"PROJCS[\"Kerguelen_Island_1949_UTM_42S\",GEOGCS[\"GCS_Kerguelen_Island_1949\",DATUM[\"D_Kerguelen_Island_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3336]]",
        3337=>"PROJCS[\"Le_Pouce_1934_Mauritius_Grid\",GEOGCS[\"GCS_Le_Pouce_1934\",DATUM[\"D_Le_Pouce_1934\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",57.52182777777778],PARAMETER[\"Standard_Parallel_1\",-20.19506944444445],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-20.19506944444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3337]]",
        3338=>"PROJCS[\"NAD_1983_Alaska_Albers\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-154.0],PARAMETER[\"Standard_Parallel_1\",55.0],PARAMETER[\"Standard_Parallel_2\",65.0],PARAMETER[\"Latitude_Of_Origin\",50.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3338]]",
        3339=>"PROJCS[\"IGCB_1955_Congo_TM_Zone_12\",GEOGCS[\"GCS_IGCB_1955\",DATUM[\"D_Institut_Geographique_du_Congo_Belge_1955\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3339]]",
        3340=>"PROJCS[\"IGCB_1955_Congo_TM_Zone_14\",GEOGCS[\"GCS_IGCB_1955\",DATUM[\"D_Institut_Geographique_du_Congo_Belge_1955\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",14.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3340]]",
        3341=>"PROJCS[\"IGCB_1955_Congo_TM_Zone_16\",GEOGCS[\"GCS_IGCB_1955\",DATUM[\"D_Institut_Geographique_du_Congo_Belge_1955\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",16.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3341]]",
        3342=>"PROJCS[\"IGCB_1955_UTM_Zone_33S\",GEOGCS[\"GCS_IGCB_1955\",DATUM[\"D_Institut_Geographique_du_Congo_Belge_1955\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3342]]",
        3343=>"PROJCS[\"Mauritania_1999_UTM_Zone_28N\",GEOGCS[\"GCS_Mauritania_1999\",DATUM[\"D_Mauritania_1999\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3343]]",
        3344=>"PROJCS[\"Mauritania_1999_UTM_Zone_29N\",GEOGCS[\"GCS_Mauritania_1999\",DATUM[\"D_Mauritania_1999\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3344]]",
        3345=>"PROJCS[\"Mauritania_1999_UTM_Zone_30N\",GEOGCS[\"GCS_Mauritania_1999\",DATUM[\"D_Mauritania_1999\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3345]]",
        3346=>"PROJCS[\"LKS_1994_Lithuania_TM\",GEOGCS[\"GCS_LKS_1994\",DATUM[\"D_Lithuania_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",0.9998],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3346]]",
        3347=>"PROJCS[\"NAD_1983_Statistics_Canada_Lambert\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6200000.0],PARAMETER[\"False_Northing\",3000000.0],PARAMETER[\"Central_Meridian\",-91.86666666666666],PARAMETER[\"Standard_Parallel_1\",49.0],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",63.390675],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3347]]",
        3348=>"PROJCS[\"NAD_1983_CSRS_Statistics_Canada_Lambert\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6200000.0],PARAMETER[\"False_Northing\",3000000.0],PARAMETER[\"Central_Meridian\",-91.86666666666666],PARAMETER[\"Standard_Parallel_1\",49.0],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",63.390675],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3348]]",
        3350=>"PROJCS[\"Pulkovo_1942_CS63_Zone_C0\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.95],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.1],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3350]]",
        3351=>"PROJCS[\"Pulkovo_1942_CS63_Zone_C1\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",1250000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.95],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.1],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3351]]",
        3352=>"PROJCS[\"Pulkovo_1942_CS63_Zone_C2\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",2250000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.95],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.1],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3352]]",
        3353=>"PROJCS[\"Mhast_Onshore_UTM_Zone_32S\",GEOGCS[\"GCS_Mhast_Onshore\",DATUM[\"D_Mhast_Onshore\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3353]]",
        3354=>"PROJCS[\"Mhast_Offshore_UTM_Zone_32S\",GEOGCS[\"GCS_Mhast_Offshore\",DATUM[\"D_Mhast_Offshore\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3354]]",
        3355=>"PROJCS[\"Egypt_Gulf_of_Suez_S-650_TL_Red_Belt\",GEOGCS[\"GCS_Egypt_Gulf_of_Suez_S-650_TL\",DATUM[\"D_Egypt_Gulf_of_Suez_S-650_TL\",SPHEROID[\"Helmert_1906\",6378200.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",615000.0],PARAMETER[\"False_Northing\",810000.0],PARAMETER[\"Central_Meridian\",31.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3355]]",
        3356=>"PROJCS[\"Grand_Cayman_1959_UTM_Zone_17N\",GEOGCS[\"GCS_Grand_Cayman_1959\",DATUM[\"D_Grand_Cayman_1959\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3356]]",
        3357=>"PROJCS[\"Little_Cayman_1961_UTM_Zone_17N\",GEOGCS[\"GCS_Little_Cayman_1961\",DATUM[\"D_Little_Cayman_1961\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3357]]",
        3358=>"PROJCS[\"NAD_1983_HARN_StatePlane_North_Carolina_FIPS_3200\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",609601.2192024384],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.0],PARAMETER[\"Standard_Parallel_1\",34.33333333333334],PARAMETER[\"Standard_Parallel_2\",36.16666666666666],PARAMETER[\"Latitude_Of_Origin\",33.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3358]]",
        3359=>"PROJCS[\"NAD_1983_HARN_StatePlane_North_Carolina_FIPS_3200_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.0],PARAMETER[\"Standard_Parallel_1\",34.33333333333334],PARAMETER[\"Standard_Parallel_2\",36.16666666666666],PARAMETER[\"Latitude_Of_Origin\",33.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3404]]",
        3360=>"PROJCS[\"NAD_1983_HARN_StatePlane_South_Carolina_FIPS_3900\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",609600.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",32.5],PARAMETER[\"Standard_Parallel_2\",34.83333333333334],PARAMETER[\"Latitude_Of_Origin\",31.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3360]]",
        3361=>"PROJCS[\"NAD_1983_HARN_StatePlane_South_Carolina_FIPS_3900_Feet_Intl\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",32.5],PARAMETER[\"Standard_Parallel_2\",34.83333333333334],PARAMETER[\"Latitude_Of_Origin\",31.83333333333333],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3361]]",
        3362=>"PROJCS[\"NAD_1983_HARN_StatePlane_Pennsylvania_North_FIPS_3701\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",40.88333333333333],PARAMETER[\"Standard_Parallel_2\",41.95],PARAMETER[\"Latitude_Of_Origin\",40.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3362]]",
        3363=>"PROJCS[\"NAD_1983_HARN_StatePlane_Pennsylvania_North_FIPS_3701_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",40.88333333333333],PARAMETER[\"Standard_Parallel_2\",41.95],PARAMETER[\"Latitude_Of_Origin\",40.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3363]]",
        3364=>"PROJCS[\"NAD_1983_HARN_StatePlane_Pennsylvania_South_FIPS_3702\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",39.93333333333333],PARAMETER[\"Standard_Parallel_2\",40.96666666666667],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3364]]",
        3365=>"PROJCS[\"NAD_1983_HARN_StatePlane_Pennsylvania_South_FIPS_3702_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",39.93333333333333],PARAMETER[\"Standard_Parallel_2\",40.96666666666667],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3365]]",
        3366=>"PROJCS[\"Hong_Kong_1963_Grid_System\",GEOGCS[\"GCS_Hong_Kong_1963\",DATUM[\"D_Hong_Kong_1963\",SPHEROID[\"Clarke_1858\",6378293.645208759,294.260676369]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",132033.92],PARAMETER[\"False_Northing\",62565.96],PARAMETER[\"Central_Meridian\",114.1785555555556],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",22.31213333333334],UNIT[\"Foot_Clarke\",0.3047972654],AUTHORITY[\"EPSG\",3407]]",
        3367=>"PROJCS[\"IGN_Astro_1960_UTM_Zone_28N\",GEOGCS[\"GCS_IGN_Astro_1960\",DATUM[\"D_IGN_Astro_1960\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3367]]",
        3368=>"PROJCS[\"IGN_Astro_1960_UTM_Zone_29N\",GEOGCS[\"GCS_IGN_Astro_1960\",DATUM[\"D_IGN_Astro_1960\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3368]]",
        3369=>"PROJCS[\"IGN_Astro_1960_UTM_Zone_30N\",GEOGCS[\"GCS_IGN_Astro_1960\",DATUM[\"D_IGN_Astro_1960\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3369]]",
        3370=>"PROJCS[\"NAD_1927_UTM_Zone_59N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3370]]",
        3371=>"PROJCS[\"NAD_1927_UTM_Zone_60N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3371]]",
        3372=>"PROJCS[\"NAD_1983_UTM_Zone_59N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3372]]",
        3373=>"PROJCS[\"NAD_1983_UTM_Zone_60N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3373]]",
        3374=>"PROJCS[\"FD_1954_UTM_Zone_29N\",GEOGCS[\"GCS_FD_1954\",DATUM[\"D_Faroe_Datum_1954\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3374]]",
        3375=>"PROJCS[\"GDM_2000_MRSO_Peninsular_Malaysia\",GEOGCS[\"GCS_GDM_2000\",DATUM[\"D_GDM_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Rectified_Skew_Orthomorphic_Natural_Origin\"],PARAMETER[\"False_Easting\",804671.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Scale_Factor\",0.99984],PARAMETER[\"Azimuth\",323.0257964666666],PARAMETER[\"Longitude_Of_Center\",102.25],PARAMETER[\"Latitude_Of_Center\",4.0],PARAMETER[\"XY_Plane_Rotation\",-36.86989764584402],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3375]]",
        3376=>"PROJCS[\"GDM_2000_BRSO_East_Malaysia\",GEOGCS[\"GCS_GDM_2000\",DATUM[\"D_GDM_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Rectified_Skew_Orthomorphic_Natural_Origin\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Scale_Factor\",0.99984],PARAMETER[\"Azimuth\",53.31580995],PARAMETER[\"Longitude_Of_Center\",115.0],PARAMETER[\"Latitude_Of_Center\",4.0],PARAMETER[\"XY_Plane_Rotation\",53.13010235415598],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3376]]",
        3377=>"PROJCS[\"GDM_2000_State_Cassini_Johor\",GEOGCS[\"GCS_GDM_2000\",DATUM[\"D_GDM_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",-14810.562],PARAMETER[\"False_Northing\",8758.32],PARAMETER[\"Central_Meridian\",103.4279362361111],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",2.121679744444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3377]]",
        3378=>"PROJCS[\"GDM_2000_State_Cassini_Negeri_Sembilan_and_Melaka\",GEOGCS[\"GCS_GDM_2000\",DATUM[\"D_GDM_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",3673.785],PARAMETER[\"False_Northing\",-4240.573],PARAMETER[\"Central_Meridian\",101.9749050416667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",2.682347636111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3378]]",
        3379=>"PROJCS[\"GDM_2000_State_Cassini_Pahang\",GEOGCS[\"GCS_GDM_2000\",DATUM[\"D_GDM_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",-7368.228],PARAMETER[\"False_Northing\",6485.858],PARAMETER[\"Central_Meridian\",102.3682989833333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",3.769388088888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3379]]",
        3380=>"PROJCS[\"GDM_2000_State_Cassini_Selangor\",GEOGCS[\"GCS_GDM_2000\",DATUM[\"D_GDM_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",-34836.161],PARAMETER[\"False_Northing\",56464.049],PARAMETER[\"Central_Meridian\",101.3891079138889],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",3.68464905],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3380]]",
        3381=>"PROJCS[\"GDM_2000_State_Cassini_Terengganu\",GEOGCS[\"GCS_GDM_2000\",DATUM[\"D_GDM_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",19594.245],PARAMETER[\"False_Northing\",3371.895],PARAMETER[\"Central_Meridian\",103.070275625],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",4.9762852],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3381]]",
        3382=>"PROJCS[\"GDM_2000_State_Cassini_Pulau_Pinang_and_Seberang_Perai\",GEOGCS[\"GCS_GDM_2000\",DATUM[\"D_GDM_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",-23.414],PARAMETER[\"False_Northing\",62.283],PARAMETER[\"Central_Meridian\",100.3443769638889],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",5.421517541666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3382]]",
        3383=>"PROJCS[\"GDM_2000_State_Cassini_Perlis\",GEOGCS[\"GCS_GDM_2000\",DATUM[\"D_GDM_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",100.6363711111111],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",5.964672713888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3383]]",
        3384=>"PROJCS[\"GDM_2000_State_Cassini_Perak\",GEOGCS[\"GCS_GDM_2000\",DATUM[\"D_GDM_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",-1.769],PARAMETER[\"False_Northing\",133454.779],PARAMETER[\"Central_Meridian\",100.8154105861111],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",4.859063022222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3384]]",
        3385=>"PROJCS[\"GDM_2000_State_Cassini_Kelantan\",GEOGCS[\"GCS_GDM_2000\",DATUM[\"D_GDM_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",13227.851],PARAMETER[\"False_Northing\",8739.894],PARAMETER[\"Central_Meridian\",102.2952416694444],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",5.972543658333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3385]]",
        3386=>"PROJCS[\"KKJ_Finland_Zone_0\",GEOGCS[\"GCS_KKJ\",DATUM[\"D_KKJ\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",18.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3386]]",
        3387=>"PROJCS[\"KKJ_Finland_Zone_5\",GEOGCS[\"GCS_KKJ\",DATUM[\"D_KKJ\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3387]]",
        3388=>"PROJCS[\"Pulkovo_1942_Caspian_Sea_Mercator\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Standard_Parallel_1\",42.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3388]]",
        3389=>"PROJCS[\"Pulkovo_1942_3_Degree_GK_Zone_60\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",60500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",180.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3389]]",
        3390=>"PROJCS[\"Pulkovo_1995_3_Degree_GK_Zone_60\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",60500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",180.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3390]]",
        3391=>"PROJCS[\"Karbala_1979_Polservice_UTM_Zone_37N\",GEOGCS[\"GCS_Karbala_1979_Polservice\",DATUM[\"D_Karbala_1979_Polservice\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3391]]",
        3392=>"PROJCS[\"Karbala_1979_Polservice_UTM_Zone_38N\",GEOGCS[\"GCS_Karbala_1979_Polservice\",DATUM[\"D_Karbala_1979_Polservice\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3392]]",
        3393=>"PROJCS[\"Karbala_1979_Polservice_UTM_Zone_39N\",GEOGCS[\"GCS_Karbala_1979_Polservice\",DATUM[\"D_Karbala_1979_Polservice\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3393]]",
        3394=>"PROJCS[\"Nahrwan_1934_Iraq_Zone\",GEOGCS[\"GCS_Nahrwan_1934\",DATUM[\"D_Nahrwan_1934\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",1166200.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Standard_Parallel_1\",32.5],PARAMETER[\"Scale_Factor\",0.9987864078],PARAMETER[\"Latitude_Of_Origin\",32.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3394]]",
        3395=>"PROJCS[\"WGS_1984_World_Mercator\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3395]]",
        3396=>"PROJCS[\"PD/83_GK_Zone_3\",GEOGCS[\"GCS_PD/83\",DATUM[\"D_Potsdam_1983\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3396]]",
        3397=>"PROJCS[\"PD/83_GK_Zone_4\",GEOGCS[\"GCS_PD/83\",DATUM[\"D_Potsdam_1983\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3397]]",
        3398=>"PROJCS[\"RD/83_GK_Zone_4\",GEOGCS[\"GCS_RD/83\",DATUM[\"D_Rauenberg_1983\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3398]]",
        3399=>"PROJCS[\"RD/83_GK_Zone_5\",GEOGCS[\"GCS_RD/83\",DATUM[\"D_Rauenberg_1983\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3399]]",
        3400=>"PROJCS[\"NAD_1983_10TM_AEP_Forest\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-115.0],PARAMETER[\"Scale_Factor\",0.9992],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3400]]",
        3401=>"PROJCS[\"NAD_1983_10TM_AEP_Resource\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-115.0],PARAMETER[\"Scale_Factor\",0.9992],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3401]]",
        3402=>"PROJCS[\"NAD_1983_CSRS_10TM_AEP_Forest\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-115.0],PARAMETER[\"Scale_Factor\",0.9992],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3402]]",
        3403=>"PROJCS[\"NAD_1983_CSRS_10TM_AEP_Resource\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-115.0],PARAMETER[\"Scale_Factor\",0.9992],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3403]]",
        3404=>"PROJCS[\"NAD_1983_HARN_StatePlane_North_Carolina_FIPS_3200_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.0],PARAMETER[\"Standard_Parallel_1\",34.33333333333334],PARAMETER[\"Standard_Parallel_2\",36.16666666666666],PARAMETER[\"Latitude_Of_Origin\",33.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3404]]",
        3405=>"PROJCS[\"VN_2000_UTM_Zone_48N\",GEOGCS[\"GCS_VN_2000\",DATUM[\"D_Vietnam_2000\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3405]]",
        3406=>"PROJCS[\"VN_2000_UTM_Zone_49N\",GEOGCS[\"GCS_VN_2000\",DATUM[\"D_Vietnam_2000\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3406]]",
        3407=>"PROJCS[\"Hong_Kong_1963_Grid_System\",GEOGCS[\"GCS_Hong_Kong_1963\",DATUM[\"D_Hong_Kong_1963\",SPHEROID[\"Clarke_1858\",6378293.645208759,294.260676369]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",132033.92],PARAMETER[\"False_Northing\",62565.96],PARAMETER[\"Central_Meridian\",114.1785555555556],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",22.31213333333334],UNIT[\"Foot_Clarke\",0.3047972654],AUTHORITY[\"EPSG\",3407]]",
        3408=>"PROJCS[\"NSIDC_EASE_Grid_North\",GEOGCS[\"GCS_Sphere_International_1924_Authalic\",DATUM[\"D_Sphere_International_1924_Authalic\",SPHEROID[\"Sphere_International_1924_Authalic\",6371228.0,0.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Latitude_Of_Origin\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3408]]",
        3409=>"PROJCS[\"NSIDC_EASE_Grid_South\",GEOGCS[\"GCS_Sphere_International_1924_Authalic\",DATUM[\"D_Sphere_International_1924_Authalic\",SPHEROID[\"Sphere_International_1924_Authalic\",6371228.0,0.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3409]]",
        3410=>"PROJCS[\"NSIDC_EASE_Grid_Global\",GEOGCS[\"GCS_Sphere_International_1924_Authalic\",DATUM[\"D_Sphere_International_1924_Authalic\",SPHEROID[\"Sphere_International_1924_Authalic\",6371228.0,0.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cylindrical_Equal_Area\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3410]]",
        3411=>"PROJCS[\"NSIDC_Sea_Ice_Polar_Stereographic_North\",GEOGCS[\"GCS_Hughes_1980\",DATUM[\"D_Hughes_1980\",SPHEROID[\"Hughes_1980\",6378273.0,298.279411123064]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic_North_Pole\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Standard_Parallel_1\",70.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3411]]",
        3412=>"PROJCS[\"NSIDC_Sea_Ice_Polar_Stereographic_South\",GEOGCS[\"GCS_Hughes_1980\",DATUM[\"D_Hughes_1980\",SPHEROID[\"Hughes_1980\",6378273.0,298.279411123064]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic_South_Pole\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",-70.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3412]]",
        3413=>"PROJCS[\"WGS_1984_NSIDC_Sea_Ice_Polar_Stereographic_North\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic_North_Pole\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Standard_Parallel_1\",70.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3413]]",
        3414=>"PROJCS[\"SVY21_Singapore_TM\",GEOGCS[\"GCS_SVY21\",DATUM[\"D_SVY21\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",28001.642],PARAMETER[\"False_Northing\",38744.572],PARAMETER[\"Central_Meridian\",103.8333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",1.366666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3414]]",
        3415=>"PROJCS[\"WGS_1972_BE_South_China_Sea_Lambert\",GEOGCS[\"GCS_WGS_1972_BE\",DATUM[\"D_WGS_1972_BE\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",114.0],PARAMETER[\"Standard_Parallel_1\",18.0],PARAMETER[\"Standard_Parallel_2\",24.0],PARAMETER[\"Latitude_Of_Origin\",21.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3415]]",
        3416=>"PROJCS[\"ETRS_1989_Austria_Lambert\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",400000.0],PARAMETER[\"Central_Meridian\",13.33333333333333],PARAMETER[\"Standard_Parallel_1\",46.0],PARAMETER[\"Standard_Parallel_2\",49.0],PARAMETER[\"Latitude_Of_Origin\",47.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3416]]",
        3417=>"PROJCS[\"NAD_1983_StatePlane_Iowa_North_FIPS_1401_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4921250.0],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",42.06666666666667],PARAMETER[\"Standard_Parallel_2\",43.26666666666667],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3417]]",
        3418=>"PROJCS[\"NAD_1983_StatePlane_Iowa_South_FIPS_1402_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",40.61666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3418]]",
        3419=>"PROJCS[\"NAD_1983_StatePlane_Kansas_North_FIPS_1501_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",38.71666666666667],PARAMETER[\"Standard_Parallel_2\",39.78333333333333],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3419]]",
        3420=>"PROJCS[\"NAD_1983_StatePlane_Kansas_South_FIPS_1502_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",1312333.333333333],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",37.26666666666667],PARAMETER[\"Standard_Parallel_2\",38.56666666666667],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3420]]",
        3421=>"PROJCS[\"NAD_1983_StatePlane_Nevada_East_FIPS_2701_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",26246666.66666666],PARAMETER[\"Central_Meridian\",-115.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3421]]",
        3422=>"PROJCS[\"NAD_1983_StatePlane_Nevada_Central_FIPS_2702_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",19685000.0],PARAMETER[\"Central_Meridian\",-116.6666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3422]]",
        3423=>"PROJCS[\"NAD_1983_StatePlane_Nevada_West_FIPS_2703_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",13123333.33333333],PARAMETER[\"Central_Meridian\",-118.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3423]]",
        3424=>"PROJCS[\"NAD_1983_StatePlane_New_Jersey_FIPS_2900_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",492125.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",38.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3424]]",
        3425=>"PROJCS[\"NAD_1983_HARN_StatePlane_Iowa_North_FIPS_1401_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4921250.0],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",42.06666666666667],PARAMETER[\"Standard_Parallel_2\",43.26666666666667],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3425]]",
        3426=>"PROJCS[\"NAD_1983_HARN_StatePlane_Iowa_South_FIPS_1402_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",40.61666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3426]]",
        3427=>"PROJCS[\"NAD_1983_HARN_StatePlane_Kansas_North_FIPS_1501_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",38.71666666666667],PARAMETER[\"Standard_Parallel_2\",39.78333333333333],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3427]]",
        3428=>"PROJCS[\"NAD_1983_HARN_StatePlane_Kansas_South_FIPS_1502_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",1312333.333333333],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",37.26666666666667],PARAMETER[\"Standard_Parallel_2\",38.56666666666667],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3428]]",
        3429=>"PROJCS[\"NAD_1983_HARN_StatePlane_Nevada_East_FIPS_2701_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",26246666.66666666],PARAMETER[\"Central_Meridian\",-115.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3429]]",
        3430=>"PROJCS[\"NAD_1983_HARN_StatePlane_Nevada_Central_FIPS_2702_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",19685000.0],PARAMETER[\"Central_Meridian\",-116.6666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3430]]",
        3431=>"PROJCS[\"NAD_1983_HARN_StatePlane_Nevada_West_FIPS_2703_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",13123333.33333333],PARAMETER[\"Central_Meridian\",-118.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3431]]",
        3432=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_Jersey_FIPS_2900_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",492125.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",38.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3432]]",
        3433=>"PROJCS[\"NAD_1983_StatePlane_Arkansas_North_FIPS_0301_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",34.93333333333333],PARAMETER[\"Standard_Parallel_2\",36.23333333333333],PARAMETER[\"Latitude_Of_Origin\",34.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3433]]",
        3434=>"PROJCS[\"NAD_1983_StatePlane_Arkansas_South_FIPS_0302_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",1312333.333333333],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",33.3],PARAMETER[\"Standard_Parallel_2\",34.76666666666667],PARAMETER[\"Latitude_Of_Origin\",32.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3434]]",
        3435=>"PROJCS[\"NAD_1983_StatePlane_Illinois_East_FIPS_1201_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.33333333333333],PARAMETER[\"Scale_Factor\",0.999975],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3435]]",
        3436=>"PROJCS[\"NAD_1983_StatePlane_Illinois_West_FIPS_1202_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2296583.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.16666666666667],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3436]]",
        3437=>"PROJCS[\"NAD_1983_StatePlane_New_Hampshire_FIPS_2800_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3437]]",
        3438=>"PROJCS[\"NAD_1983_StatePlane_Rhode_Island_FIPS_3800_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",328083.3333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Scale_Factor\",0.99999375],PARAMETER[\"Latitude_Of_Origin\",41.08333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3438]]",
        3439=>"PROJCS[\"PDO_1993_UTM_Zone_39N\",GEOGCS[\"GCS_PDO_1993\",DATUM[\"D_PDO_1993\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3439]]",
        3440=>"PROJCS[\"PDO_1993_UTM_Zone_40N\",GEOGCS[\"GCS_PDO_1993\",DATUM[\"D_PDO_1993\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3440]]",
        3441=>"PROJCS[\"NAD_1983_HARN_StatePlane_Arkansas_North_FIPS_0301_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",34.93333333333333],PARAMETER[\"Standard_Parallel_2\",36.23333333333333],PARAMETER[\"Latitude_Of_Origin\",34.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3441]]",
        3442=>"PROJCS[\"NAD_1983_HARN_StatePlane_Arkansas_South_FIPS_0302_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",1312333.333333333],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",33.3],PARAMETER[\"Standard_Parallel_2\",34.76666666666667],PARAMETER[\"Latitude_Of_Origin\",32.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3442]]",
        3443=>"PROJCS[\"NAD_1983_HARN_StatePlane_Illinois_East_FIPS_1201_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.33333333333333],PARAMETER[\"Scale_Factor\",0.999975],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3443]]",
        3444=>"PROJCS[\"NAD_1983_HARN_StatePlane_Illinois_West_FIPS_1202_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2296583.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.16666666666667],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3444]]",
        3445=>"PROJCS[\"NAD_1983_HARN_StatePlane_New_Hampshire_FIPS_2800_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3445]]",
        3446=>"PROJCS[\"NAD_1983_HARN_StatePlane_Rhode_Island_FIPS_3800_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",328083.3333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Scale_Factor\",0.99999375],PARAMETER[\"Latitude_Of_Origin\",41.08333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3446]]",
        3447=>"PROJCS[\"Belge_Lambert_2005\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",150328.0],PARAMETER[\"False_Northing\",166262.0],PARAMETER[\"Central_Meridian\",4.359215833333333],PARAMETER[\"Standard_Parallel_1\",49.83333333333334],PARAMETER[\"Standard_Parallel_2\",51.16666666666666],PARAMETER[\"Latitude_Of_Origin\",50.797815],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3447]]",
        3448=>"PROJCS[\"JAD_2001_Jamaica_Grid\",GEOGCS[\"GCS_JAD_2001\",DATUM[\"D_Jamaica_2001\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",750000.0],PARAMETER[\"False_Northing\",650000.0],PARAMETER[\"Central_Meridian\",-77.0],PARAMETER[\"Standard_Parallel_1\",18.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",18.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3448]]",
        3449=>"PROJCS[\"JAD_2001_UTM_Zone_17N\",GEOGCS[\"GCS_JAD_2001\",DATUM[\"D_Jamaica_2001\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3449]]",
        3450=>"PROJCS[\"JAD_2001_UTM_Zone_18N\",GEOGCS[\"GCS_JAD_2001\",DATUM[\"D_Jamaica_2001\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3450]]",
        3451=>"PROJCS[\"NAD_1983_StatePlane_Louisiana_North_FIPS_1701_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3280833.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.5],PARAMETER[\"Standard_Parallel_1\",31.16666666666667],PARAMETER[\"Standard_Parallel_2\",32.66666666666666],PARAMETER[\"Latitude_Of_Origin\",30.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3451]]",
        3452=>"PROJCS[\"NAD_1983_StatePlane_Louisiana_South_FIPS_1702_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3280833.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-91.33333333333333],PARAMETER[\"Standard_Parallel_1\",29.3],PARAMETER[\"Standard_Parallel_2\",30.7],PARAMETER[\"Latitude_Of_Origin\",28.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3452]]",
        3453=>"PROJCS[\"NAD_1983_StatePlane_Louisiana_Offshore_FIPS_1703_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3280833.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-91.33333333333333],PARAMETER[\"Standard_Parallel_1\",26.16666666666667],PARAMETER[\"Standard_Parallel_2\",27.83333333333333],PARAMETER[\"Latitude_Of_Origin\",25.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3453]]",
        3454=>"PROJCS[\"NAD_1983_StatePlane_South_Dakota_North_FIPS_4001_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",44.41666666666666],PARAMETER[\"Standard_Parallel_2\",45.68333333333333],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4457]]",
        3455=>"PROJCS[\"NAD_1983_StatePlane_South_Dakota_South_FIPS_4002_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",42.83333333333334],PARAMETER[\"Standard_Parallel_2\",44.4],PARAMETER[\"Latitude_Of_Origin\",42.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3455]]",
        3456=>"PROJCS[\"NAD_1983_HARN_StatePlane_Louisiana_North_FIPS_1701_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3280833.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.5],PARAMETER[\"Standard_Parallel_1\",31.16666666666667],PARAMETER[\"Standard_Parallel_2\",32.66666666666666],PARAMETER[\"Latitude_Of_Origin\",30.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3456]]",
        3457=>"PROJCS[\"NAD_1983_HARN_StatePlane_Louisiana_South_FIPS_1702_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3280833.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-91.33333333333333],PARAMETER[\"Standard_Parallel_1\",29.3],PARAMETER[\"Standard_Parallel_2\",30.7],PARAMETER[\"Latitude_Of_Origin\",28.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3457]]",
        3458=>"PROJCS[\"NAD_1983_HARN_StatePlane_South_Dakota_North_FIPS_4001_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",44.41666666666666],PARAMETER[\"Standard_Parallel_2\",45.68333333333333],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3458]]",
        3459=>"PROJCS[\"NAD_1983_HARN_StatePlane_South_Dakota_South_FIPS_4002_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",42.83333333333334],PARAMETER[\"Standard_Parallel_2\",44.4],PARAMETER[\"Latitude_Of_Origin\",42.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3459]]",
        3460=>"PROJCS[\"Fiji_1986_Fiji_Map_Grid\",GEOGCS[\"GCS_Fiji_1986\",DATUM[\"D_Fiji_1986\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",4000000.0],PARAMETER[\"Central_Meridian\",178.75],PARAMETER[\"Scale_Factor\",0.99985],PARAMETER[\"Latitude_Of_Origin\",-17.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3460]]",
        3461=>"PROJCS[\"Dabola_1981_UTM_Zone_28N\",GEOGCS[\"GCS_Dabola_1981\",DATUM[\"D_Dabola_1981\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3461]]",
        3462=>"PROJCS[\"Dabola_1981_UTM_Zone_29N\",GEOGCS[\"GCS_Dabola_1981\",DATUM[\"D_Dabola_1981\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3462]]",
        3463=>"PROJCS[\"NAD_1983_Maine_2000_Central_Zone\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.125],PARAMETER[\"Scale_Factor\",0.99998],PARAMETER[\"Latitude_Of_Origin\",43.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3463]]",
        3464=>"PROJCS[\"NAD_1983_HARN_Maine_2000_Central_Zone\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.125],PARAMETER[\"Scale_Factor\",0.99998],PARAMETER[\"Latitude_Of_Origin\",43.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3464]]",
        3465=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Alabama_East_FIPS_0101\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-85.83333333333333],PARAMETER[\"Scale_Factor\",0.99996],PARAMETER[\"Latitude_Of_Origin\",30.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3465]]",
        3466=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Alabama_West_FIPS_0102\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.5],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3466]]",
        3467=>"PROJCS[\"NAD_1983_NSRS2007_Alaska_Albers\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-154.0],PARAMETER[\"Standard_Parallel_1\",55.0],PARAMETER[\"Standard_Parallel_2\",65.0],PARAMETER[\"Latitude_Of_Origin\",50.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3467]]",
        3468=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Alaska_1_FIPS_5001\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Natural_Origin\"],PARAMETER[\"False_Easting\",5000000.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Azimuth\",-36.86989764583333],PARAMETER[\"Longitude_Of_Center\",-133.6666666666667],PARAMETER[\"Latitude_Of_Center\",57.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3468]]",
        3469=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Alaska_2_FIPS_5002\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-142.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3469]]",
        3470=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Alaska_3_FIPS_5003\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-146.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3470]]",
        3471=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Alaska_4_FIPS_5004\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-150.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3471]]",
        3472=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Alaska_5_FIPS_5005\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-154.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3472]]",
        3473=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Alaska_6_FIPS_5006\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-158.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3473]]",
        3474=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Alaska_7_FIPS_5007\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-162.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3474]]",
        3475=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Alaska_8_FIPS_5008\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-166.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3475]]",
        3476=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Alaska_9_FIPS_5009\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-170.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3476]]",
        3477=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Alaska_10_FIPS_5010\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-176.0],PARAMETER[\"Standard_Parallel_1\",51.83333333333334],PARAMETER[\"Standard_Parallel_2\",53.83333333333334],PARAMETER[\"Latitude_Of_Origin\",51.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3477]]",
        3478=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Arizona_Central_FIPS_0202\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",213360.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.9166666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3478]]",
        3479=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Arizona_Central_FIPS_0202_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.9166666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3479]]",
        3480=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Arizona_East_FIPS_0201\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",213360.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-110.1666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3480]]",
        3481=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Arizona_East_FIPS_0201_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-110.1666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3481]]",
        3482=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Arizona_West_FIPS_0203\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",213360.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-113.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3482]]",
        3483=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Arizona_West_FIPS_0203_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-113.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3483]]",
        3484=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Arkansas_North_FIPS_0301\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",34.93333333333333],PARAMETER[\"Standard_Parallel_2\",36.23333333333333],PARAMETER[\"Latitude_Of_Origin\",34.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3484]]",
        3485=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Arkansas_North_FIPS_0301_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",34.93333333333333],PARAMETER[\"Standard_Parallel_2\",36.23333333333333],PARAMETER[\"Latitude_Of_Origin\",34.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3485]]",
        3486=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Arkansas_South_FIPS_0302\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",400000.0],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",33.3],PARAMETER[\"Standard_Parallel_2\",34.76666666666667],PARAMETER[\"Latitude_Of_Origin\",32.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3486]]",
        3487=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Arkansas_South_FIPS_0302_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",1312333.333333333],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",33.3],PARAMETER[\"Standard_Parallel_2\",34.76666666666667],PARAMETER[\"Latitude_Of_Origin\",32.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3487]]",
        3488=>"PROJCS[\"NAD_1983_NSRS2007_California_Teale_Albers\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",-4000000.0],PARAMETER[\"Central_Meridian\",-120.0],PARAMETER[\"Standard_Parallel_1\",34.0],PARAMETER[\"Standard_Parallel_2\",40.5],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3488]]",
        3489=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_California_I_FIPS_0401\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Standard_Parallel_2\",41.66666666666666],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3489]]",
        3490=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_California_I_FIPS_0401_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Standard_Parallel_2\",41.66666666666666],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3490]]",
        3491=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_California_II_FIPS_0402\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",38.33333333333334],PARAMETER[\"Standard_Parallel_2\",39.83333333333334],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3491]]",
        3492=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_California_II_FIPS_0402_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",38.33333333333334],PARAMETER[\"Standard_Parallel_2\",39.83333333333334],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3492]]",
        3493=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_California_III_FIPS_0403\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",37.06666666666667],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3493]]",
        3494=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_California_III_FIPS_0403_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",37.06666666666667],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3494]]",
        3495=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_California_IV_FIPS_0404\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-119.0],PARAMETER[\"Standard_Parallel_1\",36.0],PARAMETER[\"Standard_Parallel_2\",37.25],PARAMETER[\"Latitude_Of_Origin\",35.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3495]]",
        3496=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_California_IV_FIPS_0404_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-119.0],PARAMETER[\"Standard_Parallel_1\",36.0],PARAMETER[\"Standard_Parallel_2\",37.25],PARAMETER[\"Latitude_Of_Origin\",35.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3496]]",
        3497=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_California_V_FIPS_0405\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-118.0],PARAMETER[\"Standard_Parallel_1\",34.03333333333333],PARAMETER[\"Standard_Parallel_2\",35.46666666666667],PARAMETER[\"Latitude_Of_Origin\",33.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3497]]",
        3498=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_California_V_FIPS_0405_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-118.0],PARAMETER[\"Standard_Parallel_1\",34.03333333333333],PARAMETER[\"Standard_Parallel_2\",35.46666666666667],PARAMETER[\"Latitude_Of_Origin\",33.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3498]]",
        3499=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_California_VI_FIPS_0406\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-116.25],PARAMETER[\"Standard_Parallel_1\",32.78333333333333],PARAMETER[\"Standard_Parallel_2\",33.88333333333333],PARAMETER[\"Latitude_Of_Origin\",32.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3499]]",
        3500=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_California_VI_FIPS_0406_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-116.25],PARAMETER[\"Standard_Parallel_1\",32.78333333333333],PARAMETER[\"Standard_Parallel_2\",33.88333333333333],PARAMETER[\"Latitude_Of_Origin\",32.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3500]]",
        3501=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Colorado_Central_FIPS_0502\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",914401.8289],PARAMETER[\"False_Northing\",304800.6096],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",38.45],PARAMETER[\"Standard_Parallel_2\",39.75],PARAMETER[\"Latitude_Of_Origin\",37.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3501]]",
        3502=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Colorado_Central_FIPS_0502_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.000316083],PARAMETER[\"False_Northing\",999999.999996],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",38.45],PARAMETER[\"Standard_Parallel_2\",39.75],PARAMETER[\"Latitude_Of_Origin\",37.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3502]]",
        3503=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Colorado_North_FIPS_0501\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",914401.8289],PARAMETER[\"False_Northing\",304800.6096],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",39.71666666666667],PARAMETER[\"Standard_Parallel_2\",40.78333333333333],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3503]]",
        3504=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Colorado_North_FIPS_0501_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.000316083],PARAMETER[\"False_Northing\",999999.999996],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",39.71666666666667],PARAMETER[\"Standard_Parallel_2\",40.78333333333333],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3504]]",
        3505=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Colorado_South_FIPS_0503\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",914401.8289],PARAMETER[\"False_Northing\",304800.6096],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",37.23333333333333],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3505]]",
        3506=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Colorado_South_FIPS_0503_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.000316083],PARAMETER[\"False_Northing\",999999.999996],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",37.23333333333333],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3506]]",
        3507=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Connecticut_FIPS_0600\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",304800.6096],PARAMETER[\"False_Northing\",152400.3048],PARAMETER[\"Central_Meridian\",-72.75],PARAMETER[\"Standard_Parallel_1\",41.2],PARAMETER[\"Standard_Parallel_2\",41.86666666666667],PARAMETER[\"Latitude_Of_Origin\",40.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3507]]",
        3508=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Connecticut_FIPS_0600_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",999999.999996],PARAMETER[\"False_Northing\",499999.999998],PARAMETER[\"Central_Meridian\",-72.75],PARAMETER[\"Standard_Parallel_1\",41.2],PARAMETER[\"Standard_Parallel_2\",41.86666666666667],PARAMETER[\"Latitude_Of_Origin\",40.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3508]]",
        3509=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Delaware_FIPS_0700\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.41666666666667],PARAMETER[\"Scale_Factor\",0.999995],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3509]]",
        3510=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Delaware_FIPS_0700_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.41666666666667],PARAMETER[\"Scale_Factor\",0.999995],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3510]]",
        3511=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Florida_East_FIPS_0901\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3511]]",
        3512=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Florida_East_FIPS_0901_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3512]]",
        3513=>"PROJCS[\"NAD_1983_NSRS2007_Florida_GDL_Albers\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.0],PARAMETER[\"Standard_Parallel_1\",24.0],PARAMETER[\"Standard_Parallel_2\",31.5],PARAMETER[\"Latitude_Of_Origin\",24.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3513]]",
        3514=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Florida_North_FIPS_0903\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.5],PARAMETER[\"Standard_Parallel_1\",29.58333333333333],PARAMETER[\"Standard_Parallel_2\",30.75],PARAMETER[\"Latitude_Of_Origin\",29.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3514]]",
        3515=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Florida_North_FIPS_0903_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.5],PARAMETER[\"Standard_Parallel_1\",29.58333333333333],PARAMETER[\"Standard_Parallel_2\",30.75],PARAMETER[\"Latitude_Of_Origin\",29.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3515]]",
        3516=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Florida_West_FIPS_0902\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3516]]",
        3517=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Florida_West_FIPS_0902_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3517]]",
        3518=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Georgia_East_FIPS_1001\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3518]]",
        3519=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Georgia_East_FIPS_1001_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3519]]",
        3520=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Georgia_West_FIPS_1002\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3520]]",
        3521=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Georgia_West_FIPS_1002_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2296583.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3521]]",
        3522=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Idaho_Central_FIPS_1102\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-114.0],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3522]]",
        3523=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Idaho_Central_FIPS_1102_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-114.0],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3523]]",
        3524=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Idaho_East_FIPS_1101\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-112.1666666666667],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3524]]",
        3525=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Idaho_East_FIPS_1101_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-112.1666666666667],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3525]]",
        3526=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Idaho_West_FIPS_1103\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-115.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3526]]",
        3527=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Idaho_West_FIPS_1103_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-115.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3527]]",
        3528=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Illinois_East_FIPS_1201\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.33333333333333],PARAMETER[\"Scale_Factor\",0.999975],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3528]]",
        3529=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Illinois_East_FIPS_1201_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.33333333333333],PARAMETER[\"Scale_Factor\",0.999975],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3529]]",
        3530=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Illinois_West_FIPS_1202\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.16666666666667],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3530]]",
        3531=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Illinois_West_FIPS_1202_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2296583.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.16666666666667],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3531]]",
        3532=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Indiana_East_FIPS_1301\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",250000.0],PARAMETER[\"Central_Meridian\",-85.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3532]]",
        3533=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Indiana_East_FIPS_1301_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",328083.3333333333],PARAMETER[\"False_Northing\",820208.3333333333],PARAMETER[\"Central_Meridian\",-85.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3533]]",
        3534=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Indiana_West_FIPS_1302\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",900000.0],PARAMETER[\"False_Northing\",250000.0],PARAMETER[\"Central_Meridian\",-87.08333333333333],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3534]]",
        3535=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Indiana_West_FIPS_1302_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2952750.0],PARAMETER[\"False_Northing\",820208.3333333333],PARAMETER[\"Central_Meridian\",-87.08333333333333],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3535]]",
        3536=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Iowa_North_FIPS_1401\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",42.06666666666667],PARAMETER[\"Standard_Parallel_2\",43.26666666666667],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3536]]",
        3537=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Iowa_North_FIPS_1401_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4921250.0],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",42.06666666666667],PARAMETER[\"Standard_Parallel_2\",43.26666666666667],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3537]]",
        3538=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Iowa_South_FIPS_1402\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",40.61666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3538]]",
        3539=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Iowa_South_FIPS_1402_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",40.61666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3539]]",
        3540=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Kansas_North_FIPS_1501\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",38.71666666666667],PARAMETER[\"Standard_Parallel_2\",39.78333333333333],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3540]]",
        3541=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Kansas_North_FIPS_1501_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",38.71666666666667],PARAMETER[\"Standard_Parallel_2\",39.78333333333333],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3541]]",
        3542=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Kansas_South_FIPS_1502\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",400000.0],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",37.26666666666667],PARAMETER[\"Standard_Parallel_2\",38.56666666666667],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3542]]",
        3543=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Kansas_South_FIPS_1502_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",1312333.333333333],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",37.26666666666667],PARAMETER[\"Standard_Parallel_2\",38.56666666666667],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3543]]",
        3544=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Kentucky_North_FIPS_1601\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.25],PARAMETER[\"Standard_Parallel_1\",37.96666666666667],PARAMETER[\"Standard_Parallel_2\",38.96666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3544]]",
        3545=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Kentucky_North_FIPS_1601_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.25],PARAMETER[\"Standard_Parallel_1\",37.96666666666667],PARAMETER[\"Standard_Parallel_2\",38.96666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3545]]",
        3546=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Kentucky_FIPS_1600\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Standard_Parallel_1\",37.08333333333334],PARAMETER[\"Standard_Parallel_2\",38.66666666666666],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3546]]",
        3547=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Kentucky_FIPS_1600_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4921250.0],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Standard_Parallel_1\",37.08333333333334],PARAMETER[\"Standard_Parallel_2\",38.66666666666666],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3547]]",
        3548=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Kentucky_South_FIPS_1602\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Standard_Parallel_1\",36.73333333333333],PARAMETER[\"Standard_Parallel_2\",37.93333333333333],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3548]]",
        3549=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Kentucky_South_FIPS_1602_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Standard_Parallel_1\",36.73333333333333],PARAMETER[\"Standard_Parallel_2\",37.93333333333333],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3549]]",
        3550=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Louisiana_North_FIPS_1701\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.5],PARAMETER[\"Standard_Parallel_1\",31.16666666666667],PARAMETER[\"Standard_Parallel_2\",32.66666666666666],PARAMETER[\"Latitude_Of_Origin\",30.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3550]]",
        3551=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Louisiana_North_FIPS_1701_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3280833.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.5],PARAMETER[\"Standard_Parallel_1\",31.16666666666667],PARAMETER[\"Standard_Parallel_2\",32.66666666666666],PARAMETER[\"Latitude_Of_Origin\",30.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3551]]",
        3552=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Louisiana_South_FIPS_1702\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-91.33333333333333],PARAMETER[\"Standard_Parallel_1\",29.3],PARAMETER[\"Standard_Parallel_2\",30.7],PARAMETER[\"Latitude_Of_Origin\",28.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3552]]",
        3553=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Louisiana_South_FIPS_1702_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3280833.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-91.33333333333333],PARAMETER[\"Standard_Parallel_1\",29.3],PARAMETER[\"Standard_Parallel_2\",30.7],PARAMETER[\"Latitude_Of_Origin\",28.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3553]]",
        3554=>"PROJCS[\"NAD_1983_NSRS2007_Maine_2000_Central_Zone\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.125],PARAMETER[\"Scale_Factor\",0.99998],PARAMETER[\"Latitude_Of_Origin\",43.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3554]]",
        3555=>"PROJCS[\"NAD_1983_NSRS2007_Maine_2000_East_Zone\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-67.875],PARAMETER[\"Scale_Factor\",0.99998],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3555]]",
        3556=>"PROJCS[\"NAD_1983_NSRS2007_Maine_2000_West_Zone\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.375],PARAMETER[\"Scale_Factor\",0.99998],PARAMETER[\"Latitude_Of_Origin\",42.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3556]]",
        3557=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Maine_East_FIPS_1801\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-68.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3557]]",
        3558=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Maine_West_FIPS_1802\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",900000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.16666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3558]]",
        3559=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Maryland_FIPS_1900\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.0],PARAMETER[\"Standard_Parallel_1\",38.3],PARAMETER[\"Standard_Parallel_2\",39.45],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3559]]",
        3560=>"PROJCS[\"NAD_1983_StatePlane_Utah_North_FIPS_4301_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",40.71666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3560]]",
        3561=>"PROJCS[\"Old_Hawaiian_StatePlane_Hawaii_1_FIPS_5101\",GEOGCS[\"GCS_Old_Hawaiian\",DATUM[\"D_Old_Hawaiian\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-155.5],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",18.83333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3561]]",
        3562=>"PROJCS[\"Old_Hawaiian_StatePlane_Hawaii_2_FIPS_5102\",GEOGCS[\"GCS_Old_Hawaiian\",DATUM[\"D_Old_Hawaiian\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-156.6666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",20.33333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3562]]",
        3563=>"PROJCS[\"Old_Hawaiian_StatePlane_Hawaii_3_FIPS_5103\",GEOGCS[\"GCS_Old_Hawaiian\",DATUM[\"D_Old_Hawaiian\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-158.0],PARAMETER[\"Scale_Factor\",0.99999],PARAMETER[\"Latitude_Of_Origin\",21.16666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3563]]",
        3564=>"PROJCS[\"Old_Hawaiian_StatePlane_Hawaii_4_FIPS_5104\",GEOGCS[\"GCS_Old_Hawaiian\",DATUM[\"D_Old_Hawaiian\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-159.5],PARAMETER[\"Scale_Factor\",0.99999],PARAMETER[\"Latitude_Of_Origin\",21.83333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3564]]",
        3565=>"PROJCS[\"Old_Hawaiian_StatePlane_Hawaii_5_FIPS_5105\",GEOGCS[\"GCS_Old_Hawaiian\",DATUM[\"D_Old_Hawaiian\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-160.1666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",21.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3565]]",
        3566=>"PROJCS[\"NAD_1983_StatePlane_Utah_Central_FIPS_4302_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",6561666.666666666],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",39.01666666666667],PARAMETER[\"Standard_Parallel_2\",40.65],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3566]]",
        3567=>"PROJCS[\"NAD_1983_StatePlane_Utah_South_FIPS_4303_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",9842500.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",37.21666666666667],PARAMETER[\"Standard_Parallel_2\",38.35],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3567]]",
        3568=>"PROJCS[\"NAD_1983_HARN_StatePlane_Utah_North_FIPS_4301_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",40.71666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3568]]",
        3569=>"PROJCS[\"NAD_1983_HARN_StatePlane_Utah_Central_FIPS_4302_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",6561666.666666666],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",39.01666666666667],PARAMETER[\"Standard_Parallel_2\",40.65],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3569]]",
        3570=>"PROJCS[\"NAD_1983_HARN_StatePlane_Utah_South_FIPS_4303_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",9842500.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",37.21666666666667],PARAMETER[\"Standard_Parallel_2\",38.35],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3570]]",
        3571=>"PROJCS[\"WGS_1984_North_Pole_LAEA_Bering_Sea\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",180.0],PARAMETER[\"Latitude_Of_Origin\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3571]]",
        3572=>"PROJCS[\"WGS_1984_North_Pole_LAEA_Alaska\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-150.0],PARAMETER[\"Latitude_Of_Origin\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3572]]",
        3573=>"PROJCS[\"WGS_1984_North_Pole_LAEA_Canada\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Latitude_Of_Origin\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3573]]",
        3574=>"PROJCS[\"WGS_1984_North_Pole_LAEA_Atlantic\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-40.0],PARAMETER[\"Latitude_Of_Origin\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3574]]",
        3575=>"PROJCS[\"WGS_1984_North_Pole_LAEA_Europe\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",10.0],PARAMETER[\"Latitude_Of_Origin\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3575]]",
        3576=>"PROJCS[\"WGS_1984_North_Pole_LAEA_Russia\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Latitude_Of_Origin\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3576]]",
        3577=>"PROJCS[\"GDA_1994_Australia_Albers\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",132.0],PARAMETER[\"Standard_Parallel_1\",-18.0],PARAMETER[\"Standard_Parallel_2\",-36.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3577]]",
        3578=>"PROJCS[\"NAD_1983_Yukon_Albers\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-132.5],PARAMETER[\"Standard_Parallel_1\",61.66666666666666],PARAMETER[\"Standard_Parallel_2\",68.0],PARAMETER[\"Latitude_Of_Origin\",59.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3578]]",
        3579=>"PROJCS[\"NAD_1983_CSRS_Yukon_Albers\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-132.5],PARAMETER[\"Standard_Parallel_1\",61.66666666666666],PARAMETER[\"Standard_Parallel_2\",68.0],PARAMETER[\"Latitude_Of_Origin\",59.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3579]]",
        3580=>"PROJCS[\"NAD_1983_Northwest_Territories_Lambert\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-112.0],PARAMETER[\"Standard_Parallel_1\",62.0],PARAMETER[\"Standard_Parallel_2\",70.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3580]]",
        3581=>"PROJCS[\"NAD_1983_CSRS_Northwest_Territories_Lambert\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-112.0],PARAMETER[\"Standard_Parallel_1\",62.0],PARAMETER[\"Standard_Parallel_2\",70.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3581]]",
        3582=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Maryland_FIPS_1900_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.0],PARAMETER[\"Standard_Parallel_1\",38.3],PARAMETER[\"Standard_Parallel_2\",39.45],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3582]]",
        3583=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Massachusetts_Island_FIPS_2002\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.5],PARAMETER[\"Standard_Parallel_1\",41.28333333333333],PARAMETER[\"Standard_Parallel_2\",41.48333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3583]]",
        3584=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Massachusetts_Isl_FIPS_2002_FtUS\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.5],PARAMETER[\"Standard_Parallel_1\",41.28333333333333],PARAMETER[\"Standard_Parallel_2\",41.48333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3584]]",
        3585=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Massachusetts_Mainland_FIPS_2001\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",750000.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Standard_Parallel_1\",41.71666666666667],PARAMETER[\"Standard_Parallel_2\",42.68333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3585]]",
        3586=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Massachusetts_Mnld_FIPS_2001_FtUS\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",2460625.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Standard_Parallel_1\",41.71666666666667],PARAMETER[\"Standard_Parallel_2\",42.68333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3586]]",
        3587=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Michigan_Central_FIPS_2112\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.36666666666666],PARAMETER[\"Standard_Parallel_1\",44.18333333333333],PARAMETER[\"Standard_Parallel_2\",45.7],PARAMETER[\"Latitude_Of_Origin\",43.31666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3587]]",
        3588=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Michigan_Central_FIPS_2112_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",19685039.37007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.36666666666666],PARAMETER[\"Standard_Parallel_1\",44.18333333333333],PARAMETER[\"Standard_Parallel_2\",45.7],PARAMETER[\"Latitude_Of_Origin\",43.31666666666667],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3588]]",
        3589=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Michigan_North_FIPS_2111\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",8000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Standard_Parallel_1\",45.48333333333333],PARAMETER[\"Standard_Parallel_2\",47.08333333333334],PARAMETER[\"Latitude_Of_Origin\",44.78333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3589]]",
        3590=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Michigan_North_FIPS_2111_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",26246719.16010498],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Standard_Parallel_1\",45.48333333333333],PARAMETER[\"Standard_Parallel_2\",47.08333333333334],PARAMETER[\"Latitude_Of_Origin\",44.78333333333333],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3590]]",
        3591=>"PROJCS[\"NAD_1983_NSRS2007_Michigan_GeoRef_Meters\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Natural_Origin\"],PARAMETER[\"False_Easting\",2546731.496],PARAMETER[\"False_Northing\",-4354009.816],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Azimuth\",337.25556],PARAMETER[\"Longitude_Of_Center\",-86.0],PARAMETER[\"Latitude_Of_Center\",45.30916666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3591]]",
        3592=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Michigan_South_FIPS_2113\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.36666666666666],PARAMETER[\"Standard_Parallel_1\",42.1],PARAMETER[\"Standard_Parallel_2\",43.66666666666666],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3592]]",
        3593=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Michigan_South_FIPS_2113_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",13123359.58005249],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.36666666666666],PARAMETER[\"Standard_Parallel_1\",42.1],PARAMETER[\"Standard_Parallel_2\",43.66666666666666],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3593]]",
        3594=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Minnesota_Central_FIPS_2202\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-94.25],PARAMETER[\"Standard_Parallel_1\",45.61666666666667],PARAMETER[\"Standard_Parallel_2\",47.05],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3594]]",
        3595=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Minnesota_North_FIPS_2201\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-93.1],PARAMETER[\"Standard_Parallel_1\",47.03333333333333],PARAMETER[\"Standard_Parallel_2\",48.63333333333333],PARAMETER[\"Latitude_Of_Origin\",46.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3595]]",
        3596=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Minnesota_South_FIPS_2203\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-94.0],PARAMETER[\"Standard_Parallel_1\",43.78333333333333],PARAMETER[\"Standard_Parallel_2\",45.21666666666667],PARAMETER[\"Latitude_Of_Origin\",43.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3596]]",
        3597=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Mississippi_East_FIPS_2301\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.83333333333333],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",29.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3597]]",
        3598=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Mississippi_East_FIPS_2301_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.83333333333333],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",29.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3598]]",
        3599=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Mississippi_West_FIPS_2302\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.33333333333333],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",29.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3599]]",
        3600=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Mississippi_West_FIPS_2302_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2296583.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.33333333333333],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",29.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3600]]",
        3601=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Missouri_Central_FIPS_2402\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.5],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",35.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3601]]",
        3602=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Missouri_East_FIPS_2401\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.5],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",35.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3602]]",
        3603=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Missouri_West_FIPS_2403\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",850000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-94.5],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",36.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3603]]",
        3604=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Montana_FIPS_2500\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-109.5],PARAMETER[\"Standard_Parallel_1\",45.0],PARAMETER[\"Standard_Parallel_2\",49.0],PARAMETER[\"Latitude_Of_Origin\",44.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3604]]",
        3605=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Montana_FIPS_2500_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968503.937007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-109.5],PARAMETER[\"Standard_Parallel_1\",45.0],PARAMETER[\"Standard_Parallel_2\",49.0],PARAMETER[\"Latitude_Of_Origin\",44.25],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3605]]",
        3606=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Nebraska_FIPS_2600\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Standard_Parallel_2\",43.0],PARAMETER[\"Latitude_Of_Origin\",39.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3606]]",
        3607=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Nevada_Central_FIPS_2702\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",6000000.0],PARAMETER[\"Central_Meridian\",-116.6666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3607]]",
        3608=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Nevada_Central_FIPS_2702_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",19685000.0],PARAMETER[\"Central_Meridian\",-116.6666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3608]]",
        3609=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Nevada_East_FIPS_2701\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",8000000.0],PARAMETER[\"Central_Meridian\",-115.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3609]]",
        3610=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Nevada_East_FIPS_2701_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",26246666.66666666],PARAMETER[\"Central_Meridian\",-115.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3610]]",
        3611=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Nevada_West_FIPS_2703\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",4000000.0],PARAMETER[\"Central_Meridian\",-118.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3611]]",
        3612=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Nevada_West_FIPS_2703_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",13123333.33333333],PARAMETER[\"Central_Meridian\",-118.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3612]]",
        3613=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_Hampshire_FIPS_2800\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3613]]",
        3614=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_Hampshire_FIPS_2800_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3614]]",
        3615=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_Jersey_FIPS_2900\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",38.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3615]]",
        3616=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_Jersey_FIPS_2900_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",492125.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",38.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3616]]",
        3617=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_Mexico_Central_FIPS_3002\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-106.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3617]]",
        3618=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_Mexico_Central_FIPS_3002_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-106.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3618]]",
        3619=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_Mexico_East_FIPS_3001\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",165000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-104.3333333333333],PARAMETER[\"Scale_Factor\",0.9999090909090909],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3619]]",
        3620=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_Mexico_East_FIPS_3001_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",541337.5],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-104.3333333333333],PARAMETER[\"Scale_Factor\",0.9999090909090909],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3620]]",
        3621=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_Mexico_West_FIPS_3003\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",830000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-107.8333333333333],PARAMETER[\"Scale_Factor\",0.9999166666666667],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3621]]",
        3622=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_Mexico_West_FIPS_3003_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2723091.666666666],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-107.8333333333333],PARAMETER[\"Scale_Factor\",0.9999166666666667],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3622]]",
        3623=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_York_Central_FIPS_3102\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-76.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3623]]",
        3624=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_York_Central_FIPS_3102_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",820208.3333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-76.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3624]]",
        3625=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_York_East_FIPS_3101\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",38.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3625]]",
        3626=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_York_East_FIPS_3101_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",492125.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",38.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3626]]",
        3627=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_York_Long_Island_FIPS_3104\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.0],PARAMETER[\"Standard_Parallel_1\",40.66666666666666],PARAMETER[\"Standard_Parallel_2\",41.03333333333333],PARAMETER[\"Latitude_Of_Origin\",40.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3627]]",
        3628=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_York_Long_Isl_FIPS_3104_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.0],PARAMETER[\"Standard_Parallel_1\",40.66666666666666],PARAMETER[\"Standard_Parallel_2\",41.03333333333333],PARAMETER[\"Latitude_Of_Origin\",40.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3628]]",
        3629=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_York_West_FIPS_3103\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",350000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-78.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3629]]",
        3630=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_New_York_West_FIPS_3103_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1148291.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-78.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3630]]",
        3631=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_North_Carolina_FIPS_3200\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",609601.2192024384],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.0],PARAMETER[\"Standard_Parallel_1\",34.33333333333334],PARAMETER[\"Standard_Parallel_2\",36.16666666666666],PARAMETER[\"Latitude_Of_Origin\",33.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3631]]",
        3632=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_North_Carolina_FIPS_3200_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.0],PARAMETER[\"Standard_Parallel_1\",34.33333333333334],PARAMETER[\"Standard_Parallel_2\",36.16666666666666],PARAMETER[\"Latitude_Of_Origin\",33.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3632]]",
        3633=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_North_Dakota_North_FIPS_3301\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",47.43333333333333],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3633]]",
        3634=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_North_Dakota_North_FIPS_3301_FtI\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968503.937007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",47.43333333333333],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3634]]",
        3635=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_North_Dakota_South_FIPS_3302\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",46.18333333333333],PARAMETER[\"Standard_Parallel_2\",47.48333333333333],PARAMETER[\"Latitude_Of_Origin\",45.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3635]]",
        3636=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_North_Dakota_South_FIPS_3302_FtI\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968503.937007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",46.18333333333333],PARAMETER[\"Standard_Parallel_2\",47.48333333333333],PARAMETER[\"Latitude_Of_Origin\",45.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3636]]",
        3637=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Ohio_North_FIPS_3401\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",40.43333333333333],PARAMETER[\"Standard_Parallel_2\",41.7],PARAMETER[\"Latitude_Of_Origin\",39.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3637]]",
        3638=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Ohio_South_FIPS_3402\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",38.73333333333333],PARAMETER[\"Standard_Parallel_2\",40.03333333333333],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3638]]",
        3639=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Oklahoma_North_FIPS_3501\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",35.56666666666667],PARAMETER[\"Standard_Parallel_2\",36.76666666666667],PARAMETER[\"Latitude_Of_Origin\",35.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3639]]",
        3640=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Oklahoma_North_FIPS_3501_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",35.56666666666667],PARAMETER[\"Standard_Parallel_2\",36.76666666666667],PARAMETER[\"Latitude_Of_Origin\",35.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3640]]",
        3641=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Oklahoma_South_FIPS_3502\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",33.93333333333333],PARAMETER[\"Standard_Parallel_2\",35.23333333333333],PARAMETER[\"Latitude_Of_Origin\",33.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3641]]",
        3642=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Oklahoma_South_FIPS_3502_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",33.93333333333333],PARAMETER[\"Standard_Parallel_2\",35.23333333333333],PARAMETER[\"Latitude_Of_Origin\",33.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3642]]",
        3643=>"PROJCS[\"NAD_1983_NSRS2007_Oregon_Statewide_Lambert\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",43.0],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3643]]",
        3644=>"PROJCS[\"NAD_1983_NSRS2007_Oregon_Statewide_Lambert_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312335.958005249],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",43.0],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3644]]",
        3645=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Oregon_North_FIPS_3601\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",44.33333333333334],PARAMETER[\"Standard_Parallel_2\",46.0],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3645]]",
        3646=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Oregon_North_FIPS_3601_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",8202099.737532808],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",44.33333333333334],PARAMETER[\"Standard_Parallel_2\",46.0],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3646]]",
        3647=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Oregon_South_FIPS_3602\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",42.33333333333334],PARAMETER[\"Standard_Parallel_2\",44.0],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3647]]",
        3648=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Oregon_South_FIPS_3602_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4921259.842519685],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",42.33333333333334],PARAMETER[\"Standard_Parallel_2\",44.0],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3648]]",
        3649=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Pennsylvania_North_FIPS_3701\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",40.88333333333333],PARAMETER[\"Standard_Parallel_2\",41.95],PARAMETER[\"Latitude_Of_Origin\",40.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3649]]",
        3650=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Pennsylvania_North_FIPS_3701_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",40.88333333333333],PARAMETER[\"Standard_Parallel_2\",41.95],PARAMETER[\"Latitude_Of_Origin\",40.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3650]]",
        3651=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Pennsylvania_South_FIPS_3702\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",39.93333333333333],PARAMETER[\"Standard_Parallel_2\",40.96666666666667],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3651]]",
        3652=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Pennsylvania_South_FIPS_3702_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",39.93333333333333],PARAMETER[\"Standard_Parallel_2\",40.96666666666667],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3652]]",
        3653=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Rhode_Island_FIPS_3800\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Scale_Factor\",0.99999375],PARAMETER[\"Latitude_Of_Origin\",41.08333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3653]]",
        3654=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Rhode_Island_FIPS_3800_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",328083.3333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Scale_Factor\",0.99999375],PARAMETER[\"Latitude_Of_Origin\",41.08333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3654]]",
        3655=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_South_Carolina_FIPS_3900\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",609600.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",32.5],PARAMETER[\"Standard_Parallel_2\",34.83333333333334],PARAMETER[\"Latitude_Of_Origin\",31.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3655]]",
        3656=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_South_Carolina_FIPS_3900_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",32.5],PARAMETER[\"Standard_Parallel_2\",34.83333333333334],PARAMETER[\"Latitude_Of_Origin\",31.83333333333333],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3656]]",
        3657=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_South_Dakota_North_FIPS_4001\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",44.41666666666666],PARAMETER[\"Standard_Parallel_2\",45.68333333333333],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3657]]",
        3658=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_South_Dakota_North_FIPS_4001_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",44.41666666666666],PARAMETER[\"Standard_Parallel_2\",45.68333333333333],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3658]]",
        3659=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_South_Dakota_South_FIPS_4002\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",42.83333333333334],PARAMETER[\"Standard_Parallel_2\",44.4],PARAMETER[\"Latitude_Of_Origin\",42.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3659]]",
        3660=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_South_Dakota_South_FIPS_4002_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",42.83333333333334],PARAMETER[\"Standard_Parallel_2\",44.4],PARAMETER[\"Latitude_Of_Origin\",42.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3660]]",
        3661=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Tennessee_FIPS_4100\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-86.0],PARAMETER[\"Standard_Parallel_1\",35.25],PARAMETER[\"Standard_Parallel_2\",36.41666666666666],PARAMETER[\"Latitude_Of_Origin\",34.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3661]]",
        3662=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Tennessee_FIPS_4100_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-86.0],PARAMETER[\"Standard_Parallel_1\",35.25],PARAMETER[\"Standard_Parallel_2\",36.41666666666666],PARAMETER[\"Latitude_Of_Origin\",34.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3662]]",
        3663=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Texas_Central_FIPS_4203\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",3000000.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",30.11666666666667],PARAMETER[\"Standard_Parallel_2\",31.88333333333333],PARAMETER[\"Latitude_Of_Origin\",29.66666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3663]]",
        3664=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Texas_Central_FIPS_4203_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2296583.333333333],PARAMETER[\"False_Northing\",9842500.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",30.11666666666667],PARAMETER[\"Standard_Parallel_2\",31.88333333333333],PARAMETER[\"Latitude_Of_Origin\",29.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3664]]",
        3665=>"PROJCS[\"NAD_1983_NSRS2007_Texas_Centric_Mapping_System_Albers\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",6000000.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",27.5],PARAMETER[\"Standard_Parallel_2\",35.0],PARAMETER[\"Latitude_Of_Origin\",18.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3665]]",
        3666=>"PROJCS[\"NAD_1983_NSRS2007_Texas_Centric_Mapping_System_Lambert\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",5000000.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",27.5],PARAMETER[\"Standard_Parallel_2\",35.0],PARAMETER[\"Latitude_Of_Origin\",18.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3666]]",
        3667=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Texas_North_FIPS_4201\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-101.5],PARAMETER[\"Standard_Parallel_1\",34.65],PARAMETER[\"Standard_Parallel_2\",36.18333333333333],PARAMETER[\"Latitude_Of_Origin\",34.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3667]]",
        3668=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Texas_North_FIPS_4201_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-101.5],PARAMETER[\"Standard_Parallel_1\",34.65],PARAMETER[\"Standard_Parallel_2\",36.18333333333333],PARAMETER[\"Latitude_Of_Origin\",34.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3668]]",
        3669=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Texas_North_Central_FIPS_4202\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",32.13333333333333],PARAMETER[\"Standard_Parallel_2\",33.96666666666667],PARAMETER[\"Latitude_Of_Origin\",31.66666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3669]]",
        3670=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Texas_North_Central_FIPS_4202_FtUS\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",6561666.666666666],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",32.13333333333333],PARAMETER[\"Standard_Parallel_2\",33.96666666666667],PARAMETER[\"Latitude_Of_Origin\",31.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3670]]",
        3671=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Texas_South_FIPS_4205\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",5000000.0],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",26.16666666666667],PARAMETER[\"Standard_Parallel_2\",27.83333333333333],PARAMETER[\"Latitude_Of_Origin\",25.66666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3671]]",
        3672=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Texas_South_FIPS_4205_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",16404166.66666666],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",26.16666666666667],PARAMETER[\"Standard_Parallel_2\",27.83333333333333],PARAMETER[\"Latitude_Of_Origin\",25.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3672]]",
        3673=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Texas_South_Central_FIPS_4204\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",4000000.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Standard_Parallel_1\",28.38333333333333],PARAMETER[\"Standard_Parallel_2\",30.28333333333333],PARAMETER[\"Latitude_Of_Origin\",27.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3673]]",
        3674=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Texas_South_Central_FIPS_4204_FtUS\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",13123333.33333333],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Standard_Parallel_1\",28.38333333333333],PARAMETER[\"Standard_Parallel_2\",30.28333333333333],PARAMETER[\"Latitude_Of_Origin\",27.83333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3674]]",
        3675=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Utah_Central_FIPS_4302\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",39.01666666666667],PARAMETER[\"Standard_Parallel_2\",40.65],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3675]]",
        3676=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Utah_Central_FIPS_4302_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640419.947506561],PARAMETER[\"False_Northing\",6561679.790026246],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",39.01666666666667],PARAMETER[\"Standard_Parallel_2\",40.65],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3676]]",
        3677=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Utah_Central_FIPS_4302_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",6561666.666666666],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",39.01666666666667],PARAMETER[\"Standard_Parallel_2\",40.65],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3677]]",
        3678=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Utah_North_FIPS_4301\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",40.71666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3678]]",
        3679=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Utah_North_FIPS_4301_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640419.947506561],PARAMETER[\"False_Northing\",3280839.895013123],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",40.71666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.33333333333334],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3679]]",
        3680=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Utah_North_FIPS_4301_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",40.71666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3680]]",
        3681=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Utah_South_FIPS_4303\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",3000000.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",37.21666666666667],PARAMETER[\"Standard_Parallel_2\",38.35],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3681]]",
        3682=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Utah_South_FIPS_4303_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640419.947506561],PARAMETER[\"False_Northing\",9842519.685039369],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",37.21666666666667],PARAMETER[\"Standard_Parallel_2\",38.35],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",3682]]",
        3683=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Utah_South_FIPS_4303_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",9842500.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",37.21666666666667],PARAMETER[\"Standard_Parallel_2\",38.35],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3683]]",
        3684=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Vermont_FIPS_4400\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-72.5],PARAMETER[\"Scale_Factor\",0.9999642857142857],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3684]]",
        3685=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Virginia_North_FIPS_4501\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",38.03333333333333],PARAMETER[\"Standard_Parallel_2\",39.2],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3685]]",
        3686=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Virginia_North_FIPS_4501_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",11482916.66666666],PARAMETER[\"False_Northing\",6561666.666666666],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",38.03333333333333],PARAMETER[\"Standard_Parallel_2\",39.2],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3686]]",
        3687=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Virginia_South_FIPS_4502\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",36.76666666666667],PARAMETER[\"Standard_Parallel_2\",37.96666666666667],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3687]]",
        3688=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Virginia_South_FIPS_4502_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",11482916.66666666],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",36.76666666666667],PARAMETER[\"Standard_Parallel_2\",37.96666666666667],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3688]]",
        3689=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Washington_North_FIPS_4601\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.8333333333333],PARAMETER[\"Standard_Parallel_1\",47.5],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3689]]",
        3690=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Washington_North_FIPS_4601_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.8333333333333],PARAMETER[\"Standard_Parallel_1\",47.5],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3690]]",
        3691=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Washington_South_FIPS_4602\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",45.83333333333334],PARAMETER[\"Standard_Parallel_2\",47.33333333333334],PARAMETER[\"Latitude_Of_Origin\",45.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3691]]",
        3692=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Washington_South_FIPS_4602_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",45.83333333333334],PARAMETER[\"Standard_Parallel_2\",47.33333333333334],PARAMETER[\"Latitude_Of_Origin\",45.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3692]]",
        3693=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_West_Virginia_North_FIPS_4701\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Standard_Parallel_1\",39.0],PARAMETER[\"Standard_Parallel_2\",40.25],PARAMETER[\"Latitude_Of_Origin\",38.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3693]]",
        3694=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_West_Virginia_South_FIPS_4702\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",37.48333333333333],PARAMETER[\"Standard_Parallel_2\",38.88333333333333],PARAMETER[\"Latitude_Of_Origin\",37.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3694]]",
        3695=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Wisconsin_Central_FIPS_4802\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",44.25],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3695]]",
        3696=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Wisconsin_Central_FIPS_4802_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",44.25],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3696]]",
        3697=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Wisconsin_North_FIPS_4801\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",45.56666666666667],PARAMETER[\"Standard_Parallel_2\",46.76666666666667],PARAMETER[\"Latitude_Of_Origin\",45.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3697]]",
        3698=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Wisconsin_North_FIPS_4801_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",45.56666666666667],PARAMETER[\"Standard_Parallel_2\",46.76666666666667],PARAMETER[\"Latitude_Of_Origin\",45.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3698]]",
        3699=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Wisconsin_South_FIPS_4803\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",42.73333333333333],PARAMETER[\"Standard_Parallel_2\",44.06666666666667],PARAMETER[\"Latitude_Of_Origin\",42.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3699]]",
        3700=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Wisconsin_South_FIPS_4803_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",42.73333333333333],PARAMETER[\"Standard_Parallel_2\",44.06666666666667],PARAMETER[\"Latitude_Of_Origin\",42.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3700]]",
        3701=>"PROJCS[\"NAD_1983_NSRS2007_Wisconsin_TM\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",520000.0],PARAMETER[\"False_Northing\",-4480000.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3701]]",
        3702=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Wyoming_East_FIPS_4901\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.1666666666667],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3702]]",
        3703=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Wyoming_East_Central_FIPS_4902\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-107.3333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3703]]",
        3704=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Wyoming_West_Central_FIPS_4903\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-108.75],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3704]]",
        3705=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Wyoming_West_FIPS_4904\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-110.0833333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3705]]",
        3706=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_59N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3706]]",
        3707=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_60N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3707]]",
        3708=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_1N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3708]]",
        3709=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_2N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3709]]",
        3710=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_3N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3710]]",
        3711=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_4N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3711]]",
        3712=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_5N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3712]]",
        3713=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_6N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3713]]",
        3714=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_7N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3714]]",
        3715=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_8N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3715]]",
        3716=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_9N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3716]]",
        3717=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_10N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3717]]",
        3718=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_11N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3718]]",
        3719=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_12N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3719]]",
        3720=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_13N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3720]]",
        3721=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_14N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3721]]",
        3722=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_15N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3722]]",
        3723=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_16N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3723]]",
        3724=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_17N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3724]]",
        3725=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_18N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3725]]",
        3726=>"PROJCS[\"NAD_1983_NSRS2007_UTM_Zone_19N\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3726]]",
        3727=>"PROJCS[\"Reunion_1947_TM_Reunion\",GEOGCS[\"GCS_Reunion_1947\",DATUM[\"D_Reunion_1947\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",160000.0],PARAMETER[\"False_Northing\",50000.0],PARAMETER[\"Central_Meridian\",55.53333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-21.11666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3727]]",
        3728=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Ohio_North_FIPS_3401_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",40.43333333333333],PARAMETER[\"Standard_Parallel_2\",41.7],PARAMETER[\"Latitude_Of_Origin\",39.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3728]]",
        3729=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Ohio_South_FIPS_3402_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",38.73333333333333],PARAMETER[\"Standard_Parallel_2\",40.03333333333333],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3729]]",
        3730=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Wyoming_East_FIPS_4901_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.1666666666667],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3730]]",
        3731=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Wyoming_E_Central_FIPS_4902_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-107.3333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3731]]",
        3732=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Wyoming_W_Central_FIPS_4903_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-108.75],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3732]]",
        3733=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Wyoming_West_FIPS_4904_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-110.0833333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3733]]",
        3734=>"PROJCS[\"NAD_1983_StatePlane_Ohio_North_FIPS_3401_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",40.43333333333333],PARAMETER[\"Standard_Parallel_2\",41.7],PARAMETER[\"Latitude_Of_Origin\",39.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3734]]",
        3735=>"PROJCS[\"NAD_1983_StatePlane_Ohio_South_FIPS_3402_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",38.73333333333333],PARAMETER[\"Standard_Parallel_2\",40.03333333333333],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3735]]",
        3736=>"PROJCS[\"NAD_1983_StatePlane_Wyoming_East_FIPS_4901_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.1666666666667],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3736]]",
        3737=>"PROJCS[\"NAD_1983_StatePlane_Wyoming_East_Central_FIPS_4902_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-107.3333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3737]]",
        3738=>"PROJCS[\"NAD_1983_StatePlane_Wyoming_West_Central_FIPS_4903_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-108.75],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3738]]",
        3739=>"PROJCS[\"NAD_1983_StatePlane_Wyoming_West_FIPS_4904_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-110.0833333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3739]]",
        3740=>"PROJCS[\"NAD_1983_HARN_UTM_Zone_10N\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3740]]",
        3741=>"PROJCS[\"NAD_1983_HARN_UTM_Zone_11N\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3741]]",
        3742=>"PROJCS[\"NAD_1983_HARN_UTM_Zone_12N\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3742]]",
        3743=>"PROJCS[\"NAD_1983_HARN_UTM_Zone_13N\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3743]]",
        3744=>"PROJCS[\"NAD_1983_HARN_UTM_Zone_14N\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3744]]",
        3745=>"PROJCS[\"NAD_1983_HARN_UTM_Zone_15N\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3745]]",
        3746=>"PROJCS[\"NAD_1983_HARN_UTM_Zone_16N\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3746]]",
        3747=>"PROJCS[\"NAD_1983_HARN_UTM_Zone_17N\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3747]]",
        3748=>"PROJCS[\"NAD_1983_HARN_UTM_Zone_18N\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3748]]",
        3749=>"PROJCS[\"NAD_1983_HARN_UTM_Zone_19N\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3749]]",
        3750=>"PROJCS[\"NAD_1983_HARN_UTM_Zone_4N\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3750]]",
        3751=>"PROJCS[\"NAD_1983_HARN_UTM_Zone_5N\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3751]]",
        3753=>"PROJCS[\"NAD_1983_HARN_StatePlane_Ohio_North_FIPS_3401_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",40.43333333333333],PARAMETER[\"Standard_Parallel_2\",41.7],PARAMETER[\"Latitude_Of_Origin\",39.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3753]]",
        3754=>"PROJCS[\"NAD_1983_HARN_StatePlane_Ohio_South_FIPS_3402_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",38.73333333333333],PARAMETER[\"Standard_Parallel_2\",40.03333333333333],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3754]]",
        3755=>"PROJCS[\"NAD_1983_HARN_StatePlane_Wyoming_East_FIPS_4901_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.1666666666667],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3755]]",
        3756=>"PROJCS[\"NAD_1983_HARN_StatePlane_Wyoming_East_Central_FIPS_4902_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-107.3333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3756]]",
        3757=>"PROJCS[\"NAD_1983_HARN_StatePlane_Wyoming_West_Central_FIPS_4903_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-108.75],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3757]]",
        3758=>"PROJCS[\"NAD_1983_HARN_StatePlane_Wyoming_West_FIPS_4904_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-110.0833333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3758]]",
        3759=>"PROJCS[\"NAD_1983_StatePlane_Hawaii_3_FIPS_5103_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-158.0],PARAMETER[\"Scale_Factor\",0.99999],PARAMETER[\"Latitude_Of_Origin\",21.16666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3759]]",
        3760=>"PROJCS[\"NAD_1983_HARN_StatePlane_Hawaii_3_FIPS_5103_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-158.0],PARAMETER[\"Scale_Factor\",0.99999],PARAMETER[\"Latitude_Of_Origin\",21.16666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3760]]",
        3761=>"PROJCS[\"NAD_1983_CSRS_UTM_Zone_22N\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3761]]",
        3762=>"PROJCS[\"WGS_1984_South_Georgia_Lambert\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-37.0],PARAMETER[\"Standard_Parallel_1\",-54.0],PARAMETER[\"Standard_Parallel_2\",-54.75],PARAMETER[\"Latitude_Of_Origin\",-55.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3762]]",
        3763=>"PROJCS[\"ETRS_1989_Portugal_TM06\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-8.133108333333334],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",39.66825833333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3763]]",
        3764=>"PROJCS[\"NZGD_2000_Chatham_Island_Circuit\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Central_Meridian\",-176.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3764]]",
        3765=>"PROJCS[\"HTRS96_Croatia_TM\",GEOGCS[\"GCS_HTRS96\",DATUM[\"D_Croatian_Terrestrial_Reference_System\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",16.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3765]]",
        3766=>"PROJCS[\"HTRS96_Croatia_LCC\",GEOGCS[\"GCS_HTRS96\",DATUM[\"D_Croatian_Terrestrial_Reference_System\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",16.5],PARAMETER[\"Standard_Parallel_1\",43.08333333333334],PARAMETER[\"Standard_Parallel_2\",45.91666666666666],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3766]]",
        3767=>"PROJCS[\"HTRS96_UTM_Zone_33N\",GEOGCS[\"GCS_HTRS96\",DATUM[\"D_Croatian_Terrestrial_Reference_System\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3767]]",
        3768=>"PROJCS[\"HTRS96_UTM_Zone_34N\",GEOGCS[\"GCS_HTRS96\",DATUM[\"D_Croatian_Terrestrial_Reference_System\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3768]]",
        3769=>"PROJCS[\"Bermuda_1957_UTM_Zone_20N\",GEOGCS[\"GCS_Bermuda_1957\",DATUM[\"D_Bermuda_1957\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3769]]",
        3770=>"PROJCS[\"Bermuda_2000_National_Grid\",GEOGCS[\"GCS_Bermuda_2000\",DATUM[\"D_Bermuda_2000\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",550000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-64.75],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",32.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3770]]",
        3771=>"PROJCS[\"NAD_1927_3TM_111\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3771]]",
        3772=>"PROJCS[\"NAD_1927_3TM_114\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-114.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3772]]",
        3773=>"PROJCS[\"NAD_1927_3TM_117\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3773]]",
        3775=>"PROJCS[\"NAD_1983_3TM_111\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3775]]",
        3776=>"PROJCS[\"NAD_1983_3TM_114\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-114.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3776]]",
        3777=>"PROJCS[\"NAD_1983_3TM_117\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3777]]",
        3779=>"PROJCS[\"NAD_1983_CSRS_3TM_111\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3779]]",
        3780=>"PROJCS[\"NAD_1983_CSRS_3TM_114\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-114.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3780]]",
        3781=>"PROJCS[\"NAD_1983_CSRS_3TM_117\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3781]]",
        3783=>"PROJCS[\"Pitcairn_2006_Pitcairn_TM_2006\",GEOGCS[\"GCS_Pitcairn_2006\",DATUM[\"D_Pitcairn_2006\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",14200.0],PARAMETER[\"False_Northing\",15500.0],PARAMETER[\"Central_Meridian\",-130.1129671111111],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-25.06855261111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3783]]",
        3784=>"PROJCS[\"Pitcairn_1967_UTM_Zone_9S\",GEOGCS[\"GCS_Pitcairn_1967\",DATUM[\"D_Pitcairn_1967\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3784]]",
        3785=>"PROJCS[\"WGS_1984_Web_Mercator\",GEOGCS[\"GCS_WGS_1984_Major_Auxiliary_Sphere\",DATUM[\"D_WGS_1984_Major_Auxiliary_Sphere\",SPHEROID[\"WGS_1984_Major_Auxiliary_Sphere\",6378137.0,0.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3785]]",
        3788=>"PROJCS[\"NZGD_2000_Auckland_Islands_TM_2000\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",166.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3788]]",
        3789=>"PROJCS[\"NZGD_2000_Campbell_Island_TM_2000\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",169.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3789]]",
        3790=>"PROJCS[\"NZGD_2000_Antipodes_Islands_TM_2000\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",179.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3790]]",
        3791=>"PROJCS[\"NZGD_2000_Raoul_Island_TM_2000\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-178.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3791]]",
        3793=>"PROJCS[\"NZGD_2000_Chatham_Islands_TM_2000\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-176.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3793]]",
        3794=>"PROJCS[\"Slovenia_1996_Slovene_National_Grid\",GEOGCS[\"GCS_Slovenia_1996\",DATUM[\"D_Slovenia_Geodetic_Datum_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3794]]",
        3797=>"PROJCS[\"NAD_1927_MTQ_Lambert\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.0],PARAMETER[\"Standard_Parallel_1\",46.0],PARAMETER[\"Standard_Parallel_2\",50.0],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3797]]",
        3798=>"PROJCS[\"NAD_1983_MTQ_Lambert\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.0],PARAMETER[\"Standard_Parallel_1\",46.0],PARAMETER[\"Standard_Parallel_2\",50.0],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3798]]",
        3799=>"PROJCS[\"NAD_1983_CSRS_MTQ_Lambert\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.0],PARAMETER[\"Standard_Parallel_1\",46.0],PARAMETER[\"Standard_Parallel_2\",50.0],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3799]]",
        3800=>"PROJCS[\"NAD_1927_3TM_120\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3800]]",
        3801=>"PROJCS[\"NAD_1983_3TM_120\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3801]]",
        3802=>"PROJCS[\"NAD_1983_CSRS_3TM_120\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3802]]",
        3812=>"PROJCS[\"Belge_Lambert_2008\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",649328.0],PARAMETER[\"False_Northing\",665262.0],PARAMETER[\"Central_Meridian\",4.359215833333333],PARAMETER[\"Standard_Parallel_1\",49.83333333333334],PARAMETER[\"Standard_Parallel_2\",51.16666666666666],PARAMETER[\"Latitude_Of_Origin\",50.797815],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3812]]",
        3814=>"PROJCS[\"NAD_1983_Mississippi_TM\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",1300000.0],PARAMETER[\"Central_Meridian\",-89.75],PARAMETER[\"Scale_Factor\",0.9998335],PARAMETER[\"Latitude_Of_Origin\",32.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3814]]",
        3815=>"PROJCS[\"NAD_1983_HARN_Mississippi_TM\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",1300000.0],PARAMETER[\"Central_Meridian\",-89.75],PARAMETER[\"Scale_Factor\",0.9998335],PARAMETER[\"Latitude_Of_Origin\",32.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3815]]",
        3816=>"PROJCS[\"NAD_1983_NSRS2007_Mississippi_TM\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",1300000.0],PARAMETER[\"Central_Meridian\",-89.75],PARAMETER[\"Scale_Factor\",0.9998335],PARAMETER[\"Latitude_Of_Origin\",32.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3816]]",
        3825=>"PROJCS[\"TWD_1997_TM_Penghu\",GEOGCS[\"GCS_TWD_1997\",DATUM[\"D_TWD_1997\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",119.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3825]]",
        3826=>"PROJCS[\"TWD_1997_TM_Taiwan\",GEOGCS[\"GCS_TWD_1997\",DATUM[\"D_TWD_1997\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",121.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3826]]",
        3827=>"PROJCS[\"TWD_1967_TM_Penghu\",GEOGCS[\"GCS_TWD_1967\",DATUM[\"D_TWD_1967\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",119.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3827]]",
        3828=>"PROJCS[\"TWD_1967_TM_Taiwan\",GEOGCS[\"GCS_TWD_1967\",DATUM[\"D_TWD_1967\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",121.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3828]]",
        3829=>"PROJCS[\"Hu_Tzu_Shan_UTM_Zone_51N\",GEOGCS[\"GCS_Hu_Tzu_Shan\",DATUM[\"D_Hu_Tzu_Shan\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3829]]",
        3832=>"PROJCS[\"WGS_1984_PDC_Mercator\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",150.0],PARAMETER[\"Standard_Parallel_1\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3832]]",
        3833=>"PROJCS[\"Pulkovo_1942_Adj_1958_GK_Zone_2\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3833]]",
        3834=>"PROJCS[\"Pulkovo_1942_Adj_1983_GK_Zone_2\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3834]]",
        3835=>"PROJCS[\"Pulkovo_1942_Adj_1983_GK_Zone_3\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3835]]",
        3836=>"PROJCS[\"Pulkovo_1942_Adj_1983_GK_Zone_4\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3836]]",
        3837=>"PROJCS[\"Pulkovo_1942_Adj_1958_3_Degree_GK_Zone_3\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3837]]",
        3838=>"PROJCS[\"Pulkovo_1942_Adj_1958_3_Degree_GK_Zone_4\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3838]]",
        3839=>"PROJCS[\"Pulkovo_1942_Adj_1958_3_Degree_GK_Zone_9\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",9500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3839]]",
        3840=>"PROJCS[\"Pulkovo_1942_Adj_1958_3_Degree_GK_Zone_10\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",10500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3840]]",
        3841=>"PROJCS[\"Pulkovo_1942_Adj_1983_3_Degree_GK_Zone_6\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",6500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",18.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3841]]",
        3844=>"PROJCS[\"Pulkovo_1942_Adj_58_Stereo_70\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",25.0],PARAMETER[\"Scale_Factor\",0.99975],PARAMETER[\"Latitude_Of_Origin\",46.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3844]]",
        3845=>"PROJCS[\"SWEREF99_RT90_7.5_gon_V_emulation\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500025.141],PARAMETER[\"False_Northing\",-667.282],PARAMETER[\"Central_Meridian\",11.30625],PARAMETER[\"Scale_Factor\",1.000006],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3845]]",
        3846=>"PROJCS[\"SWEREF99_RT90_5_gon_V_emulation\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500044.695],PARAMETER[\"False_Northing\",-667.13],PARAMETER[\"Central_Meridian\",13.55626666666667],PARAMETER[\"Scale_Factor\",1.0000058],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3846]]",
        3847=>"PROJCS[\"SWEREF99_RT90_2.5_gon_V_emulation\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500064.274],PARAMETER[\"False_Northing\",-667.711],PARAMETER[\"Central_Meridian\",15.80628452944445],PARAMETER[\"Scale_Factor\",1.00000561024],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3847]]",
        3848=>"PROJCS[\"SWEREF99_RT90_0_gon_emulation\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500083.521],PARAMETER[\"False_Northing\",-668.844],PARAMETER[\"Central_Meridian\",18.0563],PARAMETER[\"Scale_Factor\",1.0000054],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3848]]",
        3849=>"PROJCS[\"SWEREF99_RT90_2.5_gon_O_emulation\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500102.765],PARAMETER[\"False_Northing\",-670.706],PARAMETER[\"Central_Meridian\",20.30631666666667],PARAMETER[\"Scale_Factor\",1.0000052],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3849]]",
        3850=>"PROJCS[\"SWEREF99_RT90_5_gon_O_emulation\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500121.846],PARAMETER[\"False_Northing\",-672.557],PARAMETER[\"Central_Meridian\",22.55633333333333],PARAMETER[\"Scale_Factor\",1.0000049],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3850]]",
        3851=>"PROJCS[\"NZGD_2000_NZ_Continental_Shelf_2000\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.0],PARAMETER[\"False_Northing\",7000000.0],PARAMETER[\"Central_Meridian\",173.0],PARAMETER[\"Standard_Parallel_1\",-37.5],PARAMETER[\"Standard_Parallel_2\",-44.5],PARAMETER[\"Latitude_Of_Origin\",-41.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3851]]",
        3852=>"PROJCS[\"RSRGD2000_DGLC2000\",GEOGCS[\"GCS_RSRGD2000\",DATUM[\"D_Ross_Sea_Region_Geodetic_Datum_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",157.0],PARAMETER[\"Standard_Parallel_1\",-76.66666666666667],PARAMETER[\"Standard_Parallel_2\",-79.33333333333333],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3852]]",
        3854=>"PROJCS[\"SWEREF99_County_ST74\",GEOGCS[\"GCS_SWEREF99\",DATUM[\"D_SWEREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100182.7406],PARAMETER[\"False_Northing\",-6500620.1207],PARAMETER[\"Central_Meridian\",18.05787],PARAMETER[\"Scale_Factor\",0.99999506],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3854]]",
        3857=>"PROJCS[\"WGS_1984_Web_Mercator_Auxiliary_Sphere\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Mercator_Auxiliary_Sphere\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",0.0],PARAMETER[\"Auxiliary_Sphere_Type\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3857]]",
        3873=>"PROJCS[\"ETRS_1989_GK19FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",19500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",19.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3873]]",
        3874=>"PROJCS[\"ETRS_1989_GK20FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",20500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",20.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3874]]",
        3875=>"PROJCS[\"ETRS_1989_GK21FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",21500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3875]]",
        3876=>"PROJCS[\"ETRS_1989_GK22FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",22500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",22.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3876]]",
        3877=>"PROJCS[\"ETRS_1989_GK23FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",23500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",23.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3877]]",
        3878=>"PROJCS[\"ETRS_1989_GK24FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",24500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3878]]",
        3879=>"PROJCS[\"ETRS_1989_GK25FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",25500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",25.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3879]]",
        3880=>"PROJCS[\"ETRS_1989_GK26FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",26500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",26.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3880]]",
        3881=>"PROJCS[\"ETRS_1989_GK27FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",27500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3881]]",
        3882=>"PROJCS[\"ETRS_1989_GK28FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",28500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",28.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3882]]",
        3883=>"PROJCS[\"ETRS_1989_GK29FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",29500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",29.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3883]]",
        3884=>"PROJCS[\"ETRS_1989_GK30FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",30500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3884]]",
        3885=>"PROJCS[\"ETRS_1989_GK31FIN\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",31500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",31.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3885]]",
        3890=>"PROJCS[\"IGRS_UTM_Zone_37N\",GEOGCS[\"GCS_IGRS\",DATUM[\"D_Iraqi_Geospatial_Reference_System\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3890]]",
        3891=>"PROJCS[\"IGRS_UTM_Zone_38N\",GEOGCS[\"GCS_IGRS\",DATUM[\"D_Iraqi_Geospatial_Reference_System\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3891]]",
        3892=>"PROJCS[\"IGRS_UTM_Zone_39N\",GEOGCS[\"GCS_IGRS\",DATUM[\"D_Iraqi_Geospatial_Reference_System\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3892]]",
        3893=>"PROJCS[\"ED_1950_Iraq_National_Grid\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",46.5],PARAMETER[\"Scale_Factor\",0.9994],PARAMETER[\"Latitude_Of_Origin\",29.02626833333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3893]]",
        3907=>"PROJCS[\"MGI_1901_Balkans_5\",GEOGCS[\"GCS_MGI_1901\",DATUM[\"D_MGI_1901\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3907]]",
        3908=>"PROJCS[\"MGI_1901_Balkans_6\",GEOGCS[\"GCS_MGI_1901\",DATUM[\"D_MGI_1901\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",6500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",18.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3908]]",
        3909=>"PROJCS[\"MGI_1901_Balkans_7\",GEOGCS[\"GCS_MGI_1901\",DATUM[\"D_MGI_1901\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",7500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3909]]",
        3910=>"PROJCS[\"MGI_1901_Balkans_8\",GEOGCS[\"GCS_MGI_1901\",DATUM[\"D_MGI_1901\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",8500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3910]]",
        3911=>"PROJCS[\"MGI_1901_Slovenia_Grid\",GEOGCS[\"GCS_MGI_1901\",DATUM[\"D_MGI_1901\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3911]]",
        3912=>"PROJCS[\"MGI_1901_Slovene_National_Grid\",GEOGCS[\"GCS_MGI_1901\",DATUM[\"D_MGI_1901\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3912]]",
        3920=>"PROJCS[\"Puerto_Rico_UTM_Zone_20N\",GEOGCS[\"GCS_Puerto_Rico\",DATUM[\"D_Puerto_Rico\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3920]]",
        3942=>"PROJCS[\"RGF_1993_CC42\",GEOGCS[\"GCS_RGF_1993\",DATUM[\"D_RGF_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1700000.0],PARAMETER[\"False_Northing\",1200000.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Standard_Parallel_1\",41.25],PARAMETER[\"Standard_Parallel_2\",42.75],PARAMETER[\"Latitude_Of_Origin\",42.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3942]]",
        3943=>"PROJCS[\"RGF_1993_CC43\",GEOGCS[\"GCS_RGF_1993\",DATUM[\"D_RGF_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1700000.0],PARAMETER[\"False_Northing\",2200000.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Standard_Parallel_1\",42.25],PARAMETER[\"Standard_Parallel_2\",43.75],PARAMETER[\"Latitude_Of_Origin\",43.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3943]]",
        3944=>"PROJCS[\"RGF_1993_CC44\",GEOGCS[\"GCS_RGF_1993\",DATUM[\"D_RGF_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1700000.0],PARAMETER[\"False_Northing\",3200000.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Standard_Parallel_1\",43.25],PARAMETER[\"Standard_Parallel_2\",44.75],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3944]]",
        3945=>"PROJCS[\"RGF_1993_CC45\",GEOGCS[\"GCS_RGF_1993\",DATUM[\"D_RGF_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1700000.0],PARAMETER[\"False_Northing\",4200000.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Standard_Parallel_1\",44.25],PARAMETER[\"Standard_Parallel_2\",45.75],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3945]]",
        3946=>"PROJCS[\"RGF_1993_CC46\",GEOGCS[\"GCS_RGF_1993\",DATUM[\"D_RGF_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1700000.0],PARAMETER[\"False_Northing\",5200000.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Standard_Parallel_1\",45.25],PARAMETER[\"Standard_Parallel_2\",46.75],PARAMETER[\"Latitude_Of_Origin\",46.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3946]]",
        3947=>"PROJCS[\"RGF_1993_CC47\",GEOGCS[\"GCS_RGF_1993\",DATUM[\"D_RGF_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1700000.0],PARAMETER[\"False_Northing\",6200000.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Standard_Parallel_1\",46.25],PARAMETER[\"Standard_Parallel_2\",47.75],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3947]]",
        3948=>"PROJCS[\"RGF_1993_CC48\",GEOGCS[\"GCS_RGF_1993\",DATUM[\"D_RGF_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1700000.0],PARAMETER[\"False_Northing\",7200000.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Standard_Parallel_1\",47.25],PARAMETER[\"Standard_Parallel_2\",48.75],PARAMETER[\"Latitude_Of_Origin\",48.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3948]]",
        3949=>"PROJCS[\"RGF_1993_CC49\",GEOGCS[\"GCS_RGF_1993\",DATUM[\"D_RGF_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1700000.0],PARAMETER[\"False_Northing\",8200000.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Standard_Parallel_1\",48.25],PARAMETER[\"Standard_Parallel_2\",49.75],PARAMETER[\"Latitude_Of_Origin\",49.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3949]]",
        3950=>"PROJCS[\"RGF_1993_CC50\",GEOGCS[\"GCS_RGF_1993\",DATUM[\"D_RGF_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1700000.0],PARAMETER[\"False_Northing\",9200000.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Standard_Parallel_1\",49.25],PARAMETER[\"Standard_Parallel_2\",50.75],PARAMETER[\"Latitude_Of_Origin\",50.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3950]]",
        3968=>"PROJCS[\"NAD_1983_Virginia_Lambert\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Standard_Parallel_1\",37.0],PARAMETER[\"Standard_Parallel_2\",39.5],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3968]]",
        3969=>"PROJCS[\"NAD_1983_HARN_Virginia_Lambert\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Standard_Parallel_1\",37.0],PARAMETER[\"Standard_Parallel_2\",39.5],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3969]]",
        3970=>"PROJCS[\"NAD_1983_NSRS2007_Virginia_Lambert\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Standard_Parallel_1\",37.0],PARAMETER[\"Standard_Parallel_2\",39.5],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3970]]",
        3973=>"PROJCS[\"WGS_1984_EASE_Grid_North\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Latitude_Of_Origin\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3973]]",
        3974=>"PROJCS[\"WGS_1984_EASE_Grid_South\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3974]]",
        3975=>"PROJCS[\"WGS_1984_EASE_Grid_Global\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cylindrical_Equal_Area\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3975]]",
        3976=>"PROJCS[\"WGS_1984_NSIDC_Sea_Ice_Polar_Stereographic_South\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic_South_Pole\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",-70.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3976]]",
        3978=>"PROJCS[\"NAD_1983_Canada_Atlas_Lambert\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-95.0],PARAMETER[\"Standard_Parallel_1\",49.0],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",49.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3978]]",
        3979=>"PROJCS[\"NAD_1983_CSRS_Canada_Atlas_Lambert\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-95.0],PARAMETER[\"Standard_Parallel_1\",49.0],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",49.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3979]]",
        3986=>"PROJCS[\"Katanga_1955_Katanga_Gauss_Zone_A\",GEOGCS[\"GCS_Katanga_1955\",DATUM[\"D_Katanga_1955\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-9.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3986]]",
        3987=>"PROJCS[\"Katanga_1955_Katanga_Gauss_Zone_B\",GEOGCS[\"GCS_Katanga_1955\",DATUM[\"D_Katanga_1955\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",28.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-9.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3987]]",
        3988=>"PROJCS[\"Katanga_1955_Katanga_Gauss_Zone_C\",GEOGCS[\"GCS_Katanga_1955\",DATUM[\"D_Katanga_1955\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",26.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-9.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3988]]",
        3989=>"PROJCS[\"Katanga_1955_Katanga_Gauss_Zone_D\",GEOGCS[\"GCS_Katanga_1955\",DATUM[\"D_Katanga_1955\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-9.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3989]]",
        3991=>"PROJCS[\"Puerto_Rico_StatePlane_Puerto_Rico_FIPS_5201\",GEOGCS[\"GCS_Puerto_Rico\",DATUM[\"D_Puerto_Rico\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-66.43333333333334],PARAMETER[\"Standard_Parallel_1\",18.03333333333333],PARAMETER[\"Standard_Parallel_2\",18.43333333333333],PARAMETER[\"Latitude_Of_Origin\",17.83333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3991]]",
        3992=>"PROJCS[\"Puerto_Rico_StatePlane_Virgin_Islands_St_Croix_FIPS_5202\",GEOGCS[\"GCS_Puerto_Rico\",DATUM[\"D_Puerto_Rico\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-66.43333333333334],PARAMETER[\"Standard_Parallel_1\",18.03333333333333],PARAMETER[\"Standard_Parallel_2\",18.43333333333333],PARAMETER[\"Latitude_Of_Origin\",17.83333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",3992]]",
        3994=>"PROJCS[\"WGS_1984_Mercator_41\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",100.0],PARAMETER[\"Standard_Parallel_1\",-41.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3994]]",
        3995=>"PROJCS[\"WGS_1984_Arctic_Polar_Stereographic\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic_North_Pole\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",71.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3995]]",
        3996=>"PROJCS[\"WGS_1984_IBCAO_Polar_Stereographic\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic_North_Pole\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",75.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3996]]",
        3997=>"PROJCS[\"WGS_1984_Dubai_Local_TM\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",55.33333333333334],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",3997]]",
        4026=>"PROJCS[\"MOLDREF99_Moldova_TM\",GEOGCS[\"GCS_MOLDREF99\",DATUM[\"D_MOLDREF99\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Central_Meridian\",28.4],PARAMETER[\"Scale_Factor\",0.99994],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4026]]",
        4037=>"PROJCS[\"WGS_1984_TMzn35N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4037]]",
        4038=>"PROJCS[\"WGS_1984_TMzn36N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4038]]",
        4048=>"PROJCS[\"RGRDC_2005_Congo_TM_Zone_12\",GEOGCS[\"GCS_RGRDC_2005\",DATUM[\"D_Reseau_Geodesique_de_la_RDC_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4048]]",
        4049=>"PROJCS[\"RGRDC_2005_Congo_TM_Zone_14\",GEOGCS[\"GCS_RGRDC_2005\",DATUM[\"D_Reseau_Geodesique_de_la_RDC_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",14.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4049]]",
        4050=>"PROJCS[\"RGRDC_2005_Congo_TM_Zone_16\",GEOGCS[\"GCS_RGRDC_2005\",DATUM[\"D_Reseau_Geodesique_de_la_RDC_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",16.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4050]]",
        4051=>"PROJCS[\"RGRDC_2005_Congo_TM_Zone_18\",GEOGCS[\"GCS_RGRDC_2005\",DATUM[\"D_Reseau_Geodesique_de_la_RDC_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",18.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4051]]",
        4056=>"PROJCS[\"RGRDC_2005_Congo_TM_Zone_20\",GEOGCS[\"GCS_RGRDC_2005\",DATUM[\"D_Reseau_Geodesique_de_la_RDC_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",20.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4056]]",
        4057=>"PROJCS[\"RGRDC_2005_Congo_TM_Zone_22\",GEOGCS[\"GCS_RGRDC_2005\",DATUM[\"D_Reseau_Geodesique_de_la_RDC_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",22.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4057]]",
        4058=>"PROJCS[\"RGRDC_2005_Congo_TM_Zone_24\",GEOGCS[\"GCS_RGRDC_2005\",DATUM[\"D_Reseau_Geodesique_de_la_RDC_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4058]]",
        4059=>"PROJCS[\"RGRDC_2005_Congo_TM_Zone_26\",GEOGCS[\"GCS_RGRDC_2005\",DATUM[\"D_Reseau_Geodesique_de_la_RDC_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",26.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4059]]",
        4060=>"PROJCS[\"RGRDC_2005_Congo_TM_Zone_28\",GEOGCS[\"GCS_RGRDC_2005\",DATUM[\"D_Reseau_Geodesique_de_la_RDC_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",28.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4060]]",
        4061=>"PROJCS[\"RGRDC_2005_UTM_Zone_33S\",GEOGCS[\"GCS_RGRDC_2005\",DATUM[\"D_Reseau_Geodesique_de_la_RDC_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4061]]",
        4062=>"PROJCS[\"RGRDC_2005_UTM_Zone_34S\",GEOGCS[\"GCS_RGRDC_2005\",DATUM[\"D_Reseau_Geodesique_de_la_RDC_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4062]]",
        4063=>"PROJCS[\"RGRDC_2005_UTM_Zone_35S\",GEOGCS[\"GCS_RGRDC_2005\",DATUM[\"D_Reseau_Geodesique_de_la_RDC_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4063]]",
        4071=>"PROJCS[\"Chua_UTM_Zone_23S\",GEOGCS[\"GCS_Chua\",DATUM[\"D_Chua\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4071]]",
        4082=>"PROJCS[\"REGCAN95_UTM_Zone_27N\",GEOGCS[\"GCS_REGCAN95\",DATUM[\"D_Red_Geodesica_de_Canarias_1995\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4082]]",
        4083=>"PROJCS[\"REGCAN95_UTM_Zone_28N\",GEOGCS[\"GCS_REGCAN95\",DATUM[\"D_Red_Geodesica_de_Canarias_1995\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4083]]",
        4087=>"PROJCS[\"WGS_1984_World_Equidistant_Cylindrical\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Equidistant_Cylindrical_Ellipsoidal\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4087]]",
        4088=>"PROJCS[\"World_Equidistant_Cylindrical_(Sphere)\",GEOGCS[\"GCS_Sphere_GRS_1980_Authalic\",DATUM[\"D_Sphere_GRS_1980_Authalic\",SPHEROID[\"Sphere_GRS_1980_Authalic\",6371007.0,0.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Equidistant_Cylindrical\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4088]]",
        4093=>"PROJCS[\"ETRS_1989_DKTM1\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.99998],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4093]]",
        4094=>"PROJCS[\"ETRS_1989_DKTM2\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Central_Meridian\",10.0],PARAMETER[\"Scale_Factor\",0.99998],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4094]]",
        4095=>"PROJCS[\"ETRS_1989_DKTM3\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Central_Meridian\",11.75],PARAMETER[\"Scale_Factor\",0.99998],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4095]]",
        4096=>"PROJCS[\"ETRS_1989_DKTM4\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4096]]",
        4217=>"PROJCS[\"NAD_1983_BLM_Zone_59N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4217]]",
        4390=>"PROJCS[\"Kertau_1968_Johor_Grid\",GEOGCS[\"GCS_Kertau\",DATUM[\"D_Kertau\",SPHEROID[\"Everest_1830_Modified\",6377304.063,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",103.5627583333334],PARAMETER[\"Latitude_Of_Origin\",2.042583333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4390]]",
        4391=>"PROJCS[\"Kertau_1968_Sembilan_and_Melaka_Grid\",GEOGCS[\"GCS_Kertau\",DATUM[\"D_Kertau\",SPHEROID[\"Everest_1830_Modified\",6377304.063,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",-242.005],PARAMETER[\"False_Northing\",-948.547],PARAMETER[\"Central_Meridian\",101.9411666666667],PARAMETER[\"Latitude_Of_Origin\",2.712283333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4391]]",
        4392=>"PROJCS[\"Kertau_1968_Pahang_Grid\",GEOGCS[\"GCS_Kertau\",DATUM[\"D_Kertau\",SPHEROID[\"Everest_1830_Modified\",6377304.063,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",102.4361777777778],PARAMETER[\"Latitude_Of_Origin\",3.710972222222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4392]]",
        4393=>"PROJCS[\"Kertau_1968_Selangor_Grid\",GEOGCS[\"GCS_Kertau\",DATUM[\"D_Kertau\",SPHEROID[\"Everest_1830_Modified\",6377304.063,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",-21759.438],PARAMETER[\"False_Northing\",55960.906],PARAMETER[\"Central_Meridian\",101.5082444444444],PARAMETER[\"Latitude_Of_Origin\",3.680344444444444],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4393]]",
        4394=>"PROJCS[\"Kertau_1968_Terengganu_Grid\",GEOGCS[\"GCS_Kertau\",DATUM[\"D_Kertau\",SPHEROID[\"Everest_1830_Modified\",6377304.063,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",102.8952083333333],PARAMETER[\"Latitude_Of_Origin\",4.946141666666668],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4394]]",
        4395=>"PROJCS[\"Kertau_1968_Pinang_Grid\",GEOGCS[\"GCS_Kertau\",DATUM[\"D_Kertau\",SPHEROID[\"Everest_1830_Modified\",6377304.063,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",100.3458694444445],PARAMETER[\"Latitude_Of_Origin\",5.421325],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4395]]",
        4396=>"PROJCS[\"Kertau_1968_Kedah_and_Perlis_Grid\",GEOGCS[\"GCS_Kertau\",DATUM[\"D_Kertau\",SPHEROID[\"Everest_1830_Modified\",6377304.063,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",100.6375944444445],PARAMETER[\"Latitude_Of_Origin\",5.965147222222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4396]]",
        4397=>"PROJCS[\"Kertau_1968_Perak_Revised_Grid\",GEOGCS[\"GCS_Kertau\",DATUM[\"D_Kertau\",SPHEROID[\"Everest_1830_Modified\",6377304.063,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",133453.669],PARAMETER[\"Central_Meridian\",100.8167666666667],PARAMETER[\"Latitude_Of_Origin\",4.859380555555557],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4397]]",
        4398=>"PROJCS[\"Kertau_1968_Kelantan_Grid\",GEOGCS[\"GCS_Kertau\",DATUM[\"D_Kertau\",SPHEROID[\"Everest_1830_Modified\",6377304.063,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",102.1772916666667],PARAMETER[\"Latitude_Of_Origin\",5.893922222222224],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4398]]",
        4399=>"PROJCS[\"NAD_1927_BLM_Zone_59N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4399]]",
        4400=>"PROJCS[\"NAD_1927_BLM_Zone_60N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4400]]",
        4401=>"PROJCS[\"NAD_1927_BLM_Zone_1N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4401]]",
        4402=>"PROJCS[\"NAD_1927_BLM_Zone_2N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4402]]",
        4403=>"PROJCS[\"NAD_1927_BLM_Zone_3N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4403]]",
        4404=>"PROJCS[\"NAD_1927_BLM_Zone_4N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4404]]",
        4405=>"PROJCS[\"NAD_1927_BLM_Zone_5N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4405]]",
        4406=>"PROJCS[\"NAD_1927_BLM_Zone_6N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4406]]",
        4407=>"PROJCS[\"NAD_1927_BLM_Zone_7N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4407]]",
        4408=>"PROJCS[\"NAD_1927_BLM_Zone_8N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4408]]",
        4409=>"PROJCS[\"NAD_1927_BLM_Zone_9N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4409]]",
        4410=>"PROJCS[\"NAD_1927_BLM_Zone_10N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4410]]",
        4411=>"PROJCS[\"NAD_1927_BLM_Zone_11N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4411]]",
        4412=>"PROJCS[\"NAD_1927_BLM_Zone_12N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4412]]",
        4413=>"PROJCS[\"NAD_1927_BLM_Zone_13N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4413]]",
        4414=>"PROJCS[\"NAD_1983_HARN_Guam_Map_Grid\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Central_Meridian\",144.75],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",13.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4414]]",
        4415=>"PROJCS[\"Katanga_1955_Katanga_Lambert\",GEOGCS[\"GCS_Katanga_1955\",DATUM[\"D_Katanga_1955\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",26.0],PARAMETER[\"Standard_Parallel_1\",-6.5],PARAMETER[\"Standard_Parallel_2\",-11.5],PARAMETER[\"Latitude_Of_Origin\",-9.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4415]]",
        4417=>"PROJCS[\"Pulkovo_1942_Adj_1983_3_Degree_GK_Zone_7\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",7500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4417]]",
        4418=>"PROJCS[\"NAD_1927_BLM_Zone_18N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4418]]",
        4419=>"PROJCS[\"NAD_1927_BLM_Zone_19N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4419]]",
        4420=>"PROJCS[\"NAD_1983_BLM_Zone_60N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4420]]",
        4421=>"PROJCS[\"NAD_1983_BLM_Zone_1N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4421]]",
        4422=>"PROJCS[\"NAD_1983_BLM_Zone_2N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4422]]",
        4423=>"PROJCS[\"NAD_1983_BLM_Zone_3N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4423]]",
        4424=>"PROJCS[\"NAD_1983_BLM_Zone_4N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4424]]",
        4425=>"PROJCS[\"NAD_1983_BLM_Zone_5N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4425]]",
        4426=>"PROJCS[\"NAD_1983_BLM_Zone_6N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4426]]",
        4427=>"PROJCS[\"NAD_1983_BLM_Zone_7N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4427]]",
        4428=>"PROJCS[\"NAD_1983_BLM_Zone_8N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4428]]",
        4429=>"PROJCS[\"NAD_1983_BLM_Zone_9N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4429]]",
        4430=>"PROJCS[\"NAD_1983_BLM_Zone_10N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4430]]",
        4431=>"PROJCS[\"NAD_1983_BLM_Zone_11N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4431]]",
        4432=>"PROJCS[\"NAD_1983_BLM_Zone_12N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4432]]",
        4433=>"PROJCS[\"NAD_1983_BLM_Zone_13N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4433]]",
        4434=>"PROJCS[\"Pulkovo_1942_Adj_1983_3_Degree_GK_Zone_8\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",8500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4434]]",
        4437=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Puerto_Rico_Virgin_Isls_FIPS_5200\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Central_Meridian\",-66.43333333333334],PARAMETER[\"Standard_Parallel_1\",18.03333333333333],PARAMETER[\"Standard_Parallel_2\",18.43333333333333],PARAMETER[\"Latitude_Of_Origin\",17.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4437]]",
        4438=>"PROJCS[\"NAD_1983_BLM_Zone_18N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4438]]",
        4439=>"PROJCS[\"NAD_1983_BLM_Zone_19N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4439]]",
        4455=>"PROJCS[\"NAD_1927_StatePlane_Pennsylvania_South_FIPS_3702\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",39.93333333333333],PARAMETER[\"Standard_Parallel_2\",40.96666666666667],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4455]]",
        4456=>"PROJCS[\"NAD_1927_StatePlane_New_York_Long_Island_FIPS_3104\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-74.0],PARAMETER[\"Standard_Parallel_1\",40.66666666666666],PARAMETER[\"Standard_Parallel_2\",41.03333333333333],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4456]]",
        4457=>"PROJCS[\"NAD_1983_StatePlane_South_Dakota_North_FIPS_4001_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",44.41666666666666],PARAMETER[\"Standard_Parallel_2\",45.68333333333333],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4457]]",
        4462=>"PROJCS[\"WGS_1984_Australian_Centre_for_Remote_Sensing_Lambert\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",132.0],PARAMETER[\"Standard_Parallel_1\",-18.0],PARAMETER[\"Standard_Parallel_2\",-36.0],PARAMETER[\"Latitude_Of_Origin\",-27.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4462]]",
        4467=>"PROJCS[\"RGSPM_2006_UTM_Zone_21N\",GEOGCS[\"GCS_RGSPM_2006\",DATUM[\"D_Reseau_Geodesique_de_St_Pierre_et_Miquelon_2006\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4467]]",
        4471=>"PROJCS[\"RGM_2004_UTM_Zone_38S\",GEOGCS[\"GCS_RGM_2004\",DATUM[\"D_Reseau_Geodesique_de_Mayotte_2004\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4471]]",
        4474=>"PROJCS[\"Cadastre_1997_UTM_Zone_38S\",GEOGCS[\"GCS_Cadastre_1997\",DATUM[\"D_Cadastre_1997\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5879]]",
        4484=>"PROJCS[\"Mexican_Datum_1993_UTM_Zone_11N\",GEOGCS[\"GCS_Mexican_Datum_of_1993\",DATUM[\"D_Mexican_Datum_of_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4484]]",
        4485=>"PROJCS[\"Mexican_Datum_1993_UTM_Zone_12N\",GEOGCS[\"GCS_Mexican_Datum_of_1993\",DATUM[\"D_Mexican_Datum_of_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4485]]",
        4486=>"PROJCS[\"Mexican_Datum_1993_UTM_Zone_13N\",GEOGCS[\"GCS_Mexican_Datum_of_1993\",DATUM[\"D_Mexican_Datum_of_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4486]]",
        4487=>"PROJCS[\"Mexican_Datum_1993_UTM_Zone_14N\",GEOGCS[\"GCS_Mexican_Datum_of_1993\",DATUM[\"D_Mexican_Datum_of_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4487]]",
        4488=>"PROJCS[\"Mexican_Datum_1993_UTM_Zone_15N\",GEOGCS[\"GCS_Mexican_Datum_of_1993\",DATUM[\"D_Mexican_Datum_of_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4488]]",
        4489=>"PROJCS[\"Mexican_Datum_1993_UTM_Zone_16N\",GEOGCS[\"GCS_Mexican_Datum_of_1993\",DATUM[\"D_Mexican_Datum_of_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4489]]",
        4491=>"PROJCS[\"CGCS2000_GK_Zone_13\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",13500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4491]]",
        4492=>"PROJCS[\"CGCS2000_GK_Zone_14\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",14500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4492]]",
        4493=>"PROJCS[\"CGCS2000_GK_Zone_15\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",15500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4493]]",
        4494=>"PROJCS[\"CGCS2000_GK_Zone_16\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",16500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4494]]",
        4495=>"PROJCS[\"CGCS2000_GK_Zone_17\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",17500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4495]]",
        4496=>"PROJCS[\"CGCS2000_GK_Zone_18\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",18500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4496]]",
        4497=>"PROJCS[\"CGCS2000_GK_Zone_19\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",19500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4497]]",
        4498=>"PROJCS[\"CGCS2000_GK_Zone_20\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",20500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4498]]",
        4499=>"PROJCS[\"CGCS2000_GK_Zone_21\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",21500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4499]]",
        4500=>"PROJCS[\"CGCS2000_GK_Zone_22\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",22500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4500]]",
        4501=>"PROJCS[\"CGCS2000_GK_Zone_23\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",23500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4501]]",
        4502=>"PROJCS[\"CGCS2000_GK_CM_75E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4502]]",
        4503=>"PROJCS[\"CGCS2000_GK_CM_81E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4503]]",
        4504=>"PROJCS[\"CGCS2000_GK_CM_87E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4504]]",
        4505=>"PROJCS[\"CGCS2000_GK_CM_93E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4505]]",
        4506=>"PROJCS[\"CGCS2000_GK_CM_99E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4506]]",
        4507=>"PROJCS[\"CGCS2000_GK_CM_105E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4507]]",
        4508=>"PROJCS[\"CGCS2000_GK_CM_111E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4508]]",
        4509=>"PROJCS[\"CGCS2000_GK_CM_117E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4509]]",
        4510=>"PROJCS[\"CGCS2000_GK_CM_123E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4510]]",
        4511=>"PROJCS[\"CGCS2000_GK_CM_129E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4511]]",
        4512=>"PROJCS[\"CGCS2000_GK_CM_135E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4512]]",
        4513=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_25\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",25500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4513]]",
        4514=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_26\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",26500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",78.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4514]]",
        4515=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_27\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",27500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4515]]",
        4516=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_28\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",28500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",84.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4516]]",
        4517=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_29\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",29500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4517]]",
        4518=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_30\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",30500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4518]]",
        4519=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_31\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",31500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4519]]",
        4520=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_32\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",32500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",96.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4520]]",
        4521=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_33\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",33500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4521]]",
        4522=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_34\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",34500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",102.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4522]]",
        4523=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_35\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",35500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4523]]",
        4524=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_36\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",36500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",108.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4524]]",
        4525=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_37\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",37500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4525]]",
        4526=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_38\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",38500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",114.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4526]]",
        4527=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_39\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",39500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4527]]",
        4528=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_40\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",40500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",120.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4528]]",
        4529=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_41\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",41500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4529]]",
        4530=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_42\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",42500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",126.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4530]]",
        4531=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_43\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",43500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4531]]",
        4532=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_44\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",44500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",132.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4532]]",
        4533=>"PROJCS[\"CGCS2000_3_Degree_GK_Zone_45\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",45500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4533]]",
        4534=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_75E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4534]]",
        4535=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_78E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",78.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4535]]",
        4536=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_81E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4536]]",
        4537=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_84E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",84.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4537]]",
        4538=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_87E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4538]]",
        4539=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_90E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4539]]",
        4540=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_93E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4540]]",
        4541=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_96E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",96.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4541]]",
        4542=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_99E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4542]]",
        4543=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_102E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",102.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4543]]",
        4544=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_105E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4544]]",
        4545=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_108E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",108.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4545]]",
        4546=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_111E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4546]]",
        4547=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_114E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",114.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4547]]",
        4548=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_117E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4548]]",
        4549=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_120E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",120.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4549]]",
        4550=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_123E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4550]]",
        4551=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_126E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",126.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4551]]",
        4552=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_129E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4552]]",
        4553=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_132E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",132.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4553]]",
        4554=>"PROJCS[\"CGCS2000_3_Degree_GK_CM_135E\",GEOGCS[\"GCS_China_Geodetic_Coordinate_System_2000\",DATUM[\"D_China_2000\",SPHEROID[\"CGCS2000\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4554]]",
        4559=>"PROJCS[\"RRAF_1991_UTM_20N\",GEOGCS[\"GCS_RRAF_1991\",DATUM[\"D_RRAF_1991\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4559]]",
        4568=>"PROJCS[\"New_Beijing_Gauss_Kruger_Zone_13\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",13500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4568]]",
        4569=>"PROJCS[\"New_Beijing_Gauss_Kruger_Zone_14\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",14500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4569]]",
        4570=>"PROJCS[\"New_Beijing_Gauss_Kruger_Zone_15\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",15500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4570]]",
        4571=>"PROJCS[\"New_Beijing_Gauss_Kruger_Zone_16\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",16500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4571]]",
        4572=>"PROJCS[\"New_Beijing_Gauss_Kruger_Zone_17\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",17500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4572]]",
        4573=>"PROJCS[\"New_Beijing_Gauss_Kruger_Zone_18\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",18500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4573]]",
        4574=>"PROJCS[\"New_Beijing_Gauss_Kruger_Zone_19\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",19500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4574]]",
        4575=>"PROJCS[\"New_Beijing_Gauss_Kruger_Zone_20\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",20500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4575]]",
        4576=>"PROJCS[\"New_Beijing_Gauss_Kruger_Zone_21\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",21500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4576]]",
        4577=>"PROJCS[\"New_Beijing_Gauss_Kruger_Zone_22\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",22500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4577]]",
        4578=>"PROJCS[\"New_Beijing_Gauss_Kruger_Zone_23\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",23500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4578]]",
        4579=>"PROJCS[\"New_Beijing_Gauss_Kruger_CM_75E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4579]]",
        4580=>"PROJCS[\"New_Beijing_Gauss_Kruger_CM_81E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4580]]",
        4581=>"PROJCS[\"New_Beijing_Gauss_Kruger_CM_87E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4581]]",
        4582=>"PROJCS[\"New_Beijing_Gauss_Kruger_CM_93E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4582]]",
        4583=>"PROJCS[\"New_Beijing_Gauss_Kruger_CM_99E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4583]]",
        4584=>"PROJCS[\"New_Beijing_Gauss_Kruger_CM_105E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4584]]",
        4585=>"PROJCS[\"New_Beijing_Gauss_Kruger_CM_111E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4585]]",
        4586=>"PROJCS[\"New_Beijing_Gauss_Kruger_CM_117E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4586]]",
        4587=>"PROJCS[\"New_Beijing_Gauss_Kruger_CM_123E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4587]]",
        4588=>"PROJCS[\"New_Beijing_Gauss_Kruger_CM_129E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4588]]",
        4589=>"PROJCS[\"New_Beijing_Gauss_Kruger_CM_135E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4589]]",
        4647=>"PROJCS[\"ETRS_1989_UTM_Zone_N32\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",32500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4647]]",
        4652=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_25\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",25500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4652]]",
        4653=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_26\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",26500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",78.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4653]]",
        4654=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_27\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",27500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4654]]",
        4655=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_28\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",28500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",84.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4655]]",
        4656=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_29\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",29500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4656]]",
        4766=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_30\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",30500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4766]]",
        4767=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_31\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",31500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4767]]",
        4768=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_32\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",32500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",96.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4768]]",
        4769=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_33\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",33500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4769]]",
        4770=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_34\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",34500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",102.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4770]]",
        4771=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_35\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",35500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4771]]",
        4772=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_36\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",36500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",108.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4772]]",
        4773=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_37\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",37500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4773]]",
        4774=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_38\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",38500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",114.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4774]]",
        4775=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_39\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",39500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4775]]",
        4776=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_40\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",40500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",120.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4776]]",
        4777=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_41\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",41500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4777]]",
        4778=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_42\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",42500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",126.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4778]]",
        4779=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_43\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",43500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4779]]",
        4780=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_44\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",44500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",132.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4780]]",
        4781=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_Zone_45\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",45500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4781]]",
        4782=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_75E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4782]]",
        4783=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_78E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",78.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4783]]",
        4784=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_81E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4784]]",
        4785=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_84E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",84.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4785]]",
        4786=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_87E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4786]]",
        4787=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_90E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4787]]",
        4788=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_93E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4788]]",
        4789=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_96E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",96.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4789]]",
        4790=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_99E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4790]]",
        4791=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_102E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",102.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4791]]",
        4792=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_105E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4792]]",
        4793=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_108E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",108.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4793]]",
        4794=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_111E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4794]]",
        4795=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_114E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",114.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4795]]",
        4796=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_117E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4796]]",
        4797=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_120E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",120.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4797]]",
        4798=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_123E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4798]]",
        4799=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_126E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",126.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4799]]",
        4800=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_129E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4800]]",
        4822=>"PROJCS[\"New_Beijing_3_Degree_Gauss_Kruger_CM_135E\",GEOGCS[\"GCS_New_Beijing\",DATUM[\"D_New_Beijing\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4822]]",
        4826=>"PROJCS[\"WGS_1984_Cape_Verde_Grid\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",161587.83],PARAMETER[\"False_Northing\",128511.202],PARAMETER[\"Central_Meridian\",-24.0],PARAMETER[\"Standard_Parallel_1\",15.0],PARAMETER[\"Standard_Parallel_2\",16.66666666666667],PARAMETER[\"Latitude_Of_Origin\",15.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4826]]",
        4839=>"PROJCS[\"ETRS_1989_LCC_Germany_N-E\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",10.5],PARAMETER[\"Standard_Parallel_1\",48.66666666666666],PARAMETER[\"Standard_Parallel_2\",53.66666666666666],PARAMETER[\"Latitude_Of_Origin\",51.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",4839]]",
        5014=>"PROJCS[\"PTRA08_UTM_Zone_25N\",GEOGCS[\"GCS_PTRA08\",DATUM[\"D_PTRA08\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5014]]",
        5015=>"PROJCS[\"PTRA08_UTM_Zone_26N\",GEOGCS[\"GCS_PTRA08\",DATUM[\"D_PTRA08\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5015]]",
        5016=>"PROJCS[\"PTRA08_UTM_Zone_28N\",GEOGCS[\"GCS_PTRA08\",DATUM[\"D_PTRA08\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5016]]",
        5018=>"PROJCS[\"Lisbon_Portuguese_Grid_New\",GEOGCS[\"GCS_Lisbon\",DATUM[\"D_Lisbon\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-8.131906111111112],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",39.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5018]]",
        5048=>"PROJCS[\"ETRS_1989_TM35FIN_NE\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5048]]",
        5069=>"PROJCS[\"NAD_1927_Contiguous_USA_Albers\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-96.0],PARAMETER[\"Standard_Parallel_1\",29.5],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",23.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5069]]",
        5070=>"PROJCS[\"NAD_1983_Contiguous_USA_Albers\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-96.0],PARAMETER[\"Standard_Parallel_1\",29.5],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",23.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5070]]",
        5071=>"PROJCS[\"NAD_1983_HARN_Contiguous_USA_Albers\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-96.0],PARAMETER[\"Standard_Parallel_1\",29.5],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",23.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5071]]",
        5072=>"PROJCS[\"NAD_1983_NSRS2007_Contiguous_USA_Albers\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-96.0],PARAMETER[\"Standard_Parallel_1\",29.5],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",23.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5072]]",
        5105=>"PROJCS[\"ETRS_1989_NTM_Zone_5\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",5.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5105]]",
        5106=>"PROJCS[\"ETRS_1989_NTM_Zone_6\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",6.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5106]]",
        5107=>"PROJCS[\"ETRS_1989_NTM_Zone_7\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",7.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5107]]",
        5108=>"PROJCS[\"ETRS_1989_NTM_Zone_8\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",8.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5108]]",
        5109=>"PROJCS[\"ETRS_1989_NTM_Zone_9\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",9.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5109]]",
        5110=>"PROJCS[\"ETRS_1989_NTM_Zone_10\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",10.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5110]]",
        5111=>"PROJCS[\"ETRS_1989_NTM_Zone_11\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",11.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5111]]",
        5112=>"PROJCS[\"ETRS_1989_NTM_Zone_12\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",12.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5112]]",
        5113=>"PROJCS[\"ETRS_1989_NTM_Zone_13\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",13.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5113]]",
        5114=>"PROJCS[\"ETRS_1989_NTM_Zone_14\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",14.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5114]]",
        5115=>"PROJCS[\"ETRS_1989_NTM_Zone_15\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",15.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5115]]",
        5116=>"PROJCS[\"ETRS_1989_NTM_Zone_16\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",16.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5116]]",
        5117=>"PROJCS[\"ETRS_1989_NTM_Zone_17\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",17.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5117]]",
        5118=>"PROJCS[\"ETRS_1989_NTM_Zone_18\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",18.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5118]]",
        5119=>"PROJCS[\"ETRS_1989_NTM_Zone_19\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",19.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5119]]",
        5120=>"PROJCS[\"ETRS_1989_NTM_Zone_20\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",20.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5120]]",
        5121=>"PROJCS[\"ETRS_1989_NTM_Zone_21\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",21.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5121]]",
        5122=>"PROJCS[\"ETRS_1989_NTM_Zone_22\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",22.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5122]]",
        5123=>"PROJCS[\"ETRS_1989_NTM_Zone_23\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",23.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5123]]",
        5124=>"PROJCS[\"ETRS_1989_NTM_Zone_24\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",24.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5124]]",
        5125=>"PROJCS[\"ETRS_1989_NTM_Zone_25\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",25.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5125]]",
        5126=>"PROJCS[\"ETRS_1989_NTM_Zone_26\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",26.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5126]]",
        5127=>"PROJCS[\"ETRS_1989_NTM_Zone_27\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",27.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5127]]",
        5128=>"PROJCS[\"ETRS_1989_NTM_Zone_28\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",28.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5128]]",
        5129=>"PROJCS[\"ETRS_1989_NTM_Zone_29\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",29.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5129]]",
        5130=>"PROJCS[\"ETRS_1989_NTM_Zone_30\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",30.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5130]]",
        5167=>"PROJCS[\"Korean_1985_Korea_East_Sea_Belt\",GEOGCS[\"GCS_Korean_Datum_1985\",DATUM[\"D_Korean_Datum_1985\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",131.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5167]]",
        5168=>"PROJCS[\"Korean_1985_Korea_Central_Belt_Jeju\",GEOGCS[\"GCS_Korean_Datum_1985\",DATUM[\"D_Korean_Datum_1985\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",550000.0],PARAMETER[\"Central_Meridian\",127.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5168]]",
        5173=>"PROJCS[\"Korean_1985_Modified_Korea_West_Belt\",GEOGCS[\"GCS_Korean_Datum_1985\",DATUM[\"D_Korean_Datum_1985\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",125.0028902777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5173]]",
        5174=>"PROJCS[\"Korean_1985_Modified_Korea_Central_Belt\",GEOGCS[\"GCS_Korean_Datum_1985\",DATUM[\"D_Korean_Datum_1985\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",127.0028902777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5174]]",
        5175=>"PROJCS[\"Korean_1985_Modified_Korea_Central_Belt_Jeju\",GEOGCS[\"GCS_Korean_Datum_1985\",DATUM[\"D_Korean_Datum_1985\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",550000.0],PARAMETER[\"Central_Meridian\",127.0028902777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5175]]",
        5176=>"PROJCS[\"Korean_1985_Modified_Korea_East_Belt\",GEOGCS[\"GCS_Korean_Datum_1985\",DATUM[\"D_Korean_Datum_1985\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",129.0028902777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5176]]",
        5177=>"PROJCS[\"Korean_1985_Modified_Korea_East_Sea_Belt\",GEOGCS[\"GCS_Korean_Datum_1985\",DATUM[\"D_Korean_Datum_1985\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",131.0028902777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5177]]",
        5178=>"PROJCS[\"Korean_1985_Korea_Unified_Coordinate_System\",GEOGCS[\"GCS_Korean_Datum_1985\",DATUM[\"D_Korean_Datum_1985\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",127.5],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5178]]",
        5179=>"PROJCS[\"Korea_2000_Korea_Unified_Coordinate_System\",GEOGCS[\"GCS_Korea_2000\",DATUM[\"D_Korea_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",127.5],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5179]]",
        5180=>"PROJCS[\"Korea_2000_Korea_West_Belt\",GEOGCS[\"GCS_Korea_2000\",DATUM[\"D_Korea_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",125.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5180]]",
        5181=>"PROJCS[\"Korea_2000_Korea_Central_Belt\",GEOGCS[\"GCS_Korea_2000\",DATUM[\"D_Korea_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",127.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5181]]",
        5182=>"PROJCS[\"Korea_2000_Korea_Central_Belt_Jeju\",GEOGCS[\"GCS_Korea_2000\",DATUM[\"D_Korea_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",550000.0],PARAMETER[\"Central_Meridian\",127.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5182]]",
        5183=>"PROJCS[\"Korea_2000_Korea_East_Belt\",GEOGCS[\"GCS_Korea_2000\",DATUM[\"D_Korea_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5183]]",
        5184=>"PROJCS[\"Korea_2000_Korea_East_Sea_Belt\",GEOGCS[\"GCS_Korea_2000\",DATUM[\"D_Korea_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",131.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5184]]",
        5185=>"PROJCS[\"Korea_2000_Korea_West_Belt_2010\",GEOGCS[\"GCS_Korea_2000\",DATUM[\"D_Korea_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",600000.0],PARAMETER[\"Central_Meridian\",125.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5185]]",
        5186=>"PROJCS[\"Korea_2000_Korea_Central_Belt_2010\",GEOGCS[\"GCS_Korea_2000\",DATUM[\"D_Korea_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",600000.0],PARAMETER[\"Central_Meridian\",127.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5186]]",
        5187=>"PROJCS[\"Korea_2000_Korea_East_Belt_2010\",GEOGCS[\"GCS_Korea_2000\",DATUM[\"D_Korea_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",600000.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5187]]",
        5188=>"PROJCS[\"Korea_2000_Korea_East_Sea_Belt_2010\",GEOGCS[\"GCS_Korea_2000\",DATUM[\"D_Korea_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",600000.0],PARAMETER[\"Central_Meridian\",131.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5188]]",
        5221=>"PROJCS[\"S-JTSK_Ferro_Krovak_East_North\",GEOGCS[\"GCS_S_JTSK_Ferro\",DATUM[\"D_S_JTSK\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Krovak\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Pseudo_Standard_Parallel_1\",78.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Azimuth\",30.28813975277778],PARAMETER[\"Longitude_Of_Center\",42.5],PARAMETER[\"Latitude_Of_Center\",49.5],PARAMETER[\"X_Scale\",-1.0],PARAMETER[\"Y_Scale\",1.0],PARAMETER[\"XY_Plane_Rotation\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5221]]",
        5223=>"PROJCS[\"WGS_1984_UTM_Gabon_TM\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5223]]",
        5234=>"PROJCS[\"Kandawala_Sri_Lanka_Grid\",GEOGCS[\"GCS_Kandawala\",DATUM[\"D_Kandawala\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Central_Meridian\",80.77171111111112],PARAMETER[\"Scale_Factor\",0.9999238418],PARAMETER[\"Latitude_Of_Origin\",7.000480277777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5234]]",
        5235=>"PROJCS[\"SLD99_Sri_Lanka_Grid_1999\",GEOGCS[\"GCS_SLD99\",DATUM[\"D_Sri_Lanka_Datum_1999\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",80.77171308333334],PARAMETER[\"Scale_Factor\",0.9999238418],PARAMETER[\"Latitude_Of_Origin\",7.000471527777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5235]]",
        5243=>"PROJCS[\"ETRS_1989_LCC_Germany_E-N\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",10.5],PARAMETER[\"Standard_Parallel_1\",48.66666666666666],PARAMETER[\"Standard_Parallel_2\",53.66666666666666],PARAMETER[\"Latitude_Of_Origin\",51.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5243]]",
        5247=>"PROJCS[\"GDBD2009_GEORSO\",GEOGCS[\"GCS_GDBD2009\",DATUM[\"D_GDBD2009\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Rectified_Skew_Orthomorphic_Natural_Origin\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Scale_Factor\",0.99984],PARAMETER[\"Azimuth\",53.31580995],PARAMETER[\"Longitude_Of_Center\",115.0],PARAMETER[\"Latitude_Of_Center\",4.0],PARAMETER[\"XY_Plane_Rotation\",53.13010235415598],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5247]]",
        5253=>"PROJCS[\"TUREF_TM27\",GEOGCS[\"GCS_TUREF\",DATUM[\"D_Turkish_National_Reference_Frame\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5253]]",
        5254=>"PROJCS[\"TUREF_TM30\",GEOGCS[\"GCS_TUREF\",DATUM[\"D_Turkish_National_Reference_Frame\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5254]]",
        5255=>"PROJCS[\"TUREF_TM33\",GEOGCS[\"GCS_TUREF\",DATUM[\"D_Turkish_National_Reference_Frame\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5255]]",
        5256=>"PROJCS[\"TUREF_TM36\",GEOGCS[\"GCS_TUREF\",DATUM[\"D_Turkish_National_Reference_Frame\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",36.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5256]]",
        5257=>"PROJCS[\"TUREF_TM39\",GEOGCS[\"GCS_TUREF\",DATUM[\"D_Turkish_National_Reference_Frame\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5257]]",
        5258=>"PROJCS[\"TUREF_TM42\",GEOGCS[\"GCS_TUREF\",DATUM[\"D_Turkish_National_Reference_Frame\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",42.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5258]]",
        5259=>"PROJCS[\"TUREF_TM45\",GEOGCS[\"GCS_TUREF\",DATUM[\"D_Turkish_National_Reference_Frame\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5259]]",
        5266=>"PROJCS[\"DRUKREF_03_Bhutan_National_Grid\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5266]]",
        5269=>"PROJCS[\"TUREF_3_Degree_GK_Zone_9\",GEOGCS[\"GCS_TUREF\",DATUM[\"D_Turkish_National_Reference_Frame\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",9500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5269]]",
        5270=>"PROJCS[\"TUREF_3_Degree_GK_Zone_10\",GEOGCS[\"GCS_TUREF\",DATUM[\"D_Turkish_National_Reference_Frame\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",10500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5270]]",
        5271=>"PROJCS[\"TUREF_3_Degree_GK_Zone_11\",GEOGCS[\"GCS_TUREF\",DATUM[\"D_Turkish_National_Reference_Frame\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",11500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5271]]",
        5272=>"PROJCS[\"TUREF_3_Degree_GK_Zone_12\",GEOGCS[\"GCS_TUREF\",DATUM[\"D_Turkish_National_Reference_Frame\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",12500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",36.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5272]]",
        5273=>"PROJCS[\"TUREF_3_Degree_GK_Zone_13\",GEOGCS[\"GCS_TUREF\",DATUM[\"D_Turkish_National_Reference_Frame\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",13500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5273]]",
        5274=>"PROJCS[\"TUREF_3_Degree_GK_Zone_14\",GEOGCS[\"GCS_TUREF\",DATUM[\"D_Turkish_National_Reference_Frame\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",14500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",42.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5274]]",
        5275=>"PROJCS[\"TUREF_3_Degree_GK_Zone_15\",GEOGCS[\"GCS_TUREF\",DATUM[\"D_Turkish_National_Reference_Frame\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",15500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5275]]",
        5292=>"PROJCS[\"DRUKREF_03_Bumthang_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",90.73333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5292]]",
        5293=>"PROJCS[\"DRUKREF_03_Chhukha_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",89.55],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5293]]",
        5294=>"PROJCS[\"DRUKREF_03_Dagana_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",89.85],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5294]]",
        5295=>"PROJCS[\"DRUKREF_03_Gasa_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",90.03333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5295]]",
        5296=>"PROJCS[\"DRUKREF_03_Ha_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",90.15],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5296]]",
        5297=>"PROJCS[\"DRUKREF_03_Lhuentse_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",91.13333333333334],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5297]]",
        5298=>"PROJCS[\"DRUKREF_03_Mongar_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",91.23333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5298]]",
        5299=>"PROJCS[\"DRUKREF_03_Paro_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",89.35],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5299]]",
        5300=>"PROJCS[\"DRUKREF_03_Pemagatshel_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",91.35],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5300]]",
        5301=>"PROJCS[\"DRUKREF_03_Punakha_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",89.85],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5301]]",
        5302=>"PROJCS[\"DRUKREF_03_Samdrup_Jongkhar_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",91.56666666666666],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5302]]",
        5303=>"PROJCS[\"DRUKREF_03_Samtse_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",89.06666666666666],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5303]]",
        5304=>"PROJCS[\"DRUKREF_03_Sarpang_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",90.26666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5304]]",
        5305=>"PROJCS[\"DRUKREF_03_Thimphu_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",89.55],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5305]]",
        5306=>"PROJCS[\"DRUKREF_03_Trashigang_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",91.75],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5306]]",
        5307=>"PROJCS[\"DRUKREF_03_Trongsa_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",90.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5307]]",
        5308=>"PROJCS[\"DRUKREF_03_Tsirang_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",90.16666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5308]]",
        5309=>"PROJCS[\"DRUKREF_03_Wangdue_Phodrang_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",90.11666666666666],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5309]]",
        5310=>"PROJCS[\"DRUKREF_03_Yangtse_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",91.56666666666666],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5310]]",
        5311=>"PROJCS[\"DRUKREF_03_Zhemgang_TM\",GEOGCS[\"GCS_DRUKREF_03\",DATUM[\"D_Bhutan_National_Geodetic_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",-2500000.0],PARAMETER[\"Central_Meridian\",90.86666666666666],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5311]]",
        5316=>"PROJCS[\"ETRS_1989_FAROE_TM\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",-6000000.0],PARAMETER[\"Central_Meridian\",-7.0],PARAMETER[\"Scale_Factor\",0.999997],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5316]]",
        5320=>"PROJCS[\"NAD_1983_Teranet_Ontario_Lambert\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.0],PARAMETER[\"Standard_Parallel_1\",44.5],PARAMETER[\"Standard_Parallel_2\",54.5],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5320]]",
        5321=>"PROJCS[\"NAD_1983_CSRS_Teranet_Ontario_Lambert\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.0],PARAMETER[\"Standard_Parallel_1\",44.5],PARAMETER[\"Standard_Parallel_2\",54.5],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5321]]",
        5325=>"PROJCS[\"ISN_2004_Lambert_2004\",GEOGCS[\"GCS_ISN_2004\",DATUM[\"D_Islands_Network_2004\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1700000.0],PARAMETER[\"False_Northing\",300000.0],PARAMETER[\"Central_Meridian\",-19.0],PARAMETER[\"Standard_Parallel_1\",64.25],PARAMETER[\"Standard_Parallel_2\",65.75],PARAMETER[\"Latitude_Of_Origin\",65.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5325]]",
        5329=>"PROJCS[\"Gunung_Segara_Jakarta_NEIEZ\",GEOGCS[\"GCS_Gunung_Segara_Jakarta\",DATUM[\"D_Gunung_Segara\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Jakarta\",106.8077194444444],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Mercator\"],PARAMETER[\"False_Easting\",3900000.0],PARAMETER[\"False_Northing\",900000.0],PARAMETER[\"Central_Meridian\",3.192280555555556],PARAMETER[\"Standard_Parallel_1\",4.45405154589751],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5329]]",
        5330=>"PROJCS[\"Batavia_Jakarta_NEIEZ\",GEOGCS[\"GCS_Batavia_Jakarta\",DATUM[\"D_Batavia\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Jakarta\",106.8077194444444],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Mercator\"],PARAMETER[\"False_Easting\",3900000.0],PARAMETER[\"False_Northing\",900000.0],PARAMETER[\"Central_Meridian\",3.192280555555556],PARAMETER[\"Standard_Parallel_1\",4.45405154589751],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5330]]",
        5331=>"PROJCS[\"Makassar_Jakarta_NEIEZ\",GEOGCS[\"GCS_Makassar_Jakarta\",DATUM[\"D_Makassar\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Jakarta\",106.8077194444444],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Mercator\"],PARAMETER[\"False_Easting\",3900000.0],PARAMETER[\"False_Northing\",900000.0],PARAMETER[\"Central_Meridian\",3.192280555555556],PARAMETER[\"Standard_Parallel_1\",4.45405154589751],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5331]]",
        5337=>"PROJCS[\"Aratu_UTM_Zone_25S\",GEOGCS[\"GCS_Aratu\",DATUM[\"D_Aratu\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5337]]",
        5343=>"PROJCS[\"POSGAR_2007_Argentina_Zone_1\",GEOGCS[\"GCS_POSGAR_2007\",DATUM[\"D_POSGAR_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-72.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5343]]",
        5344=>"PROJCS[\"POSGAR_2007_Argentina_Zone_2\",GEOGCS[\"GCS_POSGAR_2007\",DATUM[\"D_POSGAR_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5344]]",
        5345=>"PROJCS[\"POSGAR_2007_Argentina_Zone_3\",GEOGCS[\"GCS_POSGAR_2007\",DATUM[\"D_POSGAR_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-66.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5345]]",
        5346=>"PROJCS[\"POSGAR_2007_Argentina_Zone_4\",GEOGCS[\"GCS_POSGAR_2007\",DATUM[\"D_POSGAR_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5346]]",
        5347=>"PROJCS[\"POSGAR_2007_Argentina_Zone_5\",GEOGCS[\"GCS_POSGAR_2007\",DATUM[\"D_POSGAR_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-60.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5347]]",
        5348=>"PROJCS[\"POSGAR_2007_Argentina_Zone_6\",GEOGCS[\"GCS_POSGAR_2007\",DATUM[\"D_POSGAR_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",6500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5348]]",
        5349=>"PROJCS[\"POSGAR_2007_Argentina_Zone_7\",GEOGCS[\"GCS_POSGAR_2007\",DATUM[\"D_POSGAR_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",7500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-54.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5349]]",
        5355=>"PROJCS[\"MARGEN_UTM_Zone_20S\",GEOGCS[\"GCS_MARGEN\",DATUM[\"D_Marco_Geodesico_Nacional\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5355]]",
        5356=>"PROJCS[\"MARGEN_UTM_Zone_19S\",GEOGCS[\"GCS_MARGEN\",DATUM[\"D_Marco_Geodesico_Nacional\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5356]]",
        5357=>"PROJCS[\"MARGEN_UTM_Zone_21S\",GEOGCS[\"GCS_MARGEN\",DATUM[\"D_Marco_Geodesico_Nacional\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5357]]",
        5361=>"PROJCS[\"SIRGAS-Chile_UTM_Zone_19S\",GEOGCS[\"GCS_SIRGAS-Chile\",DATUM[\"D_SIRGAS-Chile\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5361]]",
        5362=>"PROJCS[\"SIRGAS-Chile_UTM_Zone_18S\",GEOGCS[\"GCS_SIRGAS-Chile\",DATUM[\"D_SIRGAS-Chile\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5362]]",
        5367=>"PROJCS[\"CRTM05\",GEOGCS[\"GCS_CR05\",DATUM[\"D_Costa_Rica_2005\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5367]]",
        5382=>"PROJCS[\"SIRGAS-ROU98_UTM_Zone_21S\",GEOGCS[\"GCS_SIRGAS-ROU98\",DATUM[\"D_SIRGAS-ROU98\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5382]]",
        5383=>"PROJCS[\"SIRGAS-ROU98_UTM_Zone_22S\",GEOGCS[\"GCS_SIRGAS-ROU98\",DATUM[\"D_SIRGAS-ROU98\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5383]]",
        5387=>"PROJCS[\"Peru96_UTM_Zone_18S\",GEOGCS[\"GCS_Peru96\",DATUM[\"D_Peru96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5387]]",
        5388=>"PROJCS[\"Peru96_UTM_Zone_17S\",GEOGCS[\"GCS_Peru96\",DATUM[\"D_Peru96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5839]]",
        5389=>"PROJCS[\"Peru96_UTM_Zone_19S\",GEOGCS[\"GCS_Peru96\",DATUM[\"D_Peru96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5389]]",
        5396=>"PROJCS[\"SIRGAS_2000_UTM_Zone_26S\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5396]]",
        5456=>"PROJCS[\"Ocotepeque_1935_Costa_Rica_Norte\",GEOGCS[\"GCS_Ocotepeque_1935\",DATUM[\"D_Ocotepeque_1935\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",271820.522],PARAMETER[\"Central_Meridian\",-84.33333333333333],PARAMETER[\"Standard_Parallel_1\",10.46666666666667],PARAMETER[\"Scale_Factor\",0.99995696],PARAMETER[\"Latitude_Of_Origin\",10.46666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5456]]",
        5457=>"PROJCS[\"Ocotepeque_1935_Costa_Rica_Sur\",GEOGCS[\"GCS_Ocotepeque_1935\",DATUM[\"D_Ocotepeque_1935\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",327987.436],PARAMETER[\"Central_Meridian\",-83.66666666666667],PARAMETER[\"Standard_Parallel_1\",9.0],PARAMETER[\"Scale_Factor\",0.99995696],PARAMETER[\"Latitude_Of_Origin\",9.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5457]]",
        5459=>"PROJCS[\"Ocotepeque_1935_Guatemala_Sur\",GEOGCS[\"GCS_Ocotepeque_1935\",DATUM[\"D_Ocotepeque_1935\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",325992.681],PARAMETER[\"Central_Meridian\",-90.33333333333333],PARAMETER[\"Standard_Parallel_1\",14.9],PARAMETER[\"Scale_Factor\",0.99989906],PARAMETER[\"Latitude_Of_Origin\",14.9],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5459]]",
        5460=>"PROJCS[\"Ocotepeque_1935_El_Salvador_Lambert\",GEOGCS[\"GCS_Ocotepeque_1935\",DATUM[\"D_Ocotepeque_1935\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",295809.184],PARAMETER[\"Central_Meridian\",-89.0],PARAMETER[\"Standard_Parallel_1\",13.78333333333333],PARAMETER[\"Scale_Factor\",0.99996704],PARAMETER[\"Latitude_Of_Origin\",13.78333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5460]]",
        5461=>"PROJCS[\"Ocotepeque_1935_Nicaragua_Norte\",GEOGCS[\"GCS_Ocotepeque_1935\",DATUM[\"D_Ocotepeque_1935\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",359891.816],PARAMETER[\"Central_Meridian\",-85.5],PARAMETER[\"Standard_Parallel_1\",13.86666666666667],PARAMETER[\"Scale_Factor\",0.99990314],PARAMETER[\"Latitude_Of_Origin\",13.86666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5461]]",
        5462=>"PROJCS[\"Ocotepeque_1935_Nicaragua_Sur\",GEOGCS[\"GCS_Ocotepeque_1935\",DATUM[\"D_Ocotepeque_1935\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",288876.327],PARAMETER[\"Central_Meridian\",-85.5],PARAMETER[\"Standard_Parallel_1\",11.73333333333333],PARAMETER[\"Scale_Factor\",0.99992228],PARAMETER[\"Latitude_Of_Origin\",11.73333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5462]]",
        5463=>"PROJCS[\"SAD_1969_UTM_Zone_17N\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5463]]",
        5469=>"PROJCS[\"Panama-Colon_1911_Panama_Lambert\",GEOGCS[\"GCS_Panama-Colon_1911\",DATUM[\"D_Panama-Colon-1911\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",294865.303],PARAMETER[\"Central_Meridian\",-80.0],PARAMETER[\"Standard_Parallel_1\",8.416666666666666],PARAMETER[\"Scale_Factor\",0.99989909],PARAMETER[\"Latitude_Of_Origin\",8.416666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5469]]",
        5472=>"PROJCS[\"Panama-Colon_1911_Panama_Polyconic\",GEOGCS[\"GCS_Panama-Colon_1911\",DATUM[\"D_Panama-Colon-1911\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Polyconic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1092972.1],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Latitude_Of_Origin\",8.25],UNIT[\"Yard_Clarke\",0.9143917962],AUTHORITY[\"EPSG\",5472]]",
        5479=>"PROJCS[\"RSRGD2000_MSLC2000\",GEOGCS[\"GCS_RSRGD2000\",DATUM[\"D_Ross_Sea_Region_Geodetic_Datum_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",7000000.0],PARAMETER[\"False_Northing\",5000000.0],PARAMETER[\"Central_Meridian\",163.0],PARAMETER[\"Standard_Parallel_1\",-76.66666666666667],PARAMETER[\"Standard_Parallel_2\",-79.33333333333333],PARAMETER[\"Latitude_Of_Origin\",-78.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5479]]",
        5480=>"PROJCS[\"RSRGD2000_BCLC2000\",GEOGCS[\"GCS_RSRGD2000\",DATUM[\"D_Ross_Sea_Region_Geodetic_Datum_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",5000000.0],PARAMETER[\"False_Northing\",3000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Standard_Parallel_1\",-73.66666666666667],PARAMETER[\"Standard_Parallel_2\",-75.33333333333333],PARAMETER[\"Latitude_Of_Origin\",-74.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5480]]",
        5481=>"PROJCS[\"RSRGD2000_PCLC2000\",GEOGCS[\"GCS_RSRGD2000\",DATUM[\"D_Ross_Sea_Region_Geodetic_Datum_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",166.0],PARAMETER[\"Standard_Parallel_1\",-70.66666666666667],PARAMETER[\"Standard_Parallel_2\",-72.33333333333333],PARAMETER[\"Latitude_Of_Origin\",-71.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5481]]",
        5482=>"PROJCS[\"RSRGD2000_RSPS2000\",GEOGCS[\"GCS_RSRGD2000\",DATUM[\"D_Ross_Sea_Region_Geodetic_Datum_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic\"],PARAMETER[\"False_Easting\",5000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",180.0],PARAMETER[\"Scale_Factor\",0.994],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5482]]",
        5490=>"PROJCS[\"RGAF09_UTM_Zone_20N\",GEOGCS[\"GCS_RGAF09\",DATUM[\"Reseau_Geodesique_des_Antilles_Francaises_2009\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5490]]",
        5513=>"PROJCS[\"S-JTSK_Krovak\",GEOGCS[\"GCS_S_JTSK\",DATUM[\"D_S_JTSK\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Krovak\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Pseudo_Standard_Parallel_1\",78.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Azimuth\",30.28813975277778],PARAMETER[\"Longitude_Of_Center\",24.83333333333333],PARAMETER[\"Latitude_Of_Center\",49.5],PARAMETER[\"X_Scale\",1.0],PARAMETER[\"Y_Scale\",1.0],PARAMETER[\"XY_Plane_Rotation\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5513]]",
        5514=>"PROJCS[\"S-JTSK_Krovak_East_North\",GEOGCS[\"GCS_S_JTSK\",DATUM[\"D_S_JTSK\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Krovak\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Pseudo_Standard_Parallel_1\",78.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Azimuth\",30.28813975277778],PARAMETER[\"Longitude_Of_Center\",24.83333333333333],PARAMETER[\"Latitude_Of_Center\",49.5],PARAMETER[\"X_Scale\",-1.0],PARAMETER[\"Y_Scale\",1.0],PARAMETER[\"XY_Plane_Rotation\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5514]]",
        5518=>"PROJCS[\"Chatham_Island_1971_Map_Grid\",GEOGCS[\"GCS_Chatham_Island_1971\",DATUM[\"D_Chatham_Island_1971\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",350000.0],PARAMETER[\"False_Northing\",650000.0],PARAMETER[\"Central_Meridian\",-176.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5518]]",
        5519=>"PROJCS[\"Chatham_Islands_1979_Map_Grid\",GEOGCS[\"GCS_Chatham_Islands_1979\",DATUM[\"D_Chatham_Islands_1979\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",350000.0],PARAMETER[\"False_Northing\",650000.0],PARAMETER[\"Central_Meridian\",-176.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5519]]",
        5520=>"PROJCS[\"DHDN_3_Degree_Gauss_Zone_1\",GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5520]]",
        5523=>"PROJCS[\"WGS_1984_UTM_Gabon_TM_2011\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",11.5],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5523]]",
        5530=>"PROJCS[\"SAD_1969_96_Brazil_Polyconic\",GEOGCS[\"GCS_SAD_1969_96\",DATUM[\"D_South_American_Datum_1969_96\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Polyconic\"],PARAMETER[\"False_Easting\",5000000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-54.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5530]]",
        5531=>"PROJCS[\"SAD_1969_96_UTM_Zone_21S\",GEOGCS[\"GCS_SAD_1969_96\",DATUM[\"D_South_American_Datum_1969_96\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5531]]",
        5532=>"PROJCS[\"SAD_1969_96_UTM_Zone_22S\",GEOGCS[\"GCS_SAD_1969_96\",DATUM[\"D_South_American_Datum_1969_96\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5858]]",
        5533=>"PROJCS[\"SAD_1969_96_UTM_Zone_23S\",GEOGCS[\"GCS_SAD_1969_96\",DATUM[\"D_South_American_Datum_1969_96\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5533]]",
        5534=>"PROJCS[\"SAD_1969_96_UTM_Zone_24S\",GEOGCS[\"GCS_SAD_1969_96\",DATUM[\"D_South_American_Datum_1969_96\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5534]]",
        5535=>"PROJCS[\"SAD_1969_96_UTM_Zone_25S\",GEOGCS[\"GCS_SAD_1969_96\",DATUM[\"D_South_American_Datum_1969_96\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5535]]",
        5536=>"PROJCS[\"Corrego_Alegre_1961_UTM_Zone_21S\",GEOGCS[\"GCS_Corrego_Alegre_1961\",DATUM[\"D_Corrego_Alegre_1961\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5536]]",
        5537=>"PROJCS[\"Corrego_Alegre_1961_UTM_Zone_22S\",GEOGCS[\"GCS_Corrego_Alegre_1961\",DATUM[\"D_Corrego_Alegre_1961\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5537]]",
        5538=>"PROJCS[\"Corrego_Alegre_1961_UTM_Zone_23S\",GEOGCS[\"GCS_Corrego_Alegre_1961\",DATUM[\"D_Corrego_Alegre_1961\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5538]]",
        5539=>"PROJCS[\"Corrego_Alegre_1961_UTM_Zone_24S\",GEOGCS[\"GCS_Corrego_Alegre_1961\",DATUM[\"D_Corrego_Alegre_1961\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5539]]",
        5550=>"PROJCS[\"PNG94_PNGMG94_Zone_54\",GEOGCS[\"GCS_PNG94\",DATUM[\"D_Papua_New_Guinea_Geodetic_Datum_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5550]]",
        5551=>"PROJCS[\"PNG94_PNGMG94_Zone_55\",GEOGCS[\"GCS_PNG94\",DATUM[\"D_Papua_New_Guinea_Geodetic_Datum_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5551]]",
        5552=>"PROJCS[\"PNG94_PNGMG94_Zone_56\",GEOGCS[\"GCS_PNG94\",DATUM[\"D_Papua_New_Guinea_Geodetic_Datum_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5552]]",
        5559=>"PROJCS[\"Ocotepeque_1935_Guatemala_Norte\",GEOGCS[\"GCS_Ocotepeque_1935\",DATUM[\"D_Ocotepeque_1935\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",292209.579],PARAMETER[\"Central_Meridian\",-90.33333333333333],PARAMETER[\"Standard_Parallel_1\",16.81666666666667],PARAMETER[\"Scale_Factor\",0.99992226],PARAMETER[\"Latitude_Of_Origin\",16.81666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5559]]",
        5562=>"PROJCS[\"Ukraine_2000_GK_Zone_4\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5562]]",
        5563=>"PROJCS[\"Ukraine_2000_GK_Zone_5\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5563]]",
        5564=>"PROJCS[\"Ukraine_2000_GK_Zone_6\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",6500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5564]]",
        5565=>"PROJCS[\"Ukraine_2000_GK_Zone_7\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",7500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5565]]",
        5566=>"PROJCS[\"Ukraine_2000_GK_CM_21E\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5566]]",
        5567=>"PROJCS[\"Ukraine_2000_GK_CM_27E\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5567]]",
        5568=>"PROJCS[\"Ukraine_2000_GK_CM_33E\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5568]]",
        5569=>"PROJCS[\"Ukraine_2000_GK_CM_39E\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5569]]",
        5570=>"PROJCS[\"Ukraine_2000_3_Degree_GK_Zone_7\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",7500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5570]]",
        5571=>"PROJCS[\"Ukraine_2000_3_Degree_GK_Zone_8\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",8500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5571]]",
        5572=>"PROJCS[\"Ukraine_2000_3_Degree_GK_Zone_9\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",9500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5572]]",
        5573=>"PROJCS[\"Ukraine_2000_3_Degree_GK_Zone_10\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",10500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5573]]",
        5574=>"PROJCS[\"Ukraine_2000_3_Degree_GK_Zone_11\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",11500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5574]]",
        5575=>"PROJCS[\"Ukraine_2000_3_Degree_GK_Zone_12\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",12500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",36.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5575]]",
        5576=>"PROJCS[\"Ukraine_2000_3_Degree_GK_Zone_13\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",13500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5576]]",
        5577=>"PROJCS[\"Ukraine_2000_3_Degree_GK_CM_21E\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5577]]",
        5578=>"PROJCS[\"Ukraine_2000_3_Degree_GK_CM_24E\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5578]]",
        5579=>"PROJCS[\"Ukraine_2000_3_Degree_GK_CM_27E\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5579]]",
        5580=>"PROJCS[\"Ukraine_2000_3_Degree_GK_CM_30E\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5580]]",
        5581=>"PROJCS[\"Ukraine_2000_3_Degree_GK_CM_33E\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5581]]",
        5582=>"PROJCS[\"Ukraine_2000_3_Degree_GK_CM_36E\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",36.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5582]]",
        5583=>"PROJCS[\"Ukraine_2000_3_Degree_GK_CM_39E\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5583]]",
        5588=>"PROJCS[\"NAD_1927_New_Brunswick_Stereographic\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-66.5],PARAMETER[\"Scale_Factor\",0.999912],PARAMETER[\"Latitude_Of_Origin\",46.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",5588]]",
        5589=>"PROJCS[\"Sibun_Gorge_1922_Colony_Grid\",GEOGCS[\"GCS_Sibun_Gorge_1922\",DATUM[\"D_Sibun_Gorge_1922\",SPHEROID[\"Clarke_1858\",6378293.645208759,294.260676369]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",217259.26],PARAMETER[\"False_Northing\",445474.83],PARAMETER[\"Central_Meridian\",-88.6318575],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",17.06124194444444],UNIT[\"Foot_Clarke\",0.3047972654],AUTHORITY[\"EPSG\",5589]]",
        5596=>"PROJCS[\"FEH2010_Fehmarnbelt_TM\",GEOGCS[\"GCS_FEH2010\",DATUM[\"D_Fehmarnbelt_Datum_2010\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",11.33333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5596]]",
        5623=>"PROJCS[\"NAD_1927_StatePlane_Michigan_East_Old_FIPS_2101\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-83.66666666666667],PARAMETER[\"Scale_Factor\",0.9999428571],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",5623]]",
        5624=>"PROJCS[\"NAD_1927_StatePlane_Michigan_Central_Old_FIPS_2102\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Scale_Factor\",0.9999090909],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",5624]]",
        5625=>"PROJCS[\"NAD_1927_StatePlane_Michigan_West_Old_FIPS_2103\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.75],PARAMETER[\"Scale_Factor\",0.9999090909],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",5625]]",
        5627=>"PROJCS[\"ED_1950_TM_6_NE\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",6.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5627]]",
        5629=>"PROJCS[\"Moznet_UTM_Zone_38S\",GEOGCS[\"GCS_Moznet\",DATUM[\"D_Moznet\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5629]]",
        5631=>"PROJCS[\"Pulkovo_1942_Adj_1958_GK_Zone_2_E-N\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5631]]",
        5632=>"PROJCS[\"PTRA08_LCC_Europe\",GEOGCS[\"GCS_PTRA08\",DATUM[\"D_PTRA08\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4000000.0],PARAMETER[\"False_Northing\",2800000.0],PARAMETER[\"Central_Meridian\",10.0],PARAMETER[\"Standard_Parallel_1\",35.0],PARAMETER[\"Standard_Parallel_2\",65.0],PARAMETER[\"Latitude_Of_Origin\",52.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5632]]",
        5633=>"PROJCS[\"PTRA08_LAEA_Europe\",GEOGCS[\"GCS_PTRA08\",DATUM[\"D_PTRA08\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",4321000.0],PARAMETER[\"False_Northing\",3210000.0],PARAMETER[\"Central_Meridian\",10.0],PARAMETER[\"Latitude_Of_Origin\",52.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5633]]",
        5634=>"PROJCS[\"REGCAN95_LCC_Europe\",GEOGCS[\"GCS_REGCAN95\",DATUM[\"D_Red_Geodesica_de_Canarias_1995\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4000000.0],PARAMETER[\"False_Northing\",2800000.0],PARAMETER[\"Central_Meridian\",10.0],PARAMETER[\"Standard_Parallel_1\",35.0],PARAMETER[\"Standard_Parallel_2\",65.0],PARAMETER[\"Latitude_Of_Origin\",52.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5634]]",
        5635=>"PROJCS[\"REGCAN95_LAEA_Europe\",GEOGCS[\"GCS_REGCAN95\",DATUM[\"D_Red_Geodesica_de_Canarias_1995\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",4321000.0],PARAMETER[\"False_Northing\",3210000.0],PARAMETER[\"Central_Meridian\",10.0],PARAMETER[\"Latitude_Of_Origin\",52.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5635]]",
        5636=>"PROJCS[\"TUREF_LAEA_Europe\",GEOGCS[\"GCS_TUREF\",DATUM[\"D_Turkish_National_Reference_Frame\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",4321000.0],PARAMETER[\"False_Northing\",3210000.0],PARAMETER[\"Central_Meridian\",10.0],PARAMETER[\"Latitude_Of_Origin\",52.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5636]]",
        5637=>"PROJCS[\"TUREF_LCC_Europe\",GEOGCS[\"GCS_TUREF\",DATUM[\"D_Turkish_National_Reference_Frame\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4000000.0],PARAMETER[\"False_Northing\",2800000.0],PARAMETER[\"Central_Meridian\",10.0],PARAMETER[\"Standard_Parallel_1\",35.0],PARAMETER[\"Standard_Parallel_2\",65.0],PARAMETER[\"Latitude_Of_Origin\",52.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5637]]",
        5638=>"PROJCS[\"ISN_2004_LAEA_Europe\",GEOGCS[\"GCS_ISN_2004\",DATUM[\"D_Islands_Network_2004\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",4321000.0],PARAMETER[\"False_Northing\",3210000.0],PARAMETER[\"Central_Meridian\",10.0],PARAMETER[\"Latitude_Of_Origin\",52.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5638]]",
        5639=>"PROJCS[\"ISN_2004_LCC_Europe\",GEOGCS[\"GCS_ISN_2004\",DATUM[\"D_Islands_Network_2004\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4000000.0],PARAMETER[\"False_Northing\",2800000.0],PARAMETER[\"Central_Meridian\",10.0],PARAMETER[\"Standard_Parallel_1\",35.0],PARAMETER[\"Standard_Parallel_2\",65.0],PARAMETER[\"Latitude_Of_Origin\",52.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5639]]",
        5641=>"PROJCS[\"SIRGAS_2000_Brazil_Mercator\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Mercator\"],PARAMETER[\"False_Easting\",5000000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-43.0],PARAMETER[\"Standard_Parallel_1\",-2.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5641]]",
        5643=>"PROJCS[\"ED_1950_Southern_Permian_Basin_Lambert\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",815000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",10.0],PARAMETER[\"Standard_Parallel_1\",52.66666666666666],PARAMETER[\"Standard_Parallel_2\",54.33333333333334],PARAMETER[\"Latitude_Of_Origin\",48.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5643]]",
        5644=>"PROJCS[\"RGR_1992_UTM_39S\",GEOGCS[\"GCS_RGR_1992\",DATUM[\"D_RGR_1992\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5644]]",
        5646=>"PROJCS[\"NAD_1983_StatePlane_Vermont_FIPS_4400_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-72.5],PARAMETER[\"Scale_Factor\",0.9999642857142857],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",5646]]",
        5649=>"PROJCS[\"ETRS_1989_UTM_Zone_31N_zE-N\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",31500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5649]]",
        5650=>"PROJCS[\"ETRS_1989_UTM_Zone_33N_zE-N\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",33500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5650]]",
        5651=>"PROJCS[\"ETRS_1989_UTM_Zone_31N_N-zE\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",31500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5651]]",
        5652=>"PROJCS[\"ETRS_1989_UTM_Zone_32N_N-zE\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",32500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5652]]",
        5653=>"PROJCS[\"ETRS_1989_UTM_Zone_33N_N-zE\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",33500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5653]]",
        5654=>"PROJCS[\"NAD_1983_HARN_StatePlane_Vermont_FIPS_4400_Ft_US\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-72.5],PARAMETER[\"Scale_Factor\",0.9999642857142857],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",5654]]",
        5655=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Vermont_FIPS_4400_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-72.5],PARAMETER[\"Scale_Factor\",0.9999642857142857],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",5655]]",
        5659=>"PROJCS[\"Monte_Mario_TM_Emilia-Romagna\",GEOGCS[\"GCS_Monte_Mario\",DATUM[\"D_Monte_Mario\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500053.0],PARAMETER[\"False_Northing\",-3999820.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5659]]",
        5663=>"PROJCS[\"Pulkovo_1942_Adj_1958_GK_Zone_3_E-N\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5663]]",
        5664=>"PROJCS[\"Pulkovo_1942_Adj_1983_GK_Zone_2_E-N\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5664]]",
        5665=>"PROJCS[\"Pulkovo_1942_Adj_1983_GK_Zone_3_E-N\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5665]]",
        5666=>"PROJCS[\"PD/83_3_Degree_GK_Zone_3_E-N\",GEOGCS[\"GCS_PD/83\",DATUM[\"D_Potsdam_1983\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5666]]",
        5667=>"PROJCS[\"PD/83_3_Degree_GK_Zone_4_E-N\",GEOGCS[\"GCS_PD/83\",DATUM[\"D_Potsdam_1983\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5667]]",
        5668=>"PROJCS[\"RD/83_3_Degree_GK_Zone_4_E-N\",GEOGCS[\"GCS_RD/83\",DATUM[\"D_Rauenberg_1983\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5668]]",
        5669=>"PROJCS[\"RD/83_3_Degree_GK_Zone_5_E-N\",GEOGCS[\"GCS_RD/83\",DATUM[\"D_Rauenberg_1983\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5669]]",
        5670=>"PROJCS[\"Pulkovo_1942_Adj_1958_3_Degree_GK_Zone_3_E-N\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5670]]",
        5671=>"PROJCS[\"Pulkovo_1942_Adj_1958_3_Degree_GK_Zone_4_E-N\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5671]]",
        5672=>"PROJCS[\"Pulkovo_1942_Adj_1958_3_Degree_GK_Zone_5_E-N\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1958\",DATUM[\"D_Pulkovo_1942_Adj_1958\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5672]]",
        5673=>"PROJCS[\"Pulkovo_1942_Adj_1983_3_Degree_GK_Zone_3_E-N\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5673]]",
        5674=>"PROJCS[\"Pulkovo_1942_Adj_1983_3_Degree_GK_Zone_4_E-N\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5674]]",
        5675=>"PROJCS[\"Pulkovo_1942_Adj_1983_3_Degree_GK_Zone_5_E-N\",GEOGCS[\"GCS_Pulkovo_1942_Adj_1983\",DATUM[\"D_Pulkovo_1942_Adj_1983\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5675]]",
        5676=>"PROJCS[\"DHDN_3_Degree_GK_Zone_2_E-N\",GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",6.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5676]]",
        5677=>"PROJCS[\"DHDN_3_Degree_GK_Zone_3_E-N\",GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5677]]",
        5678=>"PROJCS[\"DHDN_3_Degree_GK_Zone_4_E-N\",GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5678]]",
        5679=>"PROJCS[\"DHDN_3_Degree_GK_Zone_5_E-N\",GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5679]]",
        5680=>"PROJCS[\"DHDN_3_Degree_GK_Zone_1_E-N\",GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5680]]",
        5682=>"PROJCS[\"DB_REF_3-Degree_GK_Zone_2_(E-N)\",GEOGCS[\"GCS_DB_REF\",DATUM[\"D_Deutsche_Bahn_Reference_System\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",6.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5682]]",
        5683=>"PROJCS[\"DB_REF_3-Degree_GK_Zone_3_(E-N)\",GEOGCS[\"GCS_DB_REF\",DATUM[\"D_Deutsche_Bahn_Reference_System\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5683]]",
        5684=>"PROJCS[\"DB_REF_3-Degree_GK_Zone_4_(E-N)\",GEOGCS[\"GCS_DB_REF\",DATUM[\"D_Deutsche_Bahn_Reference_System\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5684]]",
        5685=>"PROJCS[\"DB_REF_3-Degree_GK_Zone_5_(E-N)\",GEOGCS[\"GCS_DB_REF\",DATUM[\"D_Deutsche_Bahn_Reference_System\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5685]]",
        5700=>"PROJCS[\"NZGD_2000_UTM_Zone_1S\",GEOGCS[\"GCS_NZGD_2000\",DATUM[\"D_NZGD_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5700]]",
        5825=>"PROJCS[\"AGD_1966_ACT_Standard_Grid\",GEOGCS[\"GCS_Australian_1966\",DATUM[\"D_Australian_1966\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",600000.0],PARAMETER[\"Central_Meridian\",149.0092948305555],PARAMETER[\"Scale_Factor\",1.000086],PARAMETER[\"Latitude_Of_Origin\",-35.31773627777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5825]]",
        5836=>"PROJCS[\"Yemen_NGN_1996_UTM_Zone_37N\",GEOGCS[\"GCS_Yemen_NGN_1996\",DATUM[\"D_Yemen_NGN_1996\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5836]]",
        5837=>"PROJCS[\"Yemen_NGN_1996_UTM_Zone_40N\",GEOGCS[\"GCS_Yemen_NGN_1996\",DATUM[\"D_Yemen_NGN_1996\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5837]]",
        5839=>"PROJCS[\"Peru96_UTM_Zone_17S\",GEOGCS[\"GCS_Peru96\",DATUM[\"D_Peru96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5839]]",
        5842=>"PROJCS[\"WGS_1984_TM_12_SE\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5842]]",
        5844=>"PROJCS[\"RGRDC_2005_Congo_TM_Zone_30\",GEOGCS[\"GCS_RGRDC_2005\",DATUM[\"D_Reseau_Geodesique_de_la_RDC_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5844]]",
        5858=>"PROJCS[\"SAD_1969_96_UTM_Zone_22S\",GEOGCS[\"GCS_SAD_1969_96\",DATUM[\"D_South_American_Datum_1969_96\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5858]]",
        5875=>"PROJCS[\"SAD_1969_96_UTM_Zone_18S\",GEOGCS[\"GCS_SAD_1969_96\",DATUM[\"D_South_American_Datum_1969_96\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5875]]",
        5876=>"PROJCS[\"SAD_1969_96_UTM_Zone_19S\",GEOGCS[\"GCS_SAD_1969_96\",DATUM[\"D_South_American_Datum_1969_96\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5876]]",
        5877=>"PROJCS[\"SAD_1969_96_UTM_Zone_20S\",GEOGCS[\"GCS_SAD_1969_96\",DATUM[\"D_South_American_Datum_1969_96\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5877]]",
        5879=>"PROJCS[\"Cadastre_1997_UTM_Zone_38S\",GEOGCS[\"GCS_Cadastre_1997\",DATUM[\"D_Cadastre_1997\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5879]]",
        5880=>"PROJCS[\"SIRGAS_2000_Brazil_Polyconic\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Polyconic\"],PARAMETER[\"False_Easting\",5000000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-54.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5880]]",
        5887=>"PROJCS[\"TGD2005_Tonga_Map_Grid\",GEOGCS[\"GCS_TGD2005\",DATUM[\"D_Tonga_Geodetic_Datum_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",5000000.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5887]]",
        5890=>"PROJCS[\"JAXA_Snow_Depth_Polar_Stereographic_North\",GEOGCS[\"GCS_Hughes_1980\",DATUM[\"D_Hughes_1980\",SPHEROID[\"Hughes_1980\",6378273.0,298.279411123064]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic_North_Pole\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Standard_Parallel_1\",70.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5890]]",
        5921=>"PROJCS[\"WGS_1984_EPSG_Arctic_Regional_zone_A1\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",85.0],PARAMETER[\"Latitude_Of_Origin\",81.317226],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5921]]",
        5922=>"PROJCS[\"WGS_1984_EPSG_Arctic_Regional_zone_A2\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",85.0],PARAMETER[\"Latitude_Of_Origin\",81.317226],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5922]]",
        5923=>"PROJCS[\"WGS_1984_EPSG_Arctic_Regional_zone_A3\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",85.0],PARAMETER[\"Latitude_Of_Origin\",81.317226],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5923]]",
        5924=>"PROJCS[\"WGS_1984_EPSG_Arctic_Regional_zone_A4\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",85.0],PARAMETER[\"Latitude_Of_Origin\",81.317226],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5924]]",
        5925=>"PROJCS[\"WGS_1984_EPSG_Arctic_Regional_zone_A5\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",85.0],PARAMETER[\"Latitude_Of_Origin\",81.317226],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5925]]",
        5926=>"PROJCS[\"WGS_1984_EPSG_Arctic_Regional_zone_B1\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Standard_Parallel_1\",69.0],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",73.15574086111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5926]]",
        5927=>"PROJCS[\"WGS_1984_EPSG_Arctic_Regional_zone_B2\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Standard_Parallel_1\",69.0],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",73.15574086111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5927]]",
        5928=>"PROJCS[\"WGS_1984_EPSG_Arctic_Regional_zone_B3\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Standard_Parallel_1\",69.0],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",73.15574086111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5928]]",
        5929=>"PROJCS[\"WGS_1984_EPSG_Arctic_Regional_zone_B4\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Standard_Parallel_1\",69.0],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",73.15574086111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5929]]",
        5930=>"PROJCS[\"WGS_1984_EPSG_Arctic_Regional_zone_B5\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Standard_Parallel_1\",69.0],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",73.15574086111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5930]]",
        5931=>"PROJCS[\"WGS_1984_EPSG_Arctic_Regional_zone_C1\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Standard_Parallel_1\",61.0],PARAMETER[\"Standard_Parallel_2\",69.0],PARAMETER[\"Latitude_Of_Origin\",65.10127088888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5931]]",
        5932=>"PROJCS[\"WGS_1984_EPSG_Arctic_Regional_zone_C2\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Standard_Parallel_1\",61.0],PARAMETER[\"Standard_Parallel_2\",69.0],PARAMETER[\"Latitude_Of_Origin\",65.10127088888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5932]]",
        5933=>"PROJCS[\"WGS_1984_EPSG_Arctic_Regional_zone_C3\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Standard_Parallel_1\",61.0],PARAMETER[\"Standard_Parallel_2\",69.0],PARAMETER[\"Latitude_Of_Origin\",65.10127088888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5933]]",
        5934=>"PROJCS[\"WGS_1984_EPSG_Arctic_Regional_zone_C4\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Standard_Parallel_1\",61.0],PARAMETER[\"Standard_Parallel_2\",69.0],PARAMETER[\"Latitude_Of_Origin\",65.10127088888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5934]]",
        5935=>"PROJCS[\"WGS_1984_EPSG_Arctic_Regional_zone_C5\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Standard_Parallel_1\",61.0],PARAMETER[\"Standard_Parallel_2\",69.0],PARAMETER[\"Latitude_Of_Origin\",65.10127088888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5935]]",
        5936=>"PROJCS[\"WGS_1984_EPSG_Alaska_Polar_Stereographic\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",-150.0],PARAMETER[\"Scale_Factor\",0.994],PARAMETER[\"Latitude_Of_Origin\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5936]]",
        5937=>"PROJCS[\"WGS_1984_EPSG_Canada_Polar_Stereographic\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Scale_Factor\",0.994],PARAMETER[\"Latitude_Of_Origin\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5937]]",
        5938=>"PROJCS[\"WGS_1984_EPSG_Greenland_Polar_Stereographic\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",-33.0],PARAMETER[\"Scale_Factor\",0.994],PARAMETER[\"Latitude_Of_Origin\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5938]]",
        5939=>"PROJCS[\"WGS_1984_EPSG_Norway_Polar_Stereographic\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",18.0],PARAMETER[\"Scale_Factor\",0.994],PARAMETER[\"Latitude_Of_Origin\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5939]]",
        5940=>"PROJCS[\"WGS_1984_EPSG_Russia_Polar_Stereographic\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.994],PARAMETER[\"Latitude_Of_Origin\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5940]]",
        6050=>"PROJCS[\"GR96_EPSG_Arctic_zone_1-25\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",25500000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",-30.0],PARAMETER[\"Standard_Parallel_1\",83.66666666666667],PARAMETER[\"Standard_Parallel_2\",87.0],PARAMETER[\"Latitude_Of_Origin\",85.43711833333332],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6050]]",
        6051=>"PROJCS[\"GR96_EPSG_Arctic_zone_2-18\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",18500000.0],PARAMETER[\"False_Northing\",2500000.0],PARAMETER[\"Central_Meridian\",-52.0],PARAMETER[\"Standard_Parallel_1\",80.33333333333333],PARAMETER[\"Standard_Parallel_2\",83.66666666666667],PARAMETER[\"Latitude_Of_Origin\",82.05842488888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6051]]",
        6052=>"PROJCS[\"GR96_EPSG_Arctic_zone_2-20\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",20500000.0],PARAMETER[\"False_Northing\",2500000.0],PARAMETER[\"Central_Meridian\",-12.0],PARAMETER[\"Standard_Parallel_1\",80.33333333333333],PARAMETER[\"Standard_Parallel_2\",83.66666666666667],PARAMETER[\"Latitude_Of_Origin\",82.05842488888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6052]]",
        6053=>"PROJCS[\"GR96_EPSG_Arctic_zone_3-29\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",29500000.0],PARAMETER[\"False_Northing\",3500000.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",80.33333333333333],PARAMETER[\"Latitude_Of_Origin\",78.70733752777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6053]]",
        6054=>"PROJCS[\"GR96_EPSG_Arctic_zone_3-31\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",31500000.0],PARAMETER[\"False_Northing\",3500000.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",80.33333333333333],PARAMETER[\"Latitude_Of_Origin\",78.70733752777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6054]]",
        6055=>"PROJCS[\"GR96_EPSG_Arctic_zone_3-33\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",33500000.0],PARAMETER[\"False_Northing\",3500000.0],PARAMETER[\"Central_Meridian\",-10.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",80.33333333333333],PARAMETER[\"Latitude_Of_Origin\",78.70733752777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6055]]",
        6056=>"PROJCS[\"GR96_EPSG_Arctic_zone_4-20\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",20500000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",-64.0],PARAMETER[\"Standard_Parallel_1\",73.66666666666669],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",75.36440330555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6056]]",
        6057=>"PROJCS[\"GR96_EPSG_Arctic_zone_4-22\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",22500000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Standard_Parallel_1\",73.66666666666669],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",75.36440330555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6057]]",
        6058=>"PROJCS[\"GR96_EPSG_Arctic_zone_4-24\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",24500000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",-14.0],PARAMETER[\"Standard_Parallel_1\",73.66666666666669],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",75.36440330555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6058]]",
        6059=>"PROJCS[\"GR96_EPSG_Arctic_zone_5-41\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",41500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",-62.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6059]]",
        6060=>"PROJCS[\"GR96_EPSG_Arctic_zone_5-43\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",43500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",-42.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6060]]",
        6061=>"PROJCS[\"GR96_EPSG_Arctic_zone_5-45\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",45500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",-22.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6061]]",
        6062=>"PROJCS[\"GR96_EPSG_Arctic_zone_6-26\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",26500000.0],PARAMETER[\"False_Northing\",6500000.0],PARAMETER[\"Central_Meridian\",-56.0],PARAMETER[\"Standard_Parallel_1\",67.0],PARAMETER[\"Standard_Parallel_2\",70.33333333333333],PARAMETER[\"Latitude_Of_Origin\",68.68747555555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6062]]",
        6063=>"PROJCS[\"GR96_EPSG_Arctic_zone_6-28\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",28500000.0],PARAMETER[\"False_Northing\",6500000.0],PARAMETER[\"Central_Meridian\",-38.0],PARAMETER[\"Standard_Parallel_1\",67.0],PARAMETER[\"Standard_Parallel_2\",70.33333333333333],PARAMETER[\"Latitude_Of_Origin\",68.68747555555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6063]]",
        6064=>"PROJCS[\"GR96_EPSG_Arctic_zone_6-30\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",30500000.0],PARAMETER[\"False_Northing\",6500000.0],PARAMETER[\"Central_Meridian\",-20.0],PARAMETER[\"Standard_Parallel_1\",67.0],PARAMETER[\"Standard_Parallel_2\",70.33333333333333],PARAMETER[\"Latitude_Of_Origin\",68.68747555555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6064]]",
        6065=>"PROJCS[\"GR96_EPSG_Arctic_zone_7-11\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",11500000.0],PARAMETER[\"False_Northing\",7500000.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Standard_Parallel_1\",63.66666666666666],PARAMETER[\"Standard_Parallel_2\",67.0],PARAMETER[\"Latitude_Of_Origin\",65.35103930555557],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6065]]",
        6066=>"PROJCS[\"GR96_EPSG_Arctic_zone_7-13\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",13500000.0],PARAMETER[\"False_Northing\",7500000.0],PARAMETER[\"Central_Meridian\",-34.0],PARAMETER[\"Standard_Parallel_1\",63.66666666666666],PARAMETER[\"Standard_Parallel_2\",67.0],PARAMETER[\"Latitude_Of_Origin\",65.35103930555557],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6066]]",
        6067=>"PROJCS[\"GR96_EPSG_Arctic_zone_8-20\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",20500000.0],PARAMETER[\"False_Northing\",8500000.0],PARAMETER[\"Central_Meridian\",-52.0],PARAMETER[\"Standard_Parallel_1\",60.33333333333334],PARAMETER[\"Standard_Parallel_2\",63.66666666666666],PARAMETER[\"Latitude_Of_Origin\",62.01530688888887],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6067]]",
        6068=>"PROJCS[\"GR96_EPSG_Arctic_zone_8-22\",GEOGCS[\"GCS_Greenland_1996\",DATUM[\"D_Greenland_1996\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",22500000.0],PARAMETER[\"False_Northing\",8500000.0],PARAMETER[\"Central_Meridian\",-37.0],PARAMETER[\"Standard_Parallel_1\",60.33333333333334],PARAMETER[\"Standard_Parallel_2\",63.66666666666666],PARAMETER[\"Latitude_Of_Origin\",62.01530688888887],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6068]]",
        6069=>"PROJCS[\"ETRS_1989_EPSG_Arctic_zone_2-22\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",22500000.0],PARAMETER[\"False_Northing\",2500000.0],PARAMETER[\"Central_Meridian\",16.0],PARAMETER[\"Standard_Parallel_1\",80.33333333333333],PARAMETER[\"Standard_Parallel_2\",83.66666666666667],PARAMETER[\"Latitude_Of_Origin\",82.05842488888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6069]]",
        6070=>"PROJCS[\"ETRS_1989_EPSG_Arctic_zone_3-11\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",11500000.0],PARAMETER[\"False_Northing\",3500000.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",80.33333333333333],PARAMETER[\"Latitude_Of_Origin\",78.70733752777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6070]]",
        6071=>"PROJCS[\"ETRS_1989_EPSG_Arctic_zone_4-26\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",26500000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",10.0],PARAMETER[\"Standard_Parallel_1\",73.66666666666669],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",75.36440330555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6071]]",
        6072=>"PROJCS[\"ETRS_1989_EPSG_Arctic_zone_4-28\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",28500000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",34.0],PARAMETER[\"Standard_Parallel_1\",73.66666666666669],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",75.36440330555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6072]]",
        6073=>"PROJCS[\"ETRS_1989_EPSG_Arctic_zone_5-11\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",11500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",14.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6073]]",
        6074=>"PROJCS[\"ETRS_1989_EPSG_Arctic_zone_5-13\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",13500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",34.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6074]]",
        6075=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_2-24\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",24500000.0],PARAMETER[\"False_Northing\",2500000.0],PARAMETER[\"Central_Meridian\",53.0],PARAMETER[\"Standard_Parallel_1\",80.33333333333333],PARAMETER[\"Standard_Parallel_2\",83.66666666666667],PARAMETER[\"Latitude_Of_Origin\",82.05842488888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6075]]",
        6076=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_2-26\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",26500000.0],PARAMETER[\"False_Northing\",2500000.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Standard_Parallel_1\",80.33333333333333],PARAMETER[\"Standard_Parallel_2\",83.66666666666667],PARAMETER[\"Latitude_Of_Origin\",82.05842488888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6076]]",
        6077=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_3-13\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",13500000.0],PARAMETER[\"False_Northing\",3500000.0],PARAMETER[\"Central_Meridian\",52.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",80.33333333333333],PARAMETER[\"Latitude_Of_Origin\",78.70733752777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6077]]",
        6078=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_3-15\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",15500000.0],PARAMETER[\"False_Northing\",3500000.0],PARAMETER[\"Central_Meridian\",83.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",80.33333333333333],PARAMETER[\"Latitude_Of_Origin\",78.70733752777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6078]]",
        6079=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_3-17\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",17500000.0],PARAMETER[\"False_Northing\",3500000.0],PARAMETER[\"Central_Meridian\",114.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",80.33333333333333],PARAMETER[\"Latitude_Of_Origin\",78.70733752777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6079]]",
        6080=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_3-19\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",19500000.0],PARAMETER[\"False_Northing\",3500000.0],PARAMETER[\"Central_Meridian\",145.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",80.33333333333333],PARAMETER[\"Latitude_Of_Origin\",78.70733752777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6080]]",
        6081=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_4-30\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",30500000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",58.0],PARAMETER[\"Standard_Parallel_1\",73.66666666666669],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",75.36440330555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6081]]",
        6082=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_4-32\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",32500000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",82.0],PARAMETER[\"Standard_Parallel_1\",73.66666666666669],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",75.36440330555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6082]]",
        6083=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_4-34\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",34500000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",106.0],PARAMETER[\"Standard_Parallel_1\",73.66666666666669],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",75.36440330555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6083]]",
        6084=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_4-36\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",36500000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",130.0],PARAMETER[\"Standard_Parallel_1\",73.66666666666669],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",75.36440330555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6084]]",
        6085=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_4-38\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",38500000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",154.0],PARAMETER[\"Standard_Parallel_1\",73.66666666666669],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",75.36440330555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6085]]",
        6086=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_4-40\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",40500000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",179.0],PARAMETER[\"Standard_Parallel_1\",73.66666666666669],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",75.36440330555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6086]]",
        6087=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_5-15\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",15500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",54.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6087]]",
        6088=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_5-17\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",17500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",74.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6088]]",
        6089=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_5-19\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",19500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",95.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6089]]",
        6090=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_5-21\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",21500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",116.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6090]]",
        6091=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_5-23\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",23500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",137.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6091]]",
        6092=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_5-25\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",25500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",158.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6092]]",
        6093=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_5-27\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",27500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",179.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6093]]",
        6094=>"PROJCS[\"NAD_1983_NSRS2007_EPSG_Arctic_zone_5-29\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",29500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",-163.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6094]]",
        6095=>"PROJCS[\"NAD_1983_NSRS2007_EPSG_Arctic_zone_5-31\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",31500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6095]]",
        6096=>"PROJCS[\"NAD_1983_NSRS2007_EPSG_Arctic_zone_6-14\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",14500000.0],PARAMETER[\"False_Northing\",6500000.0],PARAMETER[\"Central_Meridian\",-165.0],PARAMETER[\"Standard_Parallel_1\",67.0],PARAMETER[\"Standard_Parallel_2\",70.33333333333333],PARAMETER[\"Latitude_Of_Origin\",68.68747555555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6096]]",
        6097=>"PROJCS[\"NAD_1983_NSRS2007_EPSG_Arctic_zone_6-16\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",16500000.0],PARAMETER[\"False_Northing\",6500000.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Standard_Parallel_1\",67.0],PARAMETER[\"Standard_Parallel_2\",70.33333333333333],PARAMETER[\"Latitude_Of_Origin\",68.68747555555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6097]]",
        6098=>"PROJCS[\"NAD_1983_CSRS_EPSG_Arctic_zone_1-23\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",23500000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",83.66666666666667],PARAMETER[\"Standard_Parallel_2\",87.0],PARAMETER[\"Latitude_Of_Origin\",85.43711833333332],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6098]]",
        6099=>"PROJCS[\"NAD_1983_CSRS_EPSG_Arctic_zone_2-14\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",14500000.0],PARAMETER[\"False_Northing\",2500000.0],PARAMETER[\"Central_Meridian\",-115.0],PARAMETER[\"Standard_Parallel_1\",80.33333333333333],PARAMETER[\"Standard_Parallel_2\",83.66666666666667],PARAMETER[\"Latitude_Of_Origin\",82.05842488888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6099]]",
        6100=>"PROJCS[\"NAD_1983_CSRS_EPSG_Arctic_zone_2-16\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",16500000.0],PARAMETER[\"False_Northing\",2500000.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Standard_Parallel_1\",80.33333333333333],PARAMETER[\"Standard_Parallel_2\",83.66666666666667],PARAMETER[\"Latitude_Of_Origin\",82.05842488888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6100]]",
        6101=>"PROJCS[\"NAD_1983_CSRS_EPSG_Arctic_zone_3-25\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",25500000.0],PARAMETER[\"False_Northing\",3500000.0],PARAMETER[\"Central_Meridian\",-129.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",80.33333333333333],PARAMETER[\"Latitude_Of_Origin\",78.70733752777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6101]]",
        6102=>"PROJCS[\"NAD_1983_CSRS_EPSG_Arctic_zone_3-27\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",27500000.0],PARAMETER[\"False_Northing\",3500000.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",80.33333333333333],PARAMETER[\"Latitude_Of_Origin\",78.70733752777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6102]]",
        6103=>"PROJCS[\"NAD_1983_CSRS_EPSG_Arctic_zone_3-29\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",29500000.0],PARAMETER[\"False_Northing\",3500000.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",80.33333333333333],PARAMETER[\"Latitude_Of_Origin\",78.70733752777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6103]]",
        6104=>"PROJCS[\"NAD_1983_CSRS_EPSG_Arctic_zone_4-14\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",14500000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",-129.0],PARAMETER[\"Standard_Parallel_1\",73.66666666666669],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",75.36440330555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6104]]",
        6105=>"PROJCS[\"NAD_1983_CSRS_EPSG_Arctic_zone_4-16\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",16500000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",-104.0],PARAMETER[\"Standard_Parallel_1\",73.66666666666669],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",75.36440330555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6105]]",
        6106=>"PROJCS[\"NAD_1983_CSRS_EPSG_Arctic_zone_4-18\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",18500000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",-79.0],PARAMETER[\"Standard_Parallel_1\",73.66666666666669],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",75.36440330555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6106]]",
        6107=>"PROJCS[\"NAD_1983_CSRS_EPSG_Arctic_zone_5-33\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",33500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",-131.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6107]]",
        6108=>"PROJCS[\"NAD_1983_CSRS_EPSG_Arctic_zone_5-35\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",35500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6108]]",
        6109=>"PROJCS[\"NAD_1983_CSRS_EPSG_Arctic_zone_5-37\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",37500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",-91.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6109]]",
        6110=>"PROJCS[\"NAD_1983_CSRS_EPSG_Arctic_zone_5-39\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",39500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",-71.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6110]]",
        6111=>"PROJCS[\"NAD_1983_CSRS_EPSG_Arctic_zone_6-18\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",18500000.0],PARAMETER[\"False_Northing\",6500000.0],PARAMETER[\"Central_Meridian\",-132.0],PARAMETER[\"Standard_Parallel_1\",67.0],PARAMETER[\"Standard_Parallel_2\",70.33333333333333],PARAMETER[\"Latitude_Of_Origin\",68.68747555555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6111]]",
        6112=>"PROJCS[\"NAD_1983_CSRS_EPSG_Arctic_zone_6-20\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",20500000.0],PARAMETER[\"False_Northing\",6500000.0],PARAMETER[\"Central_Meridian\",-113.0],PARAMETER[\"Standard_Parallel_1\",67.0],PARAMETER[\"Standard_Parallel_2\",70.33333333333333],PARAMETER[\"Latitude_Of_Origin\",68.68747555555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6112]]",
        6113=>"PROJCS[\"NAD_1983_CSRS_EPSG_Arctic_zone_6-22\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",22500000.0],PARAMETER[\"False_Northing\",6500000.0],PARAMETER[\"Central_Meridian\",-94.0],PARAMETER[\"Standard_Parallel_1\",67.0],PARAMETER[\"Standard_Parallel_2\",70.33333333333333],PARAMETER[\"Latitude_Of_Origin\",68.68747555555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6113]]",
        6114=>"PROJCS[\"NAD_1983_CSRS_EPSG_Arctic_zone_6-24\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",24500000.0],PARAMETER[\"False_Northing\",6500000.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Standard_Parallel_1\",67.0],PARAMETER[\"Standard_Parallel_2\",70.33333333333333],PARAMETER[\"Latitude_Of_Origin\",68.68747555555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6114]]",
        6115=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_1-27\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",27500000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Standard_Parallel_1\",83.66666666666667],PARAMETER[\"Standard_Parallel_2\",87.0],PARAMETER[\"Latitude_Of_Origin\",85.43711833333332],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6115]]",
        6116=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_1-29\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",29500000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Standard_Parallel_1\",83.66666666666667],PARAMETER[\"Standard_Parallel_2\",87.0],PARAMETER[\"Latitude_Of_Origin\",85.43711833333332],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6116]]",
        6117=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_1-31\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",31500000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",150.0],PARAMETER[\"Standard_Parallel_1\",83.66666666666667],PARAMETER[\"Standard_Parallel_2\",87.0],PARAMETER[\"Latitude_Of_Origin\",85.43711833333332],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6117]]",
        6118=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_1-21\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",21500000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",-150.0],PARAMETER[\"Standard_Parallel_1\",83.66666666666667],PARAMETER[\"Standard_Parallel_2\",87.0],PARAMETER[\"Latitude_Of_Origin\",85.43711833333332],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6118]]",
        6119=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_2-28\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",28500000.0],PARAMETER[\"False_Northing\",2500000.0],PARAMETER[\"Central_Meridian\",133.0],PARAMETER[\"Standard_Parallel_1\",80.33333333333333],PARAMETER[\"Standard_Parallel_2\",83.66666666666667],PARAMETER[\"Latitude_Of_Origin\",82.05842488888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6119]]",
        6120=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_2-10\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",10500000.0],PARAMETER[\"False_Northing\",2500000.0],PARAMETER[\"Central_Meridian\",166.0],PARAMETER[\"Standard_Parallel_1\",80.33333333333333],PARAMETER[\"Standard_Parallel_2\",83.66666666666667],PARAMETER[\"Latitude_Of_Origin\",82.05842488888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6120]]",
        6121=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_2-12\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",12500000.0],PARAMETER[\"False_Northing\",2500000.0],PARAMETER[\"Central_Meridian\",-154.0],PARAMETER[\"Standard_Parallel_1\",80.33333333333333],PARAMETER[\"Standard_Parallel_2\",83.66666666666667],PARAMETER[\"Latitude_Of_Origin\",82.05842488888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6121]]",
        6122=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_3-21\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",21500000.0],PARAMETER[\"False_Northing\",3500000.0],PARAMETER[\"Central_Meridian\",176.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",80.33333333333333],PARAMETER[\"Latitude_Of_Origin\",78.70733752777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6122]]",
        6123=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_3-23\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",23500000.0],PARAMETER[\"False_Northing\",3500000.0],PARAMETER[\"Central_Meridian\",-153.0],PARAMETER[\"Standard_Parallel_1\",77.0],PARAMETER[\"Standard_Parallel_2\",80.33333333333333],PARAMETER[\"Latitude_Of_Origin\",78.70733752777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6123]]",
        6124=>"PROJCS[\"WGS_1984_EPSG_Arctic_zone_4-12\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",12500000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",-155.0],PARAMETER[\"Standard_Parallel_1\",73.66666666666669],PARAMETER[\"Standard_Parallel_2\",77.0],PARAMETER[\"Latitude_Of_Origin\",75.36440330555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6124]]",
        6125=>"PROJCS[\"ETRS_1989_EPSG_Arctic_zone_5-47\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",47500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",-5.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6125]]",
        6128=>"PROJCS[\"Grand_Cayman_National_Grid_1959\",GEOGCS[\"GCS_Grand_Cayman_1959\",DATUM[\"D_Grand_Cayman_1959\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640419.9475],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6128]]",
        6129=>"PROJCS[\"Sister_Islands_National_Grid_1961\",GEOGCS[\"GCS_Little_Cayman_1961\",DATUM[\"D_Little_Cayman_1961\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640419.9475],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6129]]",
        6141=>"PROJCS[\"Cayman_Islands_National_Grid_2011\",GEOGCS[\"GCS_CIGD11\",DATUM[\"D_Cayman_Islands_Geodetic_Datum_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2950000.0],PARAMETER[\"False_Northing\",1900000.0],PARAMETER[\"Central_Meridian\",-80.56666666666666],PARAMETER[\"Standard_Parallel_1\",19.33333333333333],PARAMETER[\"Standard_Parallel_2\",19.7],PARAMETER[\"Latitude_Of_Origin\",19.33333333333333],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6391]]",
        6204=>"PROJCS[\"Macedonian_State_Coordinate_System\",GEOGCS[\"GCS_MGI_1901\",DATUM[\"D_MGI_1901\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6204]]",
        6210=>"PROJCS[\"SIRGAS_2000_UTM_Zone_23N\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6210]]",
        6211=>"PROJCS[\"SIRGAS_2000_UTM_Zone_24N\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6211]]",
        6244=>"PROJCS[\"MAGNA_Arauca_2007\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1035263.443],PARAMETER[\"False_Northing\",1275526.621],PARAMETER[\"Longitude_Of_Center\",-70.75830965555555],PARAMETER[\"Latitude_Of_Center\",7.087606391666666],PARAMETER[\"Height\",100.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6244]]",
        6245=>"PROJCS[\"MAGNA_Armenia_Quindio_2006\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1155824.666],PARAMETER[\"False_Northing\",993087.465],PARAMETER[\"Longitude_Of_Center\",-75.67348916666667],PARAMETER[\"Latitude_Of_Center\",4.532325],PARAMETER[\"Height\",1470.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6245]]",
        6246=>"PROJCS[\"MAGNA_Barranquilla_Atlantico_1997\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",917264.406],PARAMETER[\"False_Northing\",1699839.935],PARAMETER[\"Longitude_Of_Center\",-74.83433133333332],PARAMETER[\"Latitude_Of_Center\",10.92318308333333],PARAMETER[\"Height\",100.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6246]]",
        6247=>"PROJCS[\"MAGNA_Bogota_DC_2005\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",92334.879],PARAMETER[\"False_Northing\",109320.965],PARAMETER[\"Longitude_Of_Center\",-74.14659166666668],PARAMETER[\"Latitude_Of_Center\",4.680486111111112],PARAMETER[\"Height\",2550.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6247]]",
        6248=>"PROJCS[\"MAGNA_Bucaramanga_Santander_2008\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1097241.305],PARAMETER[\"False_Northing\",1274642.278],PARAMETER[\"Longitude_Of_Center\",-73.19734322222223],PARAMETER[\"Latitude_Of_Center\",7.078887141666667],PARAMETER[\"Height\",931.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6248]]",
        6249=>"PROJCS[\"MAGNA_Cali_Valle_del_Cauca_2009\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1061900.18],PARAMETER[\"False_Northing\",872364.63],PARAMETER[\"Longitude_Of_Center\",-76.5205625],PARAMETER[\"Latitude_Of_Center\",3.441883333333334],PARAMETER[\"Height\",1000.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6249]]",
        6250=>"PROJCS[\"MAGNA_Cartagena_Bolivar_2005\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",842981.41],PARAMETER[\"False_Northing\",1641887.09],PARAMETER[\"Longitude_Of_Center\",-75.51120694444444],PARAMETER[\"Latitude_Of_Center\",10.3970475],PARAMETER[\"Height\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6250]]",
        6251=>"PROJCS[\"MAGNA_Cucuta_Norte_de_Santander_2011\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",842805.406],PARAMETER[\"False_Northing\",1364404.57],PARAMETER[\"Longitude_Of_Center\",-72.50287095],PARAMETER[\"Latitude_Of_Center\",7.888936736111111],PARAMETER[\"Height\",308.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6251]]",
        6252=>"PROJCS[\"MAGNA_Florencia_Caqueta_2007\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1162300.348],PARAMETER[\"False_Northing\",671068.716],PARAMETER[\"Longitude_Of_Center\",-75.61911760277778],PARAMETER[\"Latitude_Of_Center\",1.621012294444445],PARAMETER[\"Height\",300.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6252]]",
        6253=>"PROJCS[\"MAGNA_Ibague_Tolima_2007\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",877634.33],PARAMETER[\"False_Northing\",980541.348],PARAMETER[\"Longitude_Of_Center\",-75.17992593333334],PARAMETER[\"Latitude_Of_Center\",4.419412827777778],PARAMETER[\"Height\",1100.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6253]]",
        6254=>"PROJCS[\"MAGNA_Inirida_Guainia_2008\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1019177.687],PARAMETER[\"False_Northing\",491791.326],PARAMETER[\"Longitude_Of_Center\",-67.90523208888889],PARAMETER[\"Latitude_Of_Center\",3.845438183333334],PARAMETER[\"Height\",96.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6254]]",
        6255=>"PROJCS[\"MAGNA_Leticia_Amazonas_1994\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",25978.217],PARAMETER[\"False_Northing\",27501.365],PARAMETER[\"Longitude_Of_Center\",-69.94281105833333],PARAMETER[\"Latitude_Of_Center\",-4.197684047222222],PARAMETER[\"Height\",89.7],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6255]]",
        6256=>"PROJCS[\"MAGNA_Manizales_Caldas_2011\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1173727.04],PARAMETER[\"False_Northing\",1052391.13],PARAMETER[\"Longitude_Of_Center\",-75.51109472222223],PARAMETER[\"Latitude_Of_Center\",5.068153888888888],PARAMETER[\"Height\",2100.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6256]]",
        6257=>"PROJCS[\"MAGNA_Medellin_Antioquia_2010\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",835378.647],PARAMETER[\"False_Northing\",1180816.875],PARAMETER[\"Longitude_Of_Center\",-75.56488694444444],PARAMETER[\"Latitude_Of_Center\",6.229208888888889],PARAMETER[\"Height\",1510.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6257]]",
        6258=>"PROJCS[\"MAGNA_Mitu_Vaupes_2011\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1093717.398],PARAMETER[\"False_Northing\",629997.236],PARAMETER[\"Longitude_Of_Center\",-70.23546165555555],PARAMETER[\"Latitude_Of_Center\",1.249969366666667],PARAMETER[\"Height\",170.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6258]]",
        6259=>"PROJCS[\"MAGNA_Mocoa_Putumayo_2011\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1047467.388],PARAMETER[\"False_Northing\",617828.474],PARAMETER[\"Longitude_Of_Center\",-76.65102121944444],PARAMETER[\"Latitude_Of_Center\",1.140023358333333],PARAMETER[\"Height\",655.2],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6259]]",
        6260=>"PROJCS[\"MAGNA_Monteria_Cordoba_2006\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1131814.934],PARAMETER[\"False_Northing\",1462131.119],PARAMETER[\"Longitude_Of_Center\",-75.87955333055555],PARAMETER[\"Latitude_Of_Center\",8.773085755555556],PARAMETER[\"Height\",15.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6260]]",
        6261=>"PROJCS[\"MAGNA_Neiva_Huila_2006\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",864476.923],PARAMETER[\"False_Northing\",817199.827],PARAMETER[\"Longitude_Of_Center\",-75.29643672222223],PARAMETER[\"Latitude_Of_Center\",2.942415055555556],PARAMETER[\"Height\",430.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6261]]",
        6262=>"PROJCS[\"MAGNA_Pasto_Narino_2008\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",980469.695],PARAMETER[\"False_Northing\",624555.332],PARAMETER[\"Longitude_Of_Center\",-77.25312563333334],PARAMETER[\"Latitude_Of_Center\",1.200989513888889],PARAMETER[\"Height\",2530.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6262]]",
        6263=>"PROJCS[\"MAGNA_Pereira_Risaralda_2007\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1153492.012],PARAMETER[\"False_Northing\",1024195.255],PARAMETER[\"Longitude_Of_Center\",-75.69395138888889],PARAMETER[\"Latitude_Of_Center\",4.813593611111111],PARAMETER[\"Height\",1500.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6263]]",
        6264=>"PROJCS[\"MAGNA_Popayan_Cauca_2006\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1052430.525],PARAMETER[\"False_Northing\",763366.548],PARAMETER[\"Longitude_Of_Center\",-76.6060916361111],PARAMETER[\"Latitude_Of_Center\",2.456159883333334],PARAMETER[\"Height\",1740.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6264]]",
        6265=>"PROJCS[\"MAGNA_Puerto_Carreno_Vichada_2011\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1063834.703],PARAMETER[\"False_Northing\",1175257.481],PARAMETER[\"Longitude_Of_Center\",-67.50075024722223],PARAMETER[\"Latitude_Of_Center\",6.18072141388889],PARAMETER[\"Height\",51.58],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6265]]",
        6266=>"PROJCS[\"MAGNA_Quibdo_Choco_2011\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1047273.617],PARAMETER[\"False_Northing\",1121443.09],PARAMETER[\"Longitude_Of_Center\",-76.65075385833335],PARAMETER[\"Latitude_Of_Center\",5.694247661111111],PARAMETER[\"Height\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6266]]",
        6267=>"PROJCS[\"MAGNA_Riohacha_La_Guajira_2006\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1128154.73],PARAMETER[\"False_Northing\",1767887.914],PARAMETER[\"Longitude_Of_Center\",-72.90276886944444],PARAMETER[\"Latitude_Of_Center\",11.53691332777778],PARAMETER[\"Height\",6.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6267]]",
        6268=>"PROJCS[\"MAGNA_San_Andres_2007\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",820439.298],PARAMETER[\"False_Northing\",1877357.828],PARAMETER[\"Longitude_Of_Center\",-81.72937595],PARAMETER[\"Latitude_Of_Center\",12.523794325],PARAMETER[\"Height\",6.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6268]]",
        6269=>"PROJCS[\"MAGNA_San_Jose_del_Guaviare_2011\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1159876.62],PARAMETER[\"False_Northing\",775380.342],PARAMETER[\"Longitude_Of_Center\",-72.640033325],PARAMETER[\"Latitude_Of_Center\",2.564078941666666],PARAMETER[\"Height\",185.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6269]]",
        6270=>"PROJCS[\"MAGNA_Santa_Marta_Magdalena_2007\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",983892.409],PARAMETER[\"False_Northing\",1732533.518],PARAMETER[\"Longitude_Of_Center\",-74.22500527777778],PARAMETER[\"Latitude_Of_Center\",11.21964305555556],PARAMETER[\"Height\",29.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6270]]",
        6271=>"PROJCS[\"MAGNA_Sucre_2006\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",929043.607],PARAMETER[\"False_Northing\",1466125.658],PARAMETER[\"Longitude_Of_Center\",-74.722466825],PARAMETER[\"Latitude_Of_Center\",8.810550366666668],PARAMETER[\"Height\",20.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6271]]",
        6272=>"PROJCS[\"MAGNA_Tunja_Boyaca_1997\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1080514.91],PARAMETER[\"False_Northing\",1103772.028],PARAMETER[\"Longitude_Of_Center\",-73.3519389],PARAMETER[\"Latitude_Of_Center\",5.534194738888889],PARAMETER[\"Height\",2800.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6272]]",
        6273=>"PROJCS[\"MAGNA_Valledupar_Cesar_2011\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1090979.66],PARAMETER[\"False_Northing\",1647208.93],PARAMETER[\"Longitude_Of_Center\",-73.2465713888889],PARAMETER[\"Latitude_Of_Center\",10.44726111111111],PARAMETER[\"Height\",200.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6273]]",
        6274=>"PROJCS[\"MAGNA_Villavicencio_Meta_2011\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",1050678.757],PARAMETER[\"False_Northing\",950952.124],PARAMETER[\"Longitude_Of_Center\",-73.62448598611111],PARAMETER[\"Latitude_Of_Center\",4.1553751],PARAMETER[\"Height\",427.19],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6274]]",
        6275=>"PROJCS[\"MAGNA_Yopal_Casanare_2006\",GEOGCS[\"GCS_MAGNA\",DATUM[\"D_MAGNA\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"IGAC_Plano_Cartesiano\"],PARAMETER[\"False_Easting\",851184.177],PARAMETER[\"False_Northing\",1083954.137],PARAMETER[\"Longitude_Of_Center\",-72.42004027777779],PARAMETER[\"Latitude_Of_Center\",5.353927222222222],PARAMETER[\"Height\",300.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6275]]",
        6307=>"PROJCS[\"NAD_1983_CORS96_SPCS_Puerto_Rico_and_Virgin_Islands\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Central_Meridian\",-66.43333333333337],PARAMETER[\"Standard_Parallel_1\",18.03333333333334],PARAMETER[\"Standard_Parallel_2\",18.43333333333334],PARAMETER[\"Latitude_Of_Origin\",17.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6307]]",
        6312=>"PROJCS[\"CGRS_1993_Cyprus_Local_Transverse_Mercator\",GEOGCS[\"GCS_CGRS_1993\",DATUM[\"D_Cyprus_Geodetic_Reference_System_1993\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",-3500000.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6312]]",
        6316=>"PROJCS[\"Macedonia_State_Coordinate_System_zone_7\",GEOGCS[\"GCS_MGI_1901\",DATUM[\"D_MGI_1901\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",7500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6316]]",
        6328=>"PROJCS[\"NAD_1983_2011_UTM_Zone_59N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6328]]",
        6329=>"PROJCS[\"NAD_1983_2011_UTM_Zone_60N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6329]]",
        6330=>"PROJCS[\"NAD_1983_2011_UTM_Zone_1N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6330]]",
        6331=>"PROJCS[\"NAD_1983_2011_UTM_Zone_2N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6331]]",
        6332=>"PROJCS[\"NAD_1983_2011_UTM_Zone_3N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6332]]",
        6333=>"PROJCS[\"NAD_1983_2011_UTM_Zone_4N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6333]]",
        6334=>"PROJCS[\"NAD_1983_2011_UTM_Zone_5N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6334]]",
        6335=>"PROJCS[\"NAD_1983_2011_UTM_Zone_6N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6335]]",
        6336=>"PROJCS[\"NAD_1983_2011_UTM_Zone_7N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6336]]",
        6337=>"PROJCS[\"NAD_1983_2011_UTM_Zone_8N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6337]]",
        6338=>"PROJCS[\"NAD_1983_2011_UTM_Zone_9N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6338]]",
        6339=>"PROJCS[\"NAD_1983_2011_UTM_Zone_10N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6339]]",
        6340=>"PROJCS[\"NAD_1983_2011_UTM_Zone_11N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6340]]",
        6341=>"PROJCS[\"NAD_1983_2011_UTM_Zone_12N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6341]]",
        6342=>"PROJCS[\"NAD_1983_2011_UTM_Zone_13N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6342]]",
        6343=>"PROJCS[\"NAD_1983_2011_UTM_Zone_14N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6343]]",
        6344=>"PROJCS[\"NAD_1983_2011_UTM_Zone_15N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6344]]",
        6345=>"PROJCS[\"NAD_1983_2011_UTM_Zone_16N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6345]]",
        6346=>"PROJCS[\"NAD_1983_2011_UTM_Zone_17N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6346]]",
        6347=>"PROJCS[\"NAD_1983_2011_UTM_Zone_18N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6347]]",
        6348=>"PROJCS[\"NAD_1983_2011_UTM_Zone_19N\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6348]]",
        6350=>"PROJCS[\"NAD_1983_2011_Contiguous_USA_Albers\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-96.0],PARAMETER[\"Standard_Parallel_1\",29.5],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",23.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6350]]",
        6351=>"PROJCS[\"NAD_1983_2011_EPSG_Arctic_zone_5-29\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",29500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",-163.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6351]]",
        6352=>"PROJCS[\"NAD_1983_2011_EPSG_Arctic_zone_5-31\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",31500000.0],PARAMETER[\"False_Northing\",5500000.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Standard_Parallel_1\",70.33333333333333],PARAMETER[\"Standard_Parallel_2\",73.66666666666669],PARAMETER[\"Latitude_Of_Origin\",72.02500919444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6352]]",
        6353=>"PROJCS[\"NAD_1983_2011_EPSG_Arctic_zone_6-14\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",14500000.0],PARAMETER[\"False_Northing\",6500000.0],PARAMETER[\"Central_Meridian\",-165.0],PARAMETER[\"Standard_Parallel_1\",67.0],PARAMETER[\"Standard_Parallel_2\",70.33333333333333],PARAMETER[\"Latitude_Of_Origin\",68.68747555555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6353]]",
        6354=>"PROJCS[\"NAD_1983_2011_EPSG_Arctic_zone_6-16\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",16500000.0],PARAMETER[\"False_Northing\",6500000.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Standard_Parallel_1\",67.0],PARAMETER[\"Standard_Parallel_2\",70.33333333333333],PARAMETER[\"Latitude_Of_Origin\",68.68747555555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6354]]",
        6355=>"PROJCS[\"NAD_1983_2011_StatePlane_Alabama_East_FIPS_0101\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-85.83333333333333],PARAMETER[\"Scale_Factor\",0.99996],PARAMETER[\"Latitude_Of_Origin\",30.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6355]]",
        6356=>"PROJCS[\"NAD_1983_2011_StatePlane_Alabama_West_FIPS_0102\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.5],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6356]]",
        6362=>"PROJCS[\"Mexico_ITRF92_LCC\",GEOGCS[\"GCS_Mexican_Datum_of_1993\",DATUM[\"D_Mexican_Datum_of_1993\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-102.0],PARAMETER[\"Standard_Parallel_1\",17.5],PARAMETER[\"Standard_Parallel_2\",29.5],PARAMETER[\"Latitude_Of_Origin\",12.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6362]]",
        6366=>"PROJCS[\"Mexico_ITRF2008_UTM_Zone_11N\",GEOGCS[\"GCS_Mexico_ITRF2008\",DATUM[\"D_Mexico_ITRF2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6366]]",
        6367=>"PROJCS[\"Mexico_ITRF2008_UTM_Zone_12N\",GEOGCS[\"GCS_Mexico_ITRF2008\",DATUM[\"D_Mexico_ITRF2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6367]]",
        6368=>"PROJCS[\"Mexico_ITRF2008_UTM_Zone_13N\",GEOGCS[\"GCS_Mexico_ITRF2008\",DATUM[\"D_Mexico_ITRF2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6368]]",
        6369=>"PROJCS[\"Mexico_ITRF2008_UTM_Zone_14N\",GEOGCS[\"GCS_Mexico_ITRF2008\",DATUM[\"D_Mexico_ITRF2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6369]]",
        6370=>"PROJCS[\"Mexico_ITRF2008_UTM_Zone_15N\",GEOGCS[\"GCS_Mexico_ITRF2008\",DATUM[\"D_Mexico_ITRF2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6370]]",
        6371=>"PROJCS[\"Mexico_ITRF2008_UTM_Zone_16N\",GEOGCS[\"GCS_Mexico_ITRF2008\",DATUM[\"D_Mexico_ITRF2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6371]]",
        6372=>"PROJCS[\"Mexico_ITRF2008_LCC\",GEOGCS[\"GCS_Mexico_ITRF2008\",DATUM[\"D_Mexico_ITRF2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-102.0],PARAMETER[\"Standard_Parallel_1\",17.5],PARAMETER[\"Standard_Parallel_2\",29.5],PARAMETER[\"Latitude_Of_Origin\",12.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6372]]",
        6381=>"PROJCS[\"Ukraine_2000_TM_Zone_7\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6381]]",
        6382=>"PROJCS[\"Ukraine_2000_TM_Zone_8\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6382]]",
        6383=>"PROJCS[\"Ukraine_2000_TM_Zone_9\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6383]]",
        6384=>"PROJCS[\"Ukraine_2000_TM_Zone_10\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",30.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6384]]",
        6385=>"PROJCS[\"Ukraine_2000_TM_Zone_11\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6385]]",
        6386=>"PROJCS[\"Ukraine_2000_TM_Zone_12\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",36.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6386]]",
        6387=>"PROJCS[\"Ukraine_2000_TM_Zone_13\",GEOGCS[\"GCS_Ukraine_2000\",DATUM[\"D_Ukraine_2000\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6387]]",
        6391=>"PROJCS[\"Cayman_Islands_National_Grid_2011\",GEOGCS[\"GCS_CIGD11\",DATUM[\"D_Cayman_Islands_Geodetic_Datum_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2950000.0],PARAMETER[\"False_Northing\",1900000.0],PARAMETER[\"Central_Meridian\",-80.56666666666666],PARAMETER[\"Standard_Parallel_1\",19.33333333333333],PARAMETER[\"Standard_Parallel_2\",19.7],PARAMETER[\"Latitude_Of_Origin\",19.33333333333333],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6391]]",
        6393=>"PROJCS[\"NAD_1983_2011_Alaska_Albers\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-154.0],PARAMETER[\"Standard_Parallel_1\",55.0],PARAMETER[\"Standard_Parallel_2\",65.0],PARAMETER[\"Latitude_Of_Origin\",50.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6393]]",
        6394=>"PROJCS[\"NAD_1983_2011_StatePlane_Alaska_1_FIPS_5001\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Natural_Origin\"],PARAMETER[\"False_Easting\",5000000.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Azimuth\",-36.86989764583333],PARAMETER[\"Longitude_Of_Center\",-133.6666666666667],PARAMETER[\"Latitude_Of_Center\",57.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6394]]",
        6395=>"PROJCS[\"NAD_1983_2011_StatePlane_Alaska_2_FIPS_5002\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-142.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6395]]",
        6396=>"PROJCS[\"NAD_1983_2011_StatePlane_Alaska_3_FIPS_5003\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-146.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6396]]",
        6397=>"PROJCS[\"NAD_1983_2011_StatePlane_Alaska_4_FIPS_5004\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-150.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6397]]",
        6398=>"PROJCS[\"NAD_1983_2011_StatePlane_Alaska_5_FIPS_5005\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-154.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6398]]",
        6399=>"PROJCS[\"NAD_1983_2011_StatePlane_Alaska_6_FIPS_5006\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-158.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6399]]",
        6400=>"PROJCS[\"NAD_1983_2011_StatePlane_Alaska_7_FIPS_5007\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-162.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6400]]",
        6401=>"PROJCS[\"NAD_1983_2011_StatePlane_Alaska_8_FIPS_5008\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-166.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6401]]",
        6402=>"PROJCS[\"NAD_1983_2011_StatePlane_Alaska_9_FIPS_5009\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-170.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6402]]",
        6403=>"PROJCS[\"NAD_1983_2011_StatePlane_Alaska_10_FIPS_5010\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-176.0],PARAMETER[\"Standard_Parallel_1\",51.83333333333334],PARAMETER[\"Standard_Parallel_2\",53.83333333333334],PARAMETER[\"Latitude_Of_Origin\",51.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6403]]",
        6404=>"PROJCS[\"NAD_1983_2011_StatePlane_Arizona_Central_FIPS_0202\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",213360.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.9166666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6404]]",
        6405=>"PROJCS[\"NAD_1983_2011_StatePlane_Arizona_Central_FIPS_0202_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.9166666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6405]]",
        6406=>"PROJCS[\"NAD_1983_2011_StatePlane_Arizona_East_FIPS_0201\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",213360.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-110.1666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6406]]",
        6407=>"PROJCS[\"NAD_1983_2011_StatePlane_Arizona_East_FIPS_0201_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-110.1666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6407]]",
        6408=>"PROJCS[\"NAD_1983_2011_StatePlane_Arizona_West_FIPS_0203\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",213360.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-113.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6408]]",
        6409=>"PROJCS[\"NAD_1983_2011_StatePlane_Arizona_West_FIPS_0203_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-113.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6409]]",
        6410=>"PROJCS[\"NAD_1983_2011_StatePlane_Arkansas_North_FIPS_0301\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",34.93333333333333],PARAMETER[\"Standard_Parallel_2\",36.23333333333333],PARAMETER[\"Latitude_Of_Origin\",34.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6410]]",
        6411=>"PROJCS[\"NAD_1983_2011_StatePlane_Arkansas_North_FIPS_0301_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",34.93333333333333],PARAMETER[\"Standard_Parallel_2\",36.23333333333333],PARAMETER[\"Latitude_Of_Origin\",34.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6411]]",
        6412=>"PROJCS[\"NAD_1983_2011_StatePlane_Arkansas_South_FIPS_0302\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",400000.0],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",33.3],PARAMETER[\"Standard_Parallel_2\",34.76666666666667],PARAMETER[\"Latitude_Of_Origin\",32.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6412]]",
        6413=>"PROJCS[\"NAD_1983_2011_StatePlane_Arkansas_South_FIPS_0302_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",1312333.333333333],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",33.3],PARAMETER[\"Standard_Parallel_2\",34.76666666666667],PARAMETER[\"Latitude_Of_Origin\",32.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6413]]",
        6414=>"PROJCS[\"NAD_1983_2011_California_Teale_Albers\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",-4000000.0],PARAMETER[\"Central_Meridian\",-120.0],PARAMETER[\"Standard_Parallel_1\",34.0],PARAMETER[\"Standard_Parallel_2\",40.5],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6414]]",
        6415=>"PROJCS[\"NAD_1983_2011_StatePlane_California_I_FIPS_0401\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Standard_Parallel_2\",41.66666666666666],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6415]]",
        6416=>"PROJCS[\"NAD_1983_2011_StatePlane_California_I_FIPS_0401_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Standard_Parallel_2\",41.66666666666666],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6416]]",
        6417=>"PROJCS[\"NAD_1983_2011_StatePlane_California_II_FIPS_0402\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",38.33333333333334],PARAMETER[\"Standard_Parallel_2\",39.83333333333334],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6417]]",
        6418=>"PROJCS[\"NAD_1983_2011_StatePlane_California_II_FIPS_0402_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",38.33333333333334],PARAMETER[\"Standard_Parallel_2\",39.83333333333334],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6418]]",
        6419=>"PROJCS[\"NAD_1983_2011_StatePlane_California_III_FIPS_0403\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",37.06666666666667],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6419]]",
        6420=>"PROJCS[\"NAD_1983_2011_StatePlane_California_III_FIPS_0403_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",37.06666666666667],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6420]]",
        6421=>"PROJCS[\"NAD_1983_2011_StatePlane_California_IV_FIPS_0404\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-119.0],PARAMETER[\"Standard_Parallel_1\",36.0],PARAMETER[\"Standard_Parallel_2\",37.25],PARAMETER[\"Latitude_Of_Origin\",35.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6421]]",
        6422=>"PROJCS[\"NAD_1983_2011_StatePlane_California_IV_FIPS_0404_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-119.0],PARAMETER[\"Standard_Parallel_1\",36.0],PARAMETER[\"Standard_Parallel_2\",37.25],PARAMETER[\"Latitude_Of_Origin\",35.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6422]]",
        6423=>"PROJCS[\"NAD_1983_2011_StatePlane_California_V_FIPS_0405\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-118.0],PARAMETER[\"Standard_Parallel_1\",34.03333333333333],PARAMETER[\"Standard_Parallel_2\",35.46666666666667],PARAMETER[\"Latitude_Of_Origin\",33.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6423]]",
        6424=>"PROJCS[\"NAD_1983_2011_StatePlane_California_V_FIPS_0405_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-118.0],PARAMETER[\"Standard_Parallel_1\",34.03333333333333],PARAMETER[\"Standard_Parallel_2\",35.46666666666667],PARAMETER[\"Latitude_Of_Origin\",33.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6424]]",
        6425=>"PROJCS[\"NAD_1983_2011_StatePlane_California_VI_FIPS_0406\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-116.25],PARAMETER[\"Standard_Parallel_1\",32.78333333333333],PARAMETER[\"Standard_Parallel_2\",33.88333333333333],PARAMETER[\"Latitude_Of_Origin\",32.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6425]]",
        6426=>"PROJCS[\"NAD_1983_2011_StatePlane_California_VI_FIPS_0406_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6561666.666666666],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-116.25],PARAMETER[\"Standard_Parallel_1\",32.78333333333333],PARAMETER[\"Standard_Parallel_2\",33.88333333333333],PARAMETER[\"Latitude_Of_Origin\",32.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6426]]",
        6427=>"PROJCS[\"NAD_1983_2011_StatePlane_Colorado_Central_FIPS_0502\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",914401.8289],PARAMETER[\"False_Northing\",304800.6096],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",38.45],PARAMETER[\"Standard_Parallel_2\",39.75],PARAMETER[\"Latitude_Of_Origin\",37.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6427]]",
        6428=>"PROJCS[\"NAD_1983_2011_StatePlane_Colorado_Central_FIPS_0502_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.000316083],PARAMETER[\"False_Northing\",999999.999996],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",38.45],PARAMETER[\"Standard_Parallel_2\",39.75],PARAMETER[\"Latitude_Of_Origin\",37.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6428]]",
        6429=>"PROJCS[\"NAD_1983_2011_StatePlane_Colorado_North_FIPS_0501\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",914401.8289],PARAMETER[\"False_Northing\",304800.6096],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",39.71666666666667],PARAMETER[\"Standard_Parallel_2\",40.78333333333333],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6429]]",
        6430=>"PROJCS[\"NAD_1983_2011_StatePlane_Colorado_North_FIPS_0501_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.000316083],PARAMETER[\"False_Northing\",999999.999996],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",39.71666666666667],PARAMETER[\"Standard_Parallel_2\",40.78333333333333],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6430]]",
        6431=>"PROJCS[\"NAD_1983_2011_StatePlane_Colorado_South_FIPS_0503\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",914401.8289],PARAMETER[\"False_Northing\",304800.6096],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",37.23333333333333],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6431]]",
        6432=>"PROJCS[\"NAD_1983_2011_StatePlane_Colorado_South_FIPS_0503_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.000316083],PARAMETER[\"False_Northing\",999999.999996],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",37.23333333333333],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6432]]",
        6433=>"PROJCS[\"NAD_1983_2011_StatePlane_Connecticut_FIPS_0600\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",304800.6096],PARAMETER[\"False_Northing\",152400.3048],PARAMETER[\"Central_Meridian\",-72.75],PARAMETER[\"Standard_Parallel_1\",41.2],PARAMETER[\"Standard_Parallel_2\",41.86666666666667],PARAMETER[\"Latitude_Of_Origin\",40.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6433]]",
        6434=>"PROJCS[\"NAD_1983_2011_StatePlane_Connecticut_FIPS_0600_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",999999.999996],PARAMETER[\"False_Northing\",499999.999998],PARAMETER[\"Central_Meridian\",-72.75],PARAMETER[\"Standard_Parallel_1\",41.2],PARAMETER[\"Standard_Parallel_2\",41.86666666666667],PARAMETER[\"Latitude_Of_Origin\",40.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6434]]",
        6435=>"PROJCS[\"NAD_1983_2011_StatePlane_Delaware_FIPS_0700\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.41666666666667],PARAMETER[\"Scale_Factor\",0.999995],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6435]]",
        6436=>"PROJCS[\"NAD_1983_2011_StatePlane_Delaware_FIPS_0700_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.41666666666667],PARAMETER[\"Scale_Factor\",0.999995],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6436]]",
        6437=>"PROJCS[\"NAD_1983_2011_StatePlane_Florida_East_FIPS_0901\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6437]]",
        6438=>"PROJCS[\"NAD_1983_2011_StatePlane_Florida_East_FIPS_0901_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6438]]",
        6439=>"PROJCS[\"NAD_1983_2011_Florida_GDL_Albers\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.0],PARAMETER[\"Standard_Parallel_1\",24.0],PARAMETER[\"Standard_Parallel_2\",31.5],PARAMETER[\"Latitude_Of_Origin\",24.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6439]]",
        6440=>"PROJCS[\"NAD_1983_2011_StatePlane_Florida_North_FIPS_0903\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.5],PARAMETER[\"Standard_Parallel_1\",29.58333333333333],PARAMETER[\"Standard_Parallel_2\",30.75],PARAMETER[\"Latitude_Of_Origin\",29.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6440]]",
        6441=>"PROJCS[\"NAD_1983_2011_StatePlane_Florida_North_FIPS_0903_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.5],PARAMETER[\"Standard_Parallel_1\",29.58333333333333],PARAMETER[\"Standard_Parallel_2\",30.75],PARAMETER[\"Latitude_Of_Origin\",29.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6441]]",
        6442=>"PROJCS[\"NAD_1983_2011_StatePlane_Florida_West_FIPS_0902\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6442]]",
        6443=>"PROJCS[\"NAD_1983_2011_StatePlane_Florida_West_FIPS_0902_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6443]]",
        6444=>"PROJCS[\"NAD_1983_2011_StatePlane_Georgia_East_FIPS_1001\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6444]]",
        6445=>"PROJCS[\"NAD_1983_2011_StatePlane_Georgia_East_FIPS_1001_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6445]]",
        6446=>"PROJCS[\"NAD_1983_2011_StatePlane_Georgia_West_FIPS_1002\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6446]]",
        6447=>"PROJCS[\"NAD_1983_2011_StatePlane_Georgia_West_FIPS_1002_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2296583.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6447]]",
        6448=>"PROJCS[\"NAD_1983_2011_StatePlane_Idaho_Central_FIPS_1102\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-114.0],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6448]]",
        6449=>"PROJCS[\"NAD_1983_2011_StatePlane_Idaho_Central_FIPS_1102_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-114.0],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6449]]",
        6450=>"PROJCS[\"NAD_1983_2011_StatePlane_Idaho_East_FIPS_1101\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-112.1666666666667],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6450]]",
        6451=>"PROJCS[\"NAD_1983_2011_StatePlane_Idaho_East_FIPS_1101_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-112.1666666666667],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6451]]",
        6452=>"PROJCS[\"NAD_1983_2011_StatePlane_Idaho_West_FIPS_1103\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-115.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6452]]",
        6453=>"PROJCS[\"NAD_1983_2011_StatePlane_Idaho_West_FIPS_1103_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-115.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6453]]",
        6454=>"PROJCS[\"NAD_1983_2011_StatePlane_Illinois_East_FIPS_1201\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.33333333333333],PARAMETER[\"Scale_Factor\",0.999975],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6454]]",
        6455=>"PROJCS[\"NAD_1983_2011_StatePlane_Illinois_East_FIPS_1201_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.33333333333333],PARAMETER[\"Scale_Factor\",0.999975],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6455]]",
        6456=>"PROJCS[\"NAD_1983_2011_StatePlane_Illinois_West_FIPS_1202\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.16666666666667],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6456]]",
        6457=>"PROJCS[\"NAD_1983_2011_StatePlane_Illinois_West_FIPS_1202_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2296583.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.16666666666667],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6457]]",
        6458=>"PROJCS[\"NAD_1983_2011_StatePlane_Indiana_East_FIPS_1301\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",250000.0],PARAMETER[\"Central_Meridian\",-85.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6458]]",
        6459=>"PROJCS[\"NAD_1983_2011_StatePlane_Indiana_East_FIPS_1301_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",328083.3333333333],PARAMETER[\"False_Northing\",820208.3333333333],PARAMETER[\"Central_Meridian\",-85.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6459]]",
        6460=>"PROJCS[\"NAD_1983_2011_StatePlane_Indiana_West_FIPS_1302\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",900000.0],PARAMETER[\"False_Northing\",250000.0],PARAMETER[\"Central_Meridian\",-87.08333333333333],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6460]]",
        6461=>"PROJCS[\"NAD_1983_2011_StatePlane_Indiana_West_FIPS_1302_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2952750.0],PARAMETER[\"False_Northing\",820208.3333333333],PARAMETER[\"Central_Meridian\",-87.08333333333333],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6461]]",
        6462=>"PROJCS[\"NAD_1983_2011_StatePlane_Iowa_North_FIPS_1401\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",42.06666666666667],PARAMETER[\"Standard_Parallel_2\",43.26666666666667],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6462]]",
        6463=>"PROJCS[\"NAD_1983_2011_StatePlane_Iowa_North_FIPS_1401_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4921250.0],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",42.06666666666667],PARAMETER[\"Standard_Parallel_2\",43.26666666666667],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6463]]",
        6464=>"PROJCS[\"NAD_1983_2011_StatePlane_Iowa_South_FIPS_1402\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",40.61666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6464]]",
        6465=>"PROJCS[\"NAD_1983_2011_StatePlane_Iowa_South_FIPS_1402_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",40.61666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6465]]",
        6466=>"PROJCS[\"NAD_1983_2011_StatePlane_Kansas_North_FIPS_1501\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",38.71666666666667],PARAMETER[\"Standard_Parallel_2\",39.78333333333333],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6466]]",
        6467=>"PROJCS[\"NAD_1983_2011_StatePlane_Kansas_North_FIPS_1501_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",38.71666666666667],PARAMETER[\"Standard_Parallel_2\",39.78333333333333],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6467]]",
        6468=>"PROJCS[\"NAD_1983_2011_StatePlane_Kansas_South_FIPS_1502\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",400000.0],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",37.26666666666667],PARAMETER[\"Standard_Parallel_2\",38.56666666666667],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6468]]",
        6469=>"PROJCS[\"NAD_1983_2011_StatePlane_Kansas_South_FIPS_1502_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",1312333.333333333],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",37.26666666666667],PARAMETER[\"Standard_Parallel_2\",38.56666666666667],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6469]]",
        6470=>"PROJCS[\"NAD_1983_2011_StatePlane_Kentucky_North_FIPS_1601\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.25],PARAMETER[\"Standard_Parallel_1\",37.96666666666667],PARAMETER[\"Standard_Parallel_2\",38.96666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6470]]",
        6471=>"PROJCS[\"NAD_1983_2011_StatePlane_Kentucky_North_FIPS_1601_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.25],PARAMETER[\"Standard_Parallel_1\",37.96666666666667],PARAMETER[\"Standard_Parallel_2\",38.96666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6471]]",
        6472=>"PROJCS[\"NAD_1983_2011_StatePlane_Kentucky_FIPS_1600\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Standard_Parallel_1\",37.08333333333334],PARAMETER[\"Standard_Parallel_2\",38.66666666666666],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6472]]",
        6473=>"PROJCS[\"NAD_1983_2011_StatePlane_Kentucky_FIPS_1600_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4921250.0],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Standard_Parallel_1\",37.08333333333334],PARAMETER[\"Standard_Parallel_2\",38.66666666666666],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6473]]",
        6474=>"PROJCS[\"NAD_1983_2011_StatePlane_Kentucky_South_FIPS_1602\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Standard_Parallel_1\",36.73333333333333],PARAMETER[\"Standard_Parallel_2\",37.93333333333333],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6474]]",
        6475=>"PROJCS[\"NAD_1983_2011_StatePlane_Kentucky_South_FIPS_1602_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",1640416.666666667],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Standard_Parallel_1\",36.73333333333333],PARAMETER[\"Standard_Parallel_2\",37.93333333333333],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6475]]",
        6476=>"PROJCS[\"NAD_1983_2011_StatePlane_Louisiana_North_FIPS_1701\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.5],PARAMETER[\"Standard_Parallel_1\",31.16666666666667],PARAMETER[\"Standard_Parallel_2\",32.66666666666666],PARAMETER[\"Latitude_Of_Origin\",30.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6476]]",
        6477=>"PROJCS[\"NAD_1983_2011_StatePlane_Louisiana_North_FIPS_1701_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3280833.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.5],PARAMETER[\"Standard_Parallel_1\",31.16666666666667],PARAMETER[\"Standard_Parallel_2\",32.66666666666666],PARAMETER[\"Latitude_Of_Origin\",30.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6477]]",
        6478=>"PROJCS[\"NAD_1983_2011_StatePlane_Louisiana_South_FIPS_1702\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-91.33333333333333],PARAMETER[\"Standard_Parallel_1\",29.3],PARAMETER[\"Standard_Parallel_2\",30.7],PARAMETER[\"Latitude_Of_Origin\",28.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6478]]",
        6479=>"PROJCS[\"NAD_1983_2011_StatePlane_Louisiana_South_FIPS_1702_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3280833.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-91.33333333333333],PARAMETER[\"Standard_Parallel_1\",29.3],PARAMETER[\"Standard_Parallel_2\",30.7],PARAMETER[\"Latitude_Of_Origin\",28.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6479]]",
        6480=>"PROJCS[\"NAD_1983_2011_Maine_2000_Central_Zone\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.125],PARAMETER[\"Scale_Factor\",0.99998],PARAMETER[\"Latitude_Of_Origin\",43.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6480]]",
        6481=>"PROJCS[\"NAD_1983_2011_Maine_2000_East_Zone\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-67.875],PARAMETER[\"Scale_Factor\",0.99998],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6481]]",
        6482=>"PROJCS[\"NAD_1983_2011_Maine_2000_West_Zone\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.375],PARAMETER[\"Scale_Factor\",0.99998],PARAMETER[\"Latitude_Of_Origin\",42.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6482]]",
        6483=>"PROJCS[\"NAD_1983_2011_StatePlane_Maine_East_FIPS_1801\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-68.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6483]]",
        6484=>"PROJCS[\"NAD_1983_2011_StatePlane_Maine_East_FIPS_1801_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-68.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6484]]",
        6485=>"PROJCS[\"NAD_1983_2011_StatePlane_Maine_West_FIPS_1802\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",900000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.16666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6485]]",
        6486=>"PROJCS[\"NAD_1983_2011_StatePlane_Maine_West_FIPS_1802_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2952750.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.16666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6486]]",
        6487=>"PROJCS[\"NAD_1983_2011_StatePlane_Maryland_FIPS_1900\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.0],PARAMETER[\"Standard_Parallel_1\",38.3],PARAMETER[\"Standard_Parallel_2\",39.45],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6487]]",
        6488=>"PROJCS[\"NAD_1983_2011_StatePlane_Maryland_FIPS_1900_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.0],PARAMETER[\"Standard_Parallel_1\",38.3],PARAMETER[\"Standard_Parallel_2\",39.45],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6488]]",
        6489=>"PROJCS[\"NAD_1983_2011_StatePlane_Massachusetts_Island_FIPS_2002\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.5],PARAMETER[\"Standard_Parallel_1\",41.28333333333333],PARAMETER[\"Standard_Parallel_2\",41.48333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6489]]",
        6490=>"PROJCS[\"NAD_1983_2011_StatePlane_Massachusetts_Isl_FIPS_2002_FtUS\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.5],PARAMETER[\"Standard_Parallel_1\",41.28333333333333],PARAMETER[\"Standard_Parallel_2\",41.48333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6490]]",
        6491=>"PROJCS[\"NAD_1983_2011_StatePlane_Massachusetts_Mainland_FIPS_2001\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",750000.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Standard_Parallel_1\",41.71666666666667],PARAMETER[\"Standard_Parallel_2\",42.68333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6491]]",
        6492=>"PROJCS[\"NAD_1983_2011_StatePlane_Massachusetts_Mnld_FIPS_2001_FtUS\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",2460625.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Standard_Parallel_1\",41.71666666666667],PARAMETER[\"Standard_Parallel_2\",42.68333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6492]]",
        6493=>"PROJCS[\"NAD_1983_2011_StatePlane_Michigan_Central_FIPS_2112\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.36666666666666],PARAMETER[\"Standard_Parallel_1\",44.18333333333333],PARAMETER[\"Standard_Parallel_2\",45.7],PARAMETER[\"Latitude_Of_Origin\",43.31666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6493]]",
        6494=>"PROJCS[\"NAD_1983_2011_StatePlane_Michigan_Central_FIPS_2112_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",19685039.37007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.36666666666666],PARAMETER[\"Standard_Parallel_1\",44.18333333333333],PARAMETER[\"Standard_Parallel_2\",45.7],PARAMETER[\"Latitude_Of_Origin\",43.31666666666667],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6494]]",
        6495=>"PROJCS[\"NAD_1983_2011_StatePlane_Michigan_North_FIPS_2111\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",8000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Standard_Parallel_1\",45.48333333333333],PARAMETER[\"Standard_Parallel_2\",47.08333333333334],PARAMETER[\"Latitude_Of_Origin\",44.78333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6495]]",
        6496=>"PROJCS[\"NAD_1983_2011_StatePlane_Michigan_North_FIPS_2111_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",26246719.16010498],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Standard_Parallel_1\",45.48333333333333],PARAMETER[\"Standard_Parallel_2\",47.08333333333334],PARAMETER[\"Latitude_Of_Origin\",44.78333333333333],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6496]]",
        6497=>"PROJCS[\"NAD_1983_2011_Michigan_GeoRef_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Natural_Origin\"],PARAMETER[\"False_Easting\",2546731.496],PARAMETER[\"False_Northing\",-4354009.816],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Azimuth\",337.25556],PARAMETER[\"Longitude_Of_Center\",-86.0],PARAMETER[\"Latitude_Of_Center\",45.30916666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6497]]",
        6498=>"PROJCS[\"NAD_1983_2011_StatePlane_Michigan_South_FIPS_2113\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.36666666666666],PARAMETER[\"Standard_Parallel_1\",42.1],PARAMETER[\"Standard_Parallel_2\",43.66666666666666],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6498]]",
        6499=>"PROJCS[\"NAD_1983_2011_StatePlane_Michigan_South_FIPS_2113_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",13123359.58005249],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.36666666666666],PARAMETER[\"Standard_Parallel_1\",42.1],PARAMETER[\"Standard_Parallel_2\",43.66666666666666],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6499]]",
        6500=>"PROJCS[\"NAD_1983_2011_StatePlane_Minnesota_Central_FIPS_2202\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-94.25],PARAMETER[\"Standard_Parallel_1\",45.61666666666667],PARAMETER[\"Standard_Parallel_2\",47.05],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6500]]",
        6501=>"PROJCS[\"NAD_1983_2011_StatePlane_Minnesota_Central_FIPS_2202_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-94.25],PARAMETER[\"Standard_Parallel_1\",45.61666666666667],PARAMETER[\"Standard_Parallel_2\",47.05],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6501]]",
        6502=>"PROJCS[\"NAD_1983_2011_StatePlane_Minnesota_North_FIPS_2201\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-93.1],PARAMETER[\"Standard_Parallel_1\",47.03333333333333],PARAMETER[\"Standard_Parallel_2\",48.63333333333333],PARAMETER[\"Latitude_Of_Origin\",46.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6502]]",
        6503=>"PROJCS[\"NAD_1983_2011_StatePlane_Minnesota_North_FIPS_2201_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-93.1],PARAMETER[\"Standard_Parallel_1\",47.03333333333333],PARAMETER[\"Standard_Parallel_2\",48.63333333333333],PARAMETER[\"Latitude_Of_Origin\",46.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6503]]",
        6504=>"PROJCS[\"NAD_1983_2011_StatePlane_Minnesota_South_FIPS_2203\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-94.0],PARAMETER[\"Standard_Parallel_1\",43.78333333333333],PARAMETER[\"Standard_Parallel_2\",45.21666666666667],PARAMETER[\"Latitude_Of_Origin\",43.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6504]]",
        6505=>"PROJCS[\"NAD_1983_2011_StatePlane_Minnesota_South_FIPS_2203_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-94.0],PARAMETER[\"Standard_Parallel_1\",43.78333333333333],PARAMETER[\"Standard_Parallel_2\",45.21666666666667],PARAMETER[\"Latitude_Of_Origin\",43.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6505]]",
        6506=>"PROJCS[\"NAD_1983_2011_StatePlane_Mississippi_East_FIPS_2301\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.83333333333333],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",29.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6506]]",
        6507=>"PROJCS[\"NAD_1983_2011_StatePlane_Mississippi_East_FIPS_2301_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.83333333333333],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",29.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6507]]",
        6508=>"PROJCS[\"NAD_1983_2011_Mississippi_TM\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",1300000.0],PARAMETER[\"Central_Meridian\",-89.75],PARAMETER[\"Scale_Factor\",0.9998335],PARAMETER[\"Latitude_Of_Origin\",32.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6508]]",
        6509=>"PROJCS[\"NAD_1983_2011_StatePlane_Mississippi_West_FIPS_2302\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.33333333333333],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",29.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6509]]",
        6510=>"PROJCS[\"NAD_1983_2011_StatePlane_Mississippi_West_FIPS_2302_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2296583.333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.33333333333333],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",29.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6510]]",
        6511=>"PROJCS[\"NAD_1983_2011_StatePlane_Missouri_Central_FIPS_2402\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.5],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",35.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6511]]",
        6512=>"PROJCS[\"NAD_1983_2011_StatePlane_Missouri_East_FIPS_2401\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.5],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",35.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6512]]",
        6513=>"PROJCS[\"NAD_1983_2011_StatePlane_Missouri_West_FIPS_2403\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",850000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-94.5],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",36.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6513]]",
        6514=>"PROJCS[\"NAD_1983_2011_StatePlane_Montana_FIPS_2500\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-109.5],PARAMETER[\"Standard_Parallel_1\",45.0],PARAMETER[\"Standard_Parallel_2\",49.0],PARAMETER[\"Latitude_Of_Origin\",44.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6514]]",
        6515=>"PROJCS[\"NAD_1983_2011_StatePlane_Montana_FIPS_2500_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968503.937007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-109.5],PARAMETER[\"Standard_Parallel_1\",45.0],PARAMETER[\"Standard_Parallel_2\",49.0],PARAMETER[\"Latitude_Of_Origin\",44.25],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6515]]",
        6516=>"PROJCS[\"NAD_1983_2011_StatePlane_Nebraska_FIPS_2600\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Standard_Parallel_2\",43.0],PARAMETER[\"Latitude_Of_Origin\",39.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6516]]",
        6518=>"PROJCS[\"NAD_1983_2011_StatePlane_Nevada_Central_FIPS_2702\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",6000000.0],PARAMETER[\"Central_Meridian\",-116.6666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6518]]",
        6519=>"PROJCS[\"NAD_1983_2011_StatePlane_Nevada_Central_FIPS_2702_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",19685000.0],PARAMETER[\"Central_Meridian\",-116.6666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6519]]",
        6520=>"PROJCS[\"NAD_1983_2011_StatePlane_Nevada_East_FIPS_2701\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",8000000.0],PARAMETER[\"Central_Meridian\",-115.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6520]]",
        6521=>"PROJCS[\"NAD_1983_2011_StatePlane_Nevada_East_FIPS_2701_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",26246666.66666666],PARAMETER[\"Central_Meridian\",-115.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6521]]",
        6522=>"PROJCS[\"NAD_1983_2011_StatePlane_Nevada_West_FIPS_2703\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",4000000.0],PARAMETER[\"Central_Meridian\",-118.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6522]]",
        6523=>"PROJCS[\"NAD_1983_2011_StatePlane_Nevada_West_FIPS_2703_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",13123333.33333333],PARAMETER[\"Central_Meridian\",-118.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6523]]",
        6524=>"PROJCS[\"NAD_1983_2011_StatePlane_New_Hampshire_FIPS_2800\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6524]]",
        6525=>"PROJCS[\"NAD_1983_2011_StatePlane_New_Hampshire_FIPS_2800_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6525]]",
        6526=>"PROJCS[\"NAD_1983_2011_StatePlane_New_Jersey_FIPS_2900\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",38.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6526]]",
        6527=>"PROJCS[\"NAD_1983_2011_StatePlane_New_Jersey_FIPS_2900_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",492125.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",38.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6527]]",
        6528=>"PROJCS[\"NAD_1983_2011_StatePlane_New_Mexico_Central_FIPS_3002\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-106.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6528]]",
        6529=>"PROJCS[\"NAD_1983_2011_StatePlane_New_Mexico_Central_FIPS_3002_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-106.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6529]]",
        6530=>"PROJCS[\"NAD_1983_2011_StatePlane_New_Mexico_East_FIPS_3001\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",165000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-104.3333333333333],PARAMETER[\"Scale_Factor\",0.9999090909090909],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6530]]",
        6531=>"PROJCS[\"NAD_1983_2011_StatePlane_New_Mexico_East_FIPS_3001_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",541337.5],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-104.3333333333333],PARAMETER[\"Scale_Factor\",0.9999090909090909],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6531]]",
        6532=>"PROJCS[\"NAD_1983_2011_StatePlane_New_Mexico_West_FIPS_3003\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",830000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-107.8333333333333],PARAMETER[\"Scale_Factor\",0.9999166666666667],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6532]]",
        6533=>"PROJCS[\"NAD_1983_2011_StatePlane_New_Mexico_West_FIPS_3003_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2723091.666666666],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-107.8333333333333],PARAMETER[\"Scale_Factor\",0.9999166666666667],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6533]]",
        6534=>"PROJCS[\"NAD_1983_2011_StatePlane_New_York_Central_FIPS_3102\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-76.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6534]]",
        6535=>"PROJCS[\"NAD_1983_2011_StatePlane_New_York_Central_FIPS_3102_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",820208.3333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-76.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6535]]",
        6536=>"PROJCS[\"NAD_1983_2011_StatePlane_New_York_East_FIPS_3101\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",38.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6536]]",
        6537=>"PROJCS[\"NAD_1983_2011_StatePlane_New_York_East_FIPS_3101_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",492125.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",38.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6537]]",
        6538=>"PROJCS[\"NAD_1983_2011_StatePlane_New_York_Long_Island_FIPS_3104\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.0],PARAMETER[\"Standard_Parallel_1\",40.66666666666666],PARAMETER[\"Standard_Parallel_2\",41.03333333333333],PARAMETER[\"Latitude_Of_Origin\",40.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6538]]",
        6539=>"PROJCS[\"NAD_1983_2011_StatePlane_New_York_Long_Isl_FIPS_3104_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.0],PARAMETER[\"Standard_Parallel_1\",40.66666666666666],PARAMETER[\"Standard_Parallel_2\",41.03333333333333],PARAMETER[\"Latitude_Of_Origin\",40.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6539]]",
        6540=>"PROJCS[\"NAD_1983_2011_StatePlane_New_York_West_FIPS_3103\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",350000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-78.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6540]]",
        6541=>"PROJCS[\"NAD_1983_2011_StatePlane_New_York_West_FIPS_3103_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1148291.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-78.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6541]]",
        6542=>"PROJCS[\"NAD_1983_2011_StatePlane_North_Carolina_FIPS_3200\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",609601.2192024384],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.0],PARAMETER[\"Standard_Parallel_1\",34.33333333333334],PARAMETER[\"Standard_Parallel_2\",36.16666666666666],PARAMETER[\"Latitude_Of_Origin\",33.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6542]]",
        6543=>"PROJCS[\"NAD_1983_2011_StatePlane_North_Carolina_FIPS_3200_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.0],PARAMETER[\"Standard_Parallel_1\",34.33333333333334],PARAMETER[\"Standard_Parallel_2\",36.16666666666666],PARAMETER[\"Latitude_Of_Origin\",33.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6543]]",
        6544=>"PROJCS[\"NAD_1983_2011_StatePlane_North_Dakota_North_FIPS_3301\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",47.43333333333333],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6544]]",
        6545=>"PROJCS[\"NAD_1983_2011_StatePlane_North_Dakota_North_FIPS_3301_FtI\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968503.937007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",47.43333333333333],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6545]]",
        6546=>"PROJCS[\"NAD_1983_2011_StatePlane_North_Dakota_South_FIPS_3302\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",46.18333333333333],PARAMETER[\"Standard_Parallel_2\",47.48333333333333],PARAMETER[\"Latitude_Of_Origin\",45.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6546]]",
        6547=>"PROJCS[\"NAD_1983_2011_StatePlane_North_Dakota_South_FIPS_3302_FtI\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968503.937007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",46.18333333333333],PARAMETER[\"Standard_Parallel_2\",47.48333333333333],PARAMETER[\"Latitude_Of_Origin\",45.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6547]]",
        6548=>"PROJCS[\"NAD_1983_2011_StatePlane_Ohio_North_FIPS_3401\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",40.43333333333333],PARAMETER[\"Standard_Parallel_2\",41.7],PARAMETER[\"Latitude_Of_Origin\",39.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6548]]",
        6549=>"PROJCS[\"NAD_1983_2011_StatePlane_Ohio_North_FIPS_3401_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",40.43333333333333],PARAMETER[\"Standard_Parallel_2\",41.7],PARAMETER[\"Latitude_Of_Origin\",39.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6549]]",
        6550=>"PROJCS[\"NAD_1983_2011_StatePlane_Ohio_South_FIPS_3402\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",38.73333333333333],PARAMETER[\"Standard_Parallel_2\",40.03333333333333],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6550]]",
        6551=>"PROJCS[\"NAD_1983_2011_StatePlane_Ohio_South_FIPS_3402_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",38.73333333333333],PARAMETER[\"Standard_Parallel_2\",40.03333333333333],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6551]]",
        6552=>"PROJCS[\"NAD_1983_2011_StatePlane_Oklahoma_North_FIPS_3501\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",35.56666666666667],PARAMETER[\"Standard_Parallel_2\",36.76666666666667],PARAMETER[\"Latitude_Of_Origin\",35.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6552]]",
        6553=>"PROJCS[\"NAD_1983_2011_StatePlane_Oklahoma_North_FIPS_3501_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",35.56666666666667],PARAMETER[\"Standard_Parallel_2\",36.76666666666667],PARAMETER[\"Latitude_Of_Origin\",35.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6553]]",
        6554=>"PROJCS[\"NAD_1983_2011_StatePlane_Oklahoma_South_FIPS_3502\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",33.93333333333333],PARAMETER[\"Standard_Parallel_2\",35.23333333333333],PARAMETER[\"Latitude_Of_Origin\",33.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6554]]",
        6555=>"PROJCS[\"NAD_1983_2011_StatePlane_Oklahoma_South_FIPS_3502_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",33.93333333333333],PARAMETER[\"Standard_Parallel_2\",35.23333333333333],PARAMETER[\"Latitude_Of_Origin\",33.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6555]]",
        6556=>"PROJCS[\"NAD_1983_2011_Oregon_Statewide_Lambert\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",43.0],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6556]]",
        6557=>"PROJCS[\"NAD_1983_2011_Oregon_Statewide_Lambert_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312335.958005249],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",43.0],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6557]]",
        6558=>"PROJCS[\"NAD_1983_2011_StatePlane_Oregon_North_FIPS_3601\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",44.33333333333334],PARAMETER[\"Standard_Parallel_2\",46.0],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6558]]",
        6559=>"PROJCS[\"NAD_1983_2011_StatePlane_Oregon_North_FIPS_3601_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",8202099.737532808],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",44.33333333333334],PARAMETER[\"Standard_Parallel_2\",46.0],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6559]]",
        6560=>"PROJCS[\"NAD_1983_2011_StatePlane_Oregon_South_FIPS_3602\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",42.33333333333334],PARAMETER[\"Standard_Parallel_2\",44.0],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6560]]",
        6561=>"PROJCS[\"NAD_1983_2011_StatePlane_Oregon_South_FIPS_3602_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4921259.842519685],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",42.33333333333334],PARAMETER[\"Standard_Parallel_2\",44.0],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6561]]",
        6562=>"PROJCS[\"NAD_1983_2011_StatePlane_Pennsylvania_North_FIPS_3701\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",40.88333333333333],PARAMETER[\"Standard_Parallel_2\",41.95],PARAMETER[\"Latitude_Of_Origin\",40.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6562]]",
        6563=>"PROJCS[\"NAD_1983_2011_StatePlane_Pennsylvania_North_FIPS_3701_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",40.88333333333333],PARAMETER[\"Standard_Parallel_2\",41.95],PARAMETER[\"Latitude_Of_Origin\",40.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6563]]",
        6564=>"PROJCS[\"NAD_1983_2011_StatePlane_Pennsylvania_South_FIPS_3702\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",39.93333333333333],PARAMETER[\"Standard_Parallel_2\",40.96666666666667],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6564]]",
        6565=>"PROJCS[\"NAD_1983_2011_StatePlane_Pennsylvania_South_FIPS_3702_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",39.93333333333333],PARAMETER[\"Standard_Parallel_2\",40.96666666666667],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6565]]",
        6566=>"PROJCS[\"NAD_1983_2011_StatePlane_Puerto_Rico_Virgin_Isls_FIPS_5200\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Central_Meridian\",-66.43333333333334],PARAMETER[\"Standard_Parallel_1\",18.03333333333333],PARAMETER[\"Standard_Parallel_2\",18.43333333333333],PARAMETER[\"Latitude_Of_Origin\",17.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6566]]",
        6567=>"PROJCS[\"NAD_1983_2011_StatePlane_Rhode_Island_FIPS_3800\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Scale_Factor\",0.99999375],PARAMETER[\"Latitude_Of_Origin\",41.08333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6567]]",
        6568=>"PROJCS[\"NAD_1983_2011_StatePlane_Rhode_Island_FIPS_3800_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",328083.3333333333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Scale_Factor\",0.99999375],PARAMETER[\"Latitude_Of_Origin\",41.08333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6568]]",
        6569=>"PROJCS[\"NAD_1983_2011_StatePlane_South_Carolina_FIPS_3900\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",609600.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",32.5],PARAMETER[\"Standard_Parallel_2\",34.83333333333334],PARAMETER[\"Latitude_Of_Origin\",31.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6569]]",
        6570=>"PROJCS[\"NAD_1983_2011_StatePlane_South_Carolina_FIPS_3900_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",32.5],PARAMETER[\"Standard_Parallel_2\",34.83333333333334],PARAMETER[\"Latitude_Of_Origin\",31.83333333333333],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6570]]",
        6571=>"PROJCS[\"NAD_1983_2011_StatePlane_South_Dakota_North_FIPS_4001\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",44.41666666666666],PARAMETER[\"Standard_Parallel_2\",45.68333333333333],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6571]]",
        6572=>"PROJCS[\"NAD_1983_2011_StatePlane_South_Dakota_North_FIPS_4001_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",44.41666666666666],PARAMETER[\"Standard_Parallel_2\",45.68333333333333],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6572]]",
        6573=>"PROJCS[\"NAD_1983_2011_StatePlane_South_Dakota_South_FIPS_4002\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",42.83333333333334],PARAMETER[\"Standard_Parallel_2\",44.4],PARAMETER[\"Latitude_Of_Origin\",42.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6573]]",
        6574=>"PROJCS[\"NAD_1983_2011_StatePlane_South_Dakota_South_FIPS_4002_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",42.83333333333334],PARAMETER[\"Standard_Parallel_2\",44.4],PARAMETER[\"Latitude_Of_Origin\",42.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6574]]",
        6575=>"PROJCS[\"NAD_1983_2011_StatePlane_Tennessee_FIPS_4100\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-86.0],PARAMETER[\"Standard_Parallel_1\",35.25],PARAMETER[\"Standard_Parallel_2\",36.41666666666666],PARAMETER[\"Latitude_Of_Origin\",34.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6575]]",
        6576=>"PROJCS[\"NAD_1983_2011_StatePlane_Tennessee_FIPS_4100_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-86.0],PARAMETER[\"Standard_Parallel_1\",35.25],PARAMETER[\"Standard_Parallel_2\",36.41666666666666],PARAMETER[\"Latitude_Of_Origin\",34.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6576]]",
        6577=>"PROJCS[\"NAD_1983_2011_StatePlane_Texas_Central_FIPS_4203\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",3000000.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",30.11666666666667],PARAMETER[\"Standard_Parallel_2\",31.88333333333333],PARAMETER[\"Latitude_Of_Origin\",29.66666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6577]]",
        6578=>"PROJCS[\"NAD_1983_2011_StatePlane_Texas_Central_FIPS_4203_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2296583.333333333],PARAMETER[\"False_Northing\",9842500.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",30.11666666666667],PARAMETER[\"Standard_Parallel_2\",31.88333333333333],PARAMETER[\"Latitude_Of_Origin\",29.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6578]]",
        6579=>"PROJCS[\"NAD_1983_2011_Texas_Centric_Mapping_System_Albers\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",6000000.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",27.5],PARAMETER[\"Standard_Parallel_2\",35.0],PARAMETER[\"Latitude_Of_Origin\",18.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6579]]",
        6580=>"PROJCS[\"NAD_1983_2011_Texas_Centric_Mapping_System_Lambert\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",5000000.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",27.5],PARAMETER[\"Standard_Parallel_2\",35.0],PARAMETER[\"Latitude_Of_Origin\",18.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6580]]",
        6581=>"PROJCS[\"NAD_1983_2011_StatePlane_Texas_North_FIPS_4201\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-101.5],PARAMETER[\"Standard_Parallel_1\",34.65],PARAMETER[\"Standard_Parallel_2\",36.18333333333333],PARAMETER[\"Latitude_Of_Origin\",34.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6581]]",
        6582=>"PROJCS[\"NAD_1983_2011_StatePlane_Texas_North_FIPS_4201_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-101.5],PARAMETER[\"Standard_Parallel_1\",34.65],PARAMETER[\"Standard_Parallel_2\",36.18333333333333],PARAMETER[\"Latitude_Of_Origin\",34.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6582]]",
        6583=>"PROJCS[\"NAD_1983_2011_StatePlane_Texas_North_Central_FIPS_4202\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",32.13333333333333],PARAMETER[\"Standard_Parallel_2\",33.96666666666667],PARAMETER[\"Latitude_Of_Origin\",31.66666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6583]]",
        6584=>"PROJCS[\"NAD_1983_2011_StatePlane_Texas_North_Central_FIPS_4202_FtUS\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",6561666.666666666],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",32.13333333333333],PARAMETER[\"Standard_Parallel_2\",33.96666666666667],PARAMETER[\"Latitude_Of_Origin\",31.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6584]]",
        6585=>"PROJCS[\"NAD_1983_2011_StatePlane_Texas_South_FIPS_4205\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",5000000.0],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",26.16666666666667],PARAMETER[\"Standard_Parallel_2\",27.83333333333333],PARAMETER[\"Latitude_Of_Origin\",25.66666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6585]]",
        6586=>"PROJCS[\"NAD_1983_2011_StatePlane_Texas_South_FIPS_4205_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",16404166.66666666],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",26.16666666666667],PARAMETER[\"Standard_Parallel_2\",27.83333333333333],PARAMETER[\"Latitude_Of_Origin\",25.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6586]]",
        6587=>"PROJCS[\"NAD_1983_2011_StatePlane_Texas_South_Central_FIPS_4204\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",4000000.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Standard_Parallel_1\",28.38333333333333],PARAMETER[\"Standard_Parallel_2\",30.28333333333333],PARAMETER[\"Latitude_Of_Origin\",27.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6587]]",
        6588=>"PROJCS[\"NAD_1983_2011_StatePlane_Texas_South_Central_FIPS_4204_FtUS\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",13123333.33333333],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Standard_Parallel_1\",28.38333333333333],PARAMETER[\"Standard_Parallel_2\",30.28333333333333],PARAMETER[\"Latitude_Of_Origin\",27.83333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6588]]",
        6589=>"PROJCS[\"NAD_1983_2011_StatePlane_Vermont_FIPS_4400\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-72.5],PARAMETER[\"Scale_Factor\",0.9999642857142857],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6589]]",
        6590=>"PROJCS[\"NAD_1983_2011_StatePlane_Vermont_FIPS_4400_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-72.5],PARAMETER[\"Scale_Factor\",0.9999642857142857],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6590]]",
        6591=>"PROJCS[\"NAD_1983_2011_Virginia_Lambert\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Standard_Parallel_1\",37.0],PARAMETER[\"Standard_Parallel_2\",39.5],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6591]]",
        6592=>"PROJCS[\"NAD_1983_2011_StatePlane_Virginia_North_FIPS_4501\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",38.03333333333333],PARAMETER[\"Standard_Parallel_2\",39.2],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6592]]",
        6593=>"PROJCS[\"NAD_1983_2011_StatePlane_Virginia_North_FIPS_4501_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",11482916.66666666],PARAMETER[\"False_Northing\",6561666.666666666],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",38.03333333333333],PARAMETER[\"Standard_Parallel_2\",39.2],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6593]]",
        6594=>"PROJCS[\"NAD_1983_2011_StatePlane_Virginia_South_FIPS_4502\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",36.76666666666667],PARAMETER[\"Standard_Parallel_2\",37.96666666666667],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6594]]",
        6595=>"PROJCS[\"NAD_1983_2011_StatePlane_Virginia_South_FIPS_4502_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",11482916.66666666],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",36.76666666666667],PARAMETER[\"Standard_Parallel_2\",37.96666666666667],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6595]]",
        6596=>"PROJCS[\"NAD_1983_2011_StatePlane_Washington_North_FIPS_4601\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.8333333333333],PARAMETER[\"Standard_Parallel_1\",47.5],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6596]]",
        6597=>"PROJCS[\"NAD_1983_2011_StatePlane_Washington_North_FIPS_4601_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.8333333333333],PARAMETER[\"Standard_Parallel_1\",47.5],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6597]]",
        6598=>"PROJCS[\"NAD_1983_2011_StatePlane_Washington_South_FIPS_4602\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",45.83333333333334],PARAMETER[\"Standard_Parallel_2\",47.33333333333334],PARAMETER[\"Latitude_Of_Origin\",45.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6598]]",
        6599=>"PROJCS[\"NAD_1983_2011_StatePlane_Washington_South_FIPS_4602_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",45.83333333333334],PARAMETER[\"Standard_Parallel_2\",47.33333333333334],PARAMETER[\"Latitude_Of_Origin\",45.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6599]]",
        6600=>"PROJCS[\"NAD_1983_2011_StatePlane_West_Virginia_North_FIPS_4701\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Standard_Parallel_1\",39.0],PARAMETER[\"Standard_Parallel_2\",40.25],PARAMETER[\"Latitude_Of_Origin\",38.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6600]]",
        6601=>"PROJCS[\"NAD_1983_2011_StatePlane_West_Virginia_North_FIPS_4701_FtUS\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Standard_Parallel_1\",39.0],PARAMETER[\"Standard_Parallel_2\",40.25],PARAMETER[\"Latitude_Of_Origin\",38.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6601]]",
        6602=>"PROJCS[\"NAD_1983_2011_StatePlane_West_Virginia_South_FIPS_4702\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",37.48333333333333],PARAMETER[\"Standard_Parallel_2\",38.88333333333333],PARAMETER[\"Latitude_Of_Origin\",37.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6602]]",
        6603=>"PROJCS[\"NAD_1983_2011_StatePlane_West_Virginia_South_FIPS_4702_FtUS\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",37.48333333333333],PARAMETER[\"Standard_Parallel_2\",38.88333333333333],PARAMETER[\"Latitude_Of_Origin\",37.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6603]]",
        6605=>"PROJCS[\"NAD_1983_2011_StatePlane_Wisconsin_Central_FIPS_4802_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",44.25],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6605]]",
        6606=>"PROJCS[\"NAD_1983_2011_StatePlane_Wisconsin_North_FIPS_4801\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",45.56666666666667],PARAMETER[\"Standard_Parallel_2\",46.76666666666667],PARAMETER[\"Latitude_Of_Origin\",45.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6606]]",
        6607=>"PROJCS[\"NAD_1983_2011_StatePlane_Wisconsin_North_FIPS_4801_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",45.56666666666667],PARAMETER[\"Standard_Parallel_2\",46.76666666666667],PARAMETER[\"Latitude_Of_Origin\",45.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6607]]",
        6608=>"PROJCS[\"NAD_1983_2011_StatePlane_Wisconsin_South_FIPS_4803\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",42.73333333333333],PARAMETER[\"Standard_Parallel_2\",44.06666666666667],PARAMETER[\"Latitude_Of_Origin\",42.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6608]]",
        6609=>"PROJCS[\"NAD_1983_2011_StatePlane_Wisconsin_South_FIPS_4803_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",42.73333333333333],PARAMETER[\"Standard_Parallel_2\",44.06666666666667],PARAMETER[\"Latitude_Of_Origin\",42.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6609]]",
        6610=>"PROJCS[\"NAD_1983_2011_Wisconsin_TM\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",520000.0],PARAMETER[\"False_Northing\",-4480000.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6610]]",
        6611=>"PROJCS[\"NAD_1983_2011_StatePlane_Wyoming_East_FIPS_4901\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.1666666666667],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6611]]",
        6612=>"PROJCS[\"NAD_1983_2011_StatePlane_Wyoming_East_FIPS_4901_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656166.6666666665],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.1666666666667],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6612]]",
        6613=>"PROJCS[\"NAD_1983_2011_StatePlane_Wyoming_East_Central_FIPS_4902\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-107.3333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6613]]",
        6614=>"PROJCS[\"NAD_1983_2011_StatePlane_Wyoming_E_Central_FIPS_4902_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1312333.333333333],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-107.3333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6614]]",
        6615=>"PROJCS[\"NAD_1983_2011_StatePlane_Wyoming_West_FIPS_4904\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-110.0833333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6615]]",
        6616=>"PROJCS[\"NAD_1983_2011_StatePlane_Wyoming_West_FIPS_4904_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-110.0833333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6616]]",
        6617=>"PROJCS[\"NAD_1983_2011_StatePlane_Wyoming_West_Central_FIPS_4903\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-108.75],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6617]]",
        6618=>"PROJCS[\"NAD_1983_2011_StatePlane_Wyoming_W_Central_FIPS_4903_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-108.75],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6618]]",
        6619=>"PROJCS[\"NAD_1983_2011_StatePlane_Utah_Central_FIPS_4302\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",39.01666666666667],PARAMETER[\"Standard_Parallel_2\",40.65],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6619]]",
        6620=>"PROJCS[\"NAD_1983_2011_StatePlane_Utah_North_FIPS_4301\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",40.71666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6620]]",
        6621=>"PROJCS[\"NAD_1983_2011_StatePlane_Utah_South_FIPS_4303\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",3000000.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",37.21666666666667],PARAMETER[\"Standard_Parallel_2\",38.35],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6621]]",
        6622=>"PROJCS[\"NAD_1983_CSRS_Quebec_Lambert\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-68.5],PARAMETER[\"Standard_Parallel_1\",46.0],PARAMETER[\"Standard_Parallel_2\",60.0],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6622]]",
        6623=>"PROJCS[\"NAD_1983_Quebec_Albers\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-68.5],PARAMETER[\"Standard_Parallel_1\",46.0],PARAMETER[\"Standard_Parallel_2\",60.0],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6623]]",
        6624=>"PROJCS[\"NAD_1983_CSRS_Quebec_Albers\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Albers\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-68.5],PARAMETER[\"Standard_Parallel_1\",46.0],PARAMETER[\"Standard_Parallel_2\",60.0],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6624]]",
        6625=>"PROJCS[\"NAD_1983_2011_StatePlane_Utah_Central_FIPS_4302_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",6561666.666666666],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",39.01666666666667],PARAMETER[\"Standard_Parallel_2\",40.65],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6625]]",
        6626=>"PROJCS[\"NAD_1983_2011_StatePlane_Utah_North_FIPS_4301_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",3280833.333333333],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",40.71666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6626]]",
        6627=>"PROJCS[\"NAD_1983_2011_StatePlane_Utah_South_FIPS_4303_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",9842500.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",37.21666666666667],PARAMETER[\"Standard_Parallel_2\",38.35],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6627]]",
        6628=>"PROJCS[\"NAD_1983_PA11_StatePlane_Hawaii_1_FIPS_5101\",GEOGCS[\"GCS_NAD_1983_PA11\",DATUM[\"D_NAD_1983_PA11\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-155.5],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",18.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6628]]",
        6629=>"PROJCS[\"NAD_1983_PA11_StatePlane_Hawaii_2_FIPS_5102\",GEOGCS[\"GCS_NAD_1983_PA11\",DATUM[\"D_NAD_1983_PA11\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-156.6666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",20.33333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6629]]",
        6630=>"PROJCS[\"NAD_1983_PA11_StatePlane_Hawaii_3_FIPS_5103\",GEOGCS[\"GCS_NAD_1983_PA11\",DATUM[\"D_NAD_1983_PA11\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-158.0],PARAMETER[\"Scale_Factor\",0.99999],PARAMETER[\"Latitude_Of_Origin\",21.16666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6630]]",
        6631=>"PROJCS[\"NAD_1983_PA11_StatePlane_Hawaii_4_FIPS_5104\",GEOGCS[\"GCS_NAD_1983_PA11\",DATUM[\"D_NAD_1983_PA11\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-159.5],PARAMETER[\"Scale_Factor\",0.99999],PARAMETER[\"Latitude_Of_Origin\",21.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6631]]",
        6632=>"PROJCS[\"NAD_1983_PA11_StatePlane_Hawaii_5_FIPS_5105\",GEOGCS[\"GCS_NAD_1983_PA11\",DATUM[\"D_NAD_1983_PA11\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-160.1666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",21.66666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6632]]",
        6633=>"PROJCS[\"NAD_1983_PA11_StatePlane_Hawaii_3_FIPS_5103_Feet\",GEOGCS[\"GCS_NAD_1983_PA11\",DATUM[\"D_NAD_1983_PA11\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-158.0],PARAMETER[\"Scale_Factor\",0.99999],PARAMETER[\"Latitude_Of_Origin\",21.16666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6633]]",
        6634=>"PROJCS[\"NAD_1983_PA11_UTM_Zone_4N\",GEOGCS[\"GCS_NAD_1983_PA11\",DATUM[\"D_NAD_1983_PA11\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6634]]",
        6635=>"PROJCS[\"NAD_1983_PA11_UTM_Zone_5N\",GEOGCS[\"GCS_NAD_1983_PA11\",DATUM[\"D_NAD_1983_PA11\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6635]]",
        6636=>"PROJCS[\"NAD_1983_PA11_UTM_Zone_2S\",GEOGCS[\"GCS_NAD_1983_PA11\",DATUM[\"D_NAD_1983_PA11\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6636]]",
        6637=>"PROJCS[\"NAD_1983_MA11_Guam_Map_Grid\",GEOGCS[\"GCS_NAD_1983_MA11\",DATUM[\"D_NAD_1983_MA11\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Central_Meridian\",144.75],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",13.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6637]]",
        6646=>"PROJCS[\"Karbala_1979_Iraq_National_Grid\",GEOGCS[\"GCS_Karbala_1979_Polservice\",DATUM[\"D_Karbala_1979_Polservice\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",46.5],PARAMETER[\"Scale_Factor\",0.9994],PARAMETER[\"Latitude_Of_Origin\",29.02626833333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6646]]",
        6669=>"PROJCS[\"JGD_2011_Japan_Zone_1\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",33.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6669]]",
        6670=>"PROJCS[\"JGD_2011_Japan_Zone_2\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",131.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",33.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6670]]",
        6671=>"PROJCS[\"JGD_2011_Japan_Zone_3\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",132.1666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6671]]",
        6672=>"PROJCS[\"JGD_2011_Japan_Zone_4\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",133.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",33.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6672]]",
        6673=>"PROJCS[\"JGD_2011_Japan_Zone_5\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",134.3333333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6673]]",
        6674=>"PROJCS[\"JGD_2011_Japan_Zone_6\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",136.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6674]]",
        6675=>"PROJCS[\"JGD_2011_Japan_Zone_7\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",137.1666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6675]]",
        6676=>"PROJCS[\"JGD_2011_Japan_Zone_8\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",138.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6676]]",
        6677=>"PROJCS[\"JGD_2011_Japan_Zone_9\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",139.8333333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6677]]",
        6678=>"PROJCS[\"JGD_2011_Japan_Zone_10\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",140.8333333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6678]]",
        6679=>"PROJCS[\"JGD_2011_Japan_Zone_11\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",140.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6679]]",
        6680=>"PROJCS[\"JGD_2011_Japan_Zone_12\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",142.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6680]]",
        6681=>"PROJCS[\"JGD_2011_Japan_Zone_13\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",144.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6681]]",
        6682=>"PROJCS[\"JGD_2011_Japan_Zone_14\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",142.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6682]]",
        6683=>"PROJCS[\"JGD_2011_Japan_Zone_15\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",127.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6683]]",
        6684=>"PROJCS[\"JGD_2011_Japan_Zone_16\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",124.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6684]]",
        6685=>"PROJCS[\"JGD_2011_Japan_Zone_17\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",131.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6685]]",
        6686=>"PROJCS[\"JGD_2011_Japan_Zone_18\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",136.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",20.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6686]]",
        6687=>"PROJCS[\"JGD_2011_Japan_Zone_19\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",154.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6687]]",
        6688=>"PROJCS[\"JGD_2011_UTM_Zone_51N\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6688]]",
        6689=>"PROJCS[\"JGD_2011_UTM_Zone_52N\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6689]]",
        6690=>"PROJCS[\"JGD_2011_UTM_Zone_53N\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6690]]",
        6691=>"PROJCS[\"JGD_2011_UTM_Zone_54N\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6691]]",
        6692=>"PROJCS[\"JGD_2011_UTM_Zone_55N\",GEOGCS[\"GCS_JGD_2011\",DATUM[\"D_JGD_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6692]]",
        6703=>"PROJCS[\"WGS_1984_TM_60_SW\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-60.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6703]]",
        6707=>"PROJCS[\"RDN2008_TM32\",GEOGCS[\"GCS_RDN2008\",DATUM[\"D_Rete_Dinamica_Nazionale_2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6707]]",
        6708=>"PROJCS[\"RDN2008_TM33\",GEOGCS[\"GCS_RDN2008\",DATUM[\"D_Rete_Dinamica_Nazionale_2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6708]]",
        6709=>"PROJCS[\"RDN2008_TM34\",GEOGCS[\"GCS_RDN2008\",DATUM[\"D_Rete_Dinamica_Nazionale_2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6709]]",
        6720=>"PROJCS[\"WGS_1984_CIG92\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",1300000.0],PARAMETER[\"Central_Meridian\",105.625],PARAMETER[\"Scale_Factor\",1.000024],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6720]]",
        6721=>"PROJCS[\"GDA_1994_CIG94\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",1300000.0],PARAMETER[\"Central_Meridian\",105.625],PARAMETER[\"Scale_Factor\",1.00002514],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6721]]",
        6722=>"PROJCS[\"WGS_1984_CKIG92\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",1400000.0],PARAMETER[\"Central_Meridian\",96.875],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6722]]",
        6723=>"PROJCS[\"GDA_1994_CKIG94\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",96.875],PARAMETER[\"Scale_Factor\",0.99999387],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6723]]",
        6732=>"PROJCS[\"GDA_1994_MGA_zone_41\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6732]]",
        6733=>"PROJCS[\"GDA_1994_MGA_zone_42\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6733]]",
        6734=>"PROJCS[\"GDA_1994_MGA_zone_43\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6734]]",
        6735=>"PROJCS[\"GDA_1994_MGA_zone_44\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6735]]",
        6736=>"PROJCS[\"GDA_1994_MGA_Zone_46\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6736]]",
        6737=>"PROJCS[\"GDA_1994_MGA_Zone_47\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6737]]",
        6738=>"PROJCS[\"GDA_1994_MGA_Zone_59\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6738]]",
        6784=>"PROJCS[\"OCRS_Baker_NAD_1983_CORS96_TM_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",40000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.8333333333333],PARAMETER[\"Scale_Factor\",1.00016],PARAMETER[\"Latitude_Of_Origin\",44.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6784]]",
        6785=>"PROJCS[\"OCRS_Baker_NAD_1983_CORS96_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",131233.5958005249],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.8333333333333],PARAMETER[\"Scale_Factor\",1.00016],PARAMETER[\"Latitude_Of_Origin\",44.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6785]]",
        6786=>"PROJCS[\"OCRS_Baker_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",40000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.8333333333333],PARAMETER[\"Scale_Factor\",1.00016],PARAMETER[\"Latitude_Of_Origin\",44.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6786]]",
        6787=>"PROJCS[\"OCRS_Baker_NAD_1983_2011_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",131233.5958005249],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.8333333333333],PARAMETER[\"Scale_Factor\",1.00016],PARAMETER[\"Latitude_Of_Origin\",44.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6787]]",
        6788=>"PROJCS[\"OCRS_Bend-Klamath_Falls_NAD_1983_CORS96_TM_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",80000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-121.75],PARAMETER[\"Scale_Factor\",1.0002],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6788]]",
        6789=>"PROJCS[\"OCRS_Bend-Klamath_Falls_NAD_1983_CORS96_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",262467.1916010499],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-121.75],PARAMETER[\"Scale_Factor\",1.0002],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6789]]",
        6790=>"PROJCS[\"OCRS_Bend-Klamath_Falls_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",80000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-121.75],PARAMETER[\"Scale_Factor\",1.0002],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6790]]",
        6791=>"PROJCS[\"OCRS_Bend-Klamath_Falls_NAD_1983_2011_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",262467.1916010499],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-121.75],PARAMETER[\"Scale_Factor\",1.0002],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6791]]",
        6792=>"PROJCS[\"OCRS_Bend-Redmond-Prineville_NAD_1983_CORS96_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",80000.0],PARAMETER[\"False_Northing\",130000.0],PARAMETER[\"Central_Meridian\",-121.25],PARAMETER[\"Standard_Parallel_1\",44.66666666666666],PARAMETER[\"Scale_Factor\",1.00012],PARAMETER[\"Latitude_Of_Origin\",44.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6792]]",
        6793=>"PROJCS[\"OCRS_Bend-Redmond-Prineville_NAD_1983_CORS96_LCC_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",262467.1916010499],PARAMETER[\"False_Northing\",426509.186351706],PARAMETER[\"Central_Meridian\",-121.25],PARAMETER[\"Standard_Parallel_1\",44.66666666666666],PARAMETER[\"Scale_Factor\",1.00012],PARAMETER[\"Latitude_Of_Origin\",44.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6793]]",
        6794=>"PROJCS[\"OCRS_Bend-Redmond-Prineville_NAD_1983_2011_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",80000.0],PARAMETER[\"False_Northing\",130000.0],PARAMETER[\"Central_Meridian\",-121.25],PARAMETER[\"Standard_Parallel_1\",44.66666666666666],PARAMETER[\"Scale_Factor\",1.00012],PARAMETER[\"Latitude_Of_Origin\",44.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6794]]",
        6795=>"PROJCS[\"OCRS_Bend-Redmond-Prineville_NAD_1983_2011_LCC_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",262467.1916010499],PARAMETER[\"False_Northing\",426509.186351706],PARAMETER[\"Central_Meridian\",-121.25],PARAMETER[\"Standard_Parallel_1\",44.66666666666666],PARAMETER[\"Scale_Factor\",1.00012],PARAMETER[\"Latitude_Of_Origin\",44.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6795]]",
        6796=>"PROJCS[\"OCRS_Bend-Burns_NAD_1983_CORS96_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",120000.0],PARAMETER[\"False_Northing\",60000.0],PARAMETER[\"Central_Meridian\",-119.75],PARAMETER[\"Standard_Parallel_1\",43.66666666666666],PARAMETER[\"Scale_Factor\",1.0002],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6796]]",
        6797=>"PROJCS[\"OCRS_Bend-Burns_NAD_1983_CORS96_LCC_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",393700.7874015748],PARAMETER[\"False_Northing\",196850.3937007874],PARAMETER[\"Central_Meridian\",-119.75],PARAMETER[\"Standard_Parallel_1\",43.66666666666666],PARAMETER[\"Scale_Factor\",1.0002],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6797]]",
        6798=>"PROJCS[\"OCRS_Bend-Burns_NAD_1983_2011_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",120000.0],PARAMETER[\"False_Northing\",60000.0],PARAMETER[\"Central_Meridian\",-119.75],PARAMETER[\"Standard_Parallel_1\",43.66666666666666],PARAMETER[\"Scale_Factor\",1.0002],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6798]]",
        6799=>"PROJCS[\"OCRS_Bend-Burns_NAD_1983_2011_LCC_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",393700.7874015748],PARAMETER[\"False_Northing\",196850.3937007874],PARAMETER[\"Central_Meridian\",-119.75],PARAMETER[\"Standard_Parallel_1\",43.66666666666666],PARAMETER[\"Scale_Factor\",1.0002],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6799]]",
        6800=>"PROJCS[\"OCRS_Canyonville-Grants_Pass_NAD_1983_CORS96_TM_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",40000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.3333333333333],PARAMETER[\"Scale_Factor\",1.00007],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6800]]",
        6801=>"PROJCS[\"OCRS_Canyonville-Grants_Pass_NAD_1983_CORS96_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",131233.5958005249],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.3333333333333],PARAMETER[\"Scale_Factor\",1.00007],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6801]]",
        6802=>"PROJCS[\"OCRS_Canyonville-Grants_Pass_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",40000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.3333333333333],PARAMETER[\"Scale_Factor\",1.00007],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6802]]",
        6803=>"PROJCS[\"OCRS_Canyonville-Grants_Pass_NAD_1983_2011_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",131233.5958005249],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.3333333333333],PARAMETER[\"Scale_Factor\",1.00007],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6803]]",
        6804=>"PROJCS[\"OCRS_Columbia_River_East_NAD_1983_CORS96_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",30000.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",45.66666666666666],PARAMETER[\"Scale_Factor\",1.000008],PARAMETER[\"Latitude_Of_Origin\",45.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6804]]",
        6805=>"PROJCS[\"OCRS_Columbia_River_East_NAD_1983_CORS96_LCC_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",492125.9842519685],PARAMETER[\"False_Northing\",98425.1968503937],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",45.66666666666666],PARAMETER[\"Scale_Factor\",1.000008],PARAMETER[\"Latitude_Of_Origin\",45.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6805]]",
        6806=>"PROJCS[\"OCRS_Columbia_River_East_NAD_1983_2011_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",30000.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",45.66666666666666],PARAMETER[\"Scale_Factor\",1.000008],PARAMETER[\"Latitude_Of_Origin\",45.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6806]]",
        6807=>"PROJCS[\"OCRS_Columbia_River_East_NAD_1983_2011_LCC_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",492125.9842519685],PARAMETER[\"False_Northing\",98425.1968503937],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",45.66666666666666],PARAMETER[\"Scale_Factor\",1.000008],PARAMETER[\"Latitude_Of_Origin\",45.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6807]]",
        6808=>"PROJCS[\"OCRS_Columbia_River_West_NAD_1983_CORS96_OM_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Natural_Origin\"],PARAMETER[\"False_Easting\",7000000.0],PARAMETER[\"False_Northing\",-3000000.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Azimuth\",-65.0],PARAMETER[\"Longitude_Of_Center\",-123.0],PARAMETER[\"Latitude_Of_Center\",45.91666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6808]]",
        6809=>"PROJCS[\"OCRS_Columbia_River_West_NAD_1983_CORS96_OM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Natural_Origin\"],PARAMETER[\"False_Easting\",22965879.26509186],PARAMETER[\"False_Northing\",-9842519.685039369],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Azimuth\",-65.0],PARAMETER[\"Longitude_Of_Center\",-123.0],PARAMETER[\"Latitude_Of_Center\",45.91666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6809]]",
        6810=>"PROJCS[\"OCRS_Columbia_River_West_NAD_1983_2011_OM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Natural_Origin\"],PARAMETER[\"False_Easting\",7000000.0],PARAMETER[\"False_Northing\",-3000000.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Azimuth\",-65.0],PARAMETER[\"Longitude_Of_Center\",-123.0],PARAMETER[\"Latitude_Of_Center\",45.91666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6810]]",
        6811=>"PROJCS[\"OCRS_Columbia_River_West_NAD_1983_2011_OM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Natural_Origin\"],PARAMETER[\"False_Easting\",22965879.26509186],PARAMETER[\"False_Northing\",-9842519.685039369],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Azimuth\",-65.0],PARAMETER[\"Longitude_Of_Center\",-123.0],PARAMETER[\"Latitude_Of_Center\",45.91666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6811]]",
        6812=>"PROJCS[\"OCRS_Cottage_Grove-Canyonville_NAD_1983_CORS96_TM_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.3333333333333],PARAMETER[\"Scale_Factor\",1.000023],PARAMETER[\"Latitude_Of_Origin\",42.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6812]]",
        6813=>"PROJCS[\"OCRS_Cottage_Grove-Canyonville_NAD_1983_CORS96_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",164041.9947506562],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.3333333333333],PARAMETER[\"Scale_Factor\",1.000023],PARAMETER[\"Latitude_Of_Origin\",42.83333333333334],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6813]]",
        6814=>"PROJCS[\"OCRS_Cottage_Grove-Canyonville_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.3333333333333],PARAMETER[\"Scale_Factor\",1.000023],PARAMETER[\"Latitude_Of_Origin\",42.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6814]]",
        6815=>"PROJCS[\"OCRS_Cottage_Grove-Canyonville_NAD_1983_2011_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",164041.9947506562],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.3333333333333],PARAMETER[\"Scale_Factor\",1.000023],PARAMETER[\"Latitude_Of_Origin\",42.83333333333334],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6815]]",
        6816=>"PROJCS[\"OCRS_Dufur-Madras_NAD_1983_CORS96_TM_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",80000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-121.0],PARAMETER[\"Scale_Factor\",1.00011],PARAMETER[\"Latitude_Of_Origin\",44.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6816]]",
        6817=>"PROJCS[\"OCRS_Dufur-Madras_NAD_1983_CORS96_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",262467.1916010499],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-121.0],PARAMETER[\"Scale_Factor\",1.00011],PARAMETER[\"Latitude_Of_Origin\",44.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6817]]",
        6818=>"PROJCS[\"OCRS_Dufur-Madras_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",80000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-121.0],PARAMETER[\"Scale_Factor\",1.00011],PARAMETER[\"Latitude_Of_Origin\",44.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6818]]",
        6819=>"PROJCS[\"OCRS_Dufur-Madras_NAD_1983_2011_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",262467.1916010499],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-121.0],PARAMETER[\"Scale_Factor\",1.00011],PARAMETER[\"Latitude_Of_Origin\",44.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6819]]",
        6820=>"PROJCS[\"OCRS_Eugene_NAD_1983_CORS96_TM_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.1666666666667],PARAMETER[\"Scale_Factor\",1.000015],PARAMETER[\"Latitude_Of_Origin\",43.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6820]]",
        6821=>"PROJCS[\"OCRS_Eugene_NAD_1983_CORS96_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",164041.9947506562],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.1666666666667],PARAMETER[\"Scale_Factor\",1.000015],PARAMETER[\"Latitude_Of_Origin\",43.75],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6821]]",
        6822=>"PROJCS[\"OCRS_Eugene_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.1666666666667],PARAMETER[\"Scale_Factor\",1.000015],PARAMETER[\"Latitude_Of_Origin\",43.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6822]]",
        6823=>"PROJCS[\"OCRS_Eugene_NAD_1983_2011_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",164041.9947506562],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.1666666666667],PARAMETER[\"Scale_Factor\",1.000015],PARAMETER[\"Latitude_Of_Origin\",43.75],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6823]]",
        6824=>"PROJCS[\"OCRS_Grants_Pass-Ashland_NAD_1983_CORS96_TM_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.3333333333333],PARAMETER[\"Scale_Factor\",1.000043],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6824]]",
        6825=>"PROJCS[\"OCRS_Grants_Pass-Ashland_NAD_1983_CORS96_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",164041.9947506562],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.3333333333333],PARAMETER[\"Scale_Factor\",1.000043],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6825]]",
        6826=>"PROJCS[\"OCRS_Grants_Pass-Ashland_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.3333333333333],PARAMETER[\"Scale_Factor\",1.000043],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6826]]",
        6827=>"PROJCS[\"OCRS_Grants_Pass-Ashland_NAD_1983_2011_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",164041.9947506562],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.3333333333333],PARAMETER[\"Scale_Factor\",1.000043],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6827]]",
        6828=>"PROJCS[\"OCRS_Gresham-Warm_Springs_NAD_1983_CORS96_TM_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",10000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-122.3333333333333],PARAMETER[\"Scale_Factor\",1.00005],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6828]]",
        6829=>"PROJCS[\"OCRS_Gresham-Warm_Springs_NAD_1983_CORS96_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",32808.39895013123],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-122.3333333333333],PARAMETER[\"Scale_Factor\",1.00005],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6829]]",
        6830=>"PROJCS[\"OCRS_Gresham-Warm_Springs_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",10000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-122.3333333333333],PARAMETER[\"Scale_Factor\",1.00005],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6830]]",
        6831=>"PROJCS[\"OCRS_Gresham-Warm_Springs_NAD_1983_2011_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",32808.39895013123],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-122.3333333333333],PARAMETER[\"Scale_Factor\",1.00005],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6831]]",
        6832=>"PROJCS[\"OCRS_La_Grande_NAD_1983_CORS96_TM_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",40000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-118.0],PARAMETER[\"Scale_Factor\",1.00013],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6832]]",
        6833=>"PROJCS[\"OCRS_La_Grande_NAD_1983_CORS96_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",131233.5958005249],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-118.0],PARAMETER[\"Scale_Factor\",1.00013],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6833]]",
        6834=>"PROJCS[\"OCRS_La_Grande_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",40000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-118.0],PARAMETER[\"Scale_Factor\",1.00013],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6834]]",
        6835=>"PROJCS[\"OCRS_La_Grande_NAD_1983_2011_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",131233.5958005249],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-118.0],PARAMETER[\"Scale_Factor\",1.00013],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6835]]",
        6836=>"PROJCS[\"OCRS_Ontario_NAD_1983_CORS96_TM_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",80000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",1.0001],PARAMETER[\"Latitude_Of_Origin\",43.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6836]]",
        6837=>"PROJCS[\"OCRS_Ontario_NAD_1983_CORS96_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",262467.1916010499],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",1.0001],PARAMETER[\"Latitude_Of_Origin\",43.25],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6837]]",
        6838=>"PROJCS[\"OCRS_Ontario_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",80000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",1.0001],PARAMETER[\"Latitude_Of_Origin\",43.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6838]]",
        6839=>"PROJCS[\"OCRS_Ontario_NAD_1983_2011_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",262467.1916010499],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",1.0001],PARAMETER[\"Latitude_Of_Origin\",43.25],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6839]]",
        6840=>"PROJCS[\"OCRS_Oregon_Coast_NAD_1983_CORS96_OM_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Natural_Origin\"],PARAMETER[\"False_Easting\",-300000.0],PARAMETER[\"False_Northing\",-4600000.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Azimuth\",5.0],PARAMETER[\"Longitude_Of_Center\",-124.05],PARAMETER[\"Latitude_Of_Center\",44.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6840]]",
        6841=>"PROJCS[\"OCRS_Oregon_Coast_NAD_1983_CORS96_OM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Natural_Origin\"],PARAMETER[\"False_Easting\",-984251.968503937],PARAMETER[\"False_Northing\",-15091863.51706037],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Azimuth\",5.0],PARAMETER[\"Longitude_Of_Center\",-124.05],PARAMETER[\"Latitude_Of_Center\",44.75],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6841]]",
        6842=>"PROJCS[\"OCRS_Oregon_Coast_NAD_1983_2011_OM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Natural_Origin\"],PARAMETER[\"False_Easting\",-300000.0],PARAMETER[\"False_Northing\",-4600000.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Azimuth\",5.0],PARAMETER[\"Longitude_Of_Center\",-124.05],PARAMETER[\"Latitude_Of_Center\",44.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6842]]",
        6843=>"PROJCS[\"OCRS_Oregon_Coast_NAD_1983_2011_OM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Natural_Origin\"],PARAMETER[\"False_Easting\",-984251.968503937],PARAMETER[\"False_Northing\",-15091863.51706037],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Azimuth\",5.0],PARAMETER[\"Longitude_Of_Center\",-124.05],PARAMETER[\"Latitude_Of_Center\",44.75],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6843]]",
        6844=>"PROJCS[\"OCRS_Pendleton_NAD_1983_CORS96_TM_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",60000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-119.1666666666667],PARAMETER[\"Scale_Factor\",1.000045],PARAMETER[\"Latitude_Of_Origin\",45.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6844]]",
        6845=>"PROJCS[\"OCRS_Pendleton_NAD_1983_CORS96_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",196850.3937007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-119.1666666666667],PARAMETER[\"Scale_Factor\",1.000045],PARAMETER[\"Latitude_Of_Origin\",45.25],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6845]]",
        6846=>"PROJCS[\"OCRS_Pendleton_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",60000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-119.1666666666667],PARAMETER[\"Scale_Factor\",1.000045],PARAMETER[\"Latitude_Of_Origin\",45.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6846]]",
        6847=>"PROJCS[\"OCRS_Pendleton_NAD_1983_2011_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",196850.3937007874],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-119.1666666666667],PARAMETER[\"Scale_Factor\",1.000045],PARAMETER[\"Latitude_Of_Origin\",45.25],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6847]]",
        6848=>"PROJCS[\"OCRS_Pendleton-La_Grande_NAD_1983_CORS96_TM_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",30000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-118.3333333333333],PARAMETER[\"Scale_Factor\",1.000175],PARAMETER[\"Latitude_Of_Origin\",45.08333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6848]]",
        6849=>"PROJCS[\"OCRS_Pendleton-La_Grande_NAD_1983_CORS96_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",98425.1968503937],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-118.3333333333333],PARAMETER[\"Scale_Factor\",1.000175],PARAMETER[\"Latitude_Of_Origin\",45.08333333333334],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6849]]",
        6850=>"PROJCS[\"OCRS_Pendleton-La_Grande_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",30000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-118.3333333333333],PARAMETER[\"Scale_Factor\",1.000175],PARAMETER[\"Latitude_Of_Origin\",45.08333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6850]]",
        6851=>"PROJCS[\"OCRS_Pendleton-La_Grande_NAD_1983_2011_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",98425.1968503937],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-118.3333333333333],PARAMETER[\"Scale_Factor\",1.000175],PARAMETER[\"Latitude_Of_Origin\",45.08333333333334],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6851]]",
        6852=>"PROJCS[\"OCRS_Portland_NAD_1983_CORS96_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",50000.0],PARAMETER[\"Central_Meridian\",-122.75],PARAMETER[\"Standard_Parallel_1\",45.5],PARAMETER[\"Scale_Factor\",1.000002],PARAMETER[\"Latitude_Of_Origin\",45.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6852]]",
        6853=>"PROJCS[\"OCRS_Portland_NAD_1983_CORS96_LCC_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",328083.9895013123],PARAMETER[\"False_Northing\",164041.9947506562],PARAMETER[\"Central_Meridian\",-122.75],PARAMETER[\"Standard_Parallel_1\",45.5],PARAMETER[\"Scale_Factor\",1.000002],PARAMETER[\"Latitude_Of_Origin\",45.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6853]]",
        6854=>"PROJCS[\"OCRS_Portland_NAD_1983_2011_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",50000.0],PARAMETER[\"Central_Meridian\",-122.75],PARAMETER[\"Standard_Parallel_1\",45.5],PARAMETER[\"Scale_Factor\",1.000002],PARAMETER[\"Latitude_Of_Origin\",45.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6854]]",
        6855=>"PROJCS[\"OCRS_Portland_NAD_1983_2011_LCC_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",328083.9895013123],PARAMETER[\"False_Northing\",164041.9947506562],PARAMETER[\"Central_Meridian\",-122.75],PARAMETER[\"Standard_Parallel_1\",45.5],PARAMETER[\"Scale_Factor\",1.000002],PARAMETER[\"Latitude_Of_Origin\",45.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6855]]",
        6856=>"PROJCS[\"OCRS_Salem_NAD_1983_CORS96_TM_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.0833333333333],PARAMETER[\"Scale_Factor\",1.00001],PARAMETER[\"Latitude_Of_Origin\",44.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6856]]",
        6857=>"PROJCS[\"OCRS_Salem_NAD_1983_CORS96_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",164041.9947506562],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.0833333333333],PARAMETER[\"Scale_Factor\",1.00001],PARAMETER[\"Latitude_Of_Origin\",44.33333333333334],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6857]]",
        6858=>"PROJCS[\"OCRS_Salem_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.0833333333333],PARAMETER[\"Scale_Factor\",1.00001],PARAMETER[\"Latitude_Of_Origin\",44.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6858]]",
        6859=>"PROJCS[\"OCRS_Salem_NAD_1983_2011_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",164041.9947506562],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.0833333333333],PARAMETER[\"Scale_Factor\",1.00001],PARAMETER[\"Latitude_Of_Origin\",44.33333333333334],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6859]]",
        6860=>"PROJCS[\"OCRS_Santiam_Pass_NAD_1983_CORS96_TM_Meters\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-122.5],PARAMETER[\"Scale_Factor\",1.000155],PARAMETER[\"Latitude_Of_Origin\",44.08333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6860]]",
        6861=>"PROJCS[\"OCRS_Santiam_Pass_NAD_1983_CORS96_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-122.5],PARAMETER[\"Scale_Factor\",1.000155],PARAMETER[\"Latitude_Of_Origin\",44.08333333333334],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6861]]",
        6862=>"PROJCS[\"OCRS_Santiam_Pass_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-122.5],PARAMETER[\"Scale_Factor\",1.000155],PARAMETER[\"Latitude_Of_Origin\",44.08333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6862]]",
        6863=>"PROJCS[\"OCRS_Santiam_Pass_NAD_1983_2011_TM_Feet_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-122.5],PARAMETER[\"Scale_Factor\",1.000155],PARAMETER[\"Latitude_Of_Origin\",44.08333333333334],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6863]]",
        6867=>"PROJCS[\"NAD_1983_CORS96_Oregon_Statewide_Lambert\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",43.0],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6867]]",
        6868=>"PROJCS[\"NAD_1983_CORS96_Oregon_Statewide_Lambert_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312335.958005249],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",43.0],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6868]]",
        6870=>"PROJCS[\"ETRS_1989_Albania_2010\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",20.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6870]]",
        6875=>"PROJCS[\"RDN2008_Italy_zone\",GEOGCS[\"GCS_RDN2008\",DATUM[\"D_Rete_Dinamica_Nazionale_2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",7000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",0.9985],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6875]]",
        6876=>"PROJCS[\"RDN2008_Zone_12\",GEOGCS[\"GCS_RDN2008\",DATUM[\"D_Rete_Dinamica_Nazionale_2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",3000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6876]]",
        6879=>"PROJCS[\"NAD_1983_2011_StatePlane_Wisconsin_Central_FIPS_4802\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",44.25],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6879]]",
        6880=>"PROJCS[\"NAD_1983_2011_StatePlane_Nebraska_FIPS_2600_Ft_US\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Standard_Parallel_2\",43.0],PARAMETER[\"Latitude_Of_Origin\",39.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6880]]",
        6884=>"PROJCS[\"NAD_1983_CORS96_StatePlane_Oregon_North_FIPS_3601\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",44.33333333333334],PARAMETER[\"Standard_Parallel_2\",46.0],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6884]]",
        6885=>"PROJCS[\"NAD_1983_CORS96_StatePlane_Oregon_North_FIPS_3601_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",8202099.737532808],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",44.33333333333334],PARAMETER[\"Standard_Parallel_2\",46.0],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6885]]",
        6886=>"PROJCS[\"NAD_1983_CORS96_StatePlane_Oregon_South_FIPS_3602\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",42.33333333333334],PARAMETER[\"Standard_Parallel_2\",44.0],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6886]]",
        6887=>"PROJCS[\"NAD_1983_CORS96_StatePlane_Oregon_South_FIPS_3602_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_CORS96\",DATUM[\"D_NAD_1983_CORS96\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4921259.842519685],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",42.33333333333334],PARAMETER[\"Standard_Parallel_2\",44.0],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",6887]]",
        6915=>"PROJCS[\"South_East_Island_1943_UTM_Zone_40N\",GEOGCS[\"GCS_South_East_Island_1943\",DATUM[\"D_South_East_Island_1943\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6915]]",
        6922=>"PROJCS[\"NAD_1983_Kansas_LCC\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.25],PARAMETER[\"Standard_Parallel_1\",37.5],PARAMETER[\"Standard_Parallel_2\",39.5],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6922]]",
        6923=>"PROJCS[\"NAD_1983_Kansas_LCC_ftUS\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.3333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.25],PARAMETER[\"Standard_Parallel_1\",37.5],PARAMETER[\"Standard_Parallel_2\",39.5],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6923]]",
        6924=>"PROJCS[\"NAD_1983_2011_Kansas_LCC\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.25],PARAMETER[\"Standard_Parallel_1\",37.5],PARAMETER[\"Standard_Parallel_2\",39.5],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6924]]",
        6925=>"PROJCS[\"NAD_1983_2011_Kansas_LCC_ftUS\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1312333.3333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.25],PARAMETER[\"Standard_Parallel_1\",37.5],PARAMETER[\"Standard_Parallel_2\",39.5],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",6925]]",
        6931=>"PROJCS[\"WGS_1984_EASE-Grid_2.0_North\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Latitude_Of_Origin\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6931]]",
        6932=>"PROJCS[\"WGS_1984_EASE-Grid_2.0_South\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Azimuthal_Equal_Area\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6932]]",
        6933=>"PROJCS[\"WGS_1984_EASE-Grid_2.0_Global\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cylindrical_Equal_Area\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6933]]",
        6956=>"PROJCS[\"VN_2000_TM-3_zone_481\",GEOGCS[\"GCS_VN_2000\",DATUM[\"D_Vietnam_2000\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",102.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6956]]",
        6957=>"PROJCS[\"VN_2000_TM-3_zone_482\",GEOGCS[\"GCS_VN_2000\",DATUM[\"D_Vietnam_2000\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6957]]",
        6958=>"PROJCS[\"VN_2000_TM-3_zone_491\",GEOGCS[\"GCS_VN_2000\",DATUM[\"D_Vietnam_2000\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",108.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6958]]",
        6959=>"PROJCS[\"VN_2000_TM-3_Da_Nang_zone\",GEOGCS[\"GCS_VN_2000\",DATUM[\"D_Vietnam_2000\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",107.75],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6959]]",
        6962=>"PROJCS[\"ETRS_1989_Albania_LCC_2010\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",20.0],PARAMETER[\"Standard_Parallel_1\",39.0],PARAMETER[\"Standard_Parallel_2\",43.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6962]]",
        6984=>"PROJCS[\"Israeli_Grid_05\",GEOGCS[\"IG05_Intermediate_CRS\",DATUM[\"IG05_Intermediate_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",219529.584],PARAMETER[\"False_Northing\",626907.39],PARAMETER[\"Central_Meridian\",35.20451694444444],PARAMETER[\"Scale_Factor\",1.0000067],PARAMETER[\"Latitude_Of_Origin\",31.73439361111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6984]]",
        6991=>"PROJCS[\"Israeli_Grid_05-12\",GEOGCS[\"IG05(2012)_Intermediate_CRS\",DATUM[\"IG05(2012)_Intermediate_Datum\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",219529.584],PARAMETER[\"False_Northing\",626907.39],PARAMETER[\"Central_Meridian\",35.20451694444444],PARAMETER[\"Scale_Factor\",1.0000067],PARAMETER[\"Latitude_Of_Origin\",31.73439361111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",6991]]",
        7005=>"PROJCS[\"Nahrwan_1934_UTM_zone_37N\",GEOGCS[\"GCS_Nahrwan_1934\",DATUM[\"D_Nahrwan_1934\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7005]]",
        7006=>"PROJCS[\"Nahrwan_1934_UTM_zone_38N\",GEOGCS[\"GCS_Nahrwan_1934\",DATUM[\"D_Nahrwan_1934\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7006]]",
        7007=>"PROJCS[\"Nahrwan_1934_UTM_zone_39N\",GEOGCS[\"GCS_Nahrwan_1934\",DATUM[\"D_Nahrwan_1934\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7007]]",
        7057=>"PROJCS[\"NAD_1983_(2011)_IaRCS_zone_1\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",11500000.0],PARAMETER[\"False_Northing\",9600000.0],PARAMETER[\"Central_Meridian\",-95.25],PARAMETER[\"Standard_Parallel_1\",43.2],PARAMETER[\"Scale_Factor\",1.000052],PARAMETER[\"Latitude_Of_Origin\",43.2],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7057]]",
        7058=>"PROJCS[\"NAD_1983_(2011)_IaRCS_zone_2\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",12500000.0],PARAMETER[\"False_Northing\",9800000.0],PARAMETER[\"Central_Meridian\",-92.75],PARAMETER[\"Standard_Parallel_1\",43.16666666666666],PARAMETER[\"Scale_Factor\",1.000043],PARAMETER[\"Latitude_Of_Origin\",43.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7058]]",
        7059=>"PROJCS[\"NAD_1983_(2011)_IaRCS_zone_3\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",13500000.0],PARAMETER[\"False_Northing\",8300000.0],PARAMETER[\"Central_Meridian\",-91.2],PARAMETER[\"Scale_Factor\",1.000035],PARAMETER[\"Latitude_Of_Origin\",40.25],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7059]]",
        7060=>"PROJCS[\"NAD_1983_(2011)_IaRCS_zone_4\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",14500000.0],PARAMETER[\"False_Northing\",8600000.0],PARAMETER[\"Central_Meridian\",-94.83333333333333],PARAMETER[\"Standard_Parallel_1\",42.53333333333333],PARAMETER[\"Scale_Factor\",1.000045],PARAMETER[\"Latitude_Of_Origin\",42.53333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7060]]",
        7061=>"PROJCS[\"NAD_1983_(2011)_IaRCS_zone_5\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",15500000.0],PARAMETER[\"False_Northing\",8900000.0],PARAMETER[\"Central_Meridian\",-92.25],PARAMETER[\"Standard_Parallel_1\",42.65],PARAMETER[\"Scale_Factor\",1.000032],PARAMETER[\"Latitude_Of_Origin\",42.65],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7061]]",
        7062=>"PROJCS[\"NAD_1983_(2011)_IaRCS_zone_6\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",16500000.0],PARAMETER[\"False_Northing\",6600000.0],PARAMETER[\"Central_Meridian\",-95.73333333333333],PARAMETER[\"Scale_Factor\",1.000039],PARAMETER[\"Latitude_Of_Origin\",40.25],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7062]]",
        7063=>"PROJCS[\"NAD_1983_(2011)_IaRCS_zone_7\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",17500000.0],PARAMETER[\"False_Northing\",6800000.0],PARAMETER[\"Central_Meridian\",-94.63333333333334],PARAMETER[\"Scale_Factor\",1.000045],PARAMETER[\"Latitude_Of_Origin\",40.25],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7063]]",
        7064=>"PROJCS[\"NAD_1983_(2011)_IaRCS_zone_8\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",18500000.0],PARAMETER[\"False_Northing\",7000000.0],PARAMETER[\"Central_Meridian\",-93.71666666666667],PARAMETER[\"Scale_Factor\",1.000033],PARAMETER[\"Latitude_Of_Origin\",40.25],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7064]]",
        7065=>"PROJCS[\"NAD_1983_(2011)_IaRCS_zone_9\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",19500000.0],PARAMETER[\"False_Northing\",7200000.0],PARAMETER[\"Central_Meridian\",-92.81666666666666],PARAMETER[\"Scale_Factor\",1.000027],PARAMETER[\"Latitude_Of_Origin\",40.25],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7065]]",
        7066=>"PROJCS[\"NAD_1983_(2011)_IaRCS_zone_10\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",20500000.0],PARAMETER[\"False_Northing\",8000000.0],PARAMETER[\"Central_Meridian\",-91.66666666666667],PARAMETER[\"Standard_Parallel_1\",41.83333333333334],PARAMETER[\"Scale_Factor\",1.00002],PARAMETER[\"Latitude_Of_Origin\",41.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7066]]",
        7067=>"PROJCS[\"NAD_1983_(2011)_IaRCS_zone_11\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",21500000.0],PARAMETER[\"False_Northing\",7600000.0],PARAMETER[\"Central_Meridian\",-90.53333333333333],PARAMETER[\"Scale_Factor\",1.000027],PARAMETER[\"Latitude_Of_Origin\",40.25],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7067]]",
        7068=>"PROJCS[\"NAD_1983_(2011)_IaRCS_zone_12\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",22500000.0],PARAMETER[\"False_Northing\",6200000.0],PARAMETER[\"Central_Meridian\",-93.75],PARAMETER[\"Standard_Parallel_1\",40.91666666666666],PARAMETER[\"Scale_Factor\",1.000037],PARAMETER[\"Latitude_Of_Origin\",40.91666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7068]]",
        7069=>"PROJCS[\"NAD_1983_(2011)_IaRCS_zone_13\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",23500000.0],PARAMETER[\"False_Northing\",6400000.0],PARAMETER[\"Central_Meridian\",-91.91666666666667],PARAMETER[\"Scale_Factor\",1.00002],PARAMETER[\"Latitude_Of_Origin\",40.25],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7069]]",
        7070=>"PROJCS[\"NAD_1983_(2011)_IaRCS_zone_14\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",24500000.0],PARAMETER[\"False_Northing\",6200000.0],PARAMETER[\"Central_Meridian\",-91.25],PARAMETER[\"Scale_Factor\",1.000018],PARAMETER[\"Latitude_Of_Origin\",40.25],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7070]]",
        7074=>"PROJCS[\"RGTAAF07_UTM_zone_37S\",GEOGCS[\"GCS_RGTAAF07\",DATUM[\"D_Reseau_Geodesique_des_Terres_Australes_et_Antarctiques_Francaises_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7074]]",
        7075=>"PROJCS[\"RGTAAF07_UTM_zone_38S\",GEOGCS[\"GCS_RGTAAF07\",DATUM[\"D_Reseau_Geodesique_des_Terres_Australes_et_Antarctiques_Francaises_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7075]]",
        7076=>"PROJCS[\"RGTAAF07_UTM_zone_39S\",GEOGCS[\"GCS_RGTAAF07\",DATUM[\"D_Reseau_Geodesique_des_Terres_Australes_et_Antarctiques_Francaises_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7076]]",
        7077=>"PROJCS[\"RGTAAF07_UTM_zone_40S\",GEOGCS[\"GCS_RGTAAF07\",DATUM[\"D_Reseau_Geodesique_des_Terres_Australes_et_Antarctiques_Francaises_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7077]]",
        7078=>"PROJCS[\"RGTAAF07_UTM_zone_41S\",GEOGCS[\"GCS_RGTAAF07\",DATUM[\"D_Reseau_Geodesique_des_Terres_Australes_et_Antarctiques_Francaises_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7078]]",
        7079=>"PROJCS[\"RGTAAF07_UTM_zone_42S\",GEOGCS[\"GCS_RGTAAF07\",DATUM[\"D_Reseau_Geodesique_des_Terres_Australes_et_Antarctiques_Francaises_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7079]]",
        7080=>"PROJCS[\"RGTAAF07_UTM_zone_43S\",GEOGCS[\"GCS_RGTAAF07\",DATUM[\"D_Reseau_Geodesique_des_Terres_Australes_et_Antarctiques_Francaises_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7080]]",
        7081=>"PROJCS[\"RGTAAF07_UTM_zone_44S\",GEOGCS[\"GCS_RGTAAF07\",DATUM[\"D_Reseau_Geodesique_des_Terres_Australes_et_Antarctiques_Francaises_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7081]]",
        7082=>"PROJCS[\"RGTAAF07_Terre_Adelie_Polar_Stereographic\",GEOGCS[\"GCS_RGTAAF07\",DATUM[\"D_Reseau_Geodesique_des_Terres_Australes_et_Antarctiques_Francaises_2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Polar_Stereographic_Variant_C\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Longitude_Of_Origin\",140.0],PARAMETER[\"Standard_Parallel_1\",-67.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7082]]",
        7109=>"PROJCS[\"NAD_1983_2011_RMTCRS_St_Mary_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-112.5],PARAMETER[\"Scale_Factor\",1.00016],PARAMETER[\"Latitude_Of_Origin\",48.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7109]]",
        7110=>"PROJCS[\"NAD_1983_2011_RMTCRS_Blackfeet_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-112.5],PARAMETER[\"Scale_Factor\",1.00019],PARAMETER[\"Latitude_Of_Origin\",48.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7110]]",
        7111=>"PROJCS[\"NAD_1983_2011_RMTCRS_Milk_River_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Standard_Parallel_1\",48.5],PARAMETER[\"Scale_Factor\",1.000145],PARAMETER[\"Latitude_Of_Origin\",48.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7111]]",
        7112=>"PROJCS[\"NAD_1983_2011_RMTCRS_Fort_Belknap_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",150000.0],PARAMETER[\"Central_Meridian\",-108.5],PARAMETER[\"Standard_Parallel_1\",48.5],PARAMETER[\"Scale_Factor\",1.00012],PARAMETER[\"Latitude_Of_Origin\",48.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7112]]",
        7113=>"PROJCS[\"NAD_1983_2011_RMTCRS_Fort_Peck_Assiniboine_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",48.33333333333334],PARAMETER[\"Scale_Factor\",1.00012],PARAMETER[\"Latitude_Of_Origin\",48.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7113]]",
        7114=>"PROJCS[\"NAD_1983_2011_RMTCRS_Fort_Peck_Sioux_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",50000.0],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",48.33333333333334],PARAMETER[\"Scale_Factor\",1.00009],PARAMETER[\"Latitude_Of_Origin\",48.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7114]]",
        7115=>"PROJCS[\"NAD_1983_2011_RMTCRS_Crow_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-107.75],PARAMETER[\"Scale_Factor\",1.000148],PARAMETER[\"Latitude_Of_Origin\",44.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7115]]",
        7116=>"PROJCS[\"NAD_1983_2011_RMTCRS_Bobcat_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-111.25],PARAMETER[\"Standard_Parallel_1\",46.25],PARAMETER[\"Scale_Factor\",1.000185],PARAMETER[\"Latitude_Of_Origin\",46.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7116]]",
        7117=>"PROJCS[\"NAD_1983_2011_RMTCRS_Billings_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",50000.0],PARAMETER[\"Central_Meridian\",-108.4166666666667],PARAMETER[\"Standard_Parallel_1\",45.78333333333333],PARAMETER[\"Scale_Factor\",1.0001515],PARAMETER[\"Latitude_Of_Origin\",45.78333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7117]]",
        7118=>"PROJCS[\"NAD_1983_2011_RMTCRS_Wind_River_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-108.3333333333333],PARAMETER[\"Scale_Factor\",1.00024],PARAMETER[\"Latitude_Of_Origin\",42.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7118]]",
        7119=>"PROJCS[\"NAD_1983_2011_RMTCRS_St_Mary_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",492125.9843],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-112.5],PARAMETER[\"Scale_Factor\",1.00016],PARAMETER[\"Latitude_Of_Origin\",48.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",7119]]",
        7120=>"PROJCS[\"NAD_1983_2011_RMTCRS_Blackfeet_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",328083.9895],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-112.5],PARAMETER[\"Scale_Factor\",1.00019],PARAMETER[\"Latitude_Of_Origin\",48.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",7120]]",
        7121=>"PROJCS[\"NAD_1983_2011_RMTCRS_Milk_River_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",492125.9843],PARAMETER[\"False_Northing\",656167.979],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Standard_Parallel_1\",48.5],PARAMETER[\"Scale_Factor\",1.000145],PARAMETER[\"Latitude_Of_Origin\",48.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",7121]]",
        7122=>"PROJCS[\"NAD_1983_2011_RMTCRS_Fort_Belknap_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",656167.979],PARAMETER[\"False_Northing\",492125.9843],PARAMETER[\"Central_Meridian\",-108.5],PARAMETER[\"Standard_Parallel_1\",48.5],PARAMETER[\"Scale_Factor\",1.00012],PARAMETER[\"Latitude_Of_Origin\",48.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",7122]]",
        7123=>"PROJCS[\"NAD_1983_2011_RMTCRS_Fort_Peck_Assiniboine_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",656167.979],PARAMETER[\"False_Northing\",328083.9895],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",48.33333333333334],PARAMETER[\"Scale_Factor\",1.00012],PARAMETER[\"Latitude_Of_Origin\",48.33333333333334],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",7123]]",
        7124=>"PROJCS[\"NAD_1983_2011_RMTCRS_Fort_Peck_Sioux_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",328083.9895],PARAMETER[\"False_Northing\",164041.9938],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",48.33333333333334],PARAMETER[\"Scale_Factor\",1.00009],PARAMETER[\"Latitude_Of_Origin\",48.33333333333334],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",7124]]",
        7125=>"PROJCS[\"NAD_1983_2011_RMTCRS_Crow_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",656167.979],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-107.75],PARAMETER[\"Scale_Factor\",1.000148],PARAMETER[\"Latitude_Of_Origin\",44.75],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",7125]]",
        7126=>"PROJCS[\"NAD_1983_2011_RMTCRS_Bobcat_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",328083.9895],PARAMETER[\"False_Northing\",328083.9895],PARAMETER[\"Central_Meridian\",-111.25],PARAMETER[\"Standard_Parallel_1\",46.25],PARAMETER[\"Scale_Factor\",1.000185],PARAMETER[\"Latitude_Of_Origin\",46.25],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",7126]]",
        7127=>"PROJCS[\"NAD_1983_2011_RMTCRS_Billings_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",656167.979],PARAMETER[\"False_Northing\",164041.9948],PARAMETER[\"Central_Meridian\",-108.4166666666667],PARAMETER[\"Standard_Parallel_1\",45.78333333333333],PARAMETER[\"Scale_Factor\",1.0001515],PARAMETER[\"Latitude_Of_Origin\",45.78333333333333],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",7127]]",
        7128=>"PROJCS[\"NAD_1983_2011_RMTCRS_Wind_River_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",328083.3333],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-108.3333333333333],PARAMETER[\"Scale_Factor\",1.00024],PARAMETER[\"Latitude_Of_Origin\",42.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7128]]",
        7131=>"PROJCS[\"NAD_1983_2011_San_Francisco_CS13_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",48000.0],PARAMETER[\"False_Northing\",24000.0],PARAMETER[\"Central_Meridian\",-122.45],PARAMETER[\"Scale_Factor\",1.000007],PARAMETER[\"Latitude_Of_Origin\",37.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7131]]",
        7132=>"PROJCS[\"NAD_1983_2011_San_Francisco_CS13_ftUS\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",157480.0],PARAMETER[\"False_Northing\",78740.0],PARAMETER[\"Central_Meridian\",-122.45],PARAMETER[\"Scale_Factor\",1.000007],PARAMETER[\"Latitude_Of_Origin\",37.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7132]]",
        7142=>"PROJCS[\"Palestine_1923_Palestine_Grid_TM\",GEOGCS[\"GCS_Palestine_1923\",DATUM[\"D_Palestine_1923\",SPHEROID[\"Clarke_1880_Benoit\",6378300.789,293.4663155389802]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",170251.555],PARAMETER[\"False_Northing\",126867.909],PARAMETER[\"Central_Meridian\",35.21208055555556],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",31.73409694444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7142]]",
        7257=>"PROJCS[\"NAD_1983_2011_InGCS_Adams_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-84.95],PARAMETER[\"Scale_Factor\",1.000034],PARAMETER[\"Latitude_Of_Origin\",40.55],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7257]]",
        7258=>"PROJCS[\"NAD_1983_2011_InGCS_Adams_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-84.95],PARAMETER[\"Scale_Factor\",1.000034],PARAMETER[\"Latitude_Of_Origin\",40.55],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7258]]",
        7259=>"PROJCS[\"NAD_1983_2011_InGCS_Allen_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.05],PARAMETER[\"Scale_Factor\",1.000031],PARAMETER[\"Latitude_Of_Origin\",40.9],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7259]]",
        7260=>"PROJCS[\"NAD_1983_2011_InGCS_Allen_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.05],PARAMETER[\"Scale_Factor\",1.000031],PARAMETER[\"Latitude_Of_Origin\",40.9],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7260]]",
        7261=>"PROJCS[\"NAD_1983_2011_InGCS_Bartholomew_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.85],PARAMETER[\"Scale_Factor\",1.000026],PARAMETER[\"Latitude_Of_Origin\",39.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7261]]",
        7262=>"PROJCS[\"NAD_1983_2011_InGCS_Bartholomew_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.85],PARAMETER[\"Scale_Factor\",1.000026],PARAMETER[\"Latitude_Of_Origin\",39.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7262]]",
        7263=>"PROJCS[\"NAD_1983_2011_InGCS_Benton_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-87.3],PARAMETER[\"Scale_Factor\",1.000029],PARAMETER[\"Latitude_Of_Origin\",40.45],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7263]]",
        7264=>"PROJCS[\"NAD_1983_2011_InGCS_Benton_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-87.3],PARAMETER[\"Scale_Factor\",1.000029],PARAMETER[\"Latitude_Of_Origin\",40.45],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7264]]",
        7265=>"PROJCS[\"NAD_1983_2011_InGCS_Blackford-Delaware_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.4],PARAMETER[\"Scale_Factor\",1.000038],PARAMETER[\"Latitude_Of_Origin\",40.05],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7265]]",
        7266=>"PROJCS[\"NAD_1983_2011_InGCS_Blackford-Delaware_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.4],PARAMETER[\"Scale_Factor\",1.000038],PARAMETER[\"Latitude_Of_Origin\",40.05],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7266]]",
        7267=>"PROJCS[\"NAD_1983_2011_InGCS_Boone-Hendricks_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.5],PARAMETER[\"Scale_Factor\",1.000036],PARAMETER[\"Latitude_Of_Origin\",39.6],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7267]]",
        7268=>"PROJCS[\"NAD_1983_2011_InGCS_Boone-Hendricks_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.5],PARAMETER[\"Scale_Factor\",1.000036],PARAMETER[\"Latitude_Of_Origin\",39.6],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7268]]",
        7269=>"PROJCS[\"NAD_1983_2011_InGCS_Brown_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.3],PARAMETER[\"Scale_Factor\",1.00003],PARAMETER[\"Latitude_Of_Origin\",39.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7269]]",
        7270=>"PROJCS[\"NAD_1983_2011_InGCS_Brown_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.3],PARAMETER[\"Scale_Factor\",1.00003],PARAMETER[\"Latitude_Of_Origin\",39.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7270]]",
        7271=>"PROJCS[\"NAD_1983_2011_InGCS_Carroll_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.65],PARAMETER[\"Scale_Factor\",1.000026],PARAMETER[\"Latitude_Of_Origin\",40.4],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7271]]",
        7272=>"PROJCS[\"NAD_1983_2011_InGCS_Carroll_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.65],PARAMETER[\"Scale_Factor\",1.000026],PARAMETER[\"Latitude_Of_Origin\",40.4],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7272]]",
        7273=>"PROJCS[\"NAD_1983_2011_InGCS_Cass_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.4],PARAMETER[\"Scale_Factor\",1.000028],PARAMETER[\"Latitude_Of_Origin\",40.55],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7273]]",
        7274=>"PROJCS[\"NAD_1983_2011_InGCS_Cass_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.4],PARAMETER[\"Scale_Factor\",1.000028],PARAMETER[\"Latitude_Of_Origin\",40.55],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7274]]",
        7275=>"PROJCS[\"NAD_1983_2011_InGCS_Clark-Floyd-Scott_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.6],PARAMETER[\"Scale_Factor\",1.000021],PARAMETER[\"Latitude_Of_Origin\",38.15],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7275]]",
        7276=>"PROJCS[\"NAD_1983_2011_InGCS_Clark-Floyd-Scott_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.6],PARAMETER[\"Scale_Factor\",1.000021],PARAMETER[\"Latitude_Of_Origin\",38.15],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7276]]",
        7277=>"PROJCS[\"NAD_1983_2011_InGCS_Clay_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-87.15],PARAMETER[\"Scale_Factor\",1.000024],PARAMETER[\"Latitude_Of_Origin\",39.15],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7277]]",
        7278=>"PROJCS[\"NAD_1983_2011_InGCS_Clay_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-87.15],PARAMETER[\"Scale_Factor\",1.000024],PARAMETER[\"Latitude_Of_Origin\",39.15],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7278]]",
        7279=>"PROJCS[\"NAD_1983_2011_InGCS_Clinton_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.6],PARAMETER[\"Scale_Factor\",1.000032],PARAMETER[\"Latitude_Of_Origin\",40.15],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7279]]",
        7280=>"PROJCS[\"NAD_1983_2011_InGCS_Clinton_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.6],PARAMETER[\"Scale_Factor\",1.000032],PARAMETER[\"Latitude_Of_Origin\",40.15],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7280]]",
        7281=>"PROJCS[\"NAD_1983_2011_InGCS_Crawford-Lawrence-Orange_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.5],PARAMETER[\"Scale_Factor\",1.000025],PARAMETER[\"Latitude_Of_Origin\",38.1],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7281]]",
        7282=>"PROJCS[\"NAD_1983_2011_InGCS_Crawford-Lawrence-Orange_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.5],PARAMETER[\"Scale_Factor\",1.000025],PARAMETER[\"Latitude_Of_Origin\",38.1],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7282]]",
        7283=>"PROJCS[\"NAD_1983_2011_InGCS_Daviess-Greene_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-87.1],PARAMETER[\"Scale_Factor\",1.000018],PARAMETER[\"Latitude_Of_Origin\",38.45],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7283]]",
        7284=>"PROJCS[\"NAD_1983_2011_InGCS_Daviess-Greene_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-87.1],PARAMETER[\"Scale_Factor\",1.000018],PARAMETER[\"Latitude_Of_Origin\",38.45],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7284]]",
        7285=>"PROJCS[\"NAD_1983_2011_InGCS_Dearborn-Ohio-Switzerland_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-84.9],PARAMETER[\"Scale_Factor\",1.000029],PARAMETER[\"Latitude_Of_Origin\",38.65],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7285]]",
        7286=>"PROJCS[\"NAD_1983_2011_InGCS_Dearborn-Ohio-Switzerland_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-84.9],PARAMETER[\"Scale_Factor\",1.000029],PARAMETER[\"Latitude_Of_Origin\",38.65],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7286]]",
        7287=>"PROJCS[\"NAD_1983_2011_InGCS_Decatur-Rush_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.65],PARAMETER[\"Scale_Factor\",1.000036],PARAMETER[\"Latitude_Of_Origin\",39.1],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7287]]",
        7288=>"PROJCS[\"NAD_1983_2011_InGCS_Decatur-Rush_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.65],PARAMETER[\"Scale_Factor\",1.000036],PARAMETER[\"Latitude_Of_Origin\",39.1],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7288]]",
        7289=>"PROJCS[\"NAD_1983_2011_InGCS_DeKalb_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-84.95],PARAMETER[\"Scale_Factor\",1.000036],PARAMETER[\"Latitude_Of_Origin\",41.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7289]]",
        7290=>"PROJCS[\"NAD_1983_2011_InGCS_DeKalb_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-84.95],PARAMETER[\"Scale_Factor\",1.000036],PARAMETER[\"Latitude_Of_Origin\",41.25],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7290]]",
        7291=>"PROJCS[\"NAD_1983_2011_InGCS_Dubois-Martin_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.95],PARAMETER[\"Scale_Factor\",1.00002],PARAMETER[\"Latitude_Of_Origin\",38.2],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7291]]",
        7292=>"PROJCS[\"NAD_1983_2011_InGCS_Dubois-Martin_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.95],PARAMETER[\"Scale_Factor\",1.00002],PARAMETER[\"Latitude_Of_Origin\",38.2],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7292]]",
        7293=>"PROJCS[\"NAD_1983_2011_InGCS_Elkhart-Kosciusko-Wabash_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.85],PARAMETER[\"Scale_Factor\",1.000033],PARAMETER[\"Latitude_Of_Origin\",40.65],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7293]]",
        7294=>"PROJCS[\"NAD_1983_2011_InGCS_Elkhart-Kosciusko-Wabash_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.85],PARAMETER[\"Scale_Factor\",1.000033],PARAMETER[\"Latitude_Of_Origin\",40.65],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7294]]",
        7295=>"PROJCS[\"NAD_1983_2011_InGCS_Fayette-Franklin-Union_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.05],PARAMETER[\"Scale_Factor\",1.000038],PARAMETER[\"Latitude_Of_Origin\",39.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7295]]",
        7296=>"PROJCS[\"NAD_1983_2011_InGCS_Fayette-Franklin-Union_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.05],PARAMETER[\"Scale_Factor\",1.000038],PARAMETER[\"Latitude_Of_Origin\",39.25],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7296]]",
        7297=>"PROJCS[\"NAD_1983_2011_InGCS_Fountain-Warren_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-87.3],PARAMETER[\"Scale_Factor\",1.000025],PARAMETER[\"Latitude_Of_Origin\",39.95],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7297]]",
        7298=>"PROJCS[\"NAD_1983_2011_InGCS_Fountain-Warren_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-87.3],PARAMETER[\"Scale_Factor\",1.000025],PARAMETER[\"Latitude_Of_Origin\",39.95],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7298]]",
        7299=>"PROJCS[\"NAD_1983_2011_InGCS_Fulton-Marshall-St_Joseph_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.3],PARAMETER[\"Scale_Factor\",1.000031],PARAMETER[\"Latitude_Of_Origin\",40.9],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7299]]",
        7300=>"PROJCS[\"NAD_1983_2011_InGCS_Fulton-Marshall-St_Joseph_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.3],PARAMETER[\"Scale_Factor\",1.000031],PARAMETER[\"Latitude_Of_Origin\",40.9],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7300]]",
        7301=>"PROJCS[\"NAD_1983_2011_InGCS_Gibson_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-87.65],PARAMETER[\"Scale_Factor\",1.000013],PARAMETER[\"Latitude_Of_Origin\",38.15],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7301]]",
        7302=>"PROJCS[\"NAD_1983_2011_InGCS_Gibson_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-87.65],PARAMETER[\"Scale_Factor\",1.000013],PARAMETER[\"Latitude_Of_Origin\",38.15],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7302]]",
        7303=>"PROJCS[\"NAD_1983_2011_InGCS_Grant_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.7],PARAMETER[\"Scale_Factor\",1.000034],PARAMETER[\"Latitude_Of_Origin\",40.35],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7303]]",
        7304=>"PROJCS[\"NAD_1983_2011_InGCS_Grant_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.7],PARAMETER[\"Scale_Factor\",1.000034],PARAMETER[\"Latitude_Of_Origin\",40.35],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7304]]",
        7305=>"PROJCS[\"NAD_1983_2011_InGCS_Hamilton-Tipton_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.0],PARAMETER[\"Scale_Factor\",1.000034],PARAMETER[\"Latitude_Of_Origin\",39.9],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7305]]",
        7306=>"PROJCS[\"NAD_1983_2011_InGCS_Hamilton-Tipton_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.0],PARAMETER[\"Scale_Factor\",1.000034],PARAMETER[\"Latitude_Of_Origin\",39.9],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7306]]",
        7307=>"PROJCS[\"NAD_1983_2011_InGCS_Hancock-Madison_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.8],PARAMETER[\"Scale_Factor\",1.000036],PARAMETER[\"Latitude_Of_Origin\",39.65],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7307]]",
        7308=>"PROJCS[\"NAD_1983_2011_InGCS_Hancock-Madison_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.8],PARAMETER[\"Scale_Factor\",1.000036],PARAMETER[\"Latitude_Of_Origin\",39.65],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7308]]",
        7309=>"PROJCS[\"NAD_1983_2011_InGCS_Harrison-Washington_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.15],PARAMETER[\"Scale_Factor\",1.000027],PARAMETER[\"Latitude_Of_Origin\",37.95],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7309]]",
        7310=>"PROJCS[\"NAD_1983_2011_InGCS_Harrison-Washington_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.15],PARAMETER[\"Scale_Factor\",1.000027],PARAMETER[\"Latitude_Of_Origin\",37.95],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7310]]",
        7311=>"PROJCS[\"NAD_1983_2011_InGCS_Henry_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.45],PARAMETER[\"Scale_Factor\",1.000043],PARAMETER[\"Latitude_Of_Origin\",39.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7311]]",
        7312=>"PROJCS[\"NAD_1983_2011_InGCS_Henry_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.45],PARAMETER[\"Scale_Factor\",1.000043],PARAMETER[\"Latitude_Of_Origin\",39.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7312]]",
        7313=>"PROJCS[\"NAD_1983_2011_InGCS_Howard-Miami_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.15],PARAMETER[\"Scale_Factor\",1.000031],PARAMETER[\"Latitude_Of_Origin\",40.35],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7313]]",
        7314=>"PROJCS[\"NAD_1983_2011_InGCS_Howard-Miami_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.15],PARAMETER[\"Scale_Factor\",1.000031],PARAMETER[\"Latitude_Of_Origin\",40.35],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7314]]",
        7315=>"PROJCS[\"NAD_1983_2011_InGCS_Huntington-Whitley_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.5],PARAMETER[\"Scale_Factor\",1.000034],PARAMETER[\"Latitude_Of_Origin\",40.65],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7315]]",
        7316=>"PROJCS[\"NAD_1983_2011_InGCS_Huntington-Whitley_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.5],PARAMETER[\"Scale_Factor\",1.000034],PARAMETER[\"Latitude_Of_Origin\",40.65],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7316]]",
        7317=>"PROJCS[\"NAD_1983_2011_InGCS_Jackson_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.95],PARAMETER[\"Scale_Factor\",1.000022],PARAMETER[\"Latitude_Of_Origin\",38.7],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7317]]",
        7318=>"PROJCS[\"NAD_1983_2011_InGCS_Jackson_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.95],PARAMETER[\"Scale_Factor\",1.000022],PARAMETER[\"Latitude_Of_Origin\",38.7],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7318]]",
        7319=>"PROJCS[\"NAD_1983_2011_InGCS_Jasper-Porter_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-87.1],PARAMETER[\"Scale_Factor\",1.000027],PARAMETER[\"Latitude_Of_Origin\",40.7],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7319]]",
        7320=>"PROJCS[\"NAD_1983_2011_InGCS_Jasper-Porter_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-87.1],PARAMETER[\"Scale_Factor\",1.000027],PARAMETER[\"Latitude_Of_Origin\",40.7],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7320]]",
        7321=>"PROJCS[\"NAD_1983_2011_InGCS_Jay_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.0],PARAMETER[\"Scale_Factor\",1.000038],PARAMETER[\"Latitude_Of_Origin\",40.3],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7321]]",
        7322=>"PROJCS[\"NAD_1983_2011_InGCS_Jay_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.0],PARAMETER[\"Scale_Factor\",1.000038],PARAMETER[\"Latitude_Of_Origin\",40.3],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7322]]",
        7323=>"PROJCS[\"NAD_1983_2011_InGCS_Jefferson_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.35],PARAMETER[\"Scale_Factor\",1.000028],PARAMETER[\"Latitude_Of_Origin\",38.55],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7323]]",
        7324=>"PROJCS[\"NAD_1983_2011_InGCS_Jefferson_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.35],PARAMETER[\"Scale_Factor\",1.000028],PARAMETER[\"Latitude_Of_Origin\",38.55],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7324]]",
        7325=>"PROJCS[\"NAD_1983_2011_InGCS_Jennings_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.8],PARAMETER[\"Scale_Factor\",1.000025],PARAMETER[\"Latitude_Of_Origin\",38.8],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7325]]",
        7326=>"PROJCS[\"NAD_1983_2011_InGCS_Jennings_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.8],PARAMETER[\"Scale_Factor\",1.000025],PARAMETER[\"Latitude_Of_Origin\",38.8],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7326]]",
        7327=>"PROJCS[\"NAD_1983_2011_InGCS_Johnson-Marion_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.15],PARAMETER[\"Scale_Factor\",1.000031],PARAMETER[\"Latitude_Of_Origin\",39.3],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7327]]",
        7328=>"PROJCS[\"NAD_1983_2011_InGCS_Johnson-Marion_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.15],PARAMETER[\"Scale_Factor\",1.000031],PARAMETER[\"Latitude_Of_Origin\",39.3],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7328]]",
        7329=>"PROJCS[\"NAD_1983_2011_InGCS_Knox_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-87.45],PARAMETER[\"Scale_Factor\",1.000015],PARAMETER[\"Latitude_Of_Origin\",38.4],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7329]]",
        7330=>"PROJCS[\"NAD_1983_2011_InGCS_Knox_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-87.45],PARAMETER[\"Scale_Factor\",1.000015],PARAMETER[\"Latitude_Of_Origin\",38.4],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7330]]",
        7331=>"PROJCS[\"NAD_1983_2011_InGCS_LaGrange-Noble_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.45],PARAMETER[\"Scale_Factor\",1.000037],PARAMETER[\"Latitude_Of_Origin\",41.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7331]]",
        7332=>"PROJCS[\"NAD_1983_2011_InGCS_LaGrange-Noble_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.45],PARAMETER[\"Scale_Factor\",1.000037],PARAMETER[\"Latitude_Of_Origin\",41.25],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7332]]",
        7333=>"PROJCS[\"NAD_1983_2011_InGCS_Lake-Newton_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-87.4],PARAMETER[\"Scale_Factor\",1.000026],PARAMETER[\"Latitude_Of_Origin\",40.7],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7333]]",
        7334=>"PROJCS[\"NAD_1983_2011_InGCS_Lake-Newton_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-87.4],PARAMETER[\"Scale_Factor\",1.000026],PARAMETER[\"Latitude_Of_Origin\",40.7],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7334]]",
        7335=>"PROJCS[\"NAD_1983_2011_InGCS_LaPorte-Pulaski-Starke_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.75],PARAMETER[\"Scale_Factor\",1.000027],PARAMETER[\"Latitude_Of_Origin\",40.9],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7335]]",
        7336=>"PROJCS[\"NAD_1983_2011_InGCS_LaPorte-Pulaski-Starke_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.75],PARAMETER[\"Scale_Factor\",1.000027],PARAMETER[\"Latitude_Of_Origin\",40.9],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7336]]",
        7337=>"PROJCS[\"NAD_1983_2011_InGCS_Monroe-Morgan_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.5],PARAMETER[\"Scale_Factor\",1.000028],PARAMETER[\"Latitude_Of_Origin\",38.95],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7337]]",
        7338=>"PROJCS[\"NAD_1983_2011_InGCS_Monroe-Morgan_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.5],PARAMETER[\"Scale_Factor\",1.000028],PARAMETER[\"Latitude_Of_Origin\",38.95],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7338]]",
        7339=>"PROJCS[\"NAD_1983_2011_InGCS_Montgomery-Putnam_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.95],PARAMETER[\"Scale_Factor\",1.000031],PARAMETER[\"Latitude_Of_Origin\",39.45],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7339]]",
        7340=>"PROJCS[\"NAD_1983_2011_InGCS_Montgomery-Putnam_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.95],PARAMETER[\"Scale_Factor\",1.000031],PARAMETER[\"Latitude_Of_Origin\",39.45],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7340]]",
        7341=>"PROJCS[\"NAD_1983_2011_InGCS_Owen_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.9],PARAMETER[\"Scale_Factor\",1.000026],PARAMETER[\"Latitude_Of_Origin\",39.15],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7341]]",
        7342=>"PROJCS[\"NAD_1983_2011_InGCS_Owen_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.9],PARAMETER[\"Scale_Factor\",1.000026],PARAMETER[\"Latitude_Of_Origin\",39.15],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7342]]",
        7343=>"PROJCS[\"NAD_1983_2011_InGCS_Parke-Vermillion_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-87.35],PARAMETER[\"Scale_Factor\",1.000022],PARAMETER[\"Latitude_Of_Origin\",39.6],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7343]]",
        7344=>"PROJCS[\"NAD_1983_2011_InGCS_Parke-Vermillion_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-87.35],PARAMETER[\"Scale_Factor\",1.000022],PARAMETER[\"Latitude_Of_Origin\",39.6],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7344]]",
        7345=>"PROJCS[\"NAD_1983_2011_InGCS_Perry_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.7],PARAMETER[\"Scale_Factor\",1.00002],PARAMETER[\"Latitude_Of_Origin\",37.8],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7345]]",
        7346=>"PROJCS[\"NAD_1983_2011_InGCS_Perry_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.7],PARAMETER[\"Scale_Factor\",1.00002],PARAMETER[\"Latitude_Of_Origin\",37.8],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7346]]",
        7347=>"PROJCS[\"NAD_1983_2011_InGCS_Pike-Warrick_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-87.3],PARAMETER[\"Scale_Factor\",1.000015],PARAMETER[\"Latitude_Of_Origin\",37.85],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7347]]",
        7348=>"PROJCS[\"NAD_1983_2011_InGCS_Pike-Warrick_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-87.3],PARAMETER[\"Scale_Factor\",1.000015],PARAMETER[\"Latitude_Of_Origin\",37.85],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7348]]",
        7349=>"PROJCS[\"NAD_1983_2011_InGCS_Posey_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-87.95],PARAMETER[\"Scale_Factor\",1.000013],PARAMETER[\"Latitude_Of_Origin\",37.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7349]]",
        7350=>"PROJCS[\"NAD_1983_2011_InGCS_Posey_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-87.95],PARAMETER[\"Scale_Factor\",1.000013],PARAMETER[\"Latitude_Of_Origin\",37.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7350]]",
        7351=>"PROJCS[\"NAD_1983_2011_InGCS_Randolph-Wayne_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.05],PARAMETER[\"Scale_Factor\",1.000044],PARAMETER[\"Latitude_Of_Origin\",39.7],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7351]]",
        7352=>"PROJCS[\"NAD_1983_2011_InGCS_Randolph-Wayne_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.05],PARAMETER[\"Scale_Factor\",1.000044],PARAMETER[\"Latitude_Of_Origin\",39.7],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7352]]",
        7353=>"PROJCS[\"NAD_1983_2011_InGCS_Ripley_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.3],PARAMETER[\"Scale_Factor\",1.000038],PARAMETER[\"Latitude_Of_Origin\",38.9],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7353]]",
        7354=>"PROJCS[\"NAD_1983_2011_InGCS_Ripley_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.3],PARAMETER[\"Scale_Factor\",1.000038],PARAMETER[\"Latitude_Of_Origin\",38.9],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7354]]",
        7355=>"PROJCS[\"NAD_1983_2011_InGCS_Shelby_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.9],PARAMETER[\"Scale_Factor\",1.00003],PARAMETER[\"Latitude_Of_Origin\",39.3],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7355]]",
        7356=>"PROJCS[\"NAD_1983_2011_InGCS_Shelby_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.9],PARAMETER[\"Scale_Factor\",1.00003],PARAMETER[\"Latitude_Of_Origin\",39.3],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7356]]",
        7357=>"PROJCS[\"NAD_1983_2011_InGCS_Spencer_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-87.05],PARAMETER[\"Scale_Factor\",1.000014],PARAMETER[\"Latitude_Of_Origin\",37.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7357]]",
        7358=>"PROJCS[\"NAD_1983_2011_InGCS_Spencer_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-87.05],PARAMETER[\"Scale_Factor\",1.000014],PARAMETER[\"Latitude_Of_Origin\",37.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7358]]",
        7359=>"PROJCS[\"NAD_1983_2011_InGCS_Steuben_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.0],PARAMETER[\"Scale_Factor\",1.000041],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7359]]",
        7360=>"PROJCS[\"NAD_1983_2011_InGCS_Steuben_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.0],PARAMETER[\"Scale_Factor\",1.000041],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7360]]",
        7361=>"PROJCS[\"NAD_1983_2011_InGCS_Sullivan_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-87.5],PARAMETER[\"Scale_Factor\",1.000017],PARAMETER[\"Latitude_Of_Origin\",38.9],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7361]]",
        7362=>"PROJCS[\"NAD_1983_2011_InGCS_Sullivan_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-87.5],PARAMETER[\"Scale_Factor\",1.000017],PARAMETER[\"Latitude_Of_Origin\",38.9],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7362]]",
        7363=>"PROJCS[\"NAD_1983_2011_InGCS_Tippecanoe-White_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-86.9],PARAMETER[\"Scale_Factor\",1.000026],PARAMETER[\"Latitude_Of_Origin\",40.2],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7363]]",
        7364=>"PROJCS[\"NAD_1983_2011_InGCS_Tippecanoe-White_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-86.9],PARAMETER[\"Scale_Factor\",1.000026],PARAMETER[\"Latitude_Of_Origin\",40.2],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7364]]",
        7365=>"PROJCS[\"NAD_1983_2011_InGCS_Vanderburgh_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-87.55],PARAMETER[\"Scale_Factor\",1.000015],PARAMETER[\"Latitude_Of_Origin\",37.8],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7365]]",
        7366=>"PROJCS[\"NAD_1983_2011_InGCS_Vanderburgh_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-87.55],PARAMETER[\"Scale_Factor\",1.000015],PARAMETER[\"Latitude_Of_Origin\",37.8],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7366]]",
        7367=>"PROJCS[\"NAD_1983_2011_InGCS_Vigo_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-87.45],PARAMETER[\"Scale_Factor\",1.00002],PARAMETER[\"Latitude_Of_Origin\",39.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7367]]",
        7368=>"PROJCS[\"NAD_1983_2011_InGCS_Vigo_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-87.45],PARAMETER[\"Scale_Factor\",1.00002],PARAMETER[\"Latitude_Of_Origin\",39.25],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7368]]",
        7369=>"PROJCS[\"NAD_1983_2011_InGCS_Wells_(m)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",240000.0],PARAMETER[\"False_Northing\",36000.0],PARAMETER[\"Central_Meridian\",-85.25],PARAMETER[\"Scale_Factor\",1.000034],PARAMETER[\"Latitude_Of_Origin\",40.55],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7369]]",
        7370=>"PROJCS[\"NAD_1983_2011_InGCS_Wells_(ftUS)\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",787400.0],PARAMETER[\"False_Northing\",118110.0],PARAMETER[\"Central_Meridian\",-85.25],PARAMETER[\"Scale_Factor\",1.000034],PARAMETER[\"Latitude_Of_Origin\",40.55],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7370]]",
        7374=>"PROJCS[\"ONGD14_UTM_Zone_39N\",GEOGCS[\"ONGD14\",DATUM[\"Oman_National_Geodetic_Datum_2014\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7374]]",
        7375=>"PROJCS[\"ONGD14_UTM_Zone_40N\",GEOGCS[\"ONGD14\",DATUM[\"Oman_National_Geodetic_Datum_2014\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7375]]",
        7376=>"PROJCS[\"ONGD14_UTM_Zone_41N\",GEOGCS[\"ONGD14\",DATUM[\"Oman_National_Geodetic_Datum_2014\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7376]]",
        7528=>"PROJCS[\"NAD_1983_2011_WISCRS_Adams_and_Juneau_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",147218.6942],PARAMETER[\"False_Northing\",0.0037],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Scale_Factor\",1.0000365285],PARAMETER[\"Latitude_Of_Origin\",43.36666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7528]]",
        7529=>"PROJCS[\"NAD_1983_2011_WISCRS_Ashland_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",172821.9461],PARAMETER[\"False_Northing\",0.0017],PARAMETER[\"Central_Meridian\",-90.62222222222222],PARAMETER[\"Scale_Factor\",1.0000495683],PARAMETER[\"Latitude_Of_Origin\",45.70611111111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7529]]",
        7530=>"PROJCS[\"NAD_1983_2011_WISCRS_Barron_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",93150.0],PARAMETER[\"False_Northing\",0.0029],PARAMETER[\"Central_Meridian\",-91.85],PARAMETER[\"Scale_Factor\",1.0000486665],PARAMETER[\"Latitude_Of_Origin\",45.13333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7530]]",
        7531=>"PROJCS[\"NAD_1983_2011_WISCRS_Bayfield_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",228600.4575],PARAMETER[\"False_Northing\",148551.4837],PARAMETER[\"Central_Meridian\",-91.15277777777779],PARAMETER[\"Standard_Parallel_1\",46.66964837722222],PARAMETER[\"Scale_Factor\",1.0000331195],PARAMETER[\"Latitude_Of_Origin\",46.66964837722222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7531]]",
        7532=>"PROJCS[\"NAD_1983_2011_WISCRS_Brown_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",31600.0],PARAMETER[\"False_Northing\",4600.0],PARAMETER[\"Central_Meridian\",-88.0],PARAMETER[\"Scale_Factor\",1.00002],PARAMETER[\"Latitude_Of_Origin\",43.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7532]]",
        7533=>"PROJCS[\"NAD_1983_2011_WISCRS_Buffalo_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",175260.3502],PARAMETER[\"False_Northing\",0.0048],PARAMETER[\"Central_Meridian\",-91.79722222222222],PARAMETER[\"Scale_Factor\",1.0000382778],PARAMETER[\"Latitude_Of_Origin\",43.48138888888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7533]]",
        7534=>"PROJCS[\"NAD_1983_2011_WISCRS_Burnett_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",64008.1276],PARAMETER[\"False_Northing\",59445.9043],PARAMETER[\"Central_Meridian\",-92.45777777777778],PARAMETER[\"Standard_Parallel_1\",45.89871486583333],PARAMETER[\"Scale_Factor\",1.0000383841],PARAMETER[\"Latitude_Of_Origin\",45.89871486583333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7534]]",
        7535=>"PROJCS[\"NAD_1983_2011_WISCRS_Calumet_Fond_du_Lac_Outagamie_Winnebago_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",244754.8893],PARAMETER[\"False_Northing\",0.0049],PARAMETER[\"Central_Meridian\",-88.5],PARAMETER[\"Scale_Factor\",1.0000286569],PARAMETER[\"Latitude_Of_Origin\",42.71944444444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7535]]",
        7536=>"PROJCS[\"NAD_1983_2011_WISCRS_Chippewa_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",60045.72],PARAMETER[\"False_Northing\",44091.4346],PARAMETER[\"Central_Meridian\",-91.29444444444444],PARAMETER[\"Standard_Parallel_1\",44.97785689861112],PARAMETER[\"Scale_Factor\",1.0000391127],PARAMETER[\"Latitude_Of_Origin\",44.97785689861112],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7536]]",
        7537=>"PROJCS[\"NAD_1983_2011_WISCRS_Clark_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",199949.1989],PARAMETER[\"False_Northing\",0.0086],PARAMETER[\"Central_Meridian\",-90.70833333333334],PARAMETER[\"Scale_Factor\",1.0000463003],PARAMETER[\"Latitude_Of_Origin\",43.6],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7537]]",
        7538=>"PROJCS[\"NAD_1983_2011_WISCRS_Columbia_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",169164.3381],PARAMETER[\"False_Northing\",111569.6134],PARAMETER[\"Central_Meridian\",-89.39444444444445],PARAMETER[\"Standard_Parallel_1\",43.46254664583333],PARAMETER[\"Scale_Factor\",1.00003498],PARAMETER[\"Latitude_Of_Origin\",43.46254664583333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7538]]",
        7539=>"PROJCS[\"NAD_1983_2011_WISCRS_Crawford_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",113690.6274],PARAMETER[\"False_Northing\",53703.1201],PARAMETER[\"Central_Meridian\",-90.9388888888889],PARAMETER[\"Standard_Parallel_1\",43.200055605],PARAMETER[\"Scale_Factor\",1.0000349151],PARAMETER[\"Latitude_Of_Origin\",43.200055605],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7539]]",
        7540=>"PROJCS[\"NAD_1983_2011_WISCRS_Dane_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",247193.2944],PARAMETER[\"False_Northing\",146591.9896],PARAMETER[\"Central_Meridian\",-89.42222222222223],PARAMETER[\"Standard_Parallel_1\",43.0695160375],PARAMETER[\"Scale_Factor\",1.0000384786],PARAMETER[\"Latitude_Of_Origin\",43.0695160375],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7540]]",
        7541=>"PROJCS[\"NAD_1983_2011_WISCRS_Dodge_and_Jefferson_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",263347.7263],PARAMETER[\"False_Northing\",0.0076],PARAMETER[\"Central_Meridian\",-88.775],PARAMETER[\"Scale_Factor\",1.0000346418],PARAMETER[\"Latitude_Of_Origin\",41.47222222222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7541]]",
        7542=>"PROJCS[\"NAD_1983_2011_WISCRS_Door_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",158801.1176],PARAMETER[\"False_Northing\",0.0023],PARAMETER[\"Central_Meridian\",-87.27222222222223],PARAMETER[\"Scale_Factor\",1.0000187521],PARAMETER[\"Latitude_Of_Origin\",44.4],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7542]]",
        7543=>"PROJCS[\"NAD_1983_2011_WISCRS_Douglas_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",59131.3183],PARAMETER[\"False_Northing\",0.0041],PARAMETER[\"Central_Meridian\",-91.91666666666667],PARAMETER[\"Scale_Factor\",1.0000385418],PARAMETER[\"Latitude_Of_Origin\",45.88333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7543]]",
        7544=>"PROJCS[\"NAD_1983_2011_WISCRS_Dunn_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",51816.104],PARAMETER[\"False_Northing\",0.003],PARAMETER[\"Central_Meridian\",-91.89444444444445],PARAMETER[\"Scale_Factor\",1.0000410324],PARAMETER[\"Latitude_Of_Origin\",44.40833333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7544]]",
        7545=>"PROJCS[\"NAD_1983_2011_WISCRS_EauClaire_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",120091.4402],PARAMETER[\"False_Northing\",91687.9239],PARAMETER[\"Central_Meridian\",-91.28888888888889],PARAMETER[\"Standard_Parallel_1\",44.87228112638889],PARAMETER[\"Scale_Factor\",1.000035079],PARAMETER[\"Latitude_Of_Origin\",44.87228112638889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7545]]",
        7546=>"PROJCS[\"NAD_1983_2011_WISCRS_Florence_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",133502.6683],PARAMETER[\"False_Northing\",0.0063],PARAMETER[\"Central_Meridian\",-88.14166666666668],PARAMETER[\"Scale_Factor\",1.0000552095],PARAMETER[\"Latitude_Of_Origin\",45.43888888888888],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7546]]",
        7547=>"PROJCS[\"NAD_1983_2011_WISCRS_Forest_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",275844.5533],PARAMETER[\"False_Northing\",0.0157],PARAMETER[\"Central_Meridian\",-88.63333333333334],PARAMETER[\"Scale_Factor\",1.0000673004],PARAMETER[\"Latitude_Of_Origin\",44.00555555555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7547]]",
        7548=>"PROJCS[\"NAD_1983_2011_WISCRS_Grant_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",242316.4841],PARAMETER[\"False_Northing\",0.01],PARAMETER[\"Central_Meridian\",-90.8],PARAMETER[\"Scale_Factor\",1.0000349452],PARAMETER[\"Latitude_Of_Origin\",41.41111111111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7548]]",
        7549=>"PROJCS[\"NAD_1983_2011_WISCRS_Green_and_Lafayette_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",170078.7403],PARAMETER[\"False_Northing\",45830.2947],PARAMETER[\"Central_Meridian\",-89.83888888888889],PARAMETER[\"Standard_Parallel_1\",42.63756227694444],PARAMETER[\"Scale_Factor\",1.0000390487],PARAMETER[\"Latitude_Of_Origin\",42.63756227694444],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7549]]",
        7550=>"PROJCS[\"NAD_1983_2011_WISCRS_Green_Lake_and_Marquette_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",150876.3018],PARAMETER[\"False_Northing\",79170.7795],PARAMETER[\"Central_Meridian\",-89.24166666666667],PARAMETER[\"Standard_Parallel_1\",43.80700011777778],PARAMETER[\"Scale_Factor\",1.0000344057],PARAMETER[\"Latitude_Of_Origin\",43.80700011777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7550]]",
        7551=>"PROJCS[\"NAD_1983_2011_WISCRS_Iowa_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",113081.0261],PARAMETER[\"False_Northing\",0.0045],PARAMETER[\"Central_Meridian\",-90.16111111111111],PARAMETER[\"Scale_Factor\",1.0000394961],PARAMETER[\"Latitude_Of_Origin\",42.53888888888888],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7551]]",
        7552=>"PROJCS[\"NAD_1983_2011_WISCRS_Iron_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",220980.4419],PARAMETER[\"False_Northing\",0.0085],PARAMETER[\"Central_Meridian\",-90.25555555555556],PARAMETER[\"Scale_Factor\",1.0000677153],PARAMETER[\"Latitude_Of_Origin\",45.43333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7552]]",
        7553=>"PROJCS[\"NAD_1983_2011_WISCRS_Jackson_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",27000.0],PARAMETER[\"False_Northing\",25000.0],PARAMETER[\"Central_Meridian\",-90.84429651944444],PARAMETER[\"Scale_Factor\",1.0000353],PARAMETER[\"Latitude_Of_Origin\",44.25333512777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7553]]",
        7554=>"PROJCS[\"NAD_1983_2011_WISCRS_Kenosha_Milwaukee_Ozaukee_Racine_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",185928.3728],PARAMETER[\"False_Northing\",0.0009],PARAMETER[\"Central_Meridian\",-87.89444444444445],PARAMETER[\"Scale_Factor\",1.0000260649],PARAMETER[\"Latitude_Of_Origin\",42.21666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7554]]",
        7555=>"PROJCS[\"NAD_1983_2011_WISCRS_Kewaunee_Manitowoc_Sheboygan_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",79857.7614],PARAMETER[\"False_Northing\",0.0012],PARAMETER[\"Central_Meridian\",-87.55],PARAMETER[\"Scale_Factor\",1.0000233704],PARAMETER[\"Latitude_Of_Origin\",43.26666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7555]]",
        7556=>"PROJCS[\"NAD_1983_2011_WISCRS_La_Crosse_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",130454.6598],PARAMETER[\"False_Northing\",0.0033],PARAMETER[\"Central_Meridian\",-91.31666666666666],PARAMETER[\"Scale_Factor\",1.0000319985],PARAMETER[\"Latitude_Of_Origin\",43.45111111111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7556]]",
        7557=>"PROJCS[\"NAD_1983_2011_WISCRS_Langlade_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",198425.197],PARAMETER[\"False_Northing\",105279.7829],PARAMETER[\"Central_Meridian\",-89.03333333333333],PARAMETER[\"Standard_Parallel_1\",45.15423710527778],PARAMETER[\"Scale_Factor\",1.0000627024],PARAMETER[\"Latitude_Of_Origin\",45.15423710527778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7557]]",
        7558=>"PROJCS[\"NAD_1983_2011_WISCRS_Lincoln_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",116129.0323],PARAMETER[\"False_Northing\",0.0058],PARAMETER[\"Central_Meridian\",-89.73333333333333],PARAMETER[\"Scale_Factor\",1.0000599003],PARAMETER[\"Latitude_Of_Origin\",44.84444444444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7558]]",
        7559=>"PROJCS[\"NAD_1983_2011_WISCRS_Marathon_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",74676.1493],PARAMETER[\"False_Northing\",55049.2669],PARAMETER[\"Central_Meridian\",-89.77],PARAMETER[\"Standard_Parallel_1\",44.90090442361111],PARAMETER[\"Scale_Factor\",1.000053289],PARAMETER[\"Latitude_Of_Origin\",44.90090442361111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7559]]",
        7560=>"PROJCS[\"NAD_1983_2011_WISCRS_Marinette_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",238658.8794],PARAMETER[\"False_Northing\",0.0032],PARAMETER[\"Central_Meridian\",-87.71111111111111],PARAMETER[\"Scale_Factor\",1.0000234982],PARAMETER[\"Latitude_Of_Origin\",44.69166666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7560]]",
        7561=>"PROJCS[\"NAD_1983_2011_WISCRS_Menominee_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",105461.0121],PARAMETER[\"False_Northing\",0.0029],PARAMETER[\"Central_Meridian\",-88.41666666666667],PARAMETER[\"Scale_Factor\",1.0000362499],PARAMETER[\"Latitude_Of_Origin\",44.71666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7561]]",
        7562=>"PROJCS[\"NAD_1983_2011_WISCRS_Monroe_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",204521.209],PARAMETER[\"False_Northing\",121923.9861],PARAMETER[\"Central_Meridian\",-90.64166666666668],PARAMETER[\"Standard_Parallel_1\",44.00007392861111],PARAMETER[\"Scale_Factor\",1.0000434122],PARAMETER[\"Latitude_Of_Origin\",44.00007392861111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7562]]",
        7563=>"PROJCS[\"NAD_1983_2011_WISCRS_Oconto_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",182880.3676],PARAMETER[\"False_Northing\",0.0033],PARAMETER[\"Central_Meridian\",-87.90833333333335],PARAMETER[\"Scale_Factor\",1.0000236869],PARAMETER[\"Latitude_Of_Origin\",44.39722222222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7563]]",
        7564=>"PROJCS[\"NAD_1983_2011_WISCRS_Oneida_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",70104.1401],PARAMETER[\"False_Northing\",57588.0346],PARAMETER[\"Central_Meridian\",-89.54444444444444],PARAMETER[\"Standard_Parallel_1\",45.70422377027778],PARAMETER[\"Scale_Factor\",1.0000686968],PARAMETER[\"Latitude_Of_Origin\",45.70422377027778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7564]]",
        7565=>"PROJCS[\"NAD_1983_2011_WISCRS_Pepin_and_Pierce_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",167640.3354],PARAMETER[\"False_Northing\",86033.0876],PARAMETER[\"Central_Meridian\",-92.22777777777777],PARAMETER[\"Standard_Parallel_1\",44.63614887194444],PARAMETER[\"Scale_Factor\",1.0000362977],PARAMETER[\"Latitude_Of_Origin\",44.63614887194444],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7565]]",
        7566=>"PROJCS[\"NAD_1983_2011_WISCRS_Polk_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",141732.2823],PARAMETER[\"False_Northing\",0.0059],PARAMETER[\"Central_Meridian\",-92.63333333333334],PARAMETER[\"Scale_Factor\",1.0000433849],PARAMETER[\"Latitude_Of_Origin\",44.66111111111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7566]]",
        7567=>"PROJCS[\"NAD_1983_2011_WISCRS_Portage_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",56388.1128],PARAMETER[\"False_Northing\",50022.1874],PARAMETER[\"Central_Meridian\",-89.5],PARAMETER[\"Standard_Parallel_1\",44.41682397527777],PARAMETER[\"Scale_Factor\",1.000039936],PARAMETER[\"Latitude_Of_Origin\",44.41682397527777],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7567]]",
        7568=>"PROJCS[\"NAD_1983_2011_WISCRS_Price_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",227990.8546],PARAMETER[\"False_Northing\",0.0109],PARAMETER[\"Central_Meridian\",-90.48888888888889],PARAMETER[\"Scale_Factor\",1.0000649554],PARAMETER[\"Latitude_Of_Origin\",44.55555555555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7568]]",
        7569=>"PROJCS[\"NAD_1983_2011_WISCRS_Richland_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",202387.6048],PARAMETER[\"False_Northing\",134255.4253],PARAMETER[\"Central_Meridian\",-90.43055555555556],PARAMETER[\"Standard_Parallel_1\",43.3223129275],PARAMETER[\"Scale_Factor\",1.0000375653],PARAMETER[\"Latitude_Of_Origin\",43.3223129275],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7569]]",
        7570=>"PROJCS[\"NAD_1983_2011_WISCRS_Rock_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",146304.2926],PARAMETER[\"False_Northing\",0.0068],PARAMETER[\"Central_Meridian\",-89.07222222222222],PARAMETER[\"Scale_Factor\",1.0000337311],PARAMETER[\"Latitude_Of_Origin\",41.94444444444444],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7570]]",
        7571=>"PROJCS[\"NAD_1983_2011_WISCRS_Rusk_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250546.1013],PARAMETER[\"False_Northing\",0.0234],PARAMETER[\"Central_Meridian\",-91.06666666666666],PARAMETER[\"Scale_Factor\",1.0000495976],PARAMETER[\"Latitude_Of_Origin\",43.91944444444444],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7571]]",
        7572=>"PROJCS[\"NAD_1983_2011_WISCRS_Sauk_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",185623.5716],PARAMETER[\"False_Northing\",0.0051],PARAMETER[\"Central_Meridian\",-89.9],PARAMETER[\"Scale_Factor\",1.0000373868],PARAMETER[\"Latitude_Of_Origin\",42.81944444444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7572]]",
        7573=>"PROJCS[\"NAD_1983_2011_WISCRS_Sawyer_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",216713.2336],PARAMETER[\"False_Northing\",120734.1631],PARAMETER[\"Central_Meridian\",-91.11666666666666],PARAMETER[\"Standard_Parallel_1\",45.90009913138888],PARAMETER[\"Scale_Factor\",1.0000573461],PARAMETER[\"Latitude_Of_Origin\",45.90009913138888],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7573]]",
        7574=>"PROJCS[\"NAD_1983_2011_WISCRS_Shawano_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",262433.3253],PARAMETER[\"False_Northing\",0.0096],PARAMETER[\"Central_Meridian\",-88.60555555555555],PARAMETER[\"Scale_Factor\",1.000032144],PARAMETER[\"Latitude_Of_Origin\",44.03611111111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7574]]",
        7575=>"PROJCS[\"NAD_1983_2011_WISCRS_St_Croix_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",165506.7302],PARAMETER[\"False_Northing\",0.0103],PARAMETER[\"Central_Meridian\",-92.63333333333334],PARAMETER[\"Scale_Factor\",1.0000381803],PARAMETER[\"Latitude_Of_Origin\",44.03611111111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7575]]",
        7576=>"PROJCS[\"NAD_1983_2011_WISCRS_Taylor_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",187147.5744],PARAMETER[\"False_Northing\",107746.7522],PARAMETER[\"Central_Meridian\",-90.48333333333333],PARAMETER[\"Standard_Parallel_1\",45.17782208583333],PARAMETER[\"Scale_Factor\",1.0000597566],PARAMETER[\"Latitude_Of_Origin\",45.17782208583333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7576]]",
        7577=>"PROJCS[\"NAD_1983_2011_WISCRS_Trempealeau_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",256946.9138],PARAMETER[\"False_Northing\",0.0041],PARAMETER[\"Central_Meridian\",-91.36666666666666],PARAMETER[\"Scale_Factor\",1.0000361538],PARAMETER[\"Latitude_Of_Origin\",43.16111111111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7577]]",
        7578=>"PROJCS[\"NAD_1983_2011_WISCRS_Vernon_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",222504.4451],PARAMETER[\"False_Northing\",47532.0602],PARAMETER[\"Central_Meridian\",-90.78333333333333],PARAMETER[\"Standard_Parallel_1\",43.57503293972223],PARAMETER[\"Scale_Factor\",1.0000408158],PARAMETER[\"Latitude_Of_Origin\",43.57503293972223],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7578]]",
        7579=>"PROJCS[\"NAD_1983_2011_WISCRS_Vilas_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",134417.0689],PARAMETER[\"False_Northing\",50337.1092],PARAMETER[\"Central_Meridian\",-89.48888888888889],PARAMETER[\"Standard_Parallel_1\",46.07784409055556],PARAMETER[\"Scale_Factor\",1.0000730142],PARAMETER[\"Latitude_Of_Origin\",46.07784409055556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7579]]",
        7580=>"PROJCS[\"NAD_1983_2011_WISCRS_Walworth_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",232562.8651],PARAMETER[\"False_Northing\",111088.2224],PARAMETER[\"Central_Meridian\",-88.54166666666667],PARAMETER[\"Standard_Parallel_1\",42.66946209694444],PARAMETER[\"Scale_Factor\",1.0000367192],PARAMETER[\"Latitude_Of_Origin\",42.66946209694444],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7580]]",
        7581=>"PROJCS[\"NAD_1983_2011_WISCRS_Washburn_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",234086.8682],PARAMETER[\"False_Northing\",188358.6058],PARAMETER[\"Central_Meridian\",-91.78333333333333],PARAMETER[\"Standard_Parallel_1\",45.96121983333334],PARAMETER[\"Scale_Factor\",1.0000475376],PARAMETER[\"Latitude_Of_Origin\",45.96121983333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7581]]",
        7582=>"PROJCS[\"NAD_1983_2011_WISCRS_Washington_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",120091.4415],PARAMETER[\"False_Northing\",0.003],PARAMETER[\"Central_Meridian\",-88.06388888888888],PARAMETER[\"Scale_Factor\",1.00003738],PARAMETER[\"Latitude_Of_Origin\",42.91805555555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7582]]",
        7583=>"PROJCS[\"NAD_1983_2011_WISCRS_Waukesha_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",208788.418],PARAMETER[\"False_Northing\",0.0034],PARAMETER[\"Central_Meridian\",-88.225],PARAMETER[\"Scale_Factor\",1.0000346179],PARAMETER[\"Latitude_Of_Origin\",42.56944444444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7583]]",
        7584=>"PROJCS[\"NAD_1983_2011_WISCRS_Waupaca_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",185013.9709],PARAMETER[\"False_Northing\",0.007],PARAMETER[\"Central_Meridian\",-88.81666666666666],PARAMETER[\"Scale_Factor\",1.0000333645],PARAMETER[\"Latitude_Of_Origin\",43.42027777777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7584]]",
        7585=>"PROJCS[\"NAD_1983_2011_WISCRS_Waushara_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",120091.4402],PARAMETER[\"False_Northing\",45069.7587],PARAMETER[\"Central_Meridian\",-89.24166666666667],PARAMETER[\"Standard_Parallel_1\",44.11394404583334],PARAMETER[\"Scale_Factor\",1.0000392096],PARAMETER[\"Latitude_Of_Origin\",44.11394404583334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7585]]",
        7586=>"PROJCS[\"NAD_1983_2011_WISCRS_Wood_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",208483.6173],PARAMETER[\"False_Northing\",134589.754],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",44.36259546944444],PARAMETER[\"Scale_Factor\",1.0000421209],PARAMETER[\"Latitude_Of_Origin\",44.36259546944444],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7586]]",
        7587=>"PROJCS[\"NAD_1983_2011_WISCRS_Adams_and_Juneau_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",482999.999],PARAMETER[\"False_Northing\",0.012],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Scale_Factor\",1.0000365285],PARAMETER[\"Latitude_Of_Origin\",43.36666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7587]]",
        7588=>"PROJCS[\"NAD_1983_2011_WISCRS_Ashland_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",567000.001],PARAMETER[\"False_Northing\",0.006],PARAMETER[\"Central_Meridian\",-90.62222222222222],PARAMETER[\"Scale_Factor\",1.0000495683],PARAMETER[\"Latitude_Of_Origin\",45.70611111111111],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7588]]",
        7589=>"PROJCS[\"NAD_1983_2011_WISCRS_Barron_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",305609.625],PARAMETER[\"False_Northing\",0.01],PARAMETER[\"Central_Meridian\",-91.85],PARAMETER[\"Scale_Factor\",1.0000486665],PARAMETER[\"Latitude_Of_Origin\",45.13333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7589]]",
        7590=>"PROJCS[\"NAD_1983_2011_WISCRS_Bayfield_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",750000.001],PARAMETER[\"False_Northing\",487372.659],PARAMETER[\"Central_Meridian\",-91.15277777777779],PARAMETER[\"Standard_Parallel_1\",46.66964837722222],PARAMETER[\"Scale_Factor\",1.0000331195],PARAMETER[\"Latitude_Of_Origin\",46.66964837722222],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7590]]",
        7591=>"PROJCS[\"NAD_1983_2011_WISCRS_Brown_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",103674.333],PARAMETER[\"False_Northing\",15091.833],PARAMETER[\"Central_Meridian\",-88.0],PARAMETER[\"Scale_Factor\",1.00002],PARAMETER[\"Latitude_Of_Origin\",43.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7591]]",
        7592=>"PROJCS[\"NAD_1983_2011_WISCRS_Buffalo_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",574999.999],PARAMETER[\"False_Northing\",0.016],PARAMETER[\"Central_Meridian\",-91.79722222222222],PARAMETER[\"Scale_Factor\",1.0000382778],PARAMETER[\"Latitude_Of_Origin\",43.48138888888889],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7592]]",
        7593=>"PROJCS[\"NAD_1983_2011_WISCRS_Burnett_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",209999.999],PARAMETER[\"False_Northing\",195032.104],PARAMETER[\"Central_Meridian\",-92.45777777777778],PARAMETER[\"Standard_Parallel_1\",45.89871486583333],PARAMETER[\"Scale_Factor\",1.0000383841],PARAMETER[\"Latitude_Of_Origin\",45.89871486583333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7593]]",
        7594=>"PROJCS[\"NAD_1983_2011_WISCRS_Calumet_Fond_du_Lac_Outagamie_Winnebago_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",802999.999],PARAMETER[\"False_Northing\",0.016],PARAMETER[\"Central_Meridian\",-88.5],PARAMETER[\"Scale_Factor\",1.0000286569],PARAMETER[\"Latitude_Of_Origin\",42.71944444444445],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7594]]",
        7595=>"PROJCS[\"NAD_1983_2011_WISCRS_Chippewa_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",197000.0],PARAMETER[\"False_Northing\",144656.648],PARAMETER[\"Central_Meridian\",-91.29444444444444],PARAMETER[\"Standard_Parallel_1\",44.97785689861112],PARAMETER[\"Scale_Factor\",1.0000391127],PARAMETER[\"Latitude_Of_Origin\",44.97785689861112],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7595]]",
        7596=>"PROJCS[\"NAD_1983_2011_WISCRS_Clark_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",655999.997],PARAMETER[\"False_Northing\",0.028],PARAMETER[\"Central_Meridian\",-90.70833333333334],PARAMETER[\"Scale_Factor\",1.0000463003],PARAMETER[\"Latitude_Of_Origin\",43.6],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7596]]",
        7597=>"PROJCS[\"NAD_1983_2011_WISCRS_Columbia_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",554999.999],PARAMETER[\"False_Northing\",366041.307],PARAMETER[\"Central_Meridian\",-89.39444444444445],PARAMETER[\"Standard_Parallel_1\",43.46254664583333],PARAMETER[\"Scale_Factor\",1.00003498],PARAMETER[\"Latitude_Of_Origin\",43.46254664583333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7597]]",
        7598=>"PROJCS[\"NAD_1983_2011_WISCRS_Crawford_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",373000.0],PARAMETER[\"False_Northing\",176190.987],PARAMETER[\"Central_Meridian\",-90.9388888888889],PARAMETER[\"Standard_Parallel_1\",43.200055605],PARAMETER[\"Scale_Factor\",1.0000349151],PARAMETER[\"Latitude_Of_Origin\",43.200055605],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7598]]",
        7599=>"PROJCS[\"NAD_1983_2011_WISCRS_Dane_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",811000.0],PARAMETER[\"False_Northing\",480943.886],PARAMETER[\"Central_Meridian\",-89.42222222222223],PARAMETER[\"Standard_Parallel_1\",43.0695160375],PARAMETER[\"Scale_Factor\",1.0000384786],PARAMETER[\"Latitude_Of_Origin\",43.0695160375],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7599]]",
        7600=>"PROJCS[\"NAD_1983_2011_WISCRS_Dodge_and_Jefferson_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",863999.999],PARAMETER[\"False_Northing\",0.025],PARAMETER[\"Central_Meridian\",-88.775],PARAMETER[\"Scale_Factor\",1.0000346418],PARAMETER[\"Latitude_Of_Origin\",41.47222222222222],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7600]]",
        7601=>"PROJCS[\"NAD_1983_2011_WISCRS_Door_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",521000.0],PARAMETER[\"False_Northing\",0.008],PARAMETER[\"Central_Meridian\",-87.27222222222223],PARAMETER[\"Scale_Factor\",1.0000187521],PARAMETER[\"Latitude_Of_Origin\",44.4],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7601]]",
        7602=>"PROJCS[\"NAD_1983_2011_WISCRS_Douglas_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",194000.0],PARAMETER[\"False_Northing\",0.013],PARAMETER[\"Central_Meridian\",-91.91666666666667],PARAMETER[\"Scale_Factor\",1.0000385418],PARAMETER[\"Latitude_Of_Origin\",45.88333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7602]]",
        7603=>"PROJCS[\"NAD_1983_2011_WISCRS_Dunn_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",170000.001],PARAMETER[\"False_Northing\",0.01],PARAMETER[\"Central_Meridian\",-91.89444444444445],PARAMETER[\"Scale_Factor\",1.0000410324],PARAMETER[\"Latitude_Of_Origin\",44.40833333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7603]]",
        7604=>"PROJCS[\"NAD_1983_2011_WISCRS_Eau_Claire_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",394000.0],PARAMETER[\"False_Northing\",300812.797],PARAMETER[\"Central_Meridian\",-91.28888888888889],PARAMETER[\"Standard_Parallel_1\",44.87228112638889],PARAMETER[\"Scale_Factor\",1.000035079],PARAMETER[\"Latitude_Of_Origin\",44.87228112638889],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7604]]",
        7605=>"PROJCS[\"NAD_1983_2011_WISCRS_Florence_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",438000.004],PARAMETER[\"False_Northing\",0.021],PARAMETER[\"Central_Meridian\",-88.14166666666668],PARAMETER[\"Scale_Factor\",1.0000552095],PARAMETER[\"Latitude_Of_Origin\",45.43888888888888],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7605]]",
        7606=>"PROJCS[\"NAD_1983_2011_WISCRS_Forest_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",905000.005],PARAMETER[\"False_Northing\",0.052],PARAMETER[\"Central_Meridian\",-88.63333333333334],PARAMETER[\"Scale_Factor\",1.0000673004],PARAMETER[\"Latitude_Of_Origin\",44.00555555555555],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7606]]",
        7607=>"PROJCS[\"NAD_1983_2011_WISCRS_Grant_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",794999.998],PARAMETER[\"False_Northing\",0.033],PARAMETER[\"Central_Meridian\",-90.8],PARAMETER[\"Scale_Factor\",1.0000349452],PARAMETER[\"Latitude_Of_Origin\",41.41111111111111],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7607]]",
        7608=>"PROJCS[\"NAD_1983_2011_WISCRS_Green_and_Lafayette_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",558000.0],PARAMETER[\"False_Northing\",150361.559],PARAMETER[\"Central_Meridian\",-89.83888888888889],PARAMETER[\"Standard_Parallel_1\",42.63756227694444],PARAMETER[\"Scale_Factor\",1.0000390487],PARAMETER[\"Latitude_Of_Origin\",42.63756227694444],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7608]]",
        7609=>"PROJCS[\"NAD_1983_2011_WISCRS_Green_Lake_and_Marquette_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",495000.0],PARAMETER[\"False_Northing\",259746.132],PARAMETER[\"Central_Meridian\",-89.24166666666667],PARAMETER[\"Standard_Parallel_1\",43.80700011777778],PARAMETER[\"Scale_Factor\",1.0000344057],PARAMETER[\"Latitude_Of_Origin\",43.80700011777778],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7609]]",
        7610=>"PROJCS[\"NAD_1983_2011_WISCRS_Iowa_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",371000.0],PARAMETER[\"False_Northing\",0.015],PARAMETER[\"Central_Meridian\",-90.16111111111111],PARAMETER[\"Scale_Factor\",1.0000394961],PARAMETER[\"Latitude_Of_Origin\",42.53888888888888],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7610]]",
        7611=>"PROJCS[\"NAD_1983_2011_WISCRS_Iron_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",725000.0],PARAMETER[\"False_Northing\",0.028],PARAMETER[\"Central_Meridian\",-90.25555555555556],PARAMETER[\"Scale_Factor\",1.0000677153],PARAMETER[\"Latitude_Of_Origin\",45.43333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7611]]",
        7612=>"PROJCS[\"NAD_1983_2011_WISCRS_Jackson_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",88582.5],PARAMETER[\"False_Northing\",82020.833],PARAMETER[\"Central_Meridian\",-90.84429651944444],PARAMETER[\"Scale_Factor\",1.0000353],PARAMETER[\"Latitude_Of_Origin\",44.25333512777778],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7612]]",
        7613=>"PROJCS[\"NAD_1983_2011_WISCRS_Kenosha_Milwaukee_Ozaukee_Racine_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",610000.003],PARAMETER[\"False_Northing\",0.003],PARAMETER[\"Central_Meridian\",-87.89444444444445],PARAMETER[\"Scale_Factor\",1.0000260649],PARAMETER[\"Latitude_Of_Origin\",42.21666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7613]]",
        7614=>"PROJCS[\"NAD_1983_2011_WISCRS_Kewaunee_Manitowoc_Sheboygan_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",262000.006],PARAMETER[\"False_Northing\",0.004],PARAMETER[\"Central_Meridian\",-87.55],PARAMETER[\"Scale_Factor\",1.0000233704],PARAMETER[\"Latitude_Of_Origin\",43.26666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7614]]",
        7615=>"PROJCS[\"NAD_1983_2011_WISCRS_La_Crosse_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",427999.996],PARAMETER[\"False_Northing\",0.011],PARAMETER[\"Central_Meridian\",-91.31666666666666],PARAMETER[\"Scale_Factor\",1.0000319985],PARAMETER[\"Latitude_Of_Origin\",43.45111111111111],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7615]]",
        7616=>"PROJCS[\"NAD_1983_2011_WISCRS_Langlade_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",651000.0],PARAMETER[\"False_Northing\",345405.421],PARAMETER[\"Central_Meridian\",-89.03333333333333],PARAMETER[\"Standard_Parallel_1\",45.15423710527778],PARAMETER[\"Scale_Factor\",1.0000627024],PARAMETER[\"Latitude_Of_Origin\",45.15423710527778],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7616]]",
        7617=>"PROJCS[\"NAD_1983_2011_WISCRS_Lincoln_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",381000.0],PARAMETER[\"False_Northing\",0.019],PARAMETER[\"Central_Meridian\",-89.73333333333333],PARAMETER[\"Scale_Factor\",1.0000599003],PARAMETER[\"Latitude_Of_Origin\",44.84444444444445],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7617]]",
        7618=>"PROJCS[\"NAD_1983_2011_WISCRS_Marathon_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",245000.0],PARAMETER[\"False_Northing\",180607.47],PARAMETER[\"Central_Meridian\",-89.77],PARAMETER[\"Standard_Parallel_1\",44.90090442361111],PARAMETER[\"Scale_Factor\",1.000053289],PARAMETER[\"Latitude_Of_Origin\",44.90090442361111],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7618]]",
        7619=>"PROJCS[\"NAD_1983_2011_WISCRS_Marinette_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",783000.007],PARAMETER[\"False_Northing\",0.01],PARAMETER[\"Central_Meridian\",-87.71111111111111],PARAMETER[\"Scale_Factor\",1.0000234982],PARAMETER[\"Latitude_Of_Origin\",44.69166666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7619]]",
        7620=>"PROJCS[\"NAD_1983_2011_WISCRS_Menominee_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",346000.004],PARAMETER[\"False_Northing\",0.01],PARAMETER[\"Central_Meridian\",-88.41666666666667],PARAMETER[\"Scale_Factor\",1.0000362499],PARAMETER[\"Latitude_Of_Origin\",44.71666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7620]]",
        7621=>"PROJCS[\"NAD_1983_2011_WISCRS_Monroe_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",671000.0],PARAMETER[\"False_Northing\",400012.278],PARAMETER[\"Central_Meridian\",-90.64166666666668],PARAMETER[\"Standard_Parallel_1\",44.00007392861111],PARAMETER[\"Scale_Factor\",1.0000434122],PARAMETER[\"Latitude_Of_Origin\",44.00007392861111],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7621]]",
        7622=>"PROJCS[\"NAD_1983_2011_WISCRS_Oconto_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",600000.006],PARAMETER[\"False_Northing\",0.011],PARAMETER[\"Central_Meridian\",-87.90833333333335],PARAMETER[\"Scale_Factor\",1.0000236869],PARAMETER[\"Latitude_Of_Origin\",44.39722222222222],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7622]]",
        7623=>"PROJCS[\"NAD_1983_2011_WISCRS_Oneida_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",230000.0],PARAMETER[\"False_Northing\",188936.744],PARAMETER[\"Central_Meridian\",-89.54444444444444],PARAMETER[\"Standard_Parallel_1\",45.70422377027778],PARAMETER[\"Scale_Factor\",1.0000686968],PARAMETER[\"Latitude_Of_Origin\",45.70422377027778],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7623]]",
        7624=>"PROJCS[\"NAD_1983_2011_WISCRS_Pepin_and_Pierce_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",550000.0],PARAMETER[\"False_Northing\",282260.222],PARAMETER[\"Central_Meridian\",-92.22777777777777],PARAMETER[\"Standard_Parallel_1\",44.63614887194444],PARAMETER[\"Scale_Factor\",1.0000362977],PARAMETER[\"Latitude_Of_Origin\",44.63614887194444],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7624]]",
        7625=>"PROJCS[\"NAD_1983_2011_WISCRS_Polk_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",464999.996],PARAMETER[\"False_Northing\",0.019],PARAMETER[\"Central_Meridian\",-92.63333333333334],PARAMETER[\"Scale_Factor\",1.0000433849],PARAMETER[\"Latitude_Of_Origin\",44.66111111111111],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7625]]",
        7626=>"PROJCS[\"NAD_1983_2011_WISCRS_Portage_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",185000.0],PARAMETER[\"False_Northing\",164114.46],PARAMETER[\"Central_Meridian\",-89.5],PARAMETER[\"Standard_Parallel_1\",44.41682397527777],PARAMETER[\"Scale_Factor\",1.000039936],PARAMETER[\"Latitude_Of_Origin\",44.41682397527777],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7626]]",
        7627=>"PROJCS[\"NAD_1983_2011_WISCRS_Price_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",747999.995],PARAMETER[\"False_Northing\",0.036],PARAMETER[\"Central_Meridian\",-90.48888888888889],PARAMETER[\"Scale_Factor\",1.0000649554],PARAMETER[\"Latitude_Of_Origin\",44.55555555555555],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7627]]",
        7628=>"PROJCS[\"NAD_1983_2011_WISCRS_Richland_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",664000.0],PARAMETER[\"False_Northing\",440469.675],PARAMETER[\"Central_Meridian\",-90.43055555555556],PARAMETER[\"Standard_Parallel_1\",43.3223129275],PARAMETER[\"Scale_Factor\",1.0000375653],PARAMETER[\"Latitude_Of_Origin\",43.3223129275],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7628]]",
        7629=>"PROJCS[\"NAD_1983_2011_WISCRS_Rock_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",480000.0],PARAMETER[\"False_Northing\",0.022],PARAMETER[\"Central_Meridian\",-89.07222222222222],PARAMETER[\"Scale_Factor\",1.0000337311],PARAMETER[\"Latitude_Of_Origin\",41.94444444444444],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7629]]",
        7630=>"PROJCS[\"NAD_1983_2011_WISCRS_Rusk_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",822000.001],PARAMETER[\"False_Northing\",0.077],PARAMETER[\"Central_Meridian\",-91.06666666666666],PARAMETER[\"Scale_Factor\",1.0000495976],PARAMETER[\"Latitude_Of_Origin\",43.91944444444444],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7630]]",
        7631=>"PROJCS[\"NAD_1983_2011_WISCRS_Sauk_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",609000.001],PARAMETER[\"False_Northing\",0.017],PARAMETER[\"Central_Meridian\",-89.9],PARAMETER[\"Scale_Factor\",1.0000373868],PARAMETER[\"Latitude_Of_Origin\",42.81944444444445],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7631]]",
        7632=>"PROJCS[\"NAD_1983_2011_WISCRS_Sawyer_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",711000.001],PARAMETER[\"False_Northing\",396108.667],PARAMETER[\"Central_Meridian\",-91.11666666666666],PARAMETER[\"Standard_Parallel_1\",45.90009913138888],PARAMETER[\"Scale_Factor\",1.0000573461],PARAMETER[\"Latitude_Of_Origin\",45.90009913138888],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7632]]",
        7633=>"PROJCS[\"NAD_1983_2011_WISCRS_Shawano_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",861000.001],PARAMETER[\"False_Northing\",0.031],PARAMETER[\"Central_Meridian\",-88.60555555555555],PARAMETER[\"Scale_Factor\",1.000032144],PARAMETER[\"Latitude_Of_Origin\",44.03611111111111],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7633]]",
        7634=>"PROJCS[\"NAD_1983_2011_WISCRS_St_Croix_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",542999.997],PARAMETER[\"False_Northing\",0.034],PARAMETER[\"Central_Meridian\",-92.63333333333334],PARAMETER[\"Scale_Factor\",1.0000381803],PARAMETER[\"Latitude_Of_Origin\",44.03611111111111],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7634]]",
        7635=>"PROJCS[\"NAD_1983_2011_WISCRS_Taylor_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",614000.0],PARAMETER[\"False_Northing\",353499.136],PARAMETER[\"Central_Meridian\",-90.48333333333333],PARAMETER[\"Standard_Parallel_1\",45.17782208583333],PARAMETER[\"Scale_Factor\",1.0000597566],PARAMETER[\"Latitude_Of_Origin\",45.17782208583333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7635]]",
        7636=>"PROJCS[\"NAD_1983_2011_WISCRS_Trempealeau_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",843000.0],PARAMETER[\"False_Northing\",0.013],PARAMETER[\"Central_Meridian\",-91.36666666666666],PARAMETER[\"Scale_Factor\",1.0000361538],PARAMETER[\"Latitude_Of_Origin\",43.16111111111111],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7636]]",
        7637=>"PROJCS[\"NAD_1983_2011_WISCRS_Vernon_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",730000.0],PARAMETER[\"False_Northing\",155944.768],PARAMETER[\"Central_Meridian\",-90.78333333333333],PARAMETER[\"Standard_Parallel_1\",43.57503293972223],PARAMETER[\"Scale_Factor\",1.0000408158],PARAMETER[\"Latitude_Of_Origin\",43.57503293972223],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7637]]",
        7638=>"PROJCS[\"NAD_1983_2011_WISCRS_Vilas_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",441000.0],PARAMETER[\"False_Northing\",165147.666],PARAMETER[\"Central_Meridian\",-89.48888888888889],PARAMETER[\"Standard_Parallel_1\",46.07784409055556],PARAMETER[\"Scale_Factor\",1.0000730142],PARAMETER[\"Latitude_Of_Origin\",46.07784409055556],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7638]]",
        7639=>"PROJCS[\"NAD_1983_2011_WISCRS_Walworth_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",763000.0],PARAMETER[\"False_Northing\",364461.943],PARAMETER[\"Central_Meridian\",-88.54166666666667],PARAMETER[\"Standard_Parallel_1\",42.66946209694444],PARAMETER[\"Scale_Factor\",1.0000367192],PARAMETER[\"Latitude_Of_Origin\",42.66946209694444],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7639]]",
        7640=>"PROJCS[\"NAD_1983_2011_WISCRS_Washburn_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",768000.0],PARAMETER[\"False_Northing\",617973.193],PARAMETER[\"Central_Meridian\",-91.78333333333333],PARAMETER[\"Standard_Parallel_1\",45.96121983333334],PARAMETER[\"Scale_Factor\",1.0000475376],PARAMETER[\"Latitude_Of_Origin\",45.96121983333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7640]]",
        7641=>"PROJCS[\"NAD_1983_2011_WISCRS_Washington_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",394000.004],PARAMETER[\"False_Northing\",0.01],PARAMETER[\"Central_Meridian\",-88.06388888888888],PARAMETER[\"Scale_Factor\",1.00003738],PARAMETER[\"Latitude_Of_Origin\",42.91805555555555],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7641]]",
        7642=>"PROJCS[\"NAD_1983_2011_WISCRS_Waukesha_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",685000.001],PARAMETER[\"False_Northing\",0.011],PARAMETER[\"Central_Meridian\",-88.225],PARAMETER[\"Scale_Factor\",1.0000346179],PARAMETER[\"Latitude_Of_Origin\",42.56944444444445],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7642]]",
        7643=>"PROJCS[\"NAD_1983_2011_WISCRS_Waupaca_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",607000.003],PARAMETER[\"False_Northing\",0.023],PARAMETER[\"Central_Meridian\",-88.81666666666666],PARAMETER[\"Scale_Factor\",1.0000333645],PARAMETER[\"Latitude_Of_Origin\",43.42027777777778],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7643]]",
        7644=>"PROJCS[\"NAD_1983_2011_WISCRS_Waushara_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",394000.0],PARAMETER[\"False_Northing\",147866.367],PARAMETER[\"Central_Meridian\",-89.24166666666667],PARAMETER[\"Standard_Parallel_1\",44.11394404583334],PARAMETER[\"Scale_Factor\",1.0000392096],PARAMETER[\"Latitude_Of_Origin\",44.11394404583334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7644]]",
        7645=>"PROJCS[\"NAD_1983_2011_WISCRS_Wood_Feet\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",684000.001],PARAMETER[\"False_Northing\",441566.551],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",44.36259546944444],PARAMETER[\"Scale_Factor\",1.0000421209],PARAMETER[\"Latitude_Of_Origin\",44.36259546944444],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",7645]]",
        7755=>"PROJCS[\"WGS_1984_India_NSF_LCC\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4000000.0],PARAMETER[\"False_Northing\",4000000.0],PARAMETER[\"Central_Meridian\",80.0],PARAMETER[\"Standard_Parallel_1\",12.472955],PARAMETER[\"Standard_Parallel_2\",35.17280444444444],PARAMETER[\"Latitude_Of_Origin\",24.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7755]]",
        7756=>"PROJCS[\"WGS_1984_Andhra_Pradesh\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",80.875],PARAMETER[\"Standard_Parallel_1\",13.75],PARAMETER[\"Standard_Parallel_2\",18.75],PARAMETER[\"Latitude_Of_Origin\",16.25543298],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7756]]",
        7757=>"PROJCS[\"WGS_1984_Arunachal_Pradesh\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",94.5],PARAMETER[\"Standard_Parallel_1\",27.0],PARAMETER[\"Standard_Parallel_2\",29.0],PARAMETER[\"Latitude_Of_Origin\",28.00157897],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7757]]",
        7758=>"PROJCS[\"WGS_1984_Assam\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",92.75],PARAMETER[\"Standard_Parallel_1\",24.66666666666667],PARAMETER[\"Standard_Parallel_2\",27.33333333333333],PARAMETER[\"Latitude_Of_Origin\",26.00257703],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7758]]",
        7759=>"PROJCS[\"WGS_1984_Bihar\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",85.875],PARAMETER[\"Standard_Parallel_1\",24.625],PARAMETER[\"Standard_Parallel_2\",27.125],PARAMETER[\"Latitude_Of_Origin\",25.87725247],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7759]]",
        7760=>"PROJCS[\"WGS_1984_Delhi\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",77.0],PARAMETER[\"Standard_Parallel_1\",28.375],PARAMETER[\"Standard_Parallel_2\",28.875],PARAMETER[\"Latitude_Of_Origin\",28.62510126],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7760]]",
        7761=>"PROJCS[\"WGS_1984_Gujarat\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",71.375],PARAMETER[\"Standard_Parallel_1\",20.79166666666667],PARAMETER[\"Standard_Parallel_2\",23.95833333333334],PARAMETER[\"Latitude_Of_Origin\",22.37807121],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7761]]",
        7762=>"PROJCS[\"WGS_1984_Haryana\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",76.0],PARAMETER[\"Standard_Parallel_1\",28.08333333333333],PARAMETER[\"Standard_Parallel_2\",30.41666666666667],PARAMETER[\"Latitude_Of_Origin\",29.25226266],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7762]]",
        7763=>"PROJCS[\"WGS_1984_Himachal_Pradesh\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",77.375],PARAMETER[\"Standard_Parallel_1\",30.75],PARAMETER[\"Standard_Parallel_2\",32.75],PARAMETER[\"Latitude_Of_Origin\",31.75183497],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7763]]",
        7764=>"PROJCS[\"WGS_1984_Jammu_and_Kashmir\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",76.5],PARAMETER[\"Standard_Parallel_1\",33.08333333333334],PARAMETER[\"Standard_Parallel_2\",36.41666666666666],PARAMETER[\"Latitude_Of_Origin\",34.75570874],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7764]]",
        7765=>"PROJCS[\"WGS_1984_JHARKHAND\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",85.625],PARAMETER[\"Standard_Parallel_1\",22.54166666666667],PARAMETER[\"Standard_Parallel_2\",24.70833333333333],PARAMETER[\"Latitude_Of_Origin\",23.62652682],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7765]]",
        7766=>"PROJCS[\"WGS_1984_Madhya_Pradesh\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",78.375],PARAMETER[\"Standard_Parallel_1\",22.0],PARAMETER[\"Standard_Parallel_2\",26.0],PARAMETER[\"Latitude_Of_Origin\",24.00529821],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7766]]",
        7767=>"PROJCS[\"WGS_1984_Maharashtra\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",76.75],PARAMETER[\"Standard_Parallel_1\",16.625],PARAMETER[\"Standard_Parallel_2\",21.125],PARAMETER[\"Latitude_Of_Origin\",18.88015774],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7767]]",
        7768=>"PROJCS[\"WGS_1984_Manipur\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",94.0],PARAMETER[\"Standard_Parallel_1\",24.08333333333333],PARAMETER[\"Standard_Parallel_2\",25.41666666666667],PARAMETER[\"Latitude_Of_Origin\",24.75060911],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7768]]",
        7769=>"PROJCS[\"WGS_1984_Meghalaya\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",91.375],PARAMETER[\"Standard_Parallel_1\",25.20833333333334],PARAMETER[\"Standard_Parallel_2\",26.04166666666667],PARAMETER[\"Latitude_Of_Origin\",25.62524747],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7769]]",
        7770=>"PROJCS[\"WGS_1984_Nagaland\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",94.375],PARAMETER[\"Standard_Parallel_1\",25.375],PARAMETER[\"Standard_Parallel_2\",26.875],PARAMETER[\"Latitude_Of_Origin\",26.12581974],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7770]]",
        7771=>"PROJCS[\"WGS_1984_India_Northeast\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",93.5],PARAMETER[\"Standard_Parallel_1\",23.04166666666667],PARAMETER[\"Standard_Parallel_2\",28.20833333333334],PARAMETER[\"Latitude_Of_Origin\",25.63452135],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7771]]",
        7772=>"PROJCS[\"WGS_1984_Orissa\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",84.375],PARAMETER[\"Standard_Parallel_1\",18.58333333333334],PARAMETER[\"Standard_Parallel_2\",21.91666666666667],PARAMETER[\"Latitude_Of_Origin\",20.25305174],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7772]]",
        7773=>"PROJCS[\"WGS_1984_Punjab\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",75.375],PARAMETER[\"Standard_Parallel_1\",30.0],PARAMETER[\"Standard_Parallel_2\",32.0],PARAMETER[\"Latitude_Of_Origin\",31.00178226],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7773]]",
        7774=>"PROJCS[\"WGS_1984_Rajasthan\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",73.875],PARAMETER[\"Standard_Parallel_1\",24.29166666666666],PARAMETER[\"Standard_Parallel_2\",29.45833333333334],PARAMETER[\"Latitude_Of_Origin\",26.88505546],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7774]]",
        7775=>"PROJCS[\"WGS_1984_Uttar_Pradesh\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",80.875],PARAMETER[\"Standard_Parallel_1\",24.875],PARAMETER[\"Standard_Parallel_2\",29.375],PARAMETER[\"Latitude_Of_Origin\",27.13270823],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7775]]",
        7776=>"PROJCS[\"WGS_1984_Uttaranchal\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",79.375],PARAMETER[\"Standard_Parallel_1\",29.0],PARAMETER[\"Standard_Parallel_2\",31.0],PARAMETER[\"Latitude_Of_Origin\",30.0017132],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7776]]",
        7777=>"PROJCS[\"WGS_1984_Andaman_and_Nicobar_Islands\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",93.25],PARAMETER[\"Scale_Factor\",0.9999428],PARAMETER[\"Latitude_Of_Origin\",10.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7777]]",
        7778=>"PROJCS[\"WGS_1984_Chhattisgarh\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",82.25],PARAMETER[\"Scale_Factor\",0.9998332],PARAMETER[\"Latitude_Of_Origin\",21.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7778]]",
        7779=>"PROJCS[\"WGS_1984_Goa\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",74.0],PARAMETER[\"Scale_Factor\",0.9999913],PARAMETER[\"Latitude_Of_Origin\",15.375],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7779]]",
        7780=>"PROJCS[\"WGS_1984_Karnataka\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",76.375],PARAMETER[\"Scale_Factor\",0.9998012],PARAMETER[\"Latitude_Of_Origin\",15.125],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7780]]",
        7781=>"PROJCS[\"WGS_1984_Kerala\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",76.0],PARAMETER[\"Scale_Factor\",0.9999177],PARAMETER[\"Latitude_Of_Origin\",10.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7781]]",
        7782=>"PROJCS[\"WGS_1984_Lakshadweep\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",73.125],PARAMETER[\"Scale_Factor\",0.9999536],PARAMETER[\"Latitude_Of_Origin\",10.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7782]]",
        7783=>"PROJCS[\"WGS_1984_Mizoram\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",92.75],PARAMETER[\"Scale_Factor\",0.9999821],PARAMETER[\"Latitude_Of_Origin\",23.125],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7783]]",
        7784=>"PROJCS[\"WGS_1984_Sikkim\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",88.5],PARAMETER[\"Scale_Factor\",0.9999926],PARAMETER[\"Latitude_Of_Origin\",27.625],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7784]]",
        7785=>"PROJCS[\"WGS_1984_Tamil_Nadu\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",78.375],PARAMETER[\"Scale_Factor\",0.9997942],PARAMETER[\"Latitude_Of_Origin\",10.875],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7785]]",
        7786=>"PROJCS[\"WGS_1984_Tripura\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",91.75],PARAMETER[\"Scale_Factor\",0.9999822],PARAMETER[\"Latitude_Of_Origin\",23.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7786]]",
        7787=>"PROJCS[\"WGS_1984_West_Bengal\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",87.875],PARAMETER[\"Scale_Factor\",0.9998584],PARAMETER[\"Latitude_Of_Origin\",24.375],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7787]]",
        7791=>"PROJCS[\"RDN2008_UTM_zone_32N\",GEOGCS[\"GCS_RDN2008\",DATUM[\"D_Rete_Dinamica_Nazionale_2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7791]]",
        7792=>"PROJCS[\"RDN2008_UTM_zone_33N\",GEOGCS[\"GCS_RDN2008\",DATUM[\"D_Rete_Dinamica_Nazionale_2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7792]]",
        7793=>"PROJCS[\"RDN2008_UTM_zone_34N\",GEOGCS[\"GCS_RDN2008\",DATUM[\"D_Rete_Dinamica_Nazionale_2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7793]]",
        7794=>"PROJCS[\"RDN2008_Italy_zone_(E-N)\",GEOGCS[\"GCS_RDN2008\",DATUM[\"D_Rete_Dinamica_Nazionale_2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",7000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",0.9985],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7794]]",
        7795=>"PROJCS[\"RDN2008_Zone_12_(E-N)\",GEOGCS[\"GCS_RDN2008\",DATUM[\"D_Rete_Dinamica_Nazionale_2008\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",3000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7795]]",
        7799=>"PROJCS[\"BGS2005_UTM_zone_34N_(N-E)\",GEOGCS[\"BGS2005\",DATUM[\"Bulgaria_Geodetic_System_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7799]]",
        7800=>"PROJCS[\"BGS2005_UTM_zone_35N_(N-E)\",GEOGCS[\"BGS2005\",DATUM[\"Bulgaria_Geodetic_System_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7800]]",
        7801=>"PROJCS[\"BGS2005_CCS2005\",GEOGCS[\"BGS2005\",DATUM[\"Bulgaria_Geodetic_System_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",4725824.3591],PARAMETER[\"Central_Meridian\",25.5],PARAMETER[\"Standard_Parallel_1\",42.0],PARAMETER[\"Standard_Parallel_2\",43.33333333333334],PARAMETER[\"Latitude_Of_Origin\",42.66787568333332],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7801]]",
        7803=>"PROJCS[\"BGS2005_UTM_zone_34N\",GEOGCS[\"BGS2005\",DATUM[\"Bulgaria_Geodetic_System_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7803]]",
        7804=>"PROJCS[\"BGS2005_UTM_zone_35N\",GEOGCS[\"BGS2005\",DATUM[\"Bulgaria_Geodetic_System_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7804]]",
        7805=>"PROJCS[\"BGS2005_UTM_zone_36N\",GEOGCS[\"BGS2005\",DATUM[\"Bulgaria_Geodetic_System_2005\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7805]]",
        7825=>"PROJCS[\"Pulkovo_1942_CS63_zone_X1\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",23.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.08333333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7825]]",
        7826=>"PROJCS[\"Pulkovo_1942_CS63_zone_X2\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",26.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.08333333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7826]]",
        7827=>"PROJCS[\"Pulkovo_1942_CS63_zone_X3\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",3300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",29.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.08333333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7827]]",
        7828=>"PROJCS[\"Pulkovo_1942_CS63_zone_X4\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",4300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",32.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.08333333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7828]]",
        7829=>"PROJCS[\"Pulkovo_1942_CS63_zone_X5\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",5300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",35.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.08333333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7829]]",
        7830=>"PROJCS[\"Pulkovo_1942_CS63_zone_X6\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",6300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",38.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.08333333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7830]]",
        7831=>"PROJCS[\"Pulkovo_1942_CS63_zone_X7\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",7300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",41.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.08333333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7831]]",
        7845=>"PROJCS[\"GDA2020_GA_LCC\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",134.0],PARAMETER[\"Standard_Parallel_1\",-18.0],PARAMETER[\"Standard_Parallel_2\",-36.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7845]]",
        7846=>"PROJCS[\"GDA2020_MGA_Zone_46\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7846]]",
        7847=>"PROJCS[\"GDA2020_MGA_Zone_47\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7847]]",
        7848=>"PROJCS[\"GDA2020_MGA_Zone_48\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7848]]",
        7849=>"PROJCS[\"GDA2020_MGA_Zone_49\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7849]]",
        7850=>"PROJCS[\"GDA2020_MGA_Zone_50\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7850]]",
        7851=>"PROJCS[\"GDA2020_MGA_Zone_51\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7851]]",
        7852=>"PROJCS[\"GDA2020_MGA_Zone_52\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7852]]",
        7853=>"PROJCS[\"GDA2020_MGA_Zone_53\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7853]]",
        7854=>"PROJCS[\"GDA2020_MGA_Zone_54\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7854]]",
        7855=>"PROJCS[\"GDA2020_MGA_Zone_55\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7855]]",
        7856=>"PROJCS[\"GDA2020_MGA_Zone_56\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7856]]",
        7857=>"PROJCS[\"GDA2020_MGA_Zone_57\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7857]]",
        7858=>"PROJCS[\"GDA2020_MGA_Zone_58\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7858]]",
        7859=>"PROJCS[\"GDA2020_MGA_Zone_59\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7859]]",
        7877=>"PROJCS[\"Astro_DOS_71_4_SHLG71\",GEOGCS[\"GCS_DOS_71_4\",DATUM[\"D_DOS_71_4\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",-5.716666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-15.96666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7877]]",
        7878=>"PROJCS[\"Astro_DOS_71_4_UTM_zone_30S\",GEOGCS[\"GCS_DOS_71_4\",DATUM[\"D_DOS_71_4\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7878]]",
        7882=>"PROJCS[\"St_Helena_Tritan_SHLG(Tritan)\",GEOGCS[\"St_Helena_Tritan\",DATUM[\"St_Helena_Tritan\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",299483.737],PARAMETER[\"False_Northing\",2000527.879],PARAMETER[\"Central_Meridian\",-5.716666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-15.96666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7882]]",
        7883=>"PROJCS[\"St_Helena_Tritan_UTM_zone_30S\",GEOGCS[\"St_Helena_Tritan\",DATUM[\"St_Helena_Tritan\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7883]]",
        7887=>"PROJCS[\"SHMG2015\",GEOGCS[\"SHGD2015\",DATUM[\"St_Helena_Geodetic_Datum_2015\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7887]]",
        7899=>"PROJCS[\"GDA2020_Vicgrid\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",2500000.0],PARAMETER[\"Central_Meridian\",145.0],PARAMETER[\"Standard_Parallel_1\",-36.0],PARAMETER[\"Standard_Parallel_2\",-38.0],PARAMETER[\"Latitude_Of_Origin\",-37.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7899]]",
        7991=>"PROJCS[\"NAD27_MTM_zone_10\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7991]]",
        7992=>"PROJCS[\"Malongo_1987_UTM_zone_33S\",GEOGCS[\"GCS_Malongo_1987\",DATUM[\"D_Malongo_1987\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",7992]]",
        8013=>"PROJCS[\"GDA2020_ALB2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",4100000.0],PARAMETER[\"Central_Meridian\",117.8833333333333],PARAMETER[\"Scale_Factor\",1.0000044],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8013]]",
        8014=>"PROJCS[\"GDA2020_BIO2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",60000.0],PARAMETER[\"False_Northing\",2700000.0],PARAMETER[\"Central_Meridian\",115.25],PARAMETER[\"Scale_Factor\",1.0000022],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8014]]",
        8015=>"PROJCS[\"GDA2020_BRO2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",2300000.0],PARAMETER[\"Central_Meridian\",122.3333333333333],PARAMETER[\"Scale_Factor\",1.00000298],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8015]]",
        8016=>"PROJCS[\"GDA2020_BCG2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",4000000.0],PARAMETER[\"Central_Meridian\",115.4333333333333],PARAMETER[\"Scale_Factor\",0.99999592],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8016]]",
        8017=>"PROJCS[\"GDA2020_CARN2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",3050000.0],PARAMETER[\"Central_Meridian\",113.6666666666667],PARAMETER[\"Scale_Factor\",0.99999796],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8017]]",
        8018=>"PROJCS[\"GDA2020_CIG2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",1400000.0],PARAMETER[\"Central_Meridian\",105.625],PARAMETER[\"Scale_Factor\",1.00002514],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8018]]",
        8019=>"PROJCS[\"GDA2020_CKIG2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",1600000.0],PARAMETER[\"Central_Meridian\",96.875],PARAMETER[\"Scale_Factor\",0.99999387],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8019]]",
        8020=>"PROJCS[\"GDA2020_COL2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",40000.0],PARAMETER[\"False_Northing\",4100000.0],PARAMETER[\"Central_Meridian\",115.9333333333333],PARAMETER[\"Scale_Factor\",1.000019],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8020]]",
        8021=>"PROJCS[\"GDA2020_ESP2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",4050000.0],PARAMETER[\"Central_Meridian\",121.8833333333333],PARAMETER[\"Scale_Factor\",1.0000055],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8021]]",
        8022=>"PROJCS[\"GDA2020_EXM2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",2750000.0],PARAMETER[\"Central_Meridian\",114.0666666666667],PARAMETER[\"Scale_Factor\",1.00000236],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8022]]",
        8023=>"PROJCS[\"GDA2020_GCG2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",3450000.0],PARAMETER[\"Central_Meridian\",114.5833333333333],PARAMETER[\"Scale_Factor\",1.00000628],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8023]]",
        8024=>"PROJCS[\"GDA2020_GOLD2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",60000.0],PARAMETER[\"False_Northing\",3800000.0],PARAMETER[\"Central_Meridian\",121.5],PARAMETER[\"Scale_Factor\",1.00004949],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8024]]",
        8025=>"PROJCS[\"GDA2020_JCG2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",3650000.0],PARAMETER[\"Central_Meridian\",114.9833333333333],PARAMETER[\"Scale_Factor\",1.00000314],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8025]]",
        8026=>"PROJCS[\"GDA2020_KALB2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",55000.0],PARAMETER[\"False_Northing\",3700000.0],PARAMETER[\"Central_Meridian\",114.3152777777778],PARAMETER[\"Scale_Factor\",1.000014],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8026]]",
        8027=>"PROJCS[\"GDA2020_KAR2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",2550000.0],PARAMETER[\"Central_Meridian\",116.9333333333333],PARAMETER[\"Scale_Factor\",0.9999989],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8027]]",
        8028=>"PROJCS[\"GDA2020_KUN2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",2100000.0],PARAMETER[\"Central_Meridian\",128.75],PARAMETER[\"Scale_Factor\",1.0000165],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8028]]",
        8029=>"PROJCS[\"GDA2020_LCG2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",3750000.0],PARAMETER[\"Central_Meridian\",115.3666666666667],PARAMETER[\"Scale_Factor\",1.00000157],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8029]]",
        8030=>"PROJCS[\"GDA2020_MRCG2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",4050000.0],PARAMETER[\"Central_Meridian\",115.1666666666667],PARAMETER[\"Scale_Factor\",1.0000055],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8030]]",
        8031=>"PROJCS[\"GDA2020_PCG2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",3900000.0],PARAMETER[\"Central_Meridian\",115.8166666666667],PARAMETER[\"Scale_Factor\",0.99999906],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8031]]",
        8032=>"PROJCS[\"GDA2020_PHG2020\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",2500000.0],PARAMETER[\"Central_Meridian\",118.6],PARAMETER[\"Scale_Factor\",1.00000135],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8032]]",
        8058=>"PROJCS[\"GDA2020_NSW_Lambert\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",9300000.0],PARAMETER[\"False_Northing\",4500000.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Standard_Parallel_1\",-30.75],PARAMETER[\"Standard_Parallel_2\",-35.75],PARAMETER[\"Latitude_Of_Origin\",-33.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8058]]",
        8059=>"PROJCS[\"GDA2020_South_Australia_Lambert\",GEOGCS[\"GDA2020\",DATUM[\"GDA2020\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Standard_Parallel_1\",-28.0],PARAMETER[\"Standard_Parallel_2\",-36.0],PARAMETER[\"Latitude_Of_Origin\",-32.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8059]]",
        8311=>"PROJCS[\"OCRS_Burns-Harper_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",90000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.6666666666667],PARAMETER[\"Scale_Factor\",1.00014],PARAMETER[\"Latitude_Of_Origin\",43.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8311]]",
        8312=>"PROJCS[\"OCRS_Burns-Harper_NAD_1983_2011_TM_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",295275.5906],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.6666666666667],PARAMETER[\"Scale_Factor\",1.00014],PARAMETER[\"Latitude_Of_Origin\",43.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8312]]",
        8313=>"PROJCS[\"OCRS_Canyon_City-Burns_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",20000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-119.0],PARAMETER[\"Scale_Factor\",1.00022],PARAMETER[\"Latitude_Of_Origin\",43.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8313]]",
        8314=>"PROJCS[\"OCRS_Canyon_City-Burns_NAD_1983_2011_TM_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",65616.7979],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-119.0],PARAMETER[\"Scale_Factor\",1.00022],PARAMETER[\"Latitude_Of_Origin\",43.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8314]]",
        8315=>"PROJCS[\"OCRS_Coast_Range_North_NAD_1983_2011_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",30000.0],PARAMETER[\"False_Northing\",20000.0],PARAMETER[\"Central_Meridian\",-123.4166666666667],PARAMETER[\"Standard_Parallel_1\",45.58333333333334],PARAMETER[\"Scale_Factor\",1.000045],PARAMETER[\"Latitude_Of_Origin\",45.58333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8315]]",
        8316=>"PROJCS[\"OCRS_Coast_Range_North_NAD_1983_2011_LCC_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",98425.1969],PARAMETER[\"False_Northing\",65616.7979],PARAMETER[\"Central_Meridian\",-123.4166666666667],PARAMETER[\"Standard_Parallel_1\",45.58333333333334],PARAMETER[\"Scale_Factor\",1.000045],PARAMETER[\"Latitude_Of_Origin\",45.58333333333334],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8316]]",
        8317=>"PROJCS[\"OCRS_Dayville-Prairie_City_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",20000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-119.6333333333333],PARAMETER[\"Scale_Factor\",1.00012],PARAMETER[\"Latitude_Of_Origin\",44.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8317]]",
        8318=>"PROJCS[\"OCRS_Dayville-Prairie_City_NAD_1983_2011_TM_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",65616.7979],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-119.6333333333333],PARAMETER[\"Scale_Factor\",1.00012],PARAMETER[\"Latitude_Of_Origin\",44.25],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8318]]",
        8319=>"PROJCS[\"OCRS_Denio-Burns_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",80000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-118.4166666666667],PARAMETER[\"Scale_Factor\",1.00019],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8319]]",
        8320=>"PROJCS[\"OCRS_Denio-Burns_NAD_1983_2011_TM_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",262467.1916],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-118.4166666666667],PARAMETER[\"Scale_Factor\",1.00019],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8320]]",
        8321=>"PROJCS[\"OCRS_Halfway_NAD_1983_2011_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",40000.0],PARAMETER[\"False_Northing\",70000.0],PARAMETER[\"Central_Meridian\",-117.25],PARAMETER[\"Standard_Parallel_1\",45.25],PARAMETER[\"Scale_Factor\",1.000085],PARAMETER[\"Latitude_Of_Origin\",45.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8321]]",
        8322=>"PROJCS[\"OCRS_Halfway_NAD_1983_2011_LCC_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",131233.5958],PARAMETER[\"False_Northing\",229658.7927],PARAMETER[\"Central_Meridian\",-117.25],PARAMETER[\"Standard_Parallel_1\",45.25],PARAMETER[\"Scale_Factor\",1.000085],PARAMETER[\"Latitude_Of_Origin\",45.25],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8322]]",
        8323=>"PROJCS[\"OCRS_Medford-Diamond_Lake_NAD_1983_2011_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",60000.0],PARAMETER[\"False_Northing\",-60000.0],PARAMETER[\"Central_Meridian\",-122.25],PARAMETER[\"Standard_Parallel_1\",42.0],PARAMETER[\"Scale_Factor\",1.00004],PARAMETER[\"Latitude_Of_Origin\",42.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8323]]",
        8324=>"PROJCS[\"OCRS_Medford-Diamond_Lake_NAD_1983_2011_LCC_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",196850.3937],PARAMETER[\"False_Northing\",-196850.3937],PARAMETER[\"Central_Meridian\",-122.25],PARAMETER[\"Standard_Parallel_1\",42.0],PARAMETER[\"Scale_Factor\",1.00004],PARAMETER[\"Latitude_Of_Origin\",42.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8324]]",
        8325=>"PROJCS[\"OCRS_Mitchell_NAD_1983_2011_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",30000.0],PARAMETER[\"False_Northing\",290000.0],PARAMETER[\"Central_Meridian\",-120.25],PARAMETER[\"Standard_Parallel_1\",47.0],PARAMETER[\"Scale_Factor\",0.99927],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8325]]",
        8326=>"PROJCS[\"OCRS_Mitchell_NAD_1983_2011_LCC_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",98425.1969],PARAMETER[\"False_Northing\",951443.5696],PARAMETER[\"Central_Meridian\",-120.25],PARAMETER[\"Standard_Parallel_1\",47.0],PARAMETER[\"Scale_Factor\",0.99927],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8326]]",
        8327=>"PROJCS[\"OCRS_North_Central_NAD_1983_2011_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",140000.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",46.16666666666666],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",46.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8327]]",
        8328=>"PROJCS[\"OCRS_North_Central_NAD_1983_2011_LCC_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",328083.9895],PARAMETER[\"False_Northing\",459317.5853],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",46.16666666666666],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",46.16666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8328]]",
        8329=>"PROJCS[\"OCRS_Ochoco_Summit_NAD_1983_2011_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",40000.0],PARAMETER[\"False_Northing\",-80000.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",43.5],PARAMETER[\"Scale_Factor\",1.00006],PARAMETER[\"Latitude_Of_Origin\",43.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8329]]",
        8330=>"PROJCS[\"OCRS_Ochoco_Summit_NAD_1983_2011_LCC_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",131233.5958],PARAMETER[\"False_Northing\",-262467.1916],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",43.5],PARAMETER[\"Scale_Factor\",1.00006],PARAMETER[\"Latitude_Of_Origin\",43.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8330]]",
        8331=>"PROJCS[\"OCRS_Owyhee_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",70000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.5833333333333],PARAMETER[\"Scale_Factor\",1.00018],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8331]]",
        8332=>"PROJCS[\"OCRS_Owyhee_NAD_1983_2011_TM_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",229658.7927],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.5833333333333],PARAMETER[\"Scale_Factor\",1.00018],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8332]]",
        8333=>"PROJCS[\"OCRS_Pilot_Rock-Ukiah_NAD_1983_2011_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",50000.0],PARAMETER[\"False_Northing\",130000.0],PARAMETER[\"Central_Meridian\",-119.0],PARAMETER[\"Standard_Parallel_1\",46.16666666666666],PARAMETER[\"Scale_Factor\",1.000025],PARAMETER[\"Latitude_Of_Origin\",46.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8333]]",
        8334=>"PROJCS[\"OCRS_Pilot_Rock-Ukiah_NAD_1983_2011_LCC_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",164041.9948],PARAMETER[\"False_Northing\",426509.1864],PARAMETER[\"Central_Meridian\",-119.0],PARAMETER[\"Standard_Parallel_1\",46.16666666666666],PARAMETER[\"Scale_Factor\",1.000025],PARAMETER[\"Latitude_Of_Origin\",46.16666666666666],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8334]]",
        8335=>"PROJCS[\"OCRS_Prairie_City-Brogan_NAD_1983_2011_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",60000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-118.0],PARAMETER[\"Standard_Parallel_1\",44.0],PARAMETER[\"Scale_Factor\",1.00017],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8335]]",
        8336=>"PROJCS[\"OCRS_Prairie_City-Brogan_NAD_1983_2011_LCC_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",196850.3937],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-118.0],PARAMETER[\"Standard_Parallel_1\",44.0],PARAMETER[\"Scale_Factor\",1.00017],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8336]]",
        8337=>"PROJCS[\"OCRS_Riley-Lakeview_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",70000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.3333333333333],PARAMETER[\"Scale_Factor\",1.000215],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8337]]",
        8338=>"PROJCS[\"OCRS_Riley-Lakeview_NAD_1983_2011_TM_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",229658.7927],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.3333333333333],PARAMETER[\"Scale_Factor\",1.000215],PARAMETER[\"Latitude_Of_Origin\",41.75],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8338]]",
        8339=>"PROJCS[\"OCRS_Siskiyou_Pass_NAD_1983_2011_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",10000.0],PARAMETER[\"False_Northing\",60000.0],PARAMETER[\"Central_Meridian\",-122.5833333333333],PARAMETER[\"Standard_Parallel_1\",42.5],PARAMETER[\"Scale_Factor\",1.00015],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8339]]",
        8340=>"PROJCS[\"OCRS_Siskiyou_Pass_NAD_1983_2011_LCC_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",32808.399],PARAMETER[\"False_Northing\",196850.3937],PARAMETER[\"Central_Meridian\",-122.5833333333333],PARAMETER[\"Standard_Parallel_1\",42.5],PARAMETER[\"Scale_Factor\",1.00015],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8340]]",
        8341=>"PROJCS[\"OCRS_Ukiah-Fox_NAD_1983_2011_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",30000.0],PARAMETER[\"False_Northing\",90000.0],PARAMETER[\"Central_Meridian\",-119.0],PARAMETER[\"Standard_Parallel_1\",45.25],PARAMETER[\"Scale_Factor\",1.00014],PARAMETER[\"Latitude_Of_Origin\",45.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8341]]",
        8342=>"PROJCS[\"OCRS_Ukiah-Fox_NAD_1983_2011_LCC_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",98425.1969],PARAMETER[\"False_Northing\",295275.5906],PARAMETER[\"Central_Meridian\",-119.0],PARAMETER[\"Standard_Parallel_1\",45.25],PARAMETER[\"Scale_Factor\",1.00014],PARAMETER[\"Latitude_Of_Origin\",45.25],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8342]]",
        8343=>"PROJCS[\"OCRS_Wallowa_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",60000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.5],PARAMETER[\"Scale_Factor\",1.000195],PARAMETER[\"Latitude_Of_Origin\",45.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8343]]",
        8344=>"PROJCS[\"OCRS_Wallowa_NAD_1983_2011_TM_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",196850.3937],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.5],PARAMETER[\"Scale_Factor\",1.000195],PARAMETER[\"Latitude_Of_Origin\",45.25],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8344]]",
        8345=>"PROJCS[\"OCRS_Warner_Highway_NAD_1983_2011_LCC_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",40000.0],PARAMETER[\"False_Northing\",60000.0],PARAMETER[\"Central_Meridian\",-120.0],PARAMETER[\"Standard_Parallel_1\",42.5],PARAMETER[\"Scale_Factor\",1.000245],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8345]]",
        8346=>"PROJCS[\"OCRS_Warner_Highway_NAD_1983_2011_LCC_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",131233.5958],PARAMETER[\"False_Northing\",196850.3937],PARAMETER[\"Central_Meridian\",-120.0],PARAMETER[\"Standard_Parallel_1\",42.5],PARAMETER[\"Scale_Factor\",1.000245],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8346]]",
        8347=>"PROJCS[\"OCRS_Willamette_Pass_NAD_1983_2011_TM_Meters\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",20000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Scale_Factor\",1.000223],PARAMETER[\"Latitude_Of_Origin\",43.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",8347]]",
        8348=>"PROJCS[\"OCRS_Willamette_Pass_NAD_1983_2011_TM_Ft_Intl\",GEOGCS[\"GCS_NAD_1983_2011\",DATUM[\"D_NAD_1983_2011\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",65616.7979],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Scale_Factor\",1.000223],PARAMETER[\"Latitude_Of_Origin\",43.0],UNIT[\"Foot\",0.3048],AUTHORITY[\"EPSG\",8348]]",
        20004=>"PROJCS[\"Pulkovo_1995_GK_Zone_4\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20004]]",
        20005=>"PROJCS[\"Pulkovo_1995_GK_Zone_5\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20005]]",
        20006=>"PROJCS[\"Pulkovo_1995_GK_Zone_6\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",6500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20006]]",
        20007=>"PROJCS[\"Pulkovo_1995_GK_Zone_7\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",7500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20007]]",
        20008=>"PROJCS[\"Pulkovo_1995_GK_Zone_8\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",8500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20008]]",
        20009=>"PROJCS[\"Pulkovo_1995_GK_Zone_9\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",9500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20009]]",
        20010=>"PROJCS[\"Pulkovo_1995_GK_Zone_10\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",10500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20010]]",
        20011=>"PROJCS[\"Pulkovo_1995_GK_Zone_11\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",11500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20011]]",
        20012=>"PROJCS[\"Pulkovo_1995_GK_Zone_12\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",12500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20012]]",
        20013=>"PROJCS[\"Pulkovo_1995_GK_Zone_13\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",13500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20013]]",
        20014=>"PROJCS[\"Pulkovo_1995_GK_Zone_14\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",14500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20014]]",
        20015=>"PROJCS[\"Pulkovo_1995_GK_Zone_15\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",15500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20015]]",
        20016=>"PROJCS[\"Pulkovo_1995_GK_Zone_16\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",16500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20016]]",
        20017=>"PROJCS[\"Pulkovo_1995_GK_Zone_17\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",17500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20017]]",
        20018=>"PROJCS[\"Pulkovo_1995_GK_Zone_18\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",18500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20018]]",
        20019=>"PROJCS[\"Pulkovo_1995_GK_Zone_19\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",19500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20019]]",
        20020=>"PROJCS[\"Pulkovo_1995_GK_Zone_20\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",20500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20020]]",
        20021=>"PROJCS[\"Pulkovo_1995_GK_Zone_21\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",21500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20021]]",
        20022=>"PROJCS[\"Pulkovo_1995_GK_Zone_22\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",22500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20022]]",
        20023=>"PROJCS[\"Pulkovo_1995_GK_Zone_23\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",23500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20023]]",
        20024=>"PROJCS[\"Pulkovo_1995_GK_Zone_24\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",24500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20024]]",
        20025=>"PROJCS[\"Pulkovo_1995_GK_Zone_25\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",25500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20025]]",
        20026=>"PROJCS[\"Pulkovo_1995_GK_Zone_26\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",26500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20026]]",
        20027=>"PROJCS[\"Pulkovo_1995_GK_Zone_27\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",27500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20027]]",
        20028=>"PROJCS[\"Pulkovo_1995_GK_Zone_28\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",28500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20028]]",
        20029=>"PROJCS[\"Pulkovo_1995_GK_Zone_29\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",29500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20029]]",
        20030=>"PROJCS[\"Pulkovo_1995_GK_Zone_30\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",30500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20030]]",
        20031=>"PROJCS[\"Pulkovo_1995_GK_Zone_31\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",31500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20031]]",
        20032=>"PROJCS[\"Pulkovo_1995_GK_Zone_32\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",32500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20032]]",
        20064=>"PROJCS[\"Pulkovo_1995_GK_Zone_4N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20064]]",
        20065=>"PROJCS[\"Pulkovo_1995_GK_Zone_5N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20065]]",
        20066=>"PROJCS[\"Pulkovo_1995_GK_Zone_6N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20066]]",
        20067=>"PROJCS[\"Pulkovo_1995_GK_Zone_7N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20067]]",
        20068=>"PROJCS[\"Pulkovo_1995_GK_Zone_8N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20068]]",
        20069=>"PROJCS[\"Pulkovo_1995_GK_Zone_9N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20069]]",
        20070=>"PROJCS[\"Pulkovo_1995_GK_Zone_10N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20070]]",
        20071=>"PROJCS[\"Pulkovo_1995_GK_Zone_11N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20071]]",
        20072=>"PROJCS[\"Pulkovo_1995_GK_Zone_12N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20072]]",
        20073=>"PROJCS[\"Pulkovo_1995_GK_Zone_13N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20073]]",
        20074=>"PROJCS[\"Pulkovo_1995_GK_Zone_14N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20074]]",
        20075=>"PROJCS[\"Pulkovo_1995_GK_Zone_15N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20075]]",
        20076=>"PROJCS[\"Pulkovo_1995_GK_Zone_16N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20076]]",
        20077=>"PROJCS[\"Pulkovo_1995_GK_Zone_17N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20077]]",
        20078=>"PROJCS[\"Pulkovo_1995_GK_Zone_18N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20078]]",
        20079=>"PROJCS[\"Pulkovo_1995_GK_Zone_19N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20079]]",
        20080=>"PROJCS[\"Pulkovo_1995_GK_Zone_20N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20080]]",
        20081=>"PROJCS[\"Pulkovo_1995_GK_Zone_21N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20081]]",
        20082=>"PROJCS[\"Pulkovo_1995_GK_Zone_22N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20082]]",
        20083=>"PROJCS[\"Pulkovo_1995_GK_Zone_23N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20083]]",
        20084=>"PROJCS[\"Pulkovo_1995_GK_Zone_24N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20084]]",
        20085=>"PROJCS[\"Pulkovo_1995_GK_Zone_25N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20085]]",
        20086=>"PROJCS[\"Pulkovo_1995_GK_Zone_26N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20086]]",
        20087=>"PROJCS[\"Pulkovo_1995_GK_Zone_27N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20087]]",
        20088=>"PROJCS[\"Pulkovo_1995_GK_Zone_28N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20088]]",
        20089=>"PROJCS[\"Pulkovo_1995_GK_Zone_29N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20089]]",
        20090=>"PROJCS[\"Pulkovo_1995_GK_Zone_30N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20090]]",
        20091=>"PROJCS[\"Pulkovo_1995_GK_Zone_31N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20091]]",
        20092=>"PROJCS[\"Pulkovo_1995_GK_Zone_32N\",GEOGCS[\"GCS_Pulkovo_1995\",DATUM[\"D_Pulkovo_1995\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20092]]",
        20135=>"PROJCS[\"Adindan_UTM_Zone_35N\",GEOGCS[\"GCS_Adindan\",DATUM[\"D_Adindan\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20135]]",
        20136=>"PROJCS[\"Adindan_UTM_Zone_36N\",GEOGCS[\"GCS_Adindan\",DATUM[\"D_Adindan\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20136]]",
        20137=>"PROJCS[\"Adindan_UTM_Zone_37N\",GEOGCS[\"GCS_Adindan\",DATUM[\"D_Adindan\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20137]]",
        20138=>"PROJCS[\"Adindan_UTM_Zone_38N\",GEOGCS[\"GCS_Adindan\",DATUM[\"D_Adindan\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20138]]",
        20248=>"PROJCS[\"AGD_1966_AMG_Zone_48\",GEOGCS[\"GCS_Australian_1966\",DATUM[\"D_Australian_1966\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20248]]",
        20249=>"PROJCS[\"AGD_1966_AMG_Zone_49\",GEOGCS[\"GCS_Australian_1966\",DATUM[\"D_Australian_1966\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20249]]",
        20250=>"PROJCS[\"AGD_1966_AMG_Zone_50\",GEOGCS[\"GCS_Australian_1966\",DATUM[\"D_Australian_1966\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20250]]",
        20251=>"PROJCS[\"AGD_1966_AMG_Zone_51\",GEOGCS[\"GCS_Australian_1966\",DATUM[\"D_Australian_1966\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20251]]",
        20252=>"PROJCS[\"AGD_1966_AMG_Zone_52\",GEOGCS[\"GCS_Australian_1966\",DATUM[\"D_Australian_1966\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20252]]",
        20253=>"PROJCS[\"AGD_1966_AMG_Zone_53\",GEOGCS[\"GCS_Australian_1966\",DATUM[\"D_Australian_1966\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20253]]",
        20254=>"PROJCS[\"AGD_1966_AMG_Zone_54\",GEOGCS[\"GCS_Australian_1966\",DATUM[\"D_Australian_1966\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20254]]",
        20255=>"PROJCS[\"AGD_1966_AMG_Zone_55\",GEOGCS[\"GCS_Australian_1966\",DATUM[\"D_Australian_1966\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20255]]",
        20256=>"PROJCS[\"AGD_1966_AMG_Zone_56\",GEOGCS[\"GCS_Australian_1966\",DATUM[\"D_Australian_1966\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20256]]",
        20257=>"PROJCS[\"AGD_1966_AMG_Zone_57\",GEOGCS[\"GCS_Australian_1966\",DATUM[\"D_Australian_1966\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20257]]",
        20258=>"PROJCS[\"AGD_1966_AMG_Zone_58\",GEOGCS[\"GCS_Australian_1966\",DATUM[\"D_Australian_1966\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20258]]",
        20348=>"PROJCS[\"AGD_1984_AMG_Zone_48\",GEOGCS[\"GCS_Australian_1984\",DATUM[\"D_Australian_1984\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20348]]",
        20349=>"PROJCS[\"AGD_1984_AMG_Zone_49\",GEOGCS[\"GCS_Australian_1984\",DATUM[\"D_Australian_1984\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20349]]",
        20350=>"PROJCS[\"AGD_1984_AMG_Zone_50\",GEOGCS[\"GCS_Australian_1984\",DATUM[\"D_Australian_1984\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20350]]",
        20351=>"PROJCS[\"AGD_1984_AMG_Zone_51\",GEOGCS[\"GCS_Australian_1984\",DATUM[\"D_Australian_1984\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20351]]",
        20352=>"PROJCS[\"AGD_1984_AMG_Zone_52\",GEOGCS[\"GCS_Australian_1984\",DATUM[\"D_Australian_1984\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20352]]",
        20353=>"PROJCS[\"AGD_1984_AMG_Zone_53\",GEOGCS[\"GCS_Australian_1984\",DATUM[\"D_Australian_1984\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20353]]",
        20354=>"PROJCS[\"AGD_1984_AMG_Zone_54\",GEOGCS[\"GCS_Australian_1984\",DATUM[\"D_Australian_1984\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20354]]",
        20355=>"PROJCS[\"AGD_1984_AMG_Zone_55\",GEOGCS[\"GCS_Australian_1984\",DATUM[\"D_Australian_1984\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20355]]",
        20356=>"PROJCS[\"AGD_1984_AMG_Zone_56\",GEOGCS[\"GCS_Australian_1984\",DATUM[\"D_Australian_1984\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20356]]",
        20357=>"PROJCS[\"AGD_1984_AMG_Zone_57\",GEOGCS[\"GCS_Australian_1984\",DATUM[\"D_Australian_1984\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20357]]",
        20358=>"PROJCS[\"AGD_1984_AMG_Zone_58\",GEOGCS[\"GCS_Australian_1984\",DATUM[\"D_Australian_1984\",SPHEROID[\"Australian\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20358]]",
        20436=>"PROJCS[\"Ain_el_Abd_UTM_Zone_36N\",GEOGCS[\"GCS_Ain_el_Abd_1970\",DATUM[\"D_Ain_el_Abd_1970\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20436]]",
        20437=>"PROJCS[\"Ain_el_Abd_UTM_Zone_37N\",GEOGCS[\"GCS_Ain_el_Abd_1970\",DATUM[\"D_Ain_el_Abd_1970\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20437]]",
        20438=>"PROJCS[\"Ain_el_Abd_UTM_Zone_38N\",GEOGCS[\"GCS_Ain_el_Abd_1970\",DATUM[\"D_Ain_el_Abd_1970\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20438]]",
        20439=>"PROJCS[\"Ain_el_Abd_UTM_Zone_39N\",GEOGCS[\"GCS_Ain_el_Abd_1970\",DATUM[\"D_Ain_el_Abd_1970\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20439]]",
        20440=>"PROJCS[\"Ain_el_Abd_UTM_Zone_40N\",GEOGCS[\"GCS_Ain_el_Abd_1970\",DATUM[\"D_Ain_el_Abd_1970\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20440]]",
        20499=>"PROJCS[\"Bahrain_State_Grid\",GEOGCS[\"GCS_Ain_el_Abd_1970\",DATUM[\"D_Ain_el_Abd_1970\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20499]]",
        20538=>"PROJCS[\"Afgooye_UTM_Zone_38N\",GEOGCS[\"GCS_Afgooye\",DATUM[\"D_Afgooye\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20538]]",
        20539=>"PROJCS[\"Afgooye_UTM_Zone_39N\",GEOGCS[\"GCS_Afgooye\",DATUM[\"D_Afgooye\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20539]]",
        20790=>"PROJCS[\"Portuguese_National_Grid\",GEOGCS[\"GCS_Lisbon_Lisbon\",DATUM[\"D_Lisbon\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Lisbon\",-9.131906111111112],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",300000.0],PARAMETER[\"Central_Meridian\",1.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",39.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20790]]",
        20791=>"PROJCS[\"Lisbon_Lisbon_Portuguese_Grid\",GEOGCS[\"GCS_Lisbon_Lisbon\",DATUM[\"D_Lisbon\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Lisbon\",-9.131906111111112],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",1.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",39.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20791]]",
        20822=>"PROJCS[\"Aratu_UTM_Zone_22S\",GEOGCS[\"GCS_Aratu\",DATUM[\"D_Aratu\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20822]]",
        20823=>"PROJCS[\"Aratu_UTM_Zone_23S\",GEOGCS[\"GCS_Aratu\",DATUM[\"D_Aratu\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20823]]",
        20824=>"PROJCS[\"Aratu_UTM_Zone_24S\",GEOGCS[\"GCS_Aratu\",DATUM[\"D_Aratu\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20824]]",
        20934=>"PROJCS[\"Arc_1950_UTM_Zone_34S\",GEOGCS[\"GCS_Arc_1950\",DATUM[\"D_Arc_1950\",SPHEROID[\"Clarke_1880_Arc\",6378249.145,293.466307656]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20934]]",
        20935=>"PROJCS[\"Arc_1950_UTM_Zone_35S\",GEOGCS[\"GCS_Arc_1950\",DATUM[\"D_Arc_1950\",SPHEROID[\"Clarke_1880_Arc\",6378249.145,293.466307656]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20935]]",
        20936=>"PROJCS[\"Arc_1950_UTM_Zone_36S\",GEOGCS[\"GCS_Arc_1950\",DATUM[\"D_Arc_1950\",SPHEROID[\"Clarke_1880_Arc\",6378249.145,293.466307656]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",20936]]",
        21035=>"PROJCS[\"Arc_1960_UTM_Zone_35S\",GEOGCS[\"GCS_Arc_1960\",DATUM[\"D_Arc_1960\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21035]]",
        21036=>"PROJCS[\"Arc_1960_UTM_Zone_36S\",GEOGCS[\"GCS_Arc_1960\",DATUM[\"D_Arc_1960\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21036]]",
        21037=>"PROJCS[\"Arc_1960_UTM_Zone_37S\",GEOGCS[\"GCS_Arc_1960\",DATUM[\"D_Arc_1960\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21037]]",
        21095=>"PROJCS[\"Arc_1960_UTM_Zone_35N\",GEOGCS[\"GCS_Arc_1960\",DATUM[\"D_Arc_1960\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21095]]",
        21096=>"PROJCS[\"Arc_1960_UTM_Zone_36N\",GEOGCS[\"GCS_Arc_1960\",DATUM[\"D_Arc_1960\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21096]]",
        21097=>"PROJCS[\"Arc_1960_UTM_Zone_37N\",GEOGCS[\"GCS_Arc_1960\",DATUM[\"D_Arc_1960\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21097]]",
        21148=>"PROJCS[\"Batavia_UTM_Zone_48S\",GEOGCS[\"GCS_Batavia\",DATUM[\"D_Batavia\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21148]]",
        21149=>"PROJCS[\"Batavia_UTM_Zone_49S\",GEOGCS[\"GCS_Batavia\",DATUM[\"D_Batavia\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21149]]",
        21150=>"PROJCS[\"Batavia_UTM_Zone_50S\",GEOGCS[\"GCS_Batavia\",DATUM[\"D_Batavia\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21150]]",
        21291=>"PROJCS[\"Barbados_1938_British_West_Indies_Grid\",GEOGCS[\"GCS_Barbados_1938\",DATUM[\"D_Barbados_1938\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-62.0],PARAMETER[\"Scale_Factor\",0.9995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21291]]",
        21292=>"PROJCS[\"Barbados_1938_Barbados_Grid\",GEOGCS[\"GCS_Barbados_1938\",DATUM[\"D_Barbados_1938\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",30000.0],PARAMETER[\"False_Northing\",75000.0],PARAMETER[\"Central_Meridian\",-59.55972222222222],PARAMETER[\"Scale_Factor\",0.9999986],PARAMETER[\"Latitude_Of_Origin\",13.17638888888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21292]]",
        21413=>"PROJCS[\"Beijing_1954_GK_Zone_13\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",13500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21413]]",
        21414=>"PROJCS[\"Beijing_1954_GK_Zone_14\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",14500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21414]]",
        21415=>"PROJCS[\"Beijing_1954_GK_Zone_15\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",15500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21415]]",
        21416=>"PROJCS[\"Beijing_1954_GK_Zone_16\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",16500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21416]]",
        21417=>"PROJCS[\"Beijing_1954_GK_Zone_17\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",17500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21417]]",
        21418=>"PROJCS[\"Beijing_1954_GK_Zone_18\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",18500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21418]]",
        21419=>"PROJCS[\"Beijing_1954_GK_Zone_19\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",19500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21419]]",
        21420=>"PROJCS[\"Beijing_1954_GK_Zone_20\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",20500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21420]]",
        21421=>"PROJCS[\"Beijing_1954_GK_Zone_21\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",21500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21421]]",
        21422=>"PROJCS[\"Beijing_1954_GK_Zone_22\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",22500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21422]]",
        21423=>"PROJCS[\"Beijing_1954_GK_Zone_23\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",23500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21423]]",
        21473=>"PROJCS[\"Beijing_1954_GK_Zone_13N\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21473]]",
        21474=>"PROJCS[\"Beijing_1954_GK_Zone_14N\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21474]]",
        21475=>"PROJCS[\"Beijing_1954_GK_Zone_15N\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21475]]",
        21476=>"PROJCS[\"Beijing_1954_GK_Zone_16N\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21476]]",
        21477=>"PROJCS[\"Beijing_1954_GK_Zone_17N\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21477]]",
        21478=>"PROJCS[\"Beijing_1954_GK_Zone_18N\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21478]]",
        21479=>"PROJCS[\"Beijing_1954_GK_Zone_19N\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21479]]",
        21480=>"PROJCS[\"Beijing_1954_GK_Zone_20N\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21480]]",
        21481=>"PROJCS[\"Beijing_1954_GK_Zone_21N\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21481]]",
        21482=>"PROJCS[\"Beijing_1954_GK_Zone_22N\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21482]]",
        21483=>"PROJCS[\"Beijing_1954_GK_Zone_23N\",GEOGCS[\"GCS_Beijing_1954\",DATUM[\"D_Beijing_1954\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21483]]",
        21500=>"PROJCS[\"Belge_Lambert_1950\",GEOGCS[\"GCS_Belge_1950_Brussels\",DATUM[\"D_Belge_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Brussels\",4.367975],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",5400000.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",49.83333333333334],PARAMETER[\"Standard_Parallel_2\",51.16666666666666],PARAMETER[\"Latitude_Of_Origin\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21500]]",
        21780=>"PROJCS[\"Bern_1898_Bern_LV03C\",GEOGCS[\"GCS_Bern_1898_Bern\",DATUM[\"D_Bern_1898\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Bern\",7.439583333333333],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Center\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Azimuth\",90.0],PARAMETER[\"Longitude_Of_Center\",0.0],PARAMETER[\"Latitude_Of_Center\",46.95240555555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21780]]",
        21781=>"PROJCS[\"CH1903_LV03\",GEOGCS[\"GCS_CH1903\",DATUM[\"D_CH1903\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Center\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Azimuth\",90.0],PARAMETER[\"Longitude_Of_Center\",7.439583333333333],PARAMETER[\"Latitude_Of_Center\",46.95240555555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21781]]",
        21782=>"PROJCS[\"CH1903_LV03C-G\",GEOGCS[\"GCS_CH1903\",DATUM[\"D_CH1903\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Center\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Azimuth\",90.0],PARAMETER[\"Longitude_Of_Center\",7.439583333333333],PARAMETER[\"Latitude_Of_Center\",46.95240555555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21782]]",
        21817=>"PROJCS[\"Bogota_UTM_Zone_17N\",GEOGCS[\"GCS_Bogota\",DATUM[\"D_Bogota\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21817]]",
        21818=>"PROJCS[\"Bogota_UTM_Zone_18N\",GEOGCS[\"GCS_Bogota\",DATUM[\"D_Bogota\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21818]]",
        21891=>"PROJCS[\"Colombia_West_Zone\",GEOGCS[\"GCS_Bogota\",DATUM[\"D_Bogota\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-77.08091666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",4.599047222222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21896]]",
        21892=>"PROJCS[\"Colombia_Bogota_Zone\",GEOGCS[\"GCS_Bogota\",DATUM[\"D_Bogota\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-74.08091666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",4.599047222222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21897]]",
        21893=>"PROJCS[\"Colombia_East_Central_Zone\",GEOGCS[\"GCS_Bogota\",DATUM[\"D_Bogota\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-71.08091666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",4.599047222222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21898]]",
        21894=>"PROJCS[\"Colombia_East_Zone\",GEOGCS[\"GCS_Bogota\",DATUM[\"D_Bogota\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-68.08091666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",4.599047222222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21899]]",
        21896=>"PROJCS[\"Colombia_West_Zone\",GEOGCS[\"GCS_Bogota\",DATUM[\"D_Bogota\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-77.08091666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",4.599047222222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21896]]",
        21897=>"PROJCS[\"Colombia_Bogota_Zone\",GEOGCS[\"GCS_Bogota\",DATUM[\"D_Bogota\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-74.08091666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",4.599047222222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21897]]",
        21898=>"PROJCS[\"Colombia_East_Central_Zone\",GEOGCS[\"GCS_Bogota\",DATUM[\"D_Bogota\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-71.08091666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",4.599047222222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21898]]",
        21899=>"PROJCS[\"Colombia_East_Zone\",GEOGCS[\"GCS_Bogota\",DATUM[\"D_Bogota\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-68.08091666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",4.599047222222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",21899]]",
        22032=>"PROJCS[\"Camacupa_UTM_Zone_32S\",GEOGCS[\"GCS_Camacupa\",DATUM[\"D_Camacupa\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22032]]",
        22033=>"PROJCS[\"Camacupa_UTM_Zone_33S\",GEOGCS[\"GCS_Camacupa\",DATUM[\"D_Camacupa\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22033]]",
        22091=>"PROJCS[\"Camacupa_TM_11_30_SE\",GEOGCS[\"GCS_Camacupa\",DATUM[\"D_Camacupa\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",11.5],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22091]]",
        22092=>"PROJCS[\"Camacupa_TM_12_SE\",GEOGCS[\"GCS_Camacupa\",DATUM[\"D_Camacupa\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22092]]",
        22171=>"PROJCS[\"POSGAR_1998_Argentina_Zone_1\",GEOGCS[\"GCS_POSGAR_1998\",DATUM[\"D_POSGAR_1998\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-72.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22171]]",
        22172=>"PROJCS[\"POSGAR_1998_Argentina_Zone_2\",GEOGCS[\"GCS_POSGAR_1998\",DATUM[\"D_POSGAR_1998\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22172]]",
        22173=>"PROJCS[\"POSGAR_1998_Argentina_Zone_3\",GEOGCS[\"GCS_POSGAR_1998\",DATUM[\"D_POSGAR_1998\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-66.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22173]]",
        22174=>"PROJCS[\"POSGAR_1998_Argentina_Zone_4\",GEOGCS[\"GCS_POSGAR_1998\",DATUM[\"D_POSGAR_1998\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22174]]",
        22175=>"PROJCS[\"POSGAR_1998_Argentina_Zone_5\",GEOGCS[\"GCS_POSGAR_1998\",DATUM[\"D_POSGAR_1998\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-60.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22175]]",
        22176=>"PROJCS[\"POSGAR_1998_Argentina_Zone_6\",GEOGCS[\"GCS_POSGAR_1998\",DATUM[\"D_POSGAR_1998\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",6500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22176]]",
        22177=>"PROJCS[\"POSGAR_1998_Argentina_Zone_7\",GEOGCS[\"GCS_POSGAR_1998\",DATUM[\"D_POSGAR_1998\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",7500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-54.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22177]]",
        22181=>"PROJCS[\"POSGAR_1994_Argentina_Zone_1\",GEOGCS[\"GCS_POSGAR_1994\",DATUM[\"D_POSGAR_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-72.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22181]]",
        22182=>"PROJCS[\"POSGAR_1994_Argentina_Zone_2\",GEOGCS[\"GCS_POSGAR_1994\",DATUM[\"D_POSGAR_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22182]]",
        22183=>"PROJCS[\"POSGAR_1994_Argentina_Zone_3\",GEOGCS[\"GCS_POSGAR_1994\",DATUM[\"D_POSGAR_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-66.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22183]]",
        22184=>"PROJCS[\"POSGAR_1994_Argentina_Zone_4\",GEOGCS[\"GCS_POSGAR_1994\",DATUM[\"D_POSGAR_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22184]]",
        22185=>"PROJCS[\"POSGAR_1994_Argentina_Zone_5\",GEOGCS[\"GCS_POSGAR_1994\",DATUM[\"D_POSGAR_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-60.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22185]]",
        22186=>"PROJCS[\"POSGAR_1994_Argentina_Zone_6\",GEOGCS[\"GCS_POSGAR_1994\",DATUM[\"D_POSGAR_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",6500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22186]]",
        22187=>"PROJCS[\"POSGAR_1994_Argentina_Zone_7\",GEOGCS[\"GCS_POSGAR_1994\",DATUM[\"D_POSGAR_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",7500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-54.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22187]]",
        22191=>"PROJCS[\"Argentina_Zone_1\",GEOGCS[\"GCS_Campo_Inchauspe\",DATUM[\"D_Campo_Inchauspe\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-72.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22191]]",
        22192=>"PROJCS[\"Argentina_Zone_2\",GEOGCS[\"GCS_Campo_Inchauspe\",DATUM[\"D_Campo_Inchauspe\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22192]]",
        22193=>"PROJCS[\"Argentina_Zone_3\",GEOGCS[\"GCS_Campo_Inchauspe\",DATUM[\"D_Campo_Inchauspe\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-66.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22193]]",
        22194=>"PROJCS[\"Argentina_Zone_4\",GEOGCS[\"GCS_Campo_Inchauspe\",DATUM[\"D_Campo_Inchauspe\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22194]]",
        22195=>"PROJCS[\"Argentina_Zone_5\",GEOGCS[\"GCS_Campo_Inchauspe\",DATUM[\"D_Campo_Inchauspe\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-60.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22195]]",
        22196=>"PROJCS[\"Argentina_Zone_6\",GEOGCS[\"GCS_Campo_Inchauspe\",DATUM[\"D_Campo_Inchauspe\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",6500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22196]]",
        22197=>"PROJCS[\"Argentina_Zone_7\",GEOGCS[\"GCS_Campo_Inchauspe\",DATUM[\"D_Campo_Inchauspe\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",7500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-54.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22197]]",
        22234=>"PROJCS[\"Cape_UTM_Zone_34S\",GEOGCS[\"GCS_Cape\",DATUM[\"D_Cape\",SPHEROID[\"Clarke_1880_Arc\",6378249.145,293.466307656]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22234]]",
        22235=>"PROJCS[\"Cape_UTM_Zone_35S\",GEOGCS[\"GCS_Cape\",DATUM[\"D_Cape\",SPHEROID[\"Clarke_1880_Arc\",6378249.145,293.466307656]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22235]]",
        22236=>"PROJCS[\"Cape_UTM_Zone_36S\",GEOGCS[\"GCS_Cape\",DATUM[\"D_Cape\",SPHEROID[\"Clarke_1880_Arc\",6378249.145,293.466307656]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22236]]",
        22332=>"PROJCS[\"Carthage_UTM_Zone_32N\",GEOGCS[\"GCS_Carthage\",DATUM[\"D_Carthage\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22332]]",
        22391=>"PROJCS[\"Nord_Tunisie\",GEOGCS[\"GCS_Carthage\",DATUM[\"D_Carthage\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",300000.0],PARAMETER[\"Central_Meridian\",9.9],PARAMETER[\"Standard_Parallel_1\",36.0],PARAMETER[\"Scale_Factor\",0.999625544],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22391]]",
        22392=>"PROJCS[\"Sud_Tunisie\",GEOGCS[\"GCS_Carthage\",DATUM[\"D_Carthage\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",300000.0],PARAMETER[\"Central_Meridian\",9.9],PARAMETER[\"Standard_Parallel_1\",33.3],PARAMETER[\"Scale_Factor\",0.999625769],PARAMETER[\"Latitude_Of_Origin\",33.3],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22392]]",
        22521=>"PROJCS[\"Corrego_Alegre_UTM_Zone_21S\",GEOGCS[\"GCS_Corrego_Alegre\",DATUM[\"D_Corrego_Alegre\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22521]]",
        22522=>"PROJCS[\"Corrego_Alegre_UTM_Zone_22S\",GEOGCS[\"GCS_Corrego_Alegre\",DATUM[\"D_Corrego_Alegre\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22522]]",
        22523=>"PROJCS[\"Corrego_Alegre_UTM_Zone_23S\",GEOGCS[\"GCS_Corrego_Alegre\",DATUM[\"D_Corrego_Alegre\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22523]]",
        22524=>"PROJCS[\"Corrego_Alegre_UTM_Zone_24S\",GEOGCS[\"GCS_Corrego_Alegre\",DATUM[\"D_Corrego_Alegre\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22524]]",
        22525=>"PROJCS[\"Corrego_Alegre_UTM_Zone_25S\",GEOGCS[\"GCS_Corrego_Alegre\",DATUM[\"D_Corrego_Alegre\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22525]]",
        22700=>"PROJCS[\"Deir_ez_Zor_Levant_Zone\",GEOGCS[\"GCS_Deir_ez_Zor\",DATUM[\"D_Deir_ez_Zor\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.15],PARAMETER[\"Scale_Factor\",0.9995341],PARAMETER[\"Latitude_Of_Origin\",34.2],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22700]]",
        22770=>"PROJCS[\"Deir_ez_Zor_Syria_Lambert\",GEOGCS[\"GCS_Deir_ez_Zor\",DATUM[\"D_Deir_ez_Zor\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",300000.0],PARAMETER[\"Central_Meridian\",37.35],PARAMETER[\"Standard_Parallel_1\",34.65],PARAMETER[\"Scale_Factor\",0.9996256],PARAMETER[\"Latitude_Of_Origin\",34.65],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22770]]",
        22780=>"PROJCS[\"Deir_ez_Zor_Levant_Stereographic\",GEOGCS[\"GCS_Deir_ez_Zor\",DATUM[\"D_Deir_ez_Zor\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.15],PARAMETER[\"Scale_Factor\",0.9995341],PARAMETER[\"Latitude_Of_Origin\",34.2],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22780]]",
        22832=>"PROJCS[\"Douala_UTM_Zone_32N\",GEOGCS[\"GCS_Douala\",DATUM[\"D_Douala\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22832]]",
        22991=>"PROJCS[\"Egypt_Blue_Belt\",GEOGCS[\"GCS_Egypt_1907\",DATUM[\"D_Egypt_1907\",SPHEROID[\"Helmert_1906\",6378200.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",1100000.0],PARAMETER[\"Central_Meridian\",35.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22991]]",
        22992=>"PROJCS[\"Egypt_Red_Belt\",GEOGCS[\"GCS_Egypt_1907\",DATUM[\"D_Egypt_1907\",SPHEROID[\"Helmert_1906\",6378200.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",615000.0],PARAMETER[\"False_Northing\",810000.0],PARAMETER[\"Central_Meridian\",31.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22992]]",
        22993=>"PROJCS[\"Egypt_Purple_Belt\",GEOGCS[\"GCS_Egypt_1907\",DATUM[\"D_Egypt_1907\",SPHEROID[\"Helmert_1906\",6378200.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22993]]",
        22994=>"PROJCS[\"Egypt_Extended_Purple_Belt\",GEOGCS[\"GCS_Egypt_1907\",DATUM[\"D_Egypt_1907\",SPHEROID[\"Helmert_1906\",6378200.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",1200000.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",22994]]",
        23028=>"PROJCS[\"ED_1950_UTM_Zone_28N\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23028]]",
        23029=>"PROJCS[\"ED_1950_UTM_Zone_29N\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23029]]",
        23030=>"PROJCS[\"ED_1950_UTM_Zone_30N\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23030]]",
        23031=>"PROJCS[\"ED_1950_UTM_Zone_31N\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23031]]",
        23032=>"PROJCS[\"ED_1950_UTM_Zone_32N\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23032]]",
        23033=>"PROJCS[\"ED_1950_UTM_Zone_33N\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23033]]",
        23034=>"PROJCS[\"ED_1950_UTM_Zone_34N\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23034]]",
        23035=>"PROJCS[\"ED_1950_UTM_Zone_35N\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23035]]",
        23036=>"PROJCS[\"ED_1950_UTM_Zone_36N\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23036]]",
        23037=>"PROJCS[\"ED_1950_UTM_Zone_37N\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23037]]",
        23038=>"PROJCS[\"ED_1950_UTM_Zone_38N\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23038]]",
        23090=>"PROJCS[\"ED_1950_TM_0_N\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23090]]",
        23095=>"PROJCS[\"ED_1950_TM_5_NE\",GEOGCS[\"GCS_European_1950\",DATUM[\"D_European_1950\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",5.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23095]]",
        23239=>"PROJCS[\"Fahud_UTM_Zone_39N\",GEOGCS[\"GCS_Fahud\",DATUM[\"D_Fahud\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23239]]",
        23240=>"PROJCS[\"Fahud_UTM_Zone_40N\",GEOGCS[\"GCS_Fahud\",DATUM[\"D_Fahud\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23240]]",
        23433=>"PROJCS[\"Garoua_UTM_Zone_33N\",GEOGCS[\"GCS_Garoua\",DATUM[\"D_Garoua\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2312]]",
        23700=>"PROJCS[\"Hungarian_1972_Egyseges_Orszagos_Vetuleti\",GEOGCS[\"GCS_Hungarian_1972\",DATUM[\"D_Hungarian_1972\",SPHEROID[\"GRS_1967\",6378160.0,298.247167427]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Center\"],PARAMETER[\"False_Easting\",650000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Scale_Factor\",0.99993],PARAMETER[\"Azimuth\",90.0],PARAMETER[\"Longitude_Of_Center\",19.04857177777778],PARAMETER[\"Latitude_Of_Center\",47.14439372222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23700]]",
        23830=>"PROJCS[\"DGN_1995_Indonesia_TM-3_Zone_46.2\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",94.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23830]]",
        23831=>"PROJCS[\"DGN_1995_Indonesia_TM-3_Zone_47.1\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",97.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23831]]",
        23832=>"PROJCS[\"DGN_1995_Indonesia_TM-3_Zone_47.2\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",100.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23832]]",
        23833=>"PROJCS[\"DGN_1995_Indonesia_TM-3_Zone_48.1\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",103.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23833]]",
        23834=>"PROJCS[\"DGN_1995_Indonesia_TM-3_Zone_48.2\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",106.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23834]]",
        23835=>"PROJCS[\"DGN_1995_Indonesia_TM-3_Zone_49.1\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",109.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23835]]",
        23836=>"PROJCS[\"DGN_1995_Indonesia_TM-3_Zone_49.2\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",112.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23836]]",
        23837=>"PROJCS[\"DGN_1995_Indonesia_TM-3_Zone_50.1\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",115.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23837]]",
        23838=>"PROJCS[\"DGN_1995_Indonesia_TM-3_Zone_50.2\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",118.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23838]]",
        23839=>"PROJCS[\"DGN_1995_Indonesia_TM-3_Zone_51.1\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",121.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23839]]",
        23840=>"PROJCS[\"DGN_1995_Indonesia_TM-3_Zone_51.2\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",124.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23840]]",
        23841=>"PROJCS[\"DGN_1995_Indonesia_TM-3_Zone_52.1\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",127.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23841]]",
        23842=>"PROJCS[\"DGN_1995_Indonesia_TM-3_Zone_52.2\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",130.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23842]]",
        23843=>"PROJCS[\"DGN_1995_Indonesia_TM-3_Zone_53.1\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",133.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23843]]",
        23844=>"PROJCS[\"DGN_1995_Indonesia_TM-3_Zone_53.2\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",136.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23844]]",
        23845=>"PROJCS[\"DGN_1995_Indonesia_TM-3_Zone_54.1\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1500000.0],PARAMETER[\"Central_Meridian\",139.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23845]]",
        23846=>"PROJCS[\"Indonesian_1974_UTM_Zone_46N\",GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23846]]",
        23847=>"PROJCS[\"Indonesian_1974_UTM_Zone_47N\",GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23847]]",
        23848=>"PROJCS[\"Indonesian_1974_UTM_Zone_48N\",GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23848]]",
        23849=>"PROJCS[\"Indonesian_1974_UTM_Zone_49N\",GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23849]]",
        23850=>"PROJCS[\"Indonesian_1974_UTM_Zone_50N\",GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23850]]",
        23851=>"PROJCS[\"Indonesian_1974_UTM_Zone_51N\",GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23851]]",
        23852=>"PROJCS[\"Indonesian_1974_UTM_Zone_52N\",GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23852]]",
        23853=>"PROJCS[\"Indonesian_1974_UTM_Zone_53N\",GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23853]]",
        23866=>"PROJCS[\"DGN_1995_UTM_Zone_46N\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23866]]",
        23867=>"PROJCS[\"DGN_1995_UTM_Zone_47N\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23867]]",
        23868=>"PROJCS[\"DGN_1995_UTM_Zone_48N\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23868]]",
        23869=>"PROJCS[\"DGN_1995_UTM_Zone_49N\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23869]]",
        23870=>"PROJCS[\"DGN_1995_UTM_Zone_50N\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23870]]",
        23871=>"PROJCS[\"DGN_1995_UTM_Zone_51N\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23871]]",
        23872=>"PROJCS[\"DGN_1995_UTM_Zone_52N\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23872]]",
        23877=>"PROJCS[\"DGN_1995_UTM_Zone_47S\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23877]]",
        23878=>"PROJCS[\"DGN_1995_UTM_Zone_48S\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23878]]",
        23879=>"PROJCS[\"DGN_1995_UTM_Zone_49S\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23879]]",
        23880=>"PROJCS[\"DGN_1995_UTM_Zone_50S\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23880]]",
        23881=>"PROJCS[\"DGN_1995_UTM_Zone_51S\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23881]]",
        23882=>"PROJCS[\"DGN_1995_UTM_Zone_52S\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23882]]",
        23883=>"PROJCS[\"DGN_1995_UTM_Zone_53S\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23883]]",
        23884=>"PROJCS[\"DGN_1995_UTM_Zone_54S\",GEOGCS[\"GCS_DGN_1995\",DATUM[\"D_Datum_Geodesi_Nasional_1995\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23884]]",
        23886=>"PROJCS[\"Indonesian_1974_UTM_Zone_46S\",GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23886]]",
        23887=>"PROJCS[\"Indonesian_1974_UTM_Zone_47S\",GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23887]]",
        23888=>"PROJCS[\"Indonesian_1974_UTM_Zone_48S\",GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23888]]",
        23889=>"PROJCS[\"Indonesian_1974_UTM_Zone_49S\",GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23889]]",
        23890=>"PROJCS[\"Indonesian_1974_UTM_Zone_50S\",GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23890]]",
        23891=>"PROJCS[\"Indonesian_1974_UTM_Zone_51S\",GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23891]]",
        23892=>"PROJCS[\"Indonesian_1974_UTM_Zone_52S\",GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23892]]",
        23893=>"PROJCS[\"Indonesian_1974_UTM_Zone_53S\",GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23893]]",
        23894=>"PROJCS[\"Indonesian_1974_UTM_Zone_54S\",GEOGCS[\"GCS_Indonesian_1974\",DATUM[\"D_Indonesian_1974\",SPHEROID[\"Indonesian\",6378160.0,298.247]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23894]]",
        23946=>"PROJCS[\"Indian_1954_UTM_Zone_46N\",GEOGCS[\"GCS_Indian_1954\",DATUM[\"D_Indian_1954\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23946]]",
        23947=>"PROJCS[\"Indian_1954_UTM_Zone_47N\",GEOGCS[\"GCS_Indian_1954\",DATUM[\"D_Indian_1954\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23947]]",
        23948=>"PROJCS[\"Indian_1954_UTM_Zone_48N\",GEOGCS[\"GCS_Indian_1954\",DATUM[\"D_Indian_1954\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",23948]]",
        24047=>"PROJCS[\"Indian_1975_UTM_Zone_47N\",GEOGCS[\"GCS_Indian_1975\",DATUM[\"D_Indian_1975\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24047]]",
        24048=>"PROJCS[\"Indian_1975_UTM_Zone_48N\",GEOGCS[\"GCS_Indian_1975\",DATUM[\"D_Indian_1975\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24048]]",
        24100=>"PROJCS[\"Jamaica_1875_Old_Grid\",GEOGCS[\"GCS_Jamaica_1875\",DATUM[\"D_Jamaica_1875\",SPHEROID[\"Clarke_1880\",6378249.144808011,293.4663076556253]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",550000.0],PARAMETER[\"False_Northing\",400000.0],PARAMETER[\"Central_Meridian\",-77.0],PARAMETER[\"Standard_Parallel_1\",18.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",18.0],UNIT[\"Foot_Clarke\",0.3047972654],AUTHORITY[\"EPSG\",24100]]",
        24200=>"PROJCS[\"Jamaica_Grid\",GEOGCS[\"GCS_Jamaica_1969\",DATUM[\"D_Jamaica_1969\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",150000.0],PARAMETER[\"Central_Meridian\",-77.0],PARAMETER[\"Standard_Parallel_1\",18.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",18.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24200]]",
        24305=>"PROJCS[\"Kalianpur_1937_UTM_Zone_45N\",GEOGCS[\"GCS_Kalianpur_1937\",DATUM[\"D_Kalianpur_1937\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24305]]",
        24306=>"PROJCS[\"Kalianpur_1937_UTM_Zone_46N\",GEOGCS[\"GCS_Kalianpur_1937\",DATUM[\"D_Kalianpur_1937\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24306]]",
        24311=>"PROJCS[\"Kalianpur_1962_UTM_Zone_41N\",GEOGCS[\"GCS_Kalianpur_1962\",DATUM[\"D_Kalianpur_1962\",SPHEROID[\"Everest_Definition_1962\",6377301.243,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24311]]",
        24312=>"PROJCS[\"Kalianpur_1962_UTM_Zone_42N\",GEOGCS[\"GCS_Kalianpur_1962\",DATUM[\"D_Kalianpur_1962\",SPHEROID[\"Everest_Definition_1962\",6377301.243,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24312]]",
        24313=>"PROJCS[\"Kalianpur_1962_UTM_Zone_43N\",GEOGCS[\"GCS_Kalianpur_1962\",DATUM[\"D_Kalianpur_1962\",SPHEROID[\"Everest_Definition_1962\",6377301.243,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24313]]",
        24342=>"PROJCS[\"Kalianpur_1975_UTM_Zone_42N\",GEOGCS[\"GCS_Kalianpur_1975\",DATUM[\"D_Kalianpur_1975\",SPHEROID[\"Everest_Definition_1975\",6377299.151,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24342]]",
        24343=>"PROJCS[\"Kalianpur_1975_UTM_Zone_43N\",GEOGCS[\"GCS_Kalianpur_1975\",DATUM[\"D_Kalianpur_1975\",SPHEROID[\"Everest_Definition_1975\",6377299.151,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24343]]",
        24344=>"PROJCS[\"Kalianpur_1975_UTM_Zone_44N\",GEOGCS[\"GCS_Kalianpur_1975\",DATUM[\"D_Kalianpur_1975\",SPHEROID[\"Everest_Definition_1975\",6377299.151,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24344]]",
        24345=>"PROJCS[\"Kalianpur_1975_UTM_Zone_45N\",GEOGCS[\"GCS_Kalianpur_1975\",DATUM[\"D_Kalianpur_1975\",SPHEROID[\"Everest_Definition_1975\",6377299.151,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24345]]",
        24346=>"PROJCS[\"Kalianpur_1975_UTM_Zone_46N\",GEOGCS[\"GCS_Kalianpur_1975\",DATUM[\"D_Kalianpur_1975\",SPHEROID[\"Everest_Definition_1975\",6377299.151,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24346]]",
        24347=>"PROJCS[\"Kalianpur_1975_UTM_Zone_47N\",GEOGCS[\"GCS_Kalianpur_1975\",DATUM[\"D_Kalianpur_1975\",SPHEROID[\"Everest_Definition_1975\",6377299.151,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24347]]",
        24370=>"PROJCS[\"Kalianpur_1880_India_Zone_0\",GEOGCS[\"GCS_Kalianpur_1880\",DATUM[\"D_Kalianpur_1880\",SPHEROID[\"Everest_1830\",6377299.36,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2355500.0],PARAMETER[\"False_Northing\",2590000.0],PARAMETER[\"Central_Meridian\",68.0],PARAMETER[\"Standard_Parallel_1\",39.5],PARAMETER[\"Scale_Factor\",0.99846154],PARAMETER[\"Latitude_Of_Origin\",39.5],UNIT[\"Yard_Indian\",0.9143985307444408],AUTHORITY[\"EPSG\",24370]]",
        24371=>"PROJCS[\"Kalianpur_1880_India_Zone_I\",GEOGCS[\"GCS_Kalianpur_1880\",DATUM[\"D_Kalianpur_1880\",SPHEROID[\"Everest_1830\",6377299.36,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",68.0],PARAMETER[\"Standard_Parallel_1\",32.5],PARAMETER[\"Scale_Factor\",0.99878641],PARAMETER[\"Latitude_Of_Origin\",32.5],UNIT[\"Yard_Indian\",0.9143985307444408],AUTHORITY[\"EPSG\",24371]]",
        24372=>"PROJCS[\"Kalianpur_1880_India_Zone_IIa\",GEOGCS[\"GCS_Kalianpur_1880\",DATUM[\"D_Kalianpur_1880\",SPHEROID[\"Everest_1830\",6377299.36,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",74.0],PARAMETER[\"Standard_Parallel_1\",26.0],PARAMETER[\"Scale_Factor\",0.99878641],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Yard_Indian\",0.9143985307444408],AUTHORITY[\"EPSG\",24372]]",
        24373=>"PROJCS[\"Kalianpur_1880_India_Zone_III\",GEOGCS[\"GCS_Kalianpur_1880\",DATUM[\"D_Kalianpur_1880\",SPHEROID[\"Everest_1830\",6377299.36,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",80.0],PARAMETER[\"Standard_Parallel_1\",19.0],PARAMETER[\"Scale_Factor\",0.99878641],PARAMETER[\"Latitude_Of_Origin\",19.0],UNIT[\"Yard_Indian\",0.9143985307444408],AUTHORITY[\"EPSG\",24373]]",
        24374=>"PROJCS[\"Kalianpur_1880_India_Zone_IV\",GEOGCS[\"GCS_Kalianpur_1880\",DATUM[\"D_Kalianpur_1880\",SPHEROID[\"Everest_1830\",6377299.36,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",80.0],PARAMETER[\"Standard_Parallel_1\",12.0],PARAMETER[\"Scale_Factor\",0.99878641],PARAMETER[\"Latitude_Of_Origin\",12.0],UNIT[\"Yard_Indian\",0.9143985307444408],AUTHORITY[\"EPSG\",24374]]",
        24375=>"PROJCS[\"Kalianpur_1937_India_Zone_IIb\",GEOGCS[\"GCS_Kalianpur_1937\",DATUM[\"D_Kalianpur_1937\",SPHEROID[\"Everest_Adjustment_1937\",6377276.345,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2743185.69],PARAMETER[\"False_Northing\",914395.23],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Standard_Parallel_1\",26.0],PARAMETER[\"Scale_Factor\",0.99878641],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24375]]",
        24376=>"PROJCS[\"Kalianpur_1962_India_Zone_I\",GEOGCS[\"GCS_Kalianpur_1962\",DATUM[\"D_Kalianpur_1962\",SPHEROID[\"Everest_Definition_1962\",6377301.243,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2743196.4],PARAMETER[\"False_Northing\",914398.8],PARAMETER[\"Central_Meridian\",68.0],PARAMETER[\"Standard_Parallel_1\",32.5],PARAMETER[\"Scale_Factor\",0.99878641],PARAMETER[\"Latitude_Of_Origin\",32.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24376]]",
        24377=>"PROJCS[\"Kalianpur_1962_India_Zone_IIa\",GEOGCS[\"GCS_Kalianpur_1962\",DATUM[\"D_Kalianpur_1962\",SPHEROID[\"Everest_Definition_1962\",6377301.243,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2743196.4],PARAMETER[\"False_Northing\",914398.8],PARAMETER[\"Central_Meridian\",74.0],PARAMETER[\"Standard_Parallel_1\",26.0],PARAMETER[\"Scale_Factor\",0.99878641],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24377]]",
        24378=>"PROJCS[\"Kalianpur_1975_India_Zone_I\",GEOGCS[\"GCS_Kalianpur_1975\",DATUM[\"D_Kalianpur_1975\",SPHEROID[\"Everest_Definition_1975\",6377299.151,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2743195.5],PARAMETER[\"False_Northing\",914398.5],PARAMETER[\"Central_Meridian\",68.0],PARAMETER[\"Standard_Parallel_1\",32.5],PARAMETER[\"Scale_Factor\",0.99878641],PARAMETER[\"Latitude_Of_Origin\",32.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24378]]",
        24379=>"PROJCS[\"Kalianpur_1975_India_Zone_IIa\",GEOGCS[\"GCS_Kalianpur_1975\",DATUM[\"D_Kalianpur_1975\",SPHEROID[\"Everest_Definition_1975\",6377299.151,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2743195.5],PARAMETER[\"False_Northing\",914398.5],PARAMETER[\"Central_Meridian\",74.0],PARAMETER[\"Standard_Parallel_1\",26.0],PARAMETER[\"Scale_Factor\",0.99878641],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24379]]",
        24380=>"PROJCS[\"Kalianpur_1975_India_Zone_IIb\",GEOGCS[\"GCS_Kalianpur_1975\",DATUM[\"D_Kalianpur_1975\",SPHEROID[\"Everest_Definition_1975\",6377299.151,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2743195.5],PARAMETER[\"False_Northing\",914398.5],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Standard_Parallel_1\",26.0],PARAMETER[\"Scale_Factor\",0.99878641],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24380]]",
        24381=>"PROJCS[\"Kalianpur_1975_India_Zone_III\",GEOGCS[\"GCS_Kalianpur_1975\",DATUM[\"D_Kalianpur_1975\",SPHEROID[\"Everest_Definition_1975\",6377299.151,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2743195.5],PARAMETER[\"False_Northing\",914398.5],PARAMETER[\"Central_Meridian\",80.0],PARAMETER[\"Standard_Parallel_1\",19.0],PARAMETER[\"Scale_Factor\",0.99878641],PARAMETER[\"Latitude_Of_Origin\",19.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24381]]",
        24382=>"PROJCS[\"Kalianpur_1880_India_Zone_IIb\",GEOGCS[\"GCS_Kalianpur_1880\",DATUM[\"D_Kalianpur_1880\",SPHEROID[\"Everest_1830\",6377299.36,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",90.0],PARAMETER[\"Standard_Parallel_1\",26.0],PARAMETER[\"Scale_Factor\",0.99878641],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Yard_Indian\",0.9143985307444408],AUTHORITY[\"EPSG\",24382]]",
        24383=>"PROJCS[\"Kalianpur_1975_India_Zone_IV\",GEOGCS[\"GCS_Kalianpur_1975\",DATUM[\"D_Kalianpur_1975\",SPHEROID[\"Everest_Definition_1975\",6377299.151,300.8017255]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2743195.5],PARAMETER[\"False_Northing\",914398.5],PARAMETER[\"Central_Meridian\",80.0],PARAMETER[\"Standard_Parallel_1\",12.0],PARAMETER[\"Scale_Factor\",0.99878641],PARAMETER[\"Latitude_Of_Origin\",12.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24383]]",
        24500=>"PROJCS[\"Kertau_Singapore_Grid\",GEOGCS[\"GCS_Kertau\",DATUM[\"D_Kertau\",SPHEROID[\"Everest_1830_Modified\",6377304.063,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",30000.0],PARAMETER[\"False_Northing\",30000.0],PARAMETER[\"Central_Meridian\",103.8530022222222],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",1.287646666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24500]]",
        24547=>"PROJCS[\"Kertau_UTM_Zone_47N\",GEOGCS[\"GCS_Kertau\",DATUM[\"D_Kertau\",SPHEROID[\"Everest_1830_Modified\",6377304.063,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24547]]",
        24548=>"PROJCS[\"Kertau_UTM_Zone_48N\",GEOGCS[\"GCS_Kertau\",DATUM[\"D_Kertau\",SPHEROID[\"Everest_1830_Modified\",6377304.063,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24548]]",
        24571=>"PROJCS[\"Kertau_RSO_Malaya_Chains\",GEOGCS[\"GCS_Kertau\",DATUM[\"D_Kertau\",SPHEROID[\"Everest_1830_Modified\",6377304.063,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Rectified_Skew_Orthomorphic_Natural_Origin\"],PARAMETER[\"False_Easting\",40000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Scale_Factor\",0.99984],PARAMETER[\"Azimuth\",-36.97420943711801],PARAMETER[\"Longitude_Of_Center\",102.25],PARAMETER[\"Latitude_Of_Center\",4.0],PARAMETER[\"XY_Plane_Rotation\",-36.86989764584402],UNIT[\"Chain_Benoit_1895_B\",20.11678249437587],AUTHORITY[\"EPSG\",24571]]",
        24600=>"PROJCS[\"KOC_Lambert\",GEOGCS[\"GCS_Kuwait_Oil_Company\",DATUM[\"D_Kuwait_Oil_Company\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",1166200.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Standard_Parallel_1\",32.5],PARAMETER[\"Scale_Factor\",0.998786407767],PARAMETER[\"Latitude_Of_Origin\",32.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24600]]",
        24718=>"PROJCS[\"La_Canoa_UTM_Zone_18N\",GEOGCS[\"GCS_La_Canoa\",DATUM[\"D_La_Canoa\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24718]]",
        24719=>"PROJCS[\"La_Canoa_UTM_Zone_19N\",GEOGCS[\"GCS_La_Canoa\",DATUM[\"D_La_Canoa\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24719]]",
        24720=>"PROJCS[\"La_Canoa_UTM_Zone_20N\",GEOGCS[\"GCS_La_Canoa\",DATUM[\"D_La_Canoa\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24720]]",
        24817=>"PROJCS[\"PSAD_1956_UTM_Zone_17N\",GEOGCS[\"GCS_Provisional_S_American_1956\",DATUM[\"D_Provisional_S_American_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24817]]",
        24818=>"PROJCS[\"PSAD_1956_UTM_Zone_18N\",GEOGCS[\"GCS_Provisional_S_American_1956\",DATUM[\"D_Provisional_S_American_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24818]]",
        24819=>"PROJCS[\"PSAD_1956_UTM_Zone_19N\",GEOGCS[\"GCS_Provisional_S_American_1956\",DATUM[\"D_Provisional_S_American_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24819]]",
        24820=>"PROJCS[\"PSAD_1956_UTM_Zone_20N\",GEOGCS[\"GCS_Provisional_S_American_1956\",DATUM[\"D_Provisional_S_American_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24820]]",
        24821=>"PROJCS[\"PSAD_1956_UTM_Zone_21N\",GEOGCS[\"GCS_Provisional_S_American_1956\",DATUM[\"D_Provisional_S_American_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24821]]",
        24877=>"PROJCS[\"PSAD_1956_UTM_Zone_17S\",GEOGCS[\"GCS_Provisional_S_American_1956\",DATUM[\"D_Provisional_S_American_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24877]]",
        24878=>"PROJCS[\"PSAD_1956_UTM_Zone_18S\",GEOGCS[\"GCS_Provisional_S_American_1956\",DATUM[\"D_Provisional_S_American_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24878]]",
        24879=>"PROJCS[\"PSAD_1956_UTM_Zone_19S\",GEOGCS[\"GCS_Provisional_S_American_1956\",DATUM[\"D_Provisional_S_American_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24879]]",
        24880=>"PROJCS[\"PSAD_1956_UTM_Zone_20S\",GEOGCS[\"GCS_Provisional_S_American_1956\",DATUM[\"D_Provisional_S_American_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24880]]",
        24881=>"PROJCS[\"PSAD_1956_UTM_Zone_21S\",GEOGCS[\"GCS_Provisional_S_American_1956\",DATUM[\"D_Provisional_S_American_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24881]]",
        24882=>"PROJCS[\"PSAD_1956_UTM_Zone_22S\",GEOGCS[\"GCS_Provisional_S_American_1956\",DATUM[\"D_Provisional_S_American_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24882]]",
        24891=>"PROJCS[\"Peru_West_Zone\",GEOGCS[\"GCS_Provisional_S_American_1956\",DATUM[\"D_Provisional_S_American_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",222000.0],PARAMETER[\"False_Northing\",1426834.743],PARAMETER[\"Central_Meridian\",-80.5],PARAMETER[\"Scale_Factor\",0.99983008],PARAMETER[\"Latitude_Of_Origin\",-6.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24891]]",
        24892=>"PROJCS[\"Peru_Central_Zone\",GEOGCS[\"GCS_Provisional_S_American_1956\",DATUM[\"D_Provisional_S_American_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",720000.0],PARAMETER[\"False_Northing\",1039979.159],PARAMETER[\"Central_Meridian\",-76.0],PARAMETER[\"Scale_Factor\",0.99932994],PARAMETER[\"Latitude_Of_Origin\",-9.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24892]]",
        24893=>"PROJCS[\"Peru_East_Zone\",GEOGCS[\"GCS_Provisional_S_American_1956\",DATUM[\"D_Provisional_S_American_1956\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1324000.0],PARAMETER[\"False_Northing\",1040084.558],PARAMETER[\"Central_Meridian\",-70.5],PARAMETER[\"Scale_Factor\",0.99952992],PARAMETER[\"Latitude_Of_Origin\",-9.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",24893]]",
        25000=>"PROJCS[\"Ghana_Metre_Grid\",GEOGCS[\"GCS_Leigon\",DATUM[\"D_Leigon\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",274319.51],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-1.0],PARAMETER[\"Scale_Factor\",0.99975],PARAMETER[\"Latitude_Of_Origin\",4.666666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25000]]",
        25231=>"PROJCS[\"Lome_UTM_Zone_31N\",GEOGCS[\"GCS_Lome\",DATUM[\"D_Lome\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25231]]",
        25391=>"PROJCS[\"Philippines_Zone_I\",GEOGCS[\"GCS_Luzon_1911\",DATUM[\"D_Luzon_1911\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25391]]",
        25392=>"PROJCS[\"Philippines_Zone_II\",GEOGCS[\"GCS_Luzon_1911\",DATUM[\"D_Luzon_1911\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",119.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25392]]",
        25393=>"PROJCS[\"Philippines_Zone_III\",GEOGCS[\"GCS_Luzon_1911\",DATUM[\"D_Luzon_1911\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",121.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25393]]",
        25394=>"PROJCS[\"Philippines_Zone_IV\",GEOGCS[\"GCS_Luzon_1911\",DATUM[\"D_Luzon_1911\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25394]]",
        25395=>"PROJCS[\"Philippines_Zone_V\",GEOGCS[\"GCS_Luzon_1911\",DATUM[\"D_Luzon_1911\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",125.0],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25395]]",
        25828=>"PROJCS[\"ETRS_1989_UTM_Zone_28N\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25828]]",
        25829=>"PROJCS[\"ETRS_1989_UTM_Zone_29N\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25829]]",
        25830=>"PROJCS[\"ETRS_1989_UTM_Zone_30N\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25830]]",
        25831=>"PROJCS[\"ETRS_1989_UTM_Zone_31N\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25831]]",
        25832=>"PROJCS[\"ETRS_1989_UTM_Zone_32N\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25832]]",
        25833=>"PROJCS[\"ETRS_1989_UTM_Zone_33N\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25833]]",
        25834=>"PROJCS[\"ETRS_1989_UTM_Zone_34N\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25834]]",
        25835=>"PROJCS[\"ETRS_1989_UTM_Zone_35N\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25835]]",
        25836=>"PROJCS[\"ETRS_1989_UTM_Zone_36N\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25836]]",
        25837=>"PROJCS[\"ETRS_1989_UTM_Zone_37N\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25837]]",
        25838=>"PROJCS[\"ETRS_1989_UTM_Zone_38N\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25838]]",
        25884=>"PROJCS[\"ETRS_1989_TM_Baltic_1993\",GEOGCS[\"GCS_ETRS_1989\",DATUM[\"D_ETRS_1989\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25884]]",
        25932=>"PROJCS[\"Malongo_1987_UTM_Zone_32S\",GEOGCS[\"GCS_Malongo_1987\",DATUM[\"D_Malongo_1987\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",25932]]",
        26191=>"PROJCS[\"Nord_Maroc\",GEOGCS[\"GCS_Merchich\",DATUM[\"D_Merchich\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",300000.0],PARAMETER[\"Central_Meridian\",-6.0],PARAMETER[\"Standard_Parallel_1\",37.0],PARAMETER[\"Scale_Factor\",0.999625769],PARAMETER[\"Latitude_Of_Origin\",37.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26191]]",
        26192=>"PROJCS[\"Sud_Maroc\",GEOGCS[\"GCS_Merchich\",DATUM[\"D_Merchich\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",300000.0],PARAMETER[\"Central_Meridian\",-6.0],PARAMETER[\"Standard_Parallel_1\",33.0],PARAMETER[\"Scale_Factor\",0.999615596],PARAMETER[\"Latitude_Of_Origin\",33.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26192]]",
        26193=>"PROJCS[\"Sahara\",GEOGCS[\"GCS_Merchich\",DATUM[\"D_Merchich\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1200000.0],PARAMETER[\"False_Northing\",400000.0],PARAMETER[\"Central_Meridian\",-6.0],PARAMETER[\"Standard_Parallel_1\",29.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",29.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26193]]",
        26194=>"PROJCS[\"Merchich_Sahara_Nord\",GEOGCS[\"GCS_Merchich\",DATUM[\"D_Merchich\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1200000.0],PARAMETER[\"False_Northing\",400000.0],PARAMETER[\"Central_Meridian\",-6.0],PARAMETER[\"Standard_Parallel_1\",29.0],PARAMETER[\"Scale_Factor\",0.999616304],PARAMETER[\"Latitude_Of_Origin\",29.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26194]]",
        26195=>"PROJCS[\"Merchich_Sahara_Sud\",GEOGCS[\"GCS_Merchich\",DATUM[\"D_Merchich\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",400000.0],PARAMETER[\"Central_Meridian\",-6.0],PARAMETER[\"Standard_Parallel_1\",25.0],PARAMETER[\"Scale_Factor\",0.999616437],PARAMETER[\"Latitude_Of_Origin\",25.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26195]]",
        26237=>"PROJCS[\"Massawa_UTM_Zone_37N\",GEOGCS[\"GCS_Massawa\",DATUM[\"D_Massawa\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26237]]",
        26331=>"PROJCS[\"Minna_UTM_Zone_31N\",GEOGCS[\"GCS_Minna\",DATUM[\"D_Minna\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26331]]",
        26332=>"PROJCS[\"Minna_UTM_Zone_32N\",GEOGCS[\"GCS_Minna\",DATUM[\"D_Minna\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26332]]",
        26391=>"PROJCS[\"Nigeria_West_Belt\",GEOGCS[\"GCS_Minna\",DATUM[\"D_Minna\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",230738.26],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",4.5],PARAMETER[\"Scale_Factor\",0.99975],PARAMETER[\"Latitude_Of_Origin\",4.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26391]]",
        26392=>"PROJCS[\"Nigeria_Mid_Belt\",GEOGCS[\"GCS_Minna\",DATUM[\"D_Minna\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",670553.98],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",8.5],PARAMETER[\"Scale_Factor\",0.99975],PARAMETER[\"Latitude_Of_Origin\",4.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26392]]",
        26393=>"PROJCS[\"Nigeria_East_Belt\",GEOGCS[\"GCS_Minna\",DATUM[\"D_Minna\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1110369.7],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.5],PARAMETER[\"Scale_Factor\",0.99975],PARAMETER[\"Latitude_Of_Origin\",4.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26393]]",
        26432=>"PROJCS[\"Mhast_UTM_Zone_32S\",GEOGCS[\"GCS_Mhast\",DATUM[\"D_Mhast\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26432]]",
        26591=>"PROJCS[\"Monte_Mario_Rome_Italy_1\",GEOGCS[\"GCS_Monte_Mario_Rome\",DATUM[\"D_Monte_Mario\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Rome\",12.45233333333333],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-3.45233333],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26591]]",
        26592=>"PROJCS[\"Monte_Mario_Rome_Italy_2\",GEOGCS[\"GCS_Monte_Mario_Rome\",DATUM[\"D_Monte_Mario\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Rome\",12.45233333333333],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2520000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",2.54766667],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26592]]",
        26632=>"PROJCS[\"Mporaloko_UTM_Zone_32N\",GEOGCS[\"GCS_Mporaloko\",DATUM[\"D_Mporaloko\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26632]]",
        26692=>"PROJCS[\"Mporaloko_UTM_Zone_32S\",GEOGCS[\"GCS_Mporaloko\",DATUM[\"D_Mporaloko\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26692]]",
        26701=>"PROJCS[\"NAD_1927_UTM_Zone_1N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26701]]",
        26702=>"PROJCS[\"NAD_1927_UTM_Zone_2N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26702]]",
        26703=>"PROJCS[\"NAD_1927_UTM_Zone_3N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26703]]",
        26704=>"PROJCS[\"NAD_1927_UTM_Zone_4N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26704]]",
        26705=>"PROJCS[\"NAD_1927_UTM_Zone_5N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26705]]",
        26706=>"PROJCS[\"NAD_1927_UTM_Zone_6N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26706]]",
        26707=>"PROJCS[\"NAD_1927_UTM_Zone_7N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26707]]",
        26708=>"PROJCS[\"NAD_1927_UTM_Zone_8N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26708]]",
        26709=>"PROJCS[\"NAD_1927_UTM_Zone_9N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26709]]",
        26710=>"PROJCS[\"NAD_1927_UTM_Zone_10N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26710]]",
        26711=>"PROJCS[\"NAD_1927_UTM_Zone_11N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26711]]",
        26712=>"PROJCS[\"NAD_1927_UTM_Zone_12N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26712]]",
        26713=>"PROJCS[\"NAD_1927_UTM_Zone_13N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26713]]",
        26714=>"PROJCS[\"NAD_1927_UTM_Zone_14N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26714]]",
        26715=>"PROJCS[\"NAD_1927_UTM_Zone_15N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26715]]",
        26716=>"PROJCS[\"NAD_1927_UTM_Zone_16N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26716]]",
        26717=>"PROJCS[\"NAD_1927_UTM_Zone_17N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26717]]",
        26718=>"PROJCS[\"NAD_1927_UTM_Zone_18N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26718]]",
        26719=>"PROJCS[\"NAD_1927_UTM_Zone_19N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26719]]",
        26720=>"PROJCS[\"NAD_1927_UTM_Zone_20N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26720]]",
        26721=>"PROJCS[\"NAD_1927_UTM_Zone_21N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26721]]",
        26722=>"PROJCS[\"NAD_1927_UTM_Zone_22N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26722]]",
        26729=>"PROJCS[\"NAD_1927_StatePlane_Alabama_East_FIPS_0101\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-85.83333333333333],PARAMETER[\"Scale_Factor\",0.99996],PARAMETER[\"Latitude_Of_Origin\",30.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26729]]",
        26730=>"PROJCS[\"NAD_1927_StatePlane_Alabama_West_FIPS_0102\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.5],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26730]]",
        26731=>"PROJCS[\"NAD_1927_StatePlane_Alaska_1_FIPS_5001\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Natural_Origin\"],PARAMETER[\"False_Easting\",16404166.666667],PARAMETER[\"False_Northing\",-16404166.666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Azimuth\",-36.86989764583333],PARAMETER[\"Longitude_Of_Center\",-133.6666666666667],PARAMETER[\"Latitude_Of_Center\",57.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26731]]",
        26732=>"PROJCS[\"NAD_1927_StatePlane_Alaska_2_FIPS_5002\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-142.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26732]]",
        26733=>"PROJCS[\"NAD_1927_StatePlane_Alaska_3_FIPS_5003\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-146.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26733]]",
        26734=>"PROJCS[\"NAD_1927_StatePlane_Alaska_4_FIPS_5004\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-150.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26734]]",
        26735=>"PROJCS[\"NAD_1927_StatePlane_Alaska_5_FIPS_5005\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-154.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26735]]",
        26736=>"PROJCS[\"NAD_1927_StatePlane_Alaska_6_FIPS_5006\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-158.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26736]]",
        26737=>"PROJCS[\"NAD_1927_StatePlane_Alaska_7_FIPS_5007\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-162.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26737]]",
        26738=>"PROJCS[\"NAD_1927_StatePlane_Alaska_8_FIPS_5008\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-166.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26738]]",
        26739=>"PROJCS[\"NAD_1927_StatePlane_Alaska_9_FIPS_5009\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-170.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26739]]",
        26740=>"PROJCS[\"NAD_1927_StatePlane_Alaska_10_FIPS_5010\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-176.0],PARAMETER[\"Standard_Parallel_1\",51.83333333333334],PARAMETER[\"Standard_Parallel_2\",53.83333333333334],PARAMETER[\"Latitude_Of_Origin\",51.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26740]]",
        26741=>"PROJCS[\"NAD_1927_StatePlane_California_I_FIPS_0401\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Standard_Parallel_2\",41.66666666666666],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26741]]",
        26742=>"PROJCS[\"NAD_1927_StatePlane_California_II_FIPS_0402\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",38.33333333333334],PARAMETER[\"Standard_Parallel_2\",39.83333333333334],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26742]]",
        26743=>"PROJCS[\"NAD_1927_StatePlane_California_III_FIPS_0403\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",37.06666666666667],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26743]]",
        26744=>"PROJCS[\"NAD_1927_StatePlane_California_IV_FIPS_0404\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-119.0],PARAMETER[\"Standard_Parallel_1\",36.0],PARAMETER[\"Standard_Parallel_2\",37.25],PARAMETER[\"Latitude_Of_Origin\",35.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26744]]",
        26745=>"PROJCS[\"NAD_1927_StatePlane_California_V_FIPS_0405\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-118.0],PARAMETER[\"Standard_Parallel_1\",34.03333333333333],PARAMETER[\"Standard_Parallel_2\",35.46666666666667],PARAMETER[\"Latitude_Of_Origin\",33.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26745]]",
        26746=>"PROJCS[\"NAD_1927_StatePlane_California_VI_FIPS_0406\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-116.25],PARAMETER[\"Standard_Parallel_1\",32.78333333333333],PARAMETER[\"Standard_Parallel_2\",33.88333333333333],PARAMETER[\"Latitude_Of_Origin\",32.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26746]]",
        26747=>"PROJCS[\"NAD_1927_StatePlane_California_VII_FIPS_0407\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4186692.58],PARAMETER[\"False_Northing\",4160926.74],PARAMETER[\"Central_Meridian\",-118.3333333333333],PARAMETER[\"Standard_Parallel_1\",33.86666666666667],PARAMETER[\"Standard_Parallel_2\",34.41666666666666],PARAMETER[\"Latitude_Of_Origin\",34.13333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26799]]",
        26748=>"PROJCS[\"NAD_1927_StatePlane_Arizona_East_FIPS_0201\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-110.1666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26748]]",
        26749=>"PROJCS[\"NAD_1927_StatePlane_Arizona_Central_FIPS_0202\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.9166666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26749]]",
        26750=>"PROJCS[\"NAD_1927_StatePlane_Arizona_West_FIPS_0203\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-113.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26750]]",
        26751=>"PROJCS[\"NAD_1927_StatePlane_Arkansas_North_FIPS_0301\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",34.93333333333333],PARAMETER[\"Standard_Parallel_2\",36.23333333333333],PARAMETER[\"Latitude_Of_Origin\",34.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26751]]",
        26752=>"PROJCS[\"NAD_1927_StatePlane_Arkansas_South_FIPS_0302\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",33.3],PARAMETER[\"Standard_Parallel_2\",34.76666666666667],PARAMETER[\"Latitude_Of_Origin\",32.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26752]]",
        26753=>"PROJCS[\"NAD_1927_StatePlane_Colorado_North_FIPS_0501\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",39.71666666666667],PARAMETER[\"Standard_Parallel_2\",40.78333333333333],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26753]]",
        26754=>"PROJCS[\"NAD_1927_StatePlane_Colorado_Central_FIPS_0502\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",38.45],PARAMETER[\"Standard_Parallel_2\",39.75],PARAMETER[\"Latitude_Of_Origin\",37.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26754]]",
        26755=>"PROJCS[\"NAD_1927_StatePlane_Colorado_South_FIPS_0503\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",37.23333333333333],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26755]]",
        26756=>"PROJCS[\"NAD_1927_StatePlane_Connecticut_FIPS_0600\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-72.75],PARAMETER[\"Standard_Parallel_1\",41.2],PARAMETER[\"Standard_Parallel_2\",41.86666666666667],PARAMETER[\"Latitude_Of_Origin\",40.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26756]]",
        26757=>"PROJCS[\"NAD_1927_StatePlane_Delaware_FIPS_0700\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.41666666666667],PARAMETER[\"Scale_Factor\",0.999995],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26757]]",
        26758=>"PROJCS[\"NAD_1927_StatePlane_Florida_East_FIPS_0901\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26758]]",
        26759=>"PROJCS[\"NAD_1927_StatePlane_Florida_West_FIPS_0902\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26759]]",
        26760=>"PROJCS[\"NAD_1927_StatePlane_Florida_North_FIPS_0903\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.5],PARAMETER[\"Standard_Parallel_1\",29.58333333333333],PARAMETER[\"Standard_Parallel_2\",30.75],PARAMETER[\"Latitude_Of_Origin\",29.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26760]]",
        26766=>"PROJCS[\"NAD_1927_StatePlane_Georgia_East_FIPS_1001\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26766]]",
        26767=>"PROJCS[\"NAD_1927_StatePlane_Georgia_West_FIPS_1002\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26767]]",
        26768=>"PROJCS[\"NAD_1927_StatePlane_Idaho_East_FIPS_1101\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-112.1666666666667],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26768]]",
        26769=>"PROJCS[\"NAD_1927_StatePlane_Idaho_Central_FIPS_1102\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-114.0],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26769]]",
        26770=>"PROJCS[\"NAD_1927_StatePlane_Idaho_West_FIPS_1103\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-115.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26770]]",
        26771=>"PROJCS[\"NAD_1927_StatePlane_Illinois_East_FIPS_1201\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.33333333333333],PARAMETER[\"Scale_Factor\",0.999975],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26771]]",
        26772=>"PROJCS[\"NAD_1927_StatePlane_Illinois_West_FIPS_1202\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.16666666666667],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26772]]",
        26773=>"PROJCS[\"NAD_1927_StatePlane_Indiana_East_FIPS_1301\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-85.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26773]]",
        26774=>"PROJCS[\"NAD_1927_StatePlane_Indiana_West_FIPS_1302\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.08333333333333],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26774]]",
        26775=>"PROJCS[\"NAD_1927_StatePlane_Iowa_North_FIPS_1401\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",42.06666666666667],PARAMETER[\"Standard_Parallel_2\",43.26666666666667],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26775]]",
        26776=>"PROJCS[\"NAD_1927_StatePlane_Iowa_South_FIPS_1402\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",40.61666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26776]]",
        26777=>"PROJCS[\"NAD_1927_StatePlane_Kansas_North_FIPS_1501\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",38.71666666666667],PARAMETER[\"Standard_Parallel_2\",39.78333333333333],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26777]]",
        26778=>"PROJCS[\"NAD_1927_StatePlane_Kansas_South_FIPS_1502\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",37.26666666666667],PARAMETER[\"Standard_Parallel_2\",38.56666666666667],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26778]]",
        26779=>"PROJCS[\"NAD_1927_StatePlane_Kentucky_North_FIPS_1601\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.25],PARAMETER[\"Standard_Parallel_1\",37.96666666666667],PARAMETER[\"Standard_Parallel_2\",38.96666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26779]]",
        26780=>"PROJCS[\"NAD_1927_StatePlane_Kentucky_South_FIPS_1602\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Standard_Parallel_1\",36.73333333333333],PARAMETER[\"Standard_Parallel_2\",37.93333333333333],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26780]]",
        26781=>"PROJCS[\"NAD_1927_StatePlane_Louisiana_North_FIPS_1701\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.5],PARAMETER[\"Standard_Parallel_1\",31.16666666666667],PARAMETER[\"Standard_Parallel_2\",32.66666666666666],PARAMETER[\"Latitude_Of_Origin\",30.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26781]]",
        26782=>"PROJCS[\"NAD_1927_StatePlane_Louisiana_South_FIPS_1702\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-91.33333333333333],PARAMETER[\"Standard_Parallel_1\",29.3],PARAMETER[\"Standard_Parallel_2\",30.7],PARAMETER[\"Latitude_Of_Origin\",28.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26782]]",
        26783=>"PROJCS[\"NAD_1927_StatePlane_Maine_East_FIPS_1801\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-68.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26783]]",
        26784=>"PROJCS[\"NAD_1927_StatePlane_Maine_West_FIPS_1802\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.16666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26784]]",
        26785=>"PROJCS[\"NAD_1927_StatePlane_Maryland_FIPS_1900\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.0],PARAMETER[\"Standard_Parallel_1\",38.3],PARAMETER[\"Standard_Parallel_2\",39.45],PARAMETER[\"Latitude_Of_Origin\",37.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26785]]",
        26786=>"PROJCS[\"NAD_1927_StatePlane_Massachusetts_Mainland_FIPS_2001\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Standard_Parallel_1\",41.71666666666667],PARAMETER[\"Standard_Parallel_2\",42.68333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26786]]",
        26787=>"PROJCS[\"NAD_1927_StatePlane_Massachusetts_Island_FIPS_2002\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.5],PARAMETER[\"Standard_Parallel_1\",41.28333333333333],PARAMETER[\"Standard_Parallel_2\",41.48333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26787]]",
        26791=>"PROJCS[\"NAD_1927_StatePlane_Minnesota_North_FIPS_2201\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.1],PARAMETER[\"Standard_Parallel_1\",47.03333333333333],PARAMETER[\"Standard_Parallel_2\",48.63333333333333],PARAMETER[\"Latitude_Of_Origin\",46.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26791]]",
        26792=>"PROJCS[\"NAD_1927_StatePlane_Minnesota_Central_FIPS_2202\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-94.25],PARAMETER[\"Standard_Parallel_1\",45.61666666666667],PARAMETER[\"Standard_Parallel_2\",47.05],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26792]]",
        26793=>"PROJCS[\"NAD_1927_StatePlane_Minnesota_South_FIPS_2203\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-94.0],PARAMETER[\"Standard_Parallel_1\",43.78333333333333],PARAMETER[\"Standard_Parallel_2\",45.21666666666667],PARAMETER[\"Latitude_Of_Origin\",43.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26793]]",
        26794=>"PROJCS[\"NAD_1927_StatePlane_Mississippi_East_FIPS_2301\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.83333333333333],PARAMETER[\"Scale_Factor\",0.99996],PARAMETER[\"Latitude_Of_Origin\",29.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26794]]",
        26795=>"PROJCS[\"NAD_1927_StatePlane_Mississippi_West_FIPS_2302\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.33333333333333],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",30.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26795]]",
        26796=>"PROJCS[\"NAD_1927_StatePlane_Missouri_East_FIPS_2401\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.5],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",35.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26796]]",
        26797=>"PROJCS[\"NAD_1927_StatePlane_Missouri_Central_FIPS_2402\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.5],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",35.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26797]]",
        26798=>"PROJCS[\"NAD_1927_StatePlane_Missouri_West_FIPS_2403\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-94.5],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",36.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26798]]",
        26799=>"PROJCS[\"NAD_1927_StatePlane_California_VII_FIPS_0407\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4186692.58],PARAMETER[\"False_Northing\",4160926.74],PARAMETER[\"Central_Meridian\",-118.3333333333333],PARAMETER[\"Standard_Parallel_1\",33.86666666666667],PARAMETER[\"Standard_Parallel_2\",34.41666666666666],PARAMETER[\"Latitude_Of_Origin\",34.13333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26799]]",
        26801=>"PROJCS[\"NAD_Michigan_StatePlane_Michigan_East_Old_FIPS_2101\",GEOGCS[\"GCS_North_American_Michigan\",DATUM[\"D_North_American_Michigan\",SPHEROID[\"Clarke_1866_Michigan\",6378450.047,294.978684677]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-83.66666666666667],PARAMETER[\"Scale_Factor\",0.9999428571],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26801]]",
        26802=>"PROJCS[\"NAD_Michigan_StatePlane_Michigan_Central_Old_FIPS_2102\",GEOGCS[\"GCS_North_American_Michigan\",DATUM[\"D_North_American_Michigan\",SPHEROID[\"Clarke_1866_Michigan\",6378450.047,294.978684677]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Scale_Factor\",0.9999090909],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26802]]",
        26803=>"PROJCS[\"NAD_Michigan_StatePlane_Michigan_West_Old_FIPS_2103\",GEOGCS[\"GCS_North_American_Michigan\",DATUM[\"D_North_American_Michigan\",SPHEROID[\"Clarke_1866_Michigan\",6378450.047,294.978684677]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.75],PARAMETER[\"Scale_Factor\",0.9999090909],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26803]]",
        26811=>"PROJCS[\"NAD_Michigan_StatePlane_Michigan_North_FIPS_2111\",GEOGCS[\"GCS_North_American_Michigan\",DATUM[\"D_North_American_Michigan\",SPHEROID[\"Clarke_1866_Michigan\",6378450.047,294.978684677]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Standard_Parallel_1\",45.48333333333333],PARAMETER[\"Standard_Parallel_2\",47.08333333333334],PARAMETER[\"Latitude_Of_Origin\",44.78333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26811]]",
        26812=>"PROJCS[\"NAD_Michigan_StatePlane_Michigan_Central_FIPS_2112\",GEOGCS[\"GCS_North_American_Michigan\",DATUM[\"D_North_American_Michigan\",SPHEROID[\"Clarke_1866_Michigan\",6378450.047,294.978684677]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.33333333333333],PARAMETER[\"Standard_Parallel_1\",44.18333333333333],PARAMETER[\"Standard_Parallel_2\",45.7],PARAMETER[\"Latitude_Of_Origin\",43.31666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26812]]",
        26813=>"PROJCS[\"NAD_Michigan_StatePlane_Michigan_South_FIPS_2113\",GEOGCS[\"GCS_North_American_Michigan\",DATUM[\"D_North_American_Michigan\",SPHEROID[\"Clarke_1866_Michigan\",6378450.047,294.978684677]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.33333333333333],PARAMETER[\"Standard_Parallel_1\",42.1],PARAMETER[\"Standard_Parallel_2\",43.66666666666666],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26813]]",
        26847=>"PROJCS[\"NAD_1983_StatePlane_Maine_East_FIPS_1801_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-68.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26847]]",
        26848=>"PROJCS[\"NAD_1983_StatePlane_Maine_West_FIPS_1802_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2952750.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.16666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26848]]",
        26849=>"PROJCS[\"NAD_1983_StatePlane_Minnesota_North_FIPS_2201_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-93.1],PARAMETER[\"Standard_Parallel_1\",47.03333333333333],PARAMETER[\"Standard_Parallel_2\",48.63333333333333],PARAMETER[\"Latitude_Of_Origin\",46.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26849]]",
        26850=>"PROJCS[\"NAD_1983_StatePlane_Minnesota_Central_FIPS_2202_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-94.25],PARAMETER[\"Standard_Parallel_1\",45.61666666666667],PARAMETER[\"Standard_Parallel_2\",47.05],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26850]]",
        26851=>"PROJCS[\"NAD_1983_StatePlane_Minnesota_South_FIPS_2203_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-94.0],PARAMETER[\"Standard_Parallel_1\",43.78333333333333],PARAMETER[\"Standard_Parallel_2\",45.21666666666667],PARAMETER[\"Latitude_Of_Origin\",43.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26851]]",
        26852=>"PROJCS[\"NAD_1983_StatePlane_Nebraska_FIPS_2600_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Standard_Parallel_2\",43.0],PARAMETER[\"Latitude_Of_Origin\",39.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26852]]",
        26853=>"PROJCS[\"NAD_1983_StatePlane_West_Virginia_North_FIPS_4701_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Standard_Parallel_1\",39.0],PARAMETER[\"Standard_Parallel_2\",40.25],PARAMETER[\"Latitude_Of_Origin\",38.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26853]]",
        26854=>"PROJCS[\"NAD_1983_StatePlane_West_Virginia_South_FIPS_4702_Feet\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",37.48333333333333],PARAMETER[\"Standard_Parallel_2\",38.88333333333333],PARAMETER[\"Latitude_Of_Origin\",37.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26854]]",
        26855=>"PROJCS[\"NAD_1983_HARN_StatePlane_Maine_East_FIPS_1801_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-68.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26855]]",
        26856=>"PROJCS[\"NAD_1983_HARN_StatePlane_Maine_West_FIPS_1802_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2952750.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.16666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26856]]",
        26857=>"PROJCS[\"NAD_1983_HARN_StatePlane_Minnesota_North_FIPS_2201_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-93.1],PARAMETER[\"Standard_Parallel_1\",47.03333333333333],PARAMETER[\"Standard_Parallel_2\",48.63333333333333],PARAMETER[\"Latitude_Of_Origin\",46.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26857]]",
        26858=>"PROJCS[\"NAD_1983_HARN_StatePlane_Minnesota_Central_FIPS_2202_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-94.25],PARAMETER[\"Standard_Parallel_1\",45.61666666666667],PARAMETER[\"Standard_Parallel_2\",47.05],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26858]]",
        26859=>"PROJCS[\"NAD_1983_HARN_StatePlane_Minnesota_South_FIPS_2203_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-94.0],PARAMETER[\"Standard_Parallel_1\",43.78333333333333],PARAMETER[\"Standard_Parallel_2\",45.21666666666667],PARAMETER[\"Latitude_Of_Origin\",43.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26859]]",
        26860=>"PROJCS[\"NAD_1983_HARN_StatePlane_Nebraska_FIPS_2600_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Standard_Parallel_2\",43.0],PARAMETER[\"Latitude_Of_Origin\",39.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26860]]",
        26861=>"PROJCS[\"NAD_1983_HARN_StatePlane_West_Virginia_North_FIPS_4701_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Standard_Parallel_1\",39.0],PARAMETER[\"Standard_Parallel_2\",40.25],PARAMETER[\"Latitude_Of_Origin\",38.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26861]]",
        26862=>"PROJCS[\"NAD_1983_HARN_StatePlane_West_Virginia_South_FIPS_4702_Feet\",GEOGCS[\"GCS_North_American_1983_HARN\",DATUM[\"D_North_American_1983_HARN\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",37.48333333333333],PARAMETER[\"Standard_Parallel_2\",38.88333333333333],PARAMETER[\"Latitude_Of_Origin\",37.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26862]]",
        26863=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Maine_East_FIPS_1801_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",984250.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-68.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26863]]",
        26864=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Maine_West_FIPS_1802_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2952750.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.16666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26864]]",
        26865=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Minnesota_North_FIPS_2201_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-93.1],PARAMETER[\"Standard_Parallel_1\",47.03333333333333],PARAMETER[\"Standard_Parallel_2\",48.63333333333333],PARAMETER[\"Latitude_Of_Origin\",46.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26865]]",
        26866=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Minnesota_Central_FIPS_2202_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-94.25],PARAMETER[\"Standard_Parallel_1\",45.61666666666667],PARAMETER[\"Standard_Parallel_2\",47.05],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26866]]",
        26867=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Minnesota_South_FIPS_2203_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2624666.666666666],PARAMETER[\"False_Northing\",328083.3333333333],PARAMETER[\"Central_Meridian\",-94.0],PARAMETER[\"Standard_Parallel_1\",43.78333333333333],PARAMETER[\"Standard_Parallel_2\",45.21666666666667],PARAMETER[\"Latitude_Of_Origin\",43.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26867]]",
        26868=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_Nebraska_FIPS_2600_Ft_US\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Standard_Parallel_2\",43.0],PARAMETER[\"Latitude_Of_Origin\",39.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26868]]",
        26869=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_West_Virginia_North_FIPS_4701_FtUS\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Standard_Parallel_1\",39.0],PARAMETER[\"Standard_Parallel_2\",40.25],PARAMETER[\"Latitude_Of_Origin\",38.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26869]]",
        26870=>"PROJCS[\"NAD_1983_NSRS2007_StatePlane_West_Virginia_South_FIPS_4702_FtUS\",GEOGCS[\"GCS_NAD_1983_NSRS2007\",DATUM[\"D_NAD_1983_NSRS2007\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1968500.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",37.48333333333333],PARAMETER[\"Standard_Parallel_2\",38.88333333333333],PARAMETER[\"Latitude_Of_Origin\",37.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",26870]]",
        26891=>"PROJCS[\"NAD_1983_CSRS_MTM_11\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26891]]",
        26892=>"PROJCS[\"NAD_1983_CSRS_MTM_12\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26892]]",
        26893=>"PROJCS[\"NAD_1983_CSRS_MTM_13\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26893]]",
        26894=>"PROJCS[\"NAD_1983_CSRS_MTM_14\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26894]]",
        26895=>"PROJCS[\"NAD_1983_CSRS_MTM_15\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26895]]",
        26896=>"PROJCS[\"NAD_1983_CSRS_MTM_16\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26896]]",
        26897=>"PROJCS[\"NAD_1983_CSRS_MTM_17\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-96.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26897]]",
        26898=>"PROJCS[\"NAD_1983_CSRS_MTM_1\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-53.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26898]]",
        26899=>"PROJCS[\"NAD_1983_CSRS_MTM_2\",GEOGCS[\"GCS_North_American_1983_CSRS\",DATUM[\"D_North_American_1983_CSRS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-56.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26899]]",
        26901=>"PROJCS[\"NAD_1983_UTM_Zone_1N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26901]]",
        26902=>"PROJCS[\"NAD_1983_UTM_Zone_2N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26902]]",
        26903=>"PROJCS[\"NAD_1983_UTM_Zone_3N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26903]]",
        26904=>"PROJCS[\"NAD_1983_UTM_Zone_4N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26904]]",
        26905=>"PROJCS[\"NAD_1983_UTM_Zone_5N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26905]]",
        26906=>"PROJCS[\"NAD_1983_UTM_Zone_6N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26906]]",
        26907=>"PROJCS[\"NAD_1983_UTM_Zone_7N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26907]]",
        26908=>"PROJCS[\"NAD_1983_UTM_Zone_8N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26908]]",
        26909=>"PROJCS[\"NAD_1983_UTM_Zone_9N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26909]]",
        26910=>"PROJCS[\"NAD_1983_UTM_Zone_10N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26910]]",
        26911=>"PROJCS[\"NAD_1983_UTM_Zone_11N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26911]]",
        26912=>"PROJCS[\"NAD_1983_UTM_Zone_12N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26912]]",
        26913=>"PROJCS[\"NAD_1983_UTM_Zone_13N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26913]]",
        26914=>"PROJCS[\"NAD_1983_UTM_Zone_14N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26914]]",
        26915=>"PROJCS[\"NAD_1983_UTM_Zone_15N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26915]]",
        26916=>"PROJCS[\"NAD_1983_UTM_Zone_16N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26916]]",
        26917=>"PROJCS[\"NAD_1983_UTM_Zone_17N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26917]]",
        26918=>"PROJCS[\"NAD_1983_UTM_Zone_18N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26918]]",
        26919=>"PROJCS[\"NAD_1983_UTM_Zone_19N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26919]]",
        26920=>"PROJCS[\"NAD_1983_UTM_Zone_20N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26920]]",
        26921=>"PROJCS[\"NAD_1983_UTM_Zone_21N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26921]]",
        26922=>"PROJCS[\"NAD_1983_UTM_Zone_22N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26922]]",
        26923=>"PROJCS[\"NAD_1983_UTM_Zone_23N\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26923]]",
        26929=>"PROJCS[\"NAD_1983_StatePlane_Alabama_East_FIPS_0101\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-85.83333333333333],PARAMETER[\"Scale_Factor\",0.99996],PARAMETER[\"Latitude_Of_Origin\",30.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26929]]",
        26930=>"PROJCS[\"NAD_1983_StatePlane_Alabama_West_FIPS_0102\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.5],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26930]]",
        26931=>"PROJCS[\"NAD_1983_StatePlane_Alaska_1_FIPS_5001\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Hotine_Oblique_Mercator_Azimuth_Natural_Origin\"],PARAMETER[\"False_Easting\",5000000.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Azimuth\",-36.86989764583333],PARAMETER[\"Longitude_Of_Center\",-133.6666666666667],PARAMETER[\"Latitude_Of_Center\",57.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26931]]",
        26932=>"PROJCS[\"NAD_1983_StatePlane_Alaska_2_FIPS_5002\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-142.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26932]]",
        26933=>"PROJCS[\"NAD_1983_StatePlane_Alaska_3_FIPS_5003\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-146.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26933]]",
        26934=>"PROJCS[\"NAD_1983_StatePlane_Alaska_4_FIPS_5004\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-150.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26934]]",
        26935=>"PROJCS[\"NAD_1983_StatePlane_Alaska_5_FIPS_5005\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-154.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26935]]",
        26936=>"PROJCS[\"NAD_1983_StatePlane_Alaska_6_FIPS_5006\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-158.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26936]]",
        26937=>"PROJCS[\"NAD_1983_StatePlane_Alaska_7_FIPS_5007\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-162.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26937]]",
        26938=>"PROJCS[\"NAD_1983_StatePlane_Alaska_8_FIPS_5008\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-166.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26938]]",
        26939=>"PROJCS[\"NAD_1983_StatePlane_Alaska_9_FIPS_5009\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-170.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",54.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26939]]",
        26940=>"PROJCS[\"NAD_1983_StatePlane_Alaska_10_FIPS_5010\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-176.0],PARAMETER[\"Standard_Parallel_1\",51.83333333333334],PARAMETER[\"Standard_Parallel_2\",53.83333333333334],PARAMETER[\"Latitude_Of_Origin\",51.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26940]]",
        26941=>"PROJCS[\"NAD_1983_StatePlane_California_I_FIPS_0401\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Standard_Parallel_2\",41.66666666666666],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26941]]",
        26942=>"PROJCS[\"NAD_1983_StatePlane_California_II_FIPS_0402\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-122.0],PARAMETER[\"Standard_Parallel_1\",38.33333333333334],PARAMETER[\"Standard_Parallel_2\",39.83333333333334],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26942]]",
        26943=>"PROJCS[\"NAD_1983_StatePlane_California_III_FIPS_0403\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",37.06666666666667],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26943]]",
        26944=>"PROJCS[\"NAD_1983_StatePlane_California_IV_FIPS_0404\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-119.0],PARAMETER[\"Standard_Parallel_1\",36.0],PARAMETER[\"Standard_Parallel_2\",37.25],PARAMETER[\"Latitude_Of_Origin\",35.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26944]]",
        26945=>"PROJCS[\"NAD_1983_StatePlane_California_V_FIPS_0405\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-118.0],PARAMETER[\"Standard_Parallel_1\",34.03333333333333],PARAMETER[\"Standard_Parallel_2\",35.46666666666667],PARAMETER[\"Latitude_Of_Origin\",33.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26945]]",
        26946=>"PROJCS[\"NAD_1983_StatePlane_California_VI_FIPS_0406\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-116.25],PARAMETER[\"Standard_Parallel_1\",32.78333333333333],PARAMETER[\"Standard_Parallel_2\",33.88333333333333],PARAMETER[\"Latitude_Of_Origin\",32.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26946]]",
        26948=>"PROJCS[\"NAD_1983_StatePlane_Arizona_East_FIPS_0201\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",213360.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-110.1666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26948]]",
        26949=>"PROJCS[\"NAD_1983_StatePlane_Arizona_Central_FIPS_0202\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",213360.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.9166666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26949]]",
        26950=>"PROJCS[\"NAD_1983_StatePlane_Arizona_West_FIPS_0203\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",213360.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-113.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26950]]",
        26951=>"PROJCS[\"NAD_1983_StatePlane_Arkansas_North_FIPS_0301\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",34.93333333333333],PARAMETER[\"Standard_Parallel_2\",36.23333333333333],PARAMETER[\"Latitude_Of_Origin\",34.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26951]]",
        26952=>"PROJCS[\"NAD_1983_StatePlane_Arkansas_South_FIPS_0302\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",400000.0],PARAMETER[\"Central_Meridian\",-92.0],PARAMETER[\"Standard_Parallel_1\",33.3],PARAMETER[\"Standard_Parallel_2\",34.76666666666667],PARAMETER[\"Latitude_Of_Origin\",32.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26952]]",
        26953=>"PROJCS[\"NAD_1983_StatePlane_Colorado_North_FIPS_0501\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",914401.8289],PARAMETER[\"False_Northing\",304800.6096],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",39.71666666666667],PARAMETER[\"Standard_Parallel_2\",40.78333333333333],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26953]]",
        26954=>"PROJCS[\"NAD_1983_StatePlane_Colorado_Central_FIPS_0502\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",914401.8289],PARAMETER[\"False_Northing\",304800.6096],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",38.45],PARAMETER[\"Standard_Parallel_2\",39.75],PARAMETER[\"Latitude_Of_Origin\",37.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26954]]",
        26955=>"PROJCS[\"NAD_1983_StatePlane_Colorado_South_FIPS_0503\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",914401.8289],PARAMETER[\"False_Northing\",304800.6096],PARAMETER[\"Central_Meridian\",-105.5],PARAMETER[\"Standard_Parallel_1\",37.23333333333333],PARAMETER[\"Standard_Parallel_2\",38.43333333333333],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26955]]",
        26956=>"PROJCS[\"NAD_1983_StatePlane_Connecticut_FIPS_0600\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",304800.6096],PARAMETER[\"False_Northing\",152400.3048],PARAMETER[\"Central_Meridian\",-72.75],PARAMETER[\"Standard_Parallel_1\",41.2],PARAMETER[\"Standard_Parallel_2\",41.86666666666667],PARAMETER[\"Latitude_Of_Origin\",40.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26956]]",
        26957=>"PROJCS[\"NAD_1983_StatePlane_Delaware_FIPS_0700\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.41666666666667],PARAMETER[\"Scale_Factor\",0.999995],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26957]]",
        26958=>"PROJCS[\"NAD_1983_StatePlane_Florida_East_FIPS_0901\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26958]]",
        26959=>"PROJCS[\"NAD_1983_StatePlane_Florida_West_FIPS_0902\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.0],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",24.33333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26959]]",
        26960=>"PROJCS[\"NAD_1983_StatePlane_Florida_North_FIPS_0903\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.5],PARAMETER[\"Standard_Parallel_1\",29.58333333333333],PARAMETER[\"Standard_Parallel_2\",30.75],PARAMETER[\"Latitude_Of_Origin\",29.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26960]]",
        26961=>"PROJCS[\"NAD_1983_StatePlane_Hawaii_1_FIPS_5101\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-155.5],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",18.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26961]]",
        26962=>"PROJCS[\"NAD_1983_StatePlane_Hawaii_2_FIPS_5102\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-156.6666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",20.33333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26962]]",
        26963=>"PROJCS[\"NAD_1983_StatePlane_Hawaii_3_FIPS_5103\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-158.0],PARAMETER[\"Scale_Factor\",0.99999],PARAMETER[\"Latitude_Of_Origin\",21.16666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26963]]",
        26964=>"PROJCS[\"NAD_1983_StatePlane_Hawaii_4_FIPS_5104\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-159.5],PARAMETER[\"Scale_Factor\",0.99999],PARAMETER[\"Latitude_Of_Origin\",21.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26964]]",
        26965=>"PROJCS[\"NAD_1983_StatePlane_Hawaii_5_FIPS_5105\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-160.1666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",21.66666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26965]]",
        26966=>"PROJCS[\"NAD_1983_StatePlane_Georgia_East_FIPS_1001\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26966]]",
        26967=>"PROJCS[\"NAD_1983_StatePlane_Georgia_West_FIPS_1002\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.16666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",30.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26967]]",
        26968=>"PROJCS[\"NAD_1983_StatePlane_Idaho_East_FIPS_1101\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-112.1666666666667],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26968]]",
        26969=>"PROJCS[\"NAD_1983_StatePlane_Idaho_Central_FIPS_1102\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-114.0],PARAMETER[\"Scale_Factor\",0.9999473684210526],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26969]]",
        26970=>"PROJCS[\"NAD_1983_StatePlane_Idaho_West_FIPS_1103\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-115.75],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26970]]",
        26971=>"PROJCS[\"NAD_1983_StatePlane_Illinois_East_FIPS_1201\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.33333333333333],PARAMETER[\"Scale_Factor\",0.999975],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26971]]",
        26972=>"PROJCS[\"NAD_1983_StatePlane_Illinois_West_FIPS_1202\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.16666666666667],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26972]]",
        26973=>"PROJCS[\"NAD_1983_StatePlane_Indiana_East_FIPS_1301\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",250000.0],PARAMETER[\"Central_Meridian\",-85.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26973]]",
        26974=>"PROJCS[\"NAD_1983_StatePlane_Indiana_West_FIPS_1302\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",900000.0],PARAMETER[\"False_Northing\",250000.0],PARAMETER[\"Central_Meridian\",-87.08333333333333],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26974]]",
        26975=>"PROJCS[\"NAD_1983_StatePlane_Iowa_North_FIPS_1401\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",42.06666666666667],PARAMETER[\"Standard_Parallel_2\",43.26666666666667],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26975]]",
        26976=>"PROJCS[\"NAD_1983_StatePlane_Iowa_South_FIPS_1402\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.5],PARAMETER[\"Standard_Parallel_1\",40.61666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26976]]",
        26977=>"PROJCS[\"NAD_1983_StatePlane_Kansas_North_FIPS_1501\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",38.71666666666667],PARAMETER[\"Standard_Parallel_2\",39.78333333333333],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26977]]",
        26978=>"PROJCS[\"NAD_1983_StatePlane_Kansas_South_FIPS_1502\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",400000.0],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",37.26666666666667],PARAMETER[\"Standard_Parallel_2\",38.56666666666667],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26978]]",
        26979=>"PROJCS[\"NAD_1983_StatePlane_Kentucky_North_FIPS_1601\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.25],PARAMETER[\"Standard_Parallel_1\",37.96666666666667],PARAMETER[\"Standard_Parallel_2\",38.96666666666667],PARAMETER[\"Latitude_Of_Origin\",37.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",2205]]",
        26980=>"PROJCS[\"NAD_1983_StatePlane_Kentucky_South_FIPS_1602\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",-85.75],PARAMETER[\"Standard_Parallel_1\",36.73333333333333],PARAMETER[\"Standard_Parallel_2\",37.93333333333333],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26980]]",
        26981=>"PROJCS[\"NAD_1983_StatePlane_Louisiana_North_FIPS_1701\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.5],PARAMETER[\"Standard_Parallel_1\",31.16666666666667],PARAMETER[\"Standard_Parallel_2\",32.66666666666666],PARAMETER[\"Latitude_Of_Origin\",30.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26981]]",
        26982=>"PROJCS[\"NAD_1983_StatePlane_Louisiana_South_FIPS_1702\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-91.33333333333333],PARAMETER[\"Standard_Parallel_1\",29.3],PARAMETER[\"Standard_Parallel_2\",30.7],PARAMETER[\"Latitude_Of_Origin\",28.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26982]]",
        26983=>"PROJCS[\"NAD_1983_StatePlane_Maine_East_FIPS_1801\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-68.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26983]]",
        26984=>"PROJCS[\"NAD_1983_StatePlane_Maine_West_FIPS_1802\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",900000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.16666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26984]]",
        26985=>"PROJCS[\"NAD_1983_StatePlane_Maryland_FIPS_1900\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.0],PARAMETER[\"Standard_Parallel_1\",38.3],PARAMETER[\"Standard_Parallel_2\",39.45],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26985]]",
        26986=>"PROJCS[\"NAD_1983_StatePlane_Massachusetts_Mainland_FIPS_2001\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",750000.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Standard_Parallel_1\",41.71666666666667],PARAMETER[\"Standard_Parallel_2\",42.68333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26986]]",
        26987=>"PROJCS[\"NAD_1983_StatePlane_Massachusetts_Island_FIPS_2002\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.5],PARAMETER[\"Standard_Parallel_1\",41.28333333333333],PARAMETER[\"Standard_Parallel_2\",41.48333333333333],PARAMETER[\"Latitude_Of_Origin\",41.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26987]]",
        26988=>"PROJCS[\"NAD_1983_StatePlane_Michigan_North_FIPS_2111\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",8000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Standard_Parallel_1\",45.48333333333333],PARAMETER[\"Standard_Parallel_2\",47.08333333333334],PARAMETER[\"Latitude_Of_Origin\",44.78333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26988]]",
        26989=>"PROJCS[\"NAD_1983_StatePlane_Michigan_Central_FIPS_2112\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",6000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.36666666666666],PARAMETER[\"Standard_Parallel_1\",44.18333333333333],PARAMETER[\"Standard_Parallel_2\",45.7],PARAMETER[\"Latitude_Of_Origin\",43.31666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26989]]",
        26990=>"PROJCS[\"NAD_1983_StatePlane_Michigan_South_FIPS_2113\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",4000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.36666666666666],PARAMETER[\"Standard_Parallel_1\",42.1],PARAMETER[\"Standard_Parallel_2\",43.66666666666666],PARAMETER[\"Latitude_Of_Origin\",41.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26990]]",
        26991=>"PROJCS[\"NAD_1983_StatePlane_Minnesota_North_FIPS_2201\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-93.1],PARAMETER[\"Standard_Parallel_1\",47.03333333333333],PARAMETER[\"Standard_Parallel_2\",48.63333333333333],PARAMETER[\"Latitude_Of_Origin\",46.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26991]]",
        26992=>"PROJCS[\"NAD_1983_StatePlane_Minnesota_Central_FIPS_2202\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-94.25],PARAMETER[\"Standard_Parallel_1\",45.61666666666667],PARAMETER[\"Standard_Parallel_2\",47.05],PARAMETER[\"Latitude_Of_Origin\",45.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26992]]",
        26993=>"PROJCS[\"NAD_1983_StatePlane_Minnesota_South_FIPS_2203\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-94.0],PARAMETER[\"Standard_Parallel_1\",43.78333333333333],PARAMETER[\"Standard_Parallel_2\",45.21666666666667],PARAMETER[\"Latitude_Of_Origin\",43.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26993]]",
        26994=>"PROJCS[\"NAD_1983_StatePlane_Mississippi_East_FIPS_2301\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-88.83333333333333],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",29.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26994]]",
        26995=>"PROJCS[\"NAD_1983_StatePlane_Mississippi_West_FIPS_2302\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.33333333333333],PARAMETER[\"Scale_Factor\",0.99995],PARAMETER[\"Latitude_Of_Origin\",29.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26995]]",
        26996=>"PROJCS[\"NAD_1983_StatePlane_Missouri_East_FIPS_2401\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.5],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",35.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26996]]",
        26997=>"PROJCS[\"NAD_1983_StatePlane_Missouri_Central_FIPS_2402\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-92.5],PARAMETER[\"Scale_Factor\",0.9999333333333333],PARAMETER[\"Latitude_Of_Origin\",35.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26997]]",
        26998=>"PROJCS[\"NAD_1983_StatePlane_Missouri_West_FIPS_2403\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",850000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-94.5],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",36.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",26998]]",
        27037=>"PROJCS[\"Nahrwan_1967_UTM_Zone_37N\",GEOGCS[\"GCS_Nahrwan_1967\",DATUM[\"D_Nahrwan_1967\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27037]]",
        27038=>"PROJCS[\"Nahrwan_1967_UTM_Zone_38N\",GEOGCS[\"GCS_Nahrwan_1967\",DATUM[\"D_Nahrwan_1967\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27038]]",
        27039=>"PROJCS[\"Nahrwan_1967_UTM_Zone_39N\",GEOGCS[\"GCS_Nahrwan_1967\",DATUM[\"D_Nahrwan_1967\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27039]]",
        27040=>"PROJCS[\"Nahrwan_1967_UTM_Zone_40N\",GEOGCS[\"GCS_Nahrwan_1967\",DATUM[\"D_Nahrwan_1967\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27040]]",
        27120=>"PROJCS[\"Naparima_1972_UTM_Zone_20N\",GEOGCS[\"GCS_Naparima_1972\",DATUM[\"D_Naparima_1972\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27120]]",
        27200=>"PROJCS[\"GD_1949_New_Zealand_Map_Grid\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"New_Zealand_Map_Grid\"],PARAMETER[\"False_Easting\",2510000.0],PARAMETER[\"False_Northing\",6023150.0],PARAMETER[\"Longitude_Of_Origin\",173.0],PARAMETER[\"Latitude_Of_Origin\",-41.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27200]]",
        27205=>"PROJCS[\"NZGD_1949_Mount_Eden_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",174.7643393611111],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",-36.87986527777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27205]]",
        27206=>"PROJCS[\"NZGD_1949_Bay_of_Plenty_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",176.46619725],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-37.76124980555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27206]]",
        27207=>"PROJCS[\"NZGD_1949_Poverty_Bay_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",177.8856362777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-38.62470277777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27207]]",
        27208=>"PROJCS[\"NZGD_1949_Hawkes_Bay_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",176.6736805277778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-39.65092930555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27208]]",
        27209=>"PROJCS[\"NZGD_1949_Taranaki_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",174.22801175],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-39.13575830555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27209]]",
        27210=>"PROJCS[\"NZGD_1949_Tuhirangi_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",175.6400368055556],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-39.51247038888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27210]]",
        27211=>"PROJCS[\"NZGD_1949_Wanganui_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",175.4880996111111],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-40.24194713888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27211]]",
        27212=>"PROJCS[\"NZGD_1949_Wairarapa_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",175.6473496666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-40.92553263888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27212]]",
        27213=>"PROJCS[\"NZGD_1949_Wellington_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",174.7766231111111],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-41.30131963888888],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27213]]",
        27214=>"PROJCS[\"NZGD_1949_Collingwood_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",172.6720465],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-40.71475905555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27214]]",
        27215=>"PROJCS[\"NZGD_1949_Nelson_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",173.2993168055555],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-41.27454472222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27215]]",
        27216=>"PROJCS[\"NZGD_1949_Karamea_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",172.1090281944444],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-41.28991152777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27216]]",
        27217=>"PROJCS[\"NZGD_1949_Buller_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",171.5812600555556],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-41.81080286111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27217]]",
        27218=>"PROJCS[\"NZGD_1949_Grey_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",171.5497713055556],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-42.33369427777778],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27218]]",
        27219=>"PROJCS[\"NZGD_1949_Amuri_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",173.0101333888889],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-42.68911658333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27219]]",
        27220=>"PROJCS[\"NZGD_1949_Marlborough_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",173.8020741111111],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-41.54448666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27220]]",
        27221=>"PROJCS[\"NZGD_1949_Hokitika_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",170.9799935],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-42.88632236111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27221]]",
        27222=>"PROJCS[\"NZGD_1949_Okarito_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",170.2609258333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-43.11012813888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27222]]",
        27223=>"PROJCS[\"NZGD_1949_Jacksons_Bay_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",168.606267],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-43.97780288888889],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27223]]",
        27224=>"PROJCS[\"NZGD_1949_Mount_Pleasant_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",172.7271935833333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-43.59063758333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27224]]",
        27225=>"PROJCS[\"NZGD_1949_Gawler_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",171.3607484722222],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-43.74871155555556],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27225]]",
        27226=>"PROJCS[\"NZGD_1949_Timaru_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",171.0572508333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-44.40222036111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27226]]",
        27227=>"PROJCS[\"NZGD_1949_Lindis_Peak_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",169.4677550833333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-44.73526797222222],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27227]]",
        27228=>"PROJCS[\"NZGD_1949_Mount_Nicholas_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",168.3986411944444],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-45.13290258333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27228]]",
        27229=>"PROJCS[\"NZGD_1949_Mount_York_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",167.7388617777778],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-45.56372616666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27229]]",
        27230=>"PROJCS[\"NZGD_1949_Observation_Point_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",170.6285951666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-45.81619661111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27230]]",
        27231=>"PROJCS[\"NZGD_1949_North_Taieri_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",700000.0],PARAMETER[\"Central_Meridian\",170.2825891111111],PARAMETER[\"Scale_Factor\",0.99996],PARAMETER[\"Latitude_Of_Origin\",-45.86151336111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27231]]",
        27232=>"PROJCS[\"NZGD_1949_Bluff_Circuit\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300002.66],PARAMETER[\"False_Northing\",699999.58],PARAMETER[\"Central_Meridian\",168.342872],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-46.60000961111111],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27232]]",
        27258=>"PROJCS[\"NZGD_1949_UTM_Zone_58S\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27258]]",
        27259=>"PROJCS[\"NZGD_1949_UTM_Zone_59S\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27259]]",
        27260=>"PROJCS[\"NZGD_1949_UTM_Zone_60S\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27260]]",
        27291=>"PROJCS[\"New_Zealand_North_Island\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",400000.0],PARAMETER[\"Central_Meridian\",175.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-39.0],UNIT[\"Yard_Sears\",0.9143984146160287],AUTHORITY[\"EPSG\",27291]]",
        27292=>"PROJCS[\"New_Zealand_South_Island\",GEOGCS[\"GCS_New_Zealand_1949\",DATUM[\"D_New_Zealand_1949\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",171.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",-44.0],UNIT[\"Yard_Sears\",0.9143984146160287],AUTHORITY[\"EPSG\",27292]]",
        27391=>"PROJCS[\"NGO_1948_Oslo_Norway_Zone_1\",GEOGCS[\"GCS_NGO_1948_Oslo\",DATUM[\"D_NGO_1948\",SPHEROID[\"Bessel_Modified\",6377492.018,299.1528128]],PRIMEM[\"Oslo\",10.72291666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-4.666666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27391]]",
        27392=>"PROJCS[\"NGO_1948_Oslo_Norway_Zone_2\",GEOGCS[\"GCS_NGO_1948_Oslo\",DATUM[\"D_NGO_1948\",SPHEROID[\"Bessel_Modified\",6377492.018,299.1528128]],PRIMEM[\"Oslo\",10.72291666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-2.333333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27392]]",
        27393=>"PROJCS[\"NGO_1948_Oslo_Norway_Zone_3\",GEOGCS[\"GCS_NGO_1948_Oslo\",DATUM[\"D_NGO_1948\",SPHEROID[\"Bessel_Modified\",6377492.018,299.1528128]],PRIMEM[\"Oslo\",10.72291666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27393]]",
        27394=>"PROJCS[\"NGO_1948_Oslo_Norway_Zone_4\",GEOGCS[\"GCS_NGO_1948_Oslo\",DATUM[\"D_NGO_1948\",SPHEROID[\"Bessel_Modified\",6377492.018,299.1528128]],PRIMEM[\"Oslo\",10.72291666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",2.5],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27394]]",
        27395=>"PROJCS[\"NGO_1948_Oslo_Norway_Zone_5\",GEOGCS[\"GCS_NGO_1948_Oslo\",DATUM[\"D_NGO_1948\",SPHEROID[\"Bessel_Modified\",6377492.018,299.1528128]],PRIMEM[\"Oslo\",10.72291666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",6.166666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27395]]",
        27396=>"PROJCS[\"NGO_1948_Oslo_Norway_Zone_6\",GEOGCS[\"GCS_NGO_1948_Oslo\",DATUM[\"D_NGO_1948\",SPHEROID[\"Bessel_Modified\",6377492.018,299.1528128]],PRIMEM[\"Oslo\",10.72291666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",10.16666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27396]]",
        27397=>"PROJCS[\"NGO_1948_Oslo_Norway_Zone_7\",GEOGCS[\"GCS_NGO_1948_Oslo\",DATUM[\"D_NGO_1948\",SPHEROID[\"Bessel_Modified\",6377492.018,299.1528128]],PRIMEM[\"Oslo\",10.72291666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",14.16666666666667],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27397]]",
        27398=>"PROJCS[\"NGO_1948_Oslo_Norway_Zone_8\",GEOGCS[\"GCS_NGO_1948_Oslo\",DATUM[\"D_NGO_1948\",SPHEROID[\"Bessel_Modified\",6377492.018,299.1528128]],PRIMEM[\"Oslo\",10.72291666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",18.33333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",58.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27398]]",
        27429=>"PROJCS[\"Datum_73_UTM_Zone_29N\",GEOGCS[\"GCS_Datum_73\",DATUM[\"D_Datum_73\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27429]]",
        27492=>"PROJCS[\"Datum_73_Modified_Portuguese_Grid\",GEOGCS[\"GCS_Datum_73\",DATUM[\"D_Datum_73\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",180.598],PARAMETER[\"False_Northing\",-86.99],PARAMETER[\"Central_Meridian\",-8.131906111111112],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",39.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27493]]",
        27493=>"PROJCS[\"Datum_73_Modified_Portuguese_Grid\",GEOGCS[\"GCS_Datum_73\",DATUM[\"D_Datum_73\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",180.598],PARAMETER[\"False_Northing\",-86.99],PARAMETER[\"Central_Meridian\",-8.131906111111112],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",39.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27493]]",
        27500=>"PROJCS[\"Nord_de_Guerre\",GEOGCS[\"GCS_ATF_Paris\",DATUM[\"D_ATF\",SPHEROID[\"Plessis_1817\",6376523.0,308.64]],PRIMEM[\"Paris_RGS\",2.337208333333333],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",300000.0],PARAMETER[\"Central_Meridian\",6.0],PARAMETER[\"Standard_Parallel_1\",55.0],PARAMETER[\"Scale_Factor\",0.99950908],PARAMETER[\"Latitude_Of_Origin\",55.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27500]]",
        27561=>"PROJCS[\"NTF_Paris_Lambert_Nord_France\",GEOGCS[\"GCS_NTF_Paris\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",55.0],PARAMETER[\"Scale_Factor\",0.999877341],PARAMETER[\"Latitude_Of_Origin\",55.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27561]]",
        27562=>"PROJCS[\"NTF_Paris_Lambert_Centre_France\",GEOGCS[\"GCS_NTF_Paris\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",52.0],PARAMETER[\"Scale_Factor\",0.99987742],PARAMETER[\"Latitude_Of_Origin\",52.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27562]]",
        27563=>"PROJCS[\"NTF_Paris_Lambert_Sud_France\",GEOGCS[\"GCS_NTF_Paris\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",49.0],PARAMETER[\"Scale_Factor\",0.999877499],PARAMETER[\"Latitude_Of_Origin\",49.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27563]]",
        27564=>"PROJCS[\"NTF_Paris_Lambert_Corse\",GEOGCS[\"GCS_NTF_Paris\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",234.358],PARAMETER[\"False_Northing\",185861.369],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",46.85],PARAMETER[\"Scale_Factor\",0.99994471],PARAMETER[\"Latitude_Of_Origin\",46.85],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27564]]",
        27571=>"PROJCS[\"NTF_Paris_Lambert_Zone_I\",GEOGCS[\"GCS_NTF_Paris\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",1200000.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",55.0],PARAMETER[\"Scale_Factor\",0.999877341],PARAMETER[\"Latitude_Of_Origin\",55.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27571]]",
        27572=>"PROJCS[\"NTF_Paris_Lambert_Zone_II\",GEOGCS[\"GCS_NTF_Paris\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",2200000.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",52.0],PARAMETER[\"Scale_Factor\",0.99987742],PARAMETER[\"Latitude_Of_Origin\",52.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27572]]",
        27573=>"PROJCS[\"NTF_Paris_Lambert_Zone_III\",GEOGCS[\"GCS_NTF_Paris\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",3200000.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",49.0],PARAMETER[\"Scale_Factor\",0.999877499],PARAMETER[\"Latitude_Of_Origin\",49.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27573]]",
        27574=>"PROJCS[\"NTF_Paris_Lambert_Zone_IV\",GEOGCS[\"GCS_NTF_Paris\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",234.358],PARAMETER[\"False_Northing\",4185861.369],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",46.85],PARAMETER[\"Scale_Factor\",0.99994471],PARAMETER[\"Latitude_Of_Origin\",46.85],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27574]]",
        27581=>"PROJCS[\"NTF_Paris_France_I\",GEOGCS[\"GCS_NTF_Paris\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",1200000.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",55.0],PARAMETER[\"Scale_Factor\",0.999877341],PARAMETER[\"Latitude_Of_Origin\",55.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27581]]",
        27582=>"PROJCS[\"NTF_Paris_France_II\",GEOGCS[\"GCS_NTF_Paris\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",2200000.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",52.0],PARAMETER[\"Scale_Factor\",0.99987742],PARAMETER[\"Latitude_Of_Origin\",52.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27582]]",
        27583=>"PROJCS[\"NTF_Paris_France_III\",GEOGCS[\"GCS_NTF_Paris\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",3200000.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",49.0],PARAMETER[\"Scale_Factor\",0.999877499],PARAMETER[\"Latitude_Of_Origin\",49.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27583]]",
        27584=>"PROJCS[\"NTF_Paris_France_IV\",GEOGCS[\"GCS_NTF_Paris\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",234.358],PARAMETER[\"False_Northing\",4185861.369],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",46.85],PARAMETER[\"Scale_Factor\",0.99994471],PARAMETER[\"Latitude_Of_Origin\",46.85],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27584]]",
        27591=>"PROJCS[\"NTF_Paris_Nord_France\",GEOGCS[\"GCS_NTF_Paris\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",55.0],PARAMETER[\"Scale_Factor\",0.999877341],PARAMETER[\"Latitude_Of_Origin\",55.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27591]]",
        27592=>"PROJCS[\"NTF_Paris_Centre_France\",GEOGCS[\"GCS_NTF_Paris\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",52.0],PARAMETER[\"Scale_Factor\",0.99987742],PARAMETER[\"Latitude_Of_Origin\",52.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27592]]",
        27593=>"PROJCS[\"NTF_Paris_Sud_France\",GEOGCS[\"GCS_NTF_Paris\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",49.0],PARAMETER[\"Scale_Factor\",0.999877499],PARAMETER[\"Latitude_Of_Origin\",49.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27593]]",
        27594=>"PROJCS[\"NTF_Paris_Corse\",GEOGCS[\"GCS_NTF_Paris\",DATUM[\"D_NTF\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",234.358],PARAMETER[\"False_Northing\",185861.369],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Standard_Parallel_1\",46.85],PARAMETER[\"Scale_Factor\",0.99994471],PARAMETER[\"Latitude_Of_Origin\",46.85],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27594]]",
        27700=>"PROJCS[\"British_National_Grid\",GEOGCS[\"GCS_OSGB_1936\",DATUM[\"D_OSGB_1936\",SPHEROID[\"Airy_1830\",6377563.396,299.3249646]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",-100000.0],PARAMETER[\"Central_Meridian\",-2.0],PARAMETER[\"Scale_Factor\",0.9996012717],PARAMETER[\"Latitude_Of_Origin\",49.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",27700]]",
        28191=>"PROJCS[\"Palestine_1923_Palestine_Grid\",GEOGCS[\"GCS_Palestine_1923\",DATUM[\"D_Palestine_1923\",SPHEROID[\"Clarke_1880_Benoit\",6378300.789,293.4663155389802]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",170251.555],PARAMETER[\"False_Northing\",126867.909],PARAMETER[\"Central_Meridian\",35.21208055555556],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",31.73409694444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28191]]",
        28192=>"PROJCS[\"Palestine_1923_Palestine_Belt\",GEOGCS[\"GCS_Palestine_1923\",DATUM[\"D_Palestine_1923\",SPHEROID[\"Clarke_1880_Benoit\",6378300.789,293.4663155389802]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",170251.555],PARAMETER[\"False_Northing\",1126867.909],PARAMETER[\"Central_Meridian\",35.21208055555556],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",31.73409694444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28192]]",
        28193=>"PROJCS[\"Palestine_1923_Israel_CS_Grid\",GEOGCS[\"GCS_Palestine_1923\",DATUM[\"D_Palestine_1923\",SPHEROID[\"Clarke_1880_Benoit\",6378300.789,293.4663155389802]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",170251.555],PARAMETER[\"False_Northing\",1126867.909],PARAMETER[\"Central_Meridian\",35.21208055555556],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",31.73409694444445],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28193]]",
        28232=>"PROJCS[\"Pointe_Noire_UTM_Zone_32S\",GEOGCS[\"GCS_Pointe_Noire\",DATUM[\"D_Pointe_Noire\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28232]]",
        28348=>"PROJCS[\"GDA_1994_MGA_Zone_48\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28348]]",
        28349=>"PROJCS[\"GDA_1994_MGA_Zone_49\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28349]]",
        28350=>"PROJCS[\"GDA_1994_MGA_Zone_50\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28350]]",
        28351=>"PROJCS[\"GDA_1994_MGA_Zone_51\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28351]]",
        28352=>"PROJCS[\"GDA_1994_MGA_Zone_52\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28352]]",
        28353=>"PROJCS[\"GDA_1994_MGA_Zone_53\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28353]]",
        28354=>"PROJCS[\"GDA_1994_MGA_Zone_54\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28354]]",
        28355=>"PROJCS[\"GDA_1994_MGA_Zone_55\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28355]]",
        28356=>"PROJCS[\"GDA_1994_MGA_Zone_56\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28356]]",
        28357=>"PROJCS[\"GDA_1994_MGA_Zone_57\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28357]]",
        28358=>"PROJCS[\"GDA_1994_MGA_Zone_58\",GEOGCS[\"GCS_GDA_1994\",DATUM[\"D_GDA_1994\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28358]]",
        28402=>"PROJCS[\"Pulkovo_1942_GK_Zone_2\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28402]]",
        28403=>"PROJCS[\"Pulkovo_1942_GK_Zone_3\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28403]]",
        28404=>"PROJCS[\"Pulkovo_1942_GK_Zone_4\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28404]]",
        28405=>"PROJCS[\"Pulkovo_1942_GK_Zone_5\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28405]]",
        28406=>"PROJCS[\"Pulkovo_1942_GK_Zone_6\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",6500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28406]]",
        28407=>"PROJCS[\"Pulkovo_1942_GK_Zone_7\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",7500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28407]]",
        28408=>"PROJCS[\"Pulkovo_1942_GK_Zone_8\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",8500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28408]]",
        28409=>"PROJCS[\"Pulkovo_1942_GK_Zone_9\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",9500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28409]]",
        28410=>"PROJCS[\"Pulkovo_1942_GK_Zone_10\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",10500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28410]]",
        28411=>"PROJCS[\"Pulkovo_1942_GK_Zone_11\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",11500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28411]]",
        28412=>"PROJCS[\"Pulkovo_1942_GK_Zone_12\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",12500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28412]]",
        28413=>"PROJCS[\"Pulkovo_1942_GK_Zone_13\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",13500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28413]]",
        28414=>"PROJCS[\"Pulkovo_1942_GK_Zone_14\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",14500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28414]]",
        28415=>"PROJCS[\"Pulkovo_1942_GK_Zone_15\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",15500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28415]]",
        28416=>"PROJCS[\"Pulkovo_1942_GK_Zone_16\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",16500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28416]]",
        28417=>"PROJCS[\"Pulkovo_1942_GK_Zone_17\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",17500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28417]]",
        28418=>"PROJCS[\"Pulkovo_1942_GK_Zone_18\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",18500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28418]]",
        28419=>"PROJCS[\"Pulkovo_1942_GK_Zone_19\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",19500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28419]]",
        28420=>"PROJCS[\"Pulkovo_1942_GK_Zone_20\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",20500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28420]]",
        28421=>"PROJCS[\"Pulkovo_1942_GK_Zone_21\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",21500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28421]]",
        28422=>"PROJCS[\"Pulkovo_1942_GK_Zone_22\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",22500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28422]]",
        28423=>"PROJCS[\"Pulkovo_1942_GK_Zone_23\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",23500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28423]]",
        28424=>"PROJCS[\"Pulkovo_1942_GK_Zone_24\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",24500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28424]]",
        28425=>"PROJCS[\"Pulkovo_1942_GK_Zone_25\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",25500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28425]]",
        28426=>"PROJCS[\"Pulkovo_1942_GK_Zone_26\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",26500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28426]]",
        28427=>"PROJCS[\"Pulkovo_1942_GK_Zone_27\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",27500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28427]]",
        28428=>"PROJCS[\"Pulkovo_1942_GK_Zone_28\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",28500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28428]]",
        28429=>"PROJCS[\"Pulkovo_1942_GK_Zone_29\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",29500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28429]]",
        28430=>"PROJCS[\"Pulkovo_1942_GK_Zone_30\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",30500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28430]]",
        28431=>"PROJCS[\"Pulkovo_1942_GK_Zone_31\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",31500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28431]]",
        28432=>"PROJCS[\"Pulkovo_1942_GK_Zone_32\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",32500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28432]]",
        28462=>"PROJCS[\"Pulkovo_1942_GK_Zone_2N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28462]]",
        28463=>"PROJCS[\"Pulkovo_1942_GK_Zone_3N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28463]]",
        28464=>"PROJCS[\"Pulkovo_1942_GK_Zone_4N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28464]]",
        28465=>"PROJCS[\"Pulkovo_1942_GK_Zone_5N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28465]]",
        28466=>"PROJCS[\"Pulkovo_1942_GK_Zone_6N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28466]]",
        28467=>"PROJCS[\"Pulkovo_1942_GK_Zone_7N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28467]]",
        28468=>"PROJCS[\"Pulkovo_1942_GK_Zone_8N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28468]]",
        28469=>"PROJCS[\"Pulkovo_1942_GK_Zone_9N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28469]]",
        28470=>"PROJCS[\"Pulkovo_1942_GK_Zone_10N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28470]]",
        28471=>"PROJCS[\"Pulkovo_1942_GK_Zone_11N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28471]]",
        28472=>"PROJCS[\"Pulkovo_1942_GK_Zone_12N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28472]]",
        28473=>"PROJCS[\"Pulkovo_1942_GK_Zone_13N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28473]]",
        28474=>"PROJCS[\"Pulkovo_1942_GK_Zone_14N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28474]]",
        28475=>"PROJCS[\"Pulkovo_1942_GK_Zone_15N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28475]]",
        28476=>"PROJCS[\"Pulkovo_1942_GK_Zone_16N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28476]]",
        28477=>"PROJCS[\"Pulkovo_1942_GK_Zone_17N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28477]]",
        28478=>"PROJCS[\"Pulkovo_1942_GK_Zone_18N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28478]]",
        28479=>"PROJCS[\"Pulkovo_1942_GK_Zone_19N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28479]]",
        28480=>"PROJCS[\"Pulkovo_1942_GK_Zone_20N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28480]]",
        28481=>"PROJCS[\"Pulkovo_1942_GK_Zone_21N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28481]]",
        28482=>"PROJCS[\"Pulkovo_1942_GK_Zone_22N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28482]]",
        28483=>"PROJCS[\"Pulkovo_1942_GK_Zone_23N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28483]]",
        28484=>"PROJCS[\"Pulkovo_1942_GK_Zone_24N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28484]]",
        28485=>"PROJCS[\"Pulkovo_1942_GK_Zone_25N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28485]]",
        28486=>"PROJCS[\"Pulkovo_1942_GK_Zone_26N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28486]]",
        28487=>"PROJCS[\"Pulkovo_1942_GK_Zone_27N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28487]]",
        28488=>"PROJCS[\"Pulkovo_1942_GK_Zone_28N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28488]]",
        28489=>"PROJCS[\"Pulkovo_1942_GK_Zone_29N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28489]]",
        28490=>"PROJCS[\"Pulkovo_1942_GK_Zone_30N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28490]]",
        28491=>"PROJCS[\"Pulkovo_1942_GK_Zone_31N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28491]]",
        28492=>"PROJCS[\"Pulkovo_1942_GK_Zone_32N\",GEOGCS[\"GCS_Pulkovo_1942\",DATUM[\"D_Pulkovo_1942\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28492]]",
        28600=>"PROJCS[\"Qatar_National_Grid\",GEOGCS[\"GCS_Qatar_1974\",DATUM[\"D_Qatar\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",300000.0],PARAMETER[\"Central_Meridian\",51.21666666666667],PARAMETER[\"Scale_Factor\",0.99999],PARAMETER[\"Latitude_Of_Origin\",24.45],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28600]]",
        28991=>"PROJCS[\"RD_Old\",GEOGCS[\"GCS_Amersfoort\",DATUM[\"D_Amersfoort\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",5.38763888888889],PARAMETER[\"Scale_Factor\",0.9999079],PARAMETER[\"Latitude_Of_Origin\",52.15616055555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28991]]",
        28992=>"PROJCS[\"RD_New\",GEOGCS[\"GCS_Amersfoort\",DATUM[\"D_Amersfoort\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",155000.0],PARAMETER[\"False_Northing\",463000.0],PARAMETER[\"Central_Meridian\",5.38763888888889],PARAMETER[\"Scale_Factor\",0.9999079],PARAMETER[\"Latitude_Of_Origin\",52.15616055555555],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",28992]]",
        29100=>"PROJCS[\"SAD_1969_Brazil_Polyconic\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Polyconic\"],PARAMETER[\"False_Easting\",5000000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-54.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29101]]",
        29101=>"PROJCS[\"SAD_1969_Brazil_Polyconic\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Polyconic\"],PARAMETER[\"False_Easting\",5000000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-54.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29101]]",
        29118=>"PROJCS[\"SAD_1969_UTM_Zone_18N\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29168]]",
        29119=>"PROJCS[\"SAD_1969_UTM_Zone_19N\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29169]]",
        29120=>"PROJCS[\"SAD_1969_UTM_Zone_20N\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29170]]",
        29121=>"PROJCS[\"SAD_1969_UTM_Zone_21N\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29171]]",
        29122=>"PROJCS[\"SAD_1969_UTM_Zone_22N\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29172]]",
        29168=>"PROJCS[\"SAD_1969_UTM_Zone_18N\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29168]]",
        29169=>"PROJCS[\"SAD_1969_UTM_Zone_19N\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29169]]",
        29170=>"PROJCS[\"SAD_1969_UTM_Zone_20N\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29170]]",
        29171=>"PROJCS[\"SAD_1969_UTM_Zone_21N\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29171]]",
        29172=>"PROJCS[\"SAD_1969_UTM_Zone_22N\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29172]]",
        29177=>"PROJCS[\"SAD_1969_UTM_Zone_17S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29187]]",
        29178=>"PROJCS[\"SAD_1969_UTM_Zone_18S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29188]]",
        29179=>"PROJCS[\"SAD_1969_UTM_Zone_19S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29189]]",
        29180=>"PROJCS[\"SAD_1969_UTM_Zone_20S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29190]]",
        29181=>"PROJCS[\"SAD_1969_UTM_Zone_21S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29191]]",
        29182=>"PROJCS[\"SAD_1969_UTM_Zone_22S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29192]]",
        29183=>"PROJCS[\"SAD_1969_UTM_Zone_23S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29193]]",
        29184=>"PROJCS[\"SAD_1969_UTM_Zone_24S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29194]]",
        29185=>"PROJCS[\"SAD_1969_UTM_Zone_25S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29195]]",
        29187=>"PROJCS[\"SAD_1969_UTM_Zone_17S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29187]]",
        29188=>"PROJCS[\"SAD_1969_UTM_Zone_18S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29188]]",
        29189=>"PROJCS[\"SAD_1969_UTM_Zone_19S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29189]]",
        29190=>"PROJCS[\"SAD_1969_UTM_Zone_20S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29190]]",
        29191=>"PROJCS[\"SAD_1969_UTM_Zone_21S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29191]]",
        29192=>"PROJCS[\"SAD_1969_UTM_Zone_22S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29192]]",
        29193=>"PROJCS[\"SAD_1969_UTM_Zone_23S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29193]]",
        29194=>"PROJCS[\"SAD_1969_UTM_Zone_24S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29194]]",
        29195=>"PROJCS[\"SAD_1969_UTM_Zone_25S\",GEOGCS[\"GCS_South_American_1969\",DATUM[\"D_South_American_1969\",SPHEROID[\"GRS_1967_Truncated\",6378160.0,298.25]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29195]]",
        29220=>"PROJCS[\"Sapper_Hill_1943_UTM_Zone_20S\",GEOGCS[\"GCS_Sapper_Hill_1943\",DATUM[\"D_Sapper_Hill_1943\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29220]]",
        29221=>"PROJCS[\"Sapper_Hill_1943_UTM_Zone_21S\",GEOGCS[\"GCS_Sapper_Hill_1943\",DATUM[\"D_Sapper_Hill_1943\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29221]]",
        29333=>"PROJCS[\"Schwarzeck_UTM_Zone_33S\",GEOGCS[\"GCS_Schwarzeck\",DATUM[\"D_Schwarzeck\",SPHEROID[\"Bessel_Namibia\",6377483.865280418,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29333]]",
        29635=>"PROJCS[\"Sudan_UTM_Zone_35N\",GEOGCS[\"GCS_Sudan\",DATUM[\"D_Sudan\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29635]]",
        29636=>"PROJCS[\"Sudan_UTM_Zone_36N\",GEOGCS[\"GCS_Sudan\",DATUM[\"D_Sudan\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29636]]",
        29701=>"PROJCS[\"Tananarive_1925_Paris_Laborde_Grid\",GEOGCS[\"GCS_Tananarive_1925_Paris\",DATUM[\"D_Tananarive_1925\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Paris\",2.337229166666667],UNIT[\"Grad\",0.01570796326794897]],PROJECTION[\"Laborde_Oblique_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",800000.0],PARAMETER[\"Scale_Factor\",0.9995],PARAMETER[\"Azimuth\",21.0],PARAMETER[\"Longitude_Of_Center\",49.0],PARAMETER[\"Latitude_Of_Center\",-21.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29701]]",
        29738=>"PROJCS[\"Tananarive_1925_UTM_Zone_38S\",GEOGCS[\"GCS_Tananarive_1925\",DATUM[\"D_Tananarive_1925\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29738]]",
        29739=>"PROJCS[\"Tananarive_1925_UTM_Zone_39S\",GEOGCS[\"GCS_Tananarive_1925\",DATUM[\"D_Tananarive_1925\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29739]]",
        29849=>"PROJCS[\"Timbalai_1948_UTM_Zone_49N\",GEOGCS[\"GCS_Timbalai_1948\",DATUM[\"D_Timbalai_1948\",SPHEROID[\"Everest_Definition_1967\",6377298.556,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29849]]",
        29850=>"PROJCS[\"Timbalai_1948_UTM_Zone_50N\",GEOGCS[\"GCS_Timbalai_1948\",DATUM[\"D_Timbalai_1948\",SPHEROID[\"Everest_Definition_1967\",6377298.556,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29850]]",
        29871=>"PROJCS[\"Timbalai_1948_RSO_Borneo_Chains\",GEOGCS[\"GCS_Timbalai_1948\",DATUM[\"D_Timbalai_1948\",SPHEROID[\"Everest_Definition_1967\",6377298.556,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Rectified_Skew_Orthomorphic_Natural_Origin\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Scale_Factor\",0.99984],PARAMETER[\"Azimuth\",53.31582047908623],PARAMETER[\"Longitude_Of_Center\",115.0],PARAMETER[\"Latitude_Of_Center\",4.0],PARAMETER[\"XY_Plane_Rotation\",53.13010235415598],UNIT[\"Chain_Sears\",20.11676512155263],AUTHORITY[\"EPSG\",29871]]",
        29872=>"PROJCS[\"Timbalai_1948_RSO_Borneo_Feet\",GEOGCS[\"GCS_Timbalai_1948\",DATUM[\"D_Timbalai_1948\",SPHEROID[\"Everest_Definition_1967\",6377298.556,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Rectified_Skew_Orthomorphic_Natural_Origin\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Scale_Factor\",0.99984],PARAMETER[\"Azimuth\",53.31582047908623],PARAMETER[\"Longitude_Of_Center\",115.0],PARAMETER[\"Latitude_Of_Center\",4.0],PARAMETER[\"XY_Plane_Rotation\",53.13010235415598],UNIT[\"Foot_Sears\",0.3047994715386762],AUTHORITY[\"EPSG\",29872]]",
        29873=>"PROJCS[\"Timbalai_1948_RSO_Borneo_Meters\",GEOGCS[\"GCS_Timbalai_1948\",DATUM[\"D_Timbalai_1948\",SPHEROID[\"Everest_Definition_1967\",6377298.556,300.8017]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Rectified_Skew_Orthomorphic_Natural_Origin\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Scale_Factor\",0.99984],PARAMETER[\"Azimuth\",53.31582047908623],PARAMETER[\"Longitude_Of_Center\",115.0],PARAMETER[\"Latitude_Of_Center\",4.0],PARAMETER[\"XY_Plane_Rotation\",53.13010235415598],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29873]]",
        29900=>"PROJCS[\"TM65_Irish_Grid\",GEOGCS[\"GCS_TM65\",DATUM[\"D_TM65\",SPHEROID[\"Airy_Modified\",6377340.189,299.3249646]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",250000.0],PARAMETER[\"Central_Meridian\",-8.0],PARAMETER[\"Scale_Factor\",1.000035],PARAMETER[\"Latitude_Of_Origin\",53.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29902]]",
        29901=>"PROJCS[\"OSNI_1952_Irish_National_Grid\",GEOGCS[\"GCS_OSNI_1952\",DATUM[\"D_OSNI_1952\",SPHEROID[\"Airy_1830\",6377563.396,299.3249646]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",250000.0],PARAMETER[\"Central_Meridian\",-8.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",53.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29901]]",
        29902=>"PROJCS[\"TM65_Irish_Grid\",GEOGCS[\"GCS_TM65\",DATUM[\"D_TM65\",SPHEROID[\"Airy_Modified\",6377340.189,299.3249646]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",250000.0],PARAMETER[\"Central_Meridian\",-8.0],PARAMETER[\"Scale_Factor\",1.000035],PARAMETER[\"Latitude_Of_Origin\",53.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29902]]",
        29903=>"PROJCS[\"TM75_Irish_Grid\",GEOGCS[\"GCS_TM75\",DATUM[\"D_TM75\",SPHEROID[\"Airy_Modified\",6377340.189,299.3249646]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",250000.0],PARAMETER[\"Central_Meridian\",-8.0],PARAMETER[\"Scale_Factor\",1.000035],PARAMETER[\"Latitude_Of_Origin\",53.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",29903]]",
        30161=>"PROJCS[\"Japan_Zone_1\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",33.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30161]]",
        30162=>"PROJCS[\"Japan_Zone_2\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",131.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",33.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30162]]",
        30163=>"PROJCS[\"Japan_Zone_3\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",132.1666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30163]]",
        30164=>"PROJCS[\"Japan_Zone_4\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",133.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",33.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30164]]",
        30165=>"PROJCS[\"Japan_Zone_5\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",134.3333333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30165]]",
        30166=>"PROJCS[\"Japan_Zone_6\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",136.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30166]]",
        30167=>"PROJCS[\"Japan_Zone_7\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",137.1666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30167]]",
        30168=>"PROJCS[\"Japan_Zone_8\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",138.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30168]]",
        30169=>"PROJCS[\"Japan_Zone_9\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",139.8333333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30169]]",
        30170=>"PROJCS[\"Japan_Zone_10\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",140.8333333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30170]]",
        30171=>"PROJCS[\"Japan_Zone_11\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",140.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30171]]",
        30172=>"PROJCS[\"Japan_Zone_12\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",142.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30172]]",
        30173=>"PROJCS[\"Japan_Zone_13\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",144.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30173]]",
        30174=>"PROJCS[\"Japan_Zone_14\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",142.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30174]]",
        30175=>"PROJCS[\"Japan_Zone_15\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",127.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30175]]",
        30176=>"PROJCS[\"Japan_Zone_16\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",124.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30176]]",
        30177=>"PROJCS[\"Japan_Zone_17\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",131.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30177]]",
        30178=>"PROJCS[\"Japan_Zone_18\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",136.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",20.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30178]]",
        30179=>"PROJCS[\"Japan_Zone_19\",GEOGCS[\"GCS_Tokyo\",DATUM[\"D_Tokyo\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",154.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",26.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30179]]",
        30200=>"PROJCS[\"Trinidad_1903_Trinidad_Grid\",GEOGCS[\"GCS_Trinidad_1903\",DATUM[\"D_Trinidad_1903\",SPHEROID[\"Clarke_1858\",6378293.645208759,294.260676369]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Cassini\"],PARAMETER[\"False_Easting\",430000.0],PARAMETER[\"False_Northing\",325000.0],PARAMETER[\"Central_Meridian\",-61.33333333333334],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",10.44166666666667],UNIT[\"Link_Clarke\",0.201166195164],AUTHORITY[\"EPSG\",30200]]",
        30339=>"PROJCS[\"TC_1948_UTM_Zone_39N\",GEOGCS[\"GCS_Trucial_Coast_1948\",DATUM[\"D_Trucial_Coast_1948\",SPHEROID[\"Helmert_1906\",6378200.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30339]]",
        30340=>"PROJCS[\"TC_1948_UTM_Zone_40N\",GEOGCS[\"GCS_Trucial_Coast_1948\",DATUM[\"D_Trucial_Coast_1948\",SPHEROID[\"Helmert_1906\",6378200.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30340]]",
        30491=>"PROJCS[\"Nord_Algerie_Ancienne\",GEOGCS[\"GCS_Voirol_1875\",DATUM[\"D_Voirol_1875\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",300000.0],PARAMETER[\"Central_Meridian\",2.7],PARAMETER[\"Standard_Parallel_1\",36.0],PARAMETER[\"Scale_Factor\",0.999625544],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30491]]",
        30492=>"PROJCS[\"Sud_Algerie_Ancienne\",GEOGCS[\"GCS_Voirol_1875\",DATUM[\"D_Voirol_1875\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",300000.0],PARAMETER[\"Central_Meridian\",2.7],PARAMETER[\"Standard_Parallel_1\",33.3],PARAMETER[\"Scale_Factor\",0.999625769],PARAMETER[\"Latitude_Of_Origin\",33.3],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30492]]",
        30493=>"PROJCS[\"Voirol_1879_Nord_Algerie_Ancienne\",GEOGCS[\"GCS_Voirol_1879\",DATUM[\"D_Voirol_1879\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",300000.0],PARAMETER[\"Central_Meridian\",2.7],PARAMETER[\"Standard_Parallel_1\",36.0],PARAMETER[\"Scale_Factor\",0.999625544],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30493]]",
        30494=>"PROJCS[\"Voirol_1879_Sud_Algerie_Ancienne\",GEOGCS[\"GCS_Voirol_1879\",DATUM[\"D_Voirol_1879\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",300000.0],PARAMETER[\"Central_Meridian\",2.7],PARAMETER[\"Standard_Parallel_1\",33.3],PARAMETER[\"Scale_Factor\",0.999625769],PARAMETER[\"Latitude_Of_Origin\",33.3],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30494]]",
        30729=>"PROJCS[\"Nord_Sahara_1959_UTM_Zone_29N\",GEOGCS[\"GCS_Nord_Sahara_1959\",DATUM[\"D_Nord_Sahara_1959\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30729]]",
        30730=>"PROJCS[\"Nord_Sahara_1959_UTM_Zone_30N\",GEOGCS[\"GCS_Nord_Sahara_1959\",DATUM[\"D_Nord_Sahara_1959\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30730]]",
        30731=>"PROJCS[\"Nord_Sahara_1959_UTM_Zone_31N\",GEOGCS[\"GCS_Nord_Sahara_1959\",DATUM[\"D_Nord_Sahara_1959\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30731]]",
        30732=>"PROJCS[\"Nord_Sahara_1959_UTM_Zone_32N\",GEOGCS[\"GCS_Nord_Sahara_1959\",DATUM[\"D_Nord_Sahara_1959\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30732]]",
        30791=>"PROJCS[\"Nord_Sahara_1959_Voirol_Unifie_Nord\",GEOGCS[\"GCS_Nord_Sahara_1959\",DATUM[\"D_Nord_Sahara_1959\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500135.0],PARAMETER[\"False_Northing\",300090.0],PARAMETER[\"Central_Meridian\",2.7],PARAMETER[\"Standard_Parallel_1\",36.0],PARAMETER[\"Scale_Factor\",0.999625544],PARAMETER[\"Latitude_Of_Origin\",36.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30791]]",
        30792=>"PROJCS[\"Nord_Sahara_1959_Voirol_Unifie_Sud\",GEOGCS[\"GCS_Nord_Sahara_1959\",DATUM[\"D_Nord_Sahara_1959\",SPHEROID[\"Clarke_1880_RGS\",6378249.145,293.465]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500135.0],PARAMETER[\"False_Northing\",300090.0],PARAMETER[\"Central_Meridian\",2.7],PARAMETER[\"Standard_Parallel_1\",33.3],PARAMETER[\"Scale_Factor\",0.999625769],PARAMETER[\"Latitude_Of_Origin\",33.3],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30792]]",
        30800=>"PROJCS[\"Swedish_National_Grid\",GEOGCS[\"GCS_RT38_Stockholm\",DATUM[\"D_Stockholm_1938\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Stockholm\",18.05827777777778],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-2.25],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",30800]]",
        31028=>"PROJCS[\"Yoff_1972_UTM_Zone_28N\",GEOGCS[\"GCS_Yoff\",DATUM[\"D_Yoff\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31028]]",
        31121=>"PROJCS[\"Zanderij_1972_UTM_Zone_21N\",GEOGCS[\"GCS_Zanderij\",DATUM[\"D_Zanderij\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31121]]",
        31154=>"PROJCS[\"Zanderij_TM_54_NW\",GEOGCS[\"GCS_Zanderij\",DATUM[\"D_Zanderij\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-54.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31154]]",
        31170=>"PROJCS[\"Zanderij_Suriname_Old_TM\",GEOGCS[\"GCS_Zanderij\",DATUM[\"D_Zanderij\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-55.68333333333333],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31170]]",
        31171=>"PROJCS[\"Zanderij_Suriname_TM\",GEOGCS[\"GCS_Zanderij\",DATUM[\"D_Zanderij\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-55.68333333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31171]]",
        31251=>"PROJCS[\"MGI_Ferro_Austria_GK_West\",GEOGCS[\"GCS_MGI_Ferro\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Central_Meridian\",28.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31251]]",
        31252=>"PROJCS[\"MGI_Ferro_Austria_GK_Central\",GEOGCS[\"GCS_MGI_Ferro\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Central_Meridian\",31.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31252]]",
        31253=>"PROJCS[\"MGI_Ferro_Austria_GK_East\",GEOGCS[\"GCS_MGI_Ferro\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Central_Meridian\",34.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31253]]",
        31254=>"PROJCS[\"MGI_Austria_GK_West\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Central_Meridian\",10.33333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31254]]",
        31255=>"PROJCS[\"MGI_Austria_GK_Central\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Central_Meridian\",13.33333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31255]]",
        31256=>"PROJCS[\"MGI_Austria_GK_East\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Central_Meridian\",16.33333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31256]]",
        31257=>"PROJCS[\"MGI_Austria_GK_M28\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Central_Meridian\",10.33333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31257]]",
        31258=>"PROJCS[\"MGI_Austria_GK_M31\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",450000.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Central_Meridian\",13.33333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31258]]",
        31259=>"PROJCS[\"MGI_Austria_GK_M34\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",750000.0],PARAMETER[\"False_Northing\",-5000000.0],PARAMETER[\"Central_Meridian\",16.33333333333334],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31259]]",
        31265=>"PROJCS[\"MGI_3_Degree_Gauss_Zone_5\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31265]]",
        31266=>"PROJCS[\"MGI_3_Degree_Gauss_Zone_6\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",6500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",18.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31266]]",
        31267=>"PROJCS[\"MGI_3_Degree_Gauss_Zone_7\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",7500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31267]]",
        31268=>"PROJCS[\"MGI_3_Degree_Gauss_Zone_8\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",8500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31268]]",
        31275=>"PROJCS[\"MGI_Balkans_5\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31275]]",
        31276=>"PROJCS[\"MGI_Balkans_6\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",6500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",18.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31276]]",
        31277=>"PROJCS[\"MGI_Balkans_7\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",7500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31277]]",
        31278=>"PROJCS[\"MGI_Balkans_8\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",8500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31279]]",
        31279=>"PROJCS[\"MGI_Balkans_8\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",8500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",24.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31279]]",
        31281=>"PROJCS[\"Austria_West_Zone\",GEOGCS[\"GCS_MGI_Ferro\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",28.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31281]]",
        31282=>"PROJCS[\"Austria_Central_Zone\",GEOGCS[\"GCS_MGI_Ferro\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",31.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31282]]",
        31283=>"PROJCS[\"Austria_East_Zone\",GEOGCS[\"GCS_MGI_Ferro\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",34.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31283]]",
        31284=>"PROJCS[\"MGI_M28\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",10.33333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31284]]",
        31285=>"PROJCS[\"MGI_M31\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",450000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",13.33333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31285]]",
        31286=>"PROJCS[\"MGI_M34\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",750000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",16.33333333333334],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31286]]",
        31287=>"PROJCS[\"MGI_Austria_Lambert\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",400000.0],PARAMETER[\"Central_Meridian\",13.33333333333333],PARAMETER[\"Standard_Parallel_1\",46.0],PARAMETER[\"Standard_Parallel_2\",49.0],PARAMETER[\"Latitude_Of_Origin\",47.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31287]]",
        31288=>"PROJCS[\"MGI_Ferro_M28\",GEOGCS[\"GCS_MGI_Ferro\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",28.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31288]]",
        31289=>"PROJCS[\"MGI_Ferro_M31\",GEOGCS[\"GCS_MGI_Ferro\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",450000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",31.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31289]]",
        31290=>"PROJCS[\"MGI_Ferro_M34\",GEOGCS[\"GCS_MGI_Ferro\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",750000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",34.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31290]]",
        31291=>"PROJCS[\"Austria_West_Zone\",GEOGCS[\"GCS_MGI_Ferro\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",28.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31281]]",
        31292=>"PROJCS[\"Austria_Central_Zone\",GEOGCS[\"GCS_MGI_Ferro\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",31.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31282]]",
        31293=>"PROJCS[\"Austria_East_Zone\",GEOGCS[\"GCS_MGI_Ferro\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Ferro\",-17.66666666666667],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",34.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31283]]",
        31294=>"PROJCS[\"MGI_M28\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",10.33333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31284]]",
        31295=>"PROJCS[\"MGI_M31\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",450000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",13.33333333333333],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31285]]",
        31296=>"PROJCS[\"MGI_M34\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",750000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",16.33333333333334],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31286]]",
        31297=>"PROJCS[\"MGI_Austria_Lambert\",GEOGCS[\"GCS_MGI\",DATUM[\"D_MGI\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",400000.0],PARAMETER[\"Central_Meridian\",13.33333333333333],PARAMETER[\"Standard_Parallel_1\",46.0],PARAMETER[\"Standard_Parallel_2\",49.0],PARAMETER[\"Latitude_Of_Origin\",47.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31287]]",
        31370=>"PROJCS[\"Belge_Lambert_1972\",GEOGCS[\"GCS_Belge_1972\",DATUM[\"D_Belge_1972\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",150000.01256],PARAMETER[\"False_Northing\",5400088.4378],PARAMETER[\"Central_Meridian\",4.367486666666666],PARAMETER[\"Standard_Parallel_1\",49.8333339],PARAMETER[\"Standard_Parallel_2\",51.16666723333333],PARAMETER[\"Latitude_Of_Origin\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31370]]",
        31461=>"PROJCS[\"DHDN_3_Degree_Gauss_Zone_1\",GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",5520]]",
        31462=>"PROJCS[\"DHDN_3_Degree_Gauss_Zone_2\",GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",6.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31466]]",
        31463=>"PROJCS[\"DHDN_3_Degree_Gauss_Zone_3\",GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31467]]",
        31464=>"PROJCS[\"DHDN_3_Degree_Gauss_Zone_4\",GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31468]]",
        31465=>"PROJCS[\"DHDN_3_Degree_Gauss_Zone_5\",GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31469]]",
        31466=>"PROJCS[\"DHDN_3_Degree_Gauss_Zone_2\",GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",6.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31466]]",
        31467=>"PROJCS[\"DHDN_3_Degree_Gauss_Zone_3\",GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31467]]",
        31468=>"PROJCS[\"DHDN_3_Degree_Gauss_Zone_4\",GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",4500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",12.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31468]]",
        31469=>"PROJCS[\"DHDN_3_Degree_Gauss_Zone_5\",GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Gauss_Kruger\"],PARAMETER[\"False_Easting\",5500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31469]]",
        31528=>"PROJCS[\"Conakry_1905_UTM_Zone_28N\",GEOGCS[\"GCS_Conakry_1905\",DATUM[\"D_Conakry_1905\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31528]]",
        31529=>"PROJCS[\"Conakry_1905_UTM_Zone_29N\",GEOGCS[\"GCS_Conakry_1905\",DATUM[\"D_Conakry_1905\",SPHEROID[\"Clarke_1880_IGN\",6378249.2,293.4660212936265]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31529]]",
        31600=>"PROJCS[\"Stereo_33\",GEOGCS[\"GCS_Dealul_Piscului_1933\",DATUM[\"D_Dealul_Piscului_1933\",SPHEROID[\"International_1924\",6378388.0,297.0]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",25.39246588888889],PARAMETER[\"Scale_Factor\",0.9996667],PARAMETER[\"Latitude_Of_Origin\",45.9],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31600]]",
        31700=>"PROJCS[\"Stereo_70\",GEOGCS[\"GCS_Dealul_Piscului_1970\",DATUM[\"D_Dealul_Piscului_1970\",SPHEROID[\"Krasovsky_1940\",6378245.0,298.3]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Double_Stereographic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",500000.0],PARAMETER[\"Central_Meridian\",25.0],PARAMETER[\"Scale_Factor\",0.99975],PARAMETER[\"Latitude_Of_Origin\",46.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31700]]",
        31838=>"PROJCS[\"NGN_UTM_Zone_38N\",GEOGCS[\"GCS_NGN\",DATUM[\"D_NGN\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31838]]",
        31839=>"PROJCS[\"NGN_UTM_Zone_39N\",GEOGCS[\"GCS_NGN\",DATUM[\"D_NGN\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31839]]",
        31901=>"PROJCS[\"KUDAMS_KTM\",GEOGCS[\"GCS_KUDAMS\",DATUM[\"D_Kuwait_Utility\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",48.0],PARAMETER[\"Scale_Factor\",1.0],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31901]]",
        31965=>"PROJCS[\"SIRGAS_2000_UTM_Zone_11N\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31965]]",
        31966=>"PROJCS[\"SIRGAS_2000_UTM_Zone_12N\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31966]]",
        31967=>"PROJCS[\"SIRGAS_2000_UTM_Zone_13N\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31967]]",
        31968=>"PROJCS[\"SIRGAS_2000_UTM_Zone_14N\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31968]]",
        31969=>"PROJCS[\"SIRGAS_2000_UTM_Zone_15N\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31969]]",
        31970=>"PROJCS[\"SIRGAS_2000_UTM_Zone_16N\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31970]]",
        31971=>"PROJCS[\"SIRGAS_2000_UTM_Zone_17N\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31971]]",
        31972=>"PROJCS[\"SIRGAS_2000_UTM_Zone_18N\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31972]]",
        31973=>"PROJCS[\"SIRGAS_2000_UTM_Zone_19N\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31973]]",
        31974=>"PROJCS[\"SIRGAS_2000_UTM_Zone_20N\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31974]]",
        31975=>"PROJCS[\"SIRGAS_2000_UTM_Zone_21N\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31975]]",
        31976=>"PROJCS[\"SIRGAS_2000_UTM_Zone_22N\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31976]]",
        31977=>"PROJCS[\"SIRGAS_2000_UTM_Zone_17S\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31977]]",
        31978=>"PROJCS[\"SIRGAS_2000_UTM_Zone_18S\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31978]]",
        31979=>"PROJCS[\"SIRGAS_2000_UTM_Zone_19S\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31979]]",
        31980=>"PROJCS[\"SIRGAS_2000_UTM_Zone_20S\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31980]]",
        31981=>"PROJCS[\"SIRGAS_2000_UTM_Zone_21S\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31981]]",
        31982=>"PROJCS[\"SIRGAS_2000_UTM_Zone_22S\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31982]]",
        31983=>"PROJCS[\"SIRGAS_2000_UTM_Zone_23S\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31983]]",
        31984=>"PROJCS[\"SIRGAS_2000_UTM_Zone_24S\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31984]]",
        31985=>"PROJCS[\"SIRGAS_2000_UTM_Zone_25S\",GEOGCS[\"GCS_SIRGAS_2000\",DATUM[\"D_SIRGAS_2000\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31985]]",
        31986=>"PROJCS[\"SIRGAS_UTM_Zone_17N\",GEOGCS[\"GCS_SIRGAS\",DATUM[\"D_SIRGAS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31986]]",
        31987=>"PROJCS[\"SIRGAS_UTM_Zone_18N\",GEOGCS[\"GCS_SIRGAS\",DATUM[\"D_SIRGAS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31987]]",
        31988=>"PROJCS[\"SIRGAS_UTM_Zone_19N\",GEOGCS[\"GCS_SIRGAS\",DATUM[\"D_SIRGAS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31988]]",
        31989=>"PROJCS[\"SIRGAS_UTM_Zone_20N\",GEOGCS[\"GCS_SIRGAS\",DATUM[\"D_SIRGAS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31989]]",
        31990=>"PROJCS[\"SIRGAS_UTM_Zone_21N\",GEOGCS[\"GCS_SIRGAS\",DATUM[\"D_SIRGAS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31990]]",
        31991=>"PROJCS[\"SIRGAS_UTM_Zone_22N\",GEOGCS[\"GCS_SIRGAS\",DATUM[\"D_SIRGAS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31991]]",
        31992=>"PROJCS[\"SIRGAS_UTM_Zone_17S\",GEOGCS[\"GCS_SIRGAS\",DATUM[\"D_SIRGAS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31992]]",
        31993=>"PROJCS[\"SIRGAS_UTM_Zone_18S\",GEOGCS[\"GCS_SIRGAS\",DATUM[\"D_SIRGAS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31993]]",
        31994=>"PROJCS[\"SIRGAS_UTM_Zone_19S\",GEOGCS[\"GCS_SIRGAS\",DATUM[\"D_SIRGAS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31994]]",
        31995=>"PROJCS[\"SIRGAS_UTM_Zone_20S\",GEOGCS[\"GCS_SIRGAS\",DATUM[\"D_SIRGAS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31995]]",
        31996=>"PROJCS[\"SIRGAS_UTM_Zone_21S\",GEOGCS[\"GCS_SIRGAS\",DATUM[\"D_SIRGAS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31996]]",
        31997=>"PROJCS[\"SIRGAS_UTM_Zone_22S\",GEOGCS[\"GCS_SIRGAS\",DATUM[\"D_SIRGAS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31997]]",
        31998=>"PROJCS[\"SIRGAS_UTM_Zone_23S\",GEOGCS[\"GCS_SIRGAS\",DATUM[\"D_SIRGAS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31998]]",
        31999=>"PROJCS[\"SIRGAS_UTM_Zone_24S\",GEOGCS[\"GCS_SIRGAS\",DATUM[\"D_SIRGAS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",31999]]",
        32000=>"PROJCS[\"SIRGAS_UTM_Zone_25S\",GEOGCS[\"GCS_SIRGAS\",DATUM[\"D_SIRGAS\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32000]]",
        32001=>"PROJCS[\"NAD_1927_StatePlane_Montana_North_FIPS_2501\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-109.5],PARAMETER[\"Standard_Parallel_1\",47.85],PARAMETER[\"Standard_Parallel_2\",48.71666666666667],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32001]]",
        32002=>"PROJCS[\"NAD_1927_StatePlane_Montana_Central_FIPS_2502\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-109.5],PARAMETER[\"Standard_Parallel_1\",46.45],PARAMETER[\"Standard_Parallel_2\",47.88333333333333],PARAMETER[\"Latitude_Of_Origin\",45.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32002]]",
        32003=>"PROJCS[\"NAD_1927_StatePlane_Montana_South_FIPS_2503\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-109.5],PARAMETER[\"Standard_Parallel_1\",44.86666666666667],PARAMETER[\"Standard_Parallel_2\",46.4],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32003]]",
        32005=>"PROJCS[\"NAD_1927_StatePlane_Nebraska_North_FIPS_2601\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",41.85],PARAMETER[\"Standard_Parallel_2\",42.81666666666667],PARAMETER[\"Latitude_Of_Origin\",41.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32005]]",
        32006=>"PROJCS[\"NAD_1927_StatePlane_Nebraska_South_FIPS_2602\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-99.5],PARAMETER[\"Standard_Parallel_1\",40.28333333333333],PARAMETER[\"Standard_Parallel_2\",41.71666666666667],PARAMETER[\"Latitude_Of_Origin\",39.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32006]]",
        32007=>"PROJCS[\"NAD_1927_StatePlane_Nevada_East_FIPS_2701\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-115.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32007]]",
        32008=>"PROJCS[\"NAD_1927_StatePlane_Nevada_Central_FIPS_2702\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-116.6666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32008]]",
        32009=>"PROJCS[\"NAD_1927_StatePlane_Nevada_West_FIPS_2703\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-118.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32009]]",
        32010=>"PROJCS[\"NAD_1927_StatePlane_New_Hampshire_FIPS_2800\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32010]]",
        32011=>"PROJCS[\"NAD_1927_StatePlane_New_Jersey_FIPS_2900\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.66666666666667],PARAMETER[\"Scale_Factor\",0.999975],PARAMETER[\"Latitude_Of_Origin\",38.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32011]]",
        32012=>"PROJCS[\"NAD_1927_StatePlane_New_Mexico_East_FIPS_3001\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-104.3333333333333],PARAMETER[\"Scale_Factor\",0.9999090909090909],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32012]]",
        32013=>"PROJCS[\"NAD_1927_StatePlane_New_Mexico_Central_FIPS_3002\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-106.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32013]]",
        32014=>"PROJCS[\"NAD_1927_StatePlane_New_Mexico_West_FIPS_3003\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-107.8333333333333],PARAMETER[\"Scale_Factor\",0.9999166666666667],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32014]]",
        32015=>"PROJCS[\"NAD_1927_StatePlane_New_York_East_FIPS_3101\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.33333333333333],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32015]]",
        32016=>"PROJCS[\"NAD_1927_StatePlane_New_York_Central_FIPS_3102\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-76.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32016]]",
        32017=>"PROJCS[\"NAD_1927_StatePlane_New_York_West_FIPS_3103\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-78.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32017]]",
        32018=>"PROJCS[\"NAD_1927_StatePlane_New_York_Long_Island_FIPS_3104\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-74.0],PARAMETER[\"Standard_Parallel_1\",40.66666666666666],PARAMETER[\"Standard_Parallel_2\",41.03333333333333],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4456]]",
        32019=>"PROJCS[\"NAD_1927_StatePlane_North_Carolina_FIPS_3200\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.0],PARAMETER[\"Standard_Parallel_1\",34.33333333333334],PARAMETER[\"Standard_Parallel_2\",36.16666666666666],PARAMETER[\"Latitude_Of_Origin\",33.75],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32019]]",
        32020=>"PROJCS[\"NAD_1927_StatePlane_North_Dakota_North_FIPS_3301\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",47.43333333333333],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32020]]",
        32021=>"PROJCS[\"NAD_1927_StatePlane_North_Dakota_South_FIPS_3302\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",46.18333333333333],PARAMETER[\"Standard_Parallel_2\",47.48333333333333],PARAMETER[\"Latitude_Of_Origin\",45.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32021]]",
        32022=>"PROJCS[\"NAD_1927_StatePlane_Ohio_North_FIPS_3401\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",40.43333333333333],PARAMETER[\"Standard_Parallel_2\",41.7],PARAMETER[\"Latitude_Of_Origin\",39.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32022]]",
        32023=>"PROJCS[\"NAD_1927_StatePlane_Ohio_South_FIPS_3402\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",38.73333333333333],PARAMETER[\"Standard_Parallel_2\",40.03333333333333],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32023]]",
        32024=>"PROJCS[\"NAD_1927_StatePlane_Oklahoma_North_FIPS_3501\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",35.56666666666667],PARAMETER[\"Standard_Parallel_2\",36.76666666666667],PARAMETER[\"Latitude_Of_Origin\",35.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32024]]",
        32025=>"PROJCS[\"NAD_1927_StatePlane_Oklahoma_South_FIPS_3502\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",33.93333333333333],PARAMETER[\"Standard_Parallel_2\",35.23333333333333],PARAMETER[\"Latitude_Of_Origin\",33.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32025]]",
        32026=>"PROJCS[\"NAD_1927_StatePlane_Oregon_North_FIPS_3601\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",44.33333333333334],PARAMETER[\"Standard_Parallel_2\",46.0],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32026]]",
        32027=>"PROJCS[\"NAD_1927_StatePlane_Oregon_South_FIPS_3602\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",42.33333333333334],PARAMETER[\"Standard_Parallel_2\",44.0],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32027]]",
        32028=>"PROJCS[\"NAD_1927_StatePlane_Pennsylvania_North_FIPS_3701\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",40.88333333333333],PARAMETER[\"Standard_Parallel_2\",41.95],PARAMETER[\"Latitude_Of_Origin\",40.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32028]]",
        32029=>"PROJCS[\"NAD_1927_StatePlane_Pennsylvania_South_FIPS_3702\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",39.93333333333333],PARAMETER[\"Standard_Parallel_2\",40.96666666666667],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",4455]]",
        32030=>"PROJCS[\"NAD_1927_StatePlane_Rhode_Island_FIPS_3800\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Scale_Factor\",0.99999375],PARAMETER[\"Latitude_Of_Origin\",41.08333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32030]]",
        32031=>"PROJCS[\"NAD_1927_StatePlane_South_Carolina_North_FIPS_3901\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",33.76666666666667],PARAMETER[\"Standard_Parallel_2\",34.96666666666667],PARAMETER[\"Latitude_Of_Origin\",33.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32031]]",
        32033=>"PROJCS[\"NAD_1927_StatePlane_South_Carolina_South_FIPS_3902\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",32.33333333333334],PARAMETER[\"Standard_Parallel_2\",33.66666666666666],PARAMETER[\"Latitude_Of_Origin\",31.83333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32033]]",
        32034=>"PROJCS[\"NAD_1927_StatePlane_South_Dakota_North_FIPS_4001\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",44.41666666666666],PARAMETER[\"Standard_Parallel_2\",45.68333333333333],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32034]]",
        32035=>"PROJCS[\"NAD_1927_StatePlane_South_Dakota_South_FIPS_4002\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",42.83333333333334],PARAMETER[\"Standard_Parallel_2\",44.4],PARAMETER[\"Latitude_Of_Origin\",42.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32035]]",
        32036=>"PROJCS[\"NAD_1927_StatePlane_Tennessee_FIPS_4100\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-86.0],PARAMETER[\"Standard_Parallel_1\",35.25],PARAMETER[\"Standard_Parallel_2\",36.41666666666666],PARAMETER[\"Latitude_Of_Origin\",34.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",2204]]",
        32037=>"PROJCS[\"NAD_1927_StatePlane_Texas_North_FIPS_4201\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-101.5],PARAMETER[\"Standard_Parallel_1\",34.65],PARAMETER[\"Standard_Parallel_2\",36.18333333333333],PARAMETER[\"Latitude_Of_Origin\",34.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32037]]",
        32038=>"PROJCS[\"NAD_1927_StatePlane_Texas_North_Central_FIPS_4202\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-97.5],PARAMETER[\"Standard_Parallel_1\",32.13333333333333],PARAMETER[\"Standard_Parallel_2\",33.96666666666667],PARAMETER[\"Latitude_Of_Origin\",31.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32038]]",
        32039=>"PROJCS[\"NAD_1927_StatePlane_Texas_Central_FIPS_4203\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",30.11666666666667],PARAMETER[\"Standard_Parallel_2\",31.88333333333333],PARAMETER[\"Latitude_Of_Origin\",29.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32039]]",
        32040=>"PROJCS[\"NAD_1927_StatePlane_Texas_South_Central_FIPS_4204\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Standard_Parallel_1\",28.38333333333333],PARAMETER[\"Standard_Parallel_2\",30.28333333333333],PARAMETER[\"Latitude_Of_Origin\",27.83333333333333],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32040]]",
        32041=>"PROJCS[\"NAD_1927_StatePlane_Texas_South_FIPS_4205\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",26.16666666666667],PARAMETER[\"Standard_Parallel_2\",27.83333333333333],PARAMETER[\"Latitude_Of_Origin\",25.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32041]]",
        32042=>"PROJCS[\"NAD_1927_StatePlane_Utah_North_FIPS_4301\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",40.71666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32042]]",
        32043=>"PROJCS[\"NAD_1927_StatePlane_Utah_Central_FIPS_4302\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",39.01666666666667],PARAMETER[\"Standard_Parallel_2\",40.65],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32043]]",
        32044=>"PROJCS[\"NAD_1927_StatePlane_Utah_South_FIPS_4303\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",37.21666666666667],PARAMETER[\"Standard_Parallel_2\",38.35],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32044]]",
        32045=>"PROJCS[\"NAD_1927_StatePlane_Vermont_FIPS_4400\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-72.5],PARAMETER[\"Scale_Factor\",0.9999642857142857],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32045]]",
        32046=>"PROJCS[\"NAD_1927_StatePlane_Virginia_North_FIPS_4501\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",38.03333333333333],PARAMETER[\"Standard_Parallel_2\",39.2],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32046]]",
        32047=>"PROJCS[\"NAD_1927_StatePlane_Virginia_South_FIPS_4502\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",36.76666666666667],PARAMETER[\"Standard_Parallel_2\",37.96666666666667],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32047]]",
        32048=>"PROJCS[\"NAD_1927_StatePlane_Washington_North_FIPS_4601\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.8333333333333],PARAMETER[\"Standard_Parallel_1\",47.5],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32048]]",
        32049=>"PROJCS[\"NAD_1927_StatePlane_Washington_South_FIPS_4602\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",45.83333333333334],PARAMETER[\"Standard_Parallel_2\",47.33333333333334],PARAMETER[\"Latitude_Of_Origin\",45.33333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32049]]",
        32050=>"PROJCS[\"NAD_1927_StatePlane_West_Virginia_North_FIPS_4701\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Standard_Parallel_1\",39.0],PARAMETER[\"Standard_Parallel_2\",40.25],PARAMETER[\"Latitude_Of_Origin\",38.5],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32050]]",
        32051=>"PROJCS[\"NAD_1927_StatePlane_West_Virginia_South_FIPS_4702\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",37.48333333333333],PARAMETER[\"Standard_Parallel_2\",38.88333333333333],PARAMETER[\"Latitude_Of_Origin\",37.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32051]]",
        32052=>"PROJCS[\"NAD_1927_StatePlane_Wisconsin_North_FIPS_4801\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",45.56666666666667],PARAMETER[\"Standard_Parallel_2\",46.76666666666667],PARAMETER[\"Latitude_Of_Origin\",45.16666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32052]]",
        32053=>"PROJCS[\"NAD_1927_StatePlane_Wisconsin_Central_FIPS_4802\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",44.25],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32053]]",
        32054=>"PROJCS[\"NAD_1927_StatePlane_Wisconsin_South_FIPS_4803\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",42.73333333333333],PARAMETER[\"Standard_Parallel_2\",44.06666666666667],PARAMETER[\"Latitude_Of_Origin\",42.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32054]]",
        32055=>"PROJCS[\"NAD_1927_StatePlane_Wyoming_East_FIPS_4901\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.1666666666667],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",40.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32055]]",
        32056=>"PROJCS[\"NAD_1927_StatePlane_Wyoming_East_Central_FIPS_4902\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-107.3333333333333],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",40.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32056]]",
        32057=>"PROJCS[\"NAD_1927_StatePlane_Wyoming_West_Central_FIPS_4903\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-108.75],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",40.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32057]]",
        32058=>"PROJCS[\"NAD_1927_StatePlane_Wyoming_West_FIPS_4904\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-110.0833333333333],PARAMETER[\"Scale_Factor\",0.9999411764705882],PARAMETER[\"Latitude_Of_Origin\",40.66666666666666],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32058]]",
        32061=>"PROJCS[\"NAD_1927_Guatemala_Norte\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",292209.579],PARAMETER[\"Central_Meridian\",-90.33333333333333],PARAMETER[\"Standard_Parallel_1\",16.81666666666667],PARAMETER[\"Scale_Factor\",0.99992226],PARAMETER[\"Latitude_Of_Origin\",16.81666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32061]]",
        32062=>"PROJCS[\"NAD_1927_Guatemala_Sur\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",325992.681],PARAMETER[\"Central_Meridian\",-90.33333333333333],PARAMETER[\"Standard_Parallel_1\",14.9],PARAMETER[\"Scale_Factor\",0.99989906],PARAMETER[\"Latitude_Of_Origin\",14.9],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32062]]",
        32064=>"PROJCS[\"NAD_1927_BLM_Zone_14N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32064]]",
        32065=>"PROJCS[\"NAD_1927_BLM_Zone_15N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32065]]",
        32066=>"PROJCS[\"NAD_1927_BLM_Zone_16N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32066]]",
        32067=>"PROJCS[\"NAD_1927_BLM_Zone_17N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32067]]",
        32074=>"PROJCS[\"NAD_1927_BLM_Zone_14N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32064]]",
        32075=>"PROJCS[\"NAD_1927_BLM_Zone_15N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32065]]",
        32076=>"PROJCS[\"NAD_1927_BLM_Zone_16N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32066]]",
        32077=>"PROJCS[\"NAD_1927_BLM_Zone_17N\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32067]]",
        32081=>"PROJCS[\"NAD_1927_MTM_1\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-53.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32081]]",
        32082=>"PROJCS[\"NAD_1927_MTM_2\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-56.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32082]]",
        32083=>"PROJCS[\"NAD_1927_MTM_3\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-58.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32083]]",
        32084=>"PROJCS[\"NAD_1927_MTM_4\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-61.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32084]]",
        32085=>"PROJCS[\"NAD_1927_MTM_5\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-64.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32085]]",
        32086=>"PROJCS[\"NAD_1927_MTM_6\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-67.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32086]]",
        32098=>"PROJCS[\"NAD_1927_Quebec_Lambert\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-68.5],PARAMETER[\"Standard_Parallel_1\",46.0],PARAMETER[\"Standard_Parallel_2\",60.0],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32098]]",
        32099=>"PROJCS[\"NAD_1927_StatePlane_Louisiana_Offshore_FIPS_1703\",GEOGCS[\"GCS_North_American_1927\",DATUM[\"D_North_American_1927\",SPHEROID[\"Clarke_1866\",6378206.4,294.9786982]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-91.33333333333333],PARAMETER[\"Standard_Parallel_1\",26.16666666666667],PARAMETER[\"Standard_Parallel_2\",27.83333333333333],PARAMETER[\"Latitude_Of_Origin\",25.66666666666667],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32099]]",
        32100=>"PROJCS[\"NAD_1983_StatePlane_Montana_FIPS_2500\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-109.5],PARAMETER[\"Standard_Parallel_1\",45.0],PARAMETER[\"Standard_Parallel_2\",49.0],PARAMETER[\"Latitude_Of_Origin\",44.25],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32100]]",
        32104=>"PROJCS[\"NAD_1983_StatePlane_Nebraska_FIPS_2600\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",40.0],PARAMETER[\"Standard_Parallel_2\",43.0],PARAMETER[\"Latitude_Of_Origin\",39.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32104]]",
        32107=>"PROJCS[\"NAD_1983_StatePlane_Nevada_East_FIPS_2701\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",8000000.0],PARAMETER[\"Central_Meridian\",-115.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32107]]",
        32108=>"PROJCS[\"NAD_1983_StatePlane_Nevada_Central_FIPS_2702\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",6000000.0],PARAMETER[\"Central_Meridian\",-116.6666666666667],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32108]]",
        32109=>"PROJCS[\"NAD_1983_StatePlane_Nevada_West_FIPS_2703\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",4000000.0],PARAMETER[\"Central_Meridian\",-118.5833333333333],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",34.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32109]]",
        32110=>"PROJCS[\"NAD_1983_StatePlane_New_Hampshire_FIPS_2800\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.66666666666667],PARAMETER[\"Scale_Factor\",0.9999666666666667],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32110]]",
        32111=>"PROJCS[\"NAD_1983_StatePlane_New_Jersey_FIPS_2900\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",38.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32111]]",
        32112=>"PROJCS[\"NAD_1983_StatePlane_New_Mexico_East_FIPS_3001\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",165000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-104.3333333333333],PARAMETER[\"Scale_Factor\",0.9999090909090909],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32112]]",
        32113=>"PROJCS[\"NAD_1983_StatePlane_New_Mexico_Central_FIPS_3002\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-106.25],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32113]]",
        32114=>"PROJCS[\"NAD_1983_StatePlane_New_Mexico_West_FIPS_3003\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",830000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-107.8333333333333],PARAMETER[\"Scale_Factor\",0.9999166666666667],PARAMETER[\"Latitude_Of_Origin\",31.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32114]]",
        32115=>"PROJCS[\"NAD_1983_StatePlane_New_York_East_FIPS_3101\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",150000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",38.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32115]]",
        32116=>"PROJCS[\"NAD_1983_StatePlane_New_York_Central_FIPS_3102\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",250000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-76.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32116]]",
        32117=>"PROJCS[\"NAD_1983_StatePlane_New_York_West_FIPS_3103\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",350000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-78.58333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32117]]",
        32118=>"PROJCS[\"NAD_1983_StatePlane_New_York_Long_Island_FIPS_3104\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-74.0],PARAMETER[\"Standard_Parallel_1\",40.66666666666666],PARAMETER[\"Standard_Parallel_2\",41.03333333333333],PARAMETER[\"Latitude_Of_Origin\",40.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32118]]",
        32119=>"PROJCS[\"NAD_1983_StatePlane_North_Carolina_FIPS_3200\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",609601.22],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.0],PARAMETER[\"Standard_Parallel_1\",34.33333333333334],PARAMETER[\"Standard_Parallel_2\",36.16666666666666],PARAMETER[\"Latitude_Of_Origin\",33.75],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32119]]",
        32120=>"PROJCS[\"NAD_1983_StatePlane_North_Dakota_North_FIPS_3301\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",47.43333333333333],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32120]]",
        32121=>"PROJCS[\"NAD_1983_StatePlane_North_Dakota_South_FIPS_3302\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.5],PARAMETER[\"Standard_Parallel_1\",46.18333333333333],PARAMETER[\"Standard_Parallel_2\",47.48333333333333],PARAMETER[\"Latitude_Of_Origin\",45.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32121]]",
        32122=>"PROJCS[\"NAD_1983_StatePlane_Ohio_North_FIPS_3401\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",40.43333333333333],PARAMETER[\"Standard_Parallel_2\",41.7],PARAMETER[\"Latitude_Of_Origin\",39.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32122]]",
        32123=>"PROJCS[\"NAD_1983_StatePlane_Ohio_South_FIPS_3402\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Standard_Parallel_1\",38.73333333333333],PARAMETER[\"Standard_Parallel_2\",40.03333333333333],PARAMETER[\"Latitude_Of_Origin\",38.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32123]]",
        32124=>"PROJCS[\"NAD_1983_StatePlane_Oklahoma_North_FIPS_3501\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",35.56666666666667],PARAMETER[\"Standard_Parallel_2\",36.76666666666667],PARAMETER[\"Latitude_Of_Origin\",35.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32124]]",
        32125=>"PROJCS[\"NAD_1983_StatePlane_Oklahoma_South_FIPS_3502\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-98.0],PARAMETER[\"Standard_Parallel_1\",33.93333333333333],PARAMETER[\"Standard_Parallel_2\",35.23333333333333],PARAMETER[\"Latitude_Of_Origin\",33.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32125]]",
        32126=>"PROJCS[\"NAD_1983_StatePlane_Oregon_North_FIPS_3601\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",2500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",44.33333333333334],PARAMETER[\"Standard_Parallel_2\",46.0],PARAMETER[\"Latitude_Of_Origin\",43.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32126]]",
        32127=>"PROJCS[\"NAD_1983_StatePlane_Oregon_South_FIPS_3602\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",42.33333333333334],PARAMETER[\"Standard_Parallel_2\",44.0],PARAMETER[\"Latitude_Of_Origin\",41.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32127]]",
        32128=>"PROJCS[\"NAD_1983_StatePlane_Pennsylvania_North_FIPS_3701\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",40.88333333333333],PARAMETER[\"Standard_Parallel_2\",41.95],PARAMETER[\"Latitude_Of_Origin\",40.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32128]]",
        32129=>"PROJCS[\"NAD_1983_StatePlane_Pennsylvania_South_FIPS_3702\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-77.75],PARAMETER[\"Standard_Parallel_1\",39.93333333333333],PARAMETER[\"Standard_Parallel_2\",40.96666666666667],PARAMETER[\"Latitude_Of_Origin\",39.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32129]]",
        32130=>"PROJCS[\"NAD_1983_StatePlane_Rhode_Island_FIPS_3800\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",100000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-71.5],PARAMETER[\"Scale_Factor\",0.99999375],PARAMETER[\"Latitude_Of_Origin\",41.08333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32130]]",
        32133=>"PROJCS[\"NAD_1983_StatePlane_South_Carolina_FIPS_3900\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",609600.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",32.5],PARAMETER[\"Standard_Parallel_2\",34.83333333333334],PARAMETER[\"Latitude_Of_Origin\",31.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32133]]",
        32134=>"PROJCS[\"NAD_1983_StatePlane_South_Dakota_North_FIPS_4001\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.0],PARAMETER[\"Standard_Parallel_1\",44.41666666666666],PARAMETER[\"Standard_Parallel_2\",45.68333333333333],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32134]]",
        32135=>"PROJCS[\"NAD_1983_StatePlane_South_Dakota_South_FIPS_4002\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",42.83333333333334],PARAMETER[\"Standard_Parallel_2\",44.4],PARAMETER[\"Latitude_Of_Origin\",42.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32135]]",
        32136=>"PROJCS[\"NAD_1983_StatePlane_Tennessee_FIPS_4100\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-86.0],PARAMETER[\"Standard_Parallel_1\",35.25],PARAMETER[\"Standard_Parallel_2\",36.41666666666666],PARAMETER[\"Latitude_Of_Origin\",34.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32136]]",
        32137=>"PROJCS[\"NAD_1983_StatePlane_Texas_North_FIPS_4201\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-101.5],PARAMETER[\"Standard_Parallel_1\",34.65],PARAMETER[\"Standard_Parallel_2\",36.18333333333333],PARAMETER[\"Latitude_Of_Origin\",34.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32137]]",
        32138=>"PROJCS[\"NAD_1983_StatePlane_Texas_North_Central_FIPS_4202\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",32.13333333333333],PARAMETER[\"Standard_Parallel_2\",33.96666666666667],PARAMETER[\"Latitude_Of_Origin\",31.66666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32138]]",
        32139=>"PROJCS[\"NAD_1983_StatePlane_Texas_Central_FIPS_4203\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",700000.0],PARAMETER[\"False_Northing\",3000000.0],PARAMETER[\"Central_Meridian\",-100.3333333333333],PARAMETER[\"Standard_Parallel_1\",30.11666666666667],PARAMETER[\"Standard_Parallel_2\",31.88333333333333],PARAMETER[\"Latitude_Of_Origin\",29.66666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32139]]",
        32140=>"PROJCS[\"NAD_1983_StatePlane_Texas_South_Central_FIPS_4204\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",4000000.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Standard_Parallel_1\",28.38333333333333],PARAMETER[\"Standard_Parallel_2\",30.28333333333333],PARAMETER[\"Latitude_Of_Origin\",27.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32140]]",
        32141=>"PROJCS[\"NAD_1983_StatePlane_Texas_South_FIPS_4205\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",300000.0],PARAMETER[\"False_Northing\",5000000.0],PARAMETER[\"Central_Meridian\",-98.5],PARAMETER[\"Standard_Parallel_1\",26.16666666666667],PARAMETER[\"Standard_Parallel_2\",27.83333333333333],PARAMETER[\"Latitude_Of_Origin\",25.66666666666667],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32141]]",
        32142=>"PROJCS[\"NAD_1983_StatePlane_Utah_North_FIPS_4301\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",40.71666666666667],PARAMETER[\"Standard_Parallel_2\",41.78333333333333],PARAMETER[\"Latitude_Of_Origin\",40.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32142]]",
        32143=>"PROJCS[\"NAD_1983_StatePlane_Utah_Central_FIPS_4302\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",39.01666666666667],PARAMETER[\"Standard_Parallel_2\",40.65],PARAMETER[\"Latitude_Of_Origin\",38.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32143]]",
        32144=>"PROJCS[\"NAD_1983_StatePlane_Utah_South_FIPS_4303\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",3000000.0],PARAMETER[\"Central_Meridian\",-111.5],PARAMETER[\"Standard_Parallel_1\",37.21666666666667],PARAMETER[\"Standard_Parallel_2\",38.35],PARAMETER[\"Latitude_Of_Origin\",36.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32144]]",
        32145=>"PROJCS[\"NAD_1983_StatePlane_Vermont_FIPS_4400\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-72.5],PARAMETER[\"Scale_Factor\",0.9999642857142857],PARAMETER[\"Latitude_Of_Origin\",42.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32145]]",
        32146=>"PROJCS[\"NAD_1983_StatePlane_Virginia_North_FIPS_4501\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",38.03333333333333],PARAMETER[\"Standard_Parallel_2\",39.2],PARAMETER[\"Latitude_Of_Origin\",37.66666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32146]]",
        32147=>"PROJCS[\"NAD_1983_StatePlane_Virginia_South_FIPS_4502\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",3500000.0],PARAMETER[\"False_Northing\",1000000.0],PARAMETER[\"Central_Meridian\",-78.5],PARAMETER[\"Standard_Parallel_1\",36.76666666666667],PARAMETER[\"Standard_Parallel_2\",37.96666666666667],PARAMETER[\"Latitude_Of_Origin\",36.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32147]]",
        32148=>"PROJCS[\"NAD_1983_StatePlane_Washington_North_FIPS_4601\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.8333333333333],PARAMETER[\"Standard_Parallel_1\",47.5],PARAMETER[\"Standard_Parallel_2\",48.73333333333333],PARAMETER[\"Latitude_Of_Origin\",47.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32148]]",
        32149=>"PROJCS[\"NAD_1983_StatePlane_Washington_South_FIPS_4602\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-120.5],PARAMETER[\"Standard_Parallel_1\",45.83333333333334],PARAMETER[\"Standard_Parallel_2\",47.33333333333334],PARAMETER[\"Latitude_Of_Origin\",45.33333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32149]]",
        32150=>"PROJCS[\"NAD_1983_StatePlane_West_Virginia_North_FIPS_4701\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Standard_Parallel_1\",39.0],PARAMETER[\"Standard_Parallel_2\",40.25],PARAMETER[\"Latitude_Of_Origin\",38.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32150]]",
        32151=>"PROJCS[\"NAD_1983_StatePlane_West_Virginia_South_FIPS_4702\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Standard_Parallel_1\",37.48333333333333],PARAMETER[\"Standard_Parallel_2\",38.88333333333333],PARAMETER[\"Latitude_Of_Origin\",37.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32151]]",
        32152=>"PROJCS[\"NAD_1983_StatePlane_Wisconsin_North_FIPS_4801\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",45.56666666666667],PARAMETER[\"Standard_Parallel_2\",46.76666666666667],PARAMETER[\"Latitude_Of_Origin\",45.16666666666666],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32152]]",
        32153=>"PROJCS[\"NAD_1983_StatePlane_Wisconsin_Central_FIPS_4802\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",44.25],PARAMETER[\"Standard_Parallel_2\",45.5],PARAMETER[\"Latitude_Of_Origin\",43.83333333333334],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32153]]",
        32154=>"PROJCS[\"NAD_1983_StatePlane_Wisconsin_South_FIPS_4803\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Standard_Parallel_1\",42.73333333333333],PARAMETER[\"Standard_Parallel_2\",44.06666666666667],PARAMETER[\"Latitude_Of_Origin\",42.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32154]]",
        32155=>"PROJCS[\"NAD_1983_StatePlane_Wyoming_East_FIPS_4901\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.1666666666667],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32155]]",
        32156=>"PROJCS[\"NAD_1983_StatePlane_Wyoming_East_Central_FIPS_4902\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",400000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-107.3333333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32156]]",
        32157=>"PROJCS[\"NAD_1983_StatePlane_Wyoming_West_Central_FIPS_4903\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",600000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-108.75],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32157]]",
        32158=>"PROJCS[\"NAD_1983_StatePlane_Wyoming_West_FIPS_4904\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",800000.0],PARAMETER[\"False_Northing\",100000.0],PARAMETER[\"Central_Meridian\",-110.0833333333333],PARAMETER[\"Scale_Factor\",0.9999375],PARAMETER[\"Latitude_Of_Origin\",40.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32158]]",
        32161=>"PROJCS[\"NAD_1983_StatePlane_Puerto_Rico_Virgin_Islands_FIPS_5200\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",200000.0],PARAMETER[\"False_Northing\",200000.0],PARAMETER[\"Central_Meridian\",-66.43333333333334],PARAMETER[\"Standard_Parallel_1\",18.03333333333333],PARAMETER[\"Standard_Parallel_2\",18.43333333333333],PARAMETER[\"Latitude_Of_Origin\",17.83333333333333],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32161]]",
        32164=>"PROJCS[\"NAD_1983_BLM_Zone_14N_ftUS\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32164]]",
        32165=>"PROJCS[\"NAD_1983_BLM_Zone_15N_ftUS\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32165]]",
        32166=>"PROJCS[\"NAD_1983_BLM_Zone_16N_ftUS\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32166]]",
        32167=>"PROJCS[\"NAD_1983_BLM_Zone_17N_ftUS\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32167]]",
        32180=>"PROJCS[\"NAD_1983_MTM_2_SCoPQ\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-55.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32180]]",
        32181=>"PROJCS[\"NAD_1983_MTM_1\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-53.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32181]]",
        32182=>"PROJCS[\"NAD_1983_MTM_2\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-56.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32182]]",
        32183=>"PROJCS[\"NAD_1983_MTM_3\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-58.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32183]]",
        32184=>"PROJCS[\"NAD_1983_MTM_4\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-61.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32184]]",
        32185=>"PROJCS[\"NAD_1983_MTM_5\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-64.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32185]]",
        32186=>"PROJCS[\"NAD_1983_MTM_6\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-67.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32186]]",
        32187=>"PROJCS[\"NAD_1983_MTM_7\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-70.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32187]]",
        32188=>"PROJCS[\"NAD_1983_MTM_8\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-73.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32188]]",
        32189=>"PROJCS[\"NAD_1983_MTM_9\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-76.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32189]]",
        32190=>"PROJCS[\"NAD_1983_MTM_10\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-79.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32190]]",
        32191=>"PROJCS[\"NAD_1983_MTM_11\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-82.5],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32191]]",
        32192=>"PROJCS[\"NAD_1983_MTM_12\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32192]]",
        32193=>"PROJCS[\"NAD_1983_MTM_13\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-84.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32193]]",
        32194=>"PROJCS[\"NAD_1983_MTM_14\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32194]]",
        32195=>"PROJCS[\"NAD_1983_MTM_15\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-90.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32195]]",
        32196=>"PROJCS[\"NAD_1983_MTM_16\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32196]]",
        32197=>"PROJCS[\"NAD_1983_MTM_17\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",304800.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-96.0],PARAMETER[\"Scale_Factor\",0.9999],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32197]]",
        32198=>"PROJCS[\"NAD_1983_Quebec_Lambert\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-68.5],PARAMETER[\"Standard_Parallel_1\",46.0],PARAMETER[\"Standard_Parallel_2\",60.0],PARAMETER[\"Latitude_Of_Origin\",44.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32198]]",
        32199=>"PROJCS[\"NAD_1983_StatePlane_Louisiana_Offshore_FIPS_1703\",GEOGCS[\"GCS_North_American_1983\",DATUM[\"D_North_American_1983\",SPHEROID[\"GRS_1980\",6378137.0,298.257222101]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Lambert_Conformal_Conic\"],PARAMETER[\"False_Easting\",1000000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-91.33333333333333],PARAMETER[\"Standard_Parallel_1\",26.16666666666667],PARAMETER[\"Standard_Parallel_2\",27.83333333333333],PARAMETER[\"Latitude_Of_Origin\",25.5],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32199]]",
        32201=>"PROJCS[\"WGS_1972_UTM_Zone_1N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32201]]",
        32202=>"PROJCS[\"WGS_1972_UTM_Zone_2N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32202]]",
        32203=>"PROJCS[\"WGS_1972_UTM_Zone_3N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32203]]",
        32204=>"PROJCS[\"WGS_1972_UTM_Zone_4N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32204]]",
        32205=>"PROJCS[\"WGS_1972_UTM_Zone_5N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32205]]",
        32206=>"PROJCS[\"WGS_1972_UTM_Zone_6N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32206]]",
        32207=>"PROJCS[\"WGS_1972_UTM_Zone_7N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32207]]",
        32208=>"PROJCS[\"WGS_1972_UTM_Zone_8N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32208]]",
        32209=>"PROJCS[\"WGS_1972_UTM_Zone_9N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32209]]",
        32210=>"PROJCS[\"WGS_1972_UTM_Zone_10N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32210]]",
        32211=>"PROJCS[\"WGS_1972_UTM_Zone_11N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32211]]",
        32212=>"PROJCS[\"WGS_1972_UTM_Zone_12N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32212]]",
        32213=>"PROJCS[\"WGS_1972_UTM_Zone_13N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32213]]",
        32214=>"PROJCS[\"WGS_1972_UTM_Zone_14N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32214]]",
        32215=>"PROJCS[\"WGS_1972_UTM_Zone_15N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32215]]",
        32216=>"PROJCS[\"WGS_1972_UTM_Zone_16N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32216]]",
        32217=>"PROJCS[\"WGS_1972_UTM_Zone_17N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32217]]",
        32218=>"PROJCS[\"WGS_1972_UTM_Zone_18N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32218]]",
        32219=>"PROJCS[\"WGS_1972_UTM_Zone_19N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32219]]",
        32220=>"PROJCS[\"WGS_1972_UTM_Zone_20N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32220]]",
        32221=>"PROJCS[\"WGS_1972_UTM_Zone_21N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32221]]",
        32222=>"PROJCS[\"WGS_1972_UTM_Zone_22N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32222]]",
        32223=>"PROJCS[\"WGS_1972_UTM_Zone_23N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32223]]",
        32224=>"PROJCS[\"WGS_1972_UTM_Zone_24N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32224]]",
        32225=>"PROJCS[\"WGS_1972_UTM_Zone_25N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32225]]",
        32226=>"PROJCS[\"WGS_1972_UTM_Zone_26N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32226]]",
        32227=>"PROJCS[\"WGS_1972_UTM_Zone_27N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32227]]",
        32228=>"PROJCS[\"WGS_1972_UTM_Zone_28N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32228]]",
        32229=>"PROJCS[\"WGS_1972_UTM_Zone_29N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32229]]",
        32230=>"PROJCS[\"WGS_1972_UTM_Zone_30N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32230]]",
        32231=>"PROJCS[\"WGS_1972_UTM_Zone_31N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32231]]",
        32232=>"PROJCS[\"WGS_1972_UTM_Zone_32N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32232]]",
        32233=>"PROJCS[\"WGS_1972_UTM_Zone_33N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32233]]",
        32234=>"PROJCS[\"WGS_1972_UTM_Zone_34N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32234]]",
        32235=>"PROJCS[\"WGS_1972_UTM_Zone_35N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32235]]",
        32236=>"PROJCS[\"WGS_1972_UTM_Zone_36N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32236]]",
        32237=>"PROJCS[\"WGS_1972_UTM_Zone_37N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32237]]",
        32238=>"PROJCS[\"WGS_1972_UTM_Zone_38N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32238]]",
        32239=>"PROJCS[\"WGS_1972_UTM_Zone_39N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32239]]",
        32240=>"PROJCS[\"WGS_1972_UTM_Zone_40N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32240]]",
        32241=>"PROJCS[\"WGS_1972_UTM_Zone_41N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32241]]",
        32242=>"PROJCS[\"WGS_1972_UTM_Zone_42N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32242]]",
        32243=>"PROJCS[\"WGS_1972_UTM_Zone_43N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32243]]",
        32244=>"PROJCS[\"WGS_1972_UTM_Zone_44N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32244]]",
        32245=>"PROJCS[\"WGS_1972_UTM_Zone_45N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32245]]",
        32246=>"PROJCS[\"WGS_1972_UTM_Zone_46N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32246]]",
        32247=>"PROJCS[\"WGS_1972_UTM_Zone_47N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32247]]",
        32248=>"PROJCS[\"WGS_1972_UTM_Zone_48N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32248]]",
        32249=>"PROJCS[\"WGS_1972_UTM_Zone_49N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32249]]",
        32250=>"PROJCS[\"WGS_1972_UTM_Zone_50N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32250]]",
        32251=>"PROJCS[\"WGS_1972_UTM_Zone_51N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32251]]",
        32252=>"PROJCS[\"WGS_1972_UTM_Zone_52N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32252]]",
        32253=>"PROJCS[\"WGS_1972_UTM_Zone_53N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32253]]",
        32254=>"PROJCS[\"WGS_1972_UTM_Zone_54N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32254]]",
        32255=>"PROJCS[\"WGS_1972_UTM_Zone_55N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32255]]",
        32256=>"PROJCS[\"WGS_1972_UTM_Zone_56N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32256]]",
        32257=>"PROJCS[\"WGS_1972_UTM_Zone_57N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32257]]",
        32258=>"PROJCS[\"WGS_1972_UTM_Zone_58N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32258]]",
        32259=>"PROJCS[\"WGS_1972_UTM_Zone_59N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32259]]",
        32260=>"PROJCS[\"WGS_1972_UTM_Zone_60N\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32260]]",
        32301=>"PROJCS[\"WGS_1972_UTM_Zone_1S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32301]]",
        32302=>"PROJCS[\"WGS_1972_UTM_Zone_2S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32302]]",
        32303=>"PROJCS[\"WGS_1972_UTM_Zone_3S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32303]]",
        32304=>"PROJCS[\"WGS_1972_UTM_Zone_4S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32304]]",
        32305=>"PROJCS[\"WGS_1972_UTM_Zone_5S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32305]]",
        32306=>"PROJCS[\"WGS_1972_UTM_Zone_6S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32306]]",
        32307=>"PROJCS[\"WGS_1972_UTM_Zone_7S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32307]]",
        32308=>"PROJCS[\"WGS_1972_UTM_Zone_8S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32308]]",
        32309=>"PROJCS[\"WGS_1972_UTM_Zone_9S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32309]]",
        32310=>"PROJCS[\"WGS_1972_UTM_Zone_10S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32310]]",
        32311=>"PROJCS[\"WGS_1972_UTM_Zone_11S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32311]]",
        32312=>"PROJCS[\"WGS_1972_UTM_Zone_12S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32312]]",
        32313=>"PROJCS[\"WGS_1972_UTM_Zone_13S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32313]]",
        32314=>"PROJCS[\"WGS_1972_UTM_Zone_14S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32314]]",
        32315=>"PROJCS[\"WGS_1972_UTM_Zone_15S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32315]]",
        32316=>"PROJCS[\"WGS_1972_UTM_Zone_16S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32316]]",
        32317=>"PROJCS[\"WGS_1972_UTM_Zone_17S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32317]]",
        32318=>"PROJCS[\"WGS_1972_UTM_Zone_18S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32318]]",
        32319=>"PROJCS[\"WGS_1972_UTM_Zone_19S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32319]]",
        32320=>"PROJCS[\"WGS_1972_UTM_Zone_20S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32320]]",
        32321=>"PROJCS[\"WGS_1972_UTM_Zone_21S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32321]]",
        32322=>"PROJCS[\"WGS_1972_UTM_Zone_22S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32322]]",
        32323=>"PROJCS[\"WGS_1972_UTM_Zone_23S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32323]]",
        32324=>"PROJCS[\"WGS_1972_UTM_Zone_24S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32324]]",
        32325=>"PROJCS[\"WGS_1972_UTM_Zone_25S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32325]]",
        32326=>"PROJCS[\"WGS_1972_UTM_Zone_26S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32326]]",
        32327=>"PROJCS[\"WGS_1972_UTM_Zone_27S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32327]]",
        32328=>"PROJCS[\"WGS_1972_UTM_Zone_28S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32328]]",
        32329=>"PROJCS[\"WGS_1972_UTM_Zone_29S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32329]]",
        32330=>"PROJCS[\"WGS_1972_UTM_Zone_30S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32330]]",
        32331=>"PROJCS[\"WGS_1972_UTM_Zone_31S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32331]]",
        32332=>"PROJCS[\"WGS_1972_UTM_Zone_32S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32332]]",
        32333=>"PROJCS[\"WGS_1972_UTM_Zone_33S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32333]]",
        32334=>"PROJCS[\"WGS_1972_UTM_Zone_34S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32334]]",
        32335=>"PROJCS[\"WGS_1972_UTM_Zone_35S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32335]]",
        32336=>"PROJCS[\"WGS_1972_UTM_Zone_36S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32336]]",
        32337=>"PROJCS[\"WGS_1972_UTM_Zone_37S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32337]]",
        32338=>"PROJCS[\"WGS_1972_UTM_Zone_38S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32338]]",
        32339=>"PROJCS[\"WGS_1972_UTM_Zone_39S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32339]]",
        32340=>"PROJCS[\"WGS_1972_UTM_Zone_40S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32340]]",
        32341=>"PROJCS[\"WGS_1972_UTM_Zone_41S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32341]]",
        32342=>"PROJCS[\"WGS_1972_UTM_Zone_42S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32342]]",
        32343=>"PROJCS[\"WGS_1972_UTM_Zone_43S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32343]]",
        32344=>"PROJCS[\"WGS_1972_UTM_Zone_44S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32344]]",
        32345=>"PROJCS[\"WGS_1972_UTM_Zone_45S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32345]]",
        32346=>"PROJCS[\"WGS_1972_UTM_Zone_46S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32346]]",
        32347=>"PROJCS[\"WGS_1972_UTM_Zone_47S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32347]]",
        32348=>"PROJCS[\"WGS_1972_UTM_Zone_48S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32348]]",
        32349=>"PROJCS[\"WGS_1972_UTM_Zone_49S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32349]]",
        32350=>"PROJCS[\"WGS_1972_UTM_Zone_50S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32350]]",
        32351=>"PROJCS[\"WGS_1972_UTM_Zone_51S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32351]]",
        32352=>"PROJCS[\"WGS_1972_UTM_Zone_52S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32352]]",
        32353=>"PROJCS[\"WGS_1972_UTM_Zone_53S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32353]]",
        32354=>"PROJCS[\"WGS_1972_UTM_Zone_54S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32354]]",
        32355=>"PROJCS[\"WGS_1972_UTM_Zone_55S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32355]]",
        32356=>"PROJCS[\"WGS_1972_UTM_Zone_56S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32356]]",
        32357=>"PROJCS[\"WGS_1972_UTM_Zone_57S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32357]]",
        32358=>"PROJCS[\"WGS_1972_UTM_Zone_58S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32358]]",
        32359=>"PROJCS[\"WGS_1972_UTM_Zone_59S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32359]]",
        32360=>"PROJCS[\"WGS_1972_UTM_Zone_60S\",GEOGCS[\"GCS_WGS_1972\",DATUM[\"D_WGS_1972\",SPHEROID[\"WGS_1972\",6378135.0,298.26]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32360]]",
        32601=>"PROJCS[\"WGS_1984_UTM_Zone_1N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32601]]",
        32602=>"PROJCS[\"WGS_1984_UTM_Zone_2N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32602]]",
        32603=>"PROJCS[\"WGS_1984_UTM_Zone_3N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32603]]",
        32604=>"PROJCS[\"WGS_1984_UTM_Zone_4N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32604]]",
        32605=>"PROJCS[\"WGS_1984_UTM_Zone_5N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32605]]",
        32606=>"PROJCS[\"WGS_1984_UTM_Zone_6N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32606]]",
        32607=>"PROJCS[\"WGS_1984_UTM_Zone_7N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32607]]",
        32608=>"PROJCS[\"WGS_1984_UTM_Zone_8N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32608]]",
        32609=>"PROJCS[\"WGS_1984_UTM_Zone_9N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32609]]",
        32610=>"PROJCS[\"WGS_1984_UTM_Zone_10N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32610]]",
        32611=>"PROJCS[\"WGS_1984_UTM_Zone_11N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32611]]",
        32612=>"PROJCS[\"WGS_1984_UTM_Zone_12N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32612]]",
        32613=>"PROJCS[\"WGS_1984_UTM_Zone_13N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32613]]",
        32614=>"PROJCS[\"WGS_1984_UTM_Zone_14N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32614]]",
        32615=>"PROJCS[\"WGS_1984_UTM_Zone_15N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32615]]",
        32616=>"PROJCS[\"WGS_1984_UTM_Zone_16N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32616]]",
        32617=>"PROJCS[\"WGS_1984_UTM_Zone_17N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32617]]",
        32618=>"PROJCS[\"WGS_1984_UTM_Zone_18N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32618]]",
        32619=>"PROJCS[\"WGS_1984_UTM_Zone_19N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32619]]",
        32620=>"PROJCS[\"WGS_1984_UTM_Zone_20N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32620]]",
        32621=>"PROJCS[\"WGS_1984_UTM_Zone_21N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32621]]",
        32622=>"PROJCS[\"WGS_1984_UTM_Zone_22N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32622]]",
        32623=>"PROJCS[\"WGS_1984_UTM_Zone_23N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32623]]",
        32624=>"PROJCS[\"WGS_1984_UTM_Zone_24N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32624]]",
        32625=>"PROJCS[\"WGS_1984_UTM_Zone_25N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32625]]",
        32626=>"PROJCS[\"WGS_1984_UTM_Zone_26N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32626]]",
        32627=>"PROJCS[\"WGS_1984_UTM_Zone_27N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32627]]",
        32628=>"PROJCS[\"WGS_1984_UTM_Zone_28N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32628]]",
        32629=>"PROJCS[\"WGS_1984_UTM_Zone_29N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32629]]",
        32630=>"PROJCS[\"WGS_1984_UTM_Zone_30N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32630]]",
        32631=>"PROJCS[\"WGS_1984_UTM_Zone_31N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32631]]",
        32632=>"PROJCS[\"WGS_1984_UTM_Zone_32N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32632]]",
        32633=>"PROJCS[\"WGS_1984_UTM_Zone_33N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32633]]",
        32634=>"PROJCS[\"WGS_1984_UTM_Zone_34N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32634]]",
        32635=>"PROJCS[\"WGS_1984_UTM_Zone_35N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32635]]",
        32636=>"PROJCS[\"WGS_1984_UTM_Zone_36N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32636]]",
        32637=>"PROJCS[\"WGS_1984_UTM_Zone_37N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32637]]",
        32638=>"PROJCS[\"WGS_1984_UTM_Zone_38N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32638]]",
        32639=>"PROJCS[\"WGS_1984_UTM_Zone_39N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32639]]",
        32640=>"PROJCS[\"WGS_1984_UTM_Zone_40N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32640]]",
        32641=>"PROJCS[\"WGS_1984_UTM_Zone_41N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32641]]",
        32642=>"PROJCS[\"WGS_1984_UTM_Zone_42N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32642]]",
        32643=>"PROJCS[\"WGS_1984_UTM_Zone_43N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32643]]",
        32644=>"PROJCS[\"WGS_1984_UTM_Zone_44N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32644]]",
        32645=>"PROJCS[\"WGS_1984_UTM_Zone_45N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32645]]",
        32646=>"PROJCS[\"WGS_1984_UTM_Zone_46N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32646]]",
        32647=>"PROJCS[\"WGS_1984_UTM_Zone_47N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32647]]",
        32648=>"PROJCS[\"WGS_1984_UTM_Zone_48N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32648]]",
        32649=>"PROJCS[\"WGS_1984_UTM_Zone_49N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32649]]",
        32650=>"PROJCS[\"WGS_1984_UTM_Zone_50N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32650]]",
        32651=>"PROJCS[\"WGS_1984_UTM_Zone_51N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32651]]",
        32652=>"PROJCS[\"WGS_1984_UTM_Zone_52N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32652]]",
        32653=>"PROJCS[\"WGS_1984_UTM_Zone_53N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32653]]",
        32654=>"PROJCS[\"WGS_1984_UTM_Zone_54N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32654]]",
        32655=>"PROJCS[\"WGS_1984_UTM_Zone_55N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32655]]",
        32656=>"PROJCS[\"WGS_1984_UTM_Zone_56N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32656]]",
        32657=>"PROJCS[\"WGS_1984_UTM_Zone_57N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32657]]",
        32658=>"PROJCS[\"WGS_1984_UTM_Zone_58N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32658]]",
        32659=>"PROJCS[\"WGS_1984_UTM_Zone_59N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32659]]",
        32660=>"PROJCS[\"WGS_1984_UTM_Zone_60N\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32660]]",
        32661=>"PROJCS[\"UPS_North\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Scale_Factor\",0.994],PARAMETER[\"Latitude_Of_Origin\",90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32661]]",
        32662=>"PROJCS[\"WGS_1984_Plate_Carree\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Plate_Carree\"],PARAMETER[\"False_Easting\",0.0],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32662]]",
        32664=>"PROJCS[\"WGS_1984_BLM_Zone_14N_ftUS\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32664]]",
        32665=>"PROJCS[\"WGS_1984_BLM_Zone_15N_ftUS\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32665]]",
        32666=>"PROJCS[\"WGS_1984_BLM_Zone_16N_ftUS\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32666]]",
        32667=>"PROJCS[\"WGS_1984_BLM_Zone_17N_ftUS\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",1640416.666666667],PARAMETER[\"False_Northing\",0.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Foot_US\",0.3048006096012192],AUTHORITY[\"EPSG\",32667]]",
        32701=>"PROJCS[\"WGS_1984_UTM_Zone_1S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32701]]",
        32702=>"PROJCS[\"WGS_1984_UTM_Zone_2S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32702]]",
        32703=>"PROJCS[\"WGS_1984_UTM_Zone_3S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32703]]",
        32704=>"PROJCS[\"WGS_1984_UTM_Zone_4S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32704]]",
        32705=>"PROJCS[\"WGS_1984_UTM_Zone_5S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32705]]",
        32706=>"PROJCS[\"WGS_1984_UTM_Zone_6S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32706]]",
        32707=>"PROJCS[\"WGS_1984_UTM_Zone_7S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32707]]",
        32708=>"PROJCS[\"WGS_1984_UTM_Zone_8S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32708]]",
        32709=>"PROJCS[\"WGS_1984_UTM_Zone_9S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32709]]",
        32710=>"PROJCS[\"WGS_1984_UTM_Zone_10S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32710]]",
        32711=>"PROJCS[\"WGS_1984_UTM_Zone_11S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32711]]",
        32712=>"PROJCS[\"WGS_1984_UTM_Zone_12S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32712]]",
        32713=>"PROJCS[\"WGS_1984_UTM_Zone_13S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32713]]",
        32714=>"PROJCS[\"WGS_1984_UTM_Zone_14S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32714]]",
        32715=>"PROJCS[\"WGS_1984_UTM_Zone_15S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32715]]",
        32716=>"PROJCS[\"WGS_1984_UTM_Zone_16S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32716]]",
        32717=>"PROJCS[\"WGS_1984_UTM_Zone_17S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32717]]",
        32718=>"PROJCS[\"WGS_1984_UTM_Zone_18S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32718]]",
        32719=>"PROJCS[\"WGS_1984_UTM_Zone_19S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32719]]",
        32720=>"PROJCS[\"WGS_1984_UTM_Zone_20S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32720]]",
        32721=>"PROJCS[\"WGS_1984_UTM_Zone_21S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32721]]",
        32722=>"PROJCS[\"WGS_1984_UTM_Zone_22S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32722]]",
        32723=>"PROJCS[\"WGS_1984_UTM_Zone_23S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32723]]",
        32724=>"PROJCS[\"WGS_1984_UTM_Zone_24S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32724]]",
        32725=>"PROJCS[\"WGS_1984_UTM_Zone_25S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32725]]",
        32726=>"PROJCS[\"WGS_1984_UTM_Zone_26S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32726]]",
        32727=>"PROJCS[\"WGS_1984_UTM_Zone_27S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32727]]",
        32728=>"PROJCS[\"WGS_1984_UTM_Zone_28S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32728]]",
        32729=>"PROJCS[\"WGS_1984_UTM_Zone_29S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32729]]",
        32730=>"PROJCS[\"WGS_1984_UTM_Zone_30S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",-3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32730]]",
        32731=>"PROJCS[\"WGS_1984_UTM_Zone_31S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",3.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32731]]",
        32732=>"PROJCS[\"WGS_1984_UTM_Zone_32S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",9.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32732]]",
        32733=>"PROJCS[\"WGS_1984_UTM_Zone_33S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",15.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32733]]",
        32734=>"PROJCS[\"WGS_1984_UTM_Zone_34S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",21.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32734]]",
        32735=>"PROJCS[\"WGS_1984_UTM_Zone_35S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",27.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32735]]",
        32736=>"PROJCS[\"WGS_1984_UTM_Zone_36S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",33.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32736]]",
        32737=>"PROJCS[\"WGS_1984_UTM_Zone_37S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",39.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32737]]",
        32738=>"PROJCS[\"WGS_1984_UTM_Zone_38S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",45.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32738]]",
        32739=>"PROJCS[\"WGS_1984_UTM_Zone_39S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",51.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32739]]",
        32740=>"PROJCS[\"WGS_1984_UTM_Zone_40S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",57.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32740]]",
        32741=>"PROJCS[\"WGS_1984_UTM_Zone_41S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",63.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32741]]",
        32742=>"PROJCS[\"WGS_1984_UTM_Zone_42S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",69.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32742]]",
        32743=>"PROJCS[\"WGS_1984_UTM_Zone_43S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",75.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32743]]",
        32744=>"PROJCS[\"WGS_1984_UTM_Zone_44S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",81.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32744]]",
        32745=>"PROJCS[\"WGS_1984_UTM_Zone_45S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",87.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32745]]",
        32746=>"PROJCS[\"WGS_1984_UTM_Zone_46S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",93.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32746]]",
        32747=>"PROJCS[\"WGS_1984_UTM_Zone_47S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",99.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32747]]",
        32748=>"PROJCS[\"WGS_1984_UTM_Zone_48S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",105.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32748]]",
        32749=>"PROJCS[\"WGS_1984_UTM_Zone_49S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",111.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32749]]",
        32750=>"PROJCS[\"WGS_1984_UTM_Zone_50S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",117.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32750]]",
        32751=>"PROJCS[\"WGS_1984_UTM_Zone_51S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",123.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32751]]",
        32752=>"PROJCS[\"WGS_1984_UTM_Zone_52S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",129.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32752]]",
        32753=>"PROJCS[\"WGS_1984_UTM_Zone_53S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",135.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32753]]",
        32754=>"PROJCS[\"WGS_1984_UTM_Zone_54S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",141.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32754]]",
        32755=>"PROJCS[\"WGS_1984_UTM_Zone_55S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",147.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32755]]",
        32756=>"PROJCS[\"WGS_1984_UTM_Zone_56S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",153.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32756]]",
        32757=>"PROJCS[\"WGS_1984_UTM_Zone_57S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",159.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32757]]",
        32758=>"PROJCS[\"WGS_1984_UTM_Zone_58S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",165.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32758]]",
        32759=>"PROJCS[\"WGS_1984_UTM_Zone_59S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",171.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32759]]",
        32760=>"PROJCS[\"WGS_1984_UTM_Zone_60S\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",177.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32760]]",
        32761=>"PROJCS[\"UPS_South\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Stereographic\"],PARAMETER[\"False_Easting\",2000000.0],PARAMETER[\"False_Northing\",2000000.0],PARAMETER[\"Central_Meridian\",0.0],PARAMETER[\"Scale_Factor\",0.994],PARAMETER[\"Latitude_Of_Origin\",-90.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32761]]",
        32766=>"PROJCS[\"WGS_1984_TM_36_SE\",GEOGCS[\"GCS_WGS_1984\",DATUM[\"D_WGS_1984\",SPHEROID[\"WGS_1984\",6378137.0,298.257223563]],PRIMEM[\"Greenwich\",0.0],UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"False_Easting\",500000.0],PARAMETER[\"False_Northing\",10000000.0],PARAMETER[\"Central_Meridian\",36.0],PARAMETER[\"Scale_Factor\",0.9996],PARAMETER[\"Latitude_Of_Origin\",0.0],UNIT[\"Meter\",1.0],AUTHORITY[\"EPSG\",32766]]"
    ];
    let s = match hm.get(&code) {
        Some(key) => key.to_string(),
        None => String::from("Unknown EPSG Code"),
    };
    s
}
